#include <math.h>
#include "errhand.h"
#include "swdef.h"
#include "mystdlib.h"
#include "bhtree.h"

#define MAXTREE 100000
#define MAXCLS  4196
#define MAX_TEMPLATES 250

static BHtree *bht=NULL;
static float dens=0.1;
static float rprobe=1.4;
static float rmax=0.0;
static dot_template *T[MAX_TEMPLATES];
static int nt=0;
static int buried=0;

static dot_template *appropriateTemplate(float r);
static dot_template *dotsAtPosition(float *x,dot_template *t);
static void clipper(dot_template *d,float *x,float r,int b);
static void onlyBuried(dot_template *d);

int LRinit(Atom **A,int n,float r,float d,int bf)
{
  /* initialize data structure for LR dots */

  float ra;
  int i;

  int status=0;
  if (A&&n>0) {
    freeBHtree(bht);
    bht=GenerateBHtree(n,A,n>100000?n/1000000:1);
    if (!bht) status=1;
    else {
      dens=d;
      rprobe=r;
      rmax=0;
      buried=bf;
      for (i=0;i<n;i++) {
	ra=(A[i]->atp?A[i]->atp->rvdw:1.5);
	if (ra>rmax) rmax=ra;
      }
    }
  }
  return(status);
}

void LRclose(void)
{
  int i;
  /* free data strucures use for LR dots */

  for (i=0;i<nt;i++) {free_dot_template(T[i]);} nt=0;
  freeBHtree(bht);
  bht=NULL;
}

dot_template *LRdots(Atom *a)
{
  float r,rb,dist[MAXCLS];
  dot_template *d=NULL,*t=NULL;
  Atom *B[MAXCLS];
  int j,nn;

  if (a) {
    r=(a->atp?a->atp->rvdw:1.5);
    t=appropriateTemplate(r+rprobe);
    d=dotsAtPosition(a->x,t);
    if (d&&d->n>0) {
      nn=findBHcloseAtoms(bht,a->x,r+rmax+2*rprobe,B,dist,MAXCLS);
      if (nn>MAXCLS) nn=MAXCLS;
      if (nn>0) {
	for (j=0;j<nn&&d->n>0;j++) {
	  rb=(B[j]->atp?B[j]->atp->rvdw:1.5);
	  if (B[j]!=a && dist[j]<r+rb+2*rprobe) clipper(d,B[j]->x,rb+rprobe,B[j]->flg&buried);
	}
      }
      if (buried) onlyBuried(d);
      if (d->n>0) d->x=realloc(d->x,d->n*4*sizeof(float));
      else{
	free_dot_template(d);
	d=NULL;
      }
    }
  }
  return(d);
}

static dot_template *appropriateTemplate(float r)
{
  dot_template *t;

  /* find template close to radius r */

  t=find_best_template(r,nt,T);
  if (!t || fabs(t->R-r)>0.01) {
    if (nt<MAX_TEMPLATES) {
      T[nt++]=t=make_dot_template(r,dens);
      if (!t) nt--;
    }
  }
  return(t);
}

static dot_template *dotsAtPosition(float *x,dot_template *t)
{
  dot_template *d=NULL;
  int i,j;

  /* copy dots in template to atom at position x */

  if (x && t && t->n>0) {
    d=malloc(sizeof(dot_template));
    if (d) {
      d->R=t->R;
      d->x=malloc(t->n*4*sizeof(float));
      d->n=t->n;
      if (d->x) {
	for (i=0;i<d->n;i++) {
	  for (j=0;j<3;j++) d->x[4*i+j]=x[j]+t->x[3*i+j];
	  d->x[4*i+3]=0.0;
	}
      }else{
	free(d);
	d=NULL;
      }
    }
  }
  return(d);
}

static void clipper(dot_template *d,float *x,float r,int b)
{
  int i,no=0;
  float D,dx;

  D=r*r;

  /* remove dots covered by ball B(x,r) (flag 'buried'  only if b) */

  if (d&&d->n>0) {
    if (b) {
      for (no=0,i=0;i<d->n;i++) {
	dx=(d->x[4*i  ]-x[0])*(d->x[4*i  ]-x[0])+
	   (d->x[4*i+1]-x[1])*(d->x[4*i+1]-x[1])+
	   (d->x[4*i+2]-x[2])*(d->x[4*i+2]-x[2]);
	if (dx<=D) d->x[4*i+3]=1.0;  /* mark as buried */
      }
    }else{
      for (no=0,i=0;i<d->n;i++) {
	dx=(d->x[4*i  ]-x[0])*(d->x[4*i  ]-x[0])+
	   (d->x[4*i+1]-x[1])*(d->x[4*i+1]-x[1])+
	   (d->x[4*i+2]-x[2])*(d->x[4*i+2]-x[2]);
	if (dx>D) {                 /* copy to output */
	  if (no<i) {
	    d->x[4*no  ]=d->x[4*i  ];
	    d->x[4*no+1]=d->x[4*i+1];
	    d->x[4*no+2]=d->x[4*i+2];
	    d->x[4*no+3]=d->x[4*i+3];
	  }
	  no++;
	}
      }
      d->n=no;
    }
  }
}

static void onlyBuried(dot_template *d)
{
  int i,no=0;

  /* remove dots that are not marked 'buried' */
  /* reset 'buried' flag */

  if (d&&d->n>0) {
    for (i=0;i<d->n;i++) {
      if (d->x[4*i+3]>0.5) {
	d->x[4*no]=d->x[4*i];
	d->x[4*no+1]=d->x[4*i+1];
	d->x[4*no+2]=d->x[4*i+2];
	d->x[4*no+3]=0.0;
	no++;
      }
    }
    d->n=no;
  }
}
