#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include "agglom.h"

/* List of implemented methods: */
 
typedef struct {
  char *name;
  int squared;
} agglom_method_t;

static  agglom_method_t methods[] = {
  {"single",   0},
  {"complete", 0},
  {"average",  0},
  {"weighted", 0},
  {"median",   1},
  {"centroid", 1},
  {"ward",     1},
  {"flexible", 0}
};


static int histcmp(const void *a,const void *b);


/*                 Hierarchical clustering, 8 agglomerative methods                     */
/* (c.f. Steinhausen/Langer "Clusteranalyse", Walter de Gruyter, Berlin, New York, 1977 */


/*------------------------------------------------------------------------*/

int agglom(int N,float **D,char *method,float alpha,int cpy,agglom_clhist_t *hist)
{
  /*

  input:
     N : number of objects to cluster
     D : lower triangular distance matrix (array of pointers to rows)
     method: "single"|"complete"|"average"|"weighted"|"median"|"centroid"|"ward"|"flexible"
     alpha: alpha parameter for "flexible" method, must be in [0.1...1.0], good: 0.5-0.7
     cpy : (boolean) TRUE if D must be preserved, FALSE if it may be destroyed

  output:
     hist : array of N-1 structures with {d,i,j} (dist at merge, objidx i, objidx j). 

  returns: 
     0: O.K., result in hist
     1: malloc of temporary data failed.
     2: requested unknown method.

  Notes:
     if alpha is not in [0.1...1.0], it will be forced into this range.
     "median" and "centroid" should be used only for squared euclidian D.
     "median", "centroid" may lead to non-monotonous d in hist.

  */

  int r=0;
  int i,j,k,l,nij,ia=0,ja=0,la=0,*idx=NULL,*im=NULL,*n=NULL,methodnr,nmethods;
  float **d=NULL,*dm=NULL,da;

  /* check if method is valid */

  nmethods=sizeof(methods)/sizeof(agglom_method_t);
  for (methodnr=0;methodnr<nmethods;methodnr++) if (!strcmp(method,methods[methodnr].name)) break;
  if (methodnr<nmethods) {

    /* force the alpha parameter (used only for "flexible" method) to reasonable range */
    if (!strcmp(method,"flexible")) {
      if (alpha<0.1) alpha=0.1;
      if (alpha>1.0) alpha=1.0;
    }

    if (cpy) {
      /* make a copy of the D matrix, if we are not allowed to scribble over it */
      d=malloc(N*sizeof(float *));
      if (d) {
	for (i=0;i<N;i++) {
	  d[i]=malloc((i+1)*sizeof(float));
	  if (!d[i]) break;
	  for (j=0;j<=i;j++) d[i][j]=methods[methodnr].squared?D[i][j]*D[i][j]:D[i][j];
	}
	if (i<N) {
	  for (j=0;j<i;j++) free(d[j]);
	  free(d);
	  d=NULL;
	}
      }
      idx=malloc(N*sizeof(int));
      im=malloc(N*sizeof(int));
      n=malloc(N*sizeof(int));
      dm=malloc(N*sizeof(float));
    }else{
      /* else use D for updates of distance matrix */
      d=D;
      idx=malloc(N*sizeof(int));
      im=malloc(N*sizeof(int));
      n=malloc(N*sizeof(int));
      dm=malloc(N*sizeof(float));
    }

    if (d&&idx&&n&&im&&dm) {
      
      /* put each object into a singleton */

      for (i=0;i<N;i++) {
	n[i]=1;            /* cluster size */
	idx[i]=i;          /* i-th cluster at position idx[i] */
      }

      /* initialize im and dm */

      for (i=1;i<N;i++) {
	dm[i]=d[i][0]; im[i]=0;
	for (j=0;j<i;j++) {
	  if (d[i][j]<dm[i]) {
	    dm[i]=d[i][j];
	    im[i]=j;
	  }
	}
      }

      /* Main loop: iteratively merge the two closest clusters */

      for (k=0;k<N-1;k++) {
	/* find pair of closest clusters */
	da=dm[1]; ia=1;
	for (i=2;i<N-k;i++) {
	  if (dm[i]<da) {
	    ia=i;
	    da=dm[i];
	  }
	}
	ja=im[ia];

	/* merge the two into new cluster */
	i=hist[k].i=idx[ia];      /* cluster ia ...                     */
	j=hist[k].j=idx[ja];      /* ... merged with cluster ja ...     */
	hist[k].d=da;             /* ... at distance da                 */
	nij=n[ia]+n[ja];

	/* update distance matrix */
	switch (methodnr) {
	case 0:  /* single */
	  for (la=0;la<ja;la++) {
	    l=idx[la];
	    if (d[i][l]<d[j][l]) d[j][l]=d[i][l];
	  }

	  for (la++;la<ia;la++) {
	    l=idx[la];
	    if (d[i][l]<d[l][j]) d[l][j]=d[i][l];
	  }
	  
	  for (la++;la<N-k;la++) {
	    l=idx[la];
	    if (d[l][i]<d[l][j]) d[l][j]=d[l][i];
	  }
	  break;

	case 1:  /* complete */
	  for (la=0;la<ja;la++) {
	    l=idx[la];
	    if (d[i][l]>d[j][l]) d[j][l]=d[i][l];
	  }

	  for (la++;la<ia;la++) {
	    l=idx[la];
	    if (d[i][l]>d[l][j]) d[l][j]=d[i][l];
	  }
	  
	  for (la++;la<N-k;la++) {
	    l=idx[la];
	    if (d[l][i]>d[l][j]) d[l][j]=d[l][i];
	  }
	  break;

	case 2: /* average */
	  for (la=0;la<ja;la++) {
	    l=idx[la];
	    d[j][l]=0.5*(d[j][l]+d[i][l]);
	  }

	  for (la++;la<ia;la++) {
	    l=idx[la];
	    d[l][j]=0.5*(d[l][j]+d[i][l]);
	  }
	  
	  for (la++;la<N-k;la++) {
	    l=idx[la];
	    d[l][j]=0.5*(d[l][j]+d[l][i]);
	  }
	  break;

	case 3: /* weighted */
	  for (la=0;la<ja;la++) {
	    l=idx[la];
	    d[j][l]=(n[ja]*d[j][l]+n[ia]*d[i][l])/nij;
	  }

	  for (la++;la<ia;la++) {
	    l=idx[la];
	    d[l][j]=(n[ja]*d[l][j]+n[ia]*d[i][l])/nij;
	  }
	  
	  for (la++;la<N-k;la++) {
	    l=idx[la];
	    d[l][j]=(n[ja]*d[l][j]+n[ia]*d[l][i])/nij;
	  }
	  break;

	case 4: /* median */
	  for (la=0;la<ja;la++) {
	    l=idx[la];
	    d[j][l]=0.5*(d[j][l]+d[i][l])-0.25*da;    
	  }

	  for (la++;la<ia;la++) {
	    l=idx[la];
	    d[l][j]=0.5*(d[l][j]+d[i][l])-0.25*da;    
	  }
	  
	  for (la++;la<N-k;la++) {
	    l=idx[la];
	    d[l][j]=0.5*(d[l][j]+d[l][i])-0.25*da;    
	  }
	  break;

	case 5: /* centroid */
	  for (la=0;la<ja;la++) {
	    l=idx[la];
	    d[j][l]=(n[ja]*d[j][l]+n[ia]*d[i][l])/nij-n[ia]*n[ja]*da/(nij*nij);
	  }

	  for (la++;la<ia;la++) {
	    l=idx[la];
	    d[l][j]=(n[ja]*d[l][j]+n[ia]*d[i][l])/nij-n[ia]*n[ja]*da/(nij*nij);
	  }
	  
	  for (la++;la<N-k;la++) {
	    l=idx[la];
	    d[l][j]=(n[ja]*d[l][j]+n[ia]*d[l][i])/nij-n[ia]*n[ja]*da/(nij*nij);
	  }
	  break;

	case 6: /* ward */
	  for (la=0;la<ja;la++) {
	    l=idx[la];
	    d[j][l]=((n[la]+n[ja])*d[j][l]+(n[la]+n[ia])*d[i][l]-n[la]*da)/(n[la]+nij);
	  }

	  for (la++;la<ia;la++) {
	    l=idx[la];
	    d[l][j]=((n[la]+n[ja])*d[l][j]+(n[la]+n[ia])*d[i][l]-n[la]*da)/(n[la]+nij);
	  }
	  
	  for (la++;la<N-k;la++) {
	    l=idx[la];
	    d[l][j]=((n[la]+n[ja])*d[l][j]+(n[la]+n[ia])*d[l][i]-n[la]*da)/(n[la]+nij);
	  }
	  break;

	case 7: /* flexible */
	  for (la=0;la<ja;la++) {
	    l=idx[la];
	    d[j][l]=alpha*(d[j][l]+d[i][l])+(1-2*alpha)*da;
	  }

	  for (la++;la<ia;la++) {
	    l=idx[la];
	    d[l][j]=alpha*(d[l][j]+d[i][l])+(1-2*alpha)*da;
	  }
	  
	  for (la++;la<N-k;la++) {
	    l=idx[la];
	    d[l][j]=alpha*(d[l][j]+d[l][i])+(1-2*alpha)*da;
	  }
	  break;

	}

	/* update idx pointers and n counters, im and dm */
	n[ja]=nij;
	for (i=ia+1;i<N-k;i++) {
	  if (im[i]>ia) im[i]--;
	  else if (im[i]==ia) im[i]=-1;   /* im,dm invalid */
	  idx[i-1]=idx[i];
	  n[i-1]=n[i];
	  dm[i-1]=dm[i];
	  im[i-1]=im[i];
	}
	
	/* update dm and im */
	dm[ja]=d[j][idx[0]]; im[ja]=0;
	for (la=0;la<ja;la++) {
	  if (d[j][idx[la]]<dm[ja]) {
	    im[ja]=la;
	    dm[ja]=d[j][idx[la]];
	  }
	}
	for (la=ja+1;la<N-k;la++) {
	  if (im[la]>=0&&dm[la]>=d[idx[la]][j]) {
	    im[la]=ja;
	    dm[la]=d[idx[la]][j];
	  }else if (im[la]<0||im[la]==ja) {
	    dm[la]=d[idx[la]][idx[0]]; im[la]=0;
	    for (ia=1;ia<la;ia++) {
	      if (d[idx[la]][idx[ia]]<dm[la]) {
		im[la]=ia;
		dm[la]=d[idx[la]][idx[ia]];
	      }
	    }
	  }
	}
      }
    }else r=1; /* scratch memory allocation failed */
  }else r=2;   /* unknown method */

  /* free temporary stuff */
  if (d&&cpy) {
    for (i=0;i<N;i++) free(d[i]);
    free(d);
  }

  if (r==0&&methods[methodnr].squared) 
    for (k=0;k<N-1;k++) hist[k].d=sqrt(hist[k].d<0?0:hist[k].d);

  if (n) free(n);
  if (idx) free(idx);
  if (im) free(im);
  if (dm) free(dm);

  return(r);
}

/*================================================================================*/
int slink(int N,float **D,agglom_clhist_t *hist)
{
  /* single linkage algorithm 

  input:
     N : number of objects to cluster
     D : lower triangular distance matrix (array of pointers to rows)
     method: "single"|"complete"|"average"|"weighted"|"median"|"centroid"|"ward"|"flexible"
     alpha: alpha parameter for "flexible" method, must be in [0.1...1.0], good: 0.5-0.7
     cpy : (boolean) TRUE if D must be preserved, FALSE if it may be destroyed

  output:
     hist : array of N-1 structures with {d,i,j} (dist at merge, objidx i, objidx j). 

  returns:
     0: O.K.
     1: failed to allocate scratch memory.

  Note: about the same speed as agglom(method=single), would have advantage over
     agglom D were replace by a function (each distance is accessed only once).

  */

  int r=0,i,j,k,n=0;
  float d,dmin;
  int *new;

  for (i=0;i<N-1;i++) {
    hist[i].i=-1;
    hist[i].d=FLT_MAX;
  }
  for (j=0,i=0;i<N-1;i++) {
    dmin=FLT_MAX;
    for (k=0;k<N-1;k++) {
      if (hist[k].i<0) {
	d=(k+1<j?D[j][k+1]:D[k+1][j]);
	if (d<hist[k].d) {
	  hist[k].d=d;
	  hist[k].j=j;
	}
	if (dmin>hist[k].d) {
	  dmin=hist[k].d;
	  n=k;
	}
      }
    }
    hist[n].i=n+1;
    j=n+1;
  }
  for (k=0;k<N-1;k++) {
    if (hist[k].i<hist[k].j) {
      i=hist[k].j;
      hist[k].j=hist[k].i;
      hist[k].i=i;
    }
  }

  qsort(hist,N-1,sizeof(agglom_clhist_t),histcmp);

  new=malloc(N*sizeof(int));
  if (new) {
    for (k=0;k<N;k++) new[k]=k;
    for (k=0;k<N-1;k++) {
      for (i=hist[k].i;i!=new[i];) i=new[i];
      for (j=hist[k].j;j!=new[j];) j=new[j];
      hist[k].i=(i>j?i:j);
      hist[k].j=(i>j?j:i);
      new[hist[k].i]=hist[k].j;
    }
    
  }else r=1;

  return(r);
}

static int histcmp(const void *a,const void *b)
{
  int r;

  if (((agglom_clhist_t *)a)->d>((agglom_clhist_t *)b)->d) r=1;
  else if (((agglom_clhist_t *)a)->d<((agglom_clhist_t *)b)->d) r=-1;
  else r=0;

  return(r);
}
