#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

static float max3(float B,float b,float R,float r,float D,float d, int *ar);

/*-------------------- align --------------------*/

float align(int m,int *matchSA,int n,int *matchSB,void *data,
	   float (*compare)(void *data,int i,int j),
	   float gapo,float gape,int igOverhang,int *status)
{
  /* Find best alignment of two 'sequences', SA and SB., using dynamic
     programming. 

     Target: sum(compare(i,j); matched pairs)
                 -ngaps*gapo -(ningaps-ngaps)*gape --> max,

	     where ngaps is the number of gaps in the alignment, and 
	     ningaps is the total length of all gaps. 

     Returns: score, alignment in matchSA,matchSB.
	   
     input: 
             m: number of elements in SA (input),
	     n: number of elements in SB (input),
	     gapo: gap opening penalty
	     gape: gap extension penalty
	     igOverhang: TRUE if 'overhangs' are to be ignored, i.e. if the gap
                         penalties are to be applied only to 'internal' gaps.
	     data: pointer to additional data, passed to comparison function.
     output:
             matchSA[i] = index of SB matched with SA[i] (-1: SA[i] not matched);
             matchSB[I] = index of SA matched with SB[i] (-1: SB[i] not matched);
             status: 0 OK, 1 problem with args, 2 malloc error.

     returns: score of optimal alignment, 0 if status != 0.
  */

  char  *T;
  float *pmat[8];
  float *M,*R,*Rp,*D,*Dp,*B,*Bp,d,gc,go,god,gor,ge,ged,ger,score;
  int i,j,k,ar[3],state,target;

  *status=0;

  if (n<1||m<1||!matchSA||!matchSB) {
    *status=1;
    return(0);
  }

  igOverhang=(igOverhang?1:0);   /* force to 0,1 */

  M=(float *)malloc(6*(m+1)*sizeof(float));
  T=(char *)malloc(m*n*sizeof(float));
  if (!M||!T) {
    if (T) free(T);
    if (M) free(M);
    *status=2;
    return(0);
  }

  /*

  M: scratch array, will hold two columns for each of the three scoring 
     matrices B, R, and D.

  T: trace matrix (n*m). Each element of this matrix holds 3 2-bit integers
     encoding the optimal state changes during the second (forward) pass of
     the dynamic programming algorithm.

     States are decribed by three integers (i,j,k): first i elements of SA
     are optimally aligned with first j elements of SB, and k=0,1,2 determines
     the allowed state changes as follows:

          ( i, j, 0 ) --> ( i+1, j+1, X ) , X = 0,1,2
	  ( i ,j, 1 ) --> ( i,   j+1, X ) , X = 0,1,2
	  ( i ,j, 2 ) --> ( i+1  j,   X ) , X = 0,1,2
  
     The value of X for the optimal state change from (i,j,k) is 

          X = ( T(i,j) / 2^k ) modulo 4
  */

  /***************************************************/
  /* dynamic programming, backward pass, determining */
  /* optinmal score and tracing matrix T             */
  /***************************************************/

  /* pointers to 3 x 2 columns of score matrices */
  pmat[0]=M;
  for (k=1;k<6;k++) {
    pmat[k]=pmat[k-1]+m+1;
  }

  /* in the following loop, B, R, and D keep track of the optimal score that
     can be achieved from state (i,j,0), (i,j,1), and (i,j,2) respectively.
     To save memory space, only two columns of these matrices are kept: the 
     current column j in B, R, and D, the columns from the previous iteration
     j+1 in Bp, Rp, and Dp. B(i), R(i), and D(i) are computed recursively from
     values in Bp, Rp, Dp, B(i+1), R(i+1), D(i+1).

     At the end, the T(i,j) table will contain the trace matrix for the optimal
     alignment, and B(0), R(0), D(0) will contain the best scores that can be
     achieved starting from state (0,0,0), (0,0,1) and (0,0,2) respectively.
  */

  /* initialize n-th column of score matrices */

  B=pmat[0+n%2]; R=pmat[2+n%2]; D=pmat[4+n%2]; 

  for(i=0;i<m;i++) {B[i]=-(m-i)*gape*(1-igOverhang);} B[m]=0;
  for(i=0;i<m;i++) {R[i]=-(m-i)*gape*(1-igOverhang);} R[m]=0;
  for(i=0;i<m;i++) {D[i]=-(m-i)*gape*(1-igOverhang);} D[m]=0;

#ifdef DBG
  fprintf(stderr,"%3d B: ",-1);for (i=0;i<=m;i++) fprintf(stderr," %5.1f",B[i]);fprintf(stderr,"\n");
  fprintf(stderr,"%3d R: ",-1);for (i=0;i<=m;i++) fprintf(stderr," %5.1f",R[i]);fprintf(stderr,"\n");
  fprintf(stderr,"%3d D: ",-1);for (i=0;i<=m;i++) fprintf(stderr," %5.1f",D[i]);fprintf(stderr,"\n");
  fprintf(stderr,"\n");
#endif

  go=-0.5*(gapo-gape);
  gc=-0.5*(gapo+gape);
  ge=-gape;

  /* fill in table backwards by columns */
  for (j=n-1,god=go,ged=ge;j>=0;j--) {     

    /* pointer dance: swap pointers to previous and current columns */
    B=pmat[0+j%2]; Bp=pmat[0+(j+1)%2];
    R=pmat[2+j%2]; Rp=pmat[2+(j+1)%2];
    D=pmat[4+j%2]; Dp=pmat[4+(j+1)%2];

    /* initialize m-th row of score matrices */
    B[m]=-(n-j)*gape*(1-igOverhang);
    R[m]=-(n-j)*gape*(1-igOverhang);
    D[m]=-(n-j)*gape*(1-igOverhang);

    if (j==0&&igOverhang) {
      god=0;
      ged=0;
    }

    /* fill in column j backwards */
    for (i=m-1,gor=go,ger=ge;i>=0;i--) {

      d=compare(data,i,j);      /* bonus for a match */
      if (i==0&&igOverhang) {
	gor=0;
	ger=0;
      }

      /* update scores for the three types of moves */
      B[i] = max3( Bp[i+1],   d,   Rp[i+1],  d+gc,   Dp[i+1],  d+gc,   ar+0 );
      R[i] = max3( Bp[i],   gor,   Rp[i]  ,   ger,   Dp[i],     gor,   ar+1 );
      D[i] = max3( B [i+1], god,   R [i+1],   god,   D [i+1],   ged,   ar+2 );

      /* store best moves in T array (3*2 bits) */
      T[i+j*m] = ar[0] | (ar[1]<<2) | (ar[2]<<4);

    }

#ifdef DBG
    fprintf(stderr,"%3d B: ",j);for (i=0;i<=m;i++) fprintf(stderr," %5.1f",B[i]);fprintf(stderr,"\n");
    fprintf(stderr,"%3d R: ",j);for (i=0;i<=m;i++) fprintf(stderr," %5.1f",R[i]);fprintf(stderr,"\n");
    fprintf(stderr,"%3d D: ",j);for (i=0;i<=m;i++) fprintf(stderr," %5.1f",D[i]);fprintf(stderr,"\n");
    fprintf(stderr,"\n");
#endif

  }

  /**************************************************/
  /* forward pass, read off an optimal alignment... */
  /**************************************************/

  /* read out optimal score, and determine best starting state */
  score=max3(B[0],0,R[0],0,D[0],0,&state);

  for (i=0;i<m;i++) matchSA[i]=-1;
  for (j=0;j<n;j++) matchSB[j]=-1;
       
  /* read the trace matrix T to read off the otimal alignment */
  for (i=0,j=0;i<m&&j<n;state=target) {
    target=(T[i+j*m]>>(2*state))%4;
    if      (state==1) j++;
    else if (state==2) i++;
    else{
      matchSA[i]=j;
      matchSB[j]=i;
      i++; j++;
    }
  }
      
  if (M) free(M);
  if (T) free(T);

  return(score);
}

/*------------------------------------------------------------------------*/

static float max3(float B,float b,float R,float r,float D,float d, int *ar)
{
  /* Select the best of 3 moves
     B, R, D: values of scores in target states (to which we move),
     b, r, d: gain in score made by move from source state,
     returns score for source state when best move is made (function value),
     and index of best move (0,1,2) (in *ar).
  */

  /* The following defines how to resolve ties:
     Do = 1: if two moves moves lead to the same score, pick the one from
     the state with the higher score, Do = 0: pick the first move.
  */
     
#define Do 0 && 

  float s,t;

  s=B+b;
  t=B;
  *ar=0;
  
  if (R+r>s || (Do R+r==s && R>t)) {
    s=R+r;
    t=R;
    *ar=1;
  }

  if (D+d>s || (Do D+d==s && D>t)) {
    s=D+d;
    t=D;
    *ar=2;
  }

  return(s);
}
