#include <stdio.h>
#include <unistd.h>
#include <stddef.h>
#include <math.h>
#include <time.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "strutil.h"
#include "errhand.h"
#include "transpor.h"
#include "parmset.h"

extern int graphix;

#define DEGTORAD 0.0174533 /* somewhat inaccurate, for consistency with LEAP */

#define FLG1 (1<<1)
#define FLD(rec,f,t) (strfld(fld,rec,f,t))
#define ICO(i,j) ((i)<(j)?((j)*((j)+1))/2+(i):((i)*((i)+1))/2+(j))

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

static void sort_plane(Atom *pln[3]);
static Nodechain exclusions(Atom *a);

typedef struct A_Bond {
  int f,t,tp;            /* from, to, type */
} A_Bond;

typedef struct A_Angle {
  int l,m,r,tp;          /* left, middle, right, type */
} A_Angle;

typedef struct A_Dihe {
  int i,j,k,l,tp;           /* four atom numbers, type */
  int ex;                   /* 1-4 exclusion flag? */
} A_Dihe;

typedef struct A_Impr {
  int n[4],tp;        /* four atom numbers, type */
} A_Impr;

#define itStrict  1 /* ITREE classification, use intra-residue and (i,i+1) inter bonds only */
#define itFIFO    2 /* ITREE classification, strictly FIFO tree order */

#define itMod 8
#define itX  0
#define itR  7
#define itM  5
#define itS  2
#define itB  3
#define it3  4
#define itE  1
static char *TrCode="XESB3MXM";

/* comparison function for NATEX sort */
static int cmpnatex(const void *a, const void *b)
{
  return(*(int *)a - *(int *)b);
}

static size_t my_strftime(char *s,size_t max,const char *fmt, const struct tm *tm)
{
  return strftime(s, max, fmt, tm);
}

/*========================================================================*/

int write_amber_parm(char *file,Node *mol,char *prmfile,char *modfile,int format)
{
  /***************************************/
  /* Write AMBER parameter/topology file */
  /***************************************/

  /* Various counts that will go into the counts record */
  /* -------------------------------------------------- */
  int NATOM=0;  /* total number of atoms */
  int NTYPES=0; /* total number of distinct atom types */
  int NBONH=0;  /* number of bonds containing hydrogen */
  int MBONA=0;  /* number of bonds not containing hydrogen */
  int NTHETH=0; /* number of angles containing hydrogen */
  int MTHETA=0; /* number of angles not containing hydrogen */
  int NPHIH=0;  /* number of dihedrals containing hydrogen */
  int MPHIA=0;  /* number of dihedrals not containing hydrogen */
  int NHPARM=0; /* currently not used */
  int NPARM=0;  /* currently not used */    /* something to do with sander_LES */
  int NNB=0;    /* number of excluded atoms */
  int NRES=0;   /* number of residues */
  int NBONA=0;  /* MBONA + number of constraint bonds */
  int NTHETA=0; /* MTHETA + number of constraint angles */
  int NPHIA=0;  /* MPHIA + number of constraint dihedrals */
  int NUMBND=0; /* number of unique bond types */
  int NUMANG=0; /* number of unique angle types */
  int NPTRA=0;  /* number of unique dihedral types */
  int NATYP=0;  /* number of atom types in parameter file, 0, see SOLTY below */
  int NPHB=0;   /* number of distinct 10-12 hydrogen bond pair types */
  int IFPERT=0; /* set to 1 if perturbation info is to be read in */
  int NBPER=0;  /* number of bonds to be perturbed */
  int NGPER=0;  /* number of angles to be perturbed */
  int NDPER=0;  /* number of dihedrals to be perturbed */
  int MBPER=0;  /* number of bonds with atoms completely in perturbed group */
  int MGPER=0;  /* number of angles with atoms completely in perturbed group */
  int MDPER=0;  /* number of dihedrals with atoms completely in perturbed groups */
  int IFBOX=0;  /* set to 1 if standard periodic box, 2 when truncated octahedral */
  int NMXRS=0;  /* number of atoms in the largest residue */
  int IFCAP=0;  /* set to 1 if the CAP option from edit was specified */

  int **NATEX=(int **)NULL;     /* data for non-bond exclusion list */ 
  int *NUMEX=(int *)NULL;       /* per atom exclusions */
  int *LJ=(int *)NULL;          /* numeric type of i-th atom */
  float *SOLTY=(float *)NULL;   /* currently not used */
  int *ITREE=(int *)NULL;       /* ITREE atom classification */

  int *RESINF=(int *)NULL; /* j-th entry is seqnr of first atom in residue j */

  A_Bond *bnd=(A_Bond *)NULL;          /* bond list */
  A_Angle *ang=(A_Angle *)NULL;        /* angle list */
  A_Dihe *dih=(A_Dihe *)NULL;          /* dihedral list */
  A_Impr *imp=(A_Impr *)NULL;          /* improper list */

  PARMSET PRM;
  Atom *a,*b,*c,*d,**A=(Atom **)NULL;
  Nodechain chain;
  Node *up;
  char buf[4096],sVersionHeader[81],fld[256],*p,*u;
  FILE *out=(FILE *)NULL;
  FILE *prm=(FILE *)NULL;
  int nphih,mphia,nphit=0,nxphit=0;
  int nimph,mimpa,nimpt=0;
  int i,j,k,l,n,ka,kh,*ks,status=0,sfw,bt,s,o[4];
  int itMeth=itStrict;
  double R,E;
  time_t thetime;

  /*------------------------------------------------------------------------*/
  /* count atoms, residues, check atomnames, potential names etc...         */
  /* Note: only atoms with FLG1 set will go into the output file...         */

  PRM=emptyParmSet();
  init_node_chain(&chain);
  up=(Node *)NULL;
  ForAtoms(a,mol) {
    if (a->flg&FLG1) {
      a->seqnr=NATOM++;
      if (!mvld(a->atp)) {
	err_set(1,"atom %s has no type assigned.",nodeptr_to_name(buf,(Node *)a));
	status=2;
      }
      if (a->up!=up) {
	add_to_node_chain((Node *)a,&chain);
	up=a->up;
      }
    }
  }
  
  if (NATOM==0) {
    err_set(1,"no atoms in selected molecule.");
    status=3;
  }

  /* prepare residue information */

  if (chain.n>0) {
    RESINF=(int *)malloc(chain.n*sizeof(int));
    if (mvld(RESINF)) {
      NRES=chain.n;
      for (j=0,a=(Atom *)chain.s;a;a=(Atom *)a->chain,j++) RESINF[j]=a->seqnr;
    }else {
      err_set(1,"failed to allocate scratch array RESINF.");
      status=4;
    }
    NMXRS=NATOM-RESINF[NRES-1];    /* find largest residue size */
    for (j=0;j<NRES-1;j++) {     
      if (RESINF[j+1]-RESINF[j]>NMXRS) NMXRS=RESINF[j+1]-RESINF[j];
    }
  }
  clear_node_chain(&chain);
  
  /* allocate pointer array for atoms in selected molecule */

  if (!status) {
    A=(Atom **)malloc((NATOM)*sizeof(Atom *));
    if (mvld(A)) {
      ForAtoms(a,mol) if (a->flg&FLG1) A[a->seqnr]=a;
    }else{
      err_set(1,"failed to allocate scratch array A.");
      status=4;
    }
  }

  /* allocate LJ (Lenard-Jones) type array */

  if (!status) {
    LJ=(int *)malloc((NATOM)*sizeof(int));
    if (mvld(LJ)) {
      add_to_node_chain((Node *)A[0],&chain);
      LJ[NTYPES++]=0;
      for (i=1;i<NATOM;i++) {
	for (k=0,a=(Atom *)chain.s;a;a=(Atom *)a->chain,k++) {
	  if (A[i]->atp==a->atp) {
	    LJ[i]=k;
	    break;
	  }else if (!a->chain) {
	    add_to_node_chain((Node *)A[i],&chain);
	    LJ[i]=NTYPES++;
	    break;
	  }
	}
      }
      PRM=initParmSetMasses(PRM,&chain);
      clear_node_chain(&chain);
      if (PRM.status) {
	err_set(1,"failed to allocate memory for parameter DB");
	status=4;
      }
    }else{
      err_set(1,"failed to allocate scratch array LJ.");
      status=4;
    }
  }

  /* generate bond list */
  
  if (!status) {
    for (i=0;i<NATOM;i++) {
      for (j=0;j<A[i]->nbnd;j++) {
	b=neighbour(j,A[i]);
	if (b->flg&FLG1&&b->seqnr>A[i]->seqnr) {
	  if (b->atp->z==1||A[i]->atp->z==1) NBONH++;
	  else MBONA++;
	}
      }
    }
    if (MBONA+NBONH>0) {
      bnd=(A_Bond *)malloc((MBONA+NBONH)*sizeof(A_Bond));
      if (!bnd) {
	err_set(1,"failed to allocate scratch array bnd.");
	status=4;
      }
    }
  }
  if (bnd) {
    for (ka=NBONH,kh=0,i=0;i<NATOM;i++) {
      for (j=0;j<A[i]->nbnd;j++) {
	b=neighbour(j,A[i]);
	if (b->flg&FLG1&&b->seqnr>A[i]->seqnr) {
	  ks=b->atp->z==1||A[i]->atp->z==1?&kh:&ka;
	  /* make sure that the type of .f is lexicographically <= type of .t */
	  if (strcmp(A[i]->atp->name,b->atp->name)<=0) {
	    bnd[*ks].f=i;
	    bnd[*ks].t=b->seqnr;
	  }else{
	    bnd[*ks].t=i;
	    bnd[*ks].f=b->seqnr;
	  }
	  (*ks)++;
	}
      }
    }
  }
  NBONA=MBONA+0;    /* 0 constrained distances... */

  /* generate bond type list */

  if (bnd) {
    for (j=0;j<NBONH+MBONA;j++) {     /* first time through: just counting */
      for (k=0;k<NUMBND;k++) {
	if (LJ[bnd[j].f]==LJ[bnd[bnd[k].tp].f]&&
	    LJ[bnd[j].t]==LJ[bnd[bnd[k].tp].t]) break;
      }
      if (k==NUMBND) bnd[NUMBND++].tp=j;
    }
    if (NUMBND>0) {
      PRM.bnd=(P_BOND *)malloc(NUMBND*sizeof(P_BOND));
      if (PRM.bnd) {
	PRM.nbnd=NUMBND;
	NUMBND=0;
	for (j=0;j<NBONH+MBONA;j++) { /* second time: store */
	  for (k=0;k<NUMBND;k++) {
	    if (LJ[bnd[j].f]==PRM.bnd[k].f&&
		LJ[bnd[j].t]==PRM.bnd[k].t) break;
	  }
	  if (k==NUMBND) {
	    PRM.bnd[k].f=LJ[bnd[j].f];
	    PRM.bnd[k].t=LJ[bnd[j].t];
	    PRM.bnd[k].REQ=1.5;
	    PRM.bnd[k].RK=0.0;
	    PRM.bnd[k].fnr=0;
	    NUMBND++;
	  }
	  bnd[j].tp=k;
	}
      }else{
	err_set(1,"failed to allocate scratch array for bond parameters.");
	PRM.nbnd=0;
	PRM.status=1;
	status=4;
      }
    }
  }    

  /* generate angle list */

  if (!status) {
    for (i=0;i<NATOM;i++) {     /* counting... */
      for (j=0;j<A[i]->nbnd-1;j++) { 
	b=neighbour(j,A[i]);
	if (!(b->flg&FLG1)) continue;
	for (k=j+1;k<A[i]->nbnd;k++) {
	  c=neighbour(k,A[i]);
	  if (!(c->flg&FLG1)) continue;
	  if (b->atp->z==1||c->atp->z==1) NTHETH++;
	  else MTHETA++;
	}
      }
    }
    if (NTHETH+MTHETA>0) {
      ang=(A_Angle *)malloc((NTHETH+MTHETA)*sizeof(A_Angle));
      if (!ang) {
	err_set(1,"failed to allocate scratch array ang.");
	status=4;
      }
    }
    if (ang) {
      for (ka=NTHETH,kh=0,i=0;i<NATOM;i++) {   /* storing... */
	for (j=0;j<A[i]->nbnd-1;j++) {
	  b=neighbour(j,A[i]);
	  if (!(b->flg&FLG1)) continue;
	  for (k=j+1;k<A[i]->nbnd;k++) {
	    c=neighbour(k,A[i]);
	    if (!(c->flg&FLG1)) continue;
	    ks=b->atp->z==1||c->atp->z==1?&kh:&ka;
	    ang[*ks].m=i;
	    if (strcmp(b->atp->name,c->atp->name)<=0) {
	      ang[*ks].l=b->seqnr;
	      ang[*ks].r=c->seqnr;
	    }else{
	      ang[*ks].r=b->seqnr;
	      ang[*ks].l=c->seqnr;
	    }
	    (*ks)++;
	  }
	}
      }
    }
  }
  NTHETA=MTHETA+0;  /* 0 constrained angles */

  /* generate angle type list */

  if (ang) {
    for (j=0;j<NTHETH+MTHETA;j++) { /* counting... */
      for (k=0;k<NUMANG;k++) {
	if (LJ[ang[j].m]==LJ[ang[ang[k].tp].m]&&
	    LJ[ang[j].l]==LJ[ang[ang[k].tp].l]&&
	    LJ[ang[j].r]==LJ[ang[ang[k].tp].r]) break;
      }
      if (k==NUMANG) ang[NUMANG++].tp=j;
    }
    if (NUMANG>0) {
      PRM.ang=(P_ANGL *)malloc(NUMANG*sizeof(P_ANGL));
      if (PRM.ang) {
	PRM.nang=NUMANG;
	NUMANG=0;
	for (j=0;j<NTHETH+NTHETA;j++) { /* storing */
	  for (k=0;k<NUMANG;k++) {
	    if (LJ[ang[j].m]==PRM.ang[k].m&&
		LJ[ang[j].l]==PRM.ang[k].l&&
		LJ[ang[j].r]==PRM.ang[k].r) break;
	  }
	  if (k==NUMANG) {
	    PRM.ang[k].m=LJ[ang[j].m];
	    PRM.ang[k].l=LJ[ang[j].l];
	    PRM.ang[k].r=LJ[ang[j].r];
	    PRM.ang[k].TEQ=109.5*DEGTORAD;
	    PRM.ang[k].TK=0.0;
	    PRM.ang[k].fnr=0;
	    PRM.ang[k].lev=4;
	    NUMANG++;
	  }
	  ang[j].tp=k;
	}
      }else{
	err_set(1,"failed to allocate scratch array for angle parameters.");
	PRM.status=1;
	PRM.nang=0;
	status=4;
      }
    }
  }
	  
  /* generate dihedrals list */

  if (!status) {
    nphih=0; mphia=0;
    for (j=0;j<NBONH+MBONA;j++) {           /* counting */
      b=A[bnd[j].f]; c=A[bnd[j].t];
      if (b->nbnd>1&&c->nbnd>1) {
	for (i=0;i<b->nbnd;i++) {
	  a=neighbour(i,b);
	  if (a!=c&&a->flg&FLG1) {
	    for (l=0;l<c->nbnd;l++) {
	      d=neighbour(l,c);
	      if (d!=b&&d->flg&FLG1&&d!=a) {
		if (a->atp->z==1||d->atp->z==1) nphih++;
		else mphia++;
	      }
	    }
	  }
	}
      }
    }
    if (nphih+mphia>0) {
      dih=(A_Dihe *)malloc((nphih+mphia)*sizeof(A_Dihe));
      if (!dih) {
	err_set(1,"failed to allocate scratch array dih.");
	status=4;
      }
    }
    if (dih) {
      mphia=nphih; nphih=0; 
      for (j=0;j<NBONH+MBONA;j++) {           /* storing */
	b=A[bnd[j].f]; c=A[bnd[j].t];
	if (b->nbnd>1&&c->nbnd>1) {
	  for (i=0;i<b->nbnd;i++) {
	    a=neighbour(i,b);
	    if (a!=c&&a->flg&FLG1) {
	      for (l=0;l<c->nbnd;l++) {
		d=neighbour(l,c);
		if (d!=b&&d->flg&FLG1&&d!=a) {
		  ks=(a->atp->z==1||d->atp->z==1?&nphih:&mphia);
		  if (!strcmp(b->atp->name,c->atp->name)) {
		    if (strcmp(a->atp->name,d->atp->name)>0) sfw=FALSE;
		    else sfw=TRUE;;
		  }else sfw=TRUE;
		  if (sfw) {
		    dih[*ks].i=a->seqnr; dih[*ks].j=b->seqnr; 
		    dih[*ks].k=c->seqnr; dih[*ks].l=d->seqnr; 
		  }else{
		    dih[*ks].i=d->seqnr; dih[*ks].j=c->seqnr; 
		    dih[*ks].k=b->seqnr; dih[*ks].l=a->seqnr; 
		  }
		  dih[*ks].tp=-1;
		  dih[*ks].ex=1;
		  (*ks)++;
		}
	      }
	    }
	  }
	}
      }
      mphia=mphia-nphih;
    }
  }
  
  /* generate dihedral type list */

  if (dih) {
    for (nphit=0,j=0;j<nphih+mphia;j++) {  /* counting... */
      for (k=0;k<nphit;k++) {
	if (LJ[dih[j].i]==LJ[dih[dih[k].tp].i]&&
	    LJ[dih[j].j]==LJ[dih[dih[k].tp].j]&&
	    LJ[dih[j].k]==LJ[dih[dih[k].tp].k]&&
	    LJ[dih[j].l]==LJ[dih[dih[k].tp].l]) break;
      }
      if (k==nphit) dih[nphit++].tp=j;
    }
    if (nphit>0) {
      PRM.dih=(P_DIHE *)malloc(nphit*sizeof(P_DIHE));
      if (PRM.dih) {
	for (nphit=0,j=0;j<nphih+mphia;j++) {  /* storing... */
	  for (k=0;k<nphit;k++) {
	    if (LJ[dih[j].i]==PRM.dih[k].i&&
		LJ[dih[j].j]==PRM.dih[k].j&&
		LJ[dih[j].k]==PRM.dih[k].k&&
		LJ[dih[j].l]==PRM.dih[k].l) break;
	  }
	  if (k==nphit) {
	    PRM.dih[k].m=0;
	    PRM.dih[k].exact=0;
	    PRM.dih[k].i=LJ[dih[j].i];
	    PRM.dih[k].j=LJ[dih[j].j];
	    PRM.dih[k].k=LJ[dih[j].k];
	    PRM.dih[k].l=LJ[dih[j].l];
	    PRM.dih[k].ok=FALSE;
	    nphit++;
	  }
	  dih[j].tp=k;
	}
	PRM.ndih=nphit;
      }else{
	err_set(1,"failed to allocate scratch array for dihedral parameters.");
	PRM.ndih=0;
	PRM.status=1;
	status=4;
      }
    }
  }

  /* generate impropers list */

  if (!status) {
    for (nimph=0,mimpa=0,i=0;i<NATOM;i++) {        /* counting */
      if (A[i]->oop&&A[i]->nbnd==3) {
	for (n=0,ks=&mimpa,j=0;j<A[i]->nbnd;j++) {
	  b=neighbour(j,A[i]);
	  if (b->flg&FLG1) {
	    if (b->atp->z==1) ks=&nimph;
	    n++;
	  }
	}
	if (n==3) {
	  (*ks)++;
	}
      }
    }
    if (nimph+mimpa>0) {
      imp=(A_Impr *)malloc((nimph+mimpa)*sizeof(A_Impr));
      if (!imp) {
	err_set(1,"failed to allocate scratch array imp.");
	status=4;
      }
    }
    if (imp) {
      mimpa=nimph; nimph=0; 
      for (i=0;i<NATOM;i++) {        /* storing */
	if (A[i]->oop&&A[i]->nbnd==3) {
	  for (n=0,ks=&mimpa,j=0;j<A[i]->nbnd;j++) {
	    b=neighbour(j,A[i]);
	    if (b->flg&FLG1) {
	      if (b->atp->z==1) ks=&nimph;
	      n++;
	    }
	  }
	  if (n==3) {
	    imp[*ks].n[0]=i;
	    for (j=0;j<3;j++) imp[*ks].n[j+1]=neighbour(j,A[i])->seqnr;
	    if (strcmp(A[imp[*ks].n[1]]->atp->name,A[imp[*ks].n[2]]->atp->name)>0) {
	      s=imp[*ks].n[1]; imp[*ks].n[1]=imp[*ks].n[2]; imp[*ks].n[2]=s;
	    }
	    if (strcmp(A[imp[*ks].n[2]]->atp->name,A[imp[*ks].n[3]]->atp->name)>0) {
	      s=imp[*ks].n[2]; imp[*ks].n[2]=imp[*ks].n[3]; imp[*ks].n[3]=s;
	    }
	    if (strcmp(A[imp[*ks].n[1]]->atp->name,A[imp[*ks].n[2]]->atp->name)>0) {
	      s=imp[*ks].n[1]; imp[*ks].n[1]=imp[*ks].n[2]; imp[*ks].n[2]=s;
	    }
	    (*ks)++;
	  }
	}
      }
      mimpa=mimpa-nimph;
    }
  }
  
  /* generate improper type list */

  if (imp) {
    for (nimpt=0,j=0;j<nimph+mimpa;j++) {  /* counting... */
      for (k=0;k<nimpt;k++) {
	if (LJ[imp[j].n[0]]==LJ[imp[imp[k].tp].n[0]]&&
	    LJ[imp[j].n[1]]==LJ[imp[imp[k].tp].n[1]]&&
	    LJ[imp[j].n[2]]==LJ[imp[imp[k].tp].n[2]]&&
	    LJ[imp[j].n[3]]==LJ[imp[imp[k].tp].n[3]]) break;
      }
      if (k==nimpt) imp[nimpt++].tp=j;
    }
    if (nimpt>0) {
      PRM.imp=(P_IMPR *)malloc(nimpt*sizeof(P_IMPR));
      if (PRM.imp) {
	for (nimpt=0,j=0;j<nimph+mimpa;j++) {  /* storing... */
	  for (k=0;k<nimpt;k++) {
	    if (LJ[imp[j].n[0]]==PRM.imp[k].t[0]&&
		LJ[imp[j].n[1]]==PRM.imp[k].t[1]&&
		LJ[imp[j].n[2]]==PRM.imp[k].t[2]&&
		LJ[imp[j].n[3]]==PRM.imp[k].t[3]) break;
	  }
	  if (k==nimpt) {
	    PRM.imp[k].fnr=0;
	    PRM.imp[k].lev=5;
	    for (l=0;l<4;l++) PRM.imp[k].t[l]=LJ[imp[j].n[l]];
	    nimpt++;
	  }
	  imp[j].tp=k;
	}
	PRM.nimp=nimpt;
      }else{
	err_set(1,"failed to allocate scratch array for improper parameters.");
	PRM.nimp=0;
	PRM.status=1;
	status=4;
      }
    }
  }

  /* generate exclusion list */

  if (!status) {
    NUMEX=(int *)malloc(NATOM*sizeof(int));
    NATEX=(int **)malloc(NATOM*sizeof(int *));
    if (NATEX) for(i=0;i<NATOM;i++) NATEX[i]=(int *)NULL;
    if (mvld(NATEX)&&mvld(NUMEX)) {
      for (i=0;i<NATOM;i++) {
	chain=exclusions(A[i]);
	NUMEX[i]=chain.n;
	NNB+=NUMEX[i];
	if (NUMEX[i]>0) {
	  NATEX[i]=(int *)malloc(NUMEX[i]*sizeof(int));
	  if (mvld(NATEX[i])) {
	    for (j=0,b=(Atom *)chain.s;b;b=(Atom *)b->chain,j++) NATEX[i][j]=b->seqnr;
	    if (NUMEX[i]>1) qsort(NATEX[i],NUMEX[i],sizeof(int),cmpnatex); /* Amber 6!*/
	  }else{
	    err_set(1,"failed to allocate exclusion list");  
	    status=4;
	  }
	}
	clear_node_chain(&chain);
      }
    }else{
      err_set(1,"failed to allocate exclusion list"); 
      status=4;
    }
  }

  /* set 1-4 exclusion flags in dihedrals list */

  if (!status) {
    for (k=0;k<nphih+mphia;k++) {
      dih[k].ex=1;
      for (j=0;j<NBONH+MBONA&&dih[k].ex>0;j++) {
	if ((bnd[j].f==dih[k].i&&bnd[j].t==dih[k].l)||
	    (bnd[j].t==dih[k].i&&bnd[j].f==dih[k].l)) dih[k].ex=-1;
      }
      for (j=0;j<NTHETH+MTHETA&&dih[k].ex>0;j++) {
	if ((ang[j].l==dih[k].i&&ang[j].r==dih[k].l)||
	    (ang[j].r==dih[k].i&&ang[j].l==dih[k].l)) dih[k].ex=-1;
      }
      for (j=0;j<k&&dih[k].ex>0;j++) {
	if ((dih[j].i==dih[k].i&&dih[j].l==dih[k].l)||
	    (dih[j].l==dih[k].i&&dih[j].i==dih[k].l)) dih[k].ex=-1;
      }
    }
  }

  /* generate ITREE classification */
  /* ITREE%8 : tree code (mainchain, sidechain, branch, 3-way, leaf, other) */
  /* ITREE/8 : precursor atom (only if not root) */

  if (!status) {
    ITREE=(int *)malloc(NATOM*sizeof(int));
    if (mvld(ITREE)) {
      for (i=0;i<NATOM;i++) ITREE[i]=itX;
      init_node_chain(&chain);
      for (i=0;i<NATOM;i++) {
	if (ITREE[i]==0) {
	  ITREE[i]=itR;        /* assign M(root) to first atom of connected fragment */
	  add_to_node_chain((Node *)A[i],&chain);
	  for (a=A[i];a;a=(Atom *)a->chain) { /* foreach atom in fragment... */
	    for (k=0,j=0;j<a->nbnd;j++) {
	      b=neighbour(j,a);
	      if (b->flg&FLG1&&ITREE[b->seqnr]==itX) {  /* count number k of unass. neighbours */
		if (itMeth!=itStrict||b->up==a->up||b->up==a->up->nxt) { 
		  ITREE[b->seqnr]=a->seqnr*itMod+itE;     /* assign E to unass. neighbours */
		  add_to_node_chain((Node *)b,&chain);
		  k++;
		}
	      }
	    }
	    if (ITREE[a->seqnr]!=itR) {                 /* except for root of fragment... */
	      switch (k) {                              /* assign class dep. on k */
	      case 0: ITREE[a->seqnr]=ITREE[a->seqnr]-itE+itE;break;
	      case 1: ITREE[a->seqnr]=ITREE[a->seqnr]-itE+itS;break;
	      case 2: ITREE[a->seqnr]=ITREE[a->seqnr]-itE+itB;break;
	      case 3: ITREE[a->seqnr]=ITREE[a->seqnr]-itE+it3;break;
	      default: ITREE[a->seqnr]=ITREE[a->seqnr]-itE+itX;break;
	      }
	    }
	  }
	}
      } /* chain has all atoms in LIFO tree order */
      for (a=(Atom *)chain.s;a;a=b) {  /* assign M to longest branch of each tree */
	b=(Atom *)a->chain;
	if (ITREE[a->seqnr]!=itR&&(!b||ITREE[b->seqnr]==itR)) { /* at end of longest chain */
	  while (ITREE[a->seqnr]!=itR) {  /* back to the roots */
	    k=ITREE[a->seqnr]/itMod;
	    ITREE[a->seqnr]=k*itMod+itM;
	    a=A[k];
	  }
	}
      }
      for (i=0;i<NATOM;i++) {           /* classify all Hydrogens as E */
	if (A[i]->atp->z==1) {
	  ITREE[i]=ITREE[i]/itMod;
	  ITREE[i]=ITREE[i]*itMod+itE;
	}
      }
      clear_node_chain(&chain);
    }else {
      err_set(1,"failed to allocate ITRTEE array"); 
      status=4;
    }
  }

  /* read parameter definitions */

  if (!status&&prmfile) {
    prm=fopen(prmfile,"r");
    PRM=readParmSet(prm,PRM);
    if (prm) fclose(prm);
    if (PRM.status) {
      err_set(1,"could not read AMBER parameter file %s",prmfile);
      status=5;
    }
  }
  if (!status&&modfile) {
    prm=fopen(modfile,"r");
    PRM=readParmSet(prm,PRM);
    if (prm) fclose(prm);
    if (PRM.status) {
      err_set(1,"could not read FrcMod parameter file %s",modfile);
      status=5;
    }
  }

  /* report missing parameters */

  if (!status) {
    PRM=reportParmSetMissing(PRM);
    if (PRM.status) {
      err_set(1,"missing force field parameters");
      status=6;
    }
  }

  /* proper counting of dihedrals and impropers (merged with dihedrals) */

  if (!status) { 
    for (NPTRA=0,k=0;k<PRM.ndih;k++) {
      PRM.dih[k].bt=NPTRA;
      NPTRA+=PRM.dih[k].m; /* types, with proper multiplicity */
    }
    nxphit=NPTRA;
    NPTRA+=nimpt;
    for (NPHIH=0,k=0;k<nphih;k++) {
      NPHIH+=PRM.dih[dih[k].tp].m;
    }
    NPHIH+=nimph;    
    for (MPHIA=0,k=0;k<mphia;k++) {
      MPHIA+=PRM.dih[dih[nphih+k].tp].m;
    }
    MPHIA+=mimpa;
    NPHIA=MPHIA+0;      /* 0 dihedral constraints */
  }

  /* atom number 0 (FORTRAN 1) must not occur in positions 3,4 of */
  /* dihedrals and impropers (negative atom number = flag!)       */

  if (!status) {
    for (k=0;k<nphih+mphia;k++) {
      if (dih[k].k==0||dih[k].l==0) {
	s=dih[k].l; dih[k].l=dih[k].i; dih[k].i=s;
	s=dih[k].k; dih[k].k=dih[k].j; dih[k].j=s;
      }
    }
    for (k=0;k<nimph+mimpa;k++) {
      for (j=0;j<4;j++) o[j]=imp[k].n[PRM.imp[imp[k].tp].o[j]];   /* order prescribed by type */
      if (o[2]==0||o[3]==0) for (j=0;j<4;j++) imp[k].n[j]=o[3-j]; /* backward */
      else for (j=0;j<4;j++) imp[k].n[j]=o[j];  /* forward */
    }
  }

  /* ------------------------------------------------------------------------------*/
  /*                          time to write the file                               */
  /* ------------------------------------------------------------------------------*/

  /* open output file (no default extension) */

  if (!status) {
    if ((out=fopen(file,"w"))==NULL) {
      err_set(1,"write_amber_parm: could not open %s for write\n",file);
      status=1;
    }
  }

  if (!status) {

    /* 1: write title */

    if (format==PRMTOP_FMT_NEW) {
      time(&thetime);
      my_strftime(sVersionHeader, 81,
             "%%VERSION  VERSION_STAMP = V0001.000  DATE = %m/%d/%y  %H:%M:%S",
	       localtime(&thetime));
      fprintf(out,"%s\n%s\n%s\n",
	      sVersionHeader,
	      "%FLAG TITLE",
	      "%FORMAT(20a4)");
    }

    fprintf(out,"AMBER parameter file for %s, generated by Wit!P.\n",mol->name);

    /* 2: write counts line */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG POINTERS",
	      "%FORMAT(12I6)");
    }

    fprintf(out,"%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d\n",
	    NATOM,NTYPES,NBONH,MBONA,NTHETH,MTHETA,NPHIH,MPHIA,NHPARM,NPARM,NNB,NRES);
    fprintf(out,"%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d\n",
	    NBONA,NTHETA,NPHIA,NUMBND,NUMANG,NPTRA,NATYP,NPHB,IFPERT,NBPER,NGPER,NDPER);
    fprintf(out,"%6d%6d%6d%6d%6d%6d%6d\n",
	    MBPER,MGPER,MDPER,IFBOX,NMXRS,IFCAP,0);

    /* 3: write user atom names (truncated to 4 characters) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG ATOM_NAME",
	      "%FORMAT(20a4)");
    }
    for (i=0;i<NATOM;i++) {
      fprintf(out,"%-4s",FLD(A[i]->name,1,4));
      if ((i+1)%20==0) fprintf(out,"\n");
    }
    if (i%20) fprintf(out,"\n");

    /* 4: write partial atomic charges, multiplied by 18.2223 (sqrt(332.0522)) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG CHARGE",
	      "%FORMAT(5E16.8)");
    }

    for (i=0;i<NATOM;i++) {
      fprintf(out,"%16.8e",18.2223*A[i]->q);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i%5) fprintf(out,"\n");

    /* 5: write atomic masses */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG MASS",
	      "%FORMAT(5E16.8)");
    }
    for (i=0;i<NATOM;i++) {
      fprintf(out,"%16.8e",PRM.mass[LJ[i]]);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i%5) fprintf(out,"\n");

    /* 6: write LJ types */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG ATOM_TYPE_INDEX",
	      "%FORMAT(12I6)");
    }
    for (i=0;i<NATOM;i++) {
      fprintf(out,"%6d",LJ[i]+1);
      if ((i+1)%12==0) fprintf(out,"\n");
    }
    if (i%12) fprintf(out,"\n");
    

    /* 7: write NUMEX counters (number of exclusions per atom) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG NUMBER_EXCLUDED_ATOMS",
	      "%FORMAT(12I6)");
    }

    for (i=0;i<NATOM;i++) {
      fprintf(out,"%6d",NUMEX[i]);
      if ((i+1)%12==0) fprintf(out,"\n");
    }
    if (i%12) fprintf(out,"\n");

    /* 8: write ICO pointers to LJ parameters */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG NONBONDED_PARM_INDEX",
	      "%FORMAT(12I6)");
    }

    for (k=0,i=0;i<NTYPES;i++) {
      for (j=0;j<NTYPES;j++,k++) {
	fprintf(out,"%6d",ICO(i,j)+1);
	if ((k+1)%12==0) fprintf(out,"\n");
      }
    }
    if (k==0||k%12) fprintf(out,"\n");

    /* 9: write residue labels (trunc. to 4 chars.) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG RESIDUE_LABEL",
	      "%FORMAT(20a4)");
    }

    for (j=0;j<NRES;j++) {
      p=A[RESINF[j]]->up->name;
      u=strrchr(p,'_'); if (!u) u=p+strlen(p);
      fprintf(out,"%-4s",FLD(p,1,(u-p>3 ? 4 : u-p)));
      if ((j+1)%20==0) fprintf(out,"\n");
    }
    if (j==0||j%20) fprintf(out,"\n");

    /* 10: write IPRES (atom number of first atom in i-th residue) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG RESIDUE_POINTER",
	      "%FORMAT(12I6)");
    }
    for (j=0;j<NRES;j++) {
      fprintf(out,"%6d",RESINF[j]+1);
      if ((j+1)%12==0) fprintf(out,"\n");
    }
    if (j==0||j%12) fprintf(out,"\n");

    /* 11: write RK for bond types */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG BOND_FORCE_CONSTANT",
	      "%FORMAT(5E16.8)");
    }
    for (i=0;i<NUMBND;i++) {
      fprintf(out,"%16.8e",PRM.bnd[i].RK);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");

    /* 12: write REQ for bond types */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG BOND_EQUIL_VALUE",
	      "%FORMAT(5E16.8)");
    }
    for (i=0;i<NUMBND;i++) {
      fprintf(out,"%16.8e",PRM.bnd[i].REQ);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");

    /* 13: write TK for angle types */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG ANGLE_FORCE_CONSTANT",
	      "%FORMAT(5E16.8)");
    }
    for (i=0;i<NUMANG;i++) {
      fprintf(out,"%16.8e",PRM.ang[i].TK);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");

    /* 14: write TEQ for angle types */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG ANGLE_EQUIL_VALUE",
	      "%FORMAT(5E16.8)");
    }
    for (i=0;i<NUMANG;i++) {
      fprintf(out,"%16.8e",PRM.ang[i].TEQ);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");

    /* 15: write PK for dihedral types */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG DIHEDRAL_FORCE_CONSTANT",
	      "%FORMAT(5E16.8)");
    }
    for (i=0,k=0;k<PRM.ndih;k++) {
      for (l=0;l<PRM.dih[k].m;l++,i++) {
	fprintf(out,"%16.8e",PRM.dih[k].PK[l]);
	if ((i+1)%5==0) fprintf(out,"\n");
      }
    }
    for (k=0;k<PRM.nimp;k++,i++) {
      fprintf(out,"%16.8e",PRM.imp[k].PK);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");
     
    /* 16: write PN (periodicity) for dihedral types */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG DIHEDRAL_PERIODICITY",
	      "%FORMAT(5E16.8)");
    }
    for (i=0,k=0;k<PRM.ndih;k++) {
      for (l=0;l<PRM.dih[k].m;l++,i++) {
	fprintf(out,"%16.8e",PRM.dih[k].PN[l]);
	if ((i+1)%5==0) fprintf(out,"\n");
      }
    }
    for (k=0;k<PRM.nimp;k++,i++) {
      fprintf(out,"%16.8e",PRM.imp[k].PN);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");
     
    /* 17: write PHASE for dihedral types */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG DIHEDRAL_PHASE",
	      "%FORMAT(5E16.8)");
    }
    for (i=0,k=0;k<PRM.ndih;k++) {
      for (l=0;l<PRM.dih[k].m;l++,i++) {
	fprintf(out,"%16.8e",PRM.dih[k].PHASE[l]);
	if ((i+1)%5==0) fprintf(out,"\n");
      }
    }
    for (k=0;k<PRM.nimp;k++,i++) {
      fprintf(out,"%16.8e",PRM.imp[k].PHASE);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");

    /* 18: write dummy record for SOLTY (not currently used) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG SOLTY",
	      "%FORMAT(5E16.8)");
    }
    for (i=0;i<NATYP;i++) {
      fprintf(out,"%16.8e",SOLTY[i]);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");

    /* 19: write CN1 LJ parameters */
    
    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG LENNARD_JONES_ACOEF",
	      "%FORMAT(5E16.8)");
    }
    for (k=0,i=0;i<NTYPES;i++) {
      for (j=0;j<=i;j++,k++) {
	R=(PRM.R[i]+PRM.R[j]);
	E=sqrt(PRM.E[i]*PRM.E[j]);
	fprintf(out,"%16.8e",pow(R,12)*E);
	if ((k+1)%5==0) fprintf(out,"\n");
      }
    }
    if (k==0||k%5) fprintf(out,"\n");
	
    /* 20: write CN2 LJ parameters */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG LENNARD_JONES_BCOEF",
	      "%FORMAT(5E16.8)");
    }
    for (k=0,i=0;i<NTYPES;i++) {
      for (j=0;j<=i;j++,k++) {
	R=(PRM.R[i]+PRM.R[j]);
	E=sqrt(PRM.E[i]*PRM.E[j]);
	fprintf(out,"%16.8e",2*pow(R,6)*E);
	if ((k+1)%5==0) fprintf(out,"\n");
      }
    }
    if (k==0||k%5) fprintf(out,"\n");

    /* 21: write bond H terms (at1,at2,type), obfuscated (*3) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG BONDS_INC_HYDROGEN",
	      "%FORMAT(12I6)");
    }
    for (j=0;j<NBONH;j++) {
      fprintf(out,"%6d%6d%6d",3*bnd[j].f,3*bnd[j].t,bnd[j].tp+1);
      if ((j+1)%4==0) fprintf(out,"\n");
    }
    if (j==0||j%4) fprintf(out,"\n");

    /* 22: write bond nonH terms (at1,at2,type), obfuscated */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG BONDS_WITHOUT_HYDROGEN",
	      "%FORMAT(12I6)");
    }
    for (j=0;j<NBONA;j++) {
      k=j+NBONH;
      fprintf(out,"%6d%6d%6d",3*bnd[k].f,3*bnd[k].t,bnd[k].tp+1);
      if ((j+1)%4==0) fprintf(out,"\n");
    }
    if (j==0||j%4) fprintf(out,"\n");

    /* 23: write angle H terms (at1,at2,at3,type), obfuscated */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG ANGLES_INC_HYDROGEN",
	      "%FORMAT(12I6)");
    }
    for (j=0;j<NTHETH;j++) {
      fprintf(out,"%6d%6d%6d%6d",3*ang[j].l,3*ang[j].m,3*ang[j].r,ang[j].tp+1);
      if ((j+1)%3==0) fprintf(out,"\n");
    }
    if (j==0||j%3) fprintf(out,"\n");
 
    /* 24: write angle nonH terms (at1,at2,at3,type), obfuscated */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG ANGLES_WITHOUT_HYDROGEN",
	      "%FORMAT(12I6)");
    }
    for (j=0;j<NTHETA;j++) {
      k=j+NTHETH;
      fprintf(out,"%6d%6d%6d%6d",3*ang[k].l,3*ang[k].m,3*ang[k].r,ang[k].tp+1);
      if ((j+1)%3==0) fprintf(out,"\n");
    }
    if (j==0||j%3) fprintf(out,"\n");

    /* 25: write H angle dihedral terms (at1,at2,at3,at4,type), obfuscated */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG DIHEDRALS_INC_HYDROGEN",
	      "%FORMAT(12I6)");
    }
    for (i=0,k=0;k<nphih;k++) {
      bt=PRM.dih[dih[k].tp].bt+1;      /* +1 will make FORTRAN programmers happy */
      for (j=0,s=dih[k].ex;j<PRM.dih[dih[k].tp].m;j++,i+=5) {
	fprintf(out,"%6d",3*dih[k].i); if ((i+1)%12==0) fprintf(out,"\n");
	fprintf(out,"%6d",3*dih[k].j); if ((i+2)%12==0) fprintf(out,"\n");
	fprintf(out,"%6d",s*3*dih[k].k); if ((i+3)%12==0) fprintf(out,"\n");
	fprintf(out,"%6d",3*dih[k].l); if ((i+4)%12==0) fprintf(out,"\n");
	fprintf(out,"%6d",bt+j);       if ((i+5)%12==0) fprintf(out,"\n");
	s=-1;
      }
    }
    for (k=0;k<nimph;k++,i+=5) {
      bt=nxphit+1;
      for (l=0;l<4;l++) {
	fprintf(out,"%6d",(l<2?3:-3)*imp[k].n[l]);   
	if ((i+l+1)%12==0) fprintf(out,"\n");
      }
      fprintf(out,"%6d",bt+imp[k].tp); if ((i+5)%12==0) fprintf(out,"\n");
    }
    if (i==0||i%12) fprintf(out,"\n");

    /* 26: write nonH angle dihedral terms (at1,at2,at3,at4,type), obfuscated */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG DIHEDRALS_WITHOUT_HYDROGEN",
	      "%FORMAT(12I6)");
    }
    for (i=0,l=0;l<mphia;l++) {
      k=nphih+l;
      bt=PRM.dih[dih[k].tp].bt+1;      /* +1 will make FORTRAN programmers happy */
      for (j=0,s=dih[k].ex;j<PRM.dih[dih[k].tp].m;j++,i+=5) {
	fprintf(out,"%6d",3*dih[k].i); if ((i+1)%12==0) fprintf(out,"\n");
	fprintf(out,"%6d",3*dih[k].j); if ((i+2)%12==0) fprintf(out,"\n");
	fprintf(out,"%6d",s*3*dih[k].k); if ((i+3)%12==0) fprintf(out,"\n");
	fprintf(out,"%6d",3*dih[k].l); if ((i+4)%12==0) fprintf(out,"\n");
	fprintf(out,"%6d",bt+j);       if ((i+5)%12==0) fprintf(out,"\n");
	s=-1;
      }
    }
    for (l=0;l<mimpa;l++,i+=5) {
      k=nimph+l;
      bt=nxphit+1;
      for (j=0;j<4;j++) {
	fprintf(out,"%6d",(j<2?3:-3)*imp[k].n[j]);   
	if ((i+j+1)%12==0) fprintf(out,"\n");
      }
      fprintf(out,"%6d",bt+imp[k].tp); if ((i+5)%12==0) fprintf(out,"\n");
    }
    if (i==0||i%12) fprintf(out,"\n");

    /* 27: write NATEX lists (excluded interactions) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG EXCLUDED_ATOMS_LIST",
	      "%FORMAT(12I6)");
    }
    for (k=0,i=0;i<NATOM;i++) {
      for (j=0;j<NUMEX[i];j++,k++) {
	fprintf(out,"%6d",NATEX[i][j]+1);  /* our indices are zero-based */
	if ((k+1)%12==0) fprintf(out,"\n");
      }
    }
    if (k==0||k%12) fprintf(out,"\n");

    /* 28: write ASOL H-bond parameters (not implemented) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG HBOND_ACOEF",
	      "%FORMAT(5E16.8)");
    }
    for (k=0;k<NPHB;k++) {
      fprintf(out,"%16.8e",0.0);
      if ((k+1)%5==0) fprintf(out,"\n");
    }
    if (k==0||k%5) fprintf(out,"\n"); 

    /* 29: write BSOL H-Bond parameters  (not implemented) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG HBOND_BCOEF",
	      "%FORMAT(5E16.8)");
    }
    for (k=0;k<NPHB;k++) {
      fprintf(out,"%16.8e",0.0);
      if ((k+1)%5==0) fprintf(out,"\n");
    }
    if (k==0||k%5) fprintf(out,"\n"); 

    /* 30: write HBCUT parameters (no longer used) */
    
    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG HBCUT",
	      "%FORMAT(5E16.8)");
    }
    for (i=0;i<NPHB;i++) {
      fprintf(out,"%16.8e",0.0);
      if ((i+1)%5==0) fprintf(out,"\n");
    }
    if (i==0||i%5) fprintf(out,"\n");

    /* 31: write Amber Atomtypes (trunc. 4 ) */
    
    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG AMBER_ATOM_TYPE",
	      "%FORMAT(20a4)");
    }
    for (i=0;i<NATOM;i++) {
      fprintf(out,"%-4s",FLD(A[i]->atp->name,1,4));
      if ((i+1)%20==0) fprintf(out,"\n");
    }
    if (i%20) fprintf(out,"\n");

    /* 32: write ITREE atom classification */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG TREE_CHAIN_CLASSIFICATION",
	      "%FORMAT(20a4)");
    }
    for (i=0;i<NATOM;i++) {
      fprintf(out,"%c   ",TrCode[ITREE[i]%itMod]);
      if ((i+1)%20==0) fprintf(out,"\n");
    }
    if (i%20) fprintf(out,"\n");

    /* 33: write JOIN array (1 zero per atom, not used) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG JOIN_ARRAY",
	      "%FORMAT(12I6)");
    }
    for (i=0;i<NATOM;i++) {
      fprintf(out,"%6d",0);
      if ((i+1)%12==0) fprintf(out,"\n");
    }
    if (i%12) fprintf(out,"\n");

    /* 34: write IROTAT array (1 zero per atom, not used) */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG IROTAT",
	      "%FORMAT(10I8)");
    }
    for (i=0;i<NATOM;i++) {
      fprintf(out,"%8d",0);
      if ((i+1)%10==0) fprintf(out,"\n");
    }
    if (i%10) fprintf(out,"\n");

    /* 35: write GBradii */

    if (format==PRMTOP_FMT_NEW) {
      fprintf(out,"%s\n%s\n",
	      "%FLAG RADII",
	      "%FORMAT(5E16.8)");
      for (i=0;i<NATOM;i++) {
	fprintf(out,"%16.8e",A[i]->atp->GBrad);
	if ((i+1)%5==0) fprintf(out,"\n");
      }
      if (i==0||i%5) fprintf(out,"\n");
     
      /* 36: write GBscreen */

      fprintf(out,"%s\n%s\n",
	      "%FLAG SCREEN",
	      "%FORMAT(5E16.8)");
      for (i=0;i<NATOM;i++) {
	fprintf(out,"%16.8e",A[i]->atp->GBscr);
	if ((i+1)%5==0) fprintf(out,"\n");
      }
      if (i==0||i%5) fprintf(out,"\n");
    }

  }

  /* done. if there were any errors, remove output file. */

  if (out) fclose(out);
  if (status) unlink(file);

  /* free all allocated scratch arrays */

  PRM=clearParmSet(PRM);
  if (mvld(ITREE)) free(ITREE);
  if (mvld(imp)) free(imp);
  if (mvld(dih)) free(dih);
  if (mvld(ang)) free(ang);
  if (mvld(bnd)) free(bnd);
  if (mvld(RESINF)) free(RESINF);
  if (mvld(NATEX)) {
    for (i=0;i<NATOM;i++) if (NATEX[i]) free(NATEX[i]);
    free(NATEX);
  }
  if (mvld(NUMEX)) free(NUMEX);
  if (mvld(SOLTY)) free(SOLTY);
  if (mvld(LJ)) free(LJ);
  if (mvld(A)) free(A);

  if (checkParmSetMode(PRM_SLOPPY_TEST)==PRM_SLOPPY_ONCE) checkParmSetMode(PRM_SLOPPY_NEVER);
  return(status);
}
/*========================================================================*/

static Nodechain exclusions(Atom *a)
{
  /* find atoms to be excluded in nonbonded interactions   */
  /* must have FLG1 set, seqnr higher than a, and minimum  */
  /* path length to a using flagged atoms only less than   */
  /* four.                                                 */                                              
  Nodechain excl;
  Atom *b;
  int i,k,l,n,s;

  s=a->seqnr;                    /* save for later use... */
  init_node_chain(&excl);
  for (i=0;i<a->nbnd;i++) {      /* direct neighbours... */
    b=neighbour(i,a);
    if (b->flg&FLG1) add_to_node_chain((Node *)b,&excl);
  }
  k=0;
  a=(Atom *)excl.s;
  for (l=2;l<4;l++) {            /* path lengths 2 and 3... */
    n=excl.n;
    for (;k<n;a=(Atom *)a->chain,k++) {
      for (i=0;i<a->nbnd;i++) {
	b=neighbour(i,a);
	if (b->flg&FLG1) add_to_node_chain((Node *)b,&excl);
      }
    }
  }
  excl.e=(Node *)NULL;
  a=(Atom *)excl.s;
  init_node_chain(&excl);
  for (;a;a=b) {                 /* remove if seqnr is low */
    b=(Atom *)a->chain;
    a->chain=(Node *)NULL;
    if (a->seqnr>s) add_to_node_chain((Node *)a,&excl);
  }
  return(excl);
}
/*------------------------------------------------------------------------*/

int write_amber_crd(char *file,Node *mol)
{
  int status=0,i,n;
  FILE *out;
  Atom *a;

  out=fopen(file,"w");
  if (out) {
    n=0;
    ForAtoms(a,mol) if (a->flg&FLG1&&a->nxyz>0&&a->x) n++;
    fprintf(out,"Amber coordinate file for %s, written by Wit!P\n",mol->name);
    fprintf(out,"%5d\n",n);
    i=0;
    ForAtoms(a,mol) {
      if (a->flg&FLG1&&a->nxyz>0&&a->x) {
	i++;
	fprintf(out,"%12.7f%12.7f%12.7f",a->x[0],a->x[1],a->x[2]);
	if (!(i%2)) fprintf(out,"\n");
	if (ferror(out)) break;
      }
    }
    if (i<n) {
      err_set(1,"error writing Amber coordinates file %s",file);
      status=1;
    }
    if (i==0||i%2) fprintf(out,"\n");
  }else{
    err_set(1,"failed to open Amber coordinate file %s for write.",file);
    status=1;
  }
  if (out) fclose(out);
  return(status);
}
/*------------------------------------------------------------------------*/

int read_amber_crd(UCpipe *in,Node *mol)
{
  char rec[2096],fld[13],*p;
  int status=0,i,j,k,n;
  Atom *a;
  
  n=0; ForAtoms(a,mol) if (a->flg&FLG1&&a->nxyz>0&&a->x) n++;
  if (!fgets(rec,sizeof(rec),in->fd))status=1;   /* skip title */
  if (fgets(rec,sizeof(rec),in->fd)) sscanf(rec,"%d",&i); /* get number of atoms */
  else i=0;
  if (i==n&&n>0) {
    i=0; p=rec;
    ForAtoms(a,mol) {
      if (a->flg&FLG1&&a->nxyz>0&&a->x) {
	if (i%2==0) {
	  if (!fgets(rec,sizeof(rec),in->fd)) break;
	  for (k=strlen(rec);k<72;k++) {rec[k]=' ';} rec[k]='\0';
	  p=rec;
	}
	for (j=0;j<3;j++) {
	  strncpy(fld,p,12); p+=12; fld[12]='\0';
	  a->x[j]=str2f(fld);
	}
	i=i+1;
      }
    }
    if (i<n) {
      err_set(1,"failed to read all coordinates.");
      status=1;
    }
  }else if (i!=n) {
    err_set(1,"number of atoms in molecule does not  number in file.");
    status=1;
  }
  return(status);
}

/*------------------------------------------------------------------------*/

int write_amber_prep(char *filename,Node *mol)
{
  int status=0,nat=0,nbnd=0,ncyc,nres=0,*top=NULL,*back=NULL,*nxt=NULL;
  int i,j,k,l,m=0,n,p,np=0,mbr,seq=0;
  Node *up=NULL;
  char buf[4096];
  Atom *a,*b,**atoms=NULL,*pln[3];
  Bond *bnd;
  FILE *out;
  char *code="ESB3456";

  ForAtoms(a,mol) {
    if (a->flg&FLG1) {
      nat++;
      if (!mvld(a->atp)) {
	err_set(1,"atom %s has no type assigned.",nodeptr_to_name(buf,(Node *)a));
	status=1;
      }
      if (a->up!=up) {
	up=a->up;
	nres++;
      }
    }
  }
  if (nat==0) {
    err_set(1,"no atoms to write to Amber prepfile %s",filename);
    status=2;
  }
  if (nres>1) {
    err_set(1,"multi-residue Amber prep files are not suported in this version of Wit!P.");
    status=3;
  }

  if (!status) {                           /* optional: bubble sort[sorry]  bond lists ...*/
    i=0; ForAtoms(a,up) a->seqnr=i++;
    ForAtoms(a,up) {
      if (a->flg&FLG1) {
	for (l=1;l<a->nbnd;l++) {
	  for (k=l,seq=neighbour(l,a)->seqnr;k>0;k--) {
	    if (neighbour(k-1,a)->seqnr>seq) {
	      bnd=a->bnd[k-1];
	      a->bnd[k-1]=a->bnd[k];
	      a->bnd[k]=bnd;
	    }else break;
	  }	    
	}
      }
    }
  }

  if (!status) {
    ForAtoms(a,up) {
      if (a->flg&FLG1) {
	for (j=0;j<a->nbnd;j++) {
	  if (a==a->bnd[j]->f) {
	    b=neighbour(j,a);
	    if (b->flg&FLG1) nbnd++;
	  }
	}
      }
    }

    ncyc=nbnd-nat+1;
    if (ncyc<0) {
      err_set(1,"residue to write to Amber prep file is not connected.");
      status=4;
    }
  }

  if (!status) {
    atoms=(Atom **)malloc((nat+ncyc)*sizeof(Atom *));
    back=(int *)malloc((nat+ncyc)*sizeof(int));
    top=(int *)malloc(nat*sizeof(int));
    nxt=(int *)malloc(nat*sizeof(int));
    if (!top||!atoms||!back||!nxt) {
      err_set(1,"failed to allocate scratch space in write_amber_prep.");
      status=5;
    }
  }

  if (!status) {
    
    /* find spanning tree and ring closure bonds */

    ForAtoms(a,up) if (a->flg&FLG1) a->seqnr=-1;
    ForAtoms(a,up) if (a->flg&FLG1) break;
    /* add first atom to tree */

    n=1; m=0; p=0;  /* number of atoms, number of ring closures, atoms index */
    a->seqnr=p; atoms[p]=a; top[p]=0; back[p]=-1; nxt[p]=0; 

    while (p>=0) {
      while(nxt[p]<atoms[p]->nbnd) {
	b=neighbour(nxt[p],atoms[p]); nxt[p]++;
	if ((p==0||b!=atoms[back[p]])&&  
	    b->up==up && b->flg&FLG1) {
	  if (b->seqnr<0) {
	    top[p]++;
	    b->seqnr=n;
	    atoms[n]=b;
	    top[n]=0;
	    back[n]=p;
	    nxt[n]=0;
	    p=n;
	    n++;
	  }else if (b->seqnr<atoms[p]->seqnr) {
	    atoms[nat+m]=b;
	    back[nat+m]=p;
	    m++;
	  }
	}
      }
      p=back[p];
    }

    if (n<nat) {
      err_set(1,"residue to write to Amber prep file is not connected.");
      status=4;
    }      

    if (m!=ncyc) err_set(0,"internal error: %d ring closures, expected %d.",m,ncyc);
  }

  if (!status) {
    mbr=strlen(code)-1;
    for (i=0;i<nat;i++) if (top[i]>mbr) break;
    if (i<nat) {
      err_set(1,"max. level of branching (%d) exceeded at atom %s",mbr,nodeptr_to_name(buf,(Node *)atoms[i]));
      status=6;
    }
  }

  if (!status) {
    out=fopen(filename,"w");
    if (out) {
      fprintf(out,"0 0 2\n");                                /* no DB generation, all-atom model      */
      fprintf(out,"\n");                                     /* name of database file                 */
      fprintf(out,"Amber Prep file generated by Wit!P\n");   /* descriptive header for the residue    */
      fprintf(out,"%s\n",extfil(buf,filename,".res",TRUE));  /* output filename                       */
      strcpy(buf,up->name); if (strrchr(buf,'_')) *strrchr(buf,'_')='\0';
      fprintf(out,"%s XYZ 0\n",buf);                         /* residue name, type of output, format  */
      fprintf(out,"CHANGE OMIT DU ALL\n");                   /* cartesian input, omit all dummy atoms */
      fprintf(out,"0.0\n");                                  /* cutoff for automatic bonds            */
      seq=1;
      fprintf(out,"%4d DUMM  DU  M  0 0 0  0\n",seq++);      /* 1st dummy atom                        */
      fprintf(out,"%4d DUMM  DU  M  1 0 0  0\n",seq++);      /* 2nd dummy atom                        */
      fprintf(out,"%4d DUMM  DU  M  1 1 0  0\n",seq++);      /* 3rd dummy atom                        */

      /* Atom section */
      i=0; ForAtoms(a,up) if (a->flg&FLG1) a->seqnr=++i;
      for (i=0;i<nat;i++) {
	fprintf(out,"%4d  %-6s  %-3s  %c %11.5f %11.5f %11.5f %9.6f\n",
		seq++,atoms[i]->name,atoms[i]->atp->name,i==0?'M':code[top[i]],
		atoms[i]->x[0],atoms[i]->x[1],atoms[i]->x[2],atoms[i]->q);
      }

      fprintf(out,"\n\n");                                   /* 2 blank lines terminate atom section  */
      
      /* Ring closure bonds */
      if (m>0) {
	fprintf(out,"LOOP\n");
	for (;i<nat+m;i++) {
	  fprintf(out," %-6s %-6s\n",atoms[i]->name,atoms[back[i]]->name);
	}
	fprintf(out,"\n");
      }

      /* Dihedral terms for planarity of atoms with 3 bonds and oop!=0 */
      for (i=0;i<nat;i++) {
	if (atoms[i]->oop) {
	  for (k=0,j=0;j<atoms[i]->nbnd;j++) {
	    if (neighbour(j,atoms[i])->flg&FLG1) pln[k++]=neighbour(j,atoms[i]);
	  }
	  if (k==3) {
	    if (np==0) fprintf(out,"IMPROPER\n");
	    sort_plane(pln);                                 /* put neighbours in proper order        */
	    fprintf(out," %-6s %-6s%-6s %-6s\n",pln[0]->name,pln[1]->name,atoms[i]->name,pln[2]->name);
	    np++;
	  }
	}
      }
      if (np>0) fprintf(out,"\n");

      fprintf(out,"DONE\n");
      fprintf(out,"STOP\n");
      fclose(out);
    }else{
      err_set(1,"failed to open Amber prep file %s for write.",filename);
      status=6;
    }
  }

  if (nxt) free(nxt);
  if (top) free(top);
  if (back) free(back);
  if (atoms) free(atoms);

  return(status);
}

static void sort_plane(Atom *pln[3])
{
  Atom *s;

  if (strcmp(pln[0]->atp->name,pln[1]->atp->name)>0 ||
      (strcmp(pln[0]->atp->name,pln[1]->atp->name) == 0 && pln[0]->seqnr > pln[1]->seqnr)) {
    s=pln[0];
    pln[0]=pln[1];
    pln[1]=s;
  }

  if (strcmp(pln[1]->atp->name,pln[2]->atp->name)>0 ||
      (strcmp(pln[1]->atp->name,pln[2]->atp->name) == 0 && pln[1]->seqnr > pln[2]->seqnr)) {
    s=pln[1];
    pln[1]=pln[2];
    pln[2]=s;
  }

  if (strcmp(pln[0]->atp->name,pln[1]->atp->name)>0 ||
      (strcmp(pln[0]->atp->name,pln[1]->atp->name) == 0 && pln[0]->seqnr > pln[1]->seqnr)) {
    s=pln[0];
    pln[0]=pln[1];
    pln[1]=s;
  }
}
