#include <stdio.h>
#include "mystdlib.h"
#include "bhtree.h"
#include <math.h>
#include "errhand.h"

/* Barnes - Hut Trees ? not quite, but almost...           */
/* No: more like median K-D trees, apparently invented by  */
/* Bentley: "Multidimensional Binary Search Trees Used for */
/* Associative Searching", ACM Sept. 1975 Vol 18. No 9.    */

#define MXTREE 1024

#define NSTEPS 128

static int findBHcloseAtomsInNode(BHnode *node,float *x,float cutoff,
				  Atom **atom,float *dist,int maxn);
static int markBHcloseAtomsInNode(BHnode *node,float *x,float cutoff,int flag);

static int nrfindBHcloseAtomsInNode(BHnode *node,int idx,float cutoff,
				  Atom **atom,float *dist,int maxn);
static float findBHmindistInNode(BHnode *node,float *x,float cutoff);

static Atom *findBHclosestInNode(BHnode *node,float *x,float cutoff,float *d);

static Atom *findBHnearestNeighbourInNode(BHnode *node,Atom *A,float cutoff,float *r);

static int CUT=BH_CUT_MEDIAN;       /* method to determine cutting plane */

/*-------------------- GenerateBHtree --------------------*/

BHtree *GenerateBHtree(int N,Atom **atoms,int granularity)
{
  /* 3D space-sort atoms into Barnes-Hut tree with given granularity */
  /* (max number of atoms per leaf-node) */

  BHtree *r;
  Atom **p;
  int i,k;

  /* allocate tree data structure */

  r=(BHtree *)malloc(sizeof(BHtree));
  if (!mvld(r)) return(r);
  r->atom=(Atom **)NULL;
  r->own_atoms=FALSE;
  r->n=0;
  
  /* allocate root node data structure */

  r->root=(BHnode *)malloc(sizeof(BHnode));
  if (!mvld(r->root)) {
    freeBHtree(r);
    return((BHtree *)NULL);
  }

  /* initialize root node data structure */

  r->root->atom=(Atom **)NULL;
  r->root->n=0; r->root->marked=0;
  r->root->dim=-1;
  r->root->left=(BHnode *)NULL;
  r->root->right=(BHnode *)NULL;

  r->n=N;
  if (r->n==0) {
    freeBHtree(r);
    return((BHtree *)NULL);
  }
  r->atom=atoms;
  if (!mvld(r->atom)) { 
    freeBHtree(r);
    return((BHtree *)NULL);
  }
  r->root->atom=r->atom;
  r->root->n=N;

  /* determine box dimension */

  p=r->root->atom;

  for (k=0;k<3;k++) {
    r->xmin[k]=p[0]->x[k];
    r->xmax[k]=r->xmin[k];
  }

  for (i=1;i<r->root->n;i++) {
    for (k=0;k<3;k++) {
      if (r->xmin[k] > p[i]->x[k]) r->xmin[k] = p[i]->x[k];
      if (r->xmax[k] < p[i]->x[k]) r->xmax[k] = p[i]->x[k];
    }
  }

  /* start recursive 3D space sort */

  divideBHnode(r->root,r->xmin,r->xmax,granularity);

  /* done... */

  return(r);
}


/*-------------------- generateBHtree --------------------*/

BHtree *generateBHtree(Nodechain *atoms,int granularity)
{
  BHtree *r;
  Atom **A;

  A=(Atom **)chain_to_ptr(atoms,FALSE);
  r=GenerateBHtree(atoms->n,A,granularity);
  if (mvld(r)) r->own_atoms=TRUE;
  return(r);
}

/*-------------------- find_BHnode --------------------*/

BHnode *findBHnode(BHtree *tree,float *x)
{
  /* find leaf in BH tree that contains 3D point x */

  BHnode *r;
  int k;

  if (!mvld(tree)) return((BHnode *)NULL);

  /* first check if point is in tree */

  for (k=0;k<3;k++) {
    if (x[k]<tree->xmin[k]) return((BHnode *)NULL);
    if (x[k]>tree->xmax[k]) return((BHnode *)NULL);
  }

  /* if point is in tree, locate the leaf */

  r=tree->root;
  while (mvld(r)) {
    if (r->dim<0) break;
    if (x[r->dim]<r->cut) r=r->left;
    else r=r->right;
  }

  return(r);
}
/*-------------------- freeBHtree --------------------*/

void freeBHtree(BHtree *tree)
{
  if (mvld(tree)) {
    freeBHnode(tree->root);
    if (tree->own_atoms&&mvld(tree->atom)) free(tree->atom);
    free(tree);
  }
}
/*-------------------- freeBHnode --------------------*/

void freeBHnode(BHnode *node)
{
  if (mvld(node)) {
    freeBHnode(node->left);
    freeBHnode(node->right);
    free(node);
  }
}
/*-------------------- divideBHtree --------------------*/

void divideBHnode(BHnode *node,float *xmin,float *xmax,int granularity)
{
  /* does not work well for some pathological cases... */

  float cut,dx,xminl[3],xmaxl[3],xminr[3],xmaxr[3];
  int dim,i,j,k,n[NSTEPS],lm,rm;
  Atom *a;

  /* if nothing left to divide just return */

  if (!mvld(node)) return;
  if (granularity<1 || node->n <= granularity) return;
  if (!mvld(node->atom)) return;

  /* determine dimension along which to cut */

  dim=0;
  if (xmax[1]-xmin[1] > xmax[dim]-xmin[dim]) dim=1;
  if (xmax[2]-xmin[2] > xmax[dim]-xmin[dim]) dim=2;

  /* determine position of cutting plane */

  switch (CUT) {
  case BH_CUT_MIDDLE:
    cut=0.5*(xmax[dim]+xmin[dim]);
    if (xmax[dim]-0.1 < cut || xmin[dim]+0.1 > cut) return;
    break;
  case BH_CUT_MEDIAN:
    dx=(xmax[dim]-xmin[dim])/NSTEPS;
    if (dx<0.0001) return;
    for (i=0;i<NSTEPS;i++) n[i]=0;
    for (j=0;j<node->n;j++) {
      i=(node->atom[j]->x[dim]-xmin[dim])/dx;
      if (i>=0 && i<NSTEPS) n[i]++;
    }
    for (i=1;i<NSTEPS;i++) {
      n[i]+=n[i-1];
      if (n[i]>node->n/2) break;
    }
    if (i==NSTEPS) i--;
    cut=xmin[dim]+i*dx;
    break;
  default:
    cut=0.5*(xmax[dim]+xmin[dim]);
    if (xmax[dim]-0.1 < cut || xmin[dim]+0.1 > cut) return;
    break;

  }

  /* sort atoms into left/right descendants */

  lm=0;
  rm=node->n-1;

  while (lm<rm) {
    for(;lm<node->n;lm++) if (node->atom[lm]->x[dim]>=cut) break;
    for(;rm>=0;rm--)      if (node->atom[rm]->x[dim]<cut) break;
    if (lm<rm) {
      a=node->atom[rm];
      node->atom[rm]=node->atom[lm];
      node->atom[lm]=a;
      rm--;
      lm++;
    }
  }

  if (lm==rm) {
    if (node->atom[rm]->x[dim]<cut) lm++;
    else rm--;
  }

  /* create left/right descendants */

  node->left=(BHnode *)malloc(sizeof(BHnode));
  if (!mvld(node->left)) return;
  node->left->dim=-1;
  node->left->left=(BHnode *)NULL;
  node->left->right=(BHnode *)NULL;
        
  node->right=(BHnode *)malloc(sizeof(BHnode));
  if (!mvld(node->right)) {
    freeBHnode(node->left);
    return;
  }
  node->right->dim=-1;
  node->right->left=(BHnode *)NULL;
  node->right->right=(BHnode *)NULL;

  node->left->n=rm+1; node->left->marked=0;
  node->left->atom=node->atom;

  node->right->n=node->n-(rm+1); node->right->marked=0;
  node->right->atom=node->atom+lm;

  node->cut=cut;
  node->dim=dim;

  /* if descendants are coarse, cut them up... */

  if (node->left->n > granularity) {
    for (k=0;k<3;k++) {
      xminl[k]=xmin[k];
      xmaxl[k]=xmax[k];
    }
    xmaxl[dim]=cut;
    divideBHnode(node->left,xminl,xmaxl,granularity);
  }

  if (node->right->n > granularity) {
    for (k=0;k<3;k++) {
      xminr[k]=xmin[k];
      xmaxr[k]=xmax[k];
    }
    xminr[dim]=cut;
    divideBHnode(node->right,xminr,xmaxr,granularity);
  }

  /* done... */

  return;

}
/*-------------------- findBHmindist --------------------*/

float findBHmindist(BHtree *tree,float *x)
{
  /* find minimum distance from point x to atoms in tree */

  float d;

  d=1.e30;

  if (!mvld(tree)) return(d);
  if (!mvld(tree->root)) return(d);
  if (tree->n<1) return(d);

  d=cdist(x,tree->atom[0]->x);
  d=findBHmindistInNode(tree->root,x,d);
  return(d);
}

/*-------------------- findBHclosest --------------------*/

Atom *findBHclosest(BHtree *tree,float *x,float dmax,float *d)
{
  /* find atom closest to point x, accept only atoms within dmxax */

  Atom *a;

  a=(Atom *)NULL;
  if (!mvld(tree)) return(a);
  if (!mvld(tree->root)) return(a);
  if (tree->n<1) return(a);

  a=findBHclosestInNode(tree->root,x,dmax,d);
  return(a);
}
/*-------------------- findBHclosestInNode --------------------*/

static Atom *findBHclosestInNode(BHnode *node,float *x,float cutoff,float *r)
{
  int j;
  float d[3],D,C;
  Atom *a,*b;

  a=(Atom *)NULL;
  if (!mvld(node)) return(a);
  if (node->n < 1) return(a);
  
  if (node->dim<0) {
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      d[0]=x[0]-node->atom[j]->x[0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j]->x[1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j]->x[2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      cutoff=sqrt(D);
      C=D;
      a=node->atom[j];
      *r=cutoff;
    }
  }else{
    if (x[node->dim]<node->cut) {
      a=findBHclosestInNode(node->left,x,cutoff,r);
      if (is_atom(a)) cutoff= *r;
      if (x[node->dim]>=node->cut-cutoff) {
	b=findBHclosestInNode(node->right,x,cutoff,r);
	if (is_atom(b)) a=b;
      }
    }else{
      a=findBHclosestInNode(node->right,x,cutoff,r);
      if (is_atom(a)) cutoff= *r;
      if (x[node->dim]<node->cut+cutoff) {
	b=findBHclosestInNode(node->left,x,cutoff,r);
	if (is_atom(b)) a=b;
      }
    }
  }
  return(a);
}

/*-------------------- findBHnearestNeighbour --------------------*/

Atom *findBHnearestNeighbour(BHtree *tree,Atom *A,float dmax,float *d)
{
  /* find atom closest to atom a, accept only atoms within dmxax */

  Atom *a;

  a=(Atom *)NULL;
  if (!mvld(tree)) return(a);
  if (!mvld(tree->root)) return(a);
  if (tree->n<1) return(a);

  a=findBHnearestNeighbourInNode(tree->root,A,dmax,d);
  return(a);
}

/*-------------------- findBHnearestNeighbourInNode --------------------*/

static Atom *findBHnearestNeighbourInNode(BHnode *node,Atom *A,float cutoff,float *r)
{
  int j;
  float d[3],D,C;
  Atom *a,*b;

  a=(Atom *)NULL;
  if (!mvld(node)) return(a);
  if (node->n < 1) return(a);
  
  if (node->dim<0) {
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      if (A==node->atom[j]) continue;
      d[0]=A->x[0]-node->atom[j]->x[0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=A->x[1]-node->atom[j]->x[1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=A->x[2]-node->atom[j]->x[2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      cutoff=sqrt(D);
      C=D;
      a=node->atom[j];
      *r=cutoff;
    }
  }else{
    if (A->x[node->dim]<node->cut) {
      a=findBHnearestNeighbourInNode(node->left,A,cutoff,r);
      if (is_atom(a)) cutoff= *r;
      if (A->x[node->dim]>=node->cut-cutoff) {
	b=findBHnearestNeighbourInNode(node->right,A,cutoff,r);
	if (is_atom(b)) a=b;
      }
    }else{
      a=findBHnearestNeighbourInNode(node->right,A,cutoff,r);
      if (is_atom(a)) cutoff= *r;
      if (A->x[node->dim]<node->cut+cutoff) {
	b=findBHnearestNeighbourInNode(node->left,A,cutoff,r);
	if (is_atom(b)) a=b;
      }
    }
  }
  return(a);
}

/*-------------------- findBHcloseAtoms --------------------*/

int findBHcloseAtoms(BHtree *tree,float *x,float cutoff,
		     Atom **atom,float *dist,int maxn)
{
  int i;

  /* find atoms in BH tree that are close to given point x */

  if (maxn<1 || !mvld(tree) || cutoff <=0.0) return(0);
  if (!mvld(tree->root)) return(0);

  for (i=0;i<3;i++) {
    if (x[i] < tree->xmin[i] - cutoff) break;
    if (x[i] > tree->xmax[i] + cutoff) break;
  }
  if (i<3) return(0);

  return(findBHcloseAtomsInNode(tree->root,x,cutoff,atom,dist,maxn));
}
/*-------------------- markBHcloseAtoms --------------------*/

int markBHcloseAtoms(BHtree *tree,float *x,float cutoff,int flag)
{
  int i;

  /* find atoms in BH tree that are close to given point x */

  if (!mvld(tree) || cutoff <=0.0) return(0);
  if (!mvld(tree->root)) return(0);

  for (i=0;i<3;i++) {
    if (x[i] < tree->xmin[i] - cutoff) break;
    if (x[i] > tree->xmax[i] + cutoff) break;
  }
  if (i<3) return(0);

  return(markBHcloseAtomsInNode(tree->root,x,cutoff,flag));
}
/*-------------------- nrfindBHcloseAtoms --------------------*/

int nrfindBHcloseAtoms(BHtree *tree,int idx,float cutoff,
		     Atom **atom,float *dist,int maxn)
{
  int i;

  /* non-redundant find atoms in BH tree close to given atom[idx],
     only finds atoms that come later in the list than atom idx */

  if (maxn<1 || !mvld(tree) || cutoff <=0.0) return(0);
  if (!mvld(tree->root)) return(0);
  if (idx<0 || idx>=tree->n) return(0);

  for (i=0;i<3;i++) {
    if (tree->atom[idx]->x[i] < tree->xmin[i] - cutoff) break;
    if (tree->atom[idx]->x[i] > tree->xmax[i] + cutoff) break;
  }
  if (i<3) return(0);

  return(nrfindBHcloseAtomsInNode(tree->root,idx,cutoff,atom,dist,maxn));
}
/*-------------------- findBHcloseAtomsInNode --------------------*/

static int findBHcloseAtomsInNode(BHnode *node,float *x,float cutoff,
				  Atom **atom,float *dist,int maxn)
{
  int j,n;
  float d[3],D,C;

  if (!mvld(node)) return(0);
  if (maxn<0) return(0);
  if (node->n < 1) return(0);
  
  if (node->dim<0) {
    n=0;
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      d[0]=x[0]-node->atom[j]->x[0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j]->x[1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j]->x[2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      if (n<maxn) {
	atom[n]=node->atom[j];
	dist[n]=sqrt(D);
	n++;
      }else{
	n++;
	break;
      }
    }
  }else{
    n=0;
    if (x[node->dim]<node->cut) {
      n+=findBHcloseAtomsInNode(node->left,x,cutoff,atom,dist,maxn);
      if (x[node->dim]>=node->cut-cutoff) {
	n+=findBHcloseAtomsInNode(node->right,x,cutoff,atom+n,dist+n,maxn-n);
      }
    }else{
      n+=findBHcloseAtomsInNode(node->right,x,cutoff,atom,dist,maxn);
      if (x[node->dim]<node->cut+cutoff) {
	n+=findBHcloseAtomsInNode(node->left,x,cutoff,atom+n,dist+n,maxn-n);
      }
    }
  }
  return(n);
}
/*-------------------- markBHcloseAtomsInNode --------------------*/

static int markBHcloseAtomsInNode(BHnode *node,float *x,float cutoff,int flag)
{
  int j,n,mask;
  float d[3],D,C;

  if (!mvld(node)) return(0);
  if (node->n < 1) return(0);
  if (node->n==node->marked) return(0);
  if (flag<0||flag>7) return(0);
  
  if (node->dim<0) {
    n=0; mask=(1<<flag);
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      if (mask&&node->atom[j]->flg) continue;
      d[0]=x[0]-node->atom[j]->x[0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j]->x[1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j]->x[2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      n++;
      node->atom[j]->flg|=mask;
    }
  }else{
    n=0;
    if (x[node->dim]<node->cut+cutoff) {
      n+=markBHcloseAtomsInNode(node->left,x,cutoff,flag);
    }
    if (x[node->dim]>=node->cut-cutoff) {
      n+=markBHcloseAtomsInNode(node->right,x,cutoff,flag);
    }
  }
  node->marked+=n;
  return(n);
}
/*-------------------- nrfindBHcloseAtomsInNode --------------------*/

static int nrfindBHcloseAtomsInNode(BHnode *node,int idx,float cutoff,
				  Atom **atom,float *dist,int maxn)
{
  int j,n,idl,idr,jstart;
  float *x,d[3],D,C;

  if (!mvld(node)) return(0);
  if (maxn<0) return(0);
  if (node->n < 1) return(0);
  if (idx>=node->n) return(0);
  
  x=node->atom[idx]->x;
  if (node->dim<0) {
    n=0;
    C=cutoff*cutoff;
    jstart=(idx < 0 ? 0 : idx+1);
    for (j=jstart;j<node->n;j++) {
      d[0]=x[0]-node->atom[j]->x[0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j]->x[1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j]->x[2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      if (n<maxn) {
	atom[n]=node->atom[j];
	dist[n]=sqrt(D);
	n++;
      }else{
	n++;
	break;
      }
    }
  }else{
    n=0;
    if (x[node->dim]<node->cut+cutoff) {
      idl=idx;
      n+=nrfindBHcloseAtomsInNode(node->left,idl,cutoff,atom,dist,maxn);
    }
    if (x[node->dim]>=node->cut-cutoff) {
      idr=idx-node->left->n;
      n+=nrfindBHcloseAtomsInNode(node->right,idr,cutoff,atom+n,dist+n,maxn-n);
    }
  }
  return(n);
}
/*-------------------- findBHmindistInNode --------------------*/

static float findBHmindistInNode(BHnode *node,float *x,float cutoff)
{
  int j;
  float d[3],D,C;

  if (!mvld(node)) return(cutoff);
  if (node->n < 1) return(cutoff);
  
  if (node->dim<0) {
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      d[0]=x[0]-node->atom[j]->x[0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j]->x[1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j]->x[2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      cutoff=sqrt(D);
      C=D;
    }
  }else{
    if (x[node->dim]<node->cut) {
      cutoff=findBHmindistInNode(node->left,x,cutoff);
      if (x[node->dim]>=node->cut-cutoff) {
	cutoff=findBHmindistInNode(node->right,x,cutoff);
      }
    }else{
      cutoff=findBHmindistInNode(node->right,x,cutoff);
      if (x[node->dim]<node->cut+cutoff) {
	cutoff=findBHmindistInNode(node->left,x,cutoff);
      }
    }
  }
  return(cutoff);
}
/*-------------------- mark_close --------------------*/

int mark_close(int n1,Atom **A1,int n2,Atom **A2,float maxd)
{
  /* mark atoms (flg1) in A1 that are within maxd of A2,
     mark atoms outside with flg7.                           */

  BHtree *T1,*T2;
  int i,grain,status,k;
  float d;

  status=0;
  if (n1==0||n2==0||!mvld(A1)||!mvld(A2)) return(status);

  T1=(BHtree *)NULL;
  T2=(BHtree *)NULL;
  status=1;

  /* if A1 is a subset of A2, flag all atoms of A1 and return... */

  for (i=0;i<n2;i++) A2[i]->flg|=(1<<7);
  for (k=0,i=0;i<n1;i++) {
    if (A1[i]->flg&(1<<7)) {
      A1[i]->flg|=(1<<0);
      k++;
    }
  }
  for (i=0;i<n2;i++) A2[i]->flg&=~(1<<7);
  if (k==n1) return(status);
    
  grain = ( n1 < MXTREE ? 1 : n1 / MXTREE );
  T1=GenerateBHtree(n1,A1,grain);
  if (!mvld(T1)) {
    err_set(1,"mark_close: failed to generate BH tree(1).");
    status=0;
  }
    
  grain = ( n2 < MXTREE ? 1 : n2 / MXTREE );
  T2=GenerateBHtree(n2,A2,grain);
  if (!mvld(T2)) {
    err_set(1,"mark_close: failed to generate BH tree(2).");
    T1->atom=(Atom **)NULL; freeBHtree(T1);
    status=0;
  }

  if (status) {
    for (i=0;i<n1;i++) {
      if (A1[i]->flg&(1<<0|1<<7)) continue;
      d=findBHmindist(T2,A1[i]->x);
      if (d<maxd) {
	A1[i]->flg|=(1<<0);
	if (maxd-d > 1.0) markBHcloseAtoms(T1,A1[i]->x,maxd-d,0);
      }else{
	A1[i]->flg|=(1<<7);
	if (d-maxd > 1.0) markBHcloseAtoms(T1,A1[i]->x,d-maxd,7);
      }
    }
  }

  if (mvld(T1)) {
    T1->atom=(Atom **)NULL;
    freeBHtree(T1);
  }
  if (mvld(T2)) {
    T2->atom=(Atom **)NULL;
    freeBHtree(T2);
  }

  return(status);
}
/*-------------------- set_bh_cut_method --------------------*/

int set_bh_cut_method(int m)
{
  int r;

  r=CUT;
  
  switch (m) {
  case BH_CUT_MIDDLE:
  case BH_CUT_MEDIAN:
    CUT=m;
    break;
  default:
    break;
  }

  return(r);
}
