#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "object.h"
#include "xtal.h"
#include "gcalc.h"
#include "transpor.h"
#include "math.h"
#include "bhtree.h"
#include "errhand.h"

extern int graphix;

extern Node *DefaultNode,*MolRoot;

/* ==================== Crystallography module ==================== */

static char *kwdlist = "box invert neighbours packing spacegroup symcopy xtrans symop end ' ' "
                       "command bond color cpk delete display label measure modify "
                       "reset select set split";

static char *boxkwd = "on off color";

static char *pr_menu = "around -2.0 +2.0 -1.5 +1.5 -1.0 +1.0 -0.5 +0.5 0.0";
static char *pr_prompt[3][2] = {
  {"x_min = ","x_max = "},
  {"y_min = ","y_max = "},
  {"z_min = ","z_max = "}
};

static float cf[8][3] = {
  {0.0,0.0,0.0}, {0.0,0.0,1.0}, {0.0,1.0,0.0}, {0.0,1.0,1.0},
  {1.0,0.0,0.0}, {1.0,0.0,1.0}, {1.0,1.0,0.0}, {1.0,1.0,1.0}
};

static int vect[12][2] = {
  {0,1}, {1,3}, {3,2}, {2,0},
  {4,5}, {5,7}, {7,6}, {6,4},
  {0,4}, {5,1}, {3,7}, {6,2}
};

#define DEFAULT_BOX_COLOR 2
#define MXTREE 1024

#define PACK_ERROR 0
#define PACK_RANGE 1

#define key_is(x) (streq(key->key,x))


/*-------------------- prototypes --------------------*/

void crys_box(Node *root,Xcn *xcn);
void crys_xtrans(Node *root,Xcn *xcn);
void crys_symop(Node *root,Xcn *xcnnode);
static void crys_invert(Node *root,Xcn *xcn);
static void crys_symcopy(Node *root,Xcn *xcnnode);
static void crys_packing(Node *root,Xcn *xcnnode);
static void crys_spacegroup(Node *root,Xcn *xcnnode);
static char *crys_symnode_name(char *name,int op,vec_3 dx);
static void crys_neighbours(Node *root,Xcn *xcnnode);
static int sanity_check(Node *root,Xcn *xcn);

/* ============================================================ */

void cmd_crystal(Xcn *xcn)
{
  Token *key;
  int status;
  Nodechain c;
  Node *root,*save_default;
  char *p,menu[4096];

  key=(Token *)NULL;

  /* get root node and XCN */

  status=1;
  if (is_xcn(xcn)) {
    root=xcn->up;
    if (!is_mol(root)) {
      err_set(1,"%s is not a molecule.",root->name);
      return;
    }
  }else{
    root=(Node *)NULL;
    c.n=0; c.s=(Node *)NULL; c.e=c.s; c.p=c.s;
    ForXcns(xcn,MolRoot) add_to_node_chain(xcn->up,&c);
    if (c.n==0) {
      err_set(1,"no molecule found with crystallographic info.");
      return;
    }
    *menu='\0';
    for (p=menu,root=c.s;mvld(root);root=root->chain) {
      *p++ = ' ';
      nodeptr_to_name(p,root);
      p+=strlen(p);
    }
    clear_node_chain(&c);
    key=Get_Mol_M("molecule: ",";",menu);
    if (key->status==TOK_OK) {
      for (root=(Node *)key->ptr,xcn=(Xcn *)NULL;root;root=root->up) {
	if ((xcn=first_xcn(root))) break;
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"molecule %s not found",key->text);
      status=0;
    }else status=0;
  }

  if (is_mol(root)) {
    if (!is_xcn(xcn)) {
      err_set(1,"X-tal info not found in %s",key->text);
      status=0;
    }
  }else{
    err_set(1,"X-tal info not found in %s",key->text);
    status=0;
  }

  key=free_Token(key);
  save_default=DefaultNode;
  if (status) DefaultNode=root;
    
  /* start of outer command loop */

  while(status) {
    send_header(graphix,"ARMK",0);
    start_parse();
    key=Get_Kwd("Crystal> ",";",kwdlist);
    if (key->status==TOK_OK) {
      if      (key_is("box"))         crys_box(root,xcn);
      else if (key_is("xtrans"))      crys_xtrans(root,xcn);
      else if (key_is("symop"))       crys_symop(root,xcn);
      else if (key_is("spacegroup"))  crys_spacegroup(root,xcn);
      else if (key_is("symcopy"))     crys_symcopy(root,xcn);
      else if (key_is("packing"))     crys_packing(root,xcn);
      else if (key_is("neighbours"))  crys_neighbours(root,xcn);
      else if (key_is("invert"))      crys_invert(root,xcn);
      else if (key_is("end"))         status=0;
      else{
	if (key_is("command"))        {
	  command_loop("Wit!P(Crystal)> ");
	  if (!sanity_check(root,xcn)) {
	    err_set(1,"lost my marbles, xcn and root !");
	    status=0;
	  }
	}
	else if (key_is("bond"))      cmd_bond();     
	else if (key_is("color"))     cmd_color();
	else if (key_is("cpk"))       cmd_cpk();
	else if (key_is("delete"))    {
	  cmd_delete();
	  if (!sanity_check(root,xcn)) {
	    err_set(1,"lost my marbles, xcn and root !");
	    status=0;
	  }
	}
	else if (key_is("display"))   cmd_display();
	else if (key_is("label"))     cmd_label();
	else if (key_is("measure"))   cmd_measure();
	else if (key_is("modify"))    cmd_modify();
	else if (key_is("reset"))     cmd_display_reset();
	else if (key_is("select"))    cmd_select();
	else if (key_is("set"))       cmd_set();
	else if (key_is("split"))     {
	  cmd_split();
	  if (!sanity_check(root,xcn)) {
	    err_set(1,"lost my marbles, xcn and root !");
	    status=0;
	  }
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"unknown option: %s",key->text);
      end_parse();
    }
    free_Token(key);
    if (end_parse()!=TOK_OK) {
      err_set(1,"extra characters at end of command ignored.");
    }
    err_dump();
  }
  send_header(graphix,"ARMK",0);
  if (find_node(MolRoot,save_default,OBJ_MOL|OBJ_ROOT)) DefaultNode=save_default;
  else DefaultNode=MolRoot;
}
/*-------------------- crys_box --------------------*/

void crys_box(Node *root,Xcn *xcn)
{
  Token *key;
  int color,remake;

  if (!is_mol(root) || !is_xcn(xcn)) {
    err_set(1,"internal error: bad arg in crys_box");
    return;
  }
  
  key=Get_Kwd("box option: ",";",boxkwd);
  if (key->status==TOK_OK) {
    if (key_is("on")) {
      if (!is_dsp(xcn->boxdsp)) {
	if (xcn->box_color==0) xcn->box_color=DEFAULT_BOX_COLOR;
	xcn->boxdsp=crys_make_box_dsp(root,xcn);
      }
    }else if (key_is("off")) {
      if (is_dsp(xcn->boxdsp)) {
	del_node((Node *)xcn->boxdsp);
	xcn->boxdsp=(Dspobj *)NULL;
      }
    }else if (key_is("color")) {
      free_Token(key);
      key=Get_Color("color: ",";");
      if (key->status == TOK_OK) {
	color=key->n;
	remake=color!=xcn->box_color;
	xcn->box_color=color;
	if (remake&&is_dsp(xcn->boxdsp)) {
	  lock_evtq();
	  del_node((Node *)xcn->boxdsp);
	  xcn->boxdsp=crys_make_box_dsp(root,xcn);
	  unlock_evtq();
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"unknown box color: %s",key->text);
      }
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"unknown box option: %s",key->text);
  }
  free_Token(key);
}

/*-------------------- crys_make_box_dsp --------------------*/

Dspobj *crys_make_box_dsp(Node *root,Xcn *xcn)
{
  int i,size;
  float cc[8][3];
  Dspobj *global,*box;
  
  box=(Dspobj *)NULL;
  if (is_mol(root) && is_xcn(xcn)) {
    global=glblobj(assocobj(root));
    if (is_dsp(global)) {
      box=new_dsp_obj(global,(char *)NULL,DSP_CSO);
      if (mvld(box)) {
	for (i=0;i<8;i++) v32arr(cc[i],m4xv3(xcn->fta,vec3(cf[i])));
	size=sizeof(box)+sizeof(int)+12*6*sizeof(float);
	send_bheader(graphix,"AMDS",size);
	send_bptr(graphix,box);
	send_bint(graphix,xcn->box_color);
	for (i=0;i<12;i++) {
	  send_bfloata(graphix,3,cc[vect[i][0]]);
	  send_bfloata(graphix,3,cc[vect[i][1]]);
	}
	
	size=sizeof(box)+sizeof(int)+3*sizeof(float)+1;
	
	send_bheader(graphix,"ATOB",size);
	send_bptr(graphix,box);
	send_bint(graphix,xcn->box_color);
	send_bfloata(graphix,3,cc[0]);
	send_bdata(graphix,1,"0");
	
	send_bheader(graphix,"ATOB",size);
	send_bptr(graphix,box);
	send_bint(graphix,xcn->box_color);
	send_bfloata(graphix,3,cc[4]);
	send_bdata(graphix,1,"A");
	
	send_bheader(graphix,"ATOB",size);
	send_bptr(graphix,box);
	send_bint(graphix,xcn->box_color);
	send_bfloata(graphix,3,cc[2]);
	send_bdata(graphix,1,"B");
	
	send_bheader(graphix,"ATOB",size);
	send_bptr(graphix,box);
	send_bint(graphix,xcn->box_color);
	send_bfloata(graphix,3,cc[1]);
	send_bdata(graphix,1,"C");

      }
    }
  }
  return(box);
}
/*-------------------- crys_xtrans --------------------*/

void crys_xtrans(Node *root,Xcn *xcn)
{
  Nodechain *atoms;
  Atom *a;
  Token *tok;
  vec_3 dx,org;
  int i,k,rm,size;
  float *x;
  char buff[256];

  if (!is_mol(root) || !is_xcn(xcn)) return;

  /* get translation vector */

  for (i=0;i<3;i++) {
    org.c[i]=0.0;
    sprintf(buff,"d%c = ",'x'+i);
    default_token("0");
    tok=Get_Float_M(buff,";","-4 -3 -2 -1 0 1 2 3 4");
    if (tok->status==TOK_OK) {
      dx.c[i]=tok->f;
      tok=free_Token(tok);
    }else{
      if (tok->status==TOK_NOK) {
	err_set(1,"bad floating point token: %s",tok->text);
      }
      tok=free_Token(tok);
      break;
    }
  }
  if (i!=3) return;

  /* get atoms */

  tok=Get_Atoms_D("atoms: ",";");
  if (tok->status!=TOK_OK) {
    if (tok->status==TOK_NOK) {
      err_set(1,"bad atom specification.");
    }
    tok=free_Token(tok);
    send_header(graphix,"ARMK",0);
    return;
  }
  atoms=(Nodechain *)tok->ptr;
  tok=free_Token(tok);
  rm=chain_reduce_subtree(root,atoms);
  if (rm>0) {
    err_set(1,"%d atoms not in %s ignored.",rm,nodeptr_to_name(buff,root));
  }

  dx=v3mv3(Axyz(xcn,dx),Axyz(xcn,org));

  lock_evtq();
  send_bheader(graphix,"ARMK",0);
  if (atoms->n > 0) {
    size=atoms->n*(cPtrSize()+3*sizeof(int));
    send_bheader(graphix,"NEWX",size);
    for (a=(Atom *)atoms->s;a;a=(Atom *)a->chain) {
      for (x=a->x,k=0;k<a->nxyz;k++,x+=ATM_LOCSZ) {
	for (i=0;i<3;i++) x[i]+=dx.c[i];
      }
      send_bcptr(graphix,a->cross);
      for (i=0;i<3;i++) send_bfloat(graphix,a->x[i]);
    }
  }
  unlock_evtq();
  prtGraphics("%d atoms moved.\n",atoms->n);
  clear_node_chain(atoms);
  free_node_chain(atoms);
}
/*-------------------- crys_symop --------------------*/

void crys_symop(Node *root,Xcn *xcn)
{
  Token *tok;
  Atom *a;
  Xatom *ax;
  Symop sx;
  Nodechain *atoms;
  char buff[4096],name[10];
  int i,rm,op,size;
  vec_3 zt={{0,0,0}};

  if (!is_mol(root) || !is_xcn(xcn)) return;

  /* get symm-op */

  *buff='\0';
  for (i=0;i<xcn->nsymx;i++) 
    sprintf(buff+strlen(buff),"'%s' ",crys_symnode_name(name,i,zt));
  tok=Get_Kwd("symop#: ",";",buff);
  if (tok->status!=TOK_OK) {
    if (tok->status==TOK_NOK) {
      err_set(1,"illegal syop# : %s",tok->text);
    }
    free_Token(tok);
    return;
  }
  op=tok->n;
  free_Token(tok);

  /* get atoms */

  tok=Get_Atoms_D("atoms: ",";");
  if (tok->status!=TOK_OK) {
    if (tok->status==TOK_NOK) {
      err_set(1,"bad atom specification.");
    }
    free(tok);
    send_header(graphix,"ARMK",0);
    return;
  }
  atoms=(Nodechain *)tok->ptr;
  free_Token(tok);
  rm=chain_reduce_subtree(root,atoms);
  if (rm>0) {
    err_set(1,"%d atoms not in %s ignored.",rm,nodeptr_to_name(buff,root));
  }

  /* apply symm-op to coordinates */

  lock_evtq();
  send_bheader(graphix,"ARMK",0);
  if (atoms->n > 0) {
    size=atoms->n*(cPtrSize()+3*sizeof(int));
    send_bheader(graphix,"NEWX",size);
    for (a=(Atom *)atoms->s;a;a=(Atom *)a->chain) {
      send_bcptr(graphix,a->cross);
      transform_atom(xcn->sym[op],a);
      for (i=0;i<3;i++) send_bfloat(graphix,a->x[i]);
    }
  }
  unlock_evtq();
  prtGraphics("%d atoms moved.\n",atoms->n);

  /* transform bfactors */

  if (atoms->n>0) {
    sx=symF(xcn,xcn->sym[op]);
    for (a=(Atom *)atoms->s;a;a=(Atom *)a->chain) {
      ax=first_xatom(a);
      if (mvld(ax)) {
	if (!ax->iso) {
	  RotTfact(sx,ax->bfac);
	}
      }
    }
  }

  clear_node_chain(atoms);
  free_node_chain(atoms);
}
/*-------------------- crys_spacegroup --------------------*/

static void crys_spacegroup(Node *root,Xcn *xcn)
{
  Token *tok=NULL;

  if (!is_mol(root) || !is_xcn(xcn)) return;

  tok=Get_String("new spacegroup: ",";");
  if (tok->status==TOK_OK) {
    read_spgrp(xcn,tok->text);
  }else if (tok->status==TOK_NOK) {
    err_set(1,"invalid token: %s",tok->text);
  }
  free_Token(tok);
  return;
}

/*-------------------- crys_invert --------------------*/

static void crys_invert(Node *root,Xcn *xcn)
{
  Token *tok;
  Atom *a;
  Nodechain *atoms;
  char buff[4096];
  int i,j,rm,size;
  Symop sym;

  if (!is_mol(root) || !is_xcn(xcn)) return;

  /* get atoms */

  tok=Get_Atoms_D("atoms: ",";");
  if (tok->status!=TOK_OK) {
    if (tok->status==TOK_NOK) {
      err_set(1,"bad atom specification.");
    }
    free(tok);
    send_header(graphix,"ARMK",0);
    return;
  }
  atoms=(Nodechain *)tok->ptr;
  free_Token(tok);
  rm=chain_reduce_subtree(root,atoms);
  if (rm>0) {
    err_set(1,"%d atoms not in %s ignored.",rm,nodeptr_to_name(buff,root));
  }

  /* invert */

  for (i=0;i<4;i++) for(j=0;j<4;j++) sym.r[i].c[j]=0.0;
  for (i=0;i<3;i++) sym.r[i].c[i]=-1.0;
  sym.r[3].c[3]=1.0;
  sym=symA(xcn,sym);

  lock_evtq();
  send_bheader(graphix,"ARMK",0);
  if (atoms->n > 0) {
    size=atoms->n*(cPtrSize()+3*sizeof(int));
    send_bheader(graphix,"NEWX",size);
    for (a=(Atom *)atoms->s;a;a=(Atom *)a->chain) {
      send_bcptr(graphix,a->cross);
      transform_atom(sym,a);
      for (i=0;i<3;i++) send_bfloat(graphix,a->x[i]);
    }
  }
  unlock_evtq();
  prtGraphics("%d atoms inverted.\n",atoms->n);

  clear_node_chain(atoms);
  free_node_chain(atoms);
}
/*-------------------- crys_symcopy --------------------*/

static void crys_symcopy(Node *root,Xcn *xcn)
{
  /* generate symmetry related copy of asymmetric unit */

  Token *tok;
  Node *src,*trg;
  Atom *a;
  Xatom *ax;
  Nodechain atoms;
  Symop sx,st;
  char buff[4096],name[1024],c;
  Dspobj *mdl;
  int i,op=0,id;
  vec_3 dx,org;
  vec_3 zt={{0,0,0}};

  if (!is_mol(root) || !is_xcn(xcn)) return;

  /* get source unit (at present, this is always the first unit) */

  src=first_type(root,OBJ_MOL);
  if (!is_mol(src)) {
    err_set(1,"asymmetric unit not found");
    return;
  }

  /* get symm-op */

  *buff='\0';
  for (i=0;i<xcn->nsymx;i++) 
    sprintf(buff+strlen(buff),"'%s' ",crys_symnode_name(name,i,zt));
  tok=Get_String("symop_transl: ","_;");
  if (tok->status==TOK_OK) {
    if (strlen(tok->text)==1) {
      c=tok->text[0];
      if ('a'<=c && c<='z') c+='A'-'a';
      op=c-'A';
      if (op<0 || op>=xcn->nsymx) tok->status=TOK_NOK;
    }else tok->status=TOK_NOK;
  }
  if (tok->status!=TOK_OK) {
    if (tok->status==TOK_NOK) {
      err_set(1,"illegal symop-code : %s",tok->text);
    }
    free_Token(tok);
    return;
  }

  /* get translation vector */

  for (i=0;i<3;i++) {
    org.c[i]=0.0;
    dx.c[i]=0.0;
  }

  if (tok->term=='_') {
    free_Token(tok);
    tok=Get_String("translation code: ",";");
    if (tok->status==TOK_OK) {
      if (strlen(tok->text)==3) {
	for (i=0;i<3;i++) {
	  id=tok->text[i]-'0';
	  if (0 <= id && id <= 9) {
	    if (id > 5 ) id-=10;
	    dx.c[i]=id;
	  }else tok->status=TOK_NOK;
	}
      }else tok->status=TOK_NOK;
    }
    if (tok->status==TOK_NOK) {
      err_set(1,"bad translation code : %s",tok->text);
    }
    if (tok->status!=TOK_OK) {
      free_Token(tok);
      return;
    }
  }

  crys_symnode_name(name,op,dx);
  trg=find_name(src->up,name,OBJ_MOL);
  if (mvld(trg)) {
    err_set(1,"%s already exists.",nodeptr_to_name(buff,trg));
    free_Token(tok);
    return;
  }

  /* generate copy */

  trg=tree_copy(src);
  trg->up=src->up;
  link_node(trg);
  copy_set_bits(src,trg);
  set_node_name(trg,name);

  /* apply symm-op to coordinates in target node */

  sx=xcn->sym[op]; 
  st=symF(xcn,xcn->sym[op]);
  dx=v3mv3(Axyz(xcn,dx),Axyz(xcn,org));
  for (i=0;i<3;i++) sx.r[i].c[3]+=dx.c[i];

  ForAtoms(a,trg) {
    transform_atom(sx,a);
    ax=first_xatom(a);
    if (mvld(ax)) {
      if (!ax->iso) {
	RotTfact(st,ax->bfac);
      }
    }
  }

  mdl=assocmdl(src);
  atoms=tree2chain(trg,OBJ_ATOM);
  mdl_add_atom_chain(mdl,&atoms);
  clear_node_chain(&atoms);
    
  free_Token(tok);
}
/*-------------------- crys_packing --------------------*/

static void crys_packing(Node *root,Xcn *xcn)
{
  Token *tok=(Token *)NULL;
  int i,j,k,n,done,s,mode;
  Node *src,*trg,*new;
  Nodechain atoms,*Ac;
  Dspobj *mdl;
  Atom *a,*ac;
  Xatom *ax;
  Symop Sx,Sc;
  vec_3 com,scom,dx,dx0,dc,org,offplus,offminus,pt;
  float pr[3][2],sw;
  char name[10];

  if (!is_mol(root) || !is_xcn(xcn)) return;

  src=first_type(root,OBJ_MOL);
  if (!is_mol(src)) {
    err_set(1,"asymmetric unit not found");
    return;
  }

  /* determine center of asymmetric unit */

  n=0;
  for (j=0;j<3;j++) com.c[j]=0.0;
  ForAtoms(a,src) {
    n++;
    for (j=0;j<3;j++) com.c[j]+=a->x[j];
  }
  if (n==0) {
    err_set(1,"No atoms in asymmetric unit.");
    return;
  }else{
    for (j=0;j<3;j++)  {
      com.c[j]/=n;
      org.c[j]=0.0;
    }
    com=Fxyz(xcn,com);
    org=Axyz(xcn,org);
  }

  /* get packing range in fractional coordinates */

  mode=PACK_RANGE; Ac=(Nodechain *)NULL;
  for (j=0;j<3&&mode;j++) {
    for (k=0;k<2&&mode;k++) {
      err_dump();
      tok=Get_Float_M(pr_prompt[j][k],";",pr_menu);
      if (tok->status==TOK_OK) {
	pr[j][k]=tok->f;
	tok=free_Token(tok);
      }else if (tok->status==TOK_NOK) {
	tok=Prs_Kwd(tok,"around");
	if (tok->status==TOK_OK) {
	  if (streq(tok->key,"around")) {
	    tok=free_Token(tok);
	    if (mvld(Ac)) {
	      clear_node_chain(Ac);
	      free_node_chain(Ac);
	      Ac=(Nodechain *)NULL;
	    }
	    tok=Get_Atoms_D("atoms at center: ",";");
	    send_bheader(graphix,"ARMK",0);
	    if (tok->status==TOK_OK) {
	      Ac=(Nodechain *)tok->ptr;
	      tok=free_Token(tok);
	    }else{
	      if (tok->status==TOK_NOK) err_set(1,"bad selection: %s",tok->text);
	      tok=free_Token(tok);
	    }
	    k=k-1;
	  }
	}else{
	  err_set(1,"bad float: %s",tok->text);
	  tok=free_Token(tok);
	  k=k-1;
	}
      }else{
	tok=free_Token(tok);
	clear_node_chain(Ac); 
	free_node_chain(Ac); 
	Ac=(Nodechain *)NULL;
	mode=PACK_ERROR;
      }
    }
    if (pr[j][0]>pr[j][1]) {
      sw=pr[j][0]; pr[j][0]=pr[j][1]; pr[j][1]=sw;
    }
  }

  if (mvld(Ac)) {
    for (j=0;j<3;j++) offplus.c[j]=0.0;
    offminus=offplus;
    for(ac=(Atom *)Ac->s;mvld(ac);ac=(Atom *)ac->chain) {
      if (ac->nxyz>0) {
	offplus=Fxyz(xcn,vec3(ac->x));
	offminus=offplus;
	break;
      }
    }
    if (mvld(ac)) {
      for(ac=(Atom *)ac->chain;mvld(ac);ac=(Atom *)ac->chain) {
	if (ac->nxyz>0) {
	  pt=Fxyz(xcn,vec3(ac->x));
	  for (j=0;j<3;j++) {
	    offplus.c[j]=(offplus.c[j] < pt.c[j] ? pt.c[j] : offplus.c[j]);
	    offminus.c[j]=(offminus.c[j] < pt.c[j] ? offminus.c[j] : pt.c[j]);
	  }
	}
      }
    }
    for (j=0;j<3;j++) {
      pr[j][0]+=offminus.c[j];
      pr[j][1]+=offplus.c[j];
    }
    clear_node_chain(Ac); free_node_chain(Ac); Ac=(Nodechain *)NULL;
  }
	
  tok=free_Token(tok);
  if (mode) {
    for (j=0;j<3;j++) {
      if (pr[j][0]<-4) pr[j][0]=-4;
      if (pr[j][0]>5) pr[j][0]=5;
      if (pr[j][1]<-4) pr[j][1]=-4;
      if (pr[j][1]>5) pr[j][1]=5;
    }
    
    /* Generate copies as needed */
    
    new=make_node(OBJ_MOL);                      /* node used to collect copies */
    if (!mvld(new)) {
      err_set(1,"memory full.");
      return;
    }
    
    switch (mode) {
    case PACK_RANGE:
      
      done=FALSE;
      for (s=0;s<xcn->nsymx&&!done;s++) {
	Sx=symF(xcn,xcn->sym[s]);
	scom=symdo(Sx,com);
	for (j=0;j<3;j++) {
	  k=pr[j][0]-scom.c[j];
	  if (scom.c[j]+k<pr[j][0]) k++;
	  dx0.c[j]=k;
	}
	dx=dx0;
	while(scom.c[0]+dx.c[0]<=pr[0][1]&&!done) {
	  while(scom.c[1]+dx.c[1]<=pr[1][1]&&!done) {
	    while(scom.c[2]+dx.c[2]<=pr[2][1]&&!done) {
	      crys_symnode_name(name,s,dx);
	      trg=find_name(src->up,name,OBJ_MOL);
	      if (!mvld(trg)) {
		trg=tree_copy(src);
		if (mvld(trg)) {
		  trg->up=new;
		  link_node(trg);
		  set_node_name(trg,name);
		  Sc=xcn->sym[s]; 
		  dc=v3mv3(Axyz(xcn,dx),org);
		  for (i=0;i<3;i++) Sc.r[i].c[3]+=dc.c[i];
		  ForAtoms(a,trg) {
		    transform_atom(Sc,a);
		    ax=first_xatom(a);
		    if (mvld(ax)) {
		      if (!ax->iso) {
			RotTfact(Sx,ax->bfac);
		      }
		    }
		  }
		}else done=TRUE;
	      }
	      dx.c[2]+=1.0;
	    }
	    dx.c[2]=dx0.c[2];
	    dx.c[1]+=1.0;
	  }
	  dx.c[1]=dx0.c[1];
	  dx.c[0]+=1.0;
	}
      }
      break;
      
    }
    
    /* send new nodes to graphics and link with proper root */
    
    mdl=assocmdl(src);
    atoms=tree2chain(new,OBJ_ATOM);
    mdl_add_atom_chain(mdl,&atoms);
    clear_node_chain(&atoms);
    
    while (mvld(new->fst)) {
      trg=new->fst;
      unlink_node(trg);
      trg->up=src->up;
      link_node(trg);
      copy_set_bits(src,trg);
    }
    del_node(new);
  }
}
/*-------------------- crys_neighbours --------------------*/

static void crys_neighbours(Node *root,Xcn *xcn)
{
  Nodechain c,t;
  Atom **source,**target;
  Atom *a;
  Xatom *ax;
  int i,j,k,n,N,M,NT;
  float d,dmin;
  Symop S,s,Sx;
  Token *key;

  key=Get_Atoms_D("source atoms: ",";");
  if (key->status==TOK_OK) {
    c=*(Nodechain *)key->ptr;
    free(key->ptr);
    chain_reduce_subtree(root,&c);
    if (c.n<=0) {
      err_set(1,"no atoms");
      key->status=TOK_ABORT;
    }
  }
  if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
    end_parse();
  }
  if (key->status!=TOK_OK) {
    free_Token(key);
    return;
  }

  send_header(graphix,"ARMK",0);
  free_Token(key);

  N=c.n;
  t.n=0; t.s=(Node *)NULL; t.e=t.s;
  connected_atoms(&c);
  M=c.n;
  source=(Atom **)malloc(M*sizeof(Atom *));
  if (!mvld(source)) {
    err_set(1,"memory allocation failure");
    return;
  }
  for (n=0,a=(Atom *)c.s;mvld(a);a=(Atom *)c.s) {
    source[n++]=a;
    c.s=a->chain;
    a->chain=(Node *)NULL;
  }

  key=Get_Atoms_D("target atoms: ",";");
  if (key->status==TOK_OK) {
    c=*(Nodechain *)key->ptr;
    chain_reduce_subtree(root,&c);
    if (c.n<=0) {
      err_set(1,"no target atoms.");
      key->status=TOK_ABORT;
    }
    free(key->ptr);
  }
  if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
    end_parse();
  }
  if (key->status!=TOK_OK) {
    free_Token(key);
    free(source);
    return;
  }

  send_header(graphix,"ARMK",0);
  free_Token(key);

  for (i=0;i<M;i++) source[i]->flg|=(node_in_chain((Node *)source[i],&c)<<1);
  for (a=(Atom *)c.s,c.n=0;mvld(a);a=(Atom *)c.p) {
    c.p=(Node *)a->chain;
    a->chain=(Node *)NULL;
    if (a->flg&(1<<1)) {
      a->flg&=~(1<<1);
    }else{
      if (c.n==0) c.s=(Node *)a;
      else c.e->chain=(Node *)a;
      c.e=(Node *)a;
      c.n++;
    }
  }
  
  NT=c.n;
  if (NT<=0) {
    free(source);
    return;
  }
  target=(Atom **)malloc(NT*sizeof(Atom *));
  if (!mvld(target)) {
    err_set(1,"memory allocation failure");
    free(source);
    return;
  }
  for (n=0,a=(Atom *)c.s;mvld(a);a=(Atom *)c.s) {
    target[n++]=a;
    c.s=a->chain;
    a->chain=(Node *)NULL;
  }

  for (i=0;i<N;) {
    c.n=1;
    c.s=(Node *)source[i];
    c.e=c.s;
    connected_atoms(&c);
    dmin=1.0e20;
    S=xcn->sym[0];
    for (j=i;j<N;j++) {
      if (node_in_chain((Node *)source[j],&c)) {
	for (k=0;k<NT;k++) {
	  if (!node_in_chain((Node *)target[k],&c)) {
	    s=symClosest(xcn,vec3(target[k]->x),vec3(source[j]->x),&d);
	    if (d<dmin) {
	      dmin=d;
	      S=s;
	    }
	  }
	}
      }
    }
    if (dmin<1.e20) {
      Sx=symF(xcn,S);
      for (a=(Atom *)c.s;mvld(a);a=(Atom *)a->chain) {
	transform_atom(S,a);
	ax=first_xatom(a);
	if (mvld(ax)) {
	  if (!ax->iso) {
	    RotTfact(Sx,ax->bfac);
	  }
	}
	if (!mvld(t.e)) {
	  t.n=0;
	  t.s=(Node *)a;
	}else{
	  t.e->chain=(Node *)a;
	}
	t.e=(Node *)a;
	t.n++;
      }
    }
    c.e->chain=(Node *)c.e;
    for (k=0;k<NT;k++) target[k]->chain=(Node *)NULL;
    for (i++;i<N;i++) if (!mvld(source[i]->chain)) break;
  }

  if (t.n>0) {
    t.e->chain=(Node *)NULL;
    for (n=0,a=(Atom *)t.s;mvld(a);a=(Atom *)a->chain) {
      if (cPtr(a->cross)) n++;
    }
    if (n>0) {
      n*=(cPtrSize()+3*sizeof(int));
      send_bheader(graphix,"NEWX",n);
      for (a=(Atom *)t.s;mvld(a);a=(Atom *)a->chain) {
	send_bcptr(graphix,a->cross);
	send_bfloata(graphix,3,a->x);
      }
      trflush(graphix);
    }
  }

  for (i=0;i<M;i++) source[i]->chain=(Node *)NULL;
  free(source);
  free(target);
}

/*-------------------- crys_symnode_name --------------------*/

static char *crys_symnode_name(char *name,int op,vec_3 d)
{
  int i,id,mod,maj;
  char *p;

  /* return systematic name for symmetry/translation */

  mod='Z'-'A'+1;
  maj=op/mod;
  p=name;
  if (maj>0) *p++='A'+maj-1;
  *p++='A'+(op%mod);
  *p++='_';
  for (i=0;i<3;i++) {
    id=rint(d.c[i]);
    if (id<0) id+=10;
    *p++='0'+id;
  }
  *p++='\0';
  if (streq(name+2,"000")) name[1]='\0';
  if (streq(name+3,"000")) name[2]='\0';
  return(name);
}  
/*-------------------- sanity_check --------------------*/

static int sanity_check(Node *root,Xcn *xcn)
{
  int r;
  
  if (find_node(MolRoot,root,OBJ_MOL)) {
    if (find_node(root,xcn,OBJ_XCN)) r=1;
    else r=0;
  }else r=0;
  return(r);
}
