/*
  general small molecule support for CYANA
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "mystdlib.h"
#include "swdef.h"
#include "gcalc.h"
#include "errhand.h"
#include "object.h"
#include "parser.h"
#include "cyana.h"
#include "strutil.h"
#include "transpor.h"
#include "topsym.h"

#define TRUE  1
#define FALSE 0

#define LOOP 1
#define DONE 2

#define FLG1      (1<<0)

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(x) (streq(key->key,x))

cyaOptions_t Options = {
  TRUE,                      /* generate overlap atoms in residue dictionary ? */
  TRUE,                      /* use atom names, rather than number in residue dictionary ? */
  4,                         /* rings with less atoms are rigid */
  TRUE,                      /* small bridged rings are rigid (c.f. cyaMarkRigidBonds) */
  TRUE,                      /* add pseudo atoms */
  TRUE,                      /* output atom types in all-uppercase */
  FALSE,                     /* output atom names in all-uppercase */
};

extern int graphix;

/*========================================================================*/

static Node *ligRoot(int *status)
{
  Token *key;
  Node *root=(Node *)NULL;
  
  key=Get_Mol_D("molecule: ",";");
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) root=(Node *)key->ptr;
  else if (key->status==TOK_NOK) err_set(1,"invalid molecule: %s\n",key->text);
  *status=key->status;

  key=free_Token(key);
  return(root);
}

static cyaAtom_t *treeRoot(cyaResidue_t *L)
{
  int i;
  Atom *r=NULL;
  Token *key=NULL;
  cyaAtom_t *R=NULL;
  char buff[4096],buff2[4096];
  
  key=Get_Atom("atom at root of residue tree: ",":");
  if (key->status==TOK_OK) {
    r=(Atom *)key->ptr;
    clear_atom_mark(r);
  }else if (key->status==TOK_NOK) {
    for (i=0;i<L->nat;i++) if (strceq(key->text,L->A[i]->name)) break;
    if (i<L->nat) r=L->A[i]->atm;
    else err_set(1,"bad residue tree root selection: %s",key->text);
  }
  if (r) {
    for (i=0;i<L->nat;i++) if (r==L->A[i]->atm) break;
    if (i<L->nat) R=L->A[i];
    else err_set(1,"%s is not in %s",
		 nodeptr_to_name(buff,(Node *)r),
		 nodeptr_to_name(buff2,L->root));
  }
  free_Token(key);
  return R;
}

static void reportCurrentResidue(cyaResidue_t *L)
{
  if (L) {
    prtGraphics("residue %s, %d atoms, root at %s\n",L->root->name,L->nat,L->A[0]->name);
  }
}

/*========================================================================*/

void cmd_cyana(void)
{
  Token *key=(Token *)NULL,*filename=(Token *)NULL;
  Node *root=NULL,*newR;
  int state,status;
  char *keys="bridges code molecule overlap rings root style uppercase write ' ' end";
  char *code=(char *)NULL,buff[1024];
  cyaResidue_t *L=NULL,*newL;
  cyaAtom_t *A;
  FILE *out;

  /*
  root=ligRoot(&status);
  if (root) {
    L=cyanaResidue(root,Options);
    if (L) reportCurrentResidue(L);
  }
  state=L? LOOP : DONE;
  */
  state=LOOP;

  while (state!=DONE) {
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("Cyana> ",";",keys);
    if (key->status==TOK_OK) {

      if (key_is("bridges")) {
	default_token(Options.bridgedRings?"yes":"no");
	key=free_Token(key);
	key=Get_Kwd("small bridged rings are rigid: ",";",Options.bridgedRings?"h!yes no":"yes h!no");
	if (key->status==TOK_OK) {
	  if ((Options.bridgedRings && key_is("no")) || (!Options.bridgedRings && key_is("yes"))) {
	    Options.bridgedRings=key_is("yes");
	    cyaMarkRigidBonds(L,Options);
	    cyaFindTorsions(L);
	    reportCurrentResidue(L);
	  }
	}else if (key->status==TOK_NOK) err_set(1,"invalid response: %s",key->text);
	
      }else if (key_is("code")) {
	default_token(code?code:"LI1");
	key=free_Token(key);
	key=Get_String_M("residue code: ",";","LI1 LI2 LI3 LI4 LI5");
	if (key->status==TOK_OK) {
	  if (code) free(code);
	  code=key->text;
	  key->text=(char *)NULL;
	}

      }else if (key_is("molecule")) {
	newL=NULL;
	newR=ligRoot(&status);
	if (newR && newR!=root) {
	  newL=cyanaResidue(newR,Options);
	  if (newL) {
	    freeCyaResidue(L);
	    L=newL;
	  }
	}
	err_dump();
	reportCurrentResidue(L);
	
      }else if (key_is("overlap")) {
	key=free_Token(key);
	default_token(Options.overlapAtoms?"yes":"no");
	key=Get_Kwd("add overlap atoms?  ",";","yes no");
	if (key->status==TOK_OK) {
	  Options.overlapAtoms=key_is("yes");
	}else if (key->status==TOK_NOK) err_set(1,"invalid overlap atoms mode: %s\n",key->text);

      }else if (key_is("rings")) {
	key=free_Token(key);
	sprintf(buff,"%d",Options.rigidRings);
	default_token(Options.rigidRings<3?"all": buff);
	key=Get_Int_M("rings are rigid if size <n : ",";","all 4 5 6 7 8 9 10");
	if (key->status==TOK_OK) {
	  Options.rigidRings=key->n < 0 ? 0 : key->n;
	}else if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"all");
	  if (key->status==TOK_OK) {
	    Options.rigidRings=0;
	  }
	}
	if (key->status==TOK_OK) {
	  cyaMarkRigidBonds(L,Options);
	  cyaFindTorsions(L);
	  reportCurrentResidue(L);
	}else if (key->status==TOK_NOK) err_set(1,"invalid rings mode: %s\n",key->text);

      }else if (key_is("root")) {
	if (L) {
	  A=treeRoot(L);
	  if (A) {
	    cyaNewRoot(L,A);
	    cyaFindTorsions(L);
	    reportCurrentResidue(L);
	  }
	}else{
	  err_set(1,"must select molecule before defining root.");
	}
      
      }else if (key_is("style")) {
	key=free_Token(key);
	default_token(Options.useNames?"names":"numbers");
	key=Get_Kwd("style of connectivity records: ",";","names numbers");
	if (key->status==TOK_OK) {
	  Options.useNames=key_is("names");
	}else if (key->status==TOK_NOK) err_set(1,"invalid rings mode: %s\n",key->text);

      }else if (key_is("uppercase")) {
	key=free_Token(key);
	default_token(Options.nameUpper&&Options.chsUpper?"both":
		      Options.nameUpper?"names":
		      Options.chsUpper?"types":
		      "none");
	key=Get_Kwd("use all-uppercase for: ",";","names types both none");
	if (key->status==TOK_OK) {
	  Options.chsUpper=key_is("types")||key_is("both");
	  Options.nameUpper=key_is("names")||key_is("both");
	}else if (key->status==TOK_NOK) err_set(1,"invalid all=uppercase mode: %s\n",key->text);

      }else if (key_is("write")) {
	if (L) {
	  filename=free_Token(filename);
	  filename=Get_String("filename: ",";");
	  if (filename->status==TOK_OK) {
	    out=fopen(extfil(buff,filename->text,".lib",FALSE),"w");
	    if (out) {
	      if (writeCyanaResidue(out,L,code?code:"LI1",Options)) {
		err_set(1,"ERROR: library file %s written with errors, needs manual fix.",buff);
	      }
	      fclose(out);
	      out=fopen(extfil(buff,filename->text,".cya",TRUE),"w");
	      if (out) {
		writeCyanaRestraints(out,L,code?code:"LI1",Options);
		fclose(out);
	      }else{
		err_set(1,"failed to open restraints file for write: %s\n",buff);
	      }
	    }else{
	      err_set(1,"failed to open residue definition file for write: %s\n",buff);
	    }
	  }else if (filename->status==TOK_NOK) {
	    err_set(1,"invalid token: %s\n",filename->text);
	  }
	}else err_set(1,"first select molecule, then use write.");

      }else if (key_is("end")) {
	state=DONE;
      }

    }else if (key->status==TOK_NOK) {
      err_set(1,"invalid Cyana key: %s\n",key->text);
      end_parse();
    }else state=DONE;
  }
  filename=free_Token(filename);
  key=free_Token(key);
  if (code) free(code);
  freeCyaResidue(L);
}

