#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "fileutil.h"

extern int graphix;
extern Node *DefaultNode;

#define TRUE 1
#define FALSE 0

#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define key_is(a) (streq(a,key->key))

typedef struct {
  int charge;               /* total charge, >999 = sum of formal charge */
  int hamilt;
  int mmok;
  int geo_ok;
  int inter;
  int graph;
  int mulliken;
  int pulay;
  int precise;
  int time;
  int optimize;
  int optmeth;
} mopac_options;

static void cmd_mopac_run(int mode);
static void cmd_mopac_retrieve(void);
static void cmd_mopac_setup(void);
static void mopac_retrieve(Node *root,UCpipe *F);
static char *encode_options(char *out,mopac_options O);
static int write_mopac_input(char *jobname,Node *root,mopac_options O);
static int mol_charge(Nodechain *A);

#define MOPAC_RUN      1
#define MOPAC_SUBMIT   2
#define MOPAC_WRITE    3

#define OPT_NOTHING    0
#define OPT_EVERYTHING 1
#define OPT_AM1        0
#define OPT_MINDO3     1
#define OPT_MNDO       2
#define OPT_PM3        3
#define OPT_BFGS       0
#define OPT_EIGEN      1

static mopac_options Opt = {       /* current settings, initialized to defaults */
  1000,OPT_MNDO,FALSE,FALSE,FALSE,0,0,0,0,7200,OPT_NOTHING,OPT_BFGS
};

static char *mopac_server = (char *)NULL;

/*========================================================================*/

void cmd_mopac(void)
{
  Token *key;
  char *main_keys = "retrieve run setup submit write ' ' 'end'";
  int loop;

  key=(Token *)NULL;

  for (loop=TRUE;loop;) {
    key=free_Token(key);start_parse(); err_dump();
    key=Get_Kwd("Mopac option> ",";",main_keys);
    if (key->status==TOK_OK) {
      if (key_is("retrieve"))    cmd_mopac_retrieve();
      else if (key_is("run"))    cmd_mopac_run(MOPAC_RUN);
      else if (key_is("write"))  cmd_mopac_run(MOPAC_WRITE);
      else if (key_is("submit")) cmd_mopac_run(MOPAC_SUBMIT);
      else if (key_is("setup"))  cmd_mopac_setup();
      else if (key_is("end"))    loop=FALSE;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad Mopac option: %s",key->text);
    }else loop=FALSE;
    if (end_parse()!=TOK_OK) err_set(0,"extra characters at end of command ignored.");
  }
  free_Token(key);
}
/*-------------------- cmd_mopac_retrieve --------------------*/

static void cmd_mopac_retrieve(void)
{
  Token *key;
  Node *root;
  UCpipe *F;
  
  root=(Node *)NULL;
  F=(UCpipe *)NULL;
  key=(Token *)NULL;
  key=Get_File((char *)NULL,".syb","syb-file: ",";");
  if (key->status==TOK_OK) {
    F=(UCpipe *)key->ptr;
  }else if (key->status==TOK_NOK) {
    err_set(1,"failed to open .syb-file %s",key->text);
  }
  if (key->status==TOK_OK) {
    key=free_Token(key);
    key=Get_Mol_D("molecule: ",";");
    if (key->status==TOK_OK) {
      root=(Node *)key->ptr;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad molecule selection: %s",key->text);
    }
  }
  if (key->status==TOK_OK) {
    mopac_retrieve(root,F);
  }
  UCpclose(F);
  free_Token(key);
}
/*-------------------- mopac_retrieve --------------------*/

static void mopac_retrieve(Node *root,UCpipe *F)
{
  FILE *syb;
  char rec[4096],buf[128];
  float *xyz,*chrg,homo[2],lumo[2],hof,iop,spc=0.0,rmsd;
  int nat,i,updx,updc;
  Nodechain A;
  Atom *a,**atoms;
  mat_4 rot;

  xyz=(float *)NULL;
  chrg=(float *)NULL;
  updc=FALSE;
  updx=FALSE;

  if (!F || !root) {
    err_set(1,"bad arguments in call to mopac_retrieve.");
    return;
  }
  init_node_chain(&A);
  ForAtoms(a,root) {
    if (mvld(a->atp)) {
      if (a->atp->z>0) {      /* ignore dummy atoms... */
	add_to_node_chain((Node *)a,&A);
      }
    }
  }
  if (A.n < 1) {
    err_set(1,"mopac_retrieve: %s contains no atoms.",nodeptr_to_name(rec,root));
    clear_node_chain(&A);
    return;
  }
  syb=F->fd;
  if (!fgets(rec,4095,syb)) {
    err_set(1,"unexpected EOF on file %s",F->filename);
    clear_node_chain(&A);
    return;
  }
  nat=str2i(strfld(buf,rec+4,1,4));
  if (nat!=A.n) {
    err_set(1,"number of atoms don't match, syb: %d, mol: %d.",nat,A.n);
    clear_node_chain(&A);
    return;
  }
  xyz=(float *)malloc(nat*3*sizeof(float));
  chrg=(float *)malloc(nat*sizeof(float));
  if (!mvld(xyz)||!mvld(chrg)) {
    err_set(1,"memory allocation failure.");
    clear_node_chain(&A);
    if (mvld(xyz)) free(xyz);
    if (mvld(chrg)) free(chrg);
    return;
  }
  for (i=0;i<nat;i++) {
    if (!fgets(rec,4095,syb)) break;
    xyz[3*i]=str2f(strfld(buf,rec,1,12));
    xyz[3*i+1]=str2f(strfld(buf,rec+12,1,12));
    xyz[3*i+2]=str2f(strfld(buf,rec+24,1,12));
    chrg[i]=str2f(strfld(buf,rec+36,1,12));
  }
  if (i<nat) {
    err_set(1,"unexpected EOF on file %s",F->filename);
    clear_node_chain(&A);
    free(xyz); free(chrg);
    return;
  }
  if (fgets(rec,4095,syb)) {
    homo[0]=str2f(strfld(buf,rec,1,12));
    homo[1]=str2f(strfld(buf,rec+12,1,12));
    lumo[0]=str2f(strfld(buf,rec+24,1,12));
    lumo[1]=str2f(strfld(buf,rec+36,1,12));
    prtGraphics("HOMOs : %9.3f  %9.3f Kcal/mol\n",homo[0],homo[1]);
    prtGraphics("LUMOs : %9.3f  %9.3f Kcal/mol\n",lumo[0],lumo[1]);
  }
  if (fgets(rec,4095,syb)) {
    hof=str2f(strfld(buf,rec,1,12));
    iop=str2f(strfld(buf,rec+12,1,12));
    prtGraphics("Ionization Potential = %9.3f Kcal/mol\n",iop);
    prtGraphics("Heat of Formation    = %9.3f Kcal/mol\n",hof);
  }
  while(fgets(rec,4096,syb)) {
    if (strneq("   1      MULLIKEN POPULATION",rec,strlen("   1      MULLIKEN POPULATION"))) break;
  }
  if (!feof(syb)) {
    for (i=0;i<nat;i++) {
      if (!fgets(rec,4095,syb)) break;
      chrg[i]=str2f(strfld(buf,rec+12,1,12));
    }
    if (i==nat) {
      prtGraphics("Partial charges from Mulliken Analysis.\n");
      updc=TRUE;
    }else{
      err_set(1,"Unexpected EOF in Mulliken section. No charges retrieved.");
      updc=FALSE;
    }
  }else{
    prtGraphics("MOPAC charges.\n");
    updc=TRUE;
  }
  if (updc) {
    for(spc=0.0,i=0,a=(Atom *)A.s;mvld(a);a=(Atom *)a->chain,i++) {
      a->q=chrg[i];
      spc+=chrg[i];
    }
  }
  prtGraphics("Sum of partial charges = %6.3f\n",spc);
  atoms=(Atom **)chain_to_ptr(&A,FALSE);
  if (mvld(atoms)) {
    rmsd=rms_best_mat_x(nat,atoms,xyz,&rot);
    prtGraphics("r.m.s. of coordinate shifts = %6.3f\n",rmsd);
    rot=mmxinv(rot);
    free(atoms);
    for (i=0,a=(Atom *)A.s;mvld(a);a=(Atom *)a->chain,i++) {
      a->x[0]=xyz[3*i]; a->x[1]=xyz[3*i+1]; a->x[2]=xyz[3*i+2];
      transform_atom(rot,a);
    }
    updx=TRUE;
  }else{
    err_set(1,"memory allocation failure.");
    err_set(1,"failed to retrieve new coordinates.");
  }
  if (updx) send_newx(&A);
  if (updc && auto_label()) {
    for (a=(Atom *)A.s;a;a=(Atom *)a->chain) set_atom_label(a,ATM_LAB_Q,NULL,TRUE);
    trflush(graphix);
  }
  clear_node_chain(&A);
  free(xyz);
  free(chrg);
  return;
}
/*-------------------- cmd_mopac_setup --------------------*/

static void cmd_mopac_setup(void)
{
  Token *key;
  char *setup_options = "charge ' ' AM1 MINDO/3 MNDO PM3 "
    "' ' Optimize MMOK GEO-OK "
    "' ' Inter Mulliken Graph "
    "' ' Pulay precise cpu-time "
    "' ' server ' ' list ' ' End";
  char buf[256];

  key=(Token *)NULL;
  for (;;) {
    err_dump(); key=free_Token(key); start_parse();
    key=Get_Kwd("option: ",";",setup_options);
    if (key->status==TOK_OK) {
      if (key_is("End")) break;
      else if (key_is("list")) {
	prtGraphics("current settings of Mopac options:\n");
	prtGraphics("%s\n",encode_options(buf,Opt));
	if (Opt.optimize==OPT_EVERYTHING) prtGraphics("full geometry optimization requested.\n");
	if (Opt.charge>999) prtGraphics("charge = sum of formal charges.\n");
      }
      else if (key_is("server")) {
	if (mvld(mopac_server)) default_token("'%s'",mopac_server);
	else default_token("mopac");
	key=free_Token(key);
	key=Get_String("Mopac server command: ",";");
	if (key->status==TOK_OK) {
	  if (mvld(mopac_server)) free(mopac_server);
	  mopac_server=Strdup(key->text);
	}else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      }
      else if (key_is("charge")) {
	if (Opt.charge>999) default_token("formal");
	else default_token("%d",Opt.charge);
	key=free_Token(key); 
	key=Get_Int_M("total charge = ",";","-4 -3 -2 -1 ' ' 0 ' ' 1 2 3 4 ' ' formal");
	if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"formal");
	  if (key->status==TOK_OK) key->n=1000;
	}
	if (key->status==TOK_OK) Opt.charge=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad charge specification: %s",key->text);
      }
      else if (key_is("AM1")) Opt.hamilt=0;
      else if (key_is("MINDO/3")) Opt.hamilt=1;
      else if (key_is("MNDO")) Opt.hamilt=2;
      else if (key_is("PM3")) Opt.hamilt=3;
      else if (key_is("Optimize")) {
	if (Opt.optimize==0) default_token("nothing");
	else if (Opt.optimize==1) default_token("everything");
	key=free_Token(key);
	key=Get_Kwd("optimize what ? ",";","nothing everything ' ' method");
	if (key->status==TOK_OK) {
	  if (key_is("method")) {
	    key=free_Token(key);
	    default_token(Opt.optmeth==0 ? "BFGS" :(Opt.optmeth==1 ? "Eigenv.following" : (char *)NULL));
	    key=Get_Kwd("optimization method: ",";","BFGS Eigenv.following");
	    if (key->status==TOK_OK) Opt.optmeth=key->n;
	    else if (key->status==TOK_NOK) err_set(1,"unknown optimization method: %s",key->text);
	  }else Opt.optimize=key->n;
	}else if (key->status==TOK_NOK) err_set(1,"bad optimization option: %s",key->text);
      }
      else if (key_is("MMOK")) {
	if (Opt.mmok) default_token("yes");
	else default_token("no");
	key=free_Token(key); 
	key=Get_Kwd("use molecular mechanics correction to CONH bonds ? ",";","yes no");
	if (key->status==TOK_OK) Opt.mmok=1-key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad response: %s, expected yes or no.",key->text);
      }
      else if (key_is("GEO-OK")) {
	if (Opt.geo_ok) default_token("yes");
	else default_token("no");
	key=free_Token(key); 
	key=Get_Kwd("override interatomic distance check ? ",";","yes no");
	if (key->status==TOK_OK) Opt.geo_ok=1-key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad response: %s, expected yes or no.",key->text);
      }
      else if (key_is("Inter")) {
	if (Opt.inter) default_token("yes");
	else default_token("no");
	key=free_Token(key); 
	key=Get_Kwd("print interatomic distances ? ",";","yes no");
	if (key->status==TOK_OK) Opt.inter=1-key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad response: %s, expected yes or no.",key->text);
      }
      else if (key_is("Mulliken")) {
	if (Opt.mulliken) default_token("yes");
	else default_token("no");
	key=free_Token(key); 
	key=Get_Kwd("perform Mulliken analysis ? ",";","yes no");
	if (key->status==TOK_OK) Opt.mulliken=1-key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad response: %s, expected yes or no.",key->text);
      }
      else if (key_is("Graph")) {
	if (Opt.graph) default_token("yes");
	else default_token("no");
	key=free_Token(key); 
	key=Get_Kwd("generate file for graphics ? ",";","yes no");
	if (key->status==TOK_OK) Opt.graph=1-key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad response: %s, expected yes or no.",key->text);
      }
      else if (key_is("Pulay")) {
	if (Opt.pulay) default_token("yes");
	else default_token("no");
	key=free_Token(key); 
	key=Get_Kwd("use Pulay's converger to obtain a SCF ? ",";","yes no");
	if (key->status==TOK_OK) Opt.pulay=1-key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad response: %s, expected yes or no.",key->text);
      }
      else if (key_is("precise")) {
	if (Opt.precise) default_token("yes");
	else default_token("no");
	key=free_Token(key); 
	key=Get_Kwd("criteria to be increased by 100 times ? ",";","yes no");
	if (key->status==TOK_OK) Opt.precise=1-key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad response: %s, expected yes or no.",key->text);
      }
      else if (key_is("cpu-time")) {
	default_token("%d",Opt.time);
	key=free_Token(key); 
	key=Get_Int_M("max. CPU seconds = ",";","100 200 300 40 500 600 1000 3600 7200 36000");
	if (key->status==TOK_OK) Opt.time=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad CPU time specification: %s",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad Mopac setup option: %s",key->text);
      end_parse();
    }else break;
  }
  free_Token(key);
}
/*-------------------- encode_options --------------------*/

static char *encode_options(char *out,mopac_options O)
{
  *out='\0';
  if (O.charge <=999 && O.charge !=0 ) sprintf(out+strlen(out),"CHARGE=%d ",O.charge);
  if (O.hamilt==OPT_AM1) sprintf(out+strlen(out),"AM1 ");
  else if (O.hamilt==OPT_MINDO3) sprintf(out+strlen(out),"MINDO/3 ");
  else if (O.hamilt==OPT_PM3) sprintf(out+strlen(out),"PM3 ");
  if (O.mmok) sprintf(out+strlen(out),"MMOK ");
  if (O.geo_ok) sprintf(out+strlen(out),"GEO-OK ");
  if (O.pulay) sprintf(out+strlen(out),"PULAY ");
  if (O.precise) sprintf(out+strlen(out),"PRECISE ");
  if (O.optimize!=OPT_NOTHING) {
    if (O.optmeth==OPT_EIGEN) {
      sprintf(out+strlen(out),"LET EF ");
      if (!O.geo_ok) sprintf(out+strlen(out),"GEO-OK ");
    }
  }else{
    sprintf(out+strlen(out),"1SCF ");
  }
  if (!O.inter) sprintf(out+strlen(out),"NOINTER ");
  if (O.mulliken) sprintf(out+strlen(out),"MULLIK ");
  if (O.graph) sprintf(out+strlen(out),"GRAPH ");
  sprintf(out+strlen(out),"T=%d",O.time);
  return(out);
}

/*-------------------- cmd_mopac_run --------------------*/

static void cmd_mopac_run(int mode)
{
  Token *key;
  Node *root;
  char *jobname,cmd[2048],buf[2048],buf2[2048];
  int status;
  UCpipe *R;

  jobname=(char *)NULL;
  root=(Node *)NULL;
  key=(Token *)NULL;
  key=Get_Mol_D("molecule: ",";");
  if (key->status==TOK_OK) {
    root=(Node *)key->ptr;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad molecule selection: %s",key->text);
  }
  if (key->status==TOK_OK) {
    if (mode==MOPAC_RUN) default_token("mopacjob");
    else {
      if (mvld(root->name)) default_token("%s",root->name);
      else default_token("NoName");
    }
    key=free_Token(key);
    key=Get_String("jobname: ",";");
    if (key->status==TOK_OK) jobname=Strdup(key->text);
    else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
  }
  if (key->status==TOK_OK) {
    if (write_mopac_input(jobname,root,Opt)==0) {
      err_dump();
      prtGraphics("mopac input file for job %s written.\n",jobname);
      switch (mode) {
      case MOPAC_RUN:
	sprintf(cmd,"%s %s",
		(mvld(mopac_server) ? mopac_server : "mopac"),jobname);
	prtGraphics("running command %s\n",cmd);
	rename(extfil(buf,jobname,".syb",TRUE),extfil(buf2,jobname,".syb~",TRUE));
        status=startCmd(cmd,"","",TRUE,"belownormal");
	if (status) err_set(1,"shell returns with error status %d",status);
	else{
	  R=UCpopen(extfil(buf,jobname,".syb",TRUE));
	  if (mvld(R)) {
	    mopac_retrieve(root,R);
	    UCpclose(R);
	  }else err_set(1,"failed to retrieve MOPAC result from %s",buf);
	}
	break;
      case MOPAC_SUBMIT:
	sprintf(cmd,"%s %s",
		(mvld(mopac_server) ? mopac_server : "mopac"),jobname);
	prtGraphics("sumbitting command %s\n",cmd);
	status=startCmd(cmd,"","",FALSE,"belownormal");
	if (status) err_set(1,"shell returns with error status %d",status);
	break;
      default:
	break;
      }
    }else err_set(1,"failed to write mopac input file for job %s",jobname);
  }
  free_Token(key);
  if (mvld(jobname)) free(jobname);
}
/*-------------------- write_mopac_input --------------------*/

static int write_mopac_input(char *jobname,Node *root,mopac_options O)
{
  int charge,flg;
  char buf[2048];
  Atom *a;
  Nodechain A;
  mopac_options o;
  FILE *out;

  if (!mvld(root)) {
    err_set(1,"bad root for write_mopac_input.");
    return(1);
  }
  
  init_node_chain(&A);
  ForAtoms(a,root) {
    if (mvld(a->atp)&&a->nxyz>0) {
      if (a->atp->z>0) add_to_node_chain((Node *)a,&A);
    }
  }

  if (A.n < 1) {
    err_set(1,"no atoms in molecule.");
    return(2);
  }

  if (A.n < 4) {   /* because MOPAC can't handle cart. coordinates in this case... */
    err_set(1,"sorry can't handle molecules with less than 4 atoms (a bug).");
    clear_node_chain(&A);
    return(3);
  }

  charge=mol_charge(&A);
  if (O.charge<=999 && charge != O.charge) {
    err_set(0,"total charge on molecule is %d, set charge is %d",charge,O.charge);
    err_set(0,"total charge on molecule assumed to be %d",O.charge);
    charge = O.charge;
  }else err_set(0,"total charge on molecule is %d",charge);

  o=O; o.charge=charge;
  
  out=fopen(extfil(buf,jobname,".dat",TRUE),"w");
  if (!out) {
    err_set(1,"failed to open file %s for write.",buf);
    clear_node_chain(&A);
    return(4);
  }

  if (o.optimize == OPT_EVERYTHING) flg=1;
  else flg=0;

  fprintf(out,"XYZ %s\n",encode_options(buf,o));
  fprintf(out,"Mopac input generated by Wit!P\n");
  fprintf(out,"Molecule: %s\n",(mvld(root->name) ? root->name : "<no name>"));
  for (a=(Atom *)A.s;mvld(a);a=(Atom *)a->chain) {
    fprintf(out,"%2s%10.4f %d%10.4f %d%10.4f %d\n",
	    a->atp->chs,a->x[0],flg,a->x[1],flg,a->x[2],flg);
  }
  fprintf(out,"\n");
  fclose(out);

  clear_node_chain(&A);
  return(0);
}
/*-------------------- mol_charge --------------------*/

static int mol_charge(Nodechain *A)
{
  /* determine molecular charge, assume all H present..., ignore lig,chrg */

  Atom *a;
  int charge,c;

  for (charge=0,a=(Atom *)A->s;mvld(a);a=(Atom *)a->chain) {
    c=asg_formchrg(a,eff_lig(a));
    charge+=c;                      /* check c == a->chrg ? */
  }
  return(charge);
}
