#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "fileutil.h"
#include "xtal.h"
#include "gcalc.h"
#include "errhand.h"

extern int graphix;
extern Mol *MolRoot;
extern Dspobj *DspRoot;

typedef struct {               /* PLUTO Atom structure */
  Atom  *atom; 
  Xatom *xatom;
  Dspobj *dsp;
  int   xtal;
  float radius;
  int   label;
}PAtom;

typedef struct {               /* PLUTO Bond structure */
  Bond *bnd;
  float rad;
  int   from;
  int   to;
  int   nl;
  int   cl;
}PBond;

#define TRUE 1
#define FALSE 0

#define LOOP 0
#define DONE 1
#define CLOSE 2
#define MAX_PLUTO_ATOMS 2000    /* should not exceed dimension in PLUTO */
#define MXOPEN 10

#define key_is(x) (streq(x,key->key))
#define newline(f) fprintf(f,"\n")

#define RAD(x) (((x)*3.141592653)/180.0)
#define DEG(x) (((x)*180.0)/3.141592653)

static int pluto_color[WNP_MAXCOL] =
{ 0,2,3,7,4,6,5,1,8, 8,8,8,8,8,8,8};

static void  pluto_get_atoms(void);
static void  pluto_newpic(void);
static FILE *pluto_open(void);
static void  pluto_close(void);
static void  pluto_write_atom(FILE *o,mat_4 DX,int i);
static void  pluto_reset_defaults(void);
static void  pluto_free_tmp_data(void);
static void  pluto_set_command(void);
static void  pluto_plot(void);
static mat_4 pluto_cltoan(Xcn *xcn);


static char *pluto_command=(char *)NULL;
static Mol *plutoMolRoot;
static int nopen=0;
static FILE *open_pipe[MXOPEN];
static int needs_reset=TRUE;
static PAtom *Atoms;
static PAtom defaultAtom;
static int NAtoms=0;
static PBond *Bonds;
static PBond defaultBond;
static int NBonds=0;
static FILE *pluto;
static Xcn *Xinfo=(Xcn*)NULL;
static float defaultAtomradius,defaultLsize,defaultBondrad,defaultBondnl,
  defaultTaper,defaultBondcl;
static float Lsize,Taper;
static int mode,shade;
     
static char *FMTxyz;
static char *XTLxyz = "%9.4f%9.4f%9.4f";
static char *ANGxyz = "%9.3f%9.3f%9.3f";
     
static PrjTS prj;
     
static char *mainkeys = 
  "display newpic plot pipe default color label radius bond stick solid shade "
  "end";
static char *par_keys="radius label bond";
static char *bond_keys="color lines radius taper";
static char *label_keys="on off size";

static char *atom_rad="0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5";
static char *bond_rad="0.01 0.02 0.03 0.04 0.05 0.06 0.07 0.08 0.09 0.1";
static char *bond_nl="0 1 2 3 4 5 6 7 8 9 10 15 20";
static char *taper_menu="0 1 5 10 15 20 25 30 35 40 45 50 60 70 80 90 100";
static char *lab_sizes="0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0 1.5 2.0";

#define NONE 0
#define RADIUS 1
#define ONOFF 2
#define LSIZE 3
#define BRADIUS 4
#define NLINES 5
#define BCOLOR 6

#define MODE 5
#define MODE_STICK 0
#define MODE_SOLID 1
     
/*========================================================================*/
     void cmd_pluto(void)
{
  Token *key;
  int i,do_default,apar,bpar,onoff=0,status,nl=0,cl=0;
  float rad=0;
  Nodechain *c;
  Bond *bnd;
  
  if (needs_reset) pluto_reset_defaults();
  pluto_get_atoms();
  status=LOOP;
  
  while(status==LOOP) {
    start_parse();
    apar=bpar=NONE;
    do_default=FALSE;
    key=Get_Kwd("Pluto> ",";",mainkeys);
    if (key->status==TOK_OK) {
      if (key_is("default")) {
	free_Token(key);
	key=Get_Kwd("what ? ",";",par_keys);
	do_default=TRUE;
      }
    }
    if (key->status==TOK_OK) {
      if (key_is("display")) {
	cmd_display();
	pluto_get_atoms();
      }else if (key_is("newpic")) {
	pluto_newpic();
      }else if (key_is("plot")) {
	pluto_plot();
      }else if (key_is("pipe")) {
	pluto_set_command();
      }else if (key_is("end")) {
	status=CLOSE;
      }else if (key_is("radius")) {
	free_Token(key);
	key=Get_Float_M("atom radius: ",";",atom_rad);
	if (key->status==TOK_OK) {
	  rad=key->f;
	  if (rad>0.001) apar=RADIUS;
	  else {
	    err_set(1,"radius is too small...");
	    end_parse();
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad float: %s",key->text);
	  end_parse();
	}
      }else if (key_is("label")) {
	free_Token(key);
	key=Get_Kwd("label option: ",";",label_keys);
	if (key->status==TOK_OK) {
	  if (key_is("on") || key_is("off")) {
	    apar=ONOFF;
	    onoff=key_is("on");
	  }else if (key_is("size")) {
	    free_Token(key);
	    key=Get_Float_M("label size (A): ",";",lab_sizes);
	    if (key->status==TOK_OK) {
	      if (key->f>=0) {
		if (do_default) defaultLsize=key->f;
		else Lsize=key->f;
	      }else{
		err_set(1,"bad value of label size: %s",key->text);
		end_parse();
	      }
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad label size token: %s",key->text);
	      end_parse();
	    }
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"unknown label option: %s",key->text);
	  end_parse();
	}
      }else if (key_is("bond")) {
	free_Token(key);
	key=Get_Kwd("bond parameter: ",";",bond_keys); 
	if (key->status==TOK_OK) {
	  if (key_is("radius")) {
	    free_Token(key);
	    key=Get_Float_M("radius (A): ",";",bond_rad);
	    if (key->status==TOK_OK) {
	      rad=key->f;
	      if (rad>0.0001) {
		bpar=BRADIUS;
	      }else{
		err_set(1,"bond radius too small...");
	      }
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad format of bond radius: %s",key->text);
	      end_parse();
	    }
	  }else if (key_is("lines")) {
	    free_Token(key);
	    key=Get_Int_M("number of lines = ",";",bond_nl);
	    if (key->status==TOK_OK) {
	      nl=key->n;
	      if (nl>=0) {
		bpar=NLINES;
	      }else{
		err_set(1,"bad bond lines parameter");
	      }
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad bond lines parameter: %s",key->text);
	      end_parse();
	    }
	  }else if (key_is("color")) {
	    free_Token(key);
	    key=Get_Color("color: ",";");
	    if (key->status==TOK_OK) {
	      cl=key->n;
	      if (cl>=0&&cl<WNP_MAXCOL) {
		bpar=BCOLOR;
	      }else{
		err_set(1,"bad bond color: %s",key->text);
	      }
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad bond color: %s",key->text);
	      end_parse();
	    }
	  }else if (key_is("taper")) {
	    free_Token(key);
	    key=Get_Float_M("taper= ",";",taper_menu);
	    if (key->status==TOK_OK) {
	      if (do_default) defaultTaper=key->f;
	      else Taper=key->f;
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad bond taper parameter: %s",key->text);
	      end_parse();
	    }
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"unknown bond parameter: %s",key->text);
	  end_parse();
	}
      }else if (key_is("solid")) {
	mode=MODE_SOLID;
	prtGraphics("drawing mode set to solid\n");
      }else if (key_is("stick")) {
	mode=MODE_STICK;
	prtGraphics("drawing mode set to stick\n");
      }else if (key_is("shade")) {
	free_Token(key);
	key=Get_Kwd("shade atoms ? ",";","no yes");
	if (key->status==TOK_OK) {
	  shade=key_is("yes");
	}else{
	  err_set(1,"bad reply: %s",key->text);
	  end_parse();
	}
      }else if (key_is("color")) {
	cmd_color();
      }
      
    }else if (key->status==TOK_NOK) {
      err_set(1,"unknown Pluto option: %s",key->text);
      end_parse();
    }else if (!do_default) status=DONE;
    
    if (bpar!=NONE) {
      if (do_default) {
	switch (bpar) {
	case BRADIUS:
	  defaultBond.rad=rad;
	  break;
	case NLINES:
	  defaultBond.nl=nl;
	  break;
	case BCOLOR:
	  defaultBond.cl=cl;
	  break;
	default: break;
	}
      }else{
	free_Token(key);
	key=Get_Bonds_D("bonds: ",";");
	if (key->status==TOK_OK) {
	  for (i=0;i<NBonds;i++) Bonds[i].bnd->flg|=(1<<1);
	  for (bnd=(Bond *)key->ptr;mvld(bnd);bnd=(Bond *)key->ptr) {
	    key->ptr=(Bond *)bnd->chain;
	    bnd->chain=(Bond *)NULL;
	    bnd->flg&=~(1<<1);
	  }
	  for (i=0;i<NBonds;i++) {
	    if (!(Bonds[i].bnd->flg&(1<<1))) {
	      switch(bpar) {
	      case BRADIUS:
		Bonds[i].rad=rad;
		break;
	      case NLINES:
		Bonds[i].nl=nl;
		break;
	      case BCOLOR:
		Bonds[i].cl=cl;
		break;
	      default: break;
	      }
	    }else{
	      Bonds[i].bnd->flg&=~(1<<1);
	    }
	  }
	}
	send_header(graphix,"ARMK",0);
      }
    }
    
    if (apar!=NONE) {
      if (do_default!=NONE) {
	switch(apar) {
	case RADIUS:
	  defaultAtom.radius=rad;
	  break;
	case ONOFF:
	  defaultAtom.label=onoff;
	  break;
	default:
	  break;
	}
      }else{
	free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (i=0;i<NAtoms;i++) {
	    if (in_chain(Atoms[i].atom,c)) {
	      switch(apar) {
	      case RADIUS:
		Atoms[i].radius=rad;
		break;
	      case ONOFF:
		Atoms[i].label=onoff;
		break;
	      default:
		break;
	      }
	    }
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}
	send_header(graphix,"ARMK",0);
      }
    }
    
    if (end_parse()!=TOK_OK) {
      err_set(1,"extra characters at end of command ignored.");
    }
    free_Token(key);
    err_dump();
  }
  
  pluto_free_tmp_data();
  if (status==CLOSE) pluto_close();
} 
/*-------------------- pluto_get_atoms --------------------*/

static void pluto_get_atoms(void)
{
  /* put displayed atoms into PLUTO atom list */
  
  Atom *atm;
  int i,j,k,a,b;
  Nodechain *c;
  Xcn *xcn;
  PAtom *oldAtoms;
  PBond *oldBonds;
  int multi_xcn,bad_xcn,oldNAtoms,oldNBonds;
  
  c=make_node_chain();
  ForAtoms(atm,plutoMolRoot) {
    if (atm->dsp) add_to_node_chain((Node *)atm,c);
  }
  
  /* allocate space for pluto atoms */
  
  if (c->n>MAX_PLUTO_ATOMS-1) {
    err_set(1,"too many atoms, max is %d",MAX_PLUTO_ATOMS-1);
    clear_node_chain(c);
    free_node_chain(c);
    return;
  }
  
  oldNAtoms=NAtoms;oldAtoms=Atoms;
  oldNBonds=NBonds;oldBonds=Bonds;
  NAtoms=c->n;
  
  if (NAtoms>0) {
    Atoms=(PAtom *)malloc(c->n*(sizeof(PAtom)));
    if (!mvld(Atoms)) {
      err_set(1,"memory allocation failed");
      clear_node_chain(c);
      free_node_chain(c);
      NAtoms=0;
      return;
    }
  }else{
    Atoms=(PAtom *)NULL;
  }
  
  for (i=0,atm=(Atom *)c->s;atm;i++,atm=(Atom *)atm->chain) {
    Atoms[i]=defaultAtom;
    Atoms[i].atom=atm;
  }
  clear_node_chain(c);
  for (i=0;i<NAtoms;i++) Atoms[i].atom->chain=(Node *)NULL+(i+1);
  
  for (i=0;i<oldNAtoms;i++) {            /* copy old Atoms info... */
    if (oldAtoms[i].atom->dsp) {
      k=(oldAtoms[i].atom->chain-(Node *)NULL)-1;
      Atoms[k]=oldAtoms[i];
    }
  }
  
  /* find xtal info and check consistency */
  
  Xinfo=(Xcn*)NULL;
  multi_xcn=FALSE;
  
  bad_xcn=FALSE;
  for (i=0;i<NAtoms;i++) {
    if (Atoms[i].atom->xtal) {
      Atoms[i].xatom=first_xatom(Atoms[i].atom);
      xcn=(Xcn*)find_cousin((Node *)Atoms[i].atom,OBJ_XCN);
      if (!mvld(xcn)) {
	bad_xcn=TRUE;
      }else{
	if (Xinfo && Xinfo != xcn) {
	  multi_xcn=TRUE;
	}else{
	  Xinfo=xcn;
	}
      }
    }else{
      Atoms[i].xatom=(Xatom*)NULL;
    }
  }
  
  /* find associated display objects */
  
  for (i=0;i<NAtoms;i++) {
    Atoms[i].dsp=glblobj(assocobj((Node*)Atoms[i].atom));
  }
  
  if (Xinfo) for (i=0;i<NAtoms;i++) Atoms[i].xtal=(Atoms[i].xatom ? 1 : 0);
  else for (i=0;i<NAtoms;i++) Atoms[i].xtal=0; 
  
  /* if xtal atom without xcn then we have a mess ! */
  
  if (bad_xcn) {
    err_set(1,"messed up object tree !!!");
    Xinfo=(Xcn*)NULL;
  }
  
  /* if more than one xcn in tree, we can't handle it as xtal pluto */
  
  if (multi_xcn) {
    err_set(1,"multi Xtal Pluto not supported.");
    err_set(1,"non-Xtal Pluto will be used...");
    Xinfo=(Xcn*)NULL;
  }
  
  /* generate bond list with default parameters (abusing 'chain') */
  
  for (i=0,NBonds=0;i<NAtoms;i++) {    /* just counting... */
    for (j=0;j<Atoms[i].atom->nbnd;j++) {
      a=(Atoms[i].atom->bnd[j]->f->chain-(Node *)NULL)-1;
      b=(Atoms[i].atom->bnd[j]->t->chain-(Node *)NULL)-1;
      if (a==i && a <NAtoms && b>=0 && b<NAtoms) NBonds++;
    }
  }

  if (NBonds>0) Bonds=(PBond *)malloc(NBonds*sizeof(PBond));
  else Bonds=(PBond *)NULL;

  if (Bonds) {
    for (i=0,k=0;i<NAtoms;i++) {       /* store bond info in temp. structure */
      for (j=0;j<Atoms[i].atom->nbnd;j++) {
	a=(Atoms[i].atom->bnd[j]->f->chain-(Node *)NULL)-1;
	b=(Atoms[i].atom->bnd[j]->t->chain-(Node *)NULL)-1;
	if (a==i && a <NAtoms && b>=0 && b<NAtoms) {
	  Bonds[k]=defaultBond;
	  Bonds[k].bnd=Atoms[i].atom->bnd[j];
	  Bonds[k].bnd->chain=(Bond *)NULL+(k+1);       /* bond chain abuse */
	  Bonds[k].from=a;
	  Bonds[k].to=b;
	  k++;
	}
      }
    }
    for (i=0;i<oldNBonds;i++) {                /* copy old bond info */
      a=oldBonds[i].from; b=oldBonds[i].to;
      if (oldAtoms[a].atom->dsp && oldAtoms[b].atom->dsp) {
	if (mvld(oldBonds[i].bnd)) {
	  k=(oldBonds[i].bnd->chain-(Bond *)NULL);
	  if (k>0 && k <=NBonds) {
	    Bonds[k-1].rad=oldBonds[i].rad;
	  }
	}
      }
    }
  }else if (NBonds>0) {
    err_set(1,"memory allocation failed for %d bonds",
		NBonds);
    NBonds=0;
  }
  
  for(i=0;i<NAtoms;i++) Atoms[i].atom->chain=(Node *)NULL; 
  for (i=0;i<NBonds;i++) Bonds[i].bnd->chain=(Bond *)NULL;
  if (oldNBonds>0) free(oldBonds);
  if (oldNAtoms>0) free(oldAtoms);
  free_node_chain(c);
}
/*-------------------- pluto_newpic --------------------*/

static void pluto_newpic(void)
{
  /* redraw molecule in pluto box... */
  
  FILE *o;
  int sgpdevice;
  int i;
  float rx,ry,rz,sizex,sizey,viewd,mmpA,wxl,wxh,wyl,wyh,wzl,wzh;
  char buffer[256];
  Dspobj *dspr;
  mat_4 DO,DX,DR,FP;
  
  if (NAtoms==0) {
    prtGraphics("0 Atoms, nothing to display in PLUTO\n");
    return;
  }
  
  o=pluto_open();
  if (o) {
    
    prj=getThreeDPrj();                   /* 3D projection */
    get_MVT(DspRoot);                     /* object transformations */
    
    /* set up transformation matrices ... */
    
    if (mvld(Xinfo)) {
      dspr=glblobj(assocobj((Node *)Xinfo));
      DO = (mvld(dspr) ? dspr->mtrx : ident_4());
      DX = m4xm4(Xinfo->atf,mmxinv(DO));
      FP = pluto_cltoan(Xinfo);
      FMTxyz = XTLxyz;
    }else{
      dspr = DspRoot;
      DO = dspr->mtrx;
      DX = mmxinv(DO);
      FP = ident_4();
      FMTxyz = ANGxyz;
    }

    DR=mmxinv(m4xm4(FP,DX));
    eqrxyz(&rx,&ry,&rz,DR);                   /* display rotation of root */

    sgpdevice=21;  /* sgp X graphics device */
    
    /* figure out window and viewport parameters */
    
    sizey=prj.ysize;
    sizex=prj.aspect*prj.ysize;
    
    switch (prj.viewmode) {
    case ORTHO:
      viewd=0.0;
      break;
    case PERSPECTIVE:
      viewd=prj.viewdist*prj.scrysize;
      break;
    default:
      viewd=0.0;
    }

    /* so far, everything was in A, but PLUTO wants mm, so... */

    mmpA = 190.0/sizey; /* assumes device y = 19 cm */
    viewd*=mmpA;

    /* Pluto Window */

    wxl=-0.5*sizex;
    wxh=+0.5*sizex;
    wyl=-0.5*sizey;
    wyh=+0.5*sizey;
    wzl=prj.znear-prj.slabwidth;
    wzh=prj.znear;

    /* Title (comment) line */
    
    fprintf(o,"DATA\n");
    
    /* unit cell */
    
    if (mvld(Xinfo)) {            /* Cryst. Pluto */
      
      /* unit cell geometry */
      
      fprintf(o,"CELL ");
      for (i=0;i<3;i++) fprintf(o,"%9.4f",Xinfo->cl[i]);
      for (i=0;i<3;i++) fprintf(o,"%9.4f",Xinfo->ca[i]);
      newline(o);
      
      /* space group symmetry */
      
      for (i=0;i<Xinfo->nsymx;i++) {
	sym2str(buffer,symF(Xinfo,Xinfo->sym[i]));
	fprintf(o,"SYMM %s\n",buffer);
      }
      
    }
    
    /* output atoms */
    
    for (i=0;i<NAtoms;i++) pluto_write_atom(o,DX,i);
    
    /* explicit bond list instructions */

    for (i=0;i<NBonds;i++) {
      if (Bonds[i].cl>=0&&Bonds[i].cl<WNP_MAXCOL) {
	fprintf(o,"XBOND %4d %4d %7.3f %d %d\n",
		Bonds[i].from+1,Bonds[i].to+1,
		Bonds[i].rad,Bonds[i].nl,pluto_color[Bonds[i].cl]);
      }else{
	fprintf(o,"XBOND %4d %4d %7.3f %d\n",
		Bonds[i].from+1,Bonds[i].to+1,
		Bonds[i].rad,Bonds[i].nl);
      }
    }
    
    fprintf(o,"WINDOW %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
	    wxl,wxh,wyl,wyh,wzl,wzh);
    fprintf(o,"VIEW RESET XR %8.3f YR %8.3f ZR %8.3f\n",DEG(rx),DEG(ry),DEG(rz));
    fprintf(o,"TRANS %9.3f %9.3f %9.3f\n",
	    DR.r[0].c[3],DR.r[1].c[3],DR.r[2].c[3]);
    fprintf(o,"SIZE X 248 Y 190 CHAR %7.3f\n",Lsize*mmpA);
    fprintf(o,"NOBOX\n");
    if (mode==MODE_SOLID) {
      fprintf(o,"SOLID ");
      if (shade) fprintf(o,"SHADE\n");
      else fprintf(o,"\n");
    }else{
      fprintf(o,"STICK\n");
    }
  
    switch (prj.stereomode) {
    case STEREO_NO:
      fprintf(o,"MONO %8.1f\n",viewd);
      break;
    case STEREO_LEFT:;
      fprintf(o,"STEREO LEFT %8.1f %8.1f\n",
	      viewd,prj.eyesep*prj.scrysize*mmpA);
      break;
    case STEREO_RIGHT:;
      fprintf(o,"STEREO RIGHT %8.1f %8.1f\n",
	      viewd,prj.eyesep*prj.scrysize*mmpA);
      break;
    case STEREO_RELAXED:
      fprintf(o,"STEREO %8.1f %8.1f\n",viewd,prj.eyesep*prj.scrysize*mmpA);
      break;
    case STEREO_CROSSED:
      fprintf(o,"STEREO %8.1f %8.1f\n",viewd,-prj.eyesep*prj.scrysize*mmpA);
      break;
    default:
      break;
    }

    fprintf(o,"RADII BONDS TAPER %10.3f\n",(Taper>1?Taper:1.0));

    fprintf(o,"DEVICE %d\n",sgpdevice);
    fprintf(o,"NEWPIC\n");
    
    fflush(o);
  }
}
/*-------------------- pluto_write_atom --------------------*/

static void pluto_write_atom(FILE *o,mat_4 DX,int n)
{
/* write pluto atom record to outputfile o */

  Atom *a;
  vec_3 xyz;
  mat_4 Di;
  char name[7];
  int k,label,color;

  if (n<0 || n>=NAtoms) return;

  a=Atoms[n].atom;

  if (!mvld(a)) return;

  /* coordinates record ... */
  
  Di = m4xm4(DX,Atoms[n].dsp->mtrx);
  xyz = m4xv3(Di,vec3(a->x));
  fprintf(o,"XATOM ");
  if (mvld(a->name)) {
    Strcpy(name,a->name);
    k=strlen(name); if (k<6) strncpy(name+k,"      ",6-k);
    if (k>0) fprintf(o,"%6s",name); 
    else fprintf(o,"X%d ",n+1);
  } else fprintf(o,"X%d ",n+1);
  fprintf(o,FMTxyz,xyz.c[0],xyz.c[1],xyz.c[2]);
  label=(Atoms[n].label ? 'Y' : 'N');
  color = (a->color < 0 ? -a->color : a->color);
  if (color >=0 && color < WNP_MAXCOL) {
    color=pluto_color[color];
  }else{
    color=1;
  }
  fprintf(o," %7.3f %d %c\n",Atoms[n].radius,color,label);
}

/*-------------------- file utilities --------------------*/

static FILE *pluto_open(void)
{
  char *cmd;

  if (!pluto) {
    if (mvld(pluto_command)) {
      cmd=pluto_command;
    }
    else cmd=getenv("PLUTOserver");
    if (cmd) {
      pluto=PipeOpen(cmd,"w");
      if (!pluto) {
	err_set(1,"could not connect to PLUTO command");
      }
    }else{
      err_set(1,"PLUTOserver is not defined.");
    }
  }
  return(pluto);
}

static void pluto_close(void)
{
  if (pluto) {
    fprintf(pluto,"END\n");
    fflush(pluto);
    PipeClose(pluto);
    pluto=(FILE *)NULL;
  }

  while (nopen>0) {
    nopen--;
    fprintf(open_pipe[nopen],"END\n");
    fflush(open_pipe[nopen]);
    open_pipe[nopen]=(FILE *)NULL;
  }
}

static void pluto_set_command(void)
{
  Token *key;

  default_token(getenv("PLUTOserver"));
  key=Get_String("setenv PLUTOserver = ",";");
  if (key->status==TOK_OK) {
    if (mvld(pluto_command)) free(pluto_command);
    pluto_command=Strdup(key->text);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad_token: %s",key->text);
    end_parse();
  }

  if (pluto) {
    free_Token(key);
    default_token("yes");
    key=Get_Kwd("close old PLUTO ? ",";","yes no");
    if (key->status==TOK_OK) {
      if (key_is("yes")) pluto_close();
      else {
	if (nopen>=MXOPEN) {
	  err_set(1,"too many open pluto pipes...");
	  pluto_close();
	}else{
	  open_pipe[nopen++]=pluto;
	  pluto=(FILE *)NULL; 
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
    }
  }

  free_Token(key);
}

static void pluto_plot(void)
{
  Token *key;
  char *dev=(char *)NULL,*cmd;

  if (!pluto) {
    err_set(1,"Nothing to plot...");
    end_parse();
    return;
  }

  cmd=(char *)NULL;
  key=Get_Kwd("device: ",";","HP7500 EPScript EPScolor");
  if (key->status==TOK_OK) {
    if (key_is("HP7500")) {
      cmd=getenv("HP7500plot");
      dev="HP7500";
    }else if (key_is("EPScript")) {
      cmd=getenv("EPSplot");
      dev="EPScript";
    }else if (key_is("EPScolor")) {
      cmd=getenv("EPSplot");
      dev="EPScolor";
    }
    if (cmd) {
      fprintf(pluto,"SGP PLOT %s %s\n",dev,cmd);
      fflush(pluto);
    }else{
      err_set(1,"no plot command defined for device %s",dev);
    }
  }

  free_Token(key);
}
      
    
/*-------------------- pluto_defaults --------------------*/

static void pluto_reset_defaults(void)
{
  /* set static pluto parameters to defaults */
  
  defaultAtomradius=0.075;
  defaultLsize=0.3;
  defaultBondrad=0.04;
  defaultBondnl=8;
  defaultBondcl=7;
  defaultTaper=0;

  plutoMolRoot=MolRoot;
  Lsize=defaultLsize;
  Taper=defaultTaper;
  mode=MODE_STICK;
  shade=FALSE;

  defaultAtom.atom=(Atom *)NULL;
  defaultAtom.xatom=(Xatom *)NULL;
  defaultAtom.dsp=(Dspobj *)NULL;
  defaultAtom.xtal=FALSE;
  defaultAtom.radius=defaultAtomradius;
  defaultAtom.label=FALSE;

  defaultBond.bnd=(Bond *)NULL;
  defaultBond.from=0;
  defaultBond.to=0;
  defaultBond.rad=defaultBondrad;
  defaultBond.nl=defaultBondnl;
  defaultBond.cl=defaultBondcl;

  needs_reset=FALSE;
}
/*-------------------- pluto_free_tmp_data --------------------*/

static void pluto_free_tmp_data(void)
{
  /* free memory for temporary data structures */

  NAtoms=0;
  if (mvld(Atoms)) free(Atoms);
  Atoms=(PAtom *)NULL;

  NBonds=0;
  if (mvld(Bonds)) free(Bonds);
  Bonds=(PBond *)NULL;
}
/*-------------------- pluto_ctoa --------------------*/

static mat_4 pluto_cltoan(Xcn *xcn)
{
/*  return PLUTO cell-fractional to angstrom-cartesian transformation */

  mat_4 r;
  float sa,sb,ca,cb,cg,cgstar,sgstar;
  int i,j;

  for (i=0;i<3;i++) for (j=i+1;j<4;j++) r.r[i].c[j]=0.0;

  sa=sin(RAD(xcn->ca[0]));
  ca=cos(RAD(xcn->ca[0]));
  sb=sin(RAD(xcn->ca[1]));
  cb=cos(RAD(xcn->ca[1]));
  cg=cos(RAD(xcn->ca[2]));
  cgstar=(ca*cb - cg)/(sa*sb);
  sgstar= sqrt(1.0 - cgstar*cgstar);

  r.r[0].c[0]=xcn->cl[0]*sb*sgstar;
  r.r[1].c[0]=-xcn->cl[0]*sb*cgstar;
  r.r[1].c[1]=xcn->cl[1]*sa;                  
  r.r[2].c[0]=xcn->cl[0]*cb;
  r.r[2].c[1]=xcn->cl[1]*ca;
  r.r[2].c[2]=xcn->cl[2];
  r.r[3].c[3]=1.0;

  return(r);
}
