#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "gcalc.h"
#include "errhand.h"
#include "wnprandom.h"

#if defined(darwin)
#define powf pow
#endif

extern Node *LibRoot;
extern Node *DefaultNode;
extern int graphix;

MPEOEatom MPEOEAtoms={
  (MPEOEatom *)NULL,
  (char *)NULL,(Prule *)NULL,
  {1.e6,1.0,1.e20},
  FALSE,
  {FALSE,FALSE,TRUE}
  ,0
};

MPEOEbond MPEOEBonds={
  (MPEOEbond *)NULL,
  (char *)NULL,
  (Prule *)NULL,
  0.0,
  FALSE
  ,0
};

static Node *MPEOELib=(Node *)NULL;

static MPEOEopt mpOpt={ /* options for MPEOE refinement and charge computation */
  0.0001,0.5,0.05,      /* refinement tolerance, dChi, dF                      */
  0.0001,               /* computation accuracy                                */
  0.05,0.85,            /* allowed range for f-parameters                       */
  1000,1000,            /* refinement iterations, computation iterations       */
  1,                    /* minimum occurances of parameter for refinement      */
  10                    /* progress report frequency                           */
};

static double MPEOEgran = 0.000001; /* force charges to be multiples of MPEOEgran */

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define FormalChrg(x) (x->chrg == '?' ? asg_formchrg(x,sane_formlig(x)) : x->chrg)

/*-------------------- ransign --------------------*/

static float ransign(void)
{
  return(WnpRandU()<0.5?-1:1);
}

/*-------------------- QRMS --------------------*/

static float QRMS(int n,float *p,void *G)
{
  /* trivial interface to MPEOEqrms function */

  return(MPEOEqrms(n,p,*(MPEOEinfo *)G));
}

/*-------------------- MPEOEreport --------------------*/

static int MPEOEreport(int iter,int n,float **p,float *v,int hi,int lo)
{
  /* progress report, called by SimPlexMin during parameter refinement */
  /* return FALSE to signal Abort to caller                            */

  if (mpOpt.repf>0&&iter%mpOpt.repf==0) {
    if (iter==0) prtGraphics("MPEOE refinement using %d parameters:\n",n);
    prtGraphics("iter %5d %8.5f %8.5f %8.7f\n",iter,v[lo],v[hi],v[hi]-v[lo]);
  }
  return(!checkabort());
}
/*-------------------- all_charges --------------------*/

int all_charges(Nodechain *c,int auto_label_flag,float q)
{
  Atom *a;
  int nchgd=0;
  
  for (a=(Atom *)c->s,nchgd=0;a;a=(Atom *)a->chain) {
    if (is_atom(a)) {
      a->q=q;
      if (auto_label_flag) set_atom_label(a,ATM_LAB_Q,NULL,TRUE);
      nchgd++;
    }
  }
  return(nchgd);
}

/*-------------------- bondinc_charges -----------------------------*/

int bondinc_charges(Nodechain *c,int auto_label_flag)
{
  /* simple bond increment charges, based on atomtypes... */

  Atom *a;
  int i,nchgd=0;
  Lib *lib;
  char buf[4096];

  lib=(Lib *)find_name(LibRoot,"ResLib",OBJ_LIB);
  for (a=(Atom *)c->s,nchgd=0;a;a=(Atom *)a->chain) {
    if (is_atom(a)) {
      a->q=0.0;
      for (i=0;i<a->nbnd;i++) a->q+=get_bondinc(lib,a,neighbour(i,a));
      if (a->q<-1000) {
	a->q=0;
	err_set(1,"missing charge parameters: %s",nodeptr_to_name(buf,(Node *)a));
      }
      if (auto_label_flag) set_atom_label(a,ATM_LAB_Q,NULL,TRUE);
      nchgd++;
    }
  }
  return(nchgd);
}
/*-------------------- fragment_charges --------------------*/

int fragment_charges(Nodechain *c,Lib *lib,int auto_label_flag)
{
  /* partial atomic charges based on generic fragment definitions */

  Atom *a;
  int i,bmask,umask,ubmask,N,n;
  char buf[4096],buf2[4096];
  Prule *pr;
  int C;
  float Q;

  N=c->n;
  connected_atoms(c);  /* fragment charges always work on connected fragments */
  for (n=0,a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
    if (n<N) a->q=0.0;
    a->seqnr=n++;
  }

  bmask=(1<<7);        /* use flag 7 to mark covered bonds... */
  umask=(1<<6);        /* "used" bonds */
  ubmask=bmask|umask;

  ForPrules(pr,lib) {
    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
      for (n=1;n>0&&matchPruleAtom(a,pr,0,bmask);) {
	for (i=0;i<pr->nat;i++) 
	  if (is_atom(pr->atom[i].match)) 
	    if (pr->atom[i].match->seqnr<N) pr->atom[i].match->q+=pr->atom[i].q;
	for (n=0,i=0;i<pr->nbnd;i++) {
	  if (mvld(pr->bond[i].match)) {
	    if (pr->bond[i].cover) {
	      pr->bond[i].match->flg|=ubmask;
	      n++;
	    }else{
              pr->bond[i].match->flg|=umask;
	    }
	  }
	}
      }
    }
  } 

  for (n=0,C=0,Q=0,a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain,n++) {
    if (n<N) {
      Q+=a->q; C+=FormalChrg(a);
      if (auto_label_flag) set_atom_label(a,ATM_LAB_Q,NULL,TRUE);
    }
    for (i=0;i<a->nbnd;i++) {
      if (a==a->bnd[i]->f) {
	if (!(a->bnd[i]->flg&umask)) {
	  err_set(1,"fragment parameters missing for %s -- %s",
		  nodeptr_to_name(buf,(Node *)a->bnd[i]->f),
		  nodeptr_to_name_ctxt(buf2,(Node *)(Node *)a->bnd[i]->t,(Node *)a->bnd[i]->f));
	}
	a->bnd[i]->flg&=~ubmask;
      }
    }
  }
  prtGraphics("Total charge %d (formal), %0.3f (partial)\n",C,Q);
  return(N);
}
/*-------------------- MPEOEcharges --------------------*/

int MPEOEcharges(Nodechain *c,int auto_label_flag)
{
  MPEOEinfo G;
  int i,n;
  Atom *a;
  float dq,C,Q;
  
  n=c->n;
  G=MPEOEsetup(c);
  MPEOEreport_missing(G,TRUE);
  if (n<c->n) err_set(1,"atom selection expanded to connected fragments (%d-->%d).",n,c->n);

  if (G.nat>0) {
    for (i=0;i<G.nat;i++) G.Q[i]=G.iQ[i];
    dq=MPEOEiterate(G,mpOpt.mxcomp,mpOpt.tolcomp);
    if (dq>mpOpt.tolcomp) err_set(1,"warning: MPEOE charges not converged, dq=%0.5f",dq);
    for (C=0,Q=0,i=0;i<n;i++) {
      a=G.A[i]; a->q=G.Q[i];
      Q+=a->q;
      C+=FormalChrg(a);
      if (auto_label_flag&&i<n) set_atom_label(a,ATM_LAB_Q,NULL,TRUE);
    }
    prtGraphics("Total charge %0.0f (formal), %0.3f (partial)\n",C,Q);
  }
  MPEOEfree_scratch(G);
  return(n);
}
/*-------------------- MPEOErefine --------------------*/

float MPEOErefine(Nodechain *A)
{
  MPEOEinfo G;
  float r;
  int m,n,status;

  G=MPEOEsetup_refinement(A);
  m=MPEOEreport_missing(G,FALSE);
  if (m>0) err_set(1,"warning: %d atoms/bonds have missing MPEOE parameters.",m);
  n=G.npara+G.nparb;
  if (n>0) {
    r=SimPlexMin(n,G.p,G.dp,QRMS,(void *)&G,mpOpt.tolref,mpOpt.mxref,MPEOEreport,&status);
    if (status) err_set(1,"warning: SimPlexMin did not converge. %d",status);
    MPEOEupdate(n,G.p,G);
  }else{
    err_set(1,"warning: no parameters to optimize.");
    r=MPEOEqrms(n,G.p,G);
  }
  MPEOEfree_scratch(G);
  return(r);
}
/*-------------------- MPEOEfree_scratch --------------------*/

void MPEOEfree_scratch(MPEOEinfo G)
{
  if (G.p) free(G.p);
  if (G.dp) free(G.dp);
  if (G.Mpa) free(G.Mpa);
  if (G.Mpb) free(G.Mpb);
  if (G.Q) free(G.Q);
  if (G.A) free(G.A);
  if (G.B) free(G.B);
  if (G.q) free(G.q);
  if (G.iQ) free(G.iQ);
}
/*-------------------- MPEOEinit_scratch --------------------*/

MPEOEinfo MPEOEinit_scratch(void)
{
  /* return empty scratch pad... */

  MPEOEinfo G;

  G.p=(float *)NULL; G.dp=(float *)NULL;
  G.Mpa=(MPEOEatom **)NULL; G.Mpb=(MPEOEbond **)NULL;
  G.Q=(float *)NULL; G.iQ=(float *)NULL; G.A=(Atom **)NULL; 
  G.B=(Bond **)NULL; G.q=(float *)NULL;
  G.nat=0; G.nbnd=0; G.npara=0; G.nparb=0;
  G.Latom=(MPEOEatom *)NULL;
  G.Lbond=(MPEOEbond *)NULL;

  return(G);
}
/*-------------------- MPEOEsetup --------------------*/

MPEOEinfo MPEOEsetup(Nodechain *A)
{
  /* return initialized scratch pad for MPEOE calculation on atomchain A */

  MPEOEinfo G;
  Atom *a;
  MPEOEatom *mpa;
  MPEOEbond *mpb;
  int i,j,k,bmask,umask,ubmask;
  Prule *pr;
  
  G=MPEOEinit_scratch();
  G.Latom=&MPEOEAtoms;
  G.Lbond=&MPEOEBonds;

  connected_atoms(A);    /* MPEOE method is valid only for connected components */

  /* counting atoms, bonds, etc... */

  G.nat=A->n;
  for (a=(Atom *)A->s;mvld(a);a=(Atom *)a->chain) {
    for (j=0;j<a->nbnd;j++) if (a->bnd[j]->f==a&&in_chain(a->bnd[j]->t,A)) G.nbnd++;
  }
  for (mpa=MPEOEAtoms.nxt;mvld(mpa);mpa=mpa->nxt) mpa->nu=0;
  for (mpb=MPEOEBonds.nxt;mvld(mpb);mpb=mpb->nxt) mpb->nu=0; 

  if (G.nat>0) {

    /* allocate scratch memory... */

    G.Mpa=(MPEOEatom **)malloc(G.nat*sizeof(MPEOEatom *));
    G.Q=(float *)malloc(G.nat*sizeof(float));
    G.iQ=(float *)malloc(G.nat*sizeof(float));
    G.A=(Atom **)malloc(G.nat*sizeof(Atom *));
    if (G.nbnd>0) {
      G.Mpb=(MPEOEbond **)malloc(G.nbnd*sizeof(MPEOEbond *));
      G.B=(Bond **)malloc(G.nbnd*sizeof(Bond *));
      G.q=(float *)malloc(G.nbnd*sizeof(float));
    }

    if (G.Mpa&&G.Q&&G.iQ&&G.A&&(G.nbnd<1||(G.B&&G.Mpb&&G.q))) {

      /* set up auxillary arrays */

      for (k=0,i=0,a=(Atom *)A->s;mvld(a);a=(Atom *)a->chain,i++) {
	G.A[i]=a;
	G.Mpa[i]=&MPEOEAtoms;
	for (mpa=G.Mpa[i]->nxt;mvld(mpa);mpa=mpa->nxt) {
	  for (pr=mpa->p;mvld(pr);pr=(Prule *)pr->nxt) {
	    if (matchPruleAtom(a,pr,0,0)) {
	      G.Mpa[i]=mpa;
	      mpa->nu++;
	      break;
	    }
	  }
	  if (pr) break;
	}
	a->seqnr=i;
	for (j=0;j<a->nbnd;j++) if (a->bnd[j]->f==a&&in_chain(a->bnd[j]->t,A)) G.B[k++]=a->bnd[j];
      }
      for (i=0;i<G.nbnd;i++) {
	G.Mpb[i]=&MPEOEBonds;
	for (mpb=MPEOEBonds.nxt;mvld(mpb);mpb=mpb->nxt) {
	  for (pr=mpb->p;mvld(pr);pr=(Prule *)pr->nxt) {
	    if (matchPruleBond(G.B[i],pr,0,0)) {
	      G.Mpb[i]=mpb;
	      mpb->nu++;
	      break;
	    }
	  }
	  if (pr) break;
	}
      }

      /* initial charge distribution, derived from formal charges, fragment corrections */

      bmask=(1<<7); umask=(1<<6); ubmask=bmask|umask;
      for (i=0;i<G.nat;i++) G.iQ[i]=FormalChrg(G.A[i]);
      if (mvld(MPEOELib)) {
	forPrules(pr,MPEOELib) {
	  for (i=0;i<G.nat;i++) {
	    if (matchPruleAtom(G.A[i],pr,0,bmask)) {
	      for(j=0;j<pr->nat;j++) if (!node_in_chain((Node*)pr->atom[j].match,A)) break;
	      if (j==pr->nat) {   /* apply correction only if all matched atoms in A */
		for (j=0;j<pr->nat;j++) {
		  k=pr->atom[j].match->seqnr;
		  G.iQ[k]+=pr->atom[j].q-FormalChrg(G.A[k]);
		}
		for (j=0;j<pr->nbnd;j++) {
		  pr->bond[j].match->flg|=(pr->bond[j].cover?ubmask:umask);
		}
	      }
	    }
	  }
	}
	for (i=0;i<G.nbnd;i++) G.B[i]->flg&=~ubmask;
      }
    }else{
      err_set(1,"failed to allocate scratch memeory for MPEOE");
      G.nat=0;    /* to flag that somethoing went wrong... */
    }
  }
  return(G);
}
/*-------------------- MPEOEsetup_refinement --------------------*/

MPEOEinfo MPEOEsetup_refinement(Nodechain *A)
{
  /* setup MPEOE scratchpad for MPEOE parameter refinement */

  MPEOEinfo G;
  MPEOEatom *mpa;
  MPEOEbond *mpb;
  int k;

  G=MPEOEsetup(A);

  if (G.nat>0) {
    for (G.npara=0,mpa=MPEOEAtoms.nxt;mvld(mpa);mpa=mpa->nxt) {
      if (mpa->nu>=mpOpt.minocc) {
	if (mpa->opt[0]) G.npara++;
	if (mpa->opt[1]) G.npara++;
	if (mpa->indX&&mpa->opt[2]) G.npara++;
      }
    }
    for (G.nparb=0,mpb=MPEOEBonds.nxt;mvld(mpb);mpb=mpb->nxt) if (mpb->opt&&mpb->nu>=mpOpt.minocc) G.nparb++;
    if (G.npara+G.nparb>0) {
      G.p=(float *)malloc((G.npara+G.nparb)*sizeof(float));
      G.dp=(float *)malloc((G.npara+G.nparb)*sizeof(float));
      if (!mvld(G.dp)||!mvld(G.p)) {
	err_set(1,"failed to allocate scratch memeory for MPEOE refinement");
	G.npara=0; G.nparb=0;
      }else{
	k=0;
	for (mpa=MPEOEAtoms.nxt;mvld(mpa);mpa=mpa->nxt) {
	  if (mpa->nu>=mpOpt.minocc) {
	    if (mpa->opt[0]) {G.p[k]=mpa->P[0]; G.dp[k++]=ransign()*mpOpt.dChi;}
	    if (mpa->opt[1]) {G.p[k]=mpa->P[1]; G.dp[k++]=ransign()*mpOpt.dChi;}
	    if (mpa->indX&&mpa->opt[2]) {G.p[k]=mpa->P[2]; G.dp[k++]=ransign()*mpOpt.dChi;}
	  }
	}
	for (mpb=MPEOEBonds.nxt;mvld(mpb);mpb=mpb->nxt) {
	  if (mpb->opt&&mpb->nu>=mpOpt.minocc) {
	    G.p[k]=mpb->f; G.dp[k++]=ransign()*mpOpt.dF;
	  }
	}
      }
    }
  }
  return(G);
}
/*-------------------- MPEOEreport_missing --------------------*/

int MPEOEreport_missing(MPEOEinfo G,int prt)
{
  /* report atoms and bonds with missing MPEOE parameters */

  int i,fr,to,r;
  char buf1[4096],buf2[4096];
  
  r=0;
  if (prt) err_dump();
  for (i=0;i<G.nat;i++) {
    if (!G.Mpa[i]->def) {
      r++;
      if (prt) err_set(1,"MPEOE parameters missing for %s",
		       nodeptr_to_name(buf1,(Node *)G.A[i]));
    }
  }
  if (prt) err_dump();
  for (i=0;i<G.nbnd;i++) {
    if (!G.Mpb[i]->def) {
      r++;
      fr=G.B[i]->f->seqnr; to=G.B[i]->t->seqnr;
      if (prt) err_set(1,"MPEOE parameters missing for %s -- %s",
		       nodeptr_to_name(buf1,(Node *)G.A[fr]),
		       nodeptr_to_name_ctxt(buf2,(Node *)G.A[to],(Node *)G.A[fr]));
      
    }
  }
  if (prt) err_dump();
  return(r);
}
/*-------------------- MPEOEiterate --------------------*/

#define MPEOE_MAXDQ 1.0   /* limit max. charge shift per iteration */

float MPEOEiterate(MPEOEinfo G,int nit,float dq)
{
  /* iterative partial equilibration of electonegativites, starting from current
     partial charges. Maximum of n iterations, stop if max charge shift < dq.
     return max shift in last iteration */

  Bond *b;
  float qmax,enf,ent;
  int i,n,fr,to;

  n=0;
  for(n=1,qmax=dq+1;qmax>dq&&n<=nit;n++) {
    for (i=0;i<G.nbnd;i++) {
      b=G.B[i]; fr=b->f->seqnr; to=b->t->seqnr;
      enf=G.Mpa[fr]->P[0]+G.Q[fr]*G.Mpa[fr]->P[1];
      ent=G.Mpa[to]->P[0]+G.Q[to]*G.Mpa[to]->P[1];
      if (ent>enf) G.q[i]=powf(G.Mpb[i]->f,n)*(ent-enf)/G.Mpa[fr]->P[2];
      else G.q[i]=powf(G.Mpb[i]->f,n)*(ent-enf)/G.Mpa[to]->P[2];
      G.q[i]=(G.q[i]<-MPEOE_MAXDQ ? -MPEOE_MAXDQ : (G.q[i]> MPEOE_MAXDQ ? MPEOE_MAXDQ : G.q[i]));
    }
    if (MPEOEgran!=0) for (i=0;i<G.nbnd;i++) G.q[i]=MPEOEgran*rint(G.q[i]/MPEOEgran);
    for (qmax=0.0,i=0;i<G.nbnd;i++) {
      fr=G.B[i]->f->seqnr; to=G.B[i]->t->seqnr;
      G.Q[fr]+=G.q[i]; G.Q[to]-=G.q[i];
      qmax=(qmax<fabs(G.q[i]) ? fabs(G.q[i]) : qmax);
    }
  }
  return((nit>0?qmax:0));
}
/*-------------------- MPEOEqrms --------------------*/

float MPEOEqrms(int n,float *p,MPEOEinfo G)
{
  /* compute rms of differences between partial charges on molecule, and
     partial charges computed for MPEOE parameteres in p 
     Note: atomic site occupancy factors are used as weights.    */

  float dq,w,wsum;
  int i;

  MPEOEupdate(n,p,G);         /* update MPEOE parameters from p */
  for (i=0;i<G.nat;i++) G.Q[i]=G.iQ[i];
  dq=MPEOEiterate(G,mpOpt.mxcomp,mpOpt.tolcomp);
  for (wsum=0,dq=0,i=0;i<G.nat;i++) {
    w=G.A[i]->x[ATM_SOF];
    wsum+=w;
    dq+=w*(G.A[i]->q-G.Q[i])*(G.A[i]->q-G.Q[i]);
  }
  if (wsum>0) dq=sqrt(dq/wsum);
  return(dq);
}
/*-------------------- MPEOEupdate --------------------*/

void MPEOEupdate(int n,float *p,MPEOEinfo G)
{
  /* copy parameters from p to MPEOE Library             */
  /* enforce 0.1 < A, 0.01 < B, 0.1 < X, MINF <= f <= MAXF   */ 

  int k;
  MPEOEatom *mpa;
  MPEOEbond *mpb;

  if (n!=(G.npara+G.nparb)) return;    /* numper of parameters does not match G */
  for (k=0,mpa=MPEOEAtoms.nxt;k<n&&mvld(mpa);mpa=mpa->nxt) {
    if (mpa->nu>=mpOpt.minocc) {
      if (mpa->opt[0]) {mpa->P[0]=p[k++]; if (mpa->P[0] < 0.1) mpa->P[0]=0.1;}
      if (mpa->opt[1]) {mpa->P[1]=p[k++]; if (mpa->P[1] < 0.01) mpa->P[1]=0.01;}
      if (mpa->indX) {
	if (mpa->opt[2]) {mpa->P[2]=p[k++]; if (mpa->P[2]<0.1) mpa->P[2]=0.1;}
      }else if (mpa->opt[0]||mpa->opt[1]) mpa->P[2]=mpa->P[0]+mpa->P[1];
    }
  }
  for (mpb=MPEOEBonds.nxt;k<n&&mvld(mpb);mpb=mpb->nxt) {
    if (mpb->opt&&mpb->nu>=mpOpt.minocc) {
      mpb->f=p[k++]; 
      mpb->f=(mpb->f < mpOpt.minf ? mpOpt.minf : (mpb->f>mpOpt.maxf ? mpOpt.maxf : mpb->f));
    }
  }
}
/*-------------------- MPEOEdump -------------------*/

#define OPTF(x) (x ? "*" : "")

void MPEOEdump(FILE *out)
{
  /* Dump current MPEOE library to output file. If no output file is given,
     just list the variable parameters on the terminal. */

  MPEOEatom *mpa;
  MPEOEbond *mpb;
  char buf[4096];
  Prule *pr;
  int k,no;

  if (out) {
    fprintf(out,"MPEOE reset\n");
    if (mvld(MPEOELib)) {
      fprintf(out,"\n");
      forPrules(pr,MPEOELib) {
	fprintf(out,"MPEOE init %s {",pr->name);
	for (k=0;k<pr->nat;k++) fprintf(out,"%s%0.3f",(k>0 ? ",":""),pr->atom[k].q);
	fprintf(out,"}\n");
      }
    }
  }
  for (no=0,mpa=&MPEOEAtoms;mpa;mpa=mpa->nxt) {
    if (mpa->def) {
      sprintf(buf,"MPEOE atom %s {%0.3f%s,%0.3f%s",
	      mpa->def,mpa->P[0],OPTF(mpa->opt[0]),mpa->P[1],OPTF(mpa->opt[1]));
      if (mpa->indX) sprintf(buf+strlen(buf),",%0.3f%s}",
			     mpa->P[2],OPTF(mpa->opt[2]));
      else sprintf(buf+strlen(buf),"}");
      if (out) fprintf(out,"%s%s\n",no++==0?"\n":"",buf);
      else if (mpa->opt[0]||mpa->opt[1]||mpa->opt[2]) prtGraphics("%s\n",buf);
    }
  }
  for (no=0,mpb=&MPEOEBonds;mpb;mpb=mpb->nxt) {
    if (mpb->def) {
      sprintf(buf,"MPEOE bond %s %0.3f%s",mpb->def,mpb->f,OPTF(mpb->opt));
      if (out) fprintf(out,"%s%s\n",no++==0?"\n":"",buf); 
      else if (mpb->opt) prtGraphics("%s\n",buf);
    }
  }
}
/*-------------------- MPEOEgetOpt etc... --------------------*/

MPEOEopt MPEOEgetOpt(void)
{
  return(mpOpt);
}

void MPEOEsetOpt(MPEOEopt opt)
{
  mpOpt=opt;

  /* sanity checks... */

  if (mpOpt.minocc<1) mpOpt.minocc=1;
  if (mpOpt.repf<0) mpOpt.repf=0;
  if (mpOpt.tolref<0.00000001) mpOpt.tolref=0.00000001;
}
/*-------------------- MPEOEaddInitRule --------------------*/

int MPEOEaddInitRule(Prule *pr)
{
  /* Add fragment rule for initial distribution of formal charges */

  int r;

  r=FALSE;
  if (!mvld(MPEOELib)) MPEOELib=make_node(OBJ_ROOT);
  if (mvld(MPEOELib)) {
    pr->up=MPEOELib;
    if (!link_node((Node *)pr)) pr->up=(Node *)NULL;
    else r=TRUE;
  }else;
  return(r);
}
/*-------------------- MPEOEresetInitRules --------------------*/

void MPEOEresetInitRules(void)
{
  /* reset the MPEOE library (simply delete it)... */
  MPEOELib=del_node(MPEOELib);

  /* ... and reset default atom and bond parameters */

  if (mvld(MPEOEAtoms.def)) free(MPEOEAtoms.def); 
  MPEOEAtoms.def=(char *)NULL;
  MPEOEAtoms.P[0]=1.e6;   MPEOEAtoms.opt[0]=FALSE;
  MPEOEAtoms.P[1]=1.0;    MPEOEAtoms.opt[1]=FALSE;
  MPEOEAtoms.P[2]=1.e20;  MPEOEAtoms.opt[2]=FALSE;  MPEOEAtoms.indX=TRUE;
  MPEOEAtoms.nu=0;
  
  if (mvld(MPEOEBonds.def)) free(MPEOEBonds.def); 
  MPEOEBonds.def=(char *)NULL;
  MPEOEBonds.f=0.0;       MPEOEBonds.opt=FALSE;
  MPEOEBonds.nu=0.0;

}
/*-------------------- MPEOEsetGranularity --------------------*/

void MPEOEsetGranularity(int n)
{
  switch (n) {
  case 0:
    MPEOEgran=0.0;
    break;
  case 1:
    MPEOEgran=0.1;
    break;
  case 2:
    MPEOEgran=0.01;
    break;
  case 3:
    MPEOEgran=0.001;
    break;
  case 4:
    MPEOEgran=0.0001;
    break;
  case 5:
    MPEOEgran=0.00001;
    break;
  case 6:
    MPEOEgran=0.000001;
    break;
  default:
    MPEOEgran=0.0;
    break;
  }
}
