#include <math.h>
#include "errhand.h"
#include "swdef.h"
#include "mystdlib.h"
#include "bhtree.h"
#include "xbhtree.h"

#define MAXTREE 100000
#define MAXCLS  4196
#define MAX_TEMPLATES 250

static int clipper(int n,float *x,float *y,float r);
static void sortNeighbours(int n,Atom **B,float *d);

int grid_fn_ConnollyDist(Grid *g,Nodechain *c,float r)
{
  /* Compute distance from Connolly surface generated by atoms in c */
  /* Probe radius is r. Distance is positive for points inside surface */

  int status=0,n=0,nn,nd=0,*np=NULL,nt=0,i,j,k,grain;
  float d;
  float dist[MAXCLS],dmax=0,*x=NULL,**X=NULL,*v,**pts=NULL,p[3],*s,D,st;
  Atom *a,**A=NULL,*B[MAXCLS];
  BHtree *bht=NULL;
  xBHtree *xbht=NULL;
  dot_template *T[MAX_TEMPLATES],*t;
 
  /* do distance to vdW surface first... */
  status=grid_fn_vdwdist(g,c);

  /* count atoms in or near grid */
  if (!status) {
    dmax=0;
    for (n=0,a=(Atom *)c->s;is_atom(a);a=(Atom *)a->chain) {
      if (a->atp) {
	d=a->atp->rvdw+r;
	if (a->x[0]>g->xmin[0]-d &&
	    a->x[1]>g->xmin[1]-d &&
	    a->x[2]>g->xmin[2]-d &&
	    a->x[0]<g->xmax[0]+d &&
	    a->x[1]<g->xmax[1]+d &&
	    a->x[2]<g->xmax[2]+d) n++;
	if (d>dmax) dmax=d;
      }
    }
    if (n>0) {
      A=(Atom **)malloc(n*sizeof(Atom *));
      X=malloc(n*sizeof(float *));
      np=malloc(n*sizeof(int));
      if (A) {
	for (n=0,a=(Atom *)c->s;is_atom(a);a=(Atom *)a->chain) {
	  if (a->atp) {
	    d=a->atp->rvdw+r;
	    if (a->x[0]>g->xmin[0]-d &&
		a->x[1]>g->xmin[1]-d &&
		a->x[2]>g->xmin[2]-d &&
		a->x[0]<g->xmax[0]+d &&
		a->x[1]<g->xmax[1]+d &&
		a->x[2]<g->xmax[2]+d) A[n++]=a;
	  }
	}
	bht=GenerateBHtree(n,A,n<MAXTREE?1:n/MAXTREE);
      }
    }else{
      err_set(1,"grid_fn_ConnollyDist: found no atoms in grid.");
      status=1;
    }
  }    

  if (A && X && np && bht && !status) {

    /* generate LR(r) dotted surface for atoms in or near grid                */
    /* surface will be in (X,np), where X[i] points to the dots coordinates,  */
    /* and np[i] is the number of dots fo atom A[i].                          */          

    for (i=0;i<n;i++) X[i]=NULL;
    for (i=0;i<n;i++) np[i]=0;
 
    for (i=0;i<n;i++) {
      d=A[i]->atp->rvdw+r;
      t=find_best_template(d,nt,T);
      if (!t||(fabs(t->R-d)>0.025 && nt<MAX_TEMPLATES)) {
	T[nt++]=t=make_dot_template(d,g->dx[0]);
	if (!t)  nt--;
      }
      if (t&&t->n>0) {
	x=malloc(3*t->n*sizeof(float));
	if (x) {
	  for (j=0;j<t->n;j++) 
	    for (k=0;k<3;k++) x[3*j+k]=A[i]->x[k]+t->x[3*j+k];
	  nn=findBHcloseAtoms(bht,A[i]->x,d+dmax,B,dist,MAXCLS);
	  if (nn>MAXCLS) {
	    err_set(1,"neighbour list overflow.");
	    nn=MAXCLS;
	  }
	  sortNeighbours(nn,B,dist);
	  for (j=0,np[i]=t->n;j<nn&&np[i]>0;j++) {
	    if (B[j]!=A[i]&&dist[j]<d+B[j]->atp->rvdw+r) {
	      np[i]=clipper(np[i],x,B[j]->x,B[j]->atp->rvdw+r);
	    }
	  }
	  if (np[i]>0) X[i]=realloc(x,3*np[i]*sizeof(float));
	  else free(x);
	  if (!X[i]) np[i]=0;
	  else nd+=np[i];
	  x=NULL;
	}
      }else{
	err_set(1,"failed to generate dots template for sphere with radius %0.3f\n",d);
      }
    }
    
  }else{
    err_set(1,"grid_fn_ConnollyDist: to alllocate scratch data.");
    status=2;
  }

#ifdef xxx
  /* debug: send LR dots to graphics */
  for (i=0;i<n;i++) {
    if (np[i]>0&&cPtr(A[i]->cross)) send_dots_np(A[i],np[i],X[i]);
  }
#endif

  prtGraphics("%d dots on LR surface.\n",nd);

  /* compute distance from Connolly(r) surface */

  if (!status) {

    /* put LR(r) points into xBHtree structure for fast distance */
    pts=malloc(nd*sizeof(float *));
    if (pts) {
      for (nd=0,i=0;i<n;i++) {
	for (j=0;j<np[i];j++) pts[nd++]=X[i]+3*j;
      }
      grain=g->n[0]*g->n[1]*g->n[2]<MAXTREE?1:g->n[0]*g->n[1]*g->n[2]/MAXTREE;
      xbht=GeneratexBHtree(nd,pts,grain);
    }

    if (xbht) {
      D=sqrt((g->xmax[0]-g->xmin[0])*(g->xmax[0]-g->xmin[0])+
	       (g->xmax[1]-g->xmin[1])*(g->xmax[1]-g->xmin[1])+
	       (g->xmax[2]-g->xmin[2])*(g->xmax[2]-g->xmin[2]));
      st=g->dx[0];
      for (p[2]=g->xmin[2],k=0;k<g->n[2];k++,p[2]+=g->dx[2]) {
	dotdotdot("connolly",k,g->n[2],60);
	for (p[1]=g->xmin[1],j=0;j<g->n[1];j++,p[1]+=g->dx[1]) {
	  v=g->val+g->n[0]*(g->n[1]*k+j);
	  for (d=D,p[0]=g->xmin[0],i=0;i<g->n[0];i++,p[0]+=g->dx[0],v++) {
	    if (-*v<r) {
	      *v=-r;
	      s=findxBHclosest(xbht,p,d,&d);
	      if (s) *v=d-r;
	      else prtGraphics("fugu %d %d %d\n",i,j,k);
	    }else *v=-r;
	    d+=st;
	  }
	}
      }
      dotdotdot("",0,0,0);
    }else{
      err_set(1,"failed to generate binary spatial division tree for LR dots.");
      status=3;
    }
  }


  /* free scratch data */

  if (xbht) freexBHtree(xbht);
  if (pts) free(pts);
  for (i=0;i<nt;i++) free_dot_template(T[i]);
  if (x) free(x);
  freeBHtree(bht);
  if (X) {
    for (i=0;i<n;i++) if (X[i]) free(X[i]);
    free(X);
  }
  if (A) free(A);
  
  return(status);
}

static int clipper(int n,float *x,float *y,float r)
{
  int i,no=0;
  float D,d;

  D=r*r;

  for (no=0,i=0;i<n;i++) {
    d=(x[3*i]-y[0])*(x[3*i]-y[0])+
      (x[3*i+1]-y[1])*(x[3*i+1]-y[1])+
      (x[3*i+2]-y[2])*(x[3*i+2]-y[2]);
    if (d>D) {
      if (no<i) {
	x[3*no]=x[3*i];
	x[3*no+1]=x[3*i+1];
	x[3*no+2]=x[3*i+2];
      }
      no++;
    }
  }
  return(no);
}

static void sortNeighbours(int n,Atom **B,float *d)
{
  /* sorting might help to use most efficient clippers first... */
  /* not implemented. */
}
