#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "transpor.h"
#include "swdef.h"
#include "errhand.h"
#include "cyana.h"

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

static void markComponent(cyaAtom_t *A,int c)
{
  int i;
  
  A->c=c;
  for (i=0;i<A->nn;i++) if (A->n[i]->c==0) markComponent(A->n[i],c);
}

static int markComponents(cyaResidue_t *L)
{
  int nc=0,i;

  if (L->nat>0) {
    for (i=0;i<L->nat;i++) L->A[i]->c=0;
    for (i=0;i<L->nat;i++) {
      if (L->A[i]->c==0) markComponent(L->A[i],++nc);
    }
  }
  return nc;
}

static void recursiveAddTorDefs(cyaResidue_t *L,cyaAtom_t *C)
{
  int j;
  cyaAtom_t *T[4];

  if (C->up) {
    if (C->nn>1 && !C->nb[0]->rigid) {
      T[1]=C->up;
      for (j=1;j<C->nn;j++) if (C->n[j]->up==C) break;
      if (j==C->nn) return;
      T[3]=C->n[j];
      if (T[1]->up) T[0]=T[1]->up;
      else {
	for (j=0;j<T[1]->nn;j++) if (T[1]->n[j]!=C && T[1]->n[j]->up==T[1]) break;
	T[0]=T[1]->n[j];
      }
      T[2]=C;
      L->tor[L->ntor++]=newCyaTorDef(T,T[2]->lst);
    }
  }
  for (j=0;j<C->nn;j++) if (C->n[j]->up==C) recursiveAddTorDefs(L,C->n[j]);
}

static int compareAtomNames(const void *A,const void *B)
{
  cyaAtom_t *a,*b;
  int r;

  a=*(cyaAtom_t **)A;
  b=*(cyaAtom_t **)B;

  r=strcmp(a->name,b->name);
  return(r);
}

/*-------------------------------------------------------------------------*/

cyaResidue_t *cyanaResidue(Node *root, cyaOptions_t Options)
{
  cyaResidue_t *L;
  cyaAtom_t *R;
  int n;

  L=collectAtomsAndBonds(root);
  if (L) {
    if (L->nc==1) {
      n=cyaCheckAtomNames(L,TRUE);
      if (n>1) err_set(1,"WARNING: %d bad atom names (too long, or not unique)",n);
      else if (n==1) err_set(1,"WARNING: %d bad atom name (too long, or not unique)",n);
      cyaAssignUniqueTypes(L);
      R=cyaBestRoot(L);
      if (R) cyaNewRoot(L,R);
      cyaMarkRigidBonds(L,Options);
      cyaFindTorsions(L);
    }else if (L->nc>1) {
      err_set(1,"the selected molecule has %d connected components.",L->nc);
      L=freeCyaResidue(L);
    }else{
      err_set(1,"generation of connectivity tree failed.");
      L=freeCyaResidue(L);
    }
  }else{
    err_set(1,"failed to collect atoms and bonds into Cyana residue object.");
  }
  return(L);
}

cyaResidue_t *collectAtomsAndBonds(Node *root)
{
  cyaResidue_t *L;
  Atom *a,*b;
  int i,nh=0;

  L=newCyaResidue();
  if (L) {
    L->root=root;
    ForAtoms(a,root) {
      if (a->atp->z>0) {
	if (a->atp->z==1) nh++;
	a->seqnr=L->nat;
	L->nat++;
	L->nbnd+=a->nbnd;
      }else a->seqnr=-1;
    }
    L->nbnd/=2;
    L->A=malloc((L->nat+nh)*sizeof(cyaAtom_t *));  /* room for up to nh pseudo atoms */
    L->B=malloc(L->nbnd*sizeof(cyaBond_t *));
    if (L->A && L->B) {
      L->nat=0;
      ForAtoms(a,root) {
	if (a->seqnr>=0) {
	  L->A[a->seqnr]=newCyaAtom(a);
	  L->A[a->seqnr]->seqnr=a->seqnr;
	  L->A[a->seqnr]->oseq=a->seqnr;
	  L->nat++;
	}
      }
      L->nbnd=0;
      ForAtoms(a,root) {
	if (a->seqnr>=0) {
	  for (i=0;i<a->nbnd;i++) {
	    b=neighbour(i,a);
	    if (b->seqnr > a->seqnr) {
	      L->B[L->nbnd]=newCyaBond(L->A[a->seqnr],L->A[b->seqnr]);
	      L->nbnd++;
	    }
	  }
	}
      }
      for (i=0;i<L->nbnd;i++) {
	if (L->B[i]->f->nn < CYAMAXNB) {
	  L->B[i]->f->n[L->B[i]->f->nn]=L->B[i]->t;
	  L->B[i]->f->nb[L->B[i]->f->nn++]=L->B[i];
	}
	if (L->B[i]->t->nn < CYAMAXNB) {
	  L->B[i]->t->n[L->B[i]->t->nn]=L->B[i]->f;
	  L->B[i]->t->nb[L->B[i]->t->nn++]=L->B[i];
	}
      }
    }
    L->nc=markComponents(L);
    L->ring=malloc((L->nbnd+L->nc-L->nat)*sizeof(cyaRing_t *));
    L->nrings=0;
    if (L->nc==1) {
      cyaResidueRings(L);
      cyaBFTree(L);
    }
  }
  return(L);
}

int cyaCheckAtomNames(cyaResidue_t *L,int verbose)
{
  int i,status=0;
  cyaAtom_t **A=NULL;

  if (L&&L->nat>0) {
    A=malloc(L->nat*sizeof(cyaAtom_t *));
    if (A) {
      for (i=0;i<L->nat;i++) A[i]=L->A[i];
      qsort(A,L->nat,sizeof(cyaAtom_t *),compareAtomNames);
      if (strlen(A[0]->name)>4) {
	status+=1;
	if (verbose) err_set(1,"atom name too long: %s",A[0]->name);
      }else{
	for (i=1;i<L->nat;i++) {
	  if (strlen(A[i]->name)>4 || streq(A[i]->name,A[i-1]->name)) status++;
	  if (verbose) {
	    if (strlen(A[i]->name)>4) err_set(1,"atom name too long: %s",A[i]->name);
	    if (streq(A[i]->name,A[i-1]->name)) err_set(1,"atom name not unique: %s",A[i]->name);
	  }
	}
      }
    }
  }
  if (A) free(A);
  return status;
}

void cyaMarkRigidBonds(cyaResidue_t *L,cyaOptions_t Options)
{
  int i,j,k,n,nr,done,rigrings;

  if (L) {
    rigrings=Options.rigidRings;

    for (i=0;i<L->nbnd;i++) {
      if (L->B[i]->f->nn==1 || L->B[i]->t->nn==1 ||
	  L->B[i]->bnd->bt==BT_AROMATIC ||
	  L->B[i]->bnd->bt==BT_DOUBLE ||
	  L->B[i]->bnd->bt==BT_AMIDE ||
	  (L->B[i]->rs>0 && (rigrings<1 || L->B[i]->rs < rigrings))
	  ) {
	L->B[i]->rigid=TRUE;
      }else{
	L->B[i]->rigid=FALSE;
      }
    }

    /* small bridged ring systems */
    if (Options.bridgedRings) {
      for (i=0;i<L->nrings-1;i++) {
	if (L->ring[i]->rs<7) {
	  for (k=0;k<L->ring[i]->rs;k++) L->ring[i]->bnd[k]->flg=1;
	  for (j=i+1;j<L->nrings;j++) {
	    for (n=0,k=0;k<L->ring[j]->rs;k++) n+=L->ring[j]->bnd[k]->flg;
	    if (n>1&&L->ring[i]->rs-n<=3&&L->ring[j]->rs-n<=3) {
	      for (k=0;k<L->ring[i]->rs;k++) L->ring[i]->bnd[k]->rigid=TRUE;
	      for (k=0;k<L->ring[j]->rs;k++) L->ring[j]->bnd[k]->rigid=TRUE;
	    }
	  }
	  for (k=0;k<L->ring[i]->rs;k++) L->ring[i]->bnd[k]->flg=0;
	}
      }
    }
  
    /* small rings with < 4 rotatable bonds */
    for (done=FALSE;!done;) {
      done=TRUE;
      for (i=0;i<L->nrings;i++) {
	if (L->ring[i]->rs<8) {
	  for (nr=0,j=0;j<L->ring[i]->rs;j++) if (L->ring[i]->bnd[j]->rigid) nr++;
	  if (nr<L->ring[i]->rs && L->ring[i]->rs-nr<4) {
	    for (nr=0,j=0;j<L->ring[i]->rs;j++) L->ring[i]->bnd[j]->rigid=TRUE;
	    done=FALSE;
	  }
	}
      }
    }
  }
}

void cyaFindTorsions(cyaResidue_t *L)
{
  int i,j,n;

  if (L) {
    if (L->ntor>0) {
      for (i=0;i<L->ntor;i++) freeCyaTorDef(L->tor[i]);
    }
    for (n=0,i=0;i<L->nbnd;i++) if (!L->B[i]->rc && !L->B[i]->rigid) n++;
    L->tor=realloc(L->tor,n*sizeof(cyaTorDef_t *));
    L->ntor=0;
    recursiveAddTorDefs(L,L->A[0]);
    for (i=0;i<L->nat;i++) L->A[i]->ft=NULL;
    for (i=0;i<L->ntor;i++) {
      for (j=L->tor[i]->A[2]->seqnr+1;j<=L->tor[i]->A[4]->seqnr;j++) L->A[j]->ft=L->tor[i];
    }
  }
}

