#include <stdlib.h>
#include "errhand.h"
#include "cyana.h"

/* structure used by connectivity  sort */

typedef struct srtConn_t {
  cyaAtom_t *atom;
  cyaBond_t *bond;
} srtConn_t;

/*========================================================================*/

/*-------------------- internal support functions --------------------*/

static int scndLength(cyaAtom_t *A)
{
  int lb=0,i,ib=-1;

  for (i=0;i<A->nn;i++) {
    if (A->n[i]->up==A && lb<A->n[i]->lb) {
      lb=A->n[i]->lb;
      ib=i;
    }
  }
  if (ib>=0) {
    lb=0;
    for (i=0;i<A->nn;i++) {
      if (i!=ib && A->n[i]->up==A && lb<A->n[i]->lb) {
	lb=A->n[i]->lb;
      }
    }
  }
  return lb;
}

static void reverseUp(cyaAtom_t *a,cyaAtom_t *b)
{
  if (b) {
    if (b->up!=b) reverseUp(b,b->up);
    b->up=a;
  }
}

static int recursiveCpyAtmPtr(cyaAtom_t **T,int pos,cyaAtom_t *A)
{
  int j;

  A->seqnr=pos;
  T[pos++]=A;
  for (j=0;j<A->nn;j++) {
    if (A->n[j]->up==A) {
      pos=recursiveCpyAtmPtr(T,pos,A->n[j]);
    }
  }
  return pos;
}

static int branchPrior(const cyaAtom_t *A,const cyaAtom_t *B)
{
  int r=0;

  r=A->lb-B->lb;
  if (r==0) r=A->z-B->z;
  if (r==0) r=A->oseq-B->oseq;
  return r;
}

static int reorderBranches(cyaResidue_t *L,int pos,cyaAtom_t *T)
{
  cyaAtom_t *B[CYAMAXNB];
  int j,l,nn;

  /* shortest branch first, H's before other atoms */
  L->A[pos++]=T;
  nn=0;
  for (j=0;j<T->nn;j++) {
    if (T->n[j]->up==T) {
      B[nn++]=T->n[j];
      for (l=nn-1;l>0;l--) {
	if (branchPrior(B[l-1],T->n[j])>0) B[l]=B[l-1];
	else break;
      }
      B[l]=T->n[j];
    }
  }
  for (l=0;l<nn;l++) {
    pos=reorderBranches(L,pos,B[l]);
  }
  return(pos);
}

static int compare_srtConn(const void *A,const void *B)
{
  return(((srtConn_t *)A)->atom->seqnr - ((srtConn_t *)B)->atom->seqnr);
}

static void sortConnectivity(cyaAtom_t *A)
{
  srtConn_t conn[CYAMAXNB];
  int i,n=0,nd=0;
  
  /* put parent first */
  for (i=0;i<A->nn;i++) {
    if (A->up==A->n[i]) {
      conn[n].atom=A->n[i];
      conn[n].bond=A->nb[i];
      n++;
    }
  }
  /* followed by ring-closures */
  for (i=0;i<A->nn;i++) {
    if (A->n[i]->up!=A && A->up!=A->n[i]) {
      conn[n].atom=A->n[i];
      conn[n].bond=A->nb[i];
      n++;
    }
  }
  /* followed by descendants */
  for (i=0;i<A->nn;i++) {
    if (A->n[i]->up==A) {
      conn[n].atom=A->n[i];
      conn[n].bond=A->nb[i];
      n++; nd++;
    }
  }

  qsort(conn+A->nn-nd,nd,sizeof(srtConn_t),compare_srtConn);
  for (i=0;i<A->nn;i++) {
    A->n[i]=conn[i].atom;
    A->nb[i]=conn[i].bond;
  }
}

static void addRing(cyaResidue_t *L,cyaBond_t *bnd)
{
  int j,rs=0,nr;
  cyaAtom_t *A,*B,*x,*a;
  cyaBond_t **bl=NULL;

  A=bnd->f;
  B=bnd->t;

  bl=malloc(L->nbnd*sizeof(cyaBond_t *));

  for (x=B;x;x=x->up) x->flg=2;
  for (x=A;x;x=x->up) if (x->flg==2) break;

  rs=A->dp+B->dp-2*x->dp+1;
  nr=0;
  bl[nr++]=bnd;
  for (a=B;a!=x;a=a->up,nr++) {
    for (j=0;j<a->nn;j++) if (a->up==a->n[j]) break;
    bl[nr]=a->nb[j];
  }
  nr=rs-1;
  for (a=A;a!=x;a=a->up,nr--) {
    for (j=0;j<a->nn;j++) if (a->up==a->n[j]) break;
    bl[nr]=a->nb[j];
  }
  L->ring[L->nrings++]=newCyaRing(rs,bl);

  for (x=B;x;x=x->up) x->flg=0;
  if (bl) free(bl);
}

/*========================================================================*/

/*-------------------- Connectivity tree generators --------------------*/

int cyaBFTree(cyaResidue_t *L)
{
  cyaAtom_t **A,*N;
  int i,j,k,nc=0;

  if (L&&L->nat>1) {
    A=malloc(L->nat*sizeof(cyaAtom_t *));
    if (A) {
      for (i=0;i<L->nat;i++) {
	L->A[i]->dp=0;
	L->A[i]->up=NULL;
      }
      for (i=0;i<L->nat;i++) L->A[i]->flg=0;
      for (nc=0,k=0;k<L->nat;nc++) {
	for (i=0;i<L->nat;i++) if (L->A[i]->dp==0) break;
	A[k]=L->A[i];
	A[k]->dp=1;
	for (i=k,k=i+1;i<k;i++) {
	  for (j=0;j<A[i]->nn;j++) {
	    if (!A[i]->nb[j]->rc) {
	      N=A[i]->n[j];
	      if (N!=A[i]->up) {
		if (N->dp>0) {
		  addRing(L,A[i]->nb[j]);
		  A[i]->nb[j]->rc=TRUE;
		}else{
		  N->up=A[i];
		  N->dp=N->up->dp+1;
		  A[k++]=N;
		}
	      }
	    }
	  }
	}
      }
      for (i=0;i<L->nat;i++) A[i]->seqnr=i;
      free(A);
    }
  }
  if (nc==1) cyaSortTree(L);
  return nc;
}

int cyaSortTree(cyaResidue_t *L)
{
  int i,status=0;
  cyaAtom_t **A;
  
  if (L&&L->nat>1) {
    A=malloc(L->nat*sizeof(cyaAtom_t *));
    if (A) {
      for (i=0;i<L->nat;i++) if (!L->A[i]->up||L->A[i]->up==L->A[i]) break;
      if (i<L->nat) {
	if (recursiveCpyAtmPtr(A,0,L->A[i])==L->nat) {
	  for (i=0;i<L->nat;i++) {
	    A[i]->lb=1;
	    A[i]->lst=NULL;
	  }
	  for (i=L->nat-1;i>0;i--) {
	    if (A[i]->lb+1>A[i]->up->lb) {
	      A[i]->up->lb=A[i]->lb+1;
	    }
	  }
	  reorderBranches(L,0,A[0]);
	  for (i=0;i<L->nat;i++) L->A[i]->seqnr=i;
	  for (i=0;i<L->nat;i++) {
	    sortConnectivity(L->A[i]);
	  }
	  for (i=L->nat-1;i>0;i--) {
	    if (!L->A[i]->lst) L->A[i]->lst=L->A[i];
	    if (L->A[i]->up && !L->A[i]->up->lst) L->A[i]->up->lst=L->A[i]->lst;
	  }
	}else{
	  status=2;
	}
      }else{
	status=1;
      }
      free(A);
    }else{
      status=3;
    }
  }
  return status;
}

cyaAtom_t *cyaBestRoot(cyaResidue_t *L)
{
  cyaAtom_t *bestRoot=NULL,*N;
  int lbest;

  /* NOTE: assumes L is in canonical tree form (i.e. after cyaSortTree) */

  if (L->nat>0) {
    bestRoot=L->A[0];
    lbest=L->A[0]->lb;
    bestRoot->flg=scndLength(L->A[0]);
    N=bestRoot;
    
    /* descend tree, following right-most (longest) branch */
    /* stop at half-height of tree */
    if (N->nn>0) {
      for (N=N->n[N->nn-1];N->dp<L->A[0]->lb;N=N->n[N->nn-1]) {
	N->flg = scndLength(N);
	if (N->flg < N->up->flg+1) N->flg=N->up->flg+1;
	if (lbest>N->flg) {
	  lbest=N->flg > N->lb ? N->flg : N->lb;
	  bestRoot=N;
	}
      }
    }
  }
  return bestRoot;
}

void cyaNewRoot(cyaResidue_t *L,cyaAtom_t *R)
{
  int i;

  if (L&&R&&L->nat>1) {
    if (R->z==1) {         /* avoid hydrogen at root */
      for (i=0;i<R->nn;i++) if (R->n[i]->z>1) break;
      if (i<R->nn) R=R->n[i];
    }
    for (i=0;i<L->nat;i++) if (L->A[i]==R) break;
    if (i<L->nat) {
      reverseUp(R,R->up);
      R->up=NULL;
      cyaSortTree(L);
    }
  }
}
