#include <stdio.h>
#include <stdlib.h>


int gaussj(int n,float *a,float *b)
/*************************************************************************

Linear equation solution by Gauss-Jordan elimination.

n   :  dimension of matrix a.
a   :  row-major input matrix, will be replaced by its inverse.
b   :  right hand side vector, will be replaced by solution.
       b may be a (float *)NULL pointer.

Returns non zero status if solution/inversion fails:

   1 : singular matrix,
   2 : memory allocation failed.
   3 : internal problem

**************************************************************************/
{
  int i,j,k,r,c,*work,*rows,*cols,*used;
  float *p,*q,t,T,v,piv;

  /* allocate scratch space */
  work=(int *)malloc(3*n*sizeof(int));
  if (!work) return(2);   /* failed to allocate scratch space */
  used=work;              /* used[j] TRUE iff row/column j has been "used" */
  rows=used+n;            /* rows[j] row index of pivot in j-th iteration */
  cols=rows+n;            /* cols[j] column index of pivot in j-th iteration */

  for (i=0;i<n;i++) used[i]=0;

  for (i=0;i<n;i++) {
    /* full pivot search: find a[r][c]=max{abs(a[i][j])} */
    for (T=0,r=-1,c=-1,j=0;j<n;j++) {
      if (!used[j]) {
	for (k=0,p=a+j*n;k<n;k++,p++) {
	  if (!used[k]) {
	    t=(*p>=0?*p:-(*p));
	    if (t>=T) {
	      T=t;
	      r=j;
	      c=k;
	    }
	  }
	}
      }
    }

    if (c<0||c>=n) {
      free(work);         /* internal error */
      return(3);
    }
    used[c]++;
    if (r!=c) {           /* swap rows r and c to bring pivot to (c,c) */
      for (j=0,p=a+r*n,q=a+c*n;j<n;j++,p++,q++) {
	v=*p;
	*p=*q;
	*q=v;
      }
      if (b) {
	v=b[r];
	b[r]=b[c];
	b[c]=v;
      }
    }
    rows[i]=r;
    cols[i]=c;
    piv=a[(n+1)*c];         /* pivot is now at position (c,c) */
    if (piv==0) {
      free(work);           /* singular matrix */
      return(1);
    }
    a[(n+1)*c]=1.0;
    for (k=0,p=a+c*n;k<n;k++,p++) *p/=piv;  /* divide pivot row by piv */
    if (b) b[c]/=piv;
    
    for (j=0;j<n;j++) {     /* reduce other rows */
      if (j!=c) {
	v=a[n*j+c];
	for (a[n*j+c]=0.0,k=0,p=a+j*n,q=a+c*n;k<n;k++,p++,q++) *p -= *q*v;
	if (b) b[j]-=b[c]*v;
      }
    }
  }

  /* unscramble columns (in reverse order or row swaps) */
  for (j=n-1;j>=0;j--) {
    if (rows[j]!=cols[j]) {
      for (k=0,p=a+rows[j],q=a+cols[j];k<n;k++,p+=n,q+=n) {
	v=*p;
	*p=*q;
	*q=v;
      }
    }
  }
  free(work);
  return(0);
}
