#include <math.h>
#include <stdio.h>
#include "gcalc.h"
#include "errhand.h"

int gaussj(int n, float *a,float *b);

vec_3 vec3(float *x)
{
  vec_3 v;
  int i;

  for (i=0;i<3;i++) v.c[i]=x[i];
  return(v);
}

mat_3 mat3(float *x)
{
  mat_3 m;
  int i;

  for (i=0;i<3;i++,x+=3) m.r[i]=vec3(x);
  return(m);
}

void v32arr(float *x,vec_3 v)
{
  int i;

  for (i=0;i<3;i++) x[i]=v.c[i];
}

void m32arr(float *x,mat_3 m)
{
  int i;

  for (i=0;i<3;i++,x+=3) v32arr(x,m.r[i]);
}

mat_3 ident_3(void)
{
  mat_3 m;
  int i,j;

  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) m.r[i].c[j]=0.0;
    m.r[i].c[i]=1.0;
  }
  return(m);
}

vec_3 v3n(vec_3 v)
{
/* return normalized vector */

  float s;
  vec_3 r;
  int i;

  for (s=0,i=0;i<3;i++) s+=v.c[i]*v.c[i];
  if (s>0) {
    s=sqrt(s);
    for (i=0;i<3;i++) r.c[i]=v.c[i]/s;
  }else{
    r=v;
  }
  return(r);
}

float v3l(vec_3 v)
{
/* length of vector */

  float s;
  int i;

  for (s=0,i=0;i<3;i++) s+=v.c[i]*v.c[i];
  s=(s>0 ? sqrt(s) : 0);
  return(s);
}

float v3sv3(vec_3 v,vec_3 w)
{
/* dot product */

  float s;
  int i;

  for (s=0,i=0;i<3;i++) s+=v.c[i]*w.c[i];
  return(s);
}

vec_3 v3xv3(vec_3 v,vec_3 w)
{
  vec_3 u;

  u.c[0]=v.c[1]*w.c[2]-v.c[2]*w.c[1];
  u.c[1]=v.c[2]*w.c[0]-v.c[0]*w.c[2];
  u.c[2]=v.c[0]*w.c[1]-v.c[1]*w.c[0];

  return(u);
}

vec_3 v3mv3(vec_3 v,vec_3 w)
{
/*  u = v - w */

  vec_3 u;

  u.c[0]=v.c[0]-w.c[0];
  u.c[1]=v.c[1]-w.c[1];
  u.c[2]=v.c[2]-w.c[2];

  return(u);
}

vec_3 v3pv3(vec_3 v,vec_3 w)
{
/*  u = v + w */

  vec_3 u;

  u.c[0]=v.c[0]+w.c[0];
  u.c[1]=v.c[1]+w.c[1];
  u.c[2]=v.c[2]+w.c[2];

  return(u);
}

vec_3 sxv3(float s,vec_3 v)
{
/* scalar times vector */

  vec_3 u;

  u.c[0]=s*v.c[0];
  u.c[1]=s*v.c[1];
  u.c[2]=s*v.c[2];

  return(u);
}

vec_4 sxv4(float s,vec_4 v)
{
/* scalar times vector */

  vec_4 u;

  u.c[0]=s*v.c[0];
  u.c[1]=s*v.c[1];
  u.c[2]=s*v.c[2];
  u.c[3]=s*v.c[3];

  return(u);
}

vec_3 v3ds(vec_3 v,float s)
{
/* vector v divided by s, no check for overflow... */

  vec_3 u;

  u.c[0]=v.c[0]/s;
  u.c[1]=v.c[1]/s;
  u.c[2]=v.c[2]/s;

  return(u);
}

float cdistsq(float *x,float *y)
{
/* cartesian distance between points x and y */

  float s;
  int i;
  for (s=0,i=0;i<3;i++) s+=(x[i]-y[i])*(x[i]-y[i]);
  return(s);
}

float cdist(float *x,float *y)
{
  return(sqrt(cdistsq(x,y)));
}

float cangle(float *x,float *y,float *z)
{
/* angle (radians)  x--y--z  */

  float s,sxy,szy,xy,zy,d;
  int i;

  for (s=0,sxy=0,szy=0,i=0;i<3;i++) {
    xy=x[i]-y[i];
    zy=z[i]-y[i];
    s+=xy*zy;
    sxy+=xy*xy;
    szy+=zy*zy;
  }
  d=sxy*szy;
  if (d>0) {
    s=s/sqrt(d);
    s=acos((s>1.0 ? 1.0 : (s<-1.0 ? -1.0 : s)));
  }else{
    s= 2*3.141592653;
  }
  return(s);
}

float cangled(float *p,float *q,float *r)
{
/* cangle in degrees */

  return(180.0*(cangle(p,q,r)/3.141592653));
}

float ctors(float *p,float *q,float *r,float *s)
{
/* dihedral angle between planes (p,q,r) and (q,r,s) */

  vec_3 pq,qr,rs;
  vec_3 nq,nr;
  float lnq,lnr,t;

  pq=v3mv3(vec3(q),vec3(p));
  qr=v3mv3(vec3(r),vec3(q));
  rs=v3mv3(vec3(s),vec3(r));

  nq=v3xv3(pq,qr); lnq=v3l(nq);
  nr=v3xv3(qr,rs); lnr=v3l(nr);

  if (lnq<0.000000001 || lnr<0.000000001) {
    t=2*3.141592653;
  }else{
    t=v3sv3(nq,nr)/(lnq*lnr);
    t=acos(t < -1.0 ? -1.0 : (t > 1.0 ? 1.0 : t));
    if (v3sv3(qr,v3xv3(nq,nr))<0.0) t= -t;
  }

  return(t);
}

float ctorsd(float *p,float *q,float *r,float *s)
{
/* ctors in degrees */

  return(180.0*(ctors(p,q,r,s)/3.141592653));
}


vec_4 vec4(float *x)
{
  vec_4 v;
  int i;

  for (i=0;i<4;i++) v.c[i]=x[i];
  return(v);
}

mat_4 mat4(float *x)
{
  mat_4 m;
  int i,j;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) m.r[i].c[j] = *x++;
  }
  return(m);
}

void v42arr(float *x,vec_4 v)
{
  int i;

  for (i=0;i<4;i++) x[i]=v.c[i];
}

void m42arr(float *x,mat_4 m)
{
  int i,j;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) *x++ = m.r[i].c[j];
  }
}

mat_4 ident_4(void)
{
  mat_4 m;
  int i,j;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) m.r[i].c[j]=0.0;
    m.r[i].c[i]=1.0;
  }
  return(m);
}

mat_4 m4t(mat_4 m)
{
  mat_4 r;
  int i,j;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) {
      r.r[i].c[j]=m.r[j].c[i];
    }
  }
  return(r);
}

mat_3 rot3of(mat_4 m)
{
  mat_3 r;
  int i,j;

  for (i=0;i<3;i++) for (j=0;j<3;j++) r.r[i].c[j]=m.r[i].c[j];
  return(r);
}

float m4scale(mat_4 m)
{
  float s;
  int i;

  for (s=0,i=0;i<3;i++) s+=m.r[i].c[0]*m.r[i].c[0];
  return(sqrt(s));
}

void m4factor(float r[3][3],float t[3],float *s,mat_4 m)
{
  int i,j;

  *s=m4scale(m);

  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) r[i][j]=m.r[i].c[j]/(*s);
    t[i]=m.r[i].c[3]/(*s);
  }
}

void eqrxyz(float *rx,float *ry,float *rz,mat_4 m)
{
  /* equivalent x,y,z rotation angels for given rotation matrix m:
     m = Rz*Ry*Rx, angles are in radian ...
  */
  
  float t,s;
  int i,j;
  
  /* remove scale */

  s=m4scale(m); if (s==0) s=1.0;
  for (i=0;i<3;i++) for (j=0;j<3;j++) m.r[j].c[j]/=s;

  /* extract rot angles */

  t = (m.r[2].c[0] > -1.0 ? (m.r[2].c[0] < 1.0 ? m.r[2].c[0] : 1.0) : -1.0);
  *ry=asin(t);

  if (m.r[2].c[0] > 0.99999 || m.r[2].c[0] < -0.99999) { 
    *rx=0.0;
    *rz=atan2(m.r[0].c[1],m.r[1].c[1]);
  }else{
    *rx=atan2(-m.r[2].c[1],m.r[2].c[2]);
    *rz=atan2(-m.r[1].c[0],m.r[0].c[0]);
  }
}

mat_3 m3t(mat_3 m)
{
  mat_3 r;
  int i,j;

  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      r.r[i].c[j]=m.r[j].c[i];
    }
  }
  return(r);
}

mat_4 zero_4(void)
{
  int i,j;
  mat_4 r;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) {
      r.r[i].c[j]=0.0;
    }
  }
  return(r);
}

mat_4 inv_4(mat_4 a)
{
  float t[16];
  mat_4 r;
  int status;
  char *msg[4] = {"unknown error.","singular matrix.","out of memory.","internal problem"};

  m42arr(t,a); 
  status=gaussj(4,t,(float *)NULL);
  if (status) {
    r=zero_4();
    err_set(1,"gaussj: %s",msg[status<4&&status>0 ? status : 0]);
  }else{
    r=mat4(t);
  }
  return(r);
}

void prt_mat4(mat_4 a)
{
  int i,j;
  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) fprintf(stderr,"%10.5f ",a.r[i].c[j]);
    fprintf(stderr,"\n");
  }
}

vec_3 m4xv3(mat_4 a,vec_3 b)
{
/* useful for homog. coord without perspective scale ... */

  int i,j;
  vec_3 r;
  
  for (i=0;i<3;i++) {
    for (r.c[i]=a.r[i].c[3],j=0;j<3;j++) r.c[i]+=a.r[i].c[j]*b.c[j];
  }
  return(r);
}

vec_3 m3xv3(mat_3 a,vec_3 b)
{
  int i,j;
  vec_3 r;

  for (i=0;i<3;i++) {
    for (r.c[i]=0.0,j=0;j<3;j++) r.c[i]+=a.r[i].c[j]*b.c[j];
  }
  return(r);
}

mat_3 m3xm3(mat_3 a,mat_3 b)
{
  int i,j,k;
  mat_3 c;

  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      for (c.r[i].c[j]=0.0,k=0;k<3;k++) c.r[i].c[j]+=a.r[i].c[k]*b.r[k].c[j];
    }
  }
  return(c);
}

mat_4 m4xm4(mat_4 a,mat_4 b)
{
  int i,j,k;
  mat_4 c;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) {
      for (c.r[i].c[j]=0.0,k=0;k<4;k++) c.r[i].c[j]+=a.r[i].c[k]*b.r[k].c[j];
    }
  }
  return(c);
}

mat_4 m4mm4(mat_4 a,mat_4 b)
{
  int i,j;
  mat_4 c;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) {
      c.r[i].c[j]=a.r[i].c[j]-b.r[i].c[j];
    }
  }
  return(c);
}

mat_4 rotm4(char axis,float angle)
{
  /* rotation about coordinate axis (x,y,z) by 'angle' (rad) */

  mat_4 m;
  float S,C;

  m=ident_4();
  S=sin(angle);
  C=cos(angle);

  if (axis < 'x') axis+=32;
  switch (axis) {
  case 'x':
    m.r[1].c[1] =  C;  m.r[2].c[1] = -S;
    m.r[1].c[2] =  S;  m.r[2].c[2] =  C;
    break;
  case 'y':
    m.r[0].c[0] =  C;  m.r[2].c[0] =  S;
    m.r[0].c[2] = -S;  m.r[2].c[2] =  C;
    break;
  case 'z':
    m.r[0].c[0] =  C;  m.r[1].c[0] = -S;
    m.r[0].c[1] =  S;  m.r[1].c[1] =  C;
    break;
  }

  return(m);
}

mat_4 mmxinv(mat_4 m)
{
  /* return inverse of modelling transform m; m must have
     det(rot) = 1 and no perspective ... */

  mat_4 r;
  int i,j;

  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) r.r[i].c[j]=m.r[j].c[i];
  }
  for (i=0;i<3;i++) {
    r.r[3].c[i]=0.0;
    r.r[i].c[3]=0.0;
    for (j=0;j<3;j++) r.r[i].c[3]-=r.r[i].c[j]*m.r[j].c[3];
  }
  r.r[3].c[3]=1.0;

  return(r);
}

int is_mmx(mat_4 m)
{
  /* return TRUE iff m is modelling transform */

  return(mmx_err(m)>1.e-3);
}

float mmx_err(mat_4 m)
{
  int i,j;
  float det,s,err;

  err=0;
  for (i=0;i<3;i++) err=(err < fabs(m.r[3].c[i]) ? fabs(m.r[3].c[i]) : err);
  err = (err < fabs(m.r[3].c[3]-1.0) ? fabs(m.r[3].c[3]-1.0) : err);
  for (i=0;i<3;i++) {
    for (s=0,j=0;j<3;j++) s+=m.r[j].c[i]*m.r[j].c[i];
    err = (err < fabs(sqrt(s)-1.0) ? fabs(sqrt(s)-1.0) : err);
  }
  det=m.r[0].c[0]*m.r[1].c[1]*m.r[2].c[2]
     +m.r[0].c[1]*m.r[1].c[2]*m.r[2].c[0]
     +m.r[1].c[0]*m.r[2].c[1]*m.r[0].c[2]
     -m.r[2].c[0]*m.r[1].c[1]*m.r[0].c[2]
     -m.r[1].c[0]*m.r[0].c[1]*m.r[2].c[2]
     -m.r[0].c[0]*m.r[2].c[1]*m.r[1].c[2];
  err = ( err < fabs(det-1.0) ? fabs(det-1.0) : err);
  return(err);
}

void wline2xyz(float *line,float z,float *xyz)
{
  float *pt1,*pt2;

  /*  return 3D world position xyz for point on "world-line" line at
      given z */

  pt1=line;
  pt2=line+3;

  if (fabs(pt1[2]-pt2[2])>0.00000001) {
    xyz[0]=pt1[0]+(pt2[0]-pt1[0])*(pt1[2]-z)/(pt1[2]-pt2[2]);
    xyz[1]=pt1[1]+(pt2[1]-pt1[1])*(pt1[2]-z)/(pt1[2]-pt2[2]);
    xyz[2]=z;
  }else{
    xyz[0]=pt1[0];
    xyz[1]=pt1[1];
    xyz[2]=z;
  }
}

vec_3 wline2v3(float *line,float z)
{
  float pt[3];
  
  wline2xyz(line,z,pt);
  return(vec3(pt));
}
/*-------------------- cnstr --------------------*/

int cnstr(float *p,float *x,float *y,float *z,float d3,float alp,float tau)
{

/* Construct point from x,y,z (all coordinates are cartesian) and
   torsion angle tau (degrees) and bond angle alp (degrees).
   Order of points is  x --> y --> z --> p ,
   Parameters:                tau alp d3 .     

   return: 1 iff p is well determined, 0 otherwise    */

  float x1[3],x2[3],p1,p2,l1,a,b,c,d,e,f,r[3][3],l[3],frad;
  int i,j,status;

  status=1;
  frad=3.141592653/180.0;

  for (i=0;i<3;i++) x1[i]=z[i]-y[i];
  l1=sqrt(x1[0]*x1[0]+x1[1]*x1[1]+x1[2]*x1[2]);
  p1=sqrt(x1[0]*x1[0]+x1[1]*x1[1]);

  if (p1>1.0e-8) {
    a=x1[0]/p1;
    b=x1[1]/p1;
  }else{
    a=1.0;
    b=0.0;
    status=0;
  }

  if (l1>10.e-8) {
    e=x1[2]/l1;
    f=p1/l1;
  }else{
    e=1.0;
    f=0.0;
    status=0;
  }

  x2[0]=a*e*(x[0]-y[0])+b*e*(x[1]-y[1])-f*(x[2]-y[2]);
  x2[1]=-b*(x[0]-y[0])+a*(x[1]-y[1]);
  p2=sqrt(x2[0]*x2[0]+x2[1]*x2[1]);

  if (p2>1.0e-8) {
    c=x2[0]/p2;
    d=x2[1]/p2;
  }else{
    c=1.0;
    d=0.0;
    status=0;
  }

  r[0][0]=a*c*e-b*d;
  r[0][1]=-a*d*e-b*c;
  r[0][2]=a*f;
  r[1][0]=b*c*e+a*d;
  r[1][1]=-b*d*e+a*c;
  r[1][2]=b*f;
  r[2][0]=-c*f;
  r[2][1]=d*f;
  r[2][2]=e;
  
  l[0]=d3*cos(frad*tau)*cos(frad*(alp-90));
  l[1]=d3*sin(frad*tau)*cos(frad*(alp-90));
  l[2]=d3*sin(frad*(alp-90));
  
  for (i=0;i<3;i++) {
    p[i]=z[i];
    for (j=0;j<3;j++) {
      p[i]=p[i]+r[i][j]*l[j];
    }
  }

  return(status);
}
/*-------------------- cnstrv --------------------*/

vec_3 cnstrv(vec_3 x,vec_3 y,vec_3 z,float d3,float alp,float tau)
{
  /* construct point (c.f. cnstr), return point as vector */

  float a[3],b[3],c[3],p[3];

  v32arr(a,x);
  v32arr(b,y);
  v32arr(c,z);

  cnstr(p,a,b,c,d3,alp,tau);

  return(vec3(p));
}
/*-------------------- pedestrian rotax --------------------*/

mat_4 rotax(float *a,float *b,float tau)
{
  /* 
   Build 4x4 matrix of clockwise rotation about axis a-->b
   by angle tau (deg). Result is a homogenous transformation
   matrix.

  */

  mat_4 r;
  float e[3],s,ct,st;
  int i,j;

  for (s=0,i=0;i<3;i++) {
    e[i]=b[i]-a[i];
    s+=e[i]*e[i];
  }
  if (s>0) {
    for (i=0;i<3;i++) e[i]/=sqrt(s);
    ct=cos(tau*3.141592653/180.0);
    st=sin(tau*3.141592653/180.0);
    r.r[0].c[0]=e[0]*e[0]+(1.0-e[0]*e[0])*ct;
    r.r[1].c[1]=e[1]*e[1]+(1.0-e[1]*e[1])*ct;
    r.r[2].c[2]=e[2]*e[2]+(1.0-e[2]*e[2])*ct;
    r.r[0].c[1]=e[0]*e[1]*(1-ct)-e[2]*st;
    r.r[1].c[2]=e[1]*e[2]*(1-ct)-e[0]*st;
    r.r[2].c[0]=e[2]*e[0]*(1-ct)-e[1]*st;
    r.r[1].c[0]=e[0]*e[1]*(1-ct)+e[2]*st;
    r.r[2].c[1]=e[1]*e[2]*(1-ct)+e[0]*st;
    r.r[0].c[2]=e[2]*e[0]*(1-ct)+e[1]*st;
    for (i=0;i<3;i++) {
      r.r[i].c[3]=a[i];
      for (j=0;j<3;j++) r.r[i].c[3]-=r.r[i].c[j]*a[j];
      r.r[3].c[i]=0.0;
    }
    r.r[3].c[3]=1.0;
  }else{
    r=ident_4();
  }
  return(r);
}
/*-------------------- QtoRot4 --------------------*/

mat_4 QtoRot4(float *q)
{
  /* convert quaternion q to rotation matrix */

  mat_4 r;

  r=zero_4();
  
  r.r[0].c[0]=q[0]*q[0]+q[1]*q[1]-q[2]*q[2]-q[3]*q[3];
  r.r[0].c[1]=2*(q[1]*q[2]+q[0]*q[3]);
  r.r[0].c[2]=2*(q[1]*q[3]-q[0]*q[2]);
  r.r[1].c[0]=2*(q[1]*q[2]-q[0]*q[3]);
  r.r[1].c[1]=q[0]*q[0]-q[1]*q[1]+q[2]*q[2]-q[3]*q[3];
  r.r[1].c[2]=2*(q[2]*q[3]+q[0]*q[1]);
  r.r[2].c[0]=2*(q[1]*q[3]+q[0]*q[2]);
  r.r[2].c[1]=2*(q[2]*q[3]-q[0]*q[1]);
  r.r[2].c[2]=q[0]*q[0]+q[3]*q[3]-q[1]*q[1]-q[2]*q[2];
  
  r.r[3].c[3]=1.0;

  return(r);
}
/*-------------------- EulerRot3 --------------------*/

mat_3 EulerRot3(float *tau)
{
/* Compute 3x3 rotation matrix r from three Euler angles
   tau(1)...tau(3). Angles are in radians.

   Note: tau(1) : rotation around z, x --> y,
         tau(2) : rotation around x', z' --> y' ('negative'),
         tau(3) : rotation around z'', x''--> y''.
	 */

  int i;
  float c[3],s[3];
  mat_3 R;

  for (i=0;i<3;i++) {
    c[i]=cos(tau[i]);
    s[i]=sin(tau[i]);
  }
  R.r[0].c[0]=c[0]*c[2]-s[0]*c[1]*s[2];
  R.r[0].c[1]=-c[0]*s[2]-s[0]*c[1]*c[2];
  R.r[0].c[2]=-s[0]*s[1];
  R.r[1].c[0]=s[0]*c[2]+c[0]*c[1]*s[2];
  R.r[1].c[1]=-s[0]*s[2]+c[0]*c[1]*c[2];
  R.r[1].c[2]=c[0]*s[1];
  R.r[2].c[0]=-s[1]*s[2];
  R.r[2].c[1]=-s[1]*c[2];
  R.r[2].c[2]=c[1];
  return(R);
}
/*-------------------- Rot3Euler --------------------*/

vec_3 Rot3Euler(mat_3 R)
{
  int i;
  vec_3 q;
  float pt[4][3]={{1,0,0},{0,0,0},{0,0,0},{0,0,0}};

  /* 3x3 rotation matrix to Euler angles... */

  if (fabs(R.r[0].c[2])<0.00001&&fabs(R.r[1].c[2])<0.00001) q.c[0]=0.0;
  else q.c[0]=atan2(-R.r[0].c[2],R.r[1].c[2]);
  q.c[1]=acos(R.r[2].c[2]);
  pt[0][0]=cos(q.c[0]);      /* direction of x after first rotation */
  pt[0][1]=sin(q.c[0]);
  pt[0][2]=0.0;
  for (i=0;i<3;i++) {
    pt[3][i]=R.r[i].c[0];
    pt[2][i]=R.r[i].c[2];
  }
  q.c[2]=ctors(pt[0],pt[1],pt[2],pt[3]);
  return(q);
}
/*-------------------- EulerRot3Dq --------------------*/

mat_3 EulerRot3Dq(float *tau,int k)
{
/* Derivative of Euler Rotation matrix. w.r.t. tau[k] */

  mat_3 R;
  float c[3],s[3];
  int i;

  /* compute sines and cosines */
  
  for (i=0;i<3;i++) {
    c[i]=cos(tau[i]);
    s[i]=sin(tau[i]);
  }

  /* compute derivatives... */

  switch (k) {
  case 0:
    R.r[0].c[0]=-s[0]*c[2]-c[0]*c[1]*s[2];
    R.r[0].c[1]= s[0]*s[2]-c[0]*c[1]*c[2];
    R.r[0].c[2]=-c[0]*s[1];
    R.r[1].c[0]=c[0]*c[2]-s[0]*c[1]*s[2];
    R.r[1].c[1]=-c[0]*s[2]-s[0]*c[1]*c[2];
    R.r[1].c[2]=-s[0]*s[1];
    R.r[2].c[0]=0;
    R.r[2].c[1]=0;
    R.r[2].c[2]=0;
    break;

  case 1:
    R.r[0].c[0]=s[0]*s[1]*s[2];
    R.r[0].c[1]=s[0]*s[1]*c[2];
    R.r[0].c[2]=-s[0]*c[1];
    R.r[1].c[0]=-c[0]*s[1]*s[2];
    R.r[1].c[1]=-c[0]*c[1]*c[2];
    R.r[1].c[2]=c[0]*c[1];
    R.r[2].c[0]=-c[1]*s[2];
    R.r[2].c[1]=-c[1]*c[2];
    R.r[2].c[2]=-s[1];
    break;

  case 2:
    R.r[0].c[0]=-c[0]*s[2]-s[0]*c[1]*c[2];
    R.r[0].c[1]=-c[0]*c[2]+s[0]*c[1]*s[2];
    R.r[0].c[2]=0;
    R.r[1].c[0]=-s[0]*s[2]+c[0]*c[1]*c[2];
    R.r[1].c[1]=-s[0]*c[2]-c[0]*c[1]*s[2];
    R.r[1].c[2]=0;
    R.r[2].c[0]=-s[1]*c[2];
    R.r[2].c[1]= s[1]*s[2];
    R.r[2].c[2]=0;
    break;

  default:
    R=ident_3();
    break;
  }
  return(R);
}
/*-------------------- EulerRot4 --------------------*/

mat_4 EulerRot4(float *tau)
{
/* Compute 4x4 rotation matrix r from three Euler angles
   tau(0)...tau(3). Angles are in radians. 
   tau(3)...tau(5) is translation vector. */

  int i;
  float c[3],s[3];
  mat_4 R;

  for (i=0;i<3;i++) {
    c[i]=cos(tau[i]);
    s[i]=sin(tau[i]);
  }
  R.r[0].c[3]=tau[3]; R.r[1].c[3]=tau[4]; R.r[2].c[3]=tau[5];
  R.r[3].c[0]=0; R.r[3].c[1]=0; R.r[3].c[2]=0; 
  R.r[3].c[3]=1; 
  R.r[0].c[0]=c[0]*c[2]-s[0]*c[1]*s[2];
  R.r[0].c[1]=-c[0]*s[2]-s[0]*c[1]*c[2];
  R.r[0].c[2]=-s[0]*s[1];
  R.r[1].c[0]=s[0]*c[2]+c[0]*c[1]*s[2];
  R.r[1].c[1]=-s[0]*s[2]+c[0]*c[1]*c[2];
  R.r[1].c[2]=c[0]*s[1];
  R.r[2].c[0]=-s[1]*s[2];
  R.r[2].c[1]=-s[1]*c[2];
  R.r[2].c[2]=c[1];
  return(R);
}
