/*
  Generate ligand topology and parameter files for CNS/CNS/X-Plor
  Generate ligand dictionary files for Refmac
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "mystdlib.h"
#include "swdef.h"
#include "gcalc.h"
#include "errhand.h"
#include "object.h"
#include "parser.h"
#include "genlig.h"
#include "strutil.h"
#include "transpor.h"
#include "errhand.h"

#define TRUE  1
#define FALSE 0

#define LOOP 1
#define DONE 2

#define FLG1      (1<<0)

#define HYD_NONE  0
#define HYD_POLAR 1
#define HYD_YES   2

#define GENLIG_PAR_CURRENT  1

static glLigand_t *gatherLigand(Node *root,char *code,int use_charges,int hyd_mode,int flag);
static void fixCNXAtomnames(glLigand_t *L);
static void fixDyanaAtomnames(glLigand_t *L);
static void fixRefmacAtomnames(glLigand_t *L);
static /* NON-REENTRANT */ char *rotName(char *name,char *chs);
static void ligandGeometry(glLigand_t *L);
static int isPlanarAtom(Atom *a);
static int ligandTopology(glLigand_t *L);
static int ligandParameters(glLigand_t *L,int mode);
static int ligandParametersCurrent(glLigand_t *L);
static glLigand_t *freeLigand(glLigand_t *L);
static glLigand_t *newLigand(void);
static int writeCNXtopology (char *filename,glLigand_t *L);
static int writeCNXpdb(char *filename,glLigand_t *L);
static int writeREFMACpdb(char *filename,glLigand_t *L);
static int writeCNXparameters(char *filename,glLigand_t *L);
static int writeREFMACdictionary(char *filename,glLigand_t *L);
static int writeDYANAdictionary(char *filename,glLigand_t *L);
static int writeDYANArcMacro(char *filename,glLigand_t *L);
static glLigandTree_t *ligandTree(glLigand_t *L,int root);
static glLigandTree_t *newLigandTree(int size);
static glLigandTree_t *freeLigandTree(glLigandTree_t *T);
static int ligandRings(glLigand_t *M);
static void ligandPlanes(glLigand_t *L);
static void addPlane(glLigand_t *L,int *a);
static void free_plane(glPlane_t p);
static glPlane_t new_plane(int c);
static void findAtomTypes(glLigand_t *L,Node *Library,char *rules);
static Node *genligRoot(int *status);
static int dyanaTorSort(const void *a,const void *b);
static char **singleQuotedNames(glLigand_t *L);
static char **freeQuotedNames(int n,char **names);

      
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(x) (streq(key->key,x))

#define NCNXVDW 12
static glCNXvdwpar_t CNXvdw[NCNXVDW] = {              /* should read this from file... */
  {"B",   0.0100,   2.0847,      0.0100,   2.0847},
  {"Br",  0.3200,   3.8843,      0.3200,   3.8843},
  {"C",   0.1200,   3.7418,      0.1000,   3.3854},
  {"Cl",  0.2600,   3.6705,      0.2600,   3.6705},
  {"F",   0.1000,   3.0291,      0.1500,   2.7618},
  {"H",   0.0498,   1.4254,      0.0498,   1.4254},
  {"I",   0.8000,   3.6349,      0.8000,   3.6349},
  {"N",   0.2384,   2.8509,      0.2384,   2.8509},
  {"O",   0.1591,   2.8509,      0.1591,   2.8509},
  {"P",   0.1000,   3.3854,      0.1000,   3.3854},
  {"S",   0.0430,   3.3676,      0.0430,   3.3676},
  {"Se",  0.0430,   3.5100,      0.0430,   3.5100}
};

static float genlig_conv=0.592;            /* magic conversion sigma -> k */
static float genlig_convr=1943.42;         /* same in radian, instead of deg */
static float genlig_BS_sigma=0.02;         /* default sigma for bond lengths */
static float genlig_AB_sigma=3.0;          /* default sigma for bond angles */
static float genlig_PL_sigma=1.5;          /* default sigma for planarity impropers */
static float genlig_INV_sigma=10.0;        /* default sigma for chirality impropers */
static float genlig_DBL_sigma=1.5;         /* default sigma double bond dihedrals */
static float genlig_AMIDE_sigma=3.0;       /* default sigma double amide dihedrals */
static float genlig_AROMATIC_sigma=1.5;    /* default sigma double aromatic dihedrals */
static float genlig_RINGTORS_sigma=7.5;    /* default sigma for 'soft' dihedrals in rings */
static float genlig_ZERO_sigma=360.0;      /* default sigma for unconstrained dihedrals */
static float genlig_PLANE_sigma=0.02;      /* default sigma for REFMAC plane constraints */

static int   fixNames=TRUE;                /* do we fix atom names if necessary? */
static int   rigidRings=FALSE;             /* do we keep rings rigid in Dyana? */
static float noDyanaTorsSigma=5.0;         /* do not add dihedral to Dyana if sigma < noDyanaTorsSigma */

static Node *Library = (Node *)NULL;

extern int graphix;

static char *rcmac1=
  "## 7MACROS: CYANA macro\n"
  "###\n"
  "### Parameters: weight=<float> (default: 1.0)\n"
  "###             dtol=<float>   (default: 0.0)\n"
  "###             atol=<float>   (default: 0.0)\n"
  "###             ttol=<float>   (default: 0.0)\n"   
  "###\n"
  "### Adds constraints (upper and lower limits on distances) for ring-closure\n"
  "### bonds in all residues of type %s. The optional parameter 'weight' may\n";

static char *rcmac2=
  "### be used to define the weight of the ring-closure restraints in the target\n"
  "### function. 'dtol', 'atol', and 'ttol' are tolerances (in Angstrom) for the\n"
  "###  distance restraints generated for ring-closure bondlengths (dtol),"
  "###  angles (atol), and torsions.\n"
  "\n"
  "var info echo i res\n"
  "\n"
  "syntax weight=@r=1.0 dtol=@r=0.0 atol=@r=0.0 ttol=@r=0.0\n"
  "\n"
  "echo:=off\n"
  "info:=none\n"
  "\n";

static char **singleQuotedNames(glLigand_t *L)
{
  char **names=NULL;
  int i;

  /* put single quotes around atom names that contain single or double quotes or blanks */

  if (L&&L->nat>0) {
    names=malloc(L->nat*sizeof(char *));
    if (names) {
      for (i=0;i<L->nat;i++) {
	if (strchr(L->A[i].name,'\'')||strchr(L->A[i].name,'"')||strchr(L->A[i].name,' ')) {
	  names[i]=malloc(3+strlen(L->A[i].name));
	  if (names[i]) sprintf(names[i],"'%s'",L->A[i].name);
	}else{
	  names[i]=strdup(L->A[i].name);
	}
      }
    }
  }
  return(names);
}

char **freeQuotedNames(int n,char **names)
{
  int i;

  if (names) {
    for (i=0;i<n;i++) if (names[i]) free(names[i]);
    free(names);
  }
  return NULL;
}

/*========================================================================*/

void cmd_genlig(void)
{
  Token *key=(Token *)NULL,*filename=(Token *)NULL;
  int state,status;
  char *keys="charges code hydrogens library molecule names write ' ' h!rings end";
  int use_charges=0;
  int hyd_mode=HYD_NONE;
  char *fmt=(char *)NULL;
  char *code=(char *)NULL;
  char *s;
  glLigand_t *L=(glLigand_t *)NULL;
  UCpipe *ucp;
  Node *root=(Node *)NULL,*newroot;
  Atom *a;

  rigidRings=FALSE;
  state=LOOP;
  while (state!=DONE) {
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("genlig> ",";",keys);
    if (key->status==TOK_OK) {
      if (key_is("molecule")) {
	newroot=genligRoot(&status);
	if (newroot||status==TOK_NOK) {
	  root=newroot;
	  if (L) L=freeLigand(L);
	}
      }else if (key_is("charges")) {
	key=free_Token(key);
	default_token(use_charges?"yes":"no");
	key=Get_Kwd("use charges? ",";","yes no");
	if (key->status==TOK_OK) {
	  use_charges=key_is("yes");
	}else if (key->status==TOK_NOK) {
	  err_set(1,"invalid token: %s\n",key->text);
	}
      }else if (key_is("hydrogens")) {
	key=free_Token(key);
	default_token(hyd_mode==HYD_YES?"all":hyd_mode==HYD_POLAR?"polar":"none");
	key=Get_Kwd("which hydrogens? ",";","none polar all");
	if (key->status==TOK_OK) {
	  hyd_mode=key->n;
	  if (L&&L->hyd_mode!=hyd_mode) {  /* maybe a bit harsh... */
	    L=freeLigand(L);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"invalid token: %s\n",key->text);
	}
      }else if (key_is("library")) {
	key=free_Token(key);
	key=Get_File((char *)NULL,".wlb","read library file: ",";");
	if (key->status==TOK_OK) {
	  ucp=(UCpipe *)key->ptr;
	  if (is_lib(Library)) Library=del_node(Library);
	  Library=newLib("genligLibrary");
	  if (Library) {
	    if (read_lib_ucp(ucp,(Lib *)Library)) {
	      err_set(1,"error reading library file %s",ucp->filename);
	      Library=del_node(Library);
	    }
	  }else err_set(1,"failed to allocate memory for library.\n");
	  UCpclose(ucp);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"failed to read library file %s",key->text);
	}
      }else if (key_is("code")) {
	default_token(code?code:"LI1 LI2 LI3 LI4 LI5");
	key=free_Token(key);
	key=Get_String("residue code: ",";");
	if (key->status==TOK_OK) {
	  if (code) free(code);
	  code=key->text;
	  if (L) {
	    if ((s=strchr(code,'/'))) {
	      L->segid[0]=*code;
	      strncpy(L->code,s+1,4); 
	    }else{
	      strncpy(L->code,code,4); 
	      strcpy(L->segid," ");
	    }
	    L->code[5]='\0';
	    L->segid[1]='\0';
	  }
	  key->text=(char *)NULL;
	}
      }else if (key_is("names")) {
	key=free_Token(key);
	default_token(fixNames?"yes":"no");
	key=Get_Kwd("fix atom names? ",";","yes no");
	if (key->status==TOK_OK) {
	  fixNames=key_is("yes");
	}else if (key->status==TOK_NOK) {
	  err_set(1,"invalid response: %s",key->text);
	}
      }else if (key_is("write")) {
	key=free_Token(key);
	key=Get_Kwd("format: ",";","CNX h!Dyana Refmac");
	if (key->status==TOK_OK) {
	  if (fmt) free(fmt);
	  fmt=key->key;
	  key->key=(char *)NULL;
	  filename=free_Token(filename);
	  filename=Get_String("filename: ",";");
	  if (filename->status==TOK_OK) {
	    if (!root) root=genligRoot(&status);
	    if (L) L=freeLigand(L);
	    if (!L&&root) {
	      L=gatherLigand(root,code,use_charges,hyd_mode,FLG1);
	      ForAtoms(a,root) a->flg&=~FLG1;
	    }
	    if (L) {
	      if (!fmt||streq(fmt,"CNX")) {
		genligCNX(filename->text,L);
	      }else if (streq(fmt,"Refmac")) {
		genligREFMAC(filename->text,L);
	      }else if (streq(fmt,"Dyana")) {
		genligDYANA(filename->text,L);
	      }else err_set(1,"unknown dictionary format: %s",fmt);
	    }else if (root) {
	      err_set(1,"failed to build internal ligand structure");
	    }else err_set(1,"no files written.");
	  }else if (filename->status==TOK_NOK) {
	    err_set(1,"invalid token: %s\n",filename->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"invalid dictionary format: %s\n",key->text);
	}else state=DONE;
      }else if (key_is("rings")) {
	key=free_Token(key);
	default_token(rigidRings?"rigid":"flexible");
	key=Get_Kwd("treatment of rings in DYANA: ",";","flexible rigid");
	if (key->status==TOK_OK) {
	  rigidRings=key_is("rigid");
	}else if (key->status==TOK_NOK) err_set(1,"invalid DYANA rings mode: %s\n",key->text);
      }else if (key_is("end")) {
	state=DONE;
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"invalid XRay/genlig key: %s\n",key->text);
      end_parse();
    }else state=DONE;
  }
  filename=free_Token(filename);
  key=free_Token(key);
  if (fmt) free(fmt);
  if (code) free(code);
  if (L) freeLigand(L);
}

/*-------------------- genligCNX --------------------*/

int genligCNX(char *filename,glLigand_t *L)
{
  int status=0,n;

  if (L) {
    if ((n=ligandTopology(L)==1)) {
      findAtomTypes(L,Library,"cnx");
      if (fixNames) fixCNXAtomnames(L);
      ligandGeometry(L);
      ligandParameters(L,GENLIG_PAR_CURRENT);
      writeCNXtopology(filename,L);
      writeCNXparameters(filename,L);
      writeCNXpdb(filename,L);
    }else if (n>1) {
      err_set(1,"genligCNX failed: ligand is not connected (%d pieces)",n);
      status=2;
    }else if (n<1) {
      err_set(1,"genligCNX failed: epmty ligand atom selection");
      status=3;
    }
  }else{
    err_set(1,"genligCNX: failed to generate internal datastructure for ligand");
    status=1;
  }
  return status;
}

/*-------------------- genligDYANA --------------------*/

int genligDYANA(char *filename,glLigand_t *L)
{
  int status=0,n;
  int tree=1;          /* always generate tree structure */

  findAtomTypes(L,Library,"dyana");
  if (L) {
    if ((n=ligandTopology(L)==1)) {
      if (tree) {
	L->T=freeLigandTree(L->T);
	L->T=ligandTree(L,0);
	if (!L->T) {
	  err_set(1,"genligDYANA: failed to generate tree-like structure for ligand",n);
	  status=1;
	}
      }
      if (!status) {
	if (fixNames) fixDyanaAtomnames(L);
	ligandGeometry(L);
	ligandParameters(L,GENLIG_PAR_CURRENT);
	ligandPlanes(L);
	status=writeDYANAdictionary(filename,L);
	if (!status) status=writeDYANArcMacro(filename,L);
      }
    }else if (n>1) {
      err_set(1,"genligDYANA failed: ligand is not connected (%d pieces)",n);
      status=2;
    }else if (n<1) {
      err_set(1,"genligDYANA failed: epmty ligand atom selection");
      status=3;
    }
  }else{
    err_set(1,"genligDYANA: failed to generate internal datastructure for ligand");
    status=1;
  }
  return status;
}

/*-------------------- genligREFMAC --------------------*/

int genligREFMAC(char *filename,glLigand_t *L)
{
  int status=0,n;
  int tree=1;          /* always generate tree structure */

  findAtomTypes(L,Library,"refmac");
  if (L) {
    if ((n=ligandTopology(L)==1)) {
      if (tree) {
	L->T=freeLigandTree(L->T);
	L->T=ligandTree(L,0);
	if (!L->T) {
	  err_set(1,"genligREFMAC: failed to generate tree-like structure for ligand",n);
	  status=1;
	}
      }
      if (!status) {
	if (fixNames) fixRefmacAtomnames(L);
	ligandGeometry(L);
	ligandParameters(L,GENLIG_PAR_CURRENT);
	ligandPlanes(L);
	status=writeREFMACdictionary(filename,L);
	writeREFMACpdb(filename,L);
      }
    }else if (n>1) {
      err_set(1,"genligREFMAC failed: ligand is not connected (%d pieces)",n);
      status=2;
    }else if (n<1) {
      err_set(1,"genligREFMAC failed: epmty ligand atom selection");
      status=3;
    }
  }else{
    err_set(1,"genligREFMAC: failed to generate internal datastructure for ligand");
    status=1;
  }
  return status;
}

/*-------------------- gatherLigand --------------------*/

static glLigand_t *gatherLigand(Node *root,char *code,int use_charges,int hyd_mode,int flag)
{
  glLigand_t *L;
  glAtom_t *A;
  glBond_t *B;
  Atom *a,*b,*c,*d;
  int i,j,k,n;
  char *s;

  /* collect ligand atoms (in root) into internal ligand structure */

  L=newLigand();
  if (!L) return(L);

  if (code) {
    if ((s=strchr(code,'/'))) {
      L->segid[0]=*code;
      strncpy(L->code,s+1,4); 
    }else{
      strncpy(L->code,code,4); 
      strcpy(L->segid," ");
    }
  }else{
    strcpy(L->code,"LI1");
    strcpy(L->segid," ");
  }
  L->code[5]='\0';
  L->segid[1]='\0';
  L->nonh=0;
  L->use_charges=use_charges;
  L->hyd_mode=hyd_mode;
  L->T=(glLigandTree_t *)NULL;

  /* count, mark and collect atoms */
  ForAtoms(a,root) {
    if (a->atp->z==1) {
      if (hyd_mode==HYD_POLAR) {
	if (a->nbnd>0) {
	  b=neighbour(0,a);
	  if (b->atp->z!=6&&b->atp->z!=1) a->flg|=flag;
	} 
      }else if (hyd_mode==HYD_YES) a->flg|=flag;
    }else a->flg|=flag;
    if (a->flg&flag) {
      a->seqnr=L->nat;
      L->nat++;
    }
  }
  if (L->nat>0) {
    L->A=(glAtom_t *)malloc(L->nat*sizeof(glAtom_t));
    if (!L->A) {
      L=freeLigand(L);
      return(L);
    }else {
      n=0;
      ForAtoms(a,root) {
	if (a->flg&flag) {
	  strncpy(L->A[n].name,a->name,7); L->A[n].name[7]='\0';
	  L->A[n].mass=a->atp->mass;
	  L->A[n].charge=a->q;
	  L->A[n].a=a;
	  L->A[n].nn=0;
	  L->A[n].z=a->atp->z;
	  L->A[n].atp=NULL;
	  if (a->atp->z>1) L->nonh++;
	  strncpy(L->A[n].type,a->atp->chs,2); L->A[n].type[2]='\0';
	  for (i=0;i<a->nbnd;i++) {
	    b=neighbour(i,a);
	    if (b->flg&flag) {
	      if (b->seqnr<a->seqnr) L->nbnd++;
	    }else if (b->atp->z==1) {
	      L->A[n].mass+=1.0079;
	      L->A[n].charge+=b->q;
	    }
	  }	    
	  n++;
	}
      }
    }
  }else return(L);

  /* collect bonds */
  if (L->nbnd>0) {
    L->B=(glBond_t *)malloc(L->nbnd*sizeof(glBond_t));
    if (!L->B) {
      L=freeLigand(L);
      return(L);
    }
    n=0;
    ForAtoms(a,root) {
      if (a->flg&flag) {
	for (i=0;i<a->nbnd;i++) {
	  b=neighbour(i,a);
	  if ((b->seqnr>a->seqnr)&&(b->flg&flag)) {
	    L->B[n].l=L->A+a->seqnr;
	    L->B[n].r=L->A+b->seqnr;
	    L->B[n].a[0]=a->seqnr;
	    L->B[n].a[1]=b->seqnr;
	    L->B[n].bt=a->bnd[i]->bt;
	    L->B[n].M=0;
	    L->B[n].d=0.0;
	    L->B[n].d0=0.0;
	    L->B[n].k=0.0;
	    L->B[n].sigma=0.0;
	    if (L->B[n].r->nn<MAXLIGATMNB) {
	      L->B[n].r->n[L->B[n].r->nn]=L->B[n].l;
	      L->B[n].r->b[L->B[n].r->nn]=L->B+n;
	      L->B[n].r->nn++;
	    }
	    if (L->B[n].l->nn<MAXLIGATMNB) {
	      L->B[n].l->n[L->B[n].l->nn]=L->B[n].r;
	      L->B[n].l->b[L->B[n].l->nn]=L->B+n;
	      L->B[n].l->nn++;
	    }
	    n++;
	  }
	}	    
	L->nang+=(L->A[a->seqnr].nn*(L->A[a->seqnr].nn-1))/2;
      }
    }
    /* sort bond list of each atom such that bonds involving H come last */
    for (i=0;i<L->nat;i++) {
      for (k=1;k<L->A[i].nn;k++) {
	if (L->A[i].n[k-1]->z==1 && L->A[i].n[k]->z>1) {
	  A=L->A[i].n[k-1];
	  L->A[i].n[k-1]=L->A[i].n[k];
	  L->A[i].n[k]=A;
	  B=L->A[i].b[k-1];
	  L->A[i].b[k-1]=L->A[i].b[k];
	  L->A[i].b[k]=B;
	  k-=2; if (k<0) k=0;
	}
      }
      for (k=0;k<L->A[i].nn;k++) L->A[i].nb[k]=L->A[i].n[k]->a->seqnr;
    }
  }

  /* collect angles */
  if (L->nang>0) {
    L->angle=(glAngle_t *)malloc(L->nang*sizeof(glAngle_t));
    if (!L->angle) {
      L=freeLigand(L);
      return(L);
    }
    n=0;
    for (k=0;k<L->nat;k++) {
      a=L->A[k].a;
      for (i=0;i<L->A[k].nn-1;i++) {
	b=L->A[k].n[i]->a;
	for (j=i+1;j<L->A[k].nn;j++) {
	  c=L->A[k].n[j]->a;
	  L->angle[n].w=0.0;
	  L->angle[n].w0=0.0;
	  L->angle[n].k=0.0;
	  L->angle[n].sigma=0.0;
	  L->angle[n].a[1]=a->seqnr;
	  if (b->seqnr<c->seqnr) {
	    L->angle[n].a[0]=b->seqnr;
	    L->angle[n].a[2]=c->seqnr;
	  }else{
	    L->angle[n].a[0]=c->seqnr;
	    L->angle[n].a[2]=b->seqnr;
	  }
	  n++;
	}
      }
    }
  }

  /* count dihedrals */
  for (i=0;i<L->nbnd;i++) {
    L->ndih+=(L->B[i].l->nn-1)*(L->B[i].r->nn-1);
  }
  
  /* collect dihedrals */
  if (L->ndih>0) {
    L->dihedral=(glDihedral_t *)malloc(L->ndih*sizeof(glDihedral_t));
    if (!L->ndih) {
      L=freeLigand(L);
      return(L);
    }
    n=0;
    for (i=0;i<L->nbnd;i++) {
      a=L->B[i].l->a;
      b=L->B[i].r->a;
      for (j=0;j<L->B[i].l->nn;j++) {
	c=L->B[i].l->n[j]->a;
	if (c->seqnr!=b->seqnr) {
	  for (k=0;k<L->B[i].r->nn;k++) {
	    d=L->B[i].r->n[k]->a;
	    if (d->seqnr!=a->seqnr&&c->seqnr!=d->seqnr) {
	      L->dihedral[n].type=GENL_DIH_UNK;
	      L->dihedral[n].a[0]=c->seqnr;
	      L->dihedral[n].a[1]=a->seqnr;
	      L->dihedral[n].a[2]=b->seqnr;
	      L->dihedral[n].a[3]=d->seqnr;
	      L->dihedral[n].w=0.0;
	      L->dihedral[n].k=0.0;
	      L->dihedral[n].sigma=0.0;
	      L->dihedral[n].n=1;
	      L->dihedral[n].p=0.0;
	      L->dihedral[n].b=i;
	      L->B[i].M++;
	      n++;
	    }
	  }
	}
      }
    }
    L->ndih=n;
  }

  /* count impropers */ 
  for (i=0;i<L->nat;i++) {
    if (L->A[i].nn>2) L->nimp++;
  }

  if (L->nimp>0) {
    L->improper=(glImproper_t *)malloc(L->nimp*sizeof(glImproper_t));
    if (!L) {
      L=freeLigand(L);
      return(L);
    }
    n=0;
    for (i=0;i<L->nat;i++) {
      if (L->A[i].nn==3) {
	L->improper[n].type=GENL_IMP_UNK;
	L->improper[n].a[0]=L->A[i].nb[0];
	L->improper[n].a[1]=i;
	L->improper[n].a[2]=L->A[i].nb[1];
	L->improper[n].a[3]=L->A[i].nb[2];
	L->improper[n].w=0.0;
	L->improper[n].k=0.0;
	L->improper[n].n=0;
	L->improper[n].p=0.0;
	n++;
      }else if (L->A[i].nn==4) {
	for (k=0,j=0;j<4;j++) if (L->A[L->A[i].nb[j]].z>1) k++;
	if ((k>2&&L->A[i].z==6)||(k==4&&L->A[i].z==7)) {
	  L->improper[n].type=GENL_IMP_UNK;
	  L->improper[n].w=0.0;
	  L->improper[n].k=0.0;
	  L->improper[n].n=0;
	  L->improper[n].p=0.0;
	  for (k=0,j=0;j<4;j++) {
	    if (L->A[L->A[i].nb[j]].z>1) L->improper[n].a[k++]=L->A[i].nb[j];
	  }
	  L->improper[n].a[3]=L->improper[n].a[1];
	  L->improper[n].a[1]=i;
	  n++;
	}
      }
    }
    L->nimp=n;
  }
  return(L);
}
/*-------------------- ligandGeometry --------------------*/

static void ligandGeometry(glLigand_t *L)
{
  int i;

  for (i=0;i<L->nbnd;i++) {
    L->B[i].d=cdist(L->B[i].l->a->x,
		       L->B[i].r->a->x);
  }

  for (i=0;i<L->nang;i++) {
    L->angle[i].w=cangled(L->A[L->angle[i].a[0]].a->x,
			  L->A[L->angle[i].a[1]].a->x,
			  L->A[L->angle[i].a[2]].a->x);
  }

  for (i=0;i<L->ndih;i++) {
    L->dihedral[i].w=ctorsd(L->A[L->dihedral[i].a[0]].a->x,
			    L->A[L->dihedral[i].a[1]].a->x,
			    L->A[L->dihedral[i].a[2]].a->x,
			    L->A[L->dihedral[i].a[3]].a->x);
  }

  for (i=0;i<L->nimp;i++) {
    L->improper[i].w=ctorsd(L->A[L->improper[i].a[0]].a->x,
			    L->A[L->improper[i].a[1]].a->x,
			    L->A[L->improper[i].a[2]].a->x,
			    L->A[L->improper[i].a[3]].a->x);
  }

  /* mark planar atoms */
  for (i=0;i<L->nat;i++) {
    L->A[i].planar=isPlanarAtom(L->A[i].a);
  }

}

/*-------------------- isPlanarAtom --------------------*/

static int isPlanarAtom(Atom *a)
{
  int r=FALSE;
  
  if (is_atom(a)) {
    if (a->nbnd==2) {
      r=cangled(neighbour(0,a)->x,a->x,neighbour(1,a)->x)<160;
    }else if (a->nbnd==3) {
      r=abs(ctorsd(neighbour(0,a)->x,a->x,neighbour(1,a)->x,neighbour(2,a)->x))>160;
    }
  }else r=17;
  return r;
}

/*-------------------- ligandTopology --------------------*/

static int ligandTopology(glLigand_t *L)
{
  int status;

  status=ligandRings(L);
  if (!status) {
    status=1;             /* should return number of connected components... */
  }else{
    status=-status;
  }
  return(status);
}
/*-------------------- ligandParameters --------------------*/

static int ligandParameters(glLigand_t *L,int mode)
{
  int r;

  switch (mode) {
  case GENLIG_PAR_CURRENT:
    r=ligandParametersCurrent(L);
    break;
  default:
    r=1;
    break;
  }
  return(r);
}
/*-------------------- ligandParametersCurrent --------------------*/

static int ligandParametersCurrent(glLigand_t *L)
{
  int i,k;
  float z;

  /* set all equilibrium values to current geometry */

  for (i=0;i<L->nbnd;i++) {
    L->B[i].d0=L->B[i].d;
    L->B[i].sigma=genlig_BS_sigma;
    L->B[i].k=genlig_conv/(L->B[i].sigma*L->B[i].sigma);
  }

  for (i=0;i<L->nang;i++) {
    L->angle[i].w0=L->angle[i].w;
    L->angle[i].sigma=genlig_AB_sigma;
    L->angle[i].k=genlig_convr/(L->angle[i].sigma*L->angle[i].sigma);
  }

  /* we do double, amide and aromatic torsions only... */
  /* added 8/8/2007: 'soft' ring torsions, for better ring conformations in weak electron density */
  for (i=0;i<L->ndih;i++) {
    if (L->B[L->dihedral[i].b].bt==BT_DOUBLE && L->B[L->dihedral[i].b].l->planar && L->B[L->dihedral[i].b].r->planar) {
      L->dihedral[i].n=2;
      L->dihedral[i].p=180.0;
      L->dihedral[i].sigma=genlig_DBL_sigma;
      z=L->dihedral[i].n*L->dihedral[i].sigma;
      L->dihedral[i].k=2*genlig_convr/(z*z*L->B[L->dihedral[i].b].M);
      L->dihedral[i].type=GENL_DIH_DOUBLE;
    }else if (L->B[L->dihedral[i].b].bt==BT_AMIDE) {
      L->dihedral[i].n=2;
      L->dihedral[i].p=180.0;
      L->dihedral[i].sigma=genlig_AMIDE_sigma;
      z=L->dihedral[i].n*L->dihedral[i].sigma;
      L->dihedral[i].k=2*genlig_convr/(z*z*L->B[L->dihedral[i].b].M);
      L->dihedral[i].type=GENL_DIH_AMIDE;
    }else if (L->B[L->dihedral[i].b].bt==BT_AROMATIC) {
      L->dihedral[i].n=2;
      L->dihedral[i].p=180.0;
      L->dihedral[i].sigma=genlig_AROMATIC_sigma;
      z=L->dihedral[i].n*L->dihedral[i].sigma;
      L->dihedral[i].k=2*genlig_convr/(z*z*L->B[L->dihedral[i].b].M);
      L->dihedral[i].type=GENL_DIH_AROMATIC;
    }else if (L->B[L->dihedral[i].b].rs>0&&L->B[L->dihedral[i].b].rs<10) {
      L->dihedral[i].type=GENL_DIH_RING;
      L->dihedral[i].n=3;                      /* should do better than this ...*/
      L->dihedral[i].p=L->dihedral[i].w;
      L->dihedral[i].sigma=genlig_RINGTORS_sigma;
      z=L->dihedral[i].n*L->dihedral[i].sigma;
      L->dihedral[i].k=2*genlig_convr/(z*z*L->B[L->dihedral[i].b].M);
    }else{
      L->dihedral[i].type=GENL_DIH_ZERO;
      L->dihedral[i].n=1;                      /* should do better than this ...*/
      L->dihedral[i].p=L->dihedral[i].w;
      L->dihedral[i].sigma=genlig_ZERO_sigma;
      z=L->dihedral[i].n*L->dihedral[i].sigma;
      L->dihedral[i].k=2*genlig_convr/(z*z*L->B[L->dihedral[i].b].M);
    }
  }

  /* close to planar --> force planar, close to sp3 --> keep configuration */
  for (i=0;i<L->nimp;i++) {
    k=L->improper[i].a[1];
    if (L->improper[i].w<-160||L->improper[i].w>160) {
      L->improper[i].p=180;
      L->improper[i].sigma=genlig_PL_sigma;
      L->improper[i].k=genlig_convr/(L->improper[i].sigma*L->improper[i].sigma);
      if (L->A[k].atp) {
      	if (L->A[k].atp->oop) L->improper[i].type=GENL_IMP_PLANAR;
      }else L->improper[i].type=GENL_IMP_PLANAR;
    }else if (( (L->A[k].z==6) || (L->A[k].z==7 && L->A[k].nn==4 )) &&
	      (L->improper[i].w>-150||L->improper[i].w<160)) {
      L->improper[i].p=L->improper[i].w;
      L->improper[i].sigma=genlig_INV_sigma;
      L->improper[i].k=genlig_convr/(L->improper[i].sigma*L->improper[i].sigma);
      L->improper[i].type=GENL_IMP_CHIRAL;
    }
  }

  return(0);
}

/*-------------------- ligandPlanes --------------------*/

static void ligandPlanes(glLigand_t *L)
{
  int i,j;
  float **x,**y,*Y;
  mat_4 m;

  for (i=0;i<L->ndih;i++) {
    if (L->dihedral[i].type==GENL_DIH_DOUBLE||
	L->dihedral[i].type==GENL_DIH_AMIDE||
	L->dihedral[i].type==GENL_DIH_AROMATIC) {
      addPlane(L,L->dihedral[i].a);
    }
  }

  for (i=0;i<L->nimp;i++) {
    if (L->improper[i].type==GENL_IMP_PLANAR) {
      addPlane(L,L->improper[i].a);
    }
  }

  for (i=0;i<L->nplan;i++) {
    if (L->plane[i].n>0) {
      x=malloc(L->plane[i].n*sizeof(float *));
      y=malloc(L->plane[i].n*sizeof(float *));
      Y=malloc(3*L->plane[i].n*sizeof(float));
      if (x&&y&&Y) {
	for (j=0;j<L->plane[i].n;j++) {
	  x[j]=L->A[L->plane[i].a[j]].a->x;
	  y[j]=Y+3*j;
	}
	m=PriaxMat(L->plane[i].n,x);
	ApplyTransf(m,L->plane[i].n,x,y);
	L->plane[i].rmsd=0;
	for (j=0;j<L->plane[i].n;j++) {
	  L->plane[i].d[j]=y[j][2];
	  L->plane[i].rmsd+=y[j][2]*y[j][2];
	}
	L->plane[i].rmsd=genlig_PLANE_sigma;
      }else{
	L->plane[i].rmsd=genlig_PLANE_sigma;
	for (j=0;j<L->plane[i].n;j++) L->plane[i].d[j]=L->plane[i].rmsd;
      }
      if (Y) free(Y);
      if (y) free(y);
      if (x) free(x);
    }
  }

}

static void addPlane(glLigand_t *L,int *a)
{
  int i,k,l,c,M,m;

  M=0;
  for (i=0;i<4;i++) L->A[a[i]].flg=1;

  do {
    for (m=0,k=0;k<L->nplan;k++) {
      for (c=0,l=0;l<L->plane[k].n;l++) {
	if (L->A[L->plane[k].a[l]].flg) c++;
      }
      if (c>2) {
	for (c=0,l=0;l<L->plane[k].n;l++) {
	  L->A[L->plane[k].a[l]].flg=1;
	}
	L->plane[k].n=0;
	m++;
      }
      M+=m;
    }  
  }while (m>0);

  if (M>0) {
    for (m=0,k=0;k<L->nplan;k++) {
      if (L->plane[k].n==0) {
	free_plane(L->plane[k]);
	m++;
      }else{
	L->plane[k-m]=L->plane[k];
      }
    }
    L->nplan-=m;
  }else{
    L->plane=realloc(L->plane,(L->nplan+1)*sizeof(glPlane_t));
  }
  for (c=0,i=0;i<L->nat;i++) if (L->A[i].flg) c++;
  L->plane[L->nplan]=new_plane(c);
  for (c=0,i=0;i<L->nat;i++) {
    if (L->A[i].flg) {
      L->plane[L->nplan].a[c++]=i;
    }
  }
  L->plane[L->nplan].n=c;
  L->nplan++;
  for (i=0;i<L->nat;i++) L->A[i].flg=0;
}

static void free_plane(glPlane_t p)
{
  if (p.a) free(p.a);
  if (p.d) free(p.d);
  p.a=NULL;
}

static glPlane_t new_plane(int c)
{
  glPlane_t p;

  p.n=c;
  p.a=malloc(c*sizeof(int));
  p.d=malloc(c*sizeof(float));
  p.rmsd=0;

  return(p);
}

/*-------------------- writeCNXtopology --------------------*/

static int writeCNXtopology(char *filename,glLigand_t *L)
{
  int i;
  FILE *out;
  int status=0;
  char buff[4096];

  out=fopen(extfil(buff,filename,".top",TRUE),"w");
  if (out) {
    fprintf(out,"{* Residue topology file generated by Wit!P *}\n\n");
    fprintf(out,"set echo off message off end\n");        /* for CNX, not so nice with X-Plor */
    fprintf(out,"AUTOGENERATE ANGLES=FALSE END\n\n");
    fprintf(out,"RESIDUE %s\n",L->code);
    fprintf(out,"GROUP\n");
    for (i=0;i<L->nat;i++) {
      fprintf(out,"  ATOM %-5s TYPE = %-4s CHARGE = %7.4f MASS = %7.4f  END\n",
	      L->A[i].name,L->A[i].type,
	      L->use_charges?L->A[i].charge:0.0,L->A[i].mass);
    }
    for (i=0;i<L->nbnd;i++) {
      fprintf(out,"  BOND %-5s %-5s\n",
		  L->B[i].l->name,L->B[i].r->name);
    }
    
    for (i=0;i<L->nang;i++) {
      fprintf(out,"  ANGL %-5s %-5s %-5s\n",
		  L->A[L->angle[i].a[0]].name,
		  L->A[L->angle[i].a[1]].name,
		  L->A[L->angle[i].a[2]].name);
    }
    for (i=0;i<L->ndih;i++) {
      if (L->dihedral[i].type>GENL_DIH_IGNORE) {
	fprintf(out,"  DIHE %-5s %-5s %-5s %-5s\n",
		L->A[L->dihedral[i].a[0]].name,
		L->A[L->dihedral[i].a[1]].name,
		L->A[L->dihedral[i].a[2]].name,
		L->A[L->dihedral[i].a[3]].name);
      }
    }
    for (i=0;i<L->nimp;i++) {
      if (L->improper[i].type>GENL_IMP_IGNORE) {
	fprintf(out,"  IMPR %-5s %-5s %-5s %-5s\n",
		  L->A[L->improper[i].a[0]].name,
		  L->A[L->improper[i].a[1]].name,
		  L->A[L->improper[i].a[2]].name,
		  L->A[L->improper[i].a[3]].name);
      }
    }
    fprintf(out,"END\n");
    fprintf(out,"set echo=$echo_last mess=$mess_last end\n");   /* for CNX, not so nice with X-Plor */
    if (ferror(out)) {
      err_set(1,"writeCNXtopology: error writing %s",
	      extfil(buff,filename,".top",TRUE));
      status=2;
    }
    fclose(out);
  }else{
    err_set(1,"writeCNXtopology: failed to open file %s for write",
	    extfil(buff,filename,".top",TRUE));
    status=1;
  }
  return(status);
}
/*-------------------- writeCNXparameters --------------------*/

static int writeCNXparameters(char *filename,glLigand_t *L)
{
  int i,j;
  int status=0;
  FILE *out;
  char buff[4096];
  float phi;

  out=fopen(extfil(buff,filename,".param",TRUE),"w");
  if (out) {
    fprintf(out,"{* Ligand parameter file generated by Wit!P *}\n");
    fprintf(out,"\nset echo off message off end\n\n");
    for (i=0;i<L->nbnd;i++) {
      fprintf(out,"BOND ( RESN %s and NAME %-4s ) ( RESN %s and NAME %-4s ) %8.1f %5.3f\n",
	      L->code,L->B[i].l->name,
	      L->code,L->B[i].r->name,
	      L->B[i].k,L->B[i].d0);
    }
    if (L->nbnd>0) fprintf(out,"\n");
    for (i=0;i<L->nang;i++) {
      fprintf(out,"ANGL ( RESN %s and NAME %-4s ) ( RESN %s and NAME %-4s ) ( RESN %s and NAME %-4s ) %8.1f %5.1f\n",
	      L->code,L->A[L->angle[i].a[0]].name,
	      L->code,L->A[L->angle[i].a[1]].name,
	      L->code,L->A[L->angle[i].a[2]].name,
	      L->angle[i].k,L->angle[i].w0);
    }
    if (L->nang>0) fprintf(out,"\n");
    for (i=0;i<L->ndih;i++) {
      if (L->dihedral[i].type>GENL_DIH_IGNORE) {
	phi=180-L->dihedral[i].n*L->dihedral[i].p;
	while (phi>180) phi-=360;
	while (phi<-180) phi+=360;
	if (L->dihedral[i].n>1) {
	  if (fabs(phi)<60) phi=0;
	  else if (fabs(fabs(phi)-180)<60) phi=180;
	}
	fprintf(out,"DIHE ( RESN %s and NAME %-4s ) ( RESN %s and NAME %-4s )\n"
		"     ( RESN %s and NAME %-4s ) ( RESN %s and NAME %-4s ) %6.1f %d %6.1f"
		"   {- %s %s -}\n\n",
		L->code,L->A[L->dihedral[i].a[0]].name,
		L->code,L->A[L->dihedral[i].a[1]].name,
		L->code,L->A[L->dihedral[i].a[2]].name,
		L->code,L->A[L->dihedral[i].a[3]].name, 
		L->dihedral[i].k,L->dihedral[i].n,phi,
		L->A[L->dihedral[i].a[1]].name,L->A[L->dihedral[i].a[2]].name);
      }
    }
    if (L->ndih>0) fprintf(out,"\n");
    for (i=0;i<L->nimp;i++) {
      if (L->improper[i].type>GENL_IMP_IGNORE) {
	fprintf(out,"IMPR ( RESN %s and NAME %-4s ) ( RESN %s and NAME %-4s )\n"
		"     ( RESN %s and NAME %-4s ) ( RESN %s and NAME %-4s ) %6.1f %d %6.1f"
		"   {- %s -}\n\n",
		L->code,L->A[L->improper[i].a[0]].name,
		L->code,L->A[L->improper[i].a[1]].name,
		L->code,L->A[L->improper[i].a[2]].name,
		L->code,L->A[L->improper[i].a[3]].name,
		L->improper[i].k,L->improper[i].n,L->improper[i].p,
		L->A[L->improper[i].a[1]].name);
      }
    }
    if (L->nimp>0) fprintf(out,"\n");
    for (i=0;i<NCNXVDW;i++) {
      for (j=0;j<L->nat;j++) {
	if (strceq(L->A[j].type,CNXvdw[i].type)) {
	  fprintf(out,"NONB ( RESN %s and CHEMICAL %-4s ) %7.4f %7.4f   %7.4f %7.4f\n",
		  L->code,CNXvdw[i].type,
		  CNXvdw[i].eps,CNXvdw[i].sig,CNXvdw[i].eps14,CNXvdw[i].sig14);
	  break;
	}
      }
    }

    fprintf(out,"\nset echo=$echo_last message=$mess_last end\n");
    if (ferror(out)) {
      err_set(1,"writeCNXtopology: error writing %s",
	      extfil(buff,filename,".param",TRUE));
      status=2;
    }
    fclose(out);
  }else{
    err_set(1,"writeCNXparameters: failed to open file %s for write",
	    extfil(buff,filename,".param",TRUE));
    status=1;
  }
  return(status);
}

/*-------------------- writeCNXpdb --------------------*/

static int writeCNXpdb(char *filename,glLigand_t *L)
{
  int i;
  FILE *out;
  int status=0;
  char buff[4096],chs[10],*p;

  out=fopen(extfil(buff,filename,".pdb",TRUE),"w");
  if (out) {
    fprintf(out,"REMARK Residue XPLOR/CNX PDB file generated by Wit!P\n");
    for (i=0;i<L->nat;i++) {
      strcpy(chs,L->A[i].a->atp->chs); for (p=chs;*p;p++) if (islower(*p)) *p=toupper(*p);
      fprintf(out,"%6s %4d %-4s %-4s%c%4d    %8.3f%8.3f%8.3f %5.2f %5.2f          %2s\n",
	      "HETATM",i+1,rotName(L->A[i].name,L->A[i].a->atp->chs),L->code,L->segid[0],1,
	      L->A[i].a->x[0],L->A[i].a->x[1],L->A[i].a->x[2],
	      L->A[i].a->x[ATM_SOF],L->A[i].a->x[ATM_BISO]>0?L->A[i].a->x[ATM_BISO]:20.0,
	      chs);
    }
    fprintf(out,"END\n");
    if (ferror(out)) {
      err_set(1,"writeCNXpdb: error writing %s",
	      extfil(buff,filename,".top",TRUE));
      status=2;
    }
    fclose(out);
  }else{
    err_set(1,"writeCNXpdb: failed to open file %s for write",
	    extfil(buff,filename,".pdb",TRUE));
    status=1;
  }
  return(status);
}
/*-------------------- writeDYANAdictionary --------------------*/

static int writeDYANAdictionary(char *filename,glLigand_t *L)
{
  FILE *out;
  int status=0,i,j,k,nn,nb[4],*bflg,ntor=0,*tor,*lst,sw,*T;
  char buff[4096];
  glAtom_t *A;
  float c[3];

  out=fopen(extfil(buff,filename,".drl",FALSE),"w");
  if (out) {

    /* get geometric center of ligand coordinates */
    for (k=0;k<3;k++) c[k]=0.0;
    for (i=0;i<L->nat;i++) {
      for (k=0;k<3;k++) c[k]+=L->A[i].a->x[k];
    }
    if (L->nat>1) for (k=0;k<3;k++) c[k]/=L->nat;

    /* get list of torsions to add to library entry */
    if (L->ndih>0) {
      bflg=(int *)malloc(L->nbnd*sizeof(int));
      tor=(int *)malloc(5*L->ndih*sizeof(int));
      lst=(int *)malloc(L->nat*sizeof(int));
    }else{
      bflg=(int *)NULL;
      tor=(int *)NULL;
      lst=(int *)NULL;
    }

    if (bflg&&tor&&lst) {
      /* build lst array: lst[i]=index of last atom affected by torsion with i as 
	 3rd atom. i,lst[i] are tree-indices. */
      for (i=0;i<L->nat;i++) lst[i]=i;
      for (i=L->nat-1;i>0;i--) {
	if (lst[L->T->new[L->T->up[i]]]<lst[i]) lst[L->T->new[L->T->up[i]]]=lst[i];
      }
      /* build list of torsion definitions fir dyana/cyana */
      for (i=0;i<L->nbnd;i++) bflg[i]=0;
      for (ntor=0,k=0;k<L->ndih;k++) {
	if (bflg[L->dihedral[k].b]==0) {   /* only one torsion per bond */
	  if (L->dihedral[k].type<=GENL_DIH_IGNORE) {
	    bflg[L->dihedral[k].b]=1;
	  }else if (rigidRings&&L->B[L->dihedral[k].b].rs!=0) {
	    bflg[L->dihedral[k].b]=1;
	  }else if (L->dihedral[k].sigma<noDyanaTorsSigma) {
	    bflg[L->dihedral[k].b]=1;      /* "stiff" torsion, ignore */                          
	  }else{
	    if (L->dihedral[k].a[2]==L->T->up[L->T->new[L->dihedral[k].a[1]]]) {
	      sw=L->dihedral[k].a[1]; L->dihedral[k].a[1]=L->dihedral[k].a[2]; L->dihedral[k].a[2]=sw;
	      sw=L->dihedral[k].a[0]; L->dihedral[k].a[0]=L->dihedral[k].a[3]; L->dihedral[k].a[3]=sw;
	    }
	    if (L->dihedral[k].a[1]==L->T->up[L->T->new[L->dihedral[k].a[2]]]) {
	      if (L->dihedral[k].a[2]==L->T->up[L->T->new[L->dihedral[k].a[3]]] && 
		(L->dihedral[k].a[0]==L->T->up[L->T->new[L->dihedral[k].a[1]]] ||
		 L->dihedral[k].a[1]==L->T->up[L->T->new[L->dihedral[k].a[0]]] ) ) {
		T=tor+5*ntor;
		for (j=0;j<4;j++) T[j]=L->T->new[L->dihedral[k].a[j]]+4;
		T[4]=lst[L->T->new[L->dihedral[k].a[2]]]+4;
		ntor+=1;
		bflg[L->dihedral[k].b]=1;
	      }
	    }else{
	      bflg[L->dihedral[k].b]=1;
	    }
	  }
	}
      }
      /* order torsion list so that dyana/cyana will be happy */
      qsort(tor,ntor,5*sizeof(int),dyanaTorSort);
    }
    if (bflg) free(bflg);

    /* write residue entry header line */
    fprintf(out,"%-10s%-5s%5d%5d%5d%5d\n","RESIDUE",L->code,ntor+1,L->nat+6,3,L->nat+5);

    /* write list of torsions */
    fprintf(out,"%4d %-5s%5d%5d%10.4f%5d%5d%5d%5d%5d\n",1,"LB",0,0,0.0,2,3,L->nat+5,L->nat+6,0);
    for (i=0;i<ntor;i++) {
      T=tor+5*i;
      fprintf(out,"%4d T%03d %5d%5d%10.4f%5d%5d%5d%5d%5d\n",i+2,i+1,0,0,0.0,T[0],T[1],T[2],T[3],T[4]);
    }

    /* first three atoms are pseudo-atoms for linking with PL, LL, ... residues */
    fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
	    1,"Q1","DUMMY",0,0.0,0.0,0.0,0.0,0,0,0,0,0);
    fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
	    2,"Q2","DUMMY",0,0.0,1.0,0.0,0.0,0,0,0,0,0);
    fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
	    3,"Q3","DUMMY",0,0.0,1.0,1.0,0.0,0,0,0,0,0);

    /* ligand atoms, output in tree-order */
    for (i=0;i<L->nat;i++) {
      A=L->A+L->T->nr[i];
      for (j=0;j<4;j++) nb[j]=0;
      nn=A->nn<4?A->nn:4;
      for (j=0;j<nn;j++) nb[j]=L->T->new[A->nb[j]]+4;/* +4: in L atoms are 0...nat-1, in file they are 4...nat+3 */
      fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
	      i+4,A->name,A->type,0,0.0,A->a->x[0]-c[0],A->a->x[1]-c[1],A->a->x[2]-c[2],nb[0],nb[1],nb[2],nb[3],0);
    }

    /* last three atoms are pseudo-atoms for linking with PL, LL, ... residues */
    fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
	    L->nat+4,"Q1","DUMMY",0,0.0,2.0,1.0,0.0,0,0,0,0,0);
    fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
	    L->nat+5,"Q2","DUMMY",0,0.0,2.0,0.0,0.0,0,0,0,0,0);
    fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
	    L->nat+6,"Q3","DUMMY",0,0.0,3.0,0.0,0.0,0,0,0,0,0);

    if (lst) free(lst);
    if (tor) free(tor);

    fclose(out);
  }else{
    err_set(1,"failed to open dictionary file %s for write\n",extfil(buff,filename,".drl",TRUE),"w");
    status=1;
  }
  return(status);
}
/*-------------------- dyanaTorSort --------------------*/

static int dyanaTorSort(const void *a,const void *b)
{
  int *A,*B,r=0;

  /* rotation about torsion must not affect atoms in
     torsion definitions earlier in the torsion list.
     a and b point to torsion definitions (5-tuple of
     atom numbers (tree numbering). Torsion a affects
     atoms a[3]...a[4]. These number must not occur in
     b[0],b[1],b[2],b[3] of torsions b preceeding a in
  */

  A=(int *)a;
  B=(int *)b;

  if (
      (A[0]>B[2]&&A[0]<=B[4]) ||
      (A[1]>B[2]&&A[1]<=B[4]) ||
      (A[2]>B[2]&&A[2]<=B[4]) ||
      (A[3]>B[2]&&A[3]<=B[4]) ) {
    r=+1;

  }else if (
	    (B[0]>A[2]&&B[0]<=A[4]) ||
	    (B[1]>A[2]&&B[1]<=A[4]) ||
	    (B[2]>A[2]&&B[2]<=A[4]) ||
	    (B[3]>A[2]&&B[3]<=A[4]) ) {
    r=-1;
  
  }else if (B[2]<A[2]) {
    r=+1;
  }else if (A[2]<B[2]) {
    r=-1;
  }
  return r;
}

/*-------------------- writeDYANArcMacro --------------------*/

static int writeDYANArcMacro(char *filename,glLigand_t *L)
{
  FILE *out;
  char buff[4096];
  int status=0,nc,i,j,k,l,n,t;
  float d;
  float dtol=0.000,atol=0.00,ttol=0.00;   /* 0: user should use dtol and atol macro parameters to set tolerances */

  /* generate CYANA ring-closure macro for this ligand type */

  out=fopen(extfil(buff,filename,".cya",TRUE),"w");
  if (out) {
    fprintf(out,rcmac1,L->code);
    fprintf(out,"%s",rcmac2);
    nc=L->nbnd+1-L->nat;
    if (nc>0 && !rigidRings) {
      fprintf(out,"do i 1 $nr\n");
      fprintf(out,"   if (rnam(irnum(i)).eq.'%s') then\n",L->code);
      fprintf(out,"      print \"    Added ring-closure restraint for residue $rnam(irnum(i)) $i\"\n");
      fprintf(out,"      res=irnum(i)\n");
      for (l=L->nat;l<L->nat+nc;l++) {
	i=L->T->nr[l];
	j=L->T->up[l];
	/* ring closure distance */
	d=cdist(L->A[i].a->x,L->A[j].a->x);
	fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.2f-$dtol upl=%0.2f+$dtol weight=$weight\n",
		L->A[i].name,L->A[j].name,d-dtol,d+dtol);
	/* angles at first ring closure atom */
	for (k=0;k<L->A[i].nn;k++) {
	  if (L->A[i].nb[k]!=j) {
	    n=L->A[i].nb[k];
	    d=cdist(L->A[n].a->x,L->A[j].a->x);
	    fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.2f-$atol upl=%0.2f+$atol weight=$weight\n",
		    L->A[n].name,L->A[j].name,d-atol,d+atol);
	  }
	}
	/* angles at second ring closure atom */
	for (k=0;k<L->A[j].nn;k++) {
	  if (L->A[j].nb[k]!=i) {
	    n=L->A[j].nb[k];
	    d=cdist(L->A[i].a->x,L->A[n].a->x);
	    fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.2f-$atol upl=%0.2f+$atol weight=$weight\n",
		    L->A[i].name,L->A[n].name,d-atol,d+atol);
	  }
	}
#ifdef VARIANTE1 
	/* torsions for "stiff" non-aromatic ring closure bonds */
	for (t=0;t<L->ndih;t++) {
	  if (L->dihedral[t].sigma<noDyanaTorsSigma &&    /* stiff torsion */
	      L->B[L->dihedral[t].b].rs!=0 &&             /* in a ring */
	      (( i==L->dihedral[t].a[1] && j==L->dihedral[t].a[2] ) || 
	       ( i==L->dihedral[t].a[2] && j==L->dihedral[t].a[1] ))
	      ) {
	    d=cdist(L->A[L->dihedral[t].a[0]].a->x,L->A[L->dihedral[t].a[3]].a->x);
	    fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.2f-$ttol upl=%0.2f+$ttol weight=$weight\n",
		    L->A[L->dihedral[t].a[0]].name,L->A[L->dihedral[t].a[3]].name,d-ttol,d+ttol);
	  }
	}
#endif
      }
      fprintf (out,"\n");
      for (t=0;t<L->ndih;t++) {
	if (L->dihedral[t].sigma<noDyanaTorsSigma &&    /* stiff torsion */
	    L->B[L->dihedral[t].b].rs!=0                /* in a ring */
	    ) {                                         /* overkill, many torsions do not need this... */
	  d=cdist(L->A[L->dihedral[t].a[0]].a->x,L->A[L->dihedral[t].a[3]].a->x);
	  fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.2f-$ttol upl=%0.2f+$ttol weight=$weight\n",
		  L->A[L->dihedral[t].a[0]].name,L->A[L->dihedral[t].a[3]].name,d-ttol,d+ttol);
	}
      }
      
      fprintf(out,"   end if\n");
      fprintf(out,"end do\n\n");
    }
    fprintf(out,"unset info\n");
    fclose(out);
  }else{
    err_set(1,"failed to open CYANA macro file %s for write\n",extfil(buff,filename,".cya",TRUE),"w");
    status=1;
  }

  return(status);
}

/*-------------------- writeREFMACdictionary --------------------*/

static int writeREFMACdictionary(char *filename,glLigand_t *L)
{
  FILE *out;
  char buff[4096],*bt,*pref,**qnames=NULL;
  int i,k,m,mconst=0,mvar=0,nc,status=0,period,*bflg=(int *)NULL;
  float rmsd,sigma;

  qnames=singleQuotedNames(L);
  if (!qnames) {
    err_set(1,"memory allocation failed in writeREFMACdictionary");
    return(2);
  }

  out=fopen(extfil(buff,filename,".lib",FALSE),"w");
  if (out) {
    
    fprintf(out,"%s\n%s\n%s\n\n",
	    "# ------------------------------------------------",
	    "# ---   REFMAC dictionary generated by Wit!P   ---",
	    "# ------------------------------------------------");

    fprintf(out,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
	    "data_comp_list",
	    "loop_",
	    "_chem_comp.id",
	    "_chem_comp.three_letter_code",
	    "_chem_comp.name",
	    "_chem_comp.group",
	    "_chem_comp.number_atoms_all",
	    "_chem_comp.number_atoms_nh",
	    "_chem_comp.desc_level");
    
    fprintf(out," %-8s %-8s %s %s %10d %4d   %s\n",
	    L->code,
	    L->code,
	    ".",
	    "non-polymer",
	    L->nat,
	    L->nonh,
	    ".");
 
    fprintf(out,"\n%s\n%s\n%s\n\n",
	    "# -------------------------------",
	    "# --- DESCRIPTION OF MONOMERS ---",
	    "# -------------------------------");
    
    fprintf(out,"data_comp_%s\n\n",L->code);
    fprintf(out,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
	    "loop_",
	    "_chem_comp_atom.comp_id",
	    "_chem_comp_atom.atom_id",
	    "_chem_comp_atom.type_symbol",
	    "_chem_comp_atom.type_energy",
	    "_chem_comp_atom.partial_charge",
	    "_chem_comp_atom.x",
	    "_chem_comp_atom.y",
	    "_chem_comp_atom.z");
    for (i=0;i<L->nat;i++) {       /* any order, or tree order ? */
      m=i;
      fprintf(out," %-6s %-6s %-2s %-5s",
	      L->code,
	      qnames[m],
	      L->A[m].a->atp->chs,
	      L->A[m].type);
      if (L->use_charges) fprintf(out," %7.4f",L->A[m].charge);
      else fprintf(out,"    .  ");
      fprintf(out," %9.3f %9.3f %9.3f\n",
	      L->A[m].a->x[0],
	      L->A[m].a->x[1],
	      L->A[m].a->x[2]);
    }

    if (L->T) {
      fprintf(out,"\n%s\n%s\n%s\n\n",
	      "# ---------------------------------------",
	      "# --- SPANNING TREE AND RING CLOSURES ---",
	      "# ---------------------------------------");

      fprintf(out,"%s\n%s\n%s\n%s\n%s\n%s\n",
	      "loop_",
	      "_chem_comp_tree.comp_id",
	      "_chem_comp_tree.atom_id",
	      "_chem_comp_tree.atom_back",
	      "_chem_comp_tree.atom_forward",
	      "_chem_comp_tree.connect_type");

      for (i=0;i<L->nat;i++) {
	fprintf(out," %-6s %-6s %-6s %-6s %-8s\n",
		L->code,
		qnames[L->T->nr[i]],
		L->T->up[i]<0?"n/a":qnames[L->T->up[i]],
		L->T->fw[i]<0?".":qnames[L->T->fw[i]],
		i==0?"START":i<L->nat-1?".":"END"
		);
      }
      nc=L->nbnd+1-L->nat;
      for (i=L->nat;i<L->nat+nc;i++) {
	fprintf(out," %-6s %-6s %-6s %-6s %-8s\n",
		L->code,
		qnames[L->T->nr[i]],
		qnames[L->T->up[i]],
		".",
		"ADD"
		);
      }
    }
    
    fprintf(out,"\n%s\n%s\n%s\n\n",
	    "# ------------------------",
	    "# --- BOND RESTRAINTS ---",
	    "# ------------------------");

    fprintf(out,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
	    "loop_",
	    "_chem_comp_bond.comp_id",
	    "_chem_comp_bond.atom_id_1",
	    "_chem_comp_bond.atom_id_2",
	    "_chem_comp_bond.type",
	    "_chem_comp_bond.value_dist",
	    "_chem_comp_bond.value_dist_esd");
    for (i=0;i<L->nbnd;i++) {
      bt=(L->B[i].bt==BT_SINGLE   ? "single" :
	  L->B[i].bt==BT_DOUBLE   ? "double" :
	  L->B[i].bt==BT_TRIPLE   ? "triple" :
	  L->B[i].bt==BT_AMIDE    ? "deloc" :
	  L->B[i].bt==BT_AROMATIC ? "aromatic" : "single" );
      fprintf(out," %-6s %-6s %-6s %-8s %5.3f %4.3f\n",
	      L->code,
	      qnames[L->B[i].a[0]],
	      qnames[L->B[i].a[1]],
	      bt,
	      L->B[i].d0,
	      L->B[i].sigma);
    }

    fprintf(out,"\n%s\n%s\n%s\n\n",
	    "# -------------------------",
	    "# --- ANGLE RESTRAINTS ---",
	    "# -------------------------");

    fprintf(out,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
	    "loop_",
	    "_chem_comp_angle.comp_id",
	    "_chem_comp_angle.atom_id_1",
	    "_chem_comp_angle.atom_id_2",
	    "_chem_comp_angle.atom_id_3",
	    "_chem_comp_angle.value_angle",
	    "_chem_comp_angle.value_angle_esd");
    for (i=0;i<L->nang;i++) {
      fprintf(out," %-6s %-6s %-6s %-6s %8.3f %6.3f\n",
	      L->code,
	      qnames[L->angle[i].a[0]],
	      qnames[L->angle[i].a[1]],
	      qnames[L->angle[i].a[2]],
	      L->angle[i].w0,
	      L->angle[i].sigma);
    }


    if (L->ndih>0) bflg=malloc(L->nbnd*sizeof(int));
    if (bflg) {
      for (k=0;k<L->nbnd;k++) bflg[k]=0;
      fprintf(out,"\n%s\n%s\n%s\n\n",
	      "# --------------------------",
	      "# --- TORSION RESTRAINTS ---",
	      "# --------------------------");

      fprintf(out,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
	      "loop_",
	      "_chem_comp_tor.comp_id",
	      "_chem_comp_tor.id",
	      "_chem_comp_tor.atom_id_1",
	      "_chem_comp_tor.atom_id_2",
	      "_chem_comp_tor.atom_id_3",
	      "_chem_comp_tor.atom_id_4",
	      "_chem_comp_tor.value_angle",
	      "_chem_comp_tor.value_angle_esd",
	      "_chem_comp_tor.period");

      for (m=0,k=0;k<L->ndih;k++) {
	if (bflg[L->dihedral[k].b]) continue;   /* only one torsion per bond */
	if (L->dihedral[k].type<=GENL_DIH_IGNORE) continue;
	bflg[L->dihedral[k].b]=1;
	if (L->dihedral[k].type==GENL_DIH_ZERO) {
	  sigma=90.0;
	  period=3;
	}else{
	  sigma=L->dihedral[k].sigma;
	  period=L->dihedral[k].n;
	}
	if (L->B[L->dihedral[k].b].rs) {
	  m=++mconst;
	  pref="CONST_";
	  sigma=0;
	  period=0;
	}else{
	  m=++mvar;
	  pref="var_";
	  sigma=L->dihedral[k].sigma;
	  period=L->dihedral[k].n;
	}
	fprintf(out," %-6s %6s%03d %-6s %-6s %-6s %-6s %5.0f %5.1f %2d\n",
		L->code,
		pref,m,
		qnames[L->dihedral[k].a[0]],
		qnames[L->dihedral[k].a[1]],
		qnames[L->dihedral[k].a[2]],
		qnames[L->dihedral[k].a[3]],
		L->dihedral[k].w,
		sigma,
		period
		);
      }
    }
    if (bflg) free(bflg);

    for (k=0,i=0;i<L->nimp;i++) if (L->improper[i].type==GENL_IMP_CHIRAL) k++;
    if (k>0) {
      fprintf(out,"\n%s\n%s\n%s\n\n",
	      "# -----------------------------",
	      "# --- CHIRALITY RESTRAINTS ---",
	      "# -----------------------------");

      fprintf(out,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
	      "loop_",
	      "_chem_comp_chir.comp_id",
	      "_chem_comp_chir.id",
	      "_chem_comp_chir.atom_id_centre",
	      "_chem_comp_chir.atom_id_1",
	      "_chem_comp_chir.atom_id_2",
	      "_chem_comp_chir.atom_id_3",
	      "_chem_comp_chir.volume_sign");
      
      /* Note: volume sign may be "positiv" (sic), "negativ", "both" or "anomer"
         "positiv", if (b1 x b2, b3) > 0, wher b1, b2, b3 are the three vectors
         from the center to the three neighbours, "both" if the center is allowed
         to invert during refinement, "anomer" if the center is fixed, with the
         sign of (b1 x b2, b3) determined by the input coordinates. */

      for (k=0,i=0;i<L->nimp;i++) {
	if (L->improper[i].type==GENL_IMP_CHIRAL) {
	  k++;
	  fprintf(out," %-6s %s%03d %-6s %-6s %-6s %-6s %s\n",
		  L->code,
		  "chir_",k,
		  qnames[L->improper[i].a[1]],
		  qnames[L->improper[i].a[0]],
		  qnames[L->improper[i].a[2]],
		  qnames[L->improper[i].a[3]],
		  L->improper[i].w<0 ? "positiv" : "negativ" );
	}
      }
    }

    if (L->nplan>0) {
      fprintf(out,"\n%s\n%s\n%s\n\n",
	      "# -----------------------------",
	      "# --- PLANARITY RESTRAINTS ---",
	      "# -----------------------------");

      fprintf(out,"%s\n%s\n%s\n%s\n%s\n",
	      "loop_",
	      "_chem_comp_plane_atom.comp_id",
	      "_chem_comp_plane_atom.plane_id",
	      "_chem_comp_plane_atom.atom_id",
	      "_chem_comp_plane_atom.dist_esd");
      for (i=0;i<L->nplan;i++) {
	for (k=0;k<L->plane[i].n;k++) {
	  rmsd=(L->plane[i].rmsd<0.005?0.005:L->plane[i].rmsd);
	  fprintf(out," %-6s %s%03d %-6s %6.3f\n",
		  L->code,
		  "PL-",i+1,
		  qnames[L->plane[i].a[k]],
		  rmsd);
	}
      }
    }

    if (ferror(out)) {
      err_set(1,"writeREFMACdictionary: error writing %s",
	      extfil(buff,filename,".lib",FALSE));
      status=2;
    }
    fclose(out);
  }else{
    err_set(1,"failed to open dictionary file %s for write\n",extfil(buff,filename,".lib",FALSE),"w");
    status=1;
  }
  qnames=freeQuotedNames(L->nat,qnames);
  return(status);
}

/*-------------------- writeREFMACpdb --------------------*/

static int writeREFMACpdb(char *filename,glLigand_t *L)
{
  int i;
  FILE *out;
  int status=0;
  char buff[4096],chs[10],*p;

  out=fopen(extfil(buff,filename,".pdb",TRUE),"w");
  if (out) {
    fprintf(out,"REMARK Residue REFMAC PDB file generated by Wit!P\n");
    for (i=0;i<L->nat;i++) {
      strcpy(chs,L->A[i].a->atp->chs); for (p=chs;*p;p++) if (islower(*p)) *p=toupper(*p);
      fprintf(out,"%6s %4d %-4s %-4s%c%4d    %8.3f%8.3f%8.3f %5.2f %5.2f          %2s\n",
	      "HETATM",i+1,rotName(L->A[i].name,L->A[i].a->atp->chs),
	      L->code,L->segid[0],1,
	      L->A[i].a->x[0],L->A[i].a->x[1],L->A[i].a->x[2],
	      L->A[i].a->x[ATM_SOF],L->A[i].a->x[ATM_BISO]>0?L->A[i].a->x[ATM_BISO]:20.0,
	      chs);
    }
    fprintf(out,"END\n");
    if (ferror(out)) {
      err_set(1,"writeREFMACpdb: error writing %s",
	      extfil(buff,filename,".top",TRUE));
      status=2;
    }
    fclose(out);
  }else{
    err_set(1,"writeREFMACpdb: failed to open file %s for write",
	    extfil(buff,filename,".pdb",TRUE));
    status=1;
  }
  return(status);
}
/*-------------------- fixCNXAtomnames --------------------*/

static void fixCNXAtomnames(glLigand_t *L)
{
  fixRefmacAtomnames(L);         /* use the REFMAC rules for CNX */
}

/*-------------------- fixCNXAtomnames --------------------*/

static void fixDyanaAtomnames(glLigand_t *L)
{
  fixRefmacAtomnames(L);         /* use the REFMAC rules for CNX */
}

/*-------------------- fixRefmacAtomnames --------------------*/

static void fixRefmacAtomnames(glLigand_t *L)
{
  int i,j,k,fixit;
  char *p,c;
  
  /* atomnames must be all uppercase, at most four characters, unique */
  /* atomnames must start with chemical symbol */
  /* if not, new names will be generated from seq# and chemical symbol. */

  for (fixit=FALSE,i=0;i<L->nat;i++) {
    for (p=L->A[i].name;*p;p++) *p=toupper(*p);
    if (strlen(L->A[i].name)>2+strlen(L->A[i].a->atp->chs)) fixit=TRUE;
    fixit=fixit||strncasecmp(L->A[i].name,L->A[i].a->atp->chs,strlen(L->A[i].a->atp->chs));
    for (j=0;!fixit&&j<i;j++) {
      fixit=fixit||streq(L->A[i].name,L->A[j].name);
    }
  }
  if (fixit) {
    for(i=0;i<99&&i<L->nat;i++) {
      sprintf(L->A[i].name,"%s%d",L->A[i].a->atp->chs,i+1);
      L->A[i].name[1]=toupper(L->A[i].name[1]);
    }
    for (k=0,c='A';i<L->nat;i++) {
      sprintf(L->A[i].name,"%s%d%c",L->A[i].a->atp->chs,k,c);
      if (c=='Z') {
	c='A';
	k++;
	if (k>9) break;
      }else c++;
      L->A[i].name[1]=toupper(L->A[i].name[1]);
    }
    for (k=65,c='0';i<L->nat;i++) {
      sprintf(L->A[i].name,"%s%c%c",L->A[i].a->atp->chs,k,c);
      if (c=='9') {
	c='A';
      }else if (c=='Z') {
	k++;                /* should test: if k>90, failure. */
	c='0';
      }else c++;
      L->A[i].name[1]=toupper(L->A[i].name[1]);
    }
  }
}
/*-------------------- rotName --------------------*/

static /* NON-REENTRANT */ char *rotName(char *name,char *chs)
{
  static char rotatedName[6];
    
  /* return pointer to static buffer with rotated atom name        */
  /* note: 'name' must be a 'fixed' name (i.e. at most 4 chars,    */
  /* all upper-case letters and digits, starting with chem. symbol */
  
  strcpy(rotatedName,"     ");        /* start with blank name */

  if (strlen(chs)==1) {
    if (streq(chs,"H")) {
      if (strlen(name)>3) {
	strncpy(rotatedName,name,4);
      }else{
	strncpy(rotatedName+1,name,4);
      }
      rotatedName[4]='\0';
    }else{
      strncpy(rotatedName+1,name,4);   /* single letter chem. symbol, shift right */
    }
    if (strlen(rotatedName)>4) 
      rotatedName[0]=rotatedName[4];  /* copy 5-th char into first position */
  }else{
    strncpy(rotatedName,name,4);     /* two letter chem. symbol, or no name at all */
  }
  rotatedName[4]='\0';
  return(rotatedName);
}
/*-------------------- findAtomTypes --------------------*/

static void findAtomTypes(glLigand_t *L,Node *R,char *rules)
{
  int i,k;
  Node *ResLib,*pruleset,*frules;
  Atomdef *t;
  Prule *pr;
  Frule *fr;

  for (i=0;i<L->nat;i++) L->A[i].atp=NULL;

  /* if there is a library with type rules (Prules and Frules), use it */

  if (R) {
    ResLib=find_name(R,"ResLib",OBJ_LIB);
    pruleset=find_name(find_name(R,"PruleLib",OBJ_LIB),rules,OBJ_LIB);
    if (is_lib(pruleset)) {
      for (i=0;i<L->nat;i++) { /* apply Prules (if any) ... */
	ForPrules(pr,pruleset) {
	  if (matchPruleAtom(L->A[i].a,pr,0,0)) {
	    L->A[i].atp=find_atomdef_name(ResLib,pr->name);
	    if (is_atomdef(L->A[i].atp)) break;
	  }
	}
      }
    }
    frules=find_name(find_name(R,"FruleLib",OBJ_LIB),rules,OBJ_LIB);
    if (is_lib(frules)) {
      for (i=0;i<L->nat;i++) {   /* swap real types with genlig types */
	t=L->A[i].a->atp;
	L->A[i].a->atp=L->A[i].atp;
	L->A[i].atp=t;
      }
      for (i=0;i<L->nat;i++) {   /* apply Frules (if any)... */
	ForFrules(fr,frules) {
	  if (matchFrule(L->A[i].a,fr)) {
	    for (k=0;k<fr->nat;k++) {
	      if (is_atomdef(fr->oatp[k])) {
		if (fr->oatp[k]!=fr->match[k]->atp) {
		  fr->match[k]->atp=fr->oatp[k];
		}
	      }
	    }
	    break;
	  }
	}
      }
      for (i=0;i<L->nat;i++) {   /* swap back real types */
	t=L->A[i].a->atp;
	L->A[i].a->atp=L->A[i].atp;
	L->A[i].atp=t;
      }
    }
  }

  for (i=0;i<L->nat;i++) {
    if (L->A[i].atp) {
      strncpy(L->A[i].type,L->A[i].atp->name,7);
      L->A[i].type[7]='\0';
    }else{
      strncpy(L->A[i].type,L->A[i].a->atp->chs,2); 
      L->A[i].type[2]='\0';
      if (L->A[i].type[1]>='a'&&L->A[i].type[1]<='z') L->A[i].type[1]-='a'-'A';
    }
  }
}
/*-------------------- genligRoot --------------------*/

static Node *genligRoot(int *status)
{
  Token *key;
  Node *root=(Node *)NULL;
  
  key=Get_Mol_D("molecule: ",";");
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) root=(Node *)key->ptr;
  else if (key->status==TOK_NOK) err_set(1,"invalid molecule: %s\n",key->text);
  *status=key->status;

  key=free_Token(key);
  return(root);
}
/*-------------------- ligandRings --------------------*/

static int ligandRings(glLigand_t *M)
{
  /* Foreach bond in Molecule M: determine size of smallest ring containing
     bond (0 if bond is not in a ring). Return: error status, 0 if no error. 
   
     Ring size of j-th bond will be stored in M->B[i].rs.
  */

  int i,j,k,l,n,status=0;
  glAtom_t **AL=(glAtom_t **)NULL;
  glAtom_t *m;

  if (M&&M->nat>0&&M->nbnd>0) {

    /* Allocate scratch array for explore list: */

    AL=(glAtom_t **)malloc(M->nat*sizeof(glAtom_t *));

    if (AL) {
      for (j=0;j<M->nbnd;j++) M->B[j].rs=M->nbnd+1;  /* set ring status 'unknown'                 */
      for (i=0;i<M->nat;i++) M->A[i].flg=0;          /* clear all atomflags                       */
      for (j=0;j<M->nbnd;j++) {                      /* for each bond...                          */
	M->B[j].rs=0; l=0;
	if (M->B[j].l->nn>1&&M->B[j].r->nn>1) {      /* ...which is not terminal...               */
	  M->B[j].l->flg=-1;                         /* set left atom flag                        */
	  M->B[j].r->flg=+1;                         /* set right atom flag                       */
	  AL[l++]=M->B[j].l;                         /* add left atom to explore list             */
	  AL[l++]=M->B[j].r;                         /* add right atom to explore list            */
	  for (k=0;k<l;k++) {                        /* while list has unexplored atoms...        */
	    if (AL[k]->flg<0) {                      /* if the atom is on the left of bond j:     */
	      for (n=0;n<AL[k]->nn;n++) {            /* examine each neighbor atom                */
		m=AL[k]->n[n];
		if (m->flg==0) {                     /* if neighbor has no flag                   */
		  if (AL[k]->b[n]->rs!=0) {          /* and if not bonded via known chain bond    */
		    m->flg=AL[k]->flg-1;             /* flag it (parent flag - 1)                 */
		    AL[l++]=m;                       /* and add it to list of atoms to explore    */
		  }                                  /* else ignore it                            */
		}else{                               /* atom has a flag already...                */
		  if (m->flg>1) {                    /* flag indicates ring closure               */
		    M->B[j].rs=m->flg-AL[k]->flg;    /* ring size of bond j is sum of abs(flags)  */
		    k=l;                             /* no need to continue with explore list     */
		    break;
		  }
		}
	      }
	    }else{                                   /* if the atom is on the right of bond j:    */
	      for (n=0;n<AL[k]->nn;n++) {            /* examine each neighbor atom                */
		m=AL[k]->n[n];
		if (m->flg==0) {                     /* if neighbor has no flag                   */
		  if (AL[k]->b[n]->rs!=0) {          /* and if not bonded via known chain bond    */
		    m->flg=AL[k]->flg+1;             /* flag it (parent flag + 1)                 */
		    AL[l++]=m;                       /* and add it to list of atoms to explore    */
		  }                                  /* else ignore it                            */
		}else{                               /* atom has a flag already...                */
		  if (m->flg<-1) {                   /* flag indicates ring closure               */
		    M->B[j].rs=-m->flg+AL[k]->flg;   /* ring size of bond j is sum of abs(flags)  */
		    k=l;                             /* no need to continue with explore list     */
		    break;
		  }
		}
	      }
	    }
	  }
	}else M->B[j].rs=0;                          /* terminal bonds have rs 0                  */
	for (k=0;k<l;k++) AL[k]->flg=0;              /* reset atom flags for next bonds analysis  */
      }
      status=0;                                      /* no error.                                 */
    }else{
      status=1;                                      /* error status: malloc failed.              */
      for (j=0;j<M->nbnd;j++) M->B[j].rs=M->nbnd+1;  /* undefined ring size...                    */
    }
  }

  if (AL) free(AL);

  return(status);
}

/*-------------------- glLigand functions --------------------*/

static glLigand_t *newLigand(void)
{
  glLigand_t *L;

  L=(glLigand_t *)malloc(sizeof(glLigand_t));
  if (L) {
    L->nat=0;   L->A=(glAtom_t *)NULL;
    L->nbnd=0;  L->B=(glBond_t *)NULL;
    L->nang=0;  L->angle=(glAngle_t *)NULL;
    L->ndih=0;  L->dihedral=(glDihedral_t *)NULL;
    L->nimp=0;  L->improper=(glImproper_t *)NULL;
    L->nplan=0; L->plane=(glPlane_t *)NULL;
    L->use_charges=0;
    L->hyd_mode=HYD_NONE;
  }
  return(L);
}

static glLigand_t *freeLigand(glLigand_t *L)
{
  int k;

  if (L) {
    if (L->T) L->T=freeLigandTree(L->T);
    if (L->A) free (L->A);
    if (L->B) free (L->B);
    if (L->angle) free (L->angle);
    if (L->dihedral) free (L->dihedral);
    if (L->improper) free (L->improper);
    if (L->nplan>0&&L->plane) {
      for (k=0;k<L->nplan;k++) free_plane(L->plane[k]);
      free(L->plane);
    }
    free(L);
  }
  return((glLigand_t *)NULL);
}
/*-------------------- the forest --------------------*/

static glLigandTree_t *ligandTree(glLigand_t *L,int root)
{
  glLigandTree_t *T;
  int k,p,q,i,j,l,m,n,nc;
  int nb[MAXLIGATMNB];
  int *sz=(int *)NULL,*nh=(int *)NULL,*lb=(int *)NULL,*up=(int *)NULL;

  T=newLigandTree(L->nbnd+1);
  if (T) {
    sz=malloc(L->nat*sizeof(int));
    nh=malloc(L->nat*sizeof(int));
    lb=malloc(L->nat*sizeof(int));
    up=malloc(L->nat*sizeof(int));
    if (!sz||!nh||!lb||!up) {
      if (sz) free(sz);
      if (nh) free(nh);
      if (lb) free(lb);
      if (up) free(up);
      T=freeLigandTree(T);
    }
  }
  if (T) {
    /* generate tree */
    for (k=0;k<L->nat;k++) T->new[k]=-1;
    nc=0;
    T->up[0]=-1;
    T->nr[0]=root;
    T->new[root]=0;
    for (k=1,p=0;p<k;p++) {
      m=T->nr[p];
      for (i=0;i<L->A[m].nn;i++) {
	n=L->A[m].nb[i];
	if (T->new[n]<0) {
	  T->nr[k]=n;
	  T->new[n]=k;
	  T->up[k]=m;
	  k++;
	}else if (n!=T->up[p]) {
	  if (n>m) {
	    T->nr[L->nat+nc]=m;
	    T->up[L->nat+nc]=n;
	    T->fw[L->nat+nc]=-1;
	    nc++;
	  }
	}
      }
    }
    /* check if tree covers whole molecule */
    if (k!=L->nat) {
      err_set(1,"failed to generate ligand tree: selected molecule is not connected.");
      T=freeLigandTree(T);
    }
  }
  if (T) {
    /* label nodes in tree */
    for (k=0;k<L->nat;k++) {
      lb[k]=0;                           /* longest branch */
      sz[k]=L->A[T->nr[k]].z;            /* sum of atp->z in subtree */
      nh[k]=1;                           /* number of atoms in subtree */
      up[T->nr[k]]=T->up[k];
    }
    for (k=L->nat-1;k>0;k--) {
      m=T->up[k];
      n=T->nr[k];
      if (lb[n]+1>lb[m]) lb[m]=lb[n]+1;
      sz[m]+=sz[n];
      nh[m]+=nh[n];
    }
    /* reorder branches in tree (shortest branch first...) */
    for (p=0;p<L->nat;p++) {
      m=T->nr[p];
      for (k=0,i=0;i<L->A[m].nn;i++) {
	n=L->A[m].nb[i];
	if (up[n]!=m) continue;
	for (j=0;j<k;j++) {
	  q=nb[j];
	  if (lb[n]<lb[q]) break;
	  if (lb[n]>lb[q]) continue;
	  if (sz[n]<sz[q]) break;
	  if (sz[n]>sz[q]) continue;
	  if (nh[n]<nh[q]) break;
	}
	for (l=k;l>j;l--) nb[l]=nb[l-1];
	nb[j]=n;
	k++;
      }
      for (l=p+1,i=0;i<k;i++) {
	n=nb[i];
	T->nr[l]=n;
	T->up[l]=m;
	T->fw[l]=-1;
	l+=nh[n];
      }
      if (k>0) T->fw[p]=nb[k-1];
      else T->fw[p]=-1; 
    }
    for (k=L->nat;k<L->nat+nc;k++) T->fw[k]=-1;   /* ring closure bonds */
    for (i=0;i<L->nat;i++) T->new[T->nr[i]]=i;
    if (sz) free(sz);
    if (nh) free(nh);
    if (lb) free(lb);
    if (up) free(up);
    
  }else{
    err_set(1,"I'm lost in the forest, could not find the REFMAC tree.");
  }
  return(T);
}

static glLigandTree_t *freeLigandTree(glLigandTree_t *T)
{
  if (T) {
    if (T->nr) free(T->nr);
    if (T->fw) free(T->fw);
    if (T->up) free(T->up);
    if (T->new) free(T->new);
    free(T);
  }
  return((glLigandTree_t *)NULL);
}

static glLigandTree_t *newLigandTree(int size)
{
  glLigandTree_t *T;

  T=malloc(sizeof(glLigandTree_t));
  if (T) {
    T->size=size;
    T->nr=malloc(size*sizeof(int));
    T->fw=malloc(size*sizeof(int));
    T->up=malloc(size*sizeof(int));
    T->new=malloc(size*sizeof(int));
    if (!T->nr||!T->fw||!T->up||!T->new) {
      T=freeLigandTree(T);
    }
  }
  return(T);
}
