#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gseal.h>

float ufrand(float a,float b);

float SimPlexMin(
   int     n,                             /* number of variables                     */
   float  *x,                             /* in: initial values, out: best values    */
   float  *dx,                            /* dx for initial simplex                  */
   float (*F)(int n,float *y,void *data), /* target function                         */
   void   *data,                          /* pointer to additional data needed for F */
   float  tol,                            /* tolerance (dF)                          */
   int    mxit,                           /* maximum number of iterations            */
   int   (*report)(int it,int n,float **p,float *v,int hi,int lo), /* NULL or user report function */
   int   *status                          /* exit status                             */
		 );

static float cdistsq(float *x,float *y)
{
  return ((x[0]-y[0])*(x[0]-y[0])+(x[1]-y[1])*(x[1]-y[1])+(x[2]-y[2])*(x[2]-y[2]));
}


void gsealEuler(float R[3][3],float *E)
{
  /* Compute rotation matrix R from three Euler angles in E */

  int i;
  float c[3],s[3];

  for (i=0;i<3;i++) {
    c[i]=cos(E[i]);
    s[i]=sin(E[i]);
  }
  R[0][0]=c[0]*c[2]-s[0]*c[1]*s[2];
  R[0][1]=-c[0]*s[2]-s[0]*c[1]*c[2];
  R[0][2]=-s[0]*s[1];
  R[1][0]=s[0]*c[2]+c[0]*c[1]*s[2];
  R[1][1]=-s[0]*s[2]+c[0]*c[1]*c[2];
  R[1][2]=c[0]*s[1];
  R[2][0]=-s[1]*s[2];
  R[2][1]=-s[1]*c[2];
  R[2][2]=c[1];
}

static void transform(float *y,float *x,float R[3][3],float *T)
{
  /* return y = R * x + T */
  
  int i,j;
  
  for (i=0;i<3;i++) {
    y[i]=T[i];
    for (j=0;j<3;j++) y[i]+=R[i][j]*x[j];
  }
}

float gsealValue(int n1,float *X1,int *T1,int n2,float *X2,int *T2,float alpha)
{
  int i,j;
  float r=0.0;

  if (T1 && T2) {
    for (i=0;i<n1;i++) {
      for (j=0;j<n2;j++) {
	if (T1[i]==T2[j]) {
	  r += exp(-alpha*cdistsq(X1+i*3,X2+j*3));
	}
      }
    }
  }else{
    for (i=0;i<n1;i++) {
      for (j=0;j<n2;j++) {
	r += exp(-alpha*cdistsq(X1+i*3,X2+j*3));
      }
    }
  }
  return(r);
}

static void RotTra(float *y,int n,float *x,float *T,float *E)
{
  /* Return transformed position in y = R(E) * x + T
     where R is the 3x3 rotation matrix derived from three
     Euler angles in E, and T is a translation vector. n is
     the number of points in x.
  */

  float R[3][3];
  int i;

  gsealEuler(R,E);

  for (i=0;i<n;i++) transform(y+3*i,x+3*i,R,T);
    
}

static float Target(int n,float *p,void *data)
{
  gseal_par_t *D;
  float r=0.0,*y;

  /* gseal target function for simplexmin... */

  D=(gseal_par_t *)data;
  y=malloc(D->n2*3*sizeof(float));
  RotTra(y,D->n2,D->X2,p,p+3);
  r=-gsealValue(D->n1,D->X1,D->T1,D->n2,y,D->T2,D->alpha);
  free(y);
  return(r);

}

static float bestFit(float t[3],float e[3],gseal_par_t *P)
{
  float p[6],dp[6],r=0.0;
  int i;
  int status;

  for (i=0;i<3;i++) {
    p[i]=t[i]; dp[i]=1.5;
    p[i+3]=e[i]; dp[i+3]=15.0*3.141592653/180.0;
  }

  r=SimPlexMin(6,p,dp,Target,P,0.001,10000,NULL,&status);

  for (i=0;i<3;i++) {
    t[i]=p[i];
    e[i]=p[i+3];
  }
  return(-r);
}

float gsealBestFit(int niter,float *t,float *dt,float *e,float *de,gseal_par_t *P)
{
  float tb[3],eb[3],tn[3],en[3],targb,targn;
  int iter,i;
  
  for (i=0;i<3;i++) {
    tn[i]=t[i];
    tb[i]=t[i];
    en[i]=e[i];
    eb[i]=e[i];
  }
  
  targb=-1.e32;

  for (iter=0;iter<niter;iter++) {
    targn=bestFit(tn,en,P);
    if (targn>=targb) {
      for (i=0;i<3;i++) {
	tb[i]=tn[i];
	eb[i]=en[i];
      }
      targb=targn;
    }
    for (i=0;i<3;i++) {
      tn[i]=t[i]+ufrand(-dt[i],dt[i]);
      en[i]=e[i]+ufrand(-de[i],de[i]);
    }
  }

  for (i=0;i<3;i++) {
    t[i]=tb[i];
    e[i]=eb[i];
  }
  
  return(targb);
}

