#ifndef INSANEDEF
#define INSANEDEF

#include "object.h"
#include "felix.h"
#include "list.h"
#include "parser.h"

#define HHD_RESET_ALL          1
#define HHD_RESET_SPECTRA      2
#define HHD_RESET_STRUCTS      3
#define HHD_RESET_ATOMS        4
#define HHD_RESET_PATOMS       5
#define HHD_RESET_HHDIST       6

#define N_HHD_CLASSES          6
#define HHD_CLASS_OTHER       -1  
#define HHD_CLASS_13           0
#define HHD_CLASS_14           1
#define HHD_CLASS_INTRARES     2  /* within residue */
#define HHD_CLASS_SEQUENTIAL   3  /* i -- i+1 */
#define HHD_CLASS_MEDIUM       4  /* i -- i+2 or i -- i+3 */
#define HHD_CLASS_LONG         5  /* i -- i+k, k>= 4 */

#define HHD_CONTMODE_NONE      0  /* definitions for contribution filtering */
#define HHD_CONTMODE_MINIMUM   1
#define HHD_CONTMODE_AVERAGE   2
#define HHD_CONTMODE_MAXIMUM   3

typedef struct PAtom {      /* Pseudoatom: group of H with same chem. shifts */
  char *name;           
  int n;                    /* number of atoms in group */
  float ppm[2];             /* shift of H and its precursor atom */
  int z;                    /* atomic number of precursor atom (0: undef.) */
  int *nr;                  /* pointer to list of atom number of atoms in group */
} PAtom;

typedef struct {
  float intv;               /* intensity test value in multiples of noise      */
  float maxD[3];            /* upper bound for min/max/average  distance        */
  float pntres;             /* required shift accuracy in points */
  float fcont;              /* fraction of total NOE intensity for contribution filtering */
  int class[N_HHD_CLASSES ];/* exclude 1/3 , 1/4 intrares, short medium long  ? */
  int intt;                 /* intensity test: 1: greater, 0: ignore, -1: less */
  int contmode;             /* contribution filtering (0: none, 1: minimum R, 2: average R */
} DistFilter;

typedef struct {
  int n;                   /* number of distances within cutoff */
  PAtom *h1,*h2;           /* two atoms involved in distance */
  float min,max;           /* min/max distance */
  float mean,stdd;         /* average and standard deviation */
  int class;               /* distance class: 1-3, 1-4, other */
  int valnr;               /* -1: bw/fw invalid, >-1: nr. of spectrum with bw/fw */
  int sel;                 /* selected after filtering ? */
  float fw,bw;             /* forward/backward intensities/noise  */
} HHDist;		   
			   
typedef struct {	   
  Mol **M;                 /* pointers to structures */
  FELIXMATRIXP *S;         /* pointers to spectra */
  Atom ***A;               /* pointers to hydrogens A[i][j]: j-th H in i-th structure */
  int   *p;                /* j-th hydrogen is part of pseudoatom p[j] */
  PAtom **P;               /* pointers to pseudo atom definitions */
  HHDist **D;              /* distances between pseudo atoms */
  int nm;                  /* number of structures */
  int ns;                  /* number of spectra */
  int na;                  /* number of **A pointers in ***A, ususally na==nm  */
  int nh;                  /* number of hydrogens per **A */
  int np;                  /* number of pseudo atoms */
  int nd;                  /* number of distances */
  int stv,spv;             /* structure info valid ? / spectral info valid ? */
} HHDistInfo;

#define MX_RESDEF_ATMS 25
#define MX_RESDEF_GROUPS 26

typedef struct InsaneResDef {
  char *res,*res3;
  int  nr;
  char *atm[MX_RESDEF_ATMS];
  int natm[MX_RESDEF_GROUPS];
  float ppm[MX_RESDEF_ATMS];
  int sa[MX_RESDEF_ATMS];
} InsaneResDef;

/*========================================================================*/

int             insane_sanity_check(Mol *root,int verbose);
HHDistInfo     *insane_findHHDistInfo(Mol *root);
PAtom          *insane_findPseudoAtom(Atom *a,Mol *root);
void            insane_formatHHDist(FELIXMATRIXP mat,char *line,HHDist *D,int nm);
int             insane_DistanceClass(HHDistInfo *DI,PAtom *f,PAtom *t);
HHDist          insane_DStatsI(PAtom *f,PAtom *t,Mol *root);
HHDist          insane_DStats(HHDistInfo *d,PAtom *f,PAtom *t);
void            insane_setFilter(Mol *root,char *mode);
void            insane_shifts(Mol *root);
void            insane_spectrum(Mol *root);
FELIXMATRIXP    insane_open_felixmatrix(Mol *root,char *file);
void            insane_close_felixmatrix(Mol *root,FELIXMATRIXP mat);
void            insane_spectrum_info(Mol *root,FELIXMATRIXP mat);
void            insane_show_header(FELIXMATRIXP mat);
void            insane_noise_level(Mol *root,Data *data);
void            insane_access_method(Mol *root,Data *data);
void            insane_set_reference(Mol *root,FELIXMATRIXP mat);
void            insane_peek(Mol *root);
void            insane_check(Mol *root);
void            insane_check_spectrum(int nspect,FELIXMATRIXP *spectra,
				      int nstruct,Mol **structs,
				      DistFilter filter);
Token          *Get_InsaneData(char *prmt,Mol *root,int dtype);
Data           *Get_InsaneSpectra(char *prmt,Mol *root,FELIXMATRIXP mat);
Token          *Get_InsaneDatum(char *prmt,Mol *root,int dtype);
Data           *Get_InsaneSpectrum(char *prmt,Mol *root,FELIXMATRIXP mat);
PAtom          *insane_makePAtom(char *name,int n);
void            insane_delPAtom(void *ptr);
float           insane_PADist(Atom **a,PAtom *f,PAtom *t);
void            insane_PAlookupIntensity(FELIXMATRIXP mat,PAtom *f,PAtom *t,float *fw,float *bw);
int             insane_PAlookupIntensities(HHDistInfo *DI,PAtom *f,PAtom *t,float *fw,float *bw);
void            insane_DIlookupIntensity(HHDistInfo *DI,int i);
void            insane_listHHDistances(Mol *root);
void            insane_showHHDistances(Mol *root);
void            insane_showHHDistances_add(Mol *root);
void            insane_showHHDistances_status(Mol *root,char *op);
void            insane_showHHDistances_color(Mol *root);
void            insane_showHHDistances_delete(Mol *droot);
void            insane_showHHDistances_width(Mol *root);
void            insane_getSpectra(Mol *root);
void            insane_getStructures(Mol *root);
void            insane_collectHydrogens(Mol *root);
void            insane_collectPseudoAtoms(Mol *root,int flg);
void            insane_computeHHDistances(Mol *root,float cutoff);
HHDistInfo      insane_EmptyDistInfo(void);
void            insane_ResetDistInfo(HHDistInfo *d,int what);
void            insane_filterHHDistances(Mol *root,DistFilter F);

InsaneResDef    parsedResDef(char *rec,int *status);
int             storeResDef(Mol *root,InsaneResDef *Res);

#endif
