#include <stdlib.h>
#include <math.h>
#include "minieispack.h"

void EPtred2(int n,float **a,float *d,float *e,float **z)
{
  int i,j,k,l;
  float f,g,h,hh,scale;

  /*
    Function translated manually from EISPACK tred2.f

    NOTE: all vectors are 1-based, a-la NR...

     this subroutine is a translation of the algol procedure tred2,
     num. math. 11, 181-195(1968) by martin, reinsch, and wilkinson.
     handbook for auto. comp., vol.ii-linear algebra, 212-226(1971).

     this subroutine reduces a real symmetric matrix to a
     symmetric tridiagonal matrix using and accumulating
     orthogonal similarity transformations.

     on input

        n is the order of the matrix.

        a contains the real symmetric input matrix.  only the
          lower triangle of the matrix need be supplied.
	  a is in the matrix for generated by EPmatrix.

     on output

        d contains the diagonal elements of the tridiagonal matrix.

        e contains the subdiagonal elements of the tridiagonal
          matrix in its last n-1 positions.  e(1) is set to zero.

        z contains the orthogonal transformation matrix
          produced in the reduction.
	  z is in the matrix for generated by EPmatrix.

        a and z may coincide.  if distinct, a is unaltered.

     questions and comments should be directed to burton s. garbow,
     mathematics and computer science div, argonne national laboratory

     this version dated august 1983.

     ------------------------------------------------------------------
  */  

  for (i=1;i<=n;i++) {
    for (j=i;j<=n;j++) z[j][i]=a[j][i];
    d[i]=a[n][i];
  }
  for (i=n;i>=2;i--) {
    l=i-1;
    h=0.0;
    scale=0;
    for (k=1;k<=l;k++) scale+=fabs(d[k]);
    if (l<2||scale==0.0) {
      e[i]=d[l];
      for (j=1;j<=l;j++) {
	d[j]=z[l][j];
	z[i][j]=0.0;
	z[j][i]=0.0;
      }
    }else{
      for (k=1;k<=l;k++) {
	d[k]/=scale;
	h+=d[k]*d[k];
      }
      f=d[l];
      g=(f>=0 ? -sqrt(h) : sqrt(h));
      e[i]=scale*g;
      h-=f*g;
      d[l]=f-g;
      for (j=1;j<=l;j++) e[j]=0.0;
      for (j=1;j<=l;j++) {
	f=d[j];
	z[j][i]=f;
	g=e[j]+z[j][j]*f;
	for (k=j+1;k<=l;k++) {
	  g=g+z[k][j]*d[k];
	  e[k]=e[k]+z[k][j]*f;
	}
	e[j]=g;
      }
      f=0.0;
      for (j=1;j<=l;j++) {
	e[j]/=h;
	f+=e[j]*d[j];
      }
      hh=f/(h+h);
      for (j=1;j<=l;j++) e[j]-=hh*d[j];
      for (j=1;j<=l;j++) {
	for (k=j;k<=l;k++) z[k][j]-=(d[j]*e[k]+e[j]*d[k]);
	d[j]=z[l][j];
	z[i][j]=0.0;
      }
    }
    d[i]=h;
  }
  for (i=2;i<=n;i++) {
    l=i-1;
    z[n][l]=z[l][l];
    z[l][l]=1.0;
    h=d[i];
    if (h!=0.0) {
      for (k=1;k<=l;k++) d[k]=z[k][i]/h;
      for (j=1;j<=l;j++) {
	g=0.0;
	for (k=1;k<=l;k++) g+=z[k][i]*z[k][j];
	for (k=1;k<=l;k++) z[k][j]-=g*d[k];
      }
    }
    for (k=1;k<=l;k++) z[k][i]=0.0;
  }
  for (i=1;i<=n;i++) {
    d[i]=z[n][i];
    z[n][i]=0.0;
  }
  z[n][n]=1.0;
  e[1]=0.0;
}
/*====================================================================*/

int EPtql2(int n,float *d,float *e,float **z)
{
  int i,j,k,l,m,ii,l1,l2,mml;
  float c,c2,c3=0,dl1,el1,f,g,h,p,r,s,s2=0,tst1,tst2;

  /*
    Function translated manually from EISPACK tql2.f

    NOTE: all vectors are 1-based, a-la NR...


     this subroutine is a translation of the algol procedure tql2,
     num. math. 11, 293-306(1968) by bowdler, martin, reinsch, and
     wilkinson.
     handbook for auto. comp., vol.ii-linear algebra, 227-240(1971).

     this subroutine finds the eigenvalues and eigenvectors
     of a symmetric tridiagonal matrix by the ql method.
     the eigenvectors of a full symmetric matrix can also
     be found if  tred2  has been used to reduce this
     full matrix to tridiagonal form.

     on input

        nm must be set to the row dimension of two-dimensional
          array parameters as declared in the calling program
          dimension statement.

        n is the order of the matrix.

        d contains the diagonal elements of the input matrix.

        e contains the subdiagonal elements of the input matrix
          in its last n-1 positions.  e(1) is arbitrary.

        z contains the transformation matrix produced in the
          reduction by  tred2, if performed.  if the eigenvectors
          of the tridiagonal matrix are desired, z must contain
          the identity matrix.

      on output

        d contains the eigenvalues in descending order.  if an
          error exit is made, the eigenvalues are correct but
          unordered for indices 1,2,...,ierr-1.

        e has been destroyed.

        z contains orthonormal eigenvectors of the symmetric
          tridiagonal (or full) matrix.  if an error exit is made,
          z contains the eigenvectors associated with the stored
          eigenvalues.

        ierr is set to
          zero       for normal return,
          j          if the j-th eigenvalue has not been
                     determined after 30 iterations.

     calls EPpythag for  sqrt(a*a + b*b) .

     questions and comments should be directed to burton s. garbow,
     mathematics and computer science div, argonne national laboratory

     this version dated august 1983.

     ------------------------------------------------------------------
  */

  for (i=2;i<=n;i++) e[i-1]=e[i];
  f=0.0; tst1=0.0; e[n]=0.0;
  for (l=1;l<=n;l++) {
    j=0;
    h=fabs(d[l])+fabs(e[l]);
    if (tst1<h) tst1=h;
    for (m=l;m<=n;m++) {
      tst2=tst1+fabs(e[m]);
      if (tst2==tst1) break;
    }
    if (m!=l) {
      do {
	if (j==30) return(l);
	j++;
	l1=l+1; l2=l1+1;
	g=d[l];
	p=(d[l1]-g)/(2.0*e[l]);
	r=EPpythag(p,1.0);
	d[l]=e[l]/(p+(p<0?-r:r));
	d[l1]=e[l]*(p+(p<0?-r:r));
	dl1=d[l1];
	h=g-d[l];
	for (i=l2;i<=n;i++) d[i]=d[i]-h;
	f+=h;
	p=d[m];
	c=1.0; c2=c;
	el1=e[l1];
	s=0.0;
	mml=m-l;
	for (ii=1;ii<=mml;ii++) {
	  c3=c2;
	  c2=c;
	  s2=s;
	  i=m-ii;
	  g=c*e[i];
	  h=c*p;
	  r=EPpythag(p,e[i]);
	  e[i+1]=s*r;
	  s=e[i]/r;
	  c=p/r;
	  p=c*d[i]-s*g;
	  d[i+1]=h+s*(c*g+s*d[i]);
	  for (k=1;k<=n;k++) {
	    h=z[k][i+1];
	    z[k][i+1]=s*z[k][i]+c*h;
	    z[k][i]=c*z[k][i]-s*h;
	  }
	}
	p=-s*s2*c3*el1*e[l]/dl1;
	e[l]=s*p;
	d[l]=c*p;
	tst2=tst1+fabs(e[l]);
      } while (tst2>tst1);
    }
    d[l]=d[l]+f;
  }
  for (i=1;i<n;i++) {
    k=i;
    p=d[i];
    for (j=i+1;j<=n;j++) {
      if (d[j]>=p) {
	p=d[j];
	k=j;
      }
    }
    if (k>i) {
      d[k]=d[i];
      d[i]=p;
      for (j=1;j<=n;j++) {
	p=z[j][k];
	z[j][k]=z[j][i];
	z[j][i]=p;
      }
    }
  }
  return(0);
}
  
/*====================================================================*/

int EPimtql2(int n,float *d,float *e,float **z)
{
  int i,j,k,l,m,ii,mml;
  float b,c,f,g,p,r,s,tst1,tst2;

  /*
    Function translated manually from EISPACK imtql2.f

    NOTE: all vectors are 1-based, a-la NR...

     this subroutine is a translation of the algol procedure imtql2,
     num. math. 12, 377-383(1968) by martin and wilkinson,
     as modified in num. math. 15, 450(1970) by dubrulle.
     handbook for auto. comp., vol.ii-linear algebra, 241-248(1971).

     this subroutine finds the eigenvalues and eigenvectors
     of a symmetric tridiagonal matrix by the implicit ql method.
     the eigenvectors of a full symmetric matrix can also
     be found if  tred2  has been used to reduce this
     full matrix to tridiagonal form.

     on input

        nm must be set to the row dimension of two-dimensional
          array parameters as declared in the calling program
          dimension statement.

        n is the order of the matrix.

        d contains the diagonal elements of the input matrix.

        e contains the subdiagonal elements of the input matrix
          in its last n-1 positions.  e(1) is arbitrary.

        z contains the transformation matrix produced in the
          reduction by  tred2, if performed.  if the eigenvectors
          of the tridiagonal matrix are desired, z must contain
          the identity matrix.
	  z is in EPfmatrix format.

      on output

        d contains the eigenvalues in descending order.  if an
          error exit is made, the eigenvalues are correct but
          unordered for indices 1,2,...,ierr-1.

        e has been destroyed.

        z contains orthonormal eigenvectors of the symmetric
          tridiagonal (or full) matrix.  if an error exit is made,
          z contains the eigenvectors associated with the stored
          eigenvalues.

        ierr is set to
          zero       for normal return,
          j          if the j-th eigenvalue has not been
                     determined after 30 iterations.

     calls pythag for  sqrt(a*a + b*b) .

     questions and comments should be directed to burton s. garbow,
     mathematics and computer science div, argonne national laboratory

     this version dated august 1983.

     ------------------------------------------------------------------
  */
  
  for (i=2;i<=n;i++) e[i-1]=e[i];
  e[n]=0.0;
  for (l=1;l<=n;l++) {
    j=0;
    do {
      for (m=l;m<=n;m++) {
	if (m==n) break;
	tst1=fabs(d[m])+fabs(d[m+1]);
	tst2=tst1+fabs(e[m]);
	if (tst2==tst1) break;
      }
      p=d[l];
      if (m!=l) {
	if (j==30) return(l);
	j++;
	g=(d[l+1]-p)/(2.0*e[l]);
	r=EPpythag(g,1.0);
	g=d[m]-p+e[l]/(g+(g<0 ? -fabs(r) : fabs(r)));
	s=1.0; c=1.0; p=0.0;
	mml=m-l;
	for (ii=1;ii<=mml;ii++) {
	  i=m-ii;
	  f=s*e[i];
	  b=c*e[i];
	  r=EPpythag(f,g);
	  e[i+1]=r;
	  if (r==0.0) break;
	  s=f/r;
	  c=g/r;
	  g=d[i+1]-p;
	  r=(d[i]-g)*s+2.0*c*b;
	  p=s*r;
	  d[i+1]=g+p;
	  g=c*r-b;
	  for (k=1;k<=n;k++) {
	    f=z[k][i+1];
	    z[k][i+1]=s*z[k][i]+c*f;
	    z[k][i]=c*z[k][i]-s*f;
	  }
	}
	if (ii>mml) {
	  d[l]=d[l]-p;
	  e[l]=g;
	  e[m]=0.0;
	}else{
	  d[i+1]=d[i+1]-p;
	  e[m]=0.0;
	}
      }
    } while (m != l);
  }

  for (i=1;i<n;i++) {
    k=i;
    p=d[i];
    for (j=i+1;j<=n;j++) {
      if (d[j]>=p) {
	p=d[j];
	k=j;
      }
    }
    if (k>i) {
      d[k]=d[i];
      d[i]=p;
      for (j=1;j<=n;j++) {
	p=z[j][k];
	z[j][k]=z[j][i];
	z[j][i]=p;
      }
    }
  }
  return(0);
}
/*====================================================================*/

void EPevsort(int n,float *d,float **z,float dir)
{
  /*     
     Sort eigenvectors (z) and eigenvalues (d) by eigenvalue. Sort is
     ascending for dir.eq.1, descending for dir.eq.-1.
  */
     
  int i,j,k;
  float p;

  for (i=1;i<=n-1;i++) {
    k=i;
    p=d[i];
    for (j=i+1;j<=n;j++) {
      if (dir*d[j]<dir*p) {
	k=j;
	p=d[j];
      }
    }
    if (k!=i) {
      d[k]=d[i];
      d[i]=p;
      for (j=1;j<=n;j++) {
	p=z[j][i];
	z[j][i]=z[j][k];
	z[j][k]=p;
      }
    }
  }
}
/*====================================================================*/

float EPpythag(float a,float b)
{
  /* finds sqrt(a*a+b*b) without overflow or destructive underflow */

  float p,r,s,t,u;

  p=(fabs(a) > fabs(b) ? fabs(a) : fabs(b));
  if (p>0) {
    r=(fabs(a)<fabs(b)?fabs(a):fabs(b))/p;
    r=r*r;
    for (t=4.0+r;t!=4.0;t=4.0+r) {
      s=r/t;
      u=1.0+2.0*s;
      p=u*p;
      r=(s/u)*(s/u)*r;
    }
  }
  return(p);
}

/*====================================================================*/

float **EPfmatrix(float *a,int m,int n)
{
  float **r;
  int i;

  r=(float **)malloc(m*sizeof(float *));
  if (r) {
    r[0]=a-1;
    for (i=1;i<m;i++) r[i]=r[i-1]+n;
    r--;
  }
  return(r);
}

/*====================================================================*/

float **freeEPfmatrix(float **D)
{
  if (D) free(++D);
  return((float **)NULL);
}
