#include <math.h>
#include "object.h"
#include "mystdlib.h"
#include "strutil.h"
#include "gcalc.h"
#include "errhand.h"
#include "xtal.h"
#include "swdef.h"
#include "wnpconst.h"

#define BCHUNK  4      /* granularity for bond allocation */
#define ADCHUNK 3      /* granularity for adfix allocation */

#define FOOBAR (((Atom *)NULL)->prv=(Atom *)NULL)

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

/*========================================================================*/

static int defaultAtomColor = WNP_COLOR_BYATOM;      /* default color used by make_atom */
static int defaultBondWidth = 1;

static char *obj_unknown = "unk";
static char *obj_types[OBJ_N_TYPES] = {
  "root","mol","atom","lib","type","symb","link","binc","set","xtal",
  "xatom","dsp","prule","noe","cnf","frule","flxmat"
};

/* subtype names for OBJ_DATA */
static char *data_types[] = {"none","flxmat","insane"};
 
/*-------------------- KLUDGE (actually not so bad...) --------------------*/

static char *BT_MNEM[MX_OBJ_BTYPES] = 
{ "du", "1", "2", "3", "ar", "am", "nc", "un" };

static char *BT_MNEM_FUGU = "fu";

static char BT_ORDER[MX_OBJ_BTYPES] =
{0, 1, 2, 3, '?', 1, 0, '?' };

static char BT_AROM[MX_OBJ_BTYPES] = 
{0, 0, 0, 0, 1, 0, 0, 0};

static int BT_COUNT[MX_OBJ_BTYPES] =
{0, 1, 1, 1, 1, 1, 0, 1};

static float BT_EL[MX_OBJ_BTYPES] =
{0.0, 2.0, 4.0, 6.0, 3.0, 2.0, 0.0, 2.0};   /* un-bonds assumed to be single... */

int BT_number(char *mnem)
{
  int i;

  if (!mnem) return(-1);
  for (i=0;i<MX_OBJ_BTYPES;i++) if (streq(mnem,BT_MNEM[i])) return(i);
  return(-1);
}

char *BT_mnem(int num)
{
  if (-1 < num && num < MX_OBJ_BTYPES) return(BT_MNEM[num]);
  return(BT_MNEM_FUGU);
}

char BT_order(int num)
{
  if (-1 < num && num < MX_OBJ_BTYPES) return(BT_ORDER[num]);
  else return(0);
}

int BT_arom(int num)
{
  if (-1 < num && num < MX_OBJ_BTYPES) return(BT_AROM[num]);
  else return(0);
}

int BT_count(int num)
{
  if (-1 < num && num < MX_OBJ_BTYPES) return(BT_COUNT[num]);
  else return(0);
}

int BT_el(int num)
{
  if (-1 < num && num < MX_OBJ_BTYPES) return(BT_EL[num]+0.001);
  else return(0);
}

/*-------------------- END OF KLUDGE --------------------*/

/*-------------------- Internal procedures --------------------*/

static Bond **reall_bond_list(Atom *a,int size)
{
  Bond **n;
  int i;

  if (a) {
    n=a->bnd; size=(size<0?0:size);
    size=(size<a->nbnd?a->nbnd:(size>255?255:size));
    if (size!=a->mbnd) { 
      if (size>0) { 
        n=(Bond **)malloc(size*sizeof(Bond *));
        if (n) {
	  if (a->bnd) for (i=0;i<a->nbnd;i++) n[i]=a->bnd[i];
	  for (i=a->nbnd;i<size;i++) n[i]=(Bond *)NULL;
	  if (mvld(a->bnd)) free(a->bnd);
	  a->mbnd=size;
        }
      }else{
	n=(Bond **)NULL;
	a->mbnd=0;
	if (mvld(a->bnd)) free(a->bnd);
      }
    }
    a->bnd=n;
  }else{
    n=(Bond **)NULL;
  }
  return(n);
}

static ADfix **reall_adfix_list(Xatom *a,int size)
{
  ADfix **n;
  int i;

  if (a) {
    n=a->adf;
    size=(size<a->nadf?a->nadf:(size>255?255:size));
    if (size!=a->madf) { 
      if (size>0) { 
        n=(ADfix **)malloc(size*sizeof(ADfix *));
        if (n) {
	  if (a->adf) for (i=0;i<a->nadf;i++) n[i]=a->adf[i];
	  for (i=a->nadf;i<size;i++) n[i]=(ADfix *)NULL;
	  if (mvld(a->adf)) free(a->adf);
	  a->madf=size;
        }
      }else{
	n=(ADfix **)NULL;
	a->madf=0;
	if (mvld(a->adf)) free(a->adf);
      }
    }
    a->adf=n;
  }else{
    n=(ADfix **)NULL;
  }
  return(n);
}

static Mol *make_mol(void)
{
  Mol *m;
  int k;

  m=(Mol *)malloc(sizeof(Mol));
  if (m) {
    m->dsptr=(void *)NULL;
    m->moltype=MOL_UNDEF;
    m->keep=0;
    for (k=0;k<ATOMSET_WORDS;k++) {
      m->local[k]=0;
      m->used[k]=0;
    }
  }
  return(m);
}

static Mol *copy_mol(Mol *m)
{
  Mol *r;
  
  if (mvld(m)) {
    r=make_mol();
    if (mvld(r)) {
      r->moltype=m->moltype;
    }
  }else{
    r=(Mol *)NULL;
  }
  return(r);
}

static int del_mol(Mol *m)
{
  int r;

  if (m->moltype==MOL_ROOT) {
    r=ND_DISPOSE_IGNORE;
  }else{
    r=ND_DISPOSE_DELETE;
  }
  return(r);
}

static Data *make_data(void)
{
  Data *d;

  d=(Data *)malloc(sizeof(Data));
  if (d) {
    d->dtype=OBJ_DATA_NONE;
    d->data=(void *)NULL;
    d->copy=NULL;
    d->delete=NULL;
  }
  return(d);
}

static Data *copy_data(Data *d)
{
  Data *r;
  
  if (mvld(d)) {
    if (d->copy) r=d->copy(d);
    else r=make_data();           /* data not copied, this is just a dummy... */
  }else{
    r=(Data *)NULL;
  }
  return(r);
}

static int del_data(Data *d)
{
  int r;

  if (d->delete) r=d->delete(d);
  else {
    if (d->dtype!=OBJ_DATA_NONE&&mvld(d->data)) free(d->data);
    r=ND_DISPOSE_DELETE;
  }
  return(r);
}

static Dspobj *make_dspobj(void)
{
  Dspobj *d;
  int i;

  d=(Dspobj *)malloc(sizeof(Dspobj));
  if (mvld(d)) {
    d->dsptype=DSP_ROOT;
    d->cross=cPtrNULL();
    d->root=(Node *)NULL;
    d->mtrx=ident_4();
    d->rot=ident_4();
    for (i=0;i<3;i++) {
      d->pivot[i]=0;
      d->tra[i]=0;
    }
    d->status=(1<<0);    /* display on... */
  }
  return(d);
}

static Dspobj *copy_dspobj(Dspobj *o)
{
  Dspobj *d;

  if (mvld(o)) {
    d=make_dspobj();
    if (mvld(d)) {
      d->dsptype=o->dsptype;
    }
  }else{
    d=(Dspobj *)NULL;
  }
  return(d);
}

static int del_dspobj(Dspobj *d)
{
  Atom *a;

  if (is_dsp(d)) {
    if (cPtr(d->cross)) del_graph(d);
    if (is_mol(d->root)) {
      ((Mol *)d->root)->dsptr=(Dspobj *)NULL;
      ForAtoms(a,(Mol *)d->root) a->cross=cPtrNULL();
    }
  }
  return(ND_DISPOSE_DELETE);
}

static int del_atom(Atom *a)
{
  if (a->bnd) {
    while (a->nbnd) del_bond(*a->bnd);
    if (mvld(a->bnd)) free(a->bnd);
  }
  if (mvld(a->atp)) {
    a->atp->nlnks--;
    if (a->atp->nlnks==0) del_node((Node *)a->atp);
  }
  if (mvld(a->x)) {free(a->x);} a->x=(float *)NULL; a->mxyz=0; a->nxyz=0;
  if (mvld(a->label)) free(a->label);
  if (mvld(a->alt)) free(a->alt);
  return(ND_DISPOSE_DELETE);
}

static Atom *make_atom(void)
{
  Atom *a;
  int i;

  a=(Atom *)malloc(sizeof(Atom));
  if (a) {
    a->cross=cPtrNULL();
    a->xtal=0;
    a->q=0; a->shift=1.1*WNP_NO_SHIFT;
    a->chrg='?';
    a->color=5;
    Strcpy(a->grp,"??");
    a->stass=0; a->oop=0;
    a->dsp=1; a->cosw=0;
    a->flg=0;
    a->arom=0;
    a->fe=0;
    a->bnd=(Bond **)NULL; a->mbnd=0; a->nbnd=0;
    a->x=(float *)malloc(ATM_LOCSZ*sizeof(float));
    a->alt=malloc(1*sizeof(char));
    if (a->x&&a->alt) {
      a->nxyz=0;
      a->mxyz=1;
      for (i=0;i<ATM_LOCSZ;i++) a->x[i]=0.0;
      a->x[ATM_SOF]=1.0;
    }else{
      a->mxyz=0;
      a->nxyz=0;
      del_atom(a);
      a=(Atom *)NULL;
    }
    a->atp=(Atomdef *)NULL;
    a->color=(defaultAtomColor == WNP_COLOR_BYATOM ? -5 : defaultAtomColor);
    a->lig='?';
    for (i=0;i<ATOMSET_WORDS;i++) a->set[i]=0;
    a->label=NULL; a->ldsp=1; a->lcolor=WNP_COLOR_BYATOM; a->ltype=ATM_LAB_NONE;
    a->r=-1; a->sdsp=0;
  }
  return(a);
}

static Atom *copy_atom(Atom *b)
{
/*  copy atom data b-->a, do not copy bonds */

  /* Note: flags are copied (some functions rely on it) */

  Atom *a;
  int i,j;

  if (mvld(b)) {
    a=make_atom();
    if (a) {
      a->seqnr=0;
      if (mvld(a->x)) free(a->x); 
      if (mvld(a->bnd)) free(a->bnd);
      if (mvld(a->alt)) free(a->alt);
      *a=*b;
      a->x=(float *)NULL; a->nxyz=0; a->mxyz=0;
      a->alt=NULL;
      if (a->label) a->label=strdup(b->label);
      if (mvld(a->atp)) a->atp->nlnks++;
      a->cross=cPtrNULL();
      a->bnd=(Bond **)NULL; a->mbnd=0; a->nbnd=0;
      if (reall_atom_x(a,b->nxyz)) {
	for (i=0;i<a->mxyz;i++) {
	  for (j=0;j<ATM_LOCSZ;j++) a->x[ATM_LOCSZ*i+j]=b->x[ATM_LOCSZ*i+j];
	}
	memcpy(a->alt,b->alt,a->mxyz);
	a->nxyz=a->mxyz;
      }
      for (i=0;i<ATOMSET_WORDS;i++) a->set[i]=0; /* don't copy sets... */
    }
  }else{
    a=(Atom *)NULL;
  }
  return (a);
}

static Atomset *make_atomset(void)
{
  Atomset *s;

  s=(Atomset *)malloc(sizeof(Atomset));
  if (mvld(s)) {
    s->bit= -1;
  }
  return(s);
}

static Atomset *copy_atomset(Atomset *o)
{
  Atomset *s;

  if (mvld(o)) {
    s=make_atomset();
    if (mvld(s)) {
      s->bit=o->bit;
    }
  }else{
    s=(Atomset *)NULL;
  }
  return(s);
}

static int del_atomset(Atomset *s)
{
  if (mvld(s)) {
    clear_atomset(s);
    remove_atomset_bit(s);
  }
  return(ND_DISPOSE_DELETE);
}

static NOElist *make_noelist(void)
{
  NOElist *r;
  
  r=(NOElist *)malloc(sizeof(NOElist));
  if (mvld(r)) {
    r->first=(NOEdata *)NULL;
    r->last=(NOEdata *)NULL;
    r->avtype=NOE_AVTYPE_CENTER;
  }
  return(r);
}

static NOElist *copy_noelist(NOElist *o)
{
  NOElist *r=(NOElist *)NULL;
  NOEdata *noe,*new,*old;

  if (is_NOElist(o)) {
    r=make_noelist();
    if (mvld(r)) {
      old=(NOEdata *)NULL;
      for (noe=o->first;mvld(noe);noe=noe->nxt) {
	new=copy_noedata(noe);
	if (mvld(new)) {
	  if (mvld(old)) new->prv=old;
	  else new->root=o;
	  link_noedata(new);
	  old=new;
	}else break;
      }
    }
  }
  return(r);
}

static int del_noelist(NOElist *o)
{
  NOEdata *noe,*nxt;

  if (mvld(o)) {
    for (noe=o->first;mvld(noe);noe=nxt) {
      nxt=noe->nxt;
      del_noedata(noe);
    }
  }
  return(ND_DISPOSE_DELETE);
}

static Xatom *make_xatom(void)
{
  Xatom *a;
  int i;

  a=(Xatom *)malloc(sizeof(Xatom));
  if (mvld(a)) {
    a->iso=TRUE;
    a->bfac[0]=0.05;
    a->afix_root=0;
    a->block=0;
    for (i=0;i<3;i++) {
      a->esd_x[i]=0.0;
      a->esd_bfac[i]=0.0;
      a->esd_bfac[i+3]=0.0;
    }
    for (i=0;i<10;i++) a->pfix[i]=0.0;
    a->pfix[3]=10.0;    /* SOF is fixed by default */
    a->esd_sof=0.0;
    a->peak=FALSE;
    a->peak_height=0.0;
    a->madf=0;
    a->nadf=0;
    a->adf=(ADfix **)NULL;
  }
  return(a);
}

static Xatom *copy_xatom(Xatom *b)
{
  Xatom *a;

  /* do not copy ADfixes (c.f. copy_atom )... */

  if (mvld(b)) {
    a=make_xatom();
    if (mvld(a)) {
      *a = *b;
      a->madf=0;
      a->nadf=0;
      a->adf=(ADfix **)NULL;
    }
  }else{
    a=(Xatom *)NULL;
  }
  return(a);
}

static int del_xatom(Xatom *a)
{
  if (is_atom(a->up)) ((Atom *)a->up)->xtal=0;
  if (mvld(a->adf)) {
    while(a->nadf>0) del_adfix(*a->adf);
    free(a->adf);
    a->madf=0;
  }
  return(ND_DISPOSE_DELETE);
}

static Prule *make_prule(void)
{
  Prule *r;

  r=(Prule *)malloc(sizeof(Prule ));
  if (mvld(r)) {
    r->nbnd=0;
    r->nat=0;
    r->interp=PR_HARD;
    r->atom=(PrAtom *)NULL;
    r->bond=(PrBond *)NULL;
    r->def=(char *)NULL;
  }
  return(r);
}

static Prule *copy_prule(Prule *p)
{
  Prule *r;
  int i;

  r=(Prule *)NULL;
  if (!mvld(p)) return(r);
  r=make_prule();
  if (mvld(r)) {
    r->nbnd=p->nbnd;
    if (r->nbnd>0) {
      r->bond=(PrBond *)malloc(r->nbnd*sizeof(PrBond));
      if (!mvld(r->bond)) r->nbnd=0;
      for (i=0;i<r->nbnd;i++) r->bond[i]=p->bond[i];
    }
    r->nat=p->nat;
    if (r->nat>0) {
      r->atom=(PrAtom *)malloc(r->nat*sizeof(PrAtom));
      if (!mvld(r->atom)) r->nat=0;
      for (i=0;i<r->nat;i++) r->atom[i]=p->atom[i];
    }
    r->interp=p->interp;
    r->def=Strdup(p->def);
  }
  return(r);
}

static int del_prule(Prule *p)
{
  if (mvld(p->atom)) free(p->atom);
  if (mvld(p->bond)) free(p->bond);
  return(ND_DISPOSE_DELETE);
}

static Frule *make_frule(void)
{
  Frule *r;
  
  r=(Frule *)malloc(sizeof(Frule));
  if (mvld(r)) {
    r->iatp=(Atomdef **)NULL;
    r->oatp=(Atomdef **)NULL;
    r->match=(Atom **)NULL;
    r->nitp=(int *)NULL;
    r->btp=(int *)NULL;
    r->from=(int *)NULL;
    r->to=(int *)NULL;
    r->nat=0;
    r->nbnd=0;
  }
  return(r);
}

static Frule *copy_frule(Frule *p)
{
  Frule *r;
  int i,n;

  r=(Frule *)NULL;
  if (!is_frule(p)) return(r);
  r=make_frule();
  if (mvld(r)) {
    if (mvld(p->nitp)) for (n=0,i=0;i<p->nat;i++) n+=p->nitp[i];
    else n=0;
    r->nat=p->nat;
    r->nbnd=p->nbnd;
    if (mvld(p->iatp)&&n>0) {
      r->iatp=(Atomdef **)malloc(n*sizeof(Atomdef *));
      if (mvld(r->iatp)) {
	for (i=0;i<n;i++) {
	  if (is_atomdef(p->iatp[i])) {
	    r->iatp[i]=p->iatp[i];
	    r->iatp[i]->nlnks++;
	  }else r->iatp[i]=(Atomdef *)NULL;
	}
      }
    }
    if (mvld(p->oatp)&&p->nat>0) {
      r->oatp=(Atomdef **)malloc(r->nat*sizeof(Atomdef *));
      if (mvld(r->oatp)) {
	for (i=0;i<r->nat;i++) {
	  if (is_atomdef(p->oatp[i])) {
	    r->oatp[i]=p->oatp[i];
	    r->oatp[i]->nlnks++;
	  }else r->oatp[i]=(Atomdef *)NULL;
	}
      }
    }
    if (mvld(p->match)&&p->nat>0) {
      r->match=(Atom **)malloc(p->nat*sizeof(Atom *));
      if (mvld(r->match)) for (i=0;i<p->nat;i++) r->match[i]=p->match[i];
    }
    if (mvld(p->btp)&&p->nbnd>0) {
      r->btp=(int *)malloc(p->nbnd*sizeof(int));
      if (mvld(r->btp)) for (i=0;i<p->nbnd;i++) r->btp[i]=p->btp[i];
    }
    if (mvld(p->nitp)) {
      r->nitp=(int *)malloc(r->nat*sizeof(int));
      if (mvld(r->nitp)) for (i=0;i<p->nat;i++) r->nitp[i]=p->nitp[i];
    }
    if (mvld(p->from)&&p->nbnd>0) {
      r->from=(int *)malloc(p->nbnd*sizeof(int));
      if (mvld(r->from)) for (i=0;i<p->nbnd;i++) r->from[i]=p->from[i];
    }
    if (mvld(p->to)&&p->nbnd>0) {
      r->to=(int *)malloc(p->nbnd*sizeof(int));
      if (mvld(r->to)) for (i=0;i<p->nbnd;i++) r->to[i]=p->to[i];
    }
    if ((r->nat>0&&(!mvld(r->iatp)||!mvld(r->oatp)||!mvld(r->nitp)))||
	(r->nbnd>0&&(!mvld(r->btp)||!mvld(r->from)||!mvld(r->to)))) {
      if (mvld(r->iatp)) {
	for (i=0;i<n;i++) if (is_atomdef(r->iatp[i])) r->iatp[i]->nlnks--;
	free(r->iatp);
      }
      if (mvld(r->oatp)) {
	for (i=0;i<r->nat;i++) if (is_atomdef(r->oatp[i])) r->oatp[i]->nlnks--;
	free(r->oatp);
      }
      if (mvld(r->nitp)) free(r->nitp);
      if (mvld(r->btp)) free(r->btp);
      if (mvld(r->from)) free(r->from);
      if (mvld(r->to)) free(r->to);
      free(r); r=(Frule *)NULL;
    }
  }
  return(r);
}

static int del_frule(Frule *p)
{
  int i,n=0;

  if (mvld(p->match)) free(p->match);
  if (mvld(p->nitp)) {
    for (n=0,i=0;i<p->nat;i++) n+=p->nitp[i];
    free(p->nitp);
  }
  if (mvld(p->btp)) free(p->btp);
  if (mvld(p->iatp)) {
    for (i=0;i<n;i++) {
      if (is_atomdef(p->iatp[i])) {
	p->iatp[i]->nlnks--;
	if (p->iatp[i]->nlnks<=0)  del_node((Node *)p->iatp[i]);
      }
    }
    free(p->iatp);
  }
  if (mvld(p->oatp)) {
    for (i=0;i<p->nat;i++) {
      if (is_atomdef(p->oatp[i])) {
	p->oatp[i]->nlnks--;
	if (p->oatp[i]->nlnks<=0) del_node((Node *)p->oatp[i]);
      }
    }
    free(p->oatp);
  }
  if (mvld(p->from)) free(p->from);
  if (mvld(p->to)) free(p->to);
  p->nat=0;
  p->nbnd=0;
  return(ND_DISPOSE_DELETE);
}

static hisrec *copy_history(hisrec *o)
{
  hisrec h,*l,*n;

  h.nxt=(hisrec *)NULL; l = &h;
  while (mvld(o)) {
    n=make_history();
    if (mvld(n)) {
      n->t=Strdup(o->t);
      l->nxt=n;
      l=n;
    }
    o=o->nxt;
  }
  return(h.nxt);
}
  
static Xcn *make_xcn(void)
{
  int i;
  Xcn *x;

  x=(Xcn *)malloc(sizeof(Xcn));
  if (x) {
    x->code=(char *)NULL;
    x->title=(char *)NULL;
    x->scale=(char *)NULL;
    x->lambda=1.0;
    x->boxdsp=(Dspobj *)NULL;
    x->box_color=0;
    for (i=0;i<3;i++) {
      x->cl[i]=1.0;
      x->ca[i]=90.0;
      x->esd_cl[i]=0.0;
      x->esd_ca[i]=0.0;
    }
    x->fta=ident_4();
    x->atf=ident_4();
    for (i=0;i<MX_XCN_FORMUL;i++) {
      strcpy(x->forms[i],"");
      x->formn[i]=0;
    }
    x->spgr=(char *)NULL;
    x->history=(hisrec *)NULL;
    x->lattice= P_latt;
    x->centric=0;
    x->forml=0;
    for (i=0;i<6;i+=2) {
      x->asym[i]=0.0;
      x->asym[i+1]=1.0;
    }
    x->sym[0]=ident_4();     /* should always be the first symop */
    x->nsym=1;
    x->nsymx=1;
  }
  return(x);
}

static Xcn *copy_xcn(Xcn *b)
{
  Xcn *x;

  if (mvld(b)) {
    x=(Xcn *)malloc(sizeof(Xcn));
    if (mvld(x)) {
      *x = *b;
      x->code=Strdup(b->code);
      x->title=Strdup(b->title);
      x->scale=Strdup(b->title);
      x->spgr=Strdup(b->spgr);
      x->history=copy_history(b->history);
      x->box_color=0;
      x->boxdsp=(Dspobj *)NULL;
    }
  }else{
    x=(Xcn *)NULL;
  }
  return(x);
}

static int del_xcn(Xcn *x)
{
  hisrec *h,*nh;

  if (mvld(x->code)) free(x->code);
  if (mvld(x->title)) free(x->title);
  if (mvld(x->spgr)) free(x->spgr);
  if (mvld(x->scale)) free(x->scale);
  h=x->history;
  for (h=x->history;h;h=nh) {
    nh=mvld(h->nxt);
    if (mvld(h->t)) free(h->t);
    free(h);
  }
  return(ND_DISPOSE_DELETE);
}
	

static Lib *make_lib(void)
{
  Lib *l;

  l=(Lib *)malloc(sizeof(Lib));
  return(l);
}

static Lib *copy_lib(Lib *o)
{
  Lib *l;

  if (mvld(o)) {
    l=make_lib();
  }else{
    l=(Lib *)NULL;
  }
  return(l);
}

static int del_lib(Lib *l)
{
  return(l?ND_DISPOSE_DELETE:ND_DISPOSE_IGNORE);
}

static Bondinc *make_bondinc(void)
{
  Bondinc *r;

  r=(Bondinc *)malloc(sizeof(Bondinc));
  if (mvld(r)) {
    r->dij=0.0;
    r->dji=0.0;
  }
  return(r);
}

static Bondinc *copy_bondinc(Bondinc *b)
{
  Bondinc *r;

  if (mvld(b)) {
    r=make_bondinc();
  }else{
    r=(Bondinc *)NULL;
  }
  return(r);
}

static int del_bondinc(Bondinc *b)
{
  return(b?ND_DISPOSE_DELETE:ND_DISPOSE_IGNORE);
}

static Atomdef *make_atomdef(void)
{
  Atomdef *d;

  d=(Atomdef *)malloc(sizeof(Atomdef));
  if (d) {
    d->mass=1.0;
    d->color=5;
    d->rvdw=0.2;
    d->rcov=0.0;
    Strcpy(d->geom,"*");
    Strcpy(d->chs,"??");
    d->chrg='?'; d->val=-1; d->lig='?'; d->z=-1;
    d->numtp=0;
    d->nlnks=1;
    d->oop=0; d->hbacc=0; d->hbdon=0; d->arom=0;
    d->bnd_equiv=(char *)NULL;
    d->GBrad=1.50;
    d->GBscr=0.80;
  }
  return(d);
}

static Atomdef *copy_atomdef(Atomdef *o)
{
  Atomdef *d;

  if (mvld(o)) {
    d=(Atomdef *)malloc(sizeof(Atomdef));
    if (mvld(d)) {
      *d = *o;
      d->bnd_equiv=Strdup(d->bnd_equiv);
      d->nlnks=1;
    }
  }else{
    d=(Atomdef *)NULL;
  }
  return(d);
}

static int del_atomdef(Atomdef *d)
{
  if (mvld(d)) d->nlnks--;
  if (d->nlnks<=0 && mvld(d->bnd_equiv)) free(d->bnd_equiv);
  return(d->nlnks>0 ? ND_DISPOSE_UNLINK : ND_DISPOSE_DELETE);
}

static Symbol *make_symbol(void)
{
  Symbol *s;

  s=(Symbol *)malloc(sizeof(Symbol));
  if (mvld(s)) {
    s->mass=1.0;
    s->rvdw=1.0;
    s->rcov=0.001;
    s->GBrad=1.50;
    s->GBscr=0.800;
    s->color=5;
    s->val=-1;
    s->z=-1;
  }
  return(s);
}

static Symbol *copy_symbol(Symbol *o)
{
  Symbol *s;

  if (mvld(o)) {
    s=(Symbol *)malloc(sizeof(Symbol));
    if (mvld(s)) *s = *o;
  }else{
    s=(Symbol *)NULL;
  }
  return(s);
}

static int del_symbol(Symbol *s)
{
  return(s?ND_DISPOSE_DELETE:ND_DISPOSE_IGNORE);
}

static CnfNode *make_cnf(void)
{
  CnfNode *c;

  c=(CnfNode *)malloc(sizeof(CnfNode));
  if (c) {
    c->energy=99999.0;     /* missing value.... */
    c->n=0;
    c->cseqnr=0;
    c->x=(float *)NULL;
  }
  return(c);
}

static CnfNode *copy_cnf(CnfNode *o)
{
  CnfNode *c;

  if (mvld(o)) {           /* Note: this node type can not be copied... */
    c=(CnfNode *)NULL;
  }else{
    c=(CnfNode *)NULL;
  }
  return(c);
}

static int del_cnf(CnfNode *c)
{
  if (mvld(c)) {
    if (mvld(c->x)) free(c->x);
  }
  return(ND_DISPOSE_DELETE);
}

static Link *make_link(void)
{
  Link *l;
  
  l=(Link *)malloc(sizeof(Link));
  if (l) {
    l->from=(char *)NULL;
    l->to=(char *)NULL;
    l->bo='?';
    l->bt=BT_UNKNOWN;
    l->dmin=0.0;
    l->dmax=1.0E30;
  }
  return(l);
}

static Link *copy_link(Link *o)
{
  Link *l;

  if (mvld(o)) {
    l=(Link *)malloc(sizeof(Link));
    if (mvld(l)) {
      *l = *o;
      l->from=Strdup(o->from);
      l->to=Strdup(o->to);
    }
  }else{
    l=(Link *)NULL;
  }
  return(l);
}

static int del_link(Link *l)
{
  if (mvld(l)) {
    if (mvld(l->from)) free(l->from);
    if (mvld(l->to)) free(l->to);
  }
  return(ND_DISPOSE_DELETE);
}
/*-------------------- tree_addr --------------------*/
/** not used ? **/

int tree_addr(Node *root,Node *a,int *addr)
{
  /* find 'tree address' of a in tree rooted at root */
  /* dimension of addr must be large enough to hold depth(a,root) ints */

  int n,s,i;
  Node *x;

  if (root&&a) {
    n=0;
    while (a&&a!=root) {
      if (a->up) {
	for (addr[n]=0,x=a;x;x=x->prv) addr[n]++;
	n++;
      }
      a=a->up;
    }
    if (a==root) {
      for (i=0;i<n/2;i++) {
	s=addr[i];
	addr[i]=addr[n-i-1];
	addr[n-i-1]=s;
      }
    }else n=0;
  }else n=0;
  return(n);
}
/*-------------------- locate_node --------------------*/
/** not used ? **/

Node *locate_node(Node *root,int n,int *addr)
{
  /* locate node with address (n,addr) in subtree rooted at tree */

  int l,m;
  Node *x,*r;

  r=(n<1?(Node *)NULL:root);
  for (l=0;l<n&&r;l++) {
    for (m=addr[l],x=r->fst;x&&m>0;x=(m>0?x->nxt:x)) m--;
    r=(m>0?(Node *)NULL:x);
  }
  r=(l<n?(Node *)NULL:r);
  return(r);
}
/*-------------------- locate_node_name --------------------*/
/** not used ? **/

Node *locate_node_name(Node *r1,Node *r2,Node *n2)
{
  /* find node in r1 with same name as n2 in r2 */
  /* Note: node names should be unique, or else.. */

  Node *n1,*t;

  if (n2==r2) {
    n1=r1;
  }else if (n2->up) {
    t=locate_node_name(r1,r2,n2->up);
    if (t) {
      forObjs(n1,t,n2->type) if (strceq(n1->name,n2->name)) break;
    }else n1=(Node *)NULL;
  }else{
    n1=(Node *)NULL;
  }
  return(n1);
}
/*-------------------- make_node --------------------*/

Node *make_node(int type)
{
/* create a new node of given type */

  Node *n;

  switch (type) {
  case OBJ_ROOT:
    n=(Node *)malloc(sizeof(Node));
    break;
  case OBJ_MOL:
    n=(Node *)make_mol();
    break;
  case OBJ_DSP:
    n=(Node *)make_dspobj();
    break;
  case OBJ_ATOM:
    n=(Node *)make_atom();
    break;
  case OBJ_XATOM:
    n=(Node *)make_xatom();
    break;
  case OBJ_LIB:
    n=(Node *)make_lib();
    break;
  case OBJ_ATOMDEF:
    n=(Node *)make_atomdef();
    break;
  case OBJ_SYMBOL:
    n=(Node *)make_symbol();
    break;
  case OBJ_LINK:
    n=(Node *)make_link();
    break;
  case OBJ_ATOMSET:
    n=(Node *)make_atomset();
    break;
  case OBJ_XCN:
    n=(Node *)make_xcn();
    break;
  case OBJ_PRULE:
    n=(Node *)make_prule();
    break;
  case OBJ_FRULE:
    n=(Node *)make_frule();
    break;
  case OBJ_NOELIST:
    n=(Node *)make_noelist();
    break;
  case OBJ_BONDINC:
    n=(Node *)make_bondinc();
    break;
  case OBJ_CNF:
    n=(Node *)make_cnf();
    break;
  case OBJ_DATA:
    n=(Node *)make_data();
    break;
  default:
    n=(Node *)NULL;
    break;
  }
  if (n) {
    n->type=type;
    n->chain=(Node *)NULL;
    n->prv=(Node *)NULL;
    n->nxt=(Node *)NULL;
    n->up =(Node *)NULL;
    n->fst=(Node *)NULL;
    n->lst=(Node *)NULL;
    n->name=(char *)NULL;
    n->seqnr=-1;
  }
  return(n);
}
/*-------------------- copy_node --------------------*/

Node *copy_node(Node *o)
{
/* copy a node : new node has all five pointers pointing to NULL. 
   For some node types, only part of the info contained in the node
   will be copied (e.g. bonds in atoms are not copied... ).
*/

  Node *n;

  if (mvld(o)) {
    switch (o->type) {
    case OBJ_ROOT:
      n=(Node *)malloc(sizeof(Node));
      break;
    case OBJ_MOL:
      n=(Node *)copy_mol((Mol *)o);
      break;
    case OBJ_DSP:
      n=(Node *)copy_dspobj((Dspobj *)o);
      break;
    case OBJ_ATOM:
      n=(Node *)copy_atom((Atom *)o);
      break;
    case OBJ_XATOM:
      n=(Node *)copy_xatom((Xatom *)o);
      break;
    case OBJ_LIB:
      n=(Node *)copy_lib((Lib *)o);
      break;
    case OBJ_ATOMDEF:
      n=(Node *)copy_atomdef((Atomdef *)o);
      break;
    case OBJ_SYMBOL:
      n=(Node *)copy_symbol((Symbol *)o);
      break;
    case OBJ_LINK:
      n=(Node *)copy_link((Link *)o);
      break;
    case OBJ_ATOMSET:
      n=(Node *)copy_atomset((Atomset *)o);
      break;
    case OBJ_XCN:
      n=(Node *)copy_xcn((Xcn *)o);
      break;
    case OBJ_PRULE:
      n=(Node *)copy_prule((Prule *)o);
      break;
    case OBJ_FRULE:
      n=(Node *)copy_frule((Frule *)o);
      break;
    case OBJ_NOELIST:
      n=(Node *)copy_noelist((NOElist *)o);
      break;
    case OBJ_BONDINC:
      n=(Node *)copy_bondinc((Bondinc *)o);
      break;
    case OBJ_CNF:
      n=(Node *)copy_cnf((CnfNode *)o);
      break;
    case OBJ_DATA:
      n=(Node *)copy_data((Data *)o);
      break;
    default:
      n=(Node *)NULL;
      break;
    }
    if (n) {
      n->type=o->type;
      n->chain=(Node *)NULL;
      n->prv=(Node *)NULL;
      n->nxt=(Node *)NULL;
      n->up =(Node *)NULL;
      n->fst=(Node *)NULL;
      n->lst=(Node *)NULL;
      n->name=Strdup(o->name);
    }
  }else{
    n=(Node *)NULL;
  }
  return(n);
}
/*-------------------- rdel_node --------------------*/

Node *rdel_obj(Node *n,long t)
{
  /* recursive node deletion ...                          */
  /* Delete n, if after deletion of n parent does not     */
  /* have a descendant of type t, recusively rdel parent. */

  Node *up;

  if (n) {
    up=n->up;
    del_node(n);
    if (!rfirst_type(up,t)) rdel_obj(up,t);
  }
  return(Node *)NULL;
}
/*-------------------- del_node --------------------*/

Node *del_node(Node *n)
{
/* delete a node and its descendants */

  int keep;
  Node *old;

  if (mvld(n)) {
    del_assoc_dsp(n);
    while (n->fst) {
      old=n->fst;
      del_node(n->fst);
      if (n->fst==old) {
	fprintf(stderr,"data structure is fugu\n");
      }
    }
    switch (n->type) {
    case OBJ_ROOT:                      /* never delete root */
      keep=ND_DISPOSE_IGNORE;
      break;
    case OBJ_MOL:  
      keep=del_mol((Mol *)n);
      break;
    case OBJ_DSP:
      keep=del_dspobj((Dspobj *)n);
      break;
    case OBJ_ATOM:
      keep=del_atom((Atom *)n);
      break;
    case OBJ_XATOM:
      keep=del_xatom((Xatom *)n);
      break;
    case OBJ_LIB:  
      keep=del_lib((Lib *)n);
      break;
    case OBJ_ATOMDEF:
      keep=del_atomdef((Atomdef *)n);
      break;
    case OBJ_SYMBOL:
      keep=del_symbol((Symbol *)n);
      break;
    case OBJ_LINK:
      keep=del_link((Link *)n);
      break;
    case OBJ_ATOMSET:
      keep=del_atomset((Atomset *)n);
      break;
    case OBJ_XCN:
      keep=del_xcn((Xcn *)n);
      break;
    case OBJ_PRULE:
      keep=del_prule((Prule *)n);
      break;
    case OBJ_FRULE:
      keep=del_frule((Frule *)n);
      break;
    case OBJ_NOELIST:
      keep=del_noelist((NOElist *)n);
      break;
    case OBJ_BONDINC:
      keep=del_bondinc((Bondinc *)n);
      break;
    case OBJ_CNF:
      keep=del_cnf((CnfNode *)n);
      break;
    case OBJ_DATA:
      keep=del_data((Data *)n);
      break;
    default:
      keep=ND_DISPOSE_IGNORE;
      break;
    }
    if (keep==ND_DISPOSE_UNLINK) {
      unlink_node(n);
    }else if (keep==ND_DISPOSE_DELETE) {
      if (mvld(n->name)) free(n->name);
      unlink_node(n);
      free(n);
    }
  }
  return((Node*)NULL);
}
/*-------------------- unlink_node --------------------*/

Node *unlink_node(Node *n)
{
  if (mvld(n)) {                               
    if (mvld(n->prv)) n->prv->nxt=n->nxt;
    if (mvld(n->nxt)) n->nxt->prv=n->prv;
    if (mvld(n->up)) {
      if (n->up->fst==n) n->up->fst=n->nxt;
      if (n->up->lst==n) n->up->lst=n->prv;
    }
    n->up=(Node *)NULL;
    n->prv=(Node *)NULL;
    n->nxt=(Node *)NULL;
  }else{
    n=(Node *)NULL;
  }
  return(n);
}
/*-------------------- add_node --------------------*/

Node *add_node(Node *parent,char *name,int type)
{
/* add a node to object tree */

  Node *child;

  if (parent) {
    child=make_node(type);
    if (child) {
      child->up=parent;
      link_node(child);
      if (name) child->name=Strdup(name);
    }
  }else{
    child=(Node *)NULL;
  }
  return(child);
}
/*-------------------- link_node --------------------*/

Node *link_node(Node *new)
{
/* link node to tree. new must have exactly one non-null (up,prv,nxt) ptr. */
/* return pointer to self (NULL if link failed)                            */

  int nn,i;
  Node *old,*o;

  nn=0;
  if (!new) return(new);
  if (new->up)  nn++;
  if (new->prv) nn++;
  if (new->nxt) nn++;
  if (nn!=1) return((Node *)NULL);
  if (new->up) {
    old=new->up;
    new->prv=old->lst;
    if (old->lst) {
      old->lst->nxt=new;
    }else{
      old->fst=new;
    }
    old->lst=new;
  }else if (new->prv) {
    old=new->prv;
    new->up=old->up;
    new->nxt=old->nxt;
    old->nxt=new;
    if (new->nxt) {
      new->nxt->prv=new;
    }else{
      if (new->up) if (new->up->lst==new->prv) new->up->lst=new;
    }
  }else if (new->nxt) {
    old=new->nxt;
    new->up=old->up;
    new->prv=old->prv;
    old->prv=new;
    if (new->prv) {
      new->prv->nxt=new;
    }else{
      if (new->up) if (new->up->fst==new->nxt) new->up->fst=new;
    }
  }
  if (is_mol(new)&&is_mol(new->up)) {   /* renumber set bits, if necessary... */
    renumber_atomsets(new,new->up);
    for (o=new;mvld(o->up);o=o->up) {
      if (is_mol(o)&&is_mol(o->up)) {
	for (i=0;i<ATOMSET_WORDS;i++) ((Mol *)o->up)->used[i]|=((Mol *)o)->used[i]|((Mol *)o)->local[i];
      }
    }
  }
  return(new);
}    
/*-------------------- TreeCopyAtoms --------------------*/

Mol *TreeCopyAtoms(int nat,Atom **atom,int flg,int how,Node *target,Node *rtarget)
{
  unsigned long mask[ATOMSET_WORDS],used[ATOMSET_WORDS];
  Node *r=(Node *)NULL,*u,*c,*n,*source=(Node *)NULL;
  NOElist *noe;
  Xcn *xcn;
  Atomset *s;
  Atom *f,*t,*a;
  Xatom *xat;
  
  int i,j,nc,np,ns,nt,level;

  if (!atom) nat=0;
  if (nat>0) {
    level=how/OBJ_CPYLEV;
    for (i=0;i<nat;i++) atom[i]->flg|=flg;
    for (i=0;i<nat;i++) if (!mvld(copy_parent((Node *)atom[i]))) break;
    nc=i;
    for (r=atom[0]->chain;mvld(r)&&mvld(r->up);r=r->up);  /* find common root */
    if (nc==nat) {
      if (how&OBJ_CPYALLBNDS) {    /* copy bonds... */
	for (i=0;i<nat;i++) {
	  for (j=0;j<atom[i]->nbnd;j++) {
	    f=(Atom *)atom[i]->chain;
	    t=neighbour(j,atom[i]);
	    if (t->flg&flg) {
	      if (how&OBJ_CPYINTRABNDS) add_bond_copy(f,(Atom *)t->chain,atom[i]->bnd[j]);
	    }else{
	      if (how&OBJ_CPYINTERBNDS) add_bond_copy(f,t,atom[i]->bnd[j]);
	    }
	  }
	}
      }
      if (level>0) {                  /* reduce to requested level */
	for (u=atom[0]->chain;u&&level>1;level--) u=u->up;
	if (is_mol(u)) {
	  r=u; unlink_node(r);
	  for (u=u->up;u;u=u->up) {
	    if (u->chain) u->chain->chain=(Node *)NULL;
	    del_node(u);
	  }
	}
      }else{
	if (mvld(r)&&r->fst==r->lst&&is_mol(r->fst)) { /* reduce tree: copy of Molroot */
	  r=r->fst;
	  u=r->up;
	  unlink_node(r);
	  if (u->chain) u->chain->chain=(Node *)NULL;
	  del_node(u);
	}
      }
      if (how&OBJ_CPYMINTREE) {
	while (mvld(r)&&r->fst==r->lst&&is_mol(r->fst)) { /* further reduction, if requested. */
	  r=r->fst;
	  u=r->up;
	  unlink_node(r);
	  if (u->chain) u->chain->chain=(Node *)NULL;
	  del_node(u);
	}
      }
      source=r->chain;
    }else r=del_node(r);  /* if copy incomplete, delete copy */
  }
  if (r) {
    if (target) {
      if (!rtarget) rtarget=target;
      if (!mvld(target->name)) set_node_name(target,r->name);
      while ((n=r->fst)) {
	unlink_node(n);
	n->up=target;
	link_node(n);
      }
      del_node(r);
      source->chain=target;
      u=common_parent(source,target);
      if (u!=source) {
	for (ns=0,c=source;c&&c->up!=u;c=c->up) ns++;
	for (nt=0,c=target;c&&c->up!=u&&c!=rtarget;c=c->up) nt++;
	np=nt-ns;
	for (c=target,i=0;i<np;i++,c=c->up) c->chain=source;
	for (r=source;i<=nt;i++,c=c->up,r=r->up) {
	  c->chain=r;
	  r->chain=c;
	}
	r=rtarget;
      }else{            /* target is a subtree of source... */
	r=target;
	r->chain=source;
      }
    }else{
      u=(Node *)NULL;
      r->chain=source;
    }
    if ((how&OBJ_CPYSETS)) {
      for (i=0;i<ATOMSET_WORDS;i++) {
	mask[i]=0;
	used[i]=0;
      }
      if (u==source) {
	for (c=r->chain;c;c=c->up)
	  for (i=0;i<ATOMSET_WORDS;i++) mask[i]|=((Mol *)c)->local[i];  
	for (i=0;i<ATOMSET_WORDS;i++) mask[i]=~ mask[i];
      }else{
	for (c=r->chain->up;c&&c!=u;c=c->up) 
	  for (i=0;i<ATOMSET_WORDS;i++) mask[i]|=((Mol *)c)->local[i];
	for (i=0;i<nat;i++) 
	  for (j=0;j<ATOMSET_WORDS;j++) used[j]|=atom[i]->set[j];
	ForAtomsets(s,r->chain) {
	  if (!s->chain&&s->up->chain) {
	    if (s->bit>=0&&s->bit<ATOMSET_BITS) {
	      if (used[s->bit/BITS_PER_WORD]&1<<s->bit%BITS_PER_WORD) {
		c=copy_node((Node *)s);
		if (c) {
		  c->chain=(Node *)s;
		  c->up=s->up->chain;
		  link_node(c);
		  register_atomset((Atomset *)c);
		}
	      }
	    }
	  }
	}
      }
      for (i=0;i<nat;i++) {
	a=(Atom *)atom[i]->chain;
	if (a) for (j=0;j<ATOMSET_WORDS;j++) a->set[j]=atom[i]->set[j]&~mask[j];
      }
    }
    if (how&OBJ_CPYXCN) {
      ForXcns(xcn,r->chain) {
	if (!xcn->chain&&xcn->up->chain) {
	  c=copy_node((Node *)xcn);
	  if (c) {
	    c->up=xcn->up->chain;
	    c->chain=(Node *)xcn;
	    link_node(c);
	  }
	}
      }
      for (i=0;i<nat;i++) {
	if (atom[i]->chain) {
	  forXatoms(xat,atom[i]) {
	    c=copy_node((Node *)xat);
	    if (c) c->up=atom[i]->chain;
	    link_node(c);
	  }
	}
      }
    }
    if (how&OBJ_CPYNOE) {
      ForNOElists(noe,r->chain) {
	if (!noe->chain&&noe->up->chain) {
	  c=copy_node((Node *)noe);
	  if (c) {
	    c->chain=(Node *)noe->chain;
	    c->up=noe->up->chain;
	    link_node(c);
	  }
	}
      }
    }
  }

  for (i=0;i<nat;i++) {      /* reset flags and chain pointers in old tree */
    atom[i]->flg&=~flg;
    if (atom[i]->chain) ((Atom *)atom[i]->chain)->flg=atom[i]->flg;
    for (u=(Node *)atom[i];mvld(u->up)&&u->up->chain;u=u->up) u->up->chain=(Node *)NULL;
    atom[i]->chain=(Node *)NULL;
  }
  if (r) {
    ForNodes(c,r) c->chain=(Node *)NULL;
    r->chain=(Node *)NULL;
  }

  return((Mol *)r);
}

/*-------------------- copy_parent --------------------*/

Node *copy_parent(Node *n)
{
  Node *c,*p;

  if (n) {
    if (!n->chain) {
      p=copy_parent(n->up);
      c=copy_node(n);
      if (c) c->up=p;
      link_node(c);
      n->chain=c;
      c->chain=n;
    }
  }
  if (n) return(n->chain);
  else return(n);
}
/*-------------------- tree_copy --------------------*/

Node *tree_copy(Node *r)
{
/*  tree starting at root r. return pointer to root of new tree */

  Nodechain A,B,C;
  Node *o,*n;
  Atom *a,*fo,*to,*fn,*tn;
  Bond *bo;
  int i;
  unsigned long mask[ATOMSET_WORDS];
  Atomset *s;

  if (!mvld(r)) return((Node *)NULL);

  /* collect nodes in r into chain A  */

  A.n=1; A.s=r; A.e=A.s; A.p=(Node *)NULL;
  ForObjs(o,r,OBJ_ALL) {
    A.n++;
    A.e->chain=o;
    A.e=o;
  }
  A.e->chain=(Node *)NULL;         /* should be redundant, but... */

  /* copy nodes in chain A, link through chain B */

  B.s=copy_node(A.s); B.e=B.s; B.p=(Node *)NULL;
  B.n=(B.s ? 1 : 0);
  
  for (o=A.s->chain;o;o=o->chain) {
    n=copy_node(o);
    if (n) {
      if (B.e) {
	B.n++;
	B.e->chain=n;
	B.e=n;
      }else break;
    }else break;
  }
  B.e->chain=(Node *)NULL;

  /* must have same number of nodes in both chains */

  if (B.n != A.n) {
    for (n=B.s;n;n=o) {
      o=n->chain;
      del_node(n);
    }
    clear_node_chain(&A);
    return((Node *)NULL);
  }

  /* interleave chains A and B : A->B->A->B->...  */
  /* this will make A,B invalid                   */

  C.s=A.s; C.n=A.n+B.n; C.p=(Node *)NULL;
  for (o=A.s,n=B.s; o&&n; o=C.p,n=C.e) {
    C.p=o->chain;
    C.e=n->chain;
    o->chain=n;
    n->chain=C.p;
  }
  C.e=B.e;
  B.s=A.s=B.e=A.e=(Node *)NULL;

  /* A->chain points to copy in B, allowing fast relink of nodes */
  /* Link only fst and lst of first node (root).                 */

  o=C.s; n=o->chain;
  n->fst=(mvld(o->fst) ? o->fst->chain : (Node *)NULL);
  n->lst=(mvld(o->lst) ? o->lst->chain : (Node *)NULL);
  
  while ((o=n->chain)) {
    n=o->chain;
    n->up =(mvld(o->up)  ? o->up->chain  : (Node *)NULL);
    n->prv=(mvld(o->prv) ? o->prv->chain : (Node *)NULL);
    n->nxt=(mvld(o->nxt) ? o->nxt->chain : (Node *)NULL);
    n->fst=(mvld(o->fst) ? o->fst->chain : (Node *)NULL);
    n->lst=(mvld(o->lst) ? o->lst->chain : (Node *)NULL);
  }

  /* now add bonds: use interleaved chain C to map old to new atom-ptrs */
  /* should we do afix/dfix as well ??? */

  ForAtoms(a,r) {
    for (i=0;i<a->nbnd;i++) {
      bo=a->bnd[i];
      fo=bo->f; to=bo->t;
      if (a==fo || !fo->chain) {      /* copy all inter bonds, intra only if a is "from" */
	fn=(fo->chain ? (Atom *)fo->chain : fo);
	tn=(to->chain ? (Atom *)to->chain : to);
	add_bond_copy(fn,tn,bo);
      }
    }
  }

  /* sets have been copied, but not registered, so register them now... */

  n=C.s->chain;
  for (i=0;i<ATOMSET_WORDS;i++) mask[i]=0;
  ForAtomsets(s,n) {
    if (s->bit>=0 && s->bit<ATOMSET_BITS) {
      register_atomset(s);
      mask[s->bit/BITS_PER_WORD]|=1<<s->bit%BITS_PER_WORD;
    }
  }

  /* atomset membership has not been copied, time to do it now, but need
     to mask set bits not defined in tree(n)... */

  o=C.s; n=o->chain;
  while(mvld(n)) {
    if (is_atom(n)&&is_atom(o)) {       /* redundant test, but... */
      for (i=0;i<ATOMSET_WORDS;i++) ((Atom *)n)->set[i]=((Atom *)o)->set[i]&mask[i];
    }
    o=n->chain;
    n=(mvld(o) ? o->chain : o);
  }

  /* done...  */

  n=C.s->chain;            /* second node is root of copy */
  clear_node_chain(&C);
  return(n);
}

/*-------------------- set_node_name --------------------*/

char *set_node_name(Node *node,char *name)
{
  char *r;

  r=(char *)NULL;
  if (mvld(node)) {
    r=node->name;
    if (name) node->name=Strdup(name);
    else node->name=name;
    if (mvld(r)) free(r);
    r=node->name;
  }
  return(r);
}
/*-------------------- set_atom_type --------------------*/

void set_atom_type(Atom *a,Atomdef *type)
{
  if (mvld(a)) {
    if (mvld(a->atp)) {
      if (type!=a->atp) {
	a->atp->nlnks--;
	if (a->atp->nlnks==0) del_node((Node *)a->atp);
      }
    }
    if (type!=a->atp) {
      a->atp=type;
      if (mvld(type)) {
	a->val=type->val;
	a->oop=type->oop;
	if (a->color<0) a->color=(type->color > 0 ? -type->color : type->color);
	if (a->r<-0.5) a->r=type->rvdw;
	type->nlnks++;
      }
    }
  }
}
/*-------------------- make_history --------------------*/

hisrec *make_history(void)
{
  hisrec *h;

  h=(hisrec *)malloc(sizeof(hisrec));
  if (h) {
    h->t=(char *)NULL;
    h->nxt=(hisrec *)NULL;
  }
  return(h);
}
/*-------------------- MPEOE objects (atoms and bonds) --------------------*/

MPEOEatom *make_MPEOEatom(void)
{
  MPEOEatom *r;

  r=(MPEOEatom *)malloc(sizeof(MPEOEatom));
  if (r) {
    r->def=(char *)NULL;
    r->nxt=(MPEOEatom *)NULL;
    r->p=(Prule *)NULL;
    r->P[0]=1.e6; r->P[1]=1.0; r->P[2]=1.e20; 
    r->indX=TRUE;
    r->opt[0]=FALSE; r->opt[1]=FALSE; r->opt[2]=FALSE;
    r->nu=0;
  }
  return(r);
}

MPEOEatom *del_MPEOEatom(MPEOEatom *a)
{
  Prule *p;

  if (mvld(a)) {
    if (a->def) free(a->def);
    for (p=a->p;mvld(p);p=a->p) {
      a->p=(Prule *)p->nxt;
      del_prule(p);
    }
    free(a);
  }
  return((MPEOEatom *)NULL);
}

MPEOEbond *make_MPEOEbond(void)
{
  MPEOEbond *r;

  r=(MPEOEbond *)malloc(sizeof(MPEOEbond));
  if (r) {
    r->def=(char *)NULL;
    r->nxt=(MPEOEbond *)NULL;
    r->p=(Prule *)NULL;
    r->f=0.0;
    r->opt=FALSE;
    r->nu=0;
  }
  return(r);
}

MPEOEbond *del_MPEOEbond(MPEOEbond *a)
{
  Prule *p;

  if (mvld(a)) {
    if (a->def) free(a->def);
    for (p=a->p;mvld(p);p=a->p) {
      a->p=(Prule *)p->nxt;
      del_prule(p);
    }
    free(a);
  }
  return((MPEOEbond *)NULL);
}

/*-------------------- make_bond --------------------*/

Bond *make_bond(void)
{
  Bond *b;
  
  b=(Bond *)malloc(sizeof(Bond));
  if (b) {
    b->cross=cPtrNULL();
    b->f=(Atom *)NULL;
    b->t=(Atom *)NULL;
    b->bo='?'; b->bo36=0;
    b->bt=BT_UNKNOWN;
    b->chain=(Bond *)NULL;
    b->arom=0;
    b->flg=0;
    b->cyl=0;
    b->lw=defaultBondWidth;
    b->r=0.1;
    b->labt=BND_LAB_NONE;
    b->label=NULL;
  }
  return(b);
}
/*-------------------- biso_equiv --------------------*/

float biso_equiv(Xatom *xa)
{
  float r;

  if (is_xatom(xa)&&!xa->peak) {
    if (xa->iso) {
      r=xa->bfac[0];
    }else{
      r=xa->bfac[0]*xa->bfac[1]*xa->bfac[2]
	+2*xa->bfac[3]*xa->bfac[4]*xa->bfac[5]
	-xa->bfac[3]*xa->bfac[3]*xa->bfac[0]
	-xa->bfac[4]*xa->bfac[4]*xa->bfac[1]
	-xa->bfac[5]*xa->bfac[5]*xa->bfac[2];
      r=(r > 0 ? pow(r,1.0/3.0) : 0.01);
    }
    r*=8*3.141592653*3.141592653;
  }else r=0;
  return(r);
}

/*-------------------- set_atom_biso --------------------*/

void set_atom_biso(Atom *a,float biso)
{
  Xatom *xa;
  float old;
  int i;

  a->x[ATM_BISO]=biso;
  if (a->xtal) {
    xa=first_xatom(a);
    if (xa) {
      if (xa->iso) {
	xa->bfac[0]=biso/(8*3.141592653*3.141592653);
      }else{
	old=biso_equiv(xa)/(8*3.141592653*3.141592653);
	for (i=0;i<6;i++) xa->bfac[i]*=biso/old;
      }
    }
  }
}

/*-------------------- set_atom_location --------------------*/

void set_atom_location(Atom *a,char alt)
{
  int i,k,l;
  char *locs;
  float *x;

  /* rotate specified 'alt' location to position 0 (active location) */
  /* Note: 'alt' is case sensitive */

  if (a&&a->nxyz>1) {
    for (i=0;i<a->nxyz;i++) if (alt==a->alt[i]) break;
    if (i<a->nxyz&&i>0) {
      x=malloc(ATM_LOCSZ*a->mxyz*sizeof(float));
      locs=malloc(a->mxyz);
      if (x&&locs) {
	for (k=0;k<a->nxyz;k++) {
	  for (l=0;l<ATM_LOCSZ;l++) x[ATM_LOCSZ*k+l]=a->x[ATM_LOCSZ*((k+i)%a->nxyz)+l];
	}
	for (k=0;k<a->nxyz;k++) locs[k]=a->alt[(k+i)%a->nxyz];
	free(a->x); a->x=x;
	free(a->alt); a->alt=locs;
      }else{
	if (x) free(x);
	if (locs) free(locs);
      }
    }
  }
}

/*-------------------- make_adfix --------------------*/

ADfix *make_adfix(void)
{
  ADfix *b;
  
  b=(ADfix *)malloc(sizeof(ADfix));
  if (b) {
    b->class=0;
    b->f=(Xatom *)NULL;
    b->t=(Xatom *)NULL;
    b->dist=0.0;
    b->esd=0.0;
  }
  return(b);
}
/*-------------------- add_adfix --------------------*/

ADfix *add_adfix(Xatom *f,Xatom *t)
{
  ADfix *b=(ADfix *)NULL;

  if (f && t ){
    if (f->nadf==f->madf) reall_adfix_list(f,f->madf+ADCHUNK);
    if (t->nadf==t->madf) reall_adfix_list(t,t->madf+ADCHUNK);
    if (f->nadf<f->madf && t->nadf<t->madf) {
      b=make_adfix();
      if (b) {
	b->f=f;
	b->t=t;
	f->adf[f->nadf++]=b;
	t->adf[t->nadf++]=b;
      }
    }
  }else{
    b=(ADfix *)NULL;
  }
  return(b);
}
/*-------------------- del_adfix --------------------*/

void del_adfix(ADfix *b)
{
  Xatom *a;
  int i;

  if (mvld(b)) {
    if (b->f) {
      a=b->f;
      for (i=0;i<a->nadf;i++) {
	if (b==a->adf[i]) {
	  for (i++;i<a->nadf;i++) a->adf[i-1]=a->adf[i];
	  a->nadf--;
	}
      }
      if (a->nadf<a->madf-ADCHUNK) reall_adfix_list(a,a->madf-ADCHUNK);
    }
    if (b->t) {
      a=b->t;
      for (i=0;i<a->nadf;i++) {
	if (b==a->adf[i]) {
	  for (i++;i<a->nadf;i++) a->adf[i-1]=a->adf[i];
	  a->nadf--;
	}
      }
      if (a->nadf<a->madf-ADCHUNK) reall_adfix_list(a,a->madf-ADCHUNK);
    }
    free(b);
  }
}
/*-------------------- add_bond_raw --------------------*/

Bond *add_bond_raw(Atom *f,Atom *t)
{
  Bond *b;

  if (f->nbnd==f->mbnd) reall_bond_list(f,f->mbnd+BCHUNK);
  if (t->nbnd==t->mbnd) reall_bond_list(t,t->mbnd+BCHUNK);
  if (f->nbnd<f->mbnd && t->nbnd<t->mbnd) {
    b=make_bond();
    if (b) {
      b->f=f;
      b->t=t;
      f->bnd[f->nbnd++]=b;
      t->bnd[t->nbnd++]=b;
    }
  }else{
    b=(Bond *)NULL;
  }
  return(b);
}
/*-------------------- add_bond_ng --------------------*/

Bond *add_bond_ng(Atom *f,Atom *t)
{
  Bond *b;

  if (f && t && (f!=t)) {
    b=find_bond(f,t);
    if (!b) b=add_bond_raw(f,t);
  }else{
    b=(Bond *)NULL;
  }
  return(b);
}
/*-------------------- add_bond --------------------*/

Bond *add_bond(Atom *f,Atom *t)
{
  Bond *b;
  
  b=add_bond_ng(f,t);
  if (mvld(b)) add_bond_graph(b);
  return(b);
}
/*-------------------- add_bond_copy --------------------*/

Bond *add_bond_copy(Atom *f,Atom *t,Bond *bo)
{
  Bond *bn;
  
  /* add a bond (f,t), with same properties as given bond bo */
  /* if (f,t) already exists, just copy properties from bo */
  /* does not add bond to graphics */

  bn=add_bond_ng(f,t);
  if (bn&&bo) {
    set_bt(bn,bo->bt);
    bn->label=bo->label?strdup(bo->label):NULL;
    bn->r=bo->r;
    bn->labt=bo->labt;
    bn->lw=bo->lw;
    bn->cyl=bo->cyl;
  }
  return(bn);
}

/*-------------------- del_bond --------------------*/

void del_bond(Bond *b)
{
  Atom *a;
  int i;


  if (mvld(b)) {
    if (cPtr(b->cross)) del_bond_graph(b);
    if (b->f) {
      a=b->f;
      for (i=0;i<a->nbnd;i++) {
	if (b==a->bnd[i]) {
	  for (i++;i<a->nbnd;i++) a->bnd[i-1]=a->bnd[i];
	  a->nbnd--;
	}
      }
      if (a->nbnd<a->mbnd-BCHUNK) reall_bond_list(a,a->mbnd-BCHUNK);
    }
    if (b->t) {
      a=b->t;
      for (i=0;i<a->nbnd;i++) {
	if (b==a->bnd[i]) {
	  for (i++;i<a->nbnd;i++) a->bnd[i-1]=a->bnd[i];
	  a->nbnd--;
	}
      }
      if (a->nbnd<a->mbnd-BCHUNK) reall_bond_list(a,a->mbnd-BCHUNK);
    }
    if (b->label) free(b->label);
    free(b);
  }
}
/*-------------------- find_bond --------------------*/

Bond *find_bond(Atom *f,Atom *t)
{
  int i;
  Bond *b;

  b=(Bond *)NULL;
  if (is_atom(f) && is_atom(t) && f!=t) {
    if (f->nbnd < t->nbnd) {
      for (i=0;i<f->nbnd;i++) {
	if (f->bnd[i]->f == t || f->bnd[i]->t == t) {
	  b=f->bnd[i];
	  break;
	}
      }
    }else{
      for (i=0;i<t->nbnd;i++) {
	if (t->bnd[i]->t == f || t->bnd[i]->f == f) {
	  b=t->bnd[i];
	  break;
	}
      }
    }
  }
  return(b);
}
/*-------------------- AtomToBondChain --------------------*/

Bond *AtomToBondChain(Nodechain *c)
{
  int j;
  Atom *a,*nei;
  Bond *bs,*be;
  
  be=(Bond *)NULL; bs=(Bond *)NULL;
  if (c&&c->n > 0) {
    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      if (is_atom(a)) {
	for (j=0;j<a->nbnd;j++) {
	  if (!mvld(a->bnd[j]->chain) && a->bnd[j]!=be) {
	    nei=neighbour(j,a);
	    if (node_in_chain((Node *)nei,c)) {
	      if (mvld(bs)) be->chain=a->bnd[j];
	      else bs=a->bnd[j];
	      be=a->bnd[j];
	    }
	  }
	}
      }
    }
  }
  
  return(bs);
}  

/*-------------------- make_atmmatch --------------------*/

Atmmatch *make_atmmatch(long count)
{
  Atmmatch *m;
  int i;

  m=(Atmmatch *)malloc(sizeof(Atmmatch));
  if (m) {
    m->count=count;
    m->a=(Atom **)malloc(count*sizeof(Atom *));
    m->b=(Atom **)malloc(count*sizeof(Atom *));
    if (m->a && m->b) {
      for (i=0;i<m->count;i++) {
	m->a[i]=(Atom *)NULL;
	m->b[i]=(Atom *)NULL;
      }
    }else{
      del_atmmatch(m);
      m=(Atmmatch *)NULL;
    }
  }
  return(m);
}
/*-------------------- del_atmmatch --------------------*/

void del_atmmatch(Atmmatch *m)
{
  if (mvld(m)) {
    if (mvld(m->a)) free(m->a);
    if (mvld(m->b)) free(m->b);
    free (m);
  }
}
/*-------------------- rfirst_type --------------------*/

Node *rfirst_type(Node *root,long obj_type)
{
/* recursive (DFS) find first object of given type in subtree root */

  Node *n,*s;

  if (root) {
    for (s=(Node *)NULL,n=root->fst;!s &&n;n=n->nxt) {
      s=rfirst_type(n,obj_type);
      if (!s && obj_is(n,obj_type)) s=n;
    }
  }else{
    s=(Node *)NULL;
  }
  return(s);
}
/*-------------------- rnext_type --------------------*/

Node *rnext_type(Node *root,Node *now,long obj_type)
{
/* recursive (DFS) find next object of given type in subtree root */

  Node *n,*s;

  n=now;
  if (!n) {
    s=(Node *)NULL;
  }else{
    for (s=(Node *)NULL,n=n->nxt;!s && n;n=n->nxt) {
      s=rfirst_type(n,obj_type);
      if (!s && obj_is(n,obj_type)) s=n;
    }
    if (!s) {
      n=now;
      while (n->up && n->up != root && !s) {
	n=n->up;
	if (obj_is(n,obj_type)) s=n;
	else s=rnext_type(root,n,obj_type);
      }
    }
  }
  return(s);
}
/*-------------------- first_type --------------------*/

Node *first_type(Node *root,long obj_type)
{
/*  find first child of root with given type  */

  Node *s;

  if (root) {
    for (s=root->fst;s;s=s->nxt) if (obj_is(s,obj_type)) break;
  }else{
    s=(Node *)NULL;
  }
  return(s);
}
/*-------------------- last_type --------------------*/

Node *last_type(Node *root,long obj_type)
{
/*  find first child of root with given type  */

  Node *s;

  if (root) {
    for (s=root->lst;s;s=s->prv) if (obj_is(s,obj_type)) break;
  }else{
    s=(Node *)NULL;
  }
  return(s);
}
/*-------------------- next_type --------------------*/

Node *next_type(Node *now,long obj_type)
{
/* find next object of given type */

  Node *s;

  if (now) {
    for (s=now->nxt;s;s=s->nxt) if (obj_is(s,obj_type)) break;
  }else{
    s=(Node *)NULL;
  }
  return(s);
}
/*-------------------- find_cousin --------------------*/

Node *find_cousin(Node *now,long obj_type)
{
  /* find first node of given type that is a direct descendant of an
     ancestor of 'now' */

  Node *r;

  r=(Node *)NULL;

  if (mvld(now)) {
    while (!r && mvld(now->up)) {
      now=now->up;
      r=first_type(now,obj_type);
    }
  }
  return(r);
}
/*-------------------- find_node --------------------*/

Node *find_node(Node *root,void *addr,long obj_type)
{
  /* find node with given address and type in tree with specified root */

  Node *s;

  ForObjs(s,root,obj_type) if (s==(Node *)addr) break;
  return(s);
}
/*-------------------- find_name --------------------*/

Node *find_name(Node *root,char *name,long obj_type)
{
/*   find object by name (non recursive)  */

  Node *s;

  forObjs(s,root,obj_type) if (streq(s->name,name)) break;
  return(s);
}
/*-------------------- rfind_name --------------------*/

Node *rfind_name(Node *root,char *name,long obj_type)
{
/*   find object by name (recursive)  */

  Node *s;

  ForObjs(s,root,obj_type) if (streq(s->name,name)) break;
  return(s);
}
/*-------------------- find_cname --------------------*/

Node *find_cname(Node *root,char *name,long obj_type)
{
/*   find object by name (non recursive,case insensitive)  */

  Node *s;

  forObjs(s,root,obj_type) if (strceq(s->name,name)) break;
  return(s);
}
/*-------------------- reall_atom_x --------------------*/

float *reall_atom_x(Atom *a,int size)
{
  float *n;
  char *cn;
  int i,ncpy;

  if (size<0 || size>255) return((float *)NULL);
  if (size==a->mxyz) return(a->x);
  ncpy=(size > a->nxyz ? a->nxyz : size);
  if (size>0) {
    n=(float *)malloc(ATM_LOCSZ*size*sizeof(float));
    cn=(char *)malloc(size*sizeof(char));
    if (n&&cn) {
      for (i=0;i<ATM_LOCSZ*ncpy;i++) n[i]=a->x[i];
      a->mxyz=size;
      if (mvld(a->x)) free(a->x);
      a->x=n;
      a->nxyz=ncpy;
      if (mvld(a->alt)) {
	if (ncpy>0) strncpy(cn,a->alt,ncpy);
	free(a->alt);
      }
      a->alt=cn;
    }
  }else{
    if (mvld(a->x)) free(a->x);
    a->x=(float *)NULL;
    if (a->alt) free(a->alt);
    a->alt=NULL;
    a->nxyz=0;
    a->mxyz=0;
    n=(float *)NULL;
  }
  return(n);
}

/*-------------------- newLib --------------------*/

Node *newLib(char *name)
{
  Node *LibRoot;

  /* Libraries consist of a top level LIB node with four LIB subnodes:

  ResLib: symbols,atoms,residues,links
  PruleLib: FF potentail assignment rules
  FruleLib: FF fix (re-write) rules
  CruleLib: fragment charge definitions

  The *ruleLib nodes have one subnode node per ruleset (TAFF, AMBER, MMOD, ...)
  This function creates an "empty" library tree, and puts a default atomtype definition
  into the ResLib subtree.

  */

  LibRoot=make_node(OBJ_LIB);
  if (LibRoot) {
    set_node_name(LibRoot,name?name:"DefaultLibrary");
    add_node(LibRoot,"ResLib",OBJ_LIB);    /* residue library                      */
    add_node(LibRoot->fst,"unk",OBJ_ATOMDEF);
    add_node(LibRoot,"PruleLib",OBJ_LIB);  /* ff potential type libraries          */
    add_node(LibRoot,"CruleLib",OBJ_LIB);  /* partial atomic charge rule libraries */
    add_node(LibRoot,"FruleLib",OBJ_LIB);  /* ff potential type re-write rules     */
  }
  return(LibRoot);
}

/*-------------------- assoc_dsp_obj --------------------*/

Dspobj *assocobj(Node *n)
{
  Dspobj *r;

  if (is_dsp(n)) r=(Dspobj *)n;
  else if (is_mol(n)) r=(((Mol *)n)->dsptr?(Dspobj *)((Mol *)n)->dsptr:assocobj(n->up)); 
  else if (is_atom(n)) r=assocobj(n->up);
  else if (is_xcn(n)) r=assocobj(n->up);
  else r=(Dspobj *)NULL;
  return(r);
}
/*-------------------- glblobj --------------------*/

Dspobj *glblobj(Dspobj *o)
{
/*   return pointer to global that is parent of o (to self if o is glbl) */

  while(is_dsp(o)) {
      if (((Dspobj *)o)->dsptype==DSP_GLBL) return(o);
      o=(Dspobj *)o->up;
  }
  return((Dspobj *)NULL);
}
/*-------------------- assocmdl --------------------*/

Dspobj *assocmdl(Node *m)
{
/*   return pointer to mdl associated with node */

  Dspobj *r,*g;

  r=(Dspobj *)NULL;

  for (g=(Dspobj *)NULL;mvld(m);m=m->up) {
    if (is_mol(m)) {
      if (mvld(((Mol *)m)->dsptr)) {
	g=(Dspobj *)((Mol *)m)->dsptr;
	break;
      }
    }
  }

  if (mvld(g)) {
    if (g->dsptype!=DSP_GLBL) g=(Dspobj *)g->up;
    ForDspobjs(r,g) {
      if (r->dsptype==DSP_MOL && r->root==m) break;
    }
  }

  return(r);
}

/*-------------------- node chain utilities --------------------*/

Nodechain *make_node_chain(void)
{
  Nodechain *r;

  r=(Nodechain *)malloc(sizeof(Nodechain));
  init_node_chain(r);
  return(r);
}

void init_node_chain(Nodechain *r)
{
  if (r) {
    r->n=0;
    r->s=(Node *)NULL;
    r->p=(Node *)NULL;
    r->e=(Node *)NULL;
  }
}

void clear_node_chain(Nodechain *a)
{
  Node *m,*n;

  if (a) {
    n=a->s;
    while (n) {
      m=n->chain;
      n->chain=(Node *)NULL;
      n=m;
    }
    a->n=0;
    a->s=(Node *)NULL;
    a->e=(Node *)NULL;
    a->p=(Node *)NULL;
  }
}

Node **tree2ptr(Node *r,long type,int *n)
{
  Node **p=NULL;
  Node *o;

  *n=0;
  ForObjs(o,r,type) (*n)++;
  if (*n>0) {
    p=malloc(*n*sizeof(Node *));
    if (p) {
      *n=0;
      ForObjs(o,r,type) p[(*n)++]=o;
    }
  }
  return(p);
}

Nodechain tree2chain(Node *n,long type)
{
  Nodechain r;
  Node *o;

  r.n=0;
  r.s=(Node *)NULL;
  r.e=(Node *)NULL;
  r.p=(Node *)NULL;

  if (mvld(n)) {
    ForObjs(o,n,type) {
      r.n++;
      if (!r.s) r.s=o;
      if (r.e) r.e->chain=o;
      r.e=o;
    }
    if (r.e) r.e->chain=(Node *)NULL;
  }
  return(r);
}
void add_atomset_to_chain(Atomset *s,Nodechain *c)
{
  Atom *a;

  if (!is_atomset(s)) return;
  if (is_mol(s->up)) {
    ForAtoms(a,s->up) {
      if (in_atomset(a,s)) add_to_node_chain((Node *)a,c);
    }
  }
}

void add_to_node_chain(Node *n,Nodechain *c)
{
  if (!mvld(n)) return;
  if (mvld(n->chain)||c->e==n) return;
  if (mvld(c->e)) c->e->chain=n;
  c->p=c->e;
  c->e=n;
  n->chain=(Node *)NULL;
  if (!mvld(c->s)) c->s=c->e;
  c->n++;
}

Node *RemoveHeadOfChain(Nodechain *c)
{
  Node *r;

  if (c) {
    r=c->s;
    if (r) {
      c->n--; c->s=r->chain;
      if (!c->s) c->e=(Node *)NULL;
      r->chain=(Node *)NULL;
    }
  }else r=(Node *)NULL;
  return(r);
}

Nodechain *free_node_chain(Nodechain *a)
{
  if (mvld(a)) free(a);
  return((Nodechain *)NULL);
}

int node_in_chain(Node *n,Nodechain *c)
{
  /* return TRUE iff node n is in chain c */

  int r;

  if (mvld(n)) {
    r = (n==c->e || mvld(n->chain));
  }else{
    r = FALSE;
  }
  return(r);
}

int chain_reduce_subtree(Node *root,Nodechain *c)
{
/*  reduce chain c, removing nodes not in subtree "root" */
/*  return: number of removed atoms */

  Node *n,*t;
  int rm;

  n=c->s;
  c->s=c->e=(Node *)NULL;
  c->n=0; rm=0;

  while(n) {
    c->p=n;
    for (t=n;t;t=t->up) if (t==root) break;
    if (t) {
      if (c->n > 0) c->e->chain=n;
      else c->s=n;
      c->e=n;
      c->n++;
    }else{
      rm++;
    }
    n=n->chain;
    c->p->chain=(Node *)0;
  }
  return(rm);
}
/*-------------------- chain_to_ptr --------------------*/

Node **chain_to_ptr(Nodechain *c,int fr_chain)
{
  /* dynamycally allocate array of Nodepointers, clear and free c iff fr_chain */

  Node **r=NULL;
  Node *n;
  int i;

  if (mvld(c)) {
    if (c->n > 0) {
      r=(Node **)malloc(c->n*sizeof(Node *));
      if (mvld(r)) {
	if (fr_chain) {
	  for (i=0,n=c->s;mvld(n);n=c->s) {
	    c->s=n->chain;
	    n->chain=(Node *)NULL;
	    r[i++]=n;
	  }
	}else{
	  for (i=0,n=c->s;mvld(n);n=n->chain) r[i++]=n;
	}
      }else{
	err_set(1,"chain_to_ptr: allocation failed, n=%d.",c->n);
      }
    }else{
      r=(Node **)NULL;
    }
    if (fr_chain) {
      clear_node_chain(c);
      free_node_chain(c);
    }
  }
  return(r);
}
/*-------------------- collect_by_flags --------------------*/

Nodechain *collect_by_flags(Node *n,int flags)
{
  Nodechain *r;
  Atom *a;

  r=make_node_chain();
  if (mvld(r)) {
    ForAtoms (a,n) {
      if (a->flg & flags) add_to_node_chain((Node *)a,r);
    }
  }
  return(r);
}
/*-------------------- del_node_chain --------------------*/

void del_node_chain(Nodechain *C)
{
  /* delete nodes in chain C */

  Node *a,*b,*c;
  Nodechain D;

  D.n=0; D.s=(Node *)NULL; D.e=D.s; D.p=D.s;
  for (a=C->s;mvld(a);a=b) {
    b=a->chain; a->chain=(Node *)NULL;
    for (c=a->up;c;c=c->up) 
      if (mvld(c->chain)||c==C->e||c==D.e) break;
    if (!mvld(c)) add_to_node_chain(a,&D);
  }
  for (a=D.s;mvld(a);) {
    b=a;
    a=b->chain;
    del_node(b);
  }
  C->s=(Node *)NULL;
  C->e=C->s;
  C->p=C->s;
  C->n=0;
}
/*-------------------- residue_closure --------------------*/

void residue_closure(Nodechain *c)
{
  Atom *a,*b;

  if (!c) return;

  for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
    if (is_atom(a)) {
      if (!(a->flg&(1<<7))) {
	forAtoms(b,a->up) {
	  b->flg|=(1<<7);
	  add_to_node_chain((Node *)b,c);
	}
      }
    }
  }
  for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
    if (is_atom(a)) a->flg&=~(1<<7);
  }
}

/*-------------------- ptr_to_chain --------------------*/

Nodechain *ptr_to_chain(Node **N,int n)
{
  Nodechain *r;
  int i;

  r=make_node_chain();
  if (mvld(r)) {
    if (n>0&&N) {
      r->s=N[0];
      r->n=1;
      r->p=r->s;
      for (i=1;i<n;i++) N[i-1]->chain=N[i];
      r->e=N[n-1];
      r->n=n;
      N[n-1]->chain=(Node *)NULL;
    }
  }else{
    err_set(1,"could not make nodechain...");
  }
  return(r);
}
/*-------------------- walk_order --------------------*/

int walk_order(Node *root,int n,Atom **atoms,int flg)
{
  int i,r;
  Atom *a;

/* reorder atomlist according to tree-walk sequence... 
   returns number of atoms that are in node root.

*/

  r=0; if (!atoms) n=0;
  for (i=0;i<n;i++) atoms[i]->flg|=flg;
  ForAtoms(a,root) a->flg&=~flg;
  for (i=0;i<n;i++) atoms[i]->flg^=flg;
  ForAtoms(a,root) if (a->flg&flg) atoms[r++]=a;
  for (i=0;i<r;i++) atoms[i]->flg^=flg;
  return(r);
}
/*-------------------- walk_order_chain --------------------*/

Nodechain *walk_order_chain(Node *root,Nodechain *c,int flg)
{
  Atom *a;
  Node *n;

/* reorder Nodechain according to tree-walk sequence... 
   returns pointer to modified nodechain c.
   Nodes in c that are not in root will be removed from chain.
   Non-atom nodes will be removed from chain.

   flg: flag bit used internally.

*/

  if (c&&c->n>0) {
    if (!root) root=common_root(c);
    ForAtoms(a,root) a->flg&=~flg;
    for (n=c->s;n;n=n->chain) if (is_atom(n)) ((Atom *)n)->flg|=flg;
    clear_node_chain(c);
    ForAtoms(a,root) {
      if (a->flg&flg) {
	add_to_node_chain((Node *)a,c);
	a->flg&=~flg;
      }
    }
  }
  return(c);
}

/*-------------------- available_atomset_bit --------------------*/

int available_atomset_bit(int bit,Mol *m)
{
/* return "bit", if number available for atomset in mol m,
   else return first available bit number, 
   return -1 on error or if no bit available  */

  unsigned long una[ATOMSET_WORDS],mask;
  int i;

  if (!is_mol(m)) return(-1);

  for (i=0;i<ATOMSET_WORDS;i++) una[i]=m->local[i]|m->used[i];
  for (m=(Mol *)m->up;m;m=(Mol *)m->up) {
    if (is_mol(m)) {
      for (i=0;i<ATOMSET_WORDS;i++) una[i]|=m->local[i];
    }
  }
  if (bit>=0 && bit<BITS_PER_WORD) {
    i=bit/BITS_PER_WORD;
    mask=1<<(bit%BITS_PER_WORD);
    bit=(mask&una[i] ? -1 : bit);
  }else{
    bit=-1;
  }
  if (bit<0) {
    for (i=0;i<ATOMSET_WORDS;i++) {
      mask = ~una[i];
      if (mask) {
	for(bit=0;bit<BITS_PER_WORD;bit++,mask>>=1) if (mask%2) break;
	bit+=i*BITS_PER_WORD;
	break;
      }
    }
  }
  return(bit);
}
/*-------------------- register_atomset --------------------*/

void register_atomset(Atomset *s)
{
/* update local/used bits for new atomset */

  Mol *m;
  Node *u;
  int w;
  unsigned long mask;

  if (!s) return;
  if (!is_atomset(s)) return;

  m=(Mol *)s->up;
  if (!is_mol(m)) return;

  w=s->bit/BITS_PER_WORD;
  mask=1<<(s->bit % BITS_PER_WORD);

  m->local[w]|=mask;
  for (u=m->up;u;u=u->up) if (is_mol(u)) ((Mol *)u)->used[w]|=mask;
}

/*-------------------- remove_atomset_bit --------------------*/

void remove_atomset_bit(Atomset *s)
{
  Node *p;
  Mol *m,*y;
  int w;
  unsigned long new;

  if (!s) return;
  if (!is_mol(s->up)) return;
  m=(Mol *)s->up;
  if (s->bit<0 || s->bit>=ATOMSET_BITS) return;

  w=s->bit/BITS_PER_WORD;

  m->local[w] &= ~(1<<(s->bit % BITS_PER_WORD));
  for (p=m->up;p;p=p->up) {
    if (is_mol(p)) {
      y=(Mol *)p;
      new=used_bits(p,s->bit);
      if (new==y->used[w]) break;
      else y->used[w]=new;
    }
  }
}
/*-------------------- used_bits --------------------*/

unsigned long used_bits(Node *n,int bit)
{
/* assuming all Mols in subtree of n have correct local/used set info,
   return used word for node n, and set "bit". */

  int w;
  unsigned long r;
  Node *s;
  
  r=0;

  if (n && (bit>=0 && bit<ATOMSET_BITS)) {
    w=bit/BITS_PER_WORD;
    for(s=n->fst;s;s=s->nxt) {
      if (is_mol(s)) r|=((Mol *)s)->local[w]|((Mol *)s)->used[w];
      else r|=used_bits(s,bit);
    }
  }
  return(r);
}
/*-------------------- clear_atomset --------------------*/

void clear_atomset(Atomset *s)
{
/* clear bits in atoms belonging to set s */

  unsigned long mask;
  int w;
  Mol *m;
  Atom *a;

  if (!s) return;
  if (!is_mol(s->up)) return;
  if (s->bit<0 || s->bit>=ATOMSET_BITS) return;

  w=s->bit/BITS_PER_WORD;
  mask = ~(1<<(s->bit % BITS_PER_WORD));
  m=(Mol *)s->up;
  ForAtoms(a,m) a->set[w]&=mask;
}
/*-------------------- add_to_atomset --------------------*/

void add_to_atomset(Atom *a,Atomset *s)
{
  unsigned long mask;
  int w;

  if (!is_atom(a)||!is_atomset(s)) return;       /* check for root ??? */
  if (s->bit<0 || s->bit>=ATOMSET_BITS) return;
  w=s->bit/BITS_PER_WORD;
  mask = (1<<(s->bit % BITS_PER_WORD));
  a->set[w]|=mask;
}
/*-------------------- remove_from_atomset --------------------*/

void remove_from_atomset(Atom *a,Atomset *s)
{
  unsigned long mask;
  int w;

  if (!is_atom(a)||!is_atomset(s)) return;
  if (s->bit<0 || s->bit>=ATOMSET_BITS) return;
  w=s->bit/BITS_PER_WORD;
  mask = ~(1<<(s->bit % BITS_PER_WORD));
  a->set[w]&=mask;
}
/*-------------------- copy_set_bits --------------------*/

void copy_set_bits(Node *source,Node *target)
{
  /* Set bits for atom in 'source' that are set in 'target' and
     correspond to set defined in common root. 'source' must be
     a tree_copy of target */

  unsigned long mask[ATOMSET_WORDS],any;
  int i,lev;
  Node *r;
  Atom *s,*t;

  if (source==target) return;
  r=common_parent(source,target);
  if (mvld(r)) {
    for (i=0;i<ATOMSET_WORDS;i++) mask[i]=0;
    while(r) {
      if (is_mol(r)) for(i=0;i<ATOMSET_WORDS;i++) mask[i]|=((Mol *)r)->local[i];
      r=r->up;
    }
    for (i=1,any=mask[0];i<ATOMSET_WORDS;i++) any|=mask[i];
    if (any) {
      lev=0;
      ForAtoms(s,source) lev++;
      ForAtoms(t,target) lev--;
      if (lev==0) {               /* for tree_copies, number of atoms must be the same */
	s=rfirst_atom(source);
	t=rfirst_atom(target);
	while(mvld(s)&&mvld(t)) {
	  for (i=0;i<ATOMSET_WORDS;i++) t->set[i]|=s->set[i]&mask[i];
	  s=rnext_atom(source,s);
	  t=rnext_atom(target,t);
	}
      }
    }
  }
}
/*-------------------- in_atomset --------------------*/

int in_atomset(Atom *a,Atomset *s)
{
  int w;
  unsigned long mask;

  if (!is_atom(a)||!is_atomset(s)) return(FALSE);
  if (s->bit<0 || s->bit>=ATOMSET_BITS) return(FALSE);
  w=s->bit/BITS_PER_WORD;
  mask = (1<<(s->bit % BITS_PER_WORD));
  return((a->set[w]&mask)!=0);
}
/*-------------------- define_atomset --------------------*/

Atomset *define_atomset(Mol *m,char *sname,Nodechain *c)
{
  Atomset *s;
  Atom *a;
  char buf[4096];

  chain_reduce_subtree((Node *)m,c);
  s=(Atomset *)find_cname((Node *)m,sname,OBJ_ATOMSET);
  if (mvld(s)) clear_atomset(s);
  else s=add_atomset((Mol *)m,-1,sname);
  if (mvld(s)) {
    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) add_to_atomset(a,s);
    prtGraphics("%s contains %d atoms.\n",nodeptr_to_name(buf,(Node *)s),c->n);
  }else{
    err_set(1,"sorry, could not add set %s to molecule %s",sname,nodeptr_to_name(buf,(Node*)m));
  }
  return(s);
}
/*-------------------- add_atomset --------------------*/

Atomset *add_atomset(Mol *m,int bit,char *name)
{
/* Add empty atomset to sets local in m */

  Atomset *s;

  s=(Atomset *)NULL;
  if (!m || !name) return(s);
  if (!is_mol(m)) return(s);
  
  bit=available_atomset_bit(bit,m);
  if (bit>=0) {
    s=(Atomset *)add_node((Node *)m,name,OBJ_ATOMSET);
    if (s) {
      s->bit=bit;
      register_atomset(s);
    }
  }
  return(s);
}
/*-------------------- merge_atomsets --------------------*/

void merge_atomsets(Node *r)
{
  /* merge sets in r with equal name (not case-sensitive) */
  /* not very efficinet, I am afraid... */

  Atomset *s,*t;
  Node *a,*b;
  Atom *atm;

  if (is_mol(r)) {
    for (a=r->fst;mvld(a);a=a->nxt) {
      if (is_atomset(a)) {
	s=(Atomset *)a;
	for (b=r->lst;b!=a;) {
	  if (is_atomset(b)) {
	    t=(Atomset *)b;
	    b=b->prv;
	    if (strceq(t->name,s->name)) {
	      ForAtoms(atm,r) if (in_atomset(atm,t)) add_to_atomset(atm,s);
	      del_node((Node *)t);
	    }
	  }else b=b->prv;
	}
      }
    }
  }
}
/*-------------------- renumber_atomsets --------------------*/

void renumber_atomsets(Node *root,Node *target)
{
  /* renumber atomsets in tree at root */

  Atomset *s;
  Mol *o;
  Node *n;
  Atom *a;
  int i,k,cnflct,map[ATOMSET_BITS],w,b;
  unsigned long R[ATOMSET_WORDS],T[ATOMSET_WORDS],U[ATOMSET_WORDS],L[ATOMSET_WORDS];

  if (is_mol(root)&&is_mol(target)) {
    for (i=0;i<ATOMSET_WORDS;i++) T[i]=((Mol *)target)->local[i];
    for (o=(Mol *)target;mvld(o);o=(Mol *)o->up) {
      if (is_mol(o)) for (i=0;i<ATOMSET_WORDS;i++) T[i]|=o->local[i];
    }
    for (i=0,cnflct=FALSE;i<ATOMSET_WORDS;i++) {
      R[i]=((Mol *)root)->local[i]|((Mol *)root)->used[i];
      cnflct|=R[i]&T[i];
    }
    if (cnflct) {
      for (k=0,i=0;i<ATOMSET_BITS;i++) {
	map[i]=-1;
	w=i/BITS_PER_WORD;
	b=1<<i%BITS_PER_WORD;
	if (R[w]&b) {
	  for(k++;k<ATOMSET_BITS;k++) if (!(T[k/BITS_PER_WORD]&(1<<k%BITS_PER_WORD))) break;
	  map[i]=k;
	}
      }
      ForObjs(n,root,OBJ_MOL|OBJ_ATOM|OBJ_ATOMSET) {
	switch (n->type) {
	case OBJ_ATOMSET:
	  s=(Atomset *)n;
	  if (map[s->bit]>=0&&map[s->bit]<ATOMSET_BITS) s->bit=map[s->bit];
	  else del_node(n);
	  break;
	case OBJ_MOL:
	  o=(Mol *)n;
	  for (i=0;i<ATOMSET_WORDS;i++) {
	    U[i]=0;
	    L[i]=0;
	  }
	  for (i=0;i<ATOMSET_BITS;i++) {
	    if (map[i]>=0&&map[i]<ATOMSET_BITS) {
	      w=map[i]/BITS_PER_WORD;
	      b=map[i]%BITS_PER_WORD;
	      if (o->used[i/BITS_PER_WORD]&(1<<i%BITS_PER_WORD)) U[w]|=1<<b;
	      if (o->local[i/BITS_PER_WORD]&(1<<i%BITS_PER_WORD)) L[w]|=1<<b;
	    }
	  }
	  for (i=0;i<ATOMSET_WORDS;i++) {
	    o->local[i]=L[i];
	    o->used[i]=U[i];
	  }
	  break;
	case OBJ_ATOM:
	  a=(Atom *)n;
	  for (i=0;i<ATOMSET_WORDS;i++) L[i]=0;
	  for (i=0;i<ATOMSET_BITS;i++) {
	    if (map[i]>=0&&map[i]<ATOMSET_BITS) {
	      w=map[i]/BITS_PER_WORD;
	      b=map[i]%BITS_PER_WORD;
	      if (a->set[i/BITS_PER_WORD]&(1<<i%BITS_PER_WORD)) L[w]|=1<<b;
	    }
	  }
	  for (i=0;i<ATOMSET_WORDS;i++) a->set[i]=L[i];
	  break;
	default:
	  break;
	}
      }
    }
  }
}
/*-------------------- RDeleteNodesInChain --------------------*/

int RDeleteNodesInChain(Nodechain *c)
{
  int ndel;
  Node *n,*nxt,*up;

  ndel=0;
  if (c) {
    for (c->p=(Node *)NULL,n=c->s;n;n=nxt) {
      nxt=n->chain;
      for (up=n->up;up;up=up->up) if (node_in_chain(up,c)) break;
      if (up&&up->up) {
	if (c->p) c->p->chain=nxt;
	else c->s=nxt;
	c->n--;
      }else c->p=n;
    }
    c->e=c->p;
    while ((n=RemoveHeadOfChain(c))) {
      ndel++;
      up=( is_mol(n->up)&&((Mol *)n->up)->keep==0 ? n->up : (Node *)NULL);
      del_node(n);
      if (up&&up->up) {
	if (!first_type(up,OBJ_MOL|OBJ_ATOM)) add_to_node_chain(up,c);
      }
    }
  }
  return(ndel);
}
/*-------------------- RDeleteAtomsInArray --------------------*/

void RDeleteAtomsInArray(int nat,Atom **atoms)
{
  Nodechain p;
  Node *r;
  int i;
  
  init_node_chain(&p);

  if (nat>0&&atoms) {
    /* delete atoms... */
    for (i=0;i<nat;i++) {
      if (is_atom(atoms[i])) {
	add_to_node_chain(atoms[i]->up,&p);
	del_node((Node *)atoms[i]); 
      }
    }
    /* purge empty (no atoms, no mols) nodes... */
    while ((r=RemoveHeadOfChain(&p))) {
      if (!first_type(r,OBJ_MOL|OBJ_ATOM)) {
	if (!is_mol(r)) {
	  add_to_node_chain(r->up,&p);
	  del_node(r);
	}else if (((Mol *)r)->keep==0) {
	  add_to_node_chain(r->up,&p);
	  del_node(r);
	}
      }
    }
  }
}
/*-------------------- NOE data handling --------------------*/

NOEdata *make_noedata(void)
{
  NOEdata *r;

  r=(NOEdata *)malloc(sizeof(NOEdata));
  if (mvld(r)) {
    r->root=(NOElist *)NULL;
    r->prv=(NOEdata *)NULL;
    r->nxt=(NOEdata *)NULL;
    r->sres[0]='\0'; r->tres[0]='\0';
    r->satom[0]='\0'; r->tatom[0]='\0';
    r->dupper=1.E20; r->weight=1.0;
    r->active=TRUE;
    r->def1=(char *)NULL; r->ps1=FALSE;
    r->def2=(char *)NULL; r->ps2=FALSE;
    r->sseg=(char *)NULL;
    r->tseg=(char *)NULL;
    r->sptr=(Atom *)NULL;
    r->tptr=(Atom *)NULL;
    r->cmnt=(char *)NULL;
  }
  return(r);
}

NOEdata *copy_noedata(NOEdata *noe)
{
  NOEdata *new;

  new=make_noedata();
  if (mvld(new)) {
    strcpy(new->sres,noe->sres);
    strcpy(new->satom,noe->satom);
    strcpy(new->tres,noe->tres);
    strcpy(new->tatom,noe->tatom);
    new->dupper=noe->dupper;
    new->weight=noe->weight;
    new->def1=Strdup(noe->def1);
    new->def2=Strdup(noe->def2);
    new->sseg=Strdup(noe->sseg);
    new->tseg=Strdup(noe->tseg);
    new->cmnt=Strdup(noe->cmnt);
  }                                      /* should ->active be copied ? */
  return(new);
}

void del_noedata(NOEdata *noe)
{
  NOElist *root;
  NOEdata *nxt,*prv;

  if (mvld(noe)) {
    nxt=noe->nxt;
    prv=noe->prv;
    root=noe->root;
    if (mvld(noe->def1)) free(noe->def1);
    if (mvld(noe->def2)) free(noe->def2);
    if (mvld(noe->sseg)) free(noe->sseg);
    if (mvld(noe->tseg)) free(noe->tseg);
    if (mvld(noe->cmnt)) free(noe->cmnt);
    free(noe);
    if (mvld(prv)) {
      prv->nxt=nxt;
    }else if (is_NOElist(root)) {
      root->first=nxt;
    }
    if (mvld(nxt)) nxt->prv=prv;
  }
}

NOEdata *add_noedata(NOElist *list)
{
  NOEdata *noe=(NOEdata *)NULL;

  if (is_NOElist(list)) {
    noe=make_noedata();
    if (mvld(noe)) {
      noe->root=list;
      link_noedata(noe);
    }
  }
  return(noe);
}

NOEdata *link_noedata(NOEdata *noe)
{
  NOEdata *r,*prv,*nxt;

  r=(NOEdata *)NULL;

  if (mvld(noe)) {
    if (is_NOElist(noe->root)) {
      if (!mvld(noe->prv) && !mvld(noe->nxt)) {
	if (mvld(noe->root->last)) noe->root->last->nxt=noe;
	noe->prv=noe->root->last;
	noe->root->last=noe;
	if (!mvld(noe->root->first)) noe->root->first=noe;
	r=noe;
      }
    }else if (mvld(noe->prv)) {
      if (!mvld(noe->root) && !mvld(noe->nxt)) {
	noe->root=noe->prv->root;
	nxt=noe->prv->nxt;
	noe->prv->nxt=noe;
	if (mvld(nxt)) nxt->prv=noe;
	else noe->root->last=noe;
	noe->nxt=nxt;
	r=noe;
      }
    }else if (mvld(noe->nxt)) {
      if (!mvld(noe->root) && !mvld(noe->prv)) {
	noe->root=noe->nxt->root;
	prv=noe->nxt->prv;
	noe->nxt->prv=noe;
	if (mvld(prv)) prv->nxt=noe;
	else noe->root->first=noe;
	noe->prv=prv;
	r=noe;
      }
    }
  }
  return(r);
}
/*-------------------- transform_atom --------------------*/

void transform_atom(mat_4 t,Atom *a)
{
  int i;

  /* should also handle XCN atoms... */

  if (!is_atom(a)) return;
  for (i=0;i<a->nxyz;i++) v32arr(a->x+ATM_LOCSZ*i,m4xv3(t,vec3(a->x+ATM_LOCSZ*i)));
}
/*-------------------- add_bondinc --------------------*/

Bondinc *add_bondinc(Lib *lib,char *t1,char *t2,float d12,float d21)
{
  Bondinc *r;
  Atomdef *ad1,*ad2;

  r=(Bondinc *)NULL;
  if (!is_lib(lib)) return(r);
  if (!mvld(t1) || !mvld(t2)) return(r);

  ad1=find_atomdef_name((Node *)lib,t1);
  if (is_atomdef(ad1)) {
    if (ad1->bnd_equiv) ad1=find_atomdef_name((Node *)lib,ad1->bnd_equiv);
  }
  ad2=find_atomdef_name((Node *)lib,t2);
  if (is_atomdef(ad2)) {
    if (ad2->bnd_equiv) ad2=find_atomdef_name((Node *)lib,ad2->bnd_equiv); 
  }

  if (!is_atomdef(ad1) || !is_atomdef(ad2)) return(r);
  
  if (strcmp(t1,t2)<=0) {
    forBondincs(r,ad1) if (streq(r->name,ad2->name)) break;
    if (!mvld(r)) r=(Bondinc *)add_node((Node *)ad1,ad2->name,OBJ_BONDINC);
    if (mvld(r)) {
      r->dij=d12;
      r->dji=d21;
    }
  }else{
    forBondincs(r,ad2) if (streq(r->name,ad1->name)) break;
    if (!mvld(r)) r=(Bondinc *)add_node((Node *)ad2,ad1->name,OBJ_BONDINC);
    if (mvld(r)) {
      r->dij=d21;
      r->dji=d12;
    }
  }
  return(r);
}

/*-------------------- get_bondinc --------------------*/

float get_bondinc(Lib *lib,Atom *a,Atom *b)
{
  float r;
  char *t1,*t2;
  Bondinc *B;
  Atomdef *ad;

  r=-9999.9;
  
  if (!is_lib(lib)) return(r);
  if (!is_atom(a) || !is_atom(b)) return(r);
  if (!mvld(a->atp) || !mvld(b->atp)) return(r);

  t1=(mvld(a->atp->bnd_equiv) ? a->atp->bnd_equiv : a->atp->name);
  t2=(mvld(b->atp->bnd_equiv) ? b->atp->bnd_equiv : b->atp->name);

  if (!mvld(t1) || !mvld(t2)) return(r);

  if (strcmp(t1,t2)<=0) {
    ad=find_atomdef_name((Node *)lib,t1);
    if (is_atomdef(ad)) {
      forBondincs(B,ad) if (streq(B->name,t2)) break;
      if (is_bondinc(B)) r=B->dij;
      else if (a->atp==b->atp) r=0.0;
    }
  }else{
    ad=find_atomdef_name((Node *)lib,t2);
    if (is_atomdef(ad)) {
      forBondincs(B,ad) if (streq(B->name,t1)) break;
      if (is_bondinc(B)) r=B->dji;
      else if (a->atp==b->atp) r=0.0;
    }
  }
  return(r);
}
/*-------------------- find_atom_def_name --------------------*/

Atomdef *find_atomdef_name(Node *lib,char *name)
{
  /*   find atom definition by name in library,
       return NULL if definition not found */
  
  Node *n;
  
  n=(Node *)NULL;
  if (is_lib(lib)) {
    for (n=first_type(lib,OBJ_ATOMDEF);n;n=next_type(n,OBJ_ATOMDEF)) {
      if (strcmp(name,n->name)==0) break;
    }
  }
  return((Atomdef *)n);
}
/*-------------------- find_atom_def_number --------------------*/

Atomdef *find_atomdef_number(Node *lib,int nr)
{
  /*   find atom definition by number in table starting at 'start' ,
       return NULL if definition not found */
  
  Node *n;
  
  n=(Node *)NULL;
  if (is_lib(lib)) {
    for (n=first_type(lib,OBJ_ATOMDEF);n;n=next_type(n,OBJ_ATOMDEF)) {
      if (nr==((Atomdef *)n)->numtp) break;
    }
  }
  return((Atomdef *)n);
}
/*-------------------- obj_typename --------------------*/

char *obj_typename(long type)
{
  int i;
  char *r;

  r=obj_unknown;
  for (i=0;i<OBJ_N_TYPES;i++) {
    if ((1<<i)&type) {
      r=obj_types[i];
      break;
    }
  }
  return(r);
}

/*-------------------- data_typename --------------------*/

char *data_typename(int type)
{
  char *r;

  if (type>=0&&type<sizeof(data_types)/sizeof(char *)) {
    r=data_types[type];
  }else{
    r=obj_unknown;
  }
  return(r);
}

/*-------------------- common_parent --------------------*/

Node *common_parent(Node *a,Node *b)
{
  int lev;
  Node *rs,*rt,*r;

  if (mvld(a)&&mvld(b)) {
    lev=0;
    for (rs=a;mvld(rs->up);lev++) rs=rs->up;
    for (rt=b;mvld(rt->up);lev--) rt=rt->up;
    for (rs=a;lev>0&&mvld(rs->up);rs=rs->up) lev--;
    for (rt=b;lev<0&&mvld(rt->up);rt=rt->up) lev++;
    while (rs!=rt) {
      if (!mvld(rs)||!mvld(rt)) break;
      rs=rs->up;
      rt=rt->up;
    }
    r=(rs==rt ? rs : (Node *)NULL);
  }else{
    r=(Node *)NULL;
  }
  return(r);
}
/*-------------------- common_root --------------------*/

Node *common_root(Nodechain *c)
{
  /* return pointer to common root of nodes in chain c  */
  /* Note: will fail if c is not the only chain or if   */
  /*       nodes in c do not have a common root.        */

  Node *r,*up,*n,*down;
  Nodechain tmp;
  int s;

  r=(Node *)NULL;
  init_node_chain(&tmp);

  if (c && c->s) {
    if (c->n==1) {
      r=c->s;
    }else{
      for (n=c->s;mvld(n);n=n->chain) {
	for (up=n->up;mvld(up);up=up->up) if (!in_chain(up,c)) add_to_node_chain(up,&tmp);
      }
      for (up=c->s;mvld(up);up=up->up) if (!mvld(up->up)) break;
      for (down=(Node *)NULL,r=up,s=1;s==1;r=down) {
	for (s=0,r=r->fst;mvld(r);r=r->nxt) {
	  if (r->chain || r==tmp.e || r==c->e) {
	    down=r;
	    s++;
	    if (s>1) break;
	  }
	}
      }
      r=(down ? down->up : (Node *)NULL);
    }
  }
  clear_node_chain(&tmp);
  if (r) for (up=r->up;mvld(up);up=up->up) if (in_chain(up,c)) r=up;
  return(r);
}
/*-------------------- clear_bond_chain --------------------*/

void clear_bond_chain(Bond *b)
{
  Bond *n;

  if (b) for (n=b->chain;n;b=n,n=b->chain) b->chain=(Bond *)NULL;
}
/*-------------------- SetDefaultAtomColor --------------------*/

int SetDefaultAtomColor(int color)
{
  int r;

  r=defaultAtomColor;
  defaultAtomColor=color;
  return(r);
}

/*-------------------- DefaultAtomColor --------------------*/

int DefaultAtomColor(void)
{
  return(defaultAtomColor);
}

/*-------------------- SetDefaultBondWidth--------------------*/

int SetDefaultBondWidth(int lw)
{
  int r;

  r=defaultBondWidth;
  defaultBondWidth=lw;
  return(r);
}

/*-------------------- DefaultBondWidth --------------------*/

int DefaultBondWidth(void)
{
  return(defaultBondWidth);
}

/*-------------------- data_delete_method --------------------*/

int (*data_delete_method(Data *d,int (*new)(Data *data)))(Data *data)
{
  int (*old)(Data *data);
  old=d->delete;
  d->delete=new;
  return(old);
}

/*-------------------- data_copy_method --------------------*/

Data *(*data_copy_method(Data *d,Data *(*new)(Data *data)))(Data *data)
{
  Data *(*old)(Data *data);
  old=d->copy;
  d->copy=new;
  return(old);
}
/*-------------------- dsp_is_on --------------------*/

int dsp_is_on(Dspobj *d)
{
  int r;

  if (is_dsp(d)) {
    for (r=TRUE;r&&is_dsp(d);d=(Dspobj *)d->up) r=r&&(d->status&OBJ_FLG_DSP);
  }else r=0;
  return(r);
}
