#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mystdlib.h"
#include "parmset.h"
#include "strutil.h"
#include "errhand.h"

#define DEGTORAD 0.0174533 /* somewhat inaccurate, for consistency with LEAP */

static int dihParmSetTypes(PARMSET p,char tp[4][256],int t[4]);
static int angParmSetTypes(PARMSET p,char tp[3][256],int t[3]);

static int sloppy=PRM_SLOPPY_NEVER;
static int misspar=PRM_MISSING_ZERO;

/* Tools to read AMBER paramter definition files */
/*========================================================================*/

PARMSET readParmSet(FILE *in,PARMSET p)
{
  if (in) {
    if (!p.status) {
      p.fnr++;
      if (oldParmFile(in)) p=readOldParmSet(in,p);
      else p=readFrcModSet(in,p);
    }
  }else p.status=1;
  return(p);
}

/*------------------------------------------------------------------------*/


PARMSET emptyParmSet(void)
{
  PARMSET p={
    (double *)NULL,(double *)NULL,(double *)NULL,
    (char **)NULL,
    (char ***)NULL,
    (int *)NULL,(int *)NULL,
    (P_BOND *)NULL,(P_ANGL *)NULL,(P_DIHE *)NULL,(P_IMPR *)NULL,
    (P_HBON *)NULL,
    0,0,0,0,0,0,
    0,
    0,0
  };

  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET clearParmSet(PARMSET p)
{
  int i;

  if (mvld(p.atp)) {
    for (i=0;i<p.nmass;i++) if (mvld(p.atp[i])) free(p.atp[i]);
    free(p.atp);
  }
  p.atp=(char **)NULL;
  if (mvld(p.mass)) {free(p.mass);} p.mass=(double *)NULL; p.nmass=0;
  if (mvld(p.fnb)) {free(p.fnb);} p.fnb=(int *)NULL;
  if (mvld(p.fmass)) {free(p.fmass);} p.fmass=(int *)NULL;
  if (mvld(p.R)) {free(p.R);} p.R=(double *)NULL; 
  if (mvld(p.E)) {free(p.E);} p.E=(double *)NULL; 
  if (mvld(p.bnd)) {free(p.bnd);} p.nbnd=0; p.bnd=(P_BOND *)NULL;
  if (mvld(p.ang)) {free(p.ang);} p.nang=0; p.ang=(P_ANGL *)NULL;
  if (mvld(p.dih)) {free(p.dih);} p.ndih=0; p.dih=(P_DIHE *)NULL;
  if (mvld(p.imp)) {free(p.imp);} p.nimp=0; p.imp=(P_IMPR *)NULL;
  if (mvld(p.hbo)) {free(p.hbo);} p.nhbo=0; p.hbo=(P_HBON *)NULL;
  p.status=0; p.fnr=0;
  p=clearParmSetEquivs(p);

  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET clearParmSetEquivs(PARMSET p)
{
  int i,n;

  if (p.nequiv&&p.equiv) {
    for (n=0;n<p.nequiv;n++) {
      if (p.equiv[n]) {
	for (i=0;p.equiv[n][i];i++) free(p.equiv[n][i]);
	free(p.equiv[n]);
      }
    }
    free(p.equiv);
    p.equiv=(char ***)NULL;
    p.nequiv=0;
  }
  return(p);
}
/*------------------------------------------------------------------------*/

int oldParmFile(FILE *in)
{
  /* determine if it is an old parameter file, or a new frcmod type file */
  /* uses the same method as LEAP, except that retsrictions about the    */
  /* presence or absence of MASS/NBOND sections are ignored.             */
  /* returns FALSE for a new frcmod, TRUE for an old param file.         */
  /* Works by reading through file, and fseeking to byte 0 at end.       */

  int r=TRUE;
  char rec[2048]; /* assumes no line in file is longer than this... */
  for(r=TRUE;r&&fgets(rec,sizeof(rec),in);) {
    if (!strncmp(rec,"MASS",4)||
	!strncmp(rec,"BOND",4)||
	!strncmp(rec,"ANGL",4)||
	!strncmp(rec,"DIHE",4)||
	!strncmp(rec,"IMPR",4)||
	!strncmp(rec,"HBON",4)||
	!strncmp(rec,"NBON",4)) r=FALSE;
  }
  fseek(in,0,SEEK_SET);    /* set file pointer back to start of file */
  return(r);
}
/*------------------------------------------------------------------------*/

PARMSET readOldParmSet(FILE *in,PARMSET p)
{
  int n;
  char rec[1024];            /* hopefully long enough... */
  char fld[256],dummy[256];  /*    "       "      "      */



  if (!p.status) { 
    p=clearParmSetEquivs(p);   /* should be redundant */
    if (!fgets(rec,sizeof(rec),in)) rec[0]='\0'; /* skip over title */
    p=readParmSetMasses(in,p);
    while (fgets(rec,sizeof(rec),in)) {        /* skip some obsolete data... */
      n=sscanf(rec,"%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
	       fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld,fld);
      if (n<20) break;
    }
    p=readParmSetBonds(in,p);
    p=readParmSetAngles(in,p);
    p=readParmSetDihedrals(in,p);
    p=readParmSetImpropers(in,p);
    p=readParmSetHBonds(in,p);
    p=readParmSetEquivs(in,p);
    if (fgets(rec,sizeof(rec),in)) {  /* check non-bonded parameter type, must be RE */
      n=sscanf(rec,"%s %s",dummy,fld);
      if (n==2) {
	if (strcmp(fld,"RE")) {
	  err_set(1,"only RE non-bonded parameter type supported.");
	  err_set(1,"%s",rec);
	}
      }else{
	err_set(1,"expected LABEL,KINDNB record, but got:");
	err_set(1,"%s",rec);
	p.status=1;
      }
    }else{
      err_set(1,"premature EOF in parameter file.");
      p.status=1;
    }
    p=readParmSetNBonds(in,p); 
    p=clearParmSetEquivs(p);
  }
  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET readFrcModSet(FILE *in,PARMSET p)
{
  char rec[1024],fld[256];
  int s=FALSE,n;

  if (!p.status) {
    p=clearParmSetEquivs(p);   /* should be redundant */
    if (!fgets(rec,sizeof(rec),in)) rec[0]='\0'; /* skip over title */
    for (;!p.status&&fgets(rec,sizeof(rec),in);) {
      if (!strncmp(rec,"MASS",4))      p=readParmSetMasses(in,p);
      else if (!strncmp(rec,"BOND",4)) p=readParmSetBonds(in,p);
      else if (!strncmp(rec,"ANGL",4)) p=readParmSetAngles(in,p);
      else if (!strncmp(rec,"DIHE",4)) p=readParmSetDihedrals(in,p);
      else if (!strncmp(rec,"IMPR",4)) p=readParmSetImpropers(in,p);
      else if (!strncmp(rec,"HBON",4)) p=readParmSetHBonds(in,p);
      else if (!strncmp(rec,"NONB",4)) p=readParmSetNBonds(in,p);
      else {
	n=sscanf(rec,"%20s",fld);
	if (n==1) {
	  err_set(1,"unknown keyword in FrcMod parameter file: %s",fld);
	  s=TRUE;
	}
      }
    }
  }
  if (s) p.status=s;
  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET initParmSetMasses(PARMSET p,Nodechain *c)
{
  int i;
  Atom *a;

  if (p.mass) free(p.mass);
  if (p.R) free(p.R);
  if (p.E) free(p.E);
  if (p.fnb) free(p.fnb);
  if (p.fmass) free(p.fmass);
  if (p.atp) {
    for (i=0;i<p.nmass;i++) if (p.atp[i]) free(p.atp[i]);
    free(p.atp);
  }
  if (c->n>0) {
    p.mass=(double *)malloc(c->n*sizeof(double));
    p.R=(double *)malloc(c->n*sizeof(double));
    p.E=(double *)malloc(c->n*sizeof(double));
    p.atp=(char **)malloc(c->n*sizeof(char **));
    p.fnb=(int *)malloc(c->n*sizeof(int));
    p.fmass=(int *)malloc(c->n*sizeof(int));
  }else{
    p.mass=(double *)NULL;
    p.R=(double *)NULL;
    p.E=(double *)NULL;
    p.atp=(char **)NULL;
    p.fnb=(int *)NULL;
    p.fmass=(int *)NULL;
  }    
  if (p.mass&&p.atp&&p.R&&p.E&&p.fnb&&p.fmass) {
    for (i=0,a=(Atom *)c->s;a;a=(Atom *)a->chain,i++) {
      p.atp[i]=strdup(a->atp->name);
      p.mass[i]=-1;
      p.R[i]=-1; 
      p.E[i]=0;
      p.fnb[i]=0;
      p.fmass[i]=0;
    }
    p.nmass=c->n;
  }else {
    err_set(1,"failed to allocate memory for mass parameters");
    p.status=2;
  }
  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET readParmSetMasses(FILE *in,PARMSET p)
{
  char rec[2048],fld[1024];
  double mass,polar;
  int i,n;

  while(!p.status&&fgets(rec,sizeof(rec),in)) {
    n=sscanf(rec,"%s %lf %lf",fld,&mass,&polar);
    if (n==3||n==2) {
      if (mass>=0) {
	for (i=0;i<p.nmass;i++) if (streq(p.atp[i],fld)) break;
	if (i<p.nmass&&p.fmass[i]<=p.fnr) {
	  p.mass[i]=mass;
	  p.fmass[i]=p.fnr;
	}
      }
    }else if (n<=0) break;
  }
  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET readParmSetBonds(FILE *in,PARMSET p)
{
  char rec[2048],tp[2][256],*ft,*tt;
  int i,n,f,t;
  double REQ,RK;

  while(!p.status&&fgets(rec,sizeof(rec),in)) {
    for(i=0;i<9&&rec[i];i++) if (rec[i]=='-') rec[i]=' ';
    n=sscanf(rec,"%s %s %lf %lf",tp[0],tp[1],&RK,&REQ);
    if (n==4) {
      if (strcmp(tp[0],tp[1])<=0) {
	ft=tp[0]; tt=tp[1];
      }else{
	ft=tp[1]; tt=tp[0];
      }
      for (f=0;f<p.nmass;f++) if (streq(p.atp[f],ft)) break;
      if (f<p.nmass) {
	for (t=0;t<p.nmass;t++) if (streq(p.atp[t],tt)) break;
	if (t<p.nmass) {
	  for (i=0;i<p.nbnd;i++) {
	    if (f==p.bnd[i].f&&t==p.bnd[i].t) {
	      if (p.bnd[i].fnr<p.fnr) {
		p.bnd[i].REQ=REQ;
		p.bnd[i].RK=RK;
		p.bnd[i].fnr=p.fnr;
	      }
	      break;           /* no wild cards, must be only match */
	    }
	  }
	}
      }
    }else if (n>0) {
      err_set(1,"incomprehensible bond parameter record:");
      err_set(1,"%s",rec);
      p.status=1;
    }else break;
  }
  return(p);
}

/*------------------------------------------------------------------------*/

PARMSET readParmSetAngles(FILE *in,PARMSET p)
{
  char rec[2048],tp[3][256];
  int i,n,t[3],lev;
  double TEQ,TK;

  while(!p.status&&fgets(rec,sizeof(rec),in)) {
    for(i=0;i<9&&rec[i];i++) if (rec[i]=='-') rec[i]=' ';
    n=sscanf(rec,"%s %s %s %lf %lf",tp[0],tp[1],tp[2],&TK,&TEQ);
    if (n==5) {
      lev=angParmSetTypes(p,tp,t);
      if (!p.status&&lev>=0) {
	for (i=0;i<p.nang;i++) {
	  if ((t[1]<0||t[1]==p.ang[i].m)&&
	      (lev<p.ang[i].lev||(lev==p.ang[i].lev&&p.fnr>p.ang[i].fnr)) &&
	      (((t[0]<0||t[0]==p.ang[i].l)&&(t[2]<0||t[2]==p.ang[i].r))||
	       ((t[0]<0||t[0]==p.ang[i].r)&&(t[2]<0||t[2]==p.ang[i].l)))) {
	    p.ang[i].TEQ=TEQ*DEGTORAD;
	    p.ang[i].TK=TK;
	    p.ang[i].fnr=p.fnr;
	    p.ang[i].lev=lev;
	    if (lev==0) break;
	  }
	}
      }
    }else if (n>=0) {
      err_set(1,"incomprehensible angle parameter record:");
      err_set(1,"%s",rec);
      p.status=1;
    }else break;
  }
  return(p);
}

/*------------------------------------------------------------------------*/

PARMSET readParmSetDihedrals(FILE *in,PARMSET p)
{
  char rec[2048],tp[4][256];
  int i,k,n,t[4],u[4],lev,cont=FALSE;
  double D,K,N,P,absN;

  for (i=0;i<4;i++) u[i]=-1;
  while(!p.status&&fgets(rec,sizeof(rec),in)) {
    for(i=0;i<9&&rec[i];i++) if (rec[i]=='-') rec[i]=' ';
    n=sscanf(rec,"%s %s %s %s %lf %lf %lf %lf",
	     tp[0],tp[1],tp[2],tp[3],&D,&K,&P,&N);
    if (n==8) {
      lev=dihParmSetTypes(p,tp,t);
      if (D<0.00000000001&&D>-0.00000000001) D=1;
      absN=(N<0?-N:N);
      if (cont) {
	for (i=0;i<4;i++) if (u[i]!=t[i]) break;
	if (i<4) {
	  err_set(1,"atom type mismatch in multiple dihedral term:");
	  err_set(1,"%s",rec);
	  p.status=1;
	}
      }
      if (!p.status&&lev>=0) {
	for (k=0;k<p.ndih;k++) {
	  if (
	      ((t[0]<0||t[0]==p.dih[k].i)&&(t[1]<0||t[1]==p.dih[k].j)&&
	       (t[2]<0||t[2]==p.dih[k].k)&&(t[3]<0||t[3]==p.dih[k].l))
	      ||
	      ((t[0]<0||t[0]==p.dih[k].l)&&(t[1]<0||t[1]==p.dih[k].k)&&
	       (t[2]<0||t[2]==p.dih[k].j)&&(t[3]<0||t[3]==p.dih[k].i))
	      ) {
	    if (!cont&&lev==0) {         /* if match exact --> forget all prior defs */
	      p.dih[k].m=0;
	      p.dih[k].exact=p.fnr;
	    }
	    if (p.dih[k].m>0) {          /* don't mix 4 wildcards with other terms */
	      if (lev==4) {
		if (p.dih[k].lev[0]<4) continue;
	      }else{
		if (p.dih[k].lev[0]==4) p.dih[k].m=0;
	      }
	    }
	    for (i=0;i<p.dih[k].m;i++) if (p.dih[k].PN[i]==absN) break;
	    if (i==p.dih[k].m&&
		(p.dih[k].exact==0||p.dih[k].exact==p.fnr)) {
	      if (p.dih[k].m<P_MAX_DIH_MULT) {
		p.dih[k].lev[i]=lev+1;
		p.dih[k].fnr[i]=p.fnr;
		p.dih[k].m++;
		p.dih[k].ok=TRUE;
	      }else{
		err_set(1,"max. dihedral multiplicity (%d) exceeded:");
		err_set(1,"%s",P_MAX_DIH_MULT,rec);
		p.status=1;
	      }
	    }
	    if (p.dih[k].exact==0||p.dih[k].exact==p.fnr) { /* only mix wild with exact from same file */
	      if (lev<p.dih[k].lev[i]||                     /* use first wild from last file  */
		  (lev==p.dih[k].lev[i]&&p.fnr>p.dih[k].fnr[i])) { 
		p.dih[k].lev[i]=lev;
		p.dih[k].fnr[i]=p.fnr;
		p.dih[k].PK[i]=K/D;
		p.dih[k].PN[i]=absN;
		p.dih[k].PHASE[i]=P*DEGTORAD;
	      }
	    }
	    if (lev==0) break;
	  }
	}
	cont=N<0;
	for (i=0;i<4;i++) u[i]=t[i];
      }
    }else if (n>0) {
      err_set(1,"incomprehensible dihedral record:");
      err_set(1,"%s",rec);
      p.status=1;
    }else break;
  }
  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET readParmSetImpropers(FILE *in,PARMSET p)
{
  char rec[2048],tp[4][256];
  int i,j,n,k,t[4],lev;
  double K,N,P;
  int o[6][4] = {
    {1,2,0,3}, {1,3,0,2}, {2,1,0,3}, {2,3,0,1}, {3,1,0,2}, {3,2,0,1}
  };

  /* Note: Priority rules in LEAP are unclear.                      */
  /* Our rules:                                                     */
  /*  - later files overwrite earlier files at same wild-card level */ 
  /*  - match with least wildcards wins.                            */
  /*  - use first acceptable definition.                            */
  /* The central atom is always at position 3, other atoms in any   */
  /* order. We don't allow wildcards at position 3.                 */       

  while(!p.status&&fgets(rec,sizeof(rec),in)) {
    for(i=0;i<9&&rec[i];i++) if (rec[i]=='-') rec[i]=' ';
    k=(strlen(rec)>15 ? 15 : strlen(rec));
    n=sscanf(rec,"%s %s %s %s",tp[0],tp[1],tp[2],tp[3])+
      sscanf(rec+k,"%lf %lf %lf",&K,&P,&N); /* to skip over the optional IDIV field */
    if (n==7) {
      lev=dihParmSetTypes(p,tp,t);
      if (lev>=0) {
	if (t[2]>=0) {
	  for (k=0;k<p.nimp;k++) {
	    if (lev<p.imp[k].lev||(lev==p.imp[k].lev&&p.imp[k].fnr<p.fnr)) {
	      if (t[2]==p.imp[k].t[0]) {   /* pos 3 must match first type */
		for (i=0;i<6;i++) {
		  for (j=0;j<4;j++) if (t[j]>=0&&t[j]!=p.imp[k].t[o[i][j]]) break;
		  if (j==4) break;
		}
		if (i<6) { /* found a match... */
		  p.imp[k].lev=lev;
		  p.imp[k].fnr=p.fnr;
		  p.imp[k].PK=K;
		  p.imp[k].PN=N;
		  p.imp[k].PHASE=P*DEGTORAD;
		  for (j=0;j<4;j++) p.imp[k].o[j]=o[i][j];
		  if (lev==0) break;   /* no wild cards, there can be only one match */
		}
	      }
	    }
	  }
	}else{
	  err_set(1,"wild card in position 3 of improper term not allowed:");
	  err_set(1,"%s",rec);
	  p.status=1;
	}
      }
    }else if (n>0) {
      err_set(1,"incomprehensible improper dihedral parameter record:");
      err_set(1,"%s",rec);
      p.status=1;
    }else break;
  }
  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET readParmSetHBonds(FILE *in,PARMSET p)
{
  char rec[2048],tp[2][256],*ft,*tt;
  int i,n,f,t;
  double A,B;

  /* NOT FUNCTIONAL, p.nhbo always 0 */

  while(!p.status&&fgets(rec,sizeof(rec),in)) {
    for(i=0;i<9&&rec[i];i++) if (rec[i]=='-') rec[i]=' ';
    n=sscanf(rec,"%s %s %lf %lf",tp[0],tp[1],&A,&B);
    if (n==4) {
      if (strcmp(tp[0],tp[1])<=0) {
	ft=tp[0]; tt=tp[1];
      }else{
	ft=tp[1]; tt=tp[2];
      }
      for (f=0;f<p.nmass;f++) if (streq(p.atp[f],ft)) break;
      if (f<p.nmass) {
	for (t=0;t<p.nmass;t++) if (streq(p.atp[t],tt)) break;
	if (t<p.nmass) {
	  for (i=0;i<p.nhbo;i++) {
	    if (f==p.hbo[i].f&&t==p.hbo[i].t) {
	      if (p.hbo[i].fnr<p.fnr) {
		p.hbo[i].A=A;
		p.hbo[i].B=B;
		p.hbo[i].fnr=p.fnr;
	      }
	      break;           /* no wild cards, must be only match */
	    }
	  }
	}
      }
    }else if (n>0) {
      err_set(1,"incomprehensible hbond parameter record:");
      err_set(1,"%s",rec);
      p.status=1;
    }else break;
  }
  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET readParmSetEquivs(FILE *in,PARMSET p)
{
  char rec[2048],fld[21][256],ext[256];
  long pos;
  int i,n,nequiv;
  
  pos=ftell(in);
  nequiv=0;
  while(!p.status&&fgets(rec,sizeof(rec),in)) {  /* just counting... */
    n=sscanf(rec,"%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
	     fld[0],
	     fld[1],fld[2],fld[3],fld[4],fld[5],fld[6],fld[7],fld[8],fld[9],fld[10],
	     fld[11],fld[12],fld[13],fld[14],fld[15],fld[16],fld[17],fld[18],fld[19],fld[20],
	     ext);
    if (n<=0) break;
    else if (n>21) {
      err_set(1,"too many entries in atom equivalence record:");
      err_set(1,"%s",rec);
      p.status=1;
    }else nequiv++;
  }
  p.equiv=(char ***)malloc(nequiv*sizeof(char **));
  if (p.equiv) {
    p.nequiv=nequiv;
    for (i=0;i<p.nequiv;i++) p.equiv[i]=(char **)NULL;
  }else {
    err_set(1,"memory allocation failure in readParmSetEquivs");
    p.status=1;
    p.nequiv=0;
  }
  fseek(in,pos,SEEK_SET);
  nequiv=0;
  while(!p.status&&fgets(rec,sizeof(rec),in)) {  /* now go and store... */
    n=sscanf(rec,"%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
	     fld[0],
	     fld[1],fld[2],fld[3],fld[4],fld[5],fld[6],fld[7],fld[8],fld[9],fld[10],
	     fld[11],fld[12],fld[13],fld[14],fld[15],fld[16],fld[17],fld[18],fld[19],fld[20],
	     ext);
    if (n<=0) break;
    else if (n>21) {
      err_set(1,"too many entries in atom equivalence record:");
      err_set(1,"%s",rec);
      p.status=1;
    }else if (nequiv<p.nequiv) {
      p.equiv[nequiv]=(char **)malloc((n+1)*sizeof(char *));
      if (p.equiv[nequiv]) {
	for (i=0;i<n;i++) p.equiv[nequiv][i]=strdup(fld[i]);
	p.equiv[nequiv][n]=(char *)NULL;
	nequiv++;
      }else{
	err_set(1,"memory allocation failure in readParmSetEquivs.");
	p.status=1;
      }
    }else{
      err_set(1,"too many quivalence lines in AMBER parameter file.");
      p.status=1;
    }
  }
  return(p);
}
/*------------------------------------------------------------------------*/

PARMSET readParmSetNBonds(FILE *in,PARMSET p)
{
  char rec[2048],fld[1024];
  double R,E;
  int i,k,l,n;

  while(!p.status&&fgets(rec,sizeof(rec),in)) {
    n=sscanf(rec,"%s %lf %lf",fld,&R,&E);
    if (n==3) {
      for (i=0;i<p.nmass;i++) {
	if (streq(p.atp[i],fld)) {
	  if (p.fnb[i]<=p.fnr) {
	    p.R[i]=R;
	    p.E[i]=E;
	    p.fnb[i]=p.fnr;
	  }
	  break;
	}
      }
      for (k=0;k<p.nequiv;k++) {
	if (streq(p.equiv[k][0],fld)) {
	  for (l=1;p.equiv[k][l];l++) {
	    for (i=0;i<p.nmass;i++) {
	      if (streq(p.atp[i],p.equiv[k][l])) {
		if (p.fnb[i]<p.fnr) {
		  p.R[i]=R;
		  p.E[i]=E;
		  p.fnb[i]=p.fnr;
		}
		break;
	      }
	    }
	  }
	}
      }
    }else if (n>0) {
      err_set(1,"incomprehensible nonbond parameter record:");
      err_set(1,"%s",rec);
      p.status=1;
    }else break;
  }
  return(p);
}
/*------------------------------------------------------------------------*/

int checkParmSetMode(int mode)
{
  int r;

  r=sloppy;
  if (mode==PRM_SLOPPY_NEVER||mode==PRM_SLOPPY_ONCE||mode==PRM_SLOPPY_ALWAYS) sloppy=mode;
  return(r);
}

/*------------------------------------------------------------------------*/

int missingParmSetMode(int mode)
{
  int r;

  r=misspar;
  if (mode==PRM_MISSING_FIXED||mode==PRM_MISSING_FLOPPY||mode==PRM_MISSING_ZERO) misspar=mode;
  return(r);
}

/*------------------------------------------------------------------------*/

PARMSET reportParmSetMissing(PARMSET p)
{
  /* report energy terms with missin parameters */

  int k;

  if (!p.status) {
    for (k=0;k<p.nmass;k++) {
      if (p.fmass[k]==0) {
	err_set(1,"no mass defined for %s",p.atp[k]);
	p.status=1;
      }
    }
    for (k=0;k<p.nmass;k++) {
      if (p.fnb[k]==0) {
	err_set(1,"no LJ parameters defined for %s",p.atp[k]);
	p.status=1;
      }
    }
    for (k=0;k<p.nbnd;k++) {
      if (p.bnd[k].fnr==0) {
	if (!sloppy) {
	  err_set(1,"no bond stretch parameter defined for %s %s",
		  p.atp[p.bnd[k].f],p.atp[p.bnd[k].t]);
	  p.status=1;
	}else{
	  if (misspar==PRM_MISSING_FIXED||misspar==PRM_MISSING_FLOPPY) {
	    p.bnd[k].REQ=-1;
	    p.bnd[k].RK=500;
	  }
	}
      }
    }
    for (k=0;k<p.nang;k++) {
      if (p.ang[k].fnr==0) {
	if (!sloppy) {
	  err_set(1,"no angle bend parameter defined for %s %s %s",
		  p.atp[p.ang[k].l],p.atp[p.ang[k].m],p.atp[p.ang[k].r]);
	  p.status=1;
	}else{
	  if (misspar==PRM_MISSING_FIXED||misspar==PRM_MISSING_FLOPPY) {
	    p.ang[k].TEQ=-1;
	    p.ang[k].TK=70;
	  }
	}
      }
    }
    for (k=0;k<p.ndih;k++) {
      if (!p.dih[k].ok) {
	if (!sloppy) {
	  err_set(1,"no torsion parameter defined for %s %s %s %s",
		  p.atp[p.dih[k].i],p.atp[p.dih[k].j],
		  p.atp[p.dih[k].k],p.atp[p.dih[k].l]);
	  p.status=1;
	}else{
	  if (misspar==PRM_MISSING_ZERO||misspar==PRM_MISSING_FLOPPY) {
	    p.dih[k].ok=TRUE;
	    p.dih[k].m=0;
	  }else{
	    p.dih[k].ok=TRUE;
	    p.dih[k].m=1;
	    p.dih[k].PN[0]=-1;
	    p.dih[k].PK[0]=5;
	    p.dih[k].PHASE[0]=0.0;
	  }
	}
      }
    }
    for (k=0;k<p.nimp;k++) {
      if (p.imp[k].fnr==0) {
	if (!sloppy) err_set(1,"no improper parameter defined for %s %s %s %s",
			     p.atp[p.imp[k].t[1]],p.atp[p.imp[k].t[2]],
			     p.atp[p.imp[k].t[0]],p.atp[p.imp[k].t[3]]);
	/* don't set p.status, this is just a warning */
	/* set K to zero, and other params to valid values */
	p.imp[k].PK=0.0; p.imp[k].PN=1.0; p.imp[k].PHASE=0;
	p.imp[k].o[0]=0; p.imp[k].o[1]=1; 
	p.imp[k].o[2]=2; p.imp[k].o[3]=3;
      }
    }
    for (k=0;k<p.nhbo;k++) {
      if (!p.hbo[k].fnr) {
	err_set(1,"no hydrogen bond parameters defined for %s %s",
		p.atp[p.hbo[k].f],p.atp[p.hbo[k].t]);
	p.status=1; /* should this be an error or just a warning ? */
      }
    }
  }
  if (sloppy==PRM_SLOPPY_ONCE) checkParmSetMode(PRM_SLOPPY_NEVER);
  return(p);
}
/*------------------------------------------------------------------------*/

static int angParmSetTypes(PARMSET p,char tp[3][256],int t[3])
{
  int i,k,r;

  /* translate typemnemonic to numerical types        */
  /* return number of wildcards ("X" or "?") in tp,   */
  /* or -1 if non-wild type is not known in p.        */

  for (i=0;i<3;i++) t[i]=-2;
  for (r=0,i=0;i<3&&r>=0;i++) {
    if (streq(tp[i],"X")||
	streq(tp[i],"?")) {       /* count wildcards */
      t[i]=-1;
      r++;                       
    }else{                        /* compare with known types */
      for (k=0;k<p.nmass;k++) if (streq(tp[i],p.atp[k])) break;
      if (k<p.nmass) t[i]=k;
      else r=-1;
    }
  }
  return(r);
}

/*------------------------------------------------------------------------*/

static int dihParmSetTypes(PARMSET p,char tp[4][256],int t[4])
{
  int i,k,r;

  /* translate typemnemonic to numerical types        */
  /* return number of wildcards ("X" or "?") in tp,   */
  /* or -1 if non-wild type is not known in p.        */

  for (i=0;i<4;i++) t[i]=-2;
  for (r=0,i=0;i<4&&r>=0;i++) {
    if (streq(tp[i],"X")||
	streq(tp[i],"?")) {       /* count wildcards */
      t[i]=-1;
      r++;                       
    }else{                        /* compare with known types */
      for (k=0;k<p.nmass;k++) if (streq(tp[i],p.atp[k])) break;
      if (k<p.nmass) t[i]=k;
      else r=-1;
    }
  }
  return(r);
}
