/* Note : read_mol is now always called from read_syb, which opened the input
          file and read the first record (to find out if the format is new or old
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>
#include <errno.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "strutil.h"
#include "fileutil.h"
#include "errhand.h"

#define RECSZ 132
#define MXBT 10
#define NO 0
#define YES 1
#define Return(x,y) {err_set(1,"read_mol: %s",y);return(x);}
#define Fgets(x,f) {if (!UCpgets(x,sizeof(x),f)) *x='\0';	\
                    else { \
                      while (strrchr(x,'\n')) *strrchr(x,'\n')='\0'; \
		      for (k_=strlen(x);k_<sizeof(x);k_++) x[k_]=' '; \
                      x[sizeof(x)-1]='\0';} \
                   }

extern Node *LibRoot;

static int b_type[11] = {0,1,2,3,5,4,7,7,0,7,6};

typedef struct TAFFtype {  /* hard-wired mnem/numtype table */
  char *mnem;
  int  numtp;
} TAFFtype;

static TAFFtype taff_type[] =
{
 {"C.3",   1},{"C.2",   2},{"C.1",   4},{"C.ar",  3},{"C.cat",33},
 {"H",    13},{"N.4",  31},{"N.3",   5},{"N.2",   6},{"N.1",   7},
 {"N.ar", 11},{"N.pl3",19},{"N.am", 28},{"O.3",   8},{"O.2",   9},
 {"O.co2",32},{"S.3",  10},{"S.2",  18},{"S.o",  29},{"S.o2", 30},
 {"P.3",  12},{"Br",   14},{"Cl",   15},{"F",    16},{"I",    17},
 {"Lp",   20},{"Na",   21},{"K",    22},{"Ca",   23},{"Li",   24},
 {"Al",   25},{"Du",   26},{"Si",   27},{"Zn",   95},{"B.pl3",96},
 {"B.3",  97},{"Se",   98},{"D",    99},{"Mg",  998},
 {(char *)NULL,0}
};

/**************************************************************************/

int read_mol(UCpipe *in,char *hdr1,Node *node)
     
/*   Read Tripos's MOL file format. Default extension is .MOL, NODE node must
	  be opened by caller. Returns non-zero status iff read fails.     */
     
{
  char rec[134],fld[134];
  Atom *a,*b;
  Atomdef *def,*default_type;
  Node *lib,*ResLib;
  Bond *bnd;
  int i,j,ft,k,k_,nat,nbnd,nfeat,numtp,f,t,fo,to,bt;
  
  /*   1. analyze first header record...  */
  
  if (hdr1) {
    for (i=0;i<strlen(hdr1)&&i<sizeof(rec)-1;i++) {
      if (!hdr1[i]||hdr1[i]=='\n') break;
      rec[i]=hdr1[i];
    }
    for (;i<sizeof(rec)-1;i++) rec[i]=' ';
    rec[sizeof(rec)-1]='\0';
  }else Fgets(rec,in);
  if (*rec=='\0') Return(2,"no atoms.");
  if (strncmp(rec+5,"MOL",3)!=0) Return(2,"bad format (1st header)");
  nat=str2i(strfld(fld,rec,1,4));
  if (nat<0) Return(2,"negative number of atoms.");
  if (!node->name) set_node_name(node,strfld(fld,rec,19,38));
  
  if (nat == 0) Return(0,"zero atoms");
  
  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);

  if (!mvld(ResLib)) {
    Return(-1,"Failed to open default library");
  }
  default_type=find_atomdef_name(ResLib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }

  /*   2. Read atom records...   */
  
  for(j=0;j<nat;j++) {
    Fgets(rec,in);if (*rec=='\0') Return(3,"premature EOF (atoms)");
    a=(Atom *)add_node(node,(char *)NULL,OBJ_ATOM);
    if (!a) Return(4,"failed to allocate atom");
    strfld(fld,rec,36,39);a->name=Strdup(strsqz(fld,fld));
    Strcpy(a->grp,a->name);
    for (i=0;i<3;i++) a->x[i]=str2f(strfld(fld,rec+8+i*9,1,9));
    numtp=str2i(strfld(fld,rec,5,9));
    for (i=0;taff_type[i].mnem;i++) if (numtp==taff_type[i].numtp) break;
    if (taff_type[i].mnem) def=find_atomdef_name(ResLib,taff_type[i].mnem);
    else def=find_atomdef_number(ResLib,numtp);
    if (!mvld(def)) {
      def=(Atomdef *)copy_node((Node *)default_type);
      if (mvld(def)) {
	def->up=ResLib;
	link_node((Node *)def);
	def->numtp=numtp;
	sprintf(rec,"Du.%d",numtp);
	set_node_name((Node *)def,rec);
      }
    }
    set_atom_type(a,def);
    a->nxyz=1;
  }
  
  /*    3. Read bond list ... */
  
  Fgets(rec,in);if (*rec=='\0') Return(6,"premature EOF (bonds)");
  if (strncmp(rec+5,"MOL",3)!=0) Return(7,"bad format of bond header record");
  strncpy(fld,rec,4);fld[4]='\0';
  nbnd=str2i(strfld(fld,rec,1,4));
  if (nbnd<0) Return(7,"negative number of bonds");
  
  lib=ResLib;
  if (!is_lib(lib)) lib=(Node *)NULL;

  a=(Atom *)node->fst;b=a;fo=1;to=1;
  for (i=0;i<nbnd;i++) {
    Fgets(rec,in);if (*rec=='\0') Return(6,"premature EOF (bonds)");
    f=str2i(strfld(fld,rec,5,8)); t=str2i(strfld(fld,rec,9,12));
    bt=str2i(strfld(fld,rec,22,25));
    if (f<1 || f>nat || t <1 || t >nat || t==f) Return(8,"atom numbers in bond record out of range");
    while(fo<f) {a=(Atom *)a->nxt;fo++;} while(fo>f) {a=(Atom *)a->prv;fo--;}
    while(to<t) {b=(Atom *)b->nxt;to++;} while(to>t) {b=(Atom *)b->prv;to--;}
    bnd=add_bond(a,b);
    if (!bnd) Return(9,"failed to allocate bond");
    if (bt>0 && bt <11) set_bt(bnd,b_type[bt]);
    else set_bt(bnd,BT_UNKNOWN);
  }
  
  /*   4.  Read features (currently restricted to charges only)   */
  
  Fgets(rec,in);if (*rec=='\0') Return(10,"no feateure header record");
  if (strncmp(rec+5,"MOL",3)!=0) Return(11,"bad format of faeture header record");
  nfeat=str2i(strfld(fld,rec,1,4));
  
  a=(Atom *)node->fst;
  for(i=0;i<nfeat;i++) {
    Fgets(rec,in);if (*rec=='\0') Return(10,"premature EOF in feature records");
    ft=str2i(strfld(fld,rec,1,4));
    if (ft == 4) {
      if (a==(Atom*)node->fst) {
	if ((int)str2f(strfld(fld,rec,35,42)) != nat) a=(Atom *)NULL;
	if ((int)str2f(strfld(fld,rec,43,50)) != 0) a=(Atom *)NULL;
      }
      for(k=0;k<5 && a;k++) {
	strncpy(fld,rec+4+6*k,6);fld[6]='\0';
	a->q=(float)str2l(strfld(fld,rec+4+6*k,1,6))/10000;
	a=(Atom *)a->nxt;
      }
    }
  }
  
  return(0);
}
