#include <stdio.h>
#include <math.h>
#include "gcalc.h"
#include "swdef.h"
#include "object.h"
#include "mystdlib.h"
#include "minieispack.h"
#include "xtal.h"
#include "errhand.h"
#include "bhtree.h"
#include "structal.h"

#define MXLIST 2048
#define MXTREE 2048

/*-------------------- RmsFit --------------------*/

float RmsFit(int nat,Node *smol,Atom **satoms,Atom **tatoms,mat_4 *Tra)
{
  mat_4 tra;
  float rms;

  /* r.m.s. best-fit smol(satoms) to tatoms, return rmsd, transform 
   in Tra (if not NULL)
  */

  if (Tra) *Tra=ident_4();
  if (nat>0) {
    rms=rms_best_mat(nat,satoms,tatoms,&tra);
    if (rms>=0) {
      cotransform(tra,smol);
      if (Tra) *Tra=tra;
    }
  }else{
    rms=0.0;
  }
  return(rms);
}
/*-------------------- cotransform --------------------*/

void cotransform(mat_4 Tra,Node *mol)
{
  Atom *a;
  Nodechain c;
  Xcn *xcn;
  Dspobj *o;

  /* Apply transformation Tra to node mol */

  o=glblobj(assocobj(mol));
  transform_object(o,&Tra); reset_object(o);
  ForAtoms(a,mol) transform_atom(Tra,a);
  c=tree2chain(mol,OBJ_XCN);
  for (xcn=(Xcn *)c.s;mvld(xcn);xcn=(Xcn *)xcn->nxt) {
    xcn_transform(Tra,xcn);
  }
  clear_node_chain(&c);
}
/*-------------------- rms_dist --------------------*/

float rms_dist(int nat,Atom **a,Atom **b)
{
  /* compute rms-distance between atoms in pair-list a,b */

  float r;
  int i,j;

  r=0;
  for (i=0;i<nat;i++) {
    for (j=0;j<3;j++) r+=(a[i]->x[j]-b[i]->x[j])*(a[i]->x[j]-b[i]->x[j]);
  }
  if (nat>0) r=sqrt(r/nat);
  return(r);
}
/*-------------------- rms_dist_x --------------------*/

float rms_dist_x(int nat,Atom **a,float *x)
{
  /* compute rms-distance between atoms in pair-list a,b */

  float r;
  int i,j;

  r=0;
  for (i=0;i<nat;i++) {
    for (j=0;j<3;j++,x++) r+=(a[i]->x[j]- *x)*(a[i]->x[j]- *x);
  }
  if (nat>0) r=sqrt(r/nat);
  return(r);
}
/*-------------------- rms_best_mat --------------------*/

float rms_best_mat(int n,Atom **s,Atom **t,mat_4 *r)
{
  /* Find best rot/tra r such that rmsd(r*s->x,t->x) --> min     */

  float *x,*p,rms;
  int i,j;

  if (n>0) {
    x=(float *)malloc(3*n*sizeof(float));
    if (mvld(x)) {
      for (p=x,i=0;i<n;i++) for (j=0;j<3;j++) *p++ = t[i]->x[j];
      rms=rms_best_mat_x(n,s,x,r);
      free(x);
    }else{
      rms=-1.0;
    }
  }else{
    rms=0.0;
  }
  return(rms);
}

/*-------------------- rms_best_mat_x --------------------*/

float rms_best_mat_x(int n,Atom **s,float *x,mat_4 *r)
{
  /********************************************************************/
  /* Find best rot/tra r, such that rmsd(r*s->x,x) --> min            */
  /* Return: rmsd , transformation matrix in r                        */
  /*                                                                  */
  /* S.K.Kearsley,                                                    */
  /* On the orthogonal transformation used for structural comparisons */
  /* Acta.Cryst. A45, 208-210, 1989.                                  */
  /********************************************************************/

  int i,j,k,stat;
  vec_3 cx,cy;
  float A[4][4],d[5],e[5];
  float **D,xp[3],xm[3],q[4],rms,sq,*p;

  if (n==1) {
    /* just a translation... */
    *r=ident_4();
    for (j=0;j<3;j++) {
      r->r[j].c[3]=x[j]-s[0]->x[j];
    }
    return(0.0);
  }

  stat=0;
  *r=ident_4();
  rms=-1.0;

  /* compute centroids */

  for (j=0;j<3;j++) {
    cx.c[j]=0.0;
    cy.c[j]=0.0;
  }

  for (p=x,i=0;i<n;i++) {
    for (j=0;j<3;j++) {
      cx.c[j]+= *p++;
      cy.c[j]+=s[i]->x[j];
    }
  }
    
  for (j=0;j<3;j++) {
    cx.c[j]/=n;
    cy.c[j]/=n;
  }

  for (i=0;i<4;i++) for (j=0;j<4;j++) A[i][j]=0.0;

  for (p=x,i=0;i<n;i++) {
    for (j=0;j<3;j++,p++) {
      xp[j]=(*p - cx.c[j]+s[i]->x[j]-cy.c[j]);
      xm[j]=(*p - cx.c[j]-s[i]->x[j]+cy.c[j]);
    }
    A[0][0]+=xm[0]*xm[0]+xm[1]*xm[1]+xm[2]*xm[2];
    A[1][0]+=xp[1]*xm[2]-xm[1]*xp[2];
    A[2][0]+=xm[0]*xp[2]-xp[0]*xm[2];
    A[3][0]+=xp[0]*xm[1]-xm[0]*xp[1];
    A[1][1]+=xp[1]*xp[1]+xp[2]*xp[2]+xm[0]*xm[0];
    A[2][1]+=xm[0]*xm[1]-xp[0]*xp[1];
    A[3][1]+=xm[0]*xm[2]-xp[0]*xp[2];
    A[2][2]+=xp[0]*xp[0]+xp[2]*xp[2]+xm[1]*xm[1];
    A[3][2]+=xm[1]*xm[2]-xp[1]*xp[2];
    A[3][3]+=xp[0]*xp[0]+xp[1]*xp[1]+xm[2]*xm[2];
  }
  for (i=0;i<4;i++) {
    for (j=i+1;j<4;j++) A[i][j]=A[j][i];
  }

  D=EPfmatrix((float *)A,4,4);
  if (D) {
    EPtred2(4,D,d,e,D);
    stat=EPimtql2(4,d,e,D);
    freeEPfmatrix(D);
  }else{
    stat=1;
  }
  
  if (!stat) {
    rms=(d[4] > 0 ? sqrt(d[4]/n) : 0.0 );
    for (sq=0.0,j=0;j<4;j++) {
      q[j]=A[j][3];
      sq+=q[j]*q[j];
    }
    sq=sqrt(sq);
    for (j=0;j<4;j++) q[j]/=sq;
    *r=QtoRot4(q);
    for (j=0;j<3;j++) {
      for (r->r[j].c[3]=cx.c[j],k=0;k<3;k++) r->r[j].c[3]-=r->r[j].c[k]*cy.c[k];
    }
  }
  return(rms);
}
/*-------------------- rms_best_mat_yx --------------------*/

float rms_best_mat_yx(int n,float *y,float *x,mat_4 *r)
{
  /**********************************************************/
  /* Find best rot/tra r, such that rmsd(r*y,x) --> min     */
  /* Return: rmsd , transformation matrix in r              */
  /*                                                        */
  /* Method: S.K.Kearsley, Acta.Cryst. (1989) A45 208-210   */
  /**********************************************************/

  int i,j,k,stat;
  vec_3 cx,cy;
  float A[4][4],d[5],e[5];
  float **D,xp[3],xm[3],q[4],rms,sq,*p,*py;

  if (n==1) {
    /* just a translation... */
    *r=ident_4();
    for (j=0;j<3;j++) {
      r->r[j].c[3]=x[j]-y[j];
    }
    return(0.0);
  }

  stat=0;
  *r=ident_4();
  rms=-1.0;

  /* compute centroids */

  for (j=0;j<3;j++) {
    cx.c[j]=0.0;
    cy.c[j]=0.0;
  }

  for (p=x,py=y,i=0;i<n;i++) {
    for (j=0;j<3;j++) {
      cx.c[j]+= *p++;
      cy.c[j]+= *py++;
    }
  }
    
  for (j=0;j<3;j++) {
    cx.c[j]/=n;
    cy.c[j]/=n;
  }

  for (i=0;i<4;i++) for (j=0;j<4;j++) A[i][j]=0.0;

  for (p=x,py=y,i=0;i<n;i++) {
    for (j=0;j<3;j++,p++,py++) {
      xp[j]=(*p - cx.c[j]+ *py -cy.c[j]);
      xm[j]=(*p - cx.c[j]- *py +cy.c[j]);
    }
    A[0][0]+=xm[0]*xm[0]+xm[1]*xm[1]+xm[2]*xm[2];
    A[1][0]+=xp[1]*xm[2]-xm[1]*xp[2];
    A[2][0]+=xm[0]*xp[2]-xp[0]*xm[2];
    A[3][0]+=xp[0]*xm[1]-xm[0]*xp[1];
    A[1][1]+=xp[1]*xp[1]+xp[2]*xp[2]+xm[0]*xm[0];
    A[2][1]+=xm[0]*xm[1]-xp[0]*xp[1];
    A[3][1]+=xm[0]*xm[2]-xp[0]*xp[2];
    A[2][2]+=xp[0]*xp[0]+xp[2]*xp[2]+xm[1]*xm[1];
    A[3][2]+=xm[1]*xm[2]-xp[1]*xp[2];
    A[3][3]+=xp[0]*xp[0]+xp[1]*xp[1]+xm[2]*xm[2];
  }
  for (i=0;i<4;i++) {
    for (j=i+1;j<4;j++) A[i][j]=A[j][i];
  }

  D=EPfmatrix((float *)A,4,4);
  if (D) {
    EPtred2(4,D,d,e,D);
    stat=EPimtql2(4,d,e,D);
    freeEPfmatrix(D);
  }else{
    stat=1;
  }
  
  if (!stat) {
    rms=(d[4] > 0 ? sqrt(d[4]/n) : 0.0 );
    for (sq=0.0,j=0;j<4;j++) {
      q[j]=A[j][3];
      sq+=q[j]*q[j];
    }
    sq=sqrt(sq);
    for (j=0;j<4;j++) q[j]/=sq;
    *r=QtoRot4(q);
    for (j=0;j<3;j++) {
      for (r->r[j].c[3]=cx.c[j],k=0;k<3;k++) r->r[j].c[3]-=r->r[j].c[k]*cy.c[k];
    }
  }
  return(rms);
}
/*-------------------- RmsMultiFit --------------------*/

float RmsMultiFit(int nat,int nmol,Node **mols,Atom ***atoms)
{
  int i,j,k,I,maxI;
  mat_4 tra;
  Atom *a;
  float rms,*x,*y,*p,s,S=0,A=0,dx;
  Nodechain c;

  if (nat<1 || nmol <2) return(0.0);

  s=1.0/nmol;

  /* fit all molecules to first molecule */

  x=(float *)malloc(3*nat*sizeof(float));
  if (!mvld(x)) return(-1.0);
  y=(float *)malloc(3*nat*sizeof(float)); 
  if (!mvld(y)) {free(x);return(-1);}
  
  for (p=x,i=0;i<nat;i++) {
    for (j=0;j<3;j++) *p++ = atoms[0][i]->x[j];
  }

  for (k=1;k<nmol;k++) {
    rms=rms_best_mat_x(nat,atoms[k],x,&tra);
    ForAtoms(a,mols[k]) transform_atom(tra,a);
  }

  maxI=200;              /* should be ample except for pathology ... */
  for (I=0;I<maxI;I++) {

    /* compute average position */

    for (k=1;k<nmol;k++) {
      for (p=x,i=0;i<nat;i++) {
	for (j=0;j<3;j++) *p++ += atoms[k][i]->x[j];
      }
    }
    for (p=x,j=0;j<3*nat;j++) *p++ *= s;
    if (I>0) {
      for (dx=0,j=0;j<3*nat;j++) dx=(fabs(y[j]-x[j])>dx ? fabs(y[j]-x[j]):dx);
      if (dx<0.0001) break;
    }
    for (j=0;j<3*nat;j++) y[j]=x[j];
    
    /* fit all to average */

    for (A=0,S=0,k=0;k<nmol;k++) {
      rms=rms_best_mat_x(nat,atoms[k],x,&tra);
      A+=rms;
      S+=rms*rms;
      ForAtoms(a,mols[k]) transform_atom(tra,a);
    }
    
    /* test convergence */

    A/=nmol;
    S=sqrt(S/nmol);
    
    for (p=x,i=0;i<nat;i++) {
      for (j=0;j<3;j++) *p++ = atoms[0][i]->x[j];
    }
  }

  /* send the new coordinates to the graphix box */

  err_set(1,"average r.m.s.d. to average structure = %6.3f",A);
  for (k=0;k<nmol;k++) {
    c=tree2chain(mols[k],OBJ_ATOM);
    send_newx(&c);
    clear_node_chain(&c);
  }

  if (mvld(x)) free(x);
  if (mvld(y)) free(y);
  return(S);
  
}
/*-------------------- auto_pairs --------------------*/

int auto_pairs(int m,Atom **A,int n,Atom **B,float maxd,int method)
{
  int npairs;

  switch (method) {
  case AUTO_PAIR_NEAR:
    npairs=auto_near_pairs(m,A,n,B,maxd);
    break;
  case AUTO_PAIR_STABLE:
    npairs=auto_stable_pairs(m,A,n,B,maxd);
    break;
  case AUTO_PAIR_STRUCTAL:
    npairs=auto_structal_pairs(m,A,n,B,maxd);
    break;    
  default:
    err_set(1,"internal error, bad method in auto_pairs.");
    npairs=0;
    break;
  }
  return(npairs);
}
/*-------------------- auto_stable_pairs --------------------*/

int auto_stable_pairs(int m,Atom **A,int n,Atom **B,float maxd)
{
  int npairs,i;
  BHtree *TA,*TB;
  Atom *a,*b;
  float d;
  
  /* match atom b in B with atom a in A if a is closest to b and b is
     closest to a, a stable pair  */

  npairs=0;
  TA=GenerateBHtree(m,A,( m < MXTREE ? 1 : m / MXTREE ));
  TB=GenerateBHtree(n,B,( n < MXTREE ? 1 : n / MXTREE ));
  if (mvld(TA)&&mvld(TB)) {
    for (i=0;i<n&&npairs<n&&npairs<m;i++) {
      if (B[i]->chain) continue;
      a=findBHclosest(TA,B[i]->x,maxd,&d);
      if (is_atom(a)) {
	b=findBHclosest(TB,a->x,d+0.001,&d);
	if (b==B[i]) {
	  b->chain=(Node *)a;
	  npairs++;
	}
      }
    }
    if (mvld(TA)) freeBHtree(TA);
    if (mvld(TB)) freeBHtree(TB);
  }else{
    err_set(1,"failed to generate automatic pair list.");
  }
  return(npairs);
}

/*-------------------- auto_near_pairs --------------------*/

int auto_near_pairs(int m,Atom **A,int n,Atom **B,float maxd)
{
  int npairs,i;
  BHtree *TA;
  Atom *b;
  float dmin;
  
  /* match atoms in B with nearest atom in A                     */
  /* only match atoms with distance less than maxd               */

  npairs=0;
  TA=GenerateBHtree(m,A,( m < MXTREE ? 1 : m / MXTREE ));
  if (mvld(TA)) {
    for (i=0;i<n;i++) {
      b=findBHnearestNeighbour(TA,B[i],maxd,&dmin);
      if (is_atom(b)) {
	B[i]->chain=(Node *)b;
	npairs++;
      }
    }
    freeBHtree(TA);
  }
  return(npairs);
}

/*-------------------- auto_structal_pairs --------------------*/

int auto_structal_pairs(int m,Atom **A,int n,Atom **B,float maxd)
{
  int npairs,i,*p;
  SAlires *TA,*SA;
  mat_4 tra;
  float rms;
  
  if (m>0&&n>0) {
    TA=(SAlires *)malloc(m*sizeof(SAlires));
    SA=(SAlires *)malloc(n*sizeof(SAlires));
    p=(int *)malloc(n*sizeof(int));
    if (mvld(p)&&mvld(TA)&&mvld(SA)) {
      for (i=0;i<m;i++) TA[i].x=A[i]->x;
      for (i=0;i<n;i++) SA[i].x=B[i]->x;
      npairs=algBestStructal(m,A,TA,n,B,SA,maxd,&tra,p,&rms);
      for (i=0;i<npairs;i++) B[p[i]]->chain=(Node *)A[SA[p[i]].pos];
    }else{
      err_set(1,"automatic pairs(structal): memory allocation failed,");
      npairs=0;
    }
    if (mvld(p)) free(p);
    if (mvld(TA)) free(TA);
    if (mvld(SA)) free(SA);
  }else npairs=0;
  return(npairs);
}

/*-------------------- PriaxMat --------------------*/

mat_4 PriaxMat(int n,float **x)
{
  /* return principle axis transformation, eigenvalues in .r[3] */

  int i,j,k;
  float A[3][3],t[3],**D,d[4],e[4],det;
  mat_4 r;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) r.r[i].c[j]=0.0;
    r.r[i].c[i]=1.0;
  }
  if (n>0&&x) {
    for (j=0;j<3;j++) t[j]=0.0;
    for (i=0;i<n;i++) {
      for (j=0;j<3;j++) t[j]+=x[i][j];
    }
    for (j=0;j<3;j++) t[j]/=n;
    for (j=0;j<3;j++) {
      for (k=0;k<3;k++) {
	for (A[j][k]=0,i=0;i<n;i++) A[j][k]+=(x[i][j]-t[j])*(x[i][k]-t[k]);
      }
    }
    D=EPfmatrix((float *)A,3,3);
    if (D) {
      EPtred2(3,D,d,e,D);
      if (!EPimtql2(3,d,e,D)) {
	det=A[0][0]*A[1][1]*A[2][2] +A[0][1]*A[1][2]*A[2][0] +A[1][0]*A[2][1]*A[0][2]
	   -A[2][0]*A[1][1]*A[0][2] -A[1][0]*A[0][1]*A[2][2] -A[0][0]*A[2][1]*A[1][2];
	if (det<0) for (i=0;i<3;i++) A[i][2]=-A[i][2];
	for (i=0;i<3;i++) {
	  for (j=0;j<3;j++) r.r[i].c[j]=A[j][i];
	  for (r.r[i].c[3]=0,j=0;j<3;j++) r.r[i].c[3]-=A[j][i]*t[j];
	  r.r[3].c[i]=(d[i+1]<0 ? 0 : pow(d[i+1],1.0/3.0));
	}
	r.r[3].c[3]=1.0;
      }
      freeEPfmatrix(D);
    }
  }
  return(r);
}

void ApplyTransf(mat_4 R,int n,float **x,float **y)
{
  /* Apply transformation R to x, result in y     */
  /* x,y are pointer arrays to (x,y,x) triples,   */
  /* R is 4x4 matrix, with arbitrary .r[3].       */
  /* x == y is O.K.                               */

  int i,j,k;
  float z[3];

  if (n>0&&x) {
    for (i=0;i<n;i++) {
      for (j=0;j<3;j++) for (z[j]=R.r[j].c[3],k=0;k<3;k++) z[j]+=R.r[j].c[k]*x[i][k];
      for (j=0;j<3;j++) y[i][j]=z[j];
    }
  }
}

/*-------------------- PMofI --------------------*/

mat_4 PMofI(int n,float **x,float *m)
{
  /* Return transform to principal moment of inertia, moments in .r[3] */
  /* x: array of pointers to (x,y,z) coordinates of points             */
  /* m: array of point masses (if NULL: assume all masses = 1)         */

  int i,j,i1,i2;
  float M,A[3][3],t[3],**D,d[4],e[4],det,mass;
  mat_4 r;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) r.r[i].c[j]=0.0;
    r.r[i].c[i]=1.0;
  }
  if (n>0&&x) {
    M=0.0;
    for (j=0;j<3;j++) t[j]=0.0;
    for (i=0;i<n;i++) {
      mass=(m ? m[i] : 1.0);
      for (j=0;j<3;j++) t[j]+=mass*x[i][j];
      M+=mass;
    }
    for (j=0;j<3;j++) t[j]/=M;
    for (j=0;j<3;j++) { 
      i1=(j+1)%3; i2=(j+2)%3;
      A[j][j]=0.0; A[j][i1]=0.0; A[j][i2]=0.0;
      for (i=0;i<n;i++) {
	mass=(m ? m[i] : 1.0);
	A[j][j]+=mass*((x[i][i1]-t[i1])*(x[i][i1]-t[i1])+(x[i][i2]-t[i2])*(x[i][i2]-t[i2]));
	A[j][i1]-=mass*(x[i][j]-t[j])*(x[i][i1]-t[i1]);
	A[j][i2]-=mass*(x[i][j]-t[j])*(x[i][i2]-t[i2]);
      }
    }
    D=EPfmatrix((float *)A,3,3);
    if (D) {
      EPtred2(3,D,d,e,D);
      if (!EPimtql2(3,d,e,D)) {
	EPevsort(3,d,D,1);
	det=A[0][0]*A[1][1]*A[2][2] +A[0][1]*A[1][2]*A[2][0] +A[1][0]*A[2][1]*A[0][2]
	   -A[2][0]*A[1][1]*A[0][2] -A[1][0]*A[0][1]*A[2][2] -A[0][0]*A[2][1]*A[1][2];
	if (det<0) for (i=0;i<3;i++) A[i][2]=-A[i][2];
	for (i=0;i<3;i++) {
	  for (j=0;j<3;j++) r.r[i].c[j]=A[j][i];
	  for (r.r[i].c[3]=0,j=0;j<3;j++) r.r[i].c[3]-=A[j][i]*t[j];
	  r.r[3].c[i]=(d[i+1]<0 ? 0 : d[i+1]);
	}
	r.r[3].c[3]=1.0;
      }
      freeEPfmatrix(D);
    }
  }
  return(r);
}
