#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include "rpcmethods.h"
#include "swdef.h"
#include "object.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "fileutil.h"

/* length of nap time (microseconds) in wait loops */
#define NAPTIME 20000

/* root of molecules added via RPC_LMB             */
#define LMBROOT "a_LMB"

/* will RPC_DA delete, or just undisplay globals?  */
#define DA_DELETES  1

/* RPC types: */
#define RPC_DELETE       -2
#define RPC_DONE         -1
#define RPC_NOP           0
#define RPC_CMD           3     /* send command block to parser            */
#define RPC_LMB           4     /* load mol block (MDL)                    */
#define RPC_DA            5     /* disable all ( displ global off /a_LMB)  */
#define RPC_ZOOM          6     /* zoom for SDView                         */
#define RPC_LDPDB         7     /* load pdb from data block                */
#define RPC_RDMOL         8     /* load mol types from data block          */
#define RPC_SHUTDOWN      9     /* shutdown RPC server                     */
#define RPC_PMT          10     /* get current prompt from parser          */

extern int graphix;
extern Node *MolRoot;

#ifdef withRPC

/*========================================================================*/

/* some utilities to handle the list of open RPC calls */

/*========================================================================*/

/* Open RPC calls will be kept in a doubly linked list of call structures: */

typedef struct RPCcall_t {
  struct RPCcall_t *prv,*nxt;    /* list pointers                                  */
  void *data;                    /* RPC data                                       */
  xmlrpc_value *cmdResponse;     /* the response to be sent back to the client     */
  xmlrpc_env   *cmdEnvP;         /* the xmlrpc-c environment variable (pointer to) */
  char *prompt;                  /* Wit!P's prompt string at RPC completion        */
  int  opCode;                   /* the RPC time                                   */
  int accResponse;               /* true, if prtGraphics output will be collected  */            
  int cmdErr;                    /* number of Wit!P errors during RPC execution    */
  int level;                     /* parser IOS level when RPC was received         */
  int seqnr;                     /* sequential number of RPC                       */
} RPCcall_t;

static int seqNR=0;              /* total number of RPC's received so far          */
static int nCalls=0;             /* number of open RPC's in linked list            */
static RPCcall_t *Head=NULL;     /* pointer to RPC at head of list                 */
static RPCcall_t *Tail=NULL;     /* pointer to RPC at tail of list                 */
static int dbgRPC=FALSE;         /* true if RPC list should be monitored (debug)   */
static int blocked=FALSE;        /* true if handleRPC is blocked                   */

typedef struct dataBlock_t {
  char *block;
  char *ptr;
  int size;
} dataBlock_t;

static dataBlock_t *dataBlock=NULL;

/*========================================================================*/

static void dmpRPCcall(RPCcall_t *r)
{
  /* Write info about an RPC to stderr (for debugging purposes) */

  if (r) {
    wnpRPCmutexLock();
    {
      fprintf(stderr,">>> RPC Call #%d: opCode=%d, level=%d(%d), ncalls=%d\n",
	      r->seqnr,r->opCode,r->level,parser_level(),nCalls);
      if (r->data) {
	fprintf(stderr,"%s",(char *)r->data);
      }
      fprintf(stderr,"\n>>> ---------------\n");
    }
    wnpRPCmutexUnlock();
  }else{
    fprintf(stderr,">>> RPC Call: (null)\n");
  }
}

/*========================================================================*/

static RPCcall_t *newRPCcall(void)
{
  RPCcall_t *r=NULL,*rnxt;

  /* run garbage collection (each time?) */
  for (r=Head;r;r=rnxt) {
    rnxt=r->nxt;
  }

  /* add a new call structure to the end of the list */

  if (nCalls<wnpRPCmaxCalls()) {
    r=malloc(sizeof(RPCcall_t));
    if (r) {
      wnpRPCmutexLock();
      {
	r->data=NULL;
	r->prompt=NULL;
	r->opCode=RPC_NOP;
	r->accResponse=FALSE;
	r->cmdEnvP=NULL;
	r->cmdResponse=NULL;
	r->cmdErr=-1;
	r->level=parser_level();
	r->prv=NULL;
	r->nxt=NULL;
	r->seqnr=++seqNR;
	if (!Tail) {
	  Head=Tail=r;
	}else{
	  r->prv=Tail;
	  r->prv->nxt=r;
	  Tail=r;
	}
	nCalls++;
      }
      wnpRPCmutexUnlock();
    }
  }
  if (dbgRPC) dmpRPCcall(r);
  return(r);
}

/*========================================================================*/

static RPCcall_t *deleteRPCcall(RPCcall_t *r)
{
  if (r) {
    wnpRPCmutexLock();
    {
      if (r==Tail) Tail=r->prv;
      if (r==Head) Head=r->nxt;
      if (r->prv) r->prv->nxt=r->nxt;
      if (r->nxt) r->nxt->prv=r->prv;
      if (r->prompt) free(r->prompt);
      free(r); r=NULL;
      nCalls--;
    }
    wnpRPCmutexUnlock();
  }
  
  return(r);
}

/*========================================================================*/

dataBlock_t *newBlock(int size)
{
  dataBlock_t *B;

  B=malloc(sizeof(dataBlock_t));
  if (B) {
    B->size=size;
    B->block=malloc(B->size+1);
    if (B->block) B->ptr=B->block;
    else{
      free(B);
      B=NULL;
    }
  }
  return(B);
}

dataBlock_t *deleteBlock(dataBlock_t *B)
{
  if (B) {
    if (B->block) free(B->block);
    free(B);
  }
  return(NULL);
}

int addToDataBlock(dataBlock_t *B,char *s)
{
  int a=0,l;

  if (B) {
    a=B->size-(B->ptr-B->block);
    if (a>0) {
      l=strlen(s);
      a=a>l?l:a;
      if (a>0) {
	strncpy(B->ptr,s,a);
	B->ptr+=a;
	*B->ptr='\0';
      }
    }
  }
  return(a);
}

/*========================================================================*/

/* Methods to support SDView  */

/*========================================================================*/

static xmlrpc_value *wnp_ping(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Internal:        execute immediately
     Input:           none
     Response         1 (integer)
  */
  return xmlrpc_build_value(envP, "i", 1);
}

/*========================================================================*/

static xmlrpc_value *wnp_shutdown(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Handled by main thread from RPC list. Shutdwon RPC server
     Input:  nothing
     Response: ""
  */

  RPCcall_t *rpcCall=NULL;;

  rpcCall=newRPCcall();
  if (rpcCall) {
    rpcCall->data=NULL;
    rpcCall->opCode=RPC_SHUTDOWN; 
  }else{
    xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			 "Sorry, too many pending RPC calls.");
  }
  return xmlrpc_build_value(envP, "s", "");
}

/*========================================================================*/

static xmlrpc_value *wnp_do(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Handled by main thread from RPC list.
     Input:  any PyMOL command, even if we handle only a small subset
     Response: ""
  */

  char *op;
  RPCcall_t *rpcCall=NULL;;

  rpcCall=newRPCcall();
  if (rpcCall) {
    xmlrpc_decompose_value(envP, paramArrayP, "(s)", &op);  
    if (!envP->fault_occurred) {
      if (streq(op,"disable all")) {
	rpcCall->data=NULL;
	rpcCall->opCode=RPC_DA; 
	while (rpcCall->opCode==RPC_DA) {
	  uSleep(NAPTIME);                /* could we use pthread_cond_wait instead ? */
	}
	free(op);
      }
    }
    deleteRPCcall(rpcCall);
  }else{
    xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			 "Sorry, too many pending RPC calls.");
  }
  return xmlrpc_build_value(envP, "s", "");
}

/*========================================================================*/

static xmlrpc_value *wnp_zoom(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Handled by main thread from RPC list.
     Input:  Greg's zoom command, even if we handle only the 'visible' part.
     Response: ""
  */

  char *op;
  RPCcall_t *rpcCall=NULL;;

  rpcCall=newRPCcall();
  if (rpcCall) {
    xmlrpc_decompose_value(envP, paramArrayP, "(s)", &op);  
    if (!envP->fault_occurred) {
      if (streq(op,"visible")) {
	rpcCall->data=NULL;
	rpcCall->opCode=RPC_ZOOM; 
	while (rpcCall->opCode==RPC_ZOOM) {
	  uSleep(NAPTIME);                /* could we use pthread_cond_wait instead ? */
	}
	free(op);
      }    
      deleteRPCcall(rpcCall);
    }else{
      xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			   "Sorry, too many pending RPC calls.");
    }    
  }
  return xmlrpc_build_value(envP, "s", "");
}

/*========================================================================*/

static xmlrpc_value *wnp_loadMolBlock(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Handled by main thread from RPC list.
     Input:  molBlock (MDL Mol file)
             molname  (string)
     Response: ""
   */

  char *molBlock,*name;
  RPCcall_t *rpcCall=NULL;
  xmlrpc_value *R=NULL;

  rpcCall=newRPCcall();
  if (rpcCall) {
    xmlrpc_decompose_value(envP, paramArrayP, "(ss)", &molBlock, &name);
    if (!envP->fault_occurred) {
      if (molBlock) {
	rpcCall->data=molBlock;
	rpcCall->opCode=RPC_LMB;         /* add new op code if called by wnp.loadMolBlock */
 	
	while (rpcCall->opCode==RPC_LMB) {
	  uSleep(NAPTIME);                /* could we use pthread_cond_wait instead ? */

	}
	free(molBlock);
	free(name);
      }
      if (rpcCall->opCode==RPC_DONE) R=xmlrpc_build_value(envP, "s", "");
    }
    deleteRPCcall(rpcCall);
  }else{
    xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			 "Sorry, too many pending RPC calls.");
  }
  return(R);
}

/*========================================================================*/

/* Here are our own methods */

/*========================================================================*/

static xmlrpc_value *wnp_getenv(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Internal:        execute immediately
     Input:           env. varibale name (string)
     Response         value of env. variblae (string)
  */
  char *name,*value;
  xmlrpc_value *R=NULL;

  xmlrpc_decompose_value(envP, paramArrayP, "(s)", &name);  
  if (!envP->fault_occurred) {
    if ((value=getenv(name))) R=xmlrpc_build_value(envP, "s",value);
    else R=xmlrpc_build_value(envP, "s","");
    free(name);
  }else{
    xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			 "Sorry, too many pending RPC calls.");
  }
  return (R);
}

static xmlrpc_value *wnp_getprompt(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Handled by main thread from RPC list.       
     Input:           none
     Response:        current prompt string
  */
  xmlrpc_value *R=NULL;
  RPCcall_t *rpcCall=NULL;

  rpcCall=newRPCcall();
  if (rpcCall) {
    rpcCall->cmdErr=-1;
    rpcCall->cmdEnvP=envP;
    rpcCall->opCode=RPC_PMT;
    while (rpcCall->opCode==RPC_PMT) {
      uSleep(NAPTIME);                /* could we use pthread_cond_wait instead ? */
    }
    R=xmlrpc_build_value(rpcCall->cmdEnvP, "s",rpcCall->prompt?rpcCall->prompt:"");
    deleteRPCcall(rpcCall);
  }else{
    xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			 "Sorry, too many pending RPC calls.");
  }
  return(R);
}

static xmlrpc_value *wnp_cmd(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Handled by main thread from RPC list.
     Input:  text block to be sent to Wit!P parser
     Response: prtGraphics output and error messages, prompt at end, number of errors
   */

  char *cmd=NULL;
  int nerr=0;
  xmlrpc_value *R=NULL;
  RPCcall_t *rpcCall=NULL;

  rpcCall=newRPCcall();
  if (rpcCall) {
    xmlrpc_decompose_value(envP, paramArrayP, "(s)", &cmd);
    if (!(envP->fault_occurred)) {
      if (cmd) {
	rpcCall->cmdErr=-1;
	rpcCall->cmdResponse=xmlrpc_array_new(envP);
	if (!envP->fault_occurred) {
	  rpcCall->cmdEnvP=envP;
	  rpcCall->data=cmd;
	  rpcCall->opCode=RPC_CMD;
	  while (rpcCall->opCode==RPC_CMD) {
	    uSleep(NAPTIME);                /* could we use pthread_cond_wait instead ? */
	  }
	  nerr=rpcCall->opCode==RPC_DONE?0:-2;
	}else nerr=-1;
	free(cmd);
      }
      R=xmlrpc_build_value(rpcCall->cmdEnvP, "(isA)", 
			   nerr?nerr:rpcCall->cmdErr, rpcCall->prompt?rpcCall->prompt:"", 
			   rpcCall->cmdResponse);
      xmlrpc_DECREF(rpcCall->cmdResponse);
    }
    deleteRPCcall(rpcCall);
  }else{
    xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			 "Sorry, too many pending RPC calls.");
  }
  return (R);
}  


static xmlrpc_value *wnp_loadPDB(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Handled by main thread from RPC list.
     Input:  molname  (string)
     Response: ""
   */

  /* load the current content of 'dataBlock' as a PDB */

  RPCcall_t *rpcCall=NULL;
  xmlrpc_value *R=NULL;
  char *name;

  rpcCall=newRPCcall();
  if (rpcCall) {
    xmlrpc_decompose_value(envP, paramArrayP, "(s)", &name);
    if (!envP->fault_occurred) {
      rpcCall->data=name;
      rpcCall->opCode=RPC_LDPDB;         /* add new op code if called by wnp.loadMolBlock */
 	
      while (rpcCall->opCode==RPC_LDPDB) {
	uSleep(NAPTIME);                /* could we use pthread_cond_wait instead ? */
      }
      free(name);
      if (rpcCall->opCode==RPC_DONE) R=xmlrpc_build_value(envP, "i", rpcCall->cmdErr);
    }
    deleteRPCcall(rpcCall);
  }else{
    xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			 "Sorry, too many pending RPC calls.");
  }
  return(R);
}

static xmlrpc_value *wnp_loadMOLtypes(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Handled by main thread from RPC list.
     Input:  molname  (string)
     Response: ""
   */

  /* load the current content of 'dataBlock' as a name:type  */

  RPCcall_t *rpcCall=NULL;
  xmlrpc_value *R=NULL;
  char *name;
  int nerr=0;

  rpcCall=newRPCcall();
  if (rpcCall) {
    xmlrpc_decompose_value(envP, paramArrayP, "(s)", &name);
    if (!envP->fault_occurred) {
      rpcCall->data=name;
      rpcCall->cmdEnvP=envP;
      rpcCall->opCode=RPC_RDMOL;        /* add new op code if called by wnp.loadMolBlock */
      rpcCall->cmdResponse=xmlrpc_array_new(envP);

      while (rpcCall->opCode==RPC_RDMOL) {
	uSleep(NAPTIME);                /* could we use pthread_cond_wait instead ? */
      }
      nerr=rpcCall->opCode==RPC_DONE?0:-2;
      free(name);
      if (rpcCall->opCode==RPC_DONE) { 
	R=xmlrpc_build_value(envP, "(isA)", nerr?nerr:rpcCall->cmdErr, 
			     rpcCall->prompt?rpcCall->prompt:"",
			     rpcCall->cmdResponse);
	xmlrpc_DECREF(rpcCall->cmdResponse);
      }
    }
    deleteRPCcall(rpcCall);
  }else{
    xmlrpc_env_set_fault(envP,XMLRPC_LIMIT_EXCEEDED_ERROR,
			 "Sorry, too many pending RPC calls.");
  }
  return(R);
}

static xmlrpc_value *wnp_dataBlock(xmlrpc_env *envP,xmlrpc_value *paramArrayP,void *serverContext)
{
  /* Internal: executed immediately.
     Input:  size in characters of data block to create or zero to append to current block (int)
             data, first data block (string)
     Response: number of bytes added (int)
   */
  
  /* add text data to 'dataBlock. If size>0: start a new block of given size, if size < 0: delete current block */

  int size,s;
  char *data;
  xmlrpc_value *R=NULL;

  xmlrpc_decompose_value(envP, paramArrayP, "(is)", &size, &data);
  if (!envP->fault_occurred) {
    if (size>0) {
      deleteBlock(dataBlock);
      dataBlock=newBlock(size);
    }else if (size<0) {
      dataBlock=deleteBlock(dataBlock);
    }
    s=addToDataBlock(dataBlock,data);
    free(data);
    R=xmlrpc_build_value(envP, "i", s);
  }
  return(R);
}

/*========================================================================*/

/* Register RPC methods */

/*========================================================================*/


void register_rpcmethods(xmlrpc_env *env,xmlrpc_registry *registryP)
{
  if (env && registryP) {
    /* methods to support SDView */
    xmlrpc_registry_add_method(env,registryP,NULL,"loadMolBlock",wnp_loadMolBlock,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"ping",wnp_ping,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"do",wnp_do,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"zoom",wnp_zoom,NULL);
    /* our own methods */
    xmlrpc_registry_add_method(env,registryP,NULL,"wnp.cmd",wnp_cmd,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"wnp.getenv",wnp_getenv,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"wnp.loadPDB",wnp_loadPDB,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"wnp.loadMOLtypes",wnp_loadMOLtypes,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"wnp.dataBlock",wnp_dataBlock,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"wnp.shutdown",wnp_shutdown,NULL);
    xmlrpc_registry_add_method(env,registryP,NULL,"wnp.getprompt",wnp_getprompt,NULL);
  }
}

/**************************************************************************/
/**************************************************************************/
/************************ main thread code below... ***********************/
/**************************************************************************/
/**************************************************************************/

/*========================================================================*/

static RPCcall_t *selectCurrent()
{
  RPCcall_t *p,*r=NULL;
  int level=-1;

  /* select 'current' RPC from list: highest level, first-come first-served, */

  wnpRPCmutexLock();
  {
    for (r=NULL,p=Head;p;p=p->nxt) {
      if (p->opCode>RPC_NOP&&p->level>level) {
	level=p->level;
	r=p;
      }
    }
  }
  wnpRPCmutexUnlock();
  return(r);
}

/*========================================================================*/

void wnpRPCclose(pthread_t thread)
{
  RPCcall_t *p;
  int i;

  if (nCalls>0) {
    wnpRPCmutexLock();
    {
      /* setting opCode to RPC_NOP will cause RPC to terminate in error */
      for (p=Head;p;p=p->nxt) p->opCode=RPC_NOP;
    }
    wnpRPCmutexUnlock();

    /* give the open RPC calls a chance to terminate */
    for (i=0;Head&&i<100;i++) if (Head) usleep(NAPTIME);

    /* all calls terminated, or grace time is up: cancel the server thread */
    if (thread) pthread_cancel(thread);

    /* remove any remaining RPC call structures */
    while (Tail) deleteRPCcall(Tail);

  }else  if (thread) pthread_cancel(thread);
  
  /* leave WNP_RPC_PORT undefined */
  Setenv("WNP_RPC_PORT","");
}

#endif

/*========================================================================*/

void wnpRPCprintResponse(int mode,char *msg)
{
  /* add message to RPC response (if wnp.cmd method is running)        */
  /* mode: 0 for messgae from prtGraphics, 1 for message from err_dump */

#ifdef withRPC

  xmlrpc_value *string;
  RPCcall_t *currRPC;

  currRPC=selectCurrent();

  if (currRPC&&currRPC->accResponse&&currRPC->cmdResponse) {
    string=xmlrpc_string_new(currRPC->cmdEnvP,msg);
    if (!currRPC->cmdEnvP->fault_occurred) 
      xmlrpc_array_append_item(currRPC->cmdEnvP,currRPC->cmdResponse,string);
    if (string) xmlrpc_DECREF(string);
  }

#endif
}

/*========================================================================*/

void blockRPC(void)
{
#ifdef withRPC
  blocked=TRUE;
#endif
}

void releaseRPC(void)
{
#ifdef withRPC
  blocked=FALSE;
#endif
}

int handleRPC(char *buff,int buffsz)
{

  /* check for pending RPC requests, and handle them, if there are any.     */
  /* return 1 if request entered data into parse input stream (either via   */
  /* Qmsg, if buff is NULL, or via character buffer buff.                   */

  int status=0;

#ifdef withRPC

  char *p,buffer[8196],*name;
  Qmsg *msg=NULL;
  int l,ft;
  Node *node,*root=NULL;
  Atom *a;
  Mol *mol;
  Dspobj *glbl;
  Nodechain C;
  RPCcall_t *currRPC;
  UCpipe *B;

  currRPC=selectCurrent();

  /*** to do: check parser level only for OPs that need to execute at 'acceptable' level ***/
  /* ignore it, if we ar not at the correct level of parsing */
  if (!currRPC || currRPC->level<parser_level() || blocked)  return(0);

  switch(currRPC->opCode) {

  case RPC_NOP:
    break;

  case RPC_PMT:
    if (currRPC->prompt) free(currRPC->prompt);
    currRPC->prompt=strdup(get_prompt(buffer));
    currRPC->opCode=RPC_DONE;
    break;

  case RPC_CMD:
    if (currRPC->cmdErr<0) currRPC->cmdErr=err_seqnr();
    if ( currRPC->data && *(char *)currRPC->data ) {
      p=strchr(currRPC->data,'\n');
      l=p?(p-(char *)currRPC->data)+1:strlen(currRPC->data);
      if (!buff) {
	msg=qmsg("KEYB",l,currRPC->data);
	if (msg) status=1;
      }else{
	l=l<buffsz?l:buffsz-1;
	strncpy(buff,currRPC->data,l);
	buff[l]='\0';
	status=1;
      }
      currRPC->accResponse=TRUE;
      if (p) currRPC->data=p+1;
      else currRPC->data=NULL;
    }else{
      if (currRPC->prompt) free(currRPC->prompt);
      currRPC->prompt=strdup(get_prompt(buffer));
      currRPC->cmdErr=err_seqnr()-currRPC->cmdErr;
      currRPC->accResponse=FALSE;
      currRPC->opCode=RPC_DONE;
    }
    break;

  case RPC_LMB:
    B=UCpblock(currRPC->data);
    if (B) {
      node=make_node(OBJ_MOL);
      currRPC->cmdErr=read_maccs_mol(B,node);
      if (node && currRPC->cmdErr) node=del_node(node);
      UCpclose(B);
    }else node=0;
    if (node) {
      root=find_name(MolRoot,LMBROOT,OBJ_MOL);
      if (!root) {
	root=make_node(OBJ_MOL);
	set_node_name(root,LMBROOT);
	root->up=MolRoot;
	link_node(root);
	send_molecule(MolRoot,(Mol *)root);
      }
      node->up=(root?root:MolRoot);
      link_node(node);
      send_molecule(root?root:MolRoot,(Mol *)node);
    }
    currRPC->data=NULL;
    currRPC->opCode=RPC_DONE;     
    break;

  case RPC_LDPDB:         /* obsolete... (may still be used by old scripts) */
    if (dataBlock) {
      B=UCpblock(dataBlock->block);
      if (B) {
	root=add_node(MolRoot,currRPC->data,OBJ_MOL);
	currRPC->cmdErr=read_pdb(B,NULL,root,FT_PDB);
	condAutocenterMol(root);
	condAutobondMol(root);
	send_molecule(MolRoot,(Mol *)root);
	UCpclose(B);
      }
      dataBlock=deleteBlock(dataBlock);
    }
    currRPC->data=NULL;
    currRPC->opCode=RPC_DONE;     
    break;

  case RPC_RDMOL:
    err_dump();
    currRPC->accResponse=TRUE;
    if (currRPC->cmdErr<0) currRPC->cmdErr=err_seqnr();
    if (dataBlock) {
      B=UCpblock(dataBlock->block);
      if (B) {
	if ((p=strchr(currRPC->data,':'))) {
	  *p='\0';
	  ft=ft_numberKW(currRPC->data);
	  *p=':';
	  name=p+1;
	}else{
	  ft=FT_NONE;
	  name="";
	}
	if (ft==FT_MACCS) {
	  root=make_node(OBJ_MOL);       /* maccs */
	  if (root) read_maccs_mol(B,root);
	}else if (ft==FT_CSD) {
	  root=make_node(OBJ_MOL);       /* CSD */
	  if (root) read_csd_mol(B,root);
	}else if (ft&(FT_MOL|FT_MOL2|FT_CIF|FT_PDB|FT_XPDB|FT_SHX|FT_MMOD|FT_DIANA|FT_PAR|FT_MDL)) {
	  root=readmoltypes(B,NULL,ft);  /* mol, pdb, shelx, mmod, CIF, diana */
	}else{
	  root=NULL;
	}
	if (root) {
	  set_node_name(root,name);
	  root->up=MolRoot;
	  link_node(root);
	  condAutocenterMol(root);
	  if (ft&(FT_CIF|FT_PDB|FT_XPDB|FT_DIANA|FT_PAR)) condAutobondMol(root);
	  send_molecule(MolRoot,(Mol *)root);
	}else{

	}
	UCpclose(B);
      }
      dataBlock=deleteBlock(dataBlock);
    }
    err_dump();
    currRPC->data=NULL;
    if (currRPC->prompt) free(currRPC->prompt);
    currRPC->prompt=strdup(get_prompt(buffer));
    currRPC->cmdErr=err_seqnr()-currRPC->cmdErr;
    currRPC->accResponse=FALSE;
    currRPC->opCode=RPC_DONE;
    
    break;

  case RPC_DA:
    root=find_name(MolRoot,LMBROOT,OBJ_MOL);
    if (root) {
      if (DA_DELETES) {
	while (root->fst) del_node(root->fst);
      }else{
	glbl=glblobj(assocobj((Node *)mol));
	if (glbl) {
	  glbl->status&=~OBJ_FLG_DSP;
	  if (cPtr(glbl->cross)) {
	    send_bheader(graphix,"DSPO",sizeof(int)+sizeof(mol));
	    send_bint(graphix,0);
	    send_bptr(graphix,glbl);
	  }
	}
      }
    }
    currRPC->opCode=RPC_DONE;  
    break;

  case RPC_ZOOM:
    root=find_name(MolRoot,LMBROOT,OBJ_MOL);
    if (root) {
      init_node_chain(&C);
      forMols(mol,root) {
	glbl=glblobj(assocobj((Node *)mol));
	if (glbl&&glbl->status&OBJ_FLG_DSP) {
	  ForAtoms(a,mol) add_to_node_chain((Node *)a,&C);
	}
      }
      glbl=glblobj(assocobj(MolRoot));
      if (glbl) displayFind(glbl,&C);
      clear_node_chain(&C);
    }
    currRPC->opCode=RPC_DONE;  
    break;

  case RPC_SHUTDOWN:         /* special, XML/RPC thread sends response before execution... */
    wnpRPCshutdown();        /* shutdown server, currRPC will be deleted by wnpRPCclose()  */
    send_header(graphix,"EXIT",0); /* shutdown graphics server, if it is running... */
    graphics_down("Closed connection to graphics server.");
    exit(0);                 /* sorry, no other way out (?) */

  case RPC_DONE:
    break;

  default:
    currRPC->opCode=RPC_DONE;
    break;
  }

#endif

  return(status);
}

