#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "salign.h"
#include "align.h"

#define min(a,b)    ( (a)<=(b) ? (a) : (b) )
#define upperc(x)   ( (x) >= 'A' && (x) <= 'Z' )
#define tolowerc(x) ( (x) - 'A' + 'a' )

#define TRUE 1

static float gersteinGapPenaltyFactor = 0.5;
static float gerstXcompare5Factor=1.5;

static float compareStruct(void *data,int i,int j);
typedef  struct { 
  float *dA,*dB,AV; 
  float (*func)(int i,int j,float *dA,float *dB);
} structData_t ;


static float compareGerst(void *data,int i,int j);
typedef struct {
  SAlires *A,*B;
  float *dA,*dB;
} gerstData_t;

/*------------------------------------------------------------------------*/

int salign(int m,SAlires *A,int n,SAlires *B,
	   float (*compare)(int i,int j,float *dA,float *dB),
	   float gapo,float gape,float *score)
{

  /* Structural Alignment: finds optimal sequence alignment via
     dynamic programming using the following scoring parameters:

     matched pairs (i,j):    AV - compare(i,j)
     gap open penalty:       AV * gapo
     gap extension penalty:  AV * gape

     where AV is the average value of compare(i,j), averaged over 
     all m*n pairs (i,j), and compare is a function which measures
     differences between backbone geometries in the vicinity of
     residues i and j (e.g. ResCompare5, which computes the r.m.s.
     of differences between the 6 pairs of Calpha-Calpha distances

        dist(i-2,i+2) in A and dist(i-2,i+2) in B,
        dist(i-2,i  ) in A and dist(i-2,i  ) in B,
        dist(i-2,i+1) in A and dist(i-2,i+1) in B,
        dist(i-1,i+1) in A and dist(i-1,i+1) in B,
        dist(i-1,i+2) in A and dist(i-1,i+2) in B,
        dist(i  ,i+2) in A and dist(i  ,i+2) in B).

  */

  structData_t data;

  float *dA,*dB,AV;
  int i,j,status,*matchA,*matchB;

  if (n<1||m<1) return(1);

  if (compare==ResCompare5) {
    ResCompareSetup5(m,A,n,B,&dA,&dB);
    if (!dA||!dB) {
      if (dA) free(dA);
      if (dB) free(dB);
      return(2);
    }
  }else if (compare==ResCompare6) {
    ResCompareSetup6(m,A,n,B,&dA,&dB);
    if (!dA||!dB) {
      if (dA) free(dA);
      if (dB) free(dB);
      return(2);
    }
  }else{
    dA=(float *)NULL;
    dB=(float *)NULL;
  }

  matchA=(int *)malloc(m*sizeof(int));
  matchB=(int *)malloc(n*sizeof(int));

  if (!matchA||!matchB) {
    if (dB) free(dB);
    if (dA) free(dA);
    if (matchA) free(matchA);
    if (matchB) free(matchB);
  }

  for (AV=0.0,i=0;i<m;i++) for (j=0;j<n;j++) AV+=compare(i,j,dA,dB);
  AV/=m*n;

  data.AV=AV;
  data.func=compare;
  data.dA=dA;
  data.dB=dB;

  *score=align(m,matchA,n,matchB,&data,compareStruct,AV*gapo,AV*gape,TRUE,&status)/(min(n,m)*AV);

  for (i=0;i<m;i++) A[i].pos=matchA[i];
  for (j=0;j<n;j++) B[j].pos=matchB[j];
       
  if (dB) free(dB);
  if (dA) free(dA);
  if (matchA) free(matchA);
  if (matchB) free(matchB);

  return(status);
}

/*------------------------------------------------------------------------*/

int dynGersteinX(int m,SAlires *A,int n,SAlires *B,float *score)
{
  /* Structural alignment using dynamic programming approach */
  /* modelled after M.Gerstein & M.Levitt, Comprehensive Assessment of */
  /* Automatic Structural Alignment against a Manual Standard, the     */
  /* SCOP Classification of Proteins, Protein Science (1998),7,445-456 */
  /* Only the basic method is implemented */

  /* Extension (June 27 2006): use combination of Gerstein and ResCompare5 
     scoring function to avoid matching close positions with incompatible 
     local geometry.
   */

  float *dA=NULL,*dB=NULL;
  int i,j,*matchA,*matchB,status=0;
  gerstData_t data;

  if (n<1||m<1) return(1);

  if (gerstXcompare5Factor>0) ResCompareSetup5(m,A,n,B,&dA,&dB);
  if (gerstXcompare5Factor>0&&(!dA||!dB)) {
    if (dA) free(dA);
    if (dB) free(dB);
    return(2);
  }

  matchA=(int *)malloc(m*sizeof(int));
  matchB=(int *)malloc(n*sizeof(int));

  if (!matchA||!matchB) {
    if (dA) free(dA);
    if (dB) free(dB);
    if (matchA) free(matchA);
    if (matchB) free(matchB);
    return(2);
  }

  data.A=A;
  data.B=B;
  data.dA=dA;
  data.dB=dB;

  *score=align(m,matchA,n,matchB,&data,compareGerst,gersteinGapPenaltyFactor,0,TRUE,&status)/(min(n,m));

  for (i=0;i<m;i++) A[i].pos=matchA[i];
  for (j=0;j<n;j++) B[j].pos=matchB[j];
       
  if (matchA) free(matchA);
  if (matchB) free(matchB);
  if (dA) free(dA);
  if (dB) free(dB);
  
  return(status);
}
/*------------------------------------------------------------------------*/

static float compareStruct(void *data,int i,int j)
{
  structData_t *S;

  /* Comparison function used for matches in 'salign' */

  S=data;
  return(S->AV-S->func(i,j,S->dA,S->dB));
}

/*------------------------------------------------------------------------*/

static float compareGerst(void *data,int i,int j)
{
  gerstData_t *S;
  float dij;
  float r=0,d02=2.24*2.24;

  S=data;

  dij=((S->A[i].x[0]-S->B[j].x[0])*(S->A[i].x[0]-S->B[j].x[0])+
       (S->A[i].x[1]-S->B[j].x[1])*(S->A[i].x[1]-S->B[j].x[1])+
       (S->A[i].x[2]-S->B[j].x[2])*(S->A[i].x[2]-S->B[j].x[2]));

  r=1/(1+dij/d02);

  if (gerstXcompare5Factor>0) 
    r+=(1-gerstXcompare5Factor*ResCompare5(i,j,S->dA,S->dB));

  return(r);
}

/*------------------------------------------------------------------------*/

/* Access function to set and read parameters */
float GetGersteinGapPenalty(void)
{
  return(gersteinGapPenaltyFactor);
}

void SetGersteinGapPenalty(float factor)
{
  gersteinGapPenaltyFactor=factor;
}

float GetGersteinCompare5Factor(void)
{
  return(gerstXcompare5Factor);
}

void SetGersteinCompare5Factor(float m)
{
  gerstXcompare5Factor=m;
}
