#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>

#define SHRINK    0.5
#define CONTRACT  0.5
#define GROW      2.0

/* Function minimization using simplex method (Nelder & Mead)                  */
/* From S.S. Rao, Optimization, Theory and Applications, Wiley Eastern Limited */

float SimPlexMin(
   int     n,                             /* number of variables                     */
   float  *x,                             /* in: initial values, out: best values    */
   float  *dx,                            /* dx for initial simplex                  */
   float (*F)(int n,float *y,void *data), /* target function                         */
   void   *data,                          /* pointer to additional data needed for F */
   float  tol,                            /* tolerance (dF)                          */
   int    mxit,                           /* maximum number of iterations            */
   int   (*report)(int it,int n,float **p,float *v,int hi,int lo), /* NULL or user report function */
   int   *status                          /* exit status                             */
   )

     /* exit status: 0, OK, 1 misc. errors., 2, not converged */

{
  int i,j,hi,nhi,lo,iter;
  float **p,*pr,*pe,*pb,*v,vr,ve,r;

  *status=0;

  /* Allocate local arrays */

  p=(float **)malloc((n+1)*sizeof(float *));
  if (!p) {
    *status=1;
    return(0);
  }
  for(i=0;i<n+1;i++) p[i]=(float *)NULL;

  for (i=0;!*status&&i<n+1;i++) {
    p[i]=(float *)malloc(n*sizeof(float));
    if (!p[i]) *status=1;
  }

  pr=(float *)malloc(n*sizeof(float));
  pe=(float *)malloc(n*sizeof(float));
  pb=(float *)malloc(n*sizeof(float));
  v=(float *)malloc((n+1)*sizeof(float));

  if (!status||!pr||!pe||!pb||!v) {
    *status=1;
    for (i=0;i<=n;i++) if (p[i]) free(p[i]);
    free(p);
    if (pr) free(pr);
    if (pe) free(pe);
    if (pb) free(pb);
    if (v) free(v);
    return(0);
  }
  if (i<0) {
    *status=1;
    return(0);
  }
  
  /* initialize simplex ... */

  for (j=0;j<n;j++) p[0][j]=x[j];
  for (i=0;i<n;i++) {
    for (j=0;j<n;j++) p[i+1][j]=p[0][j];
    p[i+1][i]+=dx[i];
  }
  for (hi=0,lo=0,i=0;i<n+1;i++) {
    v[i]=F(n,p[i],data);
    if (v[i]<v[lo]) lo=i;
    if (v[i]>v[hi]) hi=i;
  }

  /* iterate... */

  for(iter=0;iter<mxit;iter++) {                                    /* find high and low points */
    for (hi=0,lo=0,i=1;i<n+1;i++) {
      if (v[i]<v[lo]) lo=i;
      if (v[i]>v[hi]) hi=i;
    }
    for (nhi=lo,i=0;i<n+1;i++) if (i!=hi&&v[nhi]<v[i]) nhi=i;       /* find next highest ... */
    if (report) if (!report(iter,n,p,v,hi,lo)) break;               /* report progress */
    if (v[hi]-v[lo]<tol) break;                                     /* done if all values within tol */
    for (j=0;j<n;j++) {              
      for (pb[j]=0,i=0;i<n+1;i++) if (i!=hi) pb[j]+=p[i][j];
      pb[j]/=n;                                                     /* centroid of face opposite hi */
      pr[j]=2*pb[j]-p[hi][j];                                       /* reflected hi                 */
    }
    vr=F(n,pr,data);
    if (vr<v[lo]) {                 
      for (j=0;j<n;j++) pe[j]=GROW*pr[j]+(1-GROW)*pb[j];            /* expansion... */
      ve=F(n,pe,data);
      if (ve<v[lo]) for (v[hi]=ve,j=0;j<n;j++) p[hi][j]=pe[j];      /* accept expanded point */
      else for (v[hi]=vr,j=0;j<n;j++) p[hi][j]=pr[j];               /* accept reflected point */
    }else if (vr>=v[nhi]) {
      if (vr<v[hi]) for (v[hi]=vr,j=0;j<n;j++) p[hi][j]=pr[j];      /* accept reflected point */
      for (j=0;j<n;j++) pe[j]=CONTRACT*p[hi][j]+(1-CONTRACT)*pb[j]; /* interploate ... */
      ve=F(n,pe,data);
      if (ve>=v[hi]) {
	for (i=0;i<=n;i++) {
	  if (i!=lo) {                                              /* shrink towards lo */
	    for (j=0;j<n;j++) p[i][j]=SHRINK*p[i][j]+(1-SHRINK)*p[lo][j];
	    v[i]=F(n,p[i],data);
	  }
	}
      }else for (v[hi]=ve,j=0;j<n;j++) p[hi][j]=pe[j];              /* accept interapolated point */
    }else for (v[hi]=vr,j=0;j<n;j++) p[hi][j]=pr[j];
  }

  /* done... copy best point to x */

  if (v[hi]-v[lo]<tol) *status=0;
  else *status=2;
  if (mxit>0) {
    for (j=0;j<n;j++) x[j]=p[lo][j];
    r=v[lo];
  }else r=v[0];
      
  /* free local data */

  for(i=0;i<n+1;i++) free(p[i]);
  free(p); free(v); free(pr); free(pe); free(pb);

  return(r);
}
