/*   strutil.c   */

#include <stdio.h>
#include <errno.h>
#include "errhand.h"
#include <ctype.h>
#include "mystdlib.h"
#if defined(sgi) || defined(linux) || defined (cygwin) || defined (darwin) || defined(mingw)
#include <math.h>
#include <limits.h>
#endif
#if defined (cygwin) || defined(mingw)
#include <float.h>
#endif
#include "strutil.h"

/* if some of the constants are still undefined, then try this... */
#ifndef FLT_MAX
#define FLT_MAX 1.7e38
#endif
#ifndef FLT_MIN
#define FLT_MIN 3.4e-38
#endif
#ifndef LONG_MAX
#define LONG_MAX  2147483647
#endif
#ifndef SHRT_MAX
#define SHRT_MAX 32767
#endif

#define TRUE 1
#define FALSE 0

/*-------------------- local procedures --------------------*/

static char *mystrdup(const char *s)  /* may use memsafe stuff */
{
  char *t;
  t=(char *)malloc(strlen(s)+1);
  if (t) strcpy(t,s);
  return(t);
}

/*=====================================================================*/

double str2d(char *s)
     
     /*  Convert ASCIIZ string to double precision number.
	 Set errno to EINVAL if s is not a double,
	 set errno to ERANGE on overflow.                   */
     
{
  double d;
  int sign,expo,point,syntax_ok,edig,mdig,esign;
  char *p;
  
  /* check syntax ... */
  
  if (!s) {
    d=0;
    err_set(1,"strd2 : bad operand(empty field)");
    errno=EINVAL;
    return(d);
  }
  
  for(p=s;isspace((int)(*p));p++);
  sign = (*p == '-' || *p == '+'); if (sign) p++;
  for(mdig=0;isdigit((int)(*p));p++,mdig++);
  point = (*p == '.'); if (point) p++;
  for(;isdigit((int)(*p));p++,mdig++);
  expo = (*p == 'e' || *p == 'E'); if (expo) p++;
  if (expo) {
    esign = (*p == '-' || *p == '+'); if (esign) p++;
    for(edig=0;isdigit((int)(*p));p++,edig++);
  }else edig=0;
  for (;isspace((int)(*p));p++);
  syntax_ok = (!expo || (edig && mdig)) && (!point || mdig)
    && (!sign || mdig) && *p == '\0';
  
  errno=0;
  if (mdig==0) d=0.0; else d=atof(s);
#ifdef sgi
  if (d > HUGE_VAL || d < -HUGE_VAL) {
    errno=ERANGE;
    err_set(1,"str2d: operand range(%s)",s);
  }  
#endif
  if (!syntax_ok) {
    errno=EINVAL;
    err_set(1,"str2d: syntax(%s)",s);
  }
  return(d);
}


double str2dsigma(char *s,double *sigma)
     
     /*  Convert ASCIIZ string to double precision number.
	 Set errno to EINVAL if s is not a double,
	 set errno to ERANGE on overflow. 

	 This function also returns sigma ( <float>(sigma), in CIF format )

      */
     
{
  double d;
  int sign,expo,point,syntax_ok,edig,mdig,esign,sgmaexp,sgma,exp,exps;
  char *p;
  
  /* check syntax ... */
  
  if (!s) {
    d=0;
    err_set(1,"strd2 : bad operand(empty field)");
    errno=EINVAL;
    return(d);
  }
  
  sgmaexp=0;exp=0;exps=1;*sigma=0;
  for(p=s;isspace((int)(*p));p++);
  sign = (*p == '-' || *p == '+'); if (sign) p++;
  for(d=0,mdig=0;isdigit((int)(*p));p++,mdig++);
  point = (*p == '.'); if (point) p++;
  for(;isdigit((int)(*p));p++,mdig++,sgmaexp--);
  expo = (*p == 'e' || *p == 'E'); if (expo) p++;
  if (expo) {
    esign = (*p == '-' || *p == '+'); if (*p=='-') exps=-1; if (esign) p++;
    for(exp=0,edig=0;isdigit((int)(*p));edig++,p++) exp=10*exp+(int)(*p)-'0'; 
  }else edig=0;
  if (*p=='(') {
    p++;
    sgmaexp+=exps*exp;
    for (sgma=0;isdigit((int)(*p));p++) sgma=sgma*10+(int)(*p)-'0';
    if (*p==')') {
      *sigma=sgma*pow(10.0,sgmaexp);
      p++; 
    }else err_set(1,"str2dsigma: syntax(%s)",s);
  }
  for (;isspace((int)(*p));p++);
  syntax_ok = (!expo || (edig && mdig)) && (!point || mdig)
    && (!sign || mdig) && *p == '\0';
  
  errno=0;
  if (mdig==0) d=0.0; else d=atof(s);
#ifdef sgi
  if (d > HUGE_VAL || d < -HUGE_VAL) {
    errno=ERANGE;
    err_set(1,"str2d: operand range(%s)",s);
  }  
#endif
  if (!syntax_ok) {
    errno=EINVAL;
    err_set(1,"str2d: syntax(%s)",s);
  }
  return(d);
}

/*=====================================================================*/

float str2f(char *s)
     
     /*  Convert ASCIIZ string to single precision number.
	 Set errno to EINVAL if wrong syntax,
	 set errno to ERANGE if out of range (overflow)   */
     
{
  double d;
  
  d=str2d(s);if (errno==EINVAL) return(0.0);
#if defined(sgi)
  if (d < FLT_MIN && d > -FLT_MIN) return(0.0);
#endif
  if (d > FLT_MAX || d < -FLT_MAX) {
    errno=ERANGE;
    err_set(1,"str2f: operand range(%s)",s);
    return(d < 0.0 ? -FLT_MAX : FLT_MAX);
  }
  return((float)d);
}

/*=====================================================================*/

long str2l(char *s)
     
     /*  Convert ASCIIZ string to long integer number.
	 Set errno to EINVAL if wrong syntax,
	 set errno to ERANGE if out of range (overflow)   */
     
{
  long l;
  double d;
  
  d=str2d(s);
  if (errno==EINVAL) return(0);
  if (d > LONG_MAX || d < -LONG_MAX) {
    errno=ERANGE;
    err_set(1,"str2l: operand range(%s)",s);
    return(d < 0.0 ? -LONG_MAX : LONG_MAX);
  }
  l=(long)d;
  if ((double)l - d != 0.0) {
    err_set(1,"str2l: bad operand(%s)",s);
    errno=EINVAL;
    return(0);
  }
  return(l);
}
/*=====================================================================*/

short str2s(char *s)
     
     /*  Convert ASCIIZ string to short integer number.
	 Set errno to EINVAL if wrong syntax,
	 set errno to ERANGE if out of range (overflow)   */
     
{
  short l;
  double d;
  
  d=str2d(s);
  if (errno==EINVAL) return(0);
  if (d > SHRT_MAX || d < -SHRT_MAX) {
    err_set(1,"str2s: operand range(%s)",s);
    errno=ERANGE;
    return(d < 0.0 ? -SHRT_MAX : SHRT_MAX);
  }
  l=d;
  if ((double)l - d != 0.0) {
    err_set(1,"str2s: bad operand(%s)",s);
    errno=EINVAL;
    return(0);
  }
  return(l);
}
/*=====================================================================*/

int str2i(char *s)
     
     /*  Convert ASCIIZ string to integer number.
	 Set errno to EINVAL if wrong syntax,
	 set errno to ERANGE if out of range (overflow)   */
     
{
  int l;
  double d;
  
  d=str2d(s);
  if (errno==EINVAL) return(0);
  if (d > INT_MAX || d < -INT_MAX) {
    err_set(1,"str2i: operand range(%s)",s);
    errno=ERANGE;
    return(d < 0.0 ? -INT_MAX : INT_MAX);
  }
  l=d;
  if ((double)l - d != 0.0) {
    err_set(1,"str2i: bad operand(%s)",s);
    errno=EINVAL;
    return(0);
  }
  return(l);
}
/*======================================================================*/

void StrCpy(char *t,char *x,int len)
{
  if (t) {
   t[0]='\0';
   if(x) {
     if (strlen(x)<len) strcpy(t,x);
     else if (len>0) {
       strncpy(t,x,len-1);
       t[len-1]='\0';
     }
   }
  }
}

char *Strdup(const char *p)
{
  return( p ? mystrdup(p) : (char *)NULL);
}

char *strfld(char *target,char *source,int from,int to)
{
  /* extract field from source string, strip trailing whitespace */
  /* Note: from=1 is source[0]  */
  
  int k;
  
  if (!source) {
    if (target) *target='\0';
    return(target);
  }
  if (!target) return(target);
  if (to>=from) {
    strncpy(target,source+from-1,to-from+1);
    target[to-from+1]='\0';
    k=to-from;
    while (k>=0) {
      if (!isspace(target[k])) break;
      target[k--]='\0';
    }
  }else *target='\0';
  return(target);
}

int streq(char *s1,char *s2)
{
  if (s1==s2) return(TRUE);
  else if (!s1 || !s2) return(FALSE);
  else return(strcmp(s1,s2)==0);
}
    
int strceq(char *s1,char *s2)
{
  if (s1==s2) return(TRUE);
  else if (!s1 || !s2) return(FALSE);
  else return(strcasecmp(s1,s2)==0);
}
    
int strneq(char *s1,char *s2,int n)
{
  if (s1==s2) return(TRUE);
  else if (!s1 || !s2) return(FALSE);
  else return(strncmp(s1,s2,n)==0);
}
    
int strnceq(char *s1,char *s2,int n)
{
  if (s1==s2) return(TRUE);
  else if (!s1 || !s2) return(FALSE);
  else return(strncasecmp(s1,s2,n)==0);
}
    
char *strsqz(char *target,char *source)
{
  /* squeeze out blanks from string */
  
  char *s,*t;
  
  s=source; t=target;
  if (!s) return(target);
  while (*s) {
    if (*s!=' ') *t++ = *s;
    s++;
  }
  *t = '\0';
  return(target);
}

char *Strcat(char *s1,char *s2)
{
  /* append s2 to s1, free s1, return pointer to dynamically allocated result */

  char *s;
  int l1,l2;

  l1=(s1 ? strlen(s1) :0);
  l2=(s2 ? strlen(s2) :0);

  s=(char *)malloc((l1+l2+1)*sizeof(char));
  if (s) {
    *s='\0';
    if (s1) strcpy(s,s1);
    if (s2) strcpy(s+l1,s2);
  }
  if (s1) free(s1);
  return(s);
}

int kwd_match(char *kwd,char* kwd_list,char *match)
{     
/* Return kwd# (zero based) if kwd matches a keyword from kwd_list; 
   return the matched keyword in match. If kwd does not match unambiguously 
   a word from kwd_list the function will return -1. Comparison is 
   case-insensitive. If a keyword in kwd_list is "quoted", kwd must have
   the same length for a match
   
   Handles special keywords (second char is '!')
      h!... ("hidden") matches iff ... does,
      t!... ("title")  never matches.

 */
     
  char *s;
  Parse k;
  int lkwd,nr=0,nt,l=0;
  
  if (!kwd || !kwd_list) return(-1);
  if (strlen(kwd)==0) return(-1);

  /* Set up parsing structure for keyword list */
  
  k.buf = kwd_list;
  k.pos = 0;
  k.sep = "";
  k.tok = (char *)NULL;
  
  /* Initialize a few other things...          */
  
  lkwd = strlen(kwd);         /* length of keyword to match */
  s = (char *)NULL;           /* not found yet... */
  
  /* Loop over tokens in keyword list */
  
  for (nt=0;get_token(&k);nt++) {
    if (k.ltok>2&&k.tks[1]=='!') {
      if (k.tks[0]=='t') continue;
      else if (k.tks[0]=='h') {
	k.tks+=2; k.ltok-=2;
      }
    }
    if (k.quoted && k.ltok != lkwd) continue;
    if (k.ltok < lkwd) continue;            /* too short to match */
    if (strncasecmp(k.tks,kwd,lkwd)==0) {
      if (s == (char *)NULL) {
	s = k.tks;
	nr = nt;
	l = k.ltok;
	if (k.ltok == lkwd) break;
      }
      else {
	if (k.ltok==lkwd) {
	  nr=nt;
	  s=k.tks;
	  l = k.ltok;
	}else{
	  s = (char *)NULL;
	}
	break;
      }
    }
  }
  if (s) {
    if (match) {
      strncpy(match,s,l);
      match[l]='\0';
    }
  }else{
    nr = -1;
  }
  return(nr);
}

/*========================================================================*/


#define ANY '*'
#define ONE '?'
#define XONE '%'
#define XONE2 '?'
#define DIG '+'
#define DIGS '#'
#define QUO '\\'
#define FALSE 0
#define TRUE 1
#define YES 1
#define NO 0

#define is_special(c) ((c) == ANY || (c) == ONE || (c) == QUO)
#define is_xspecial(c) ((c) == ANY || (c) == XONE || (c) == XONE2 || (c) == DIG || (c) == DIGS || (c) == QUO)
#define is_normal(c) (!is_special(c))
#define is_xnormal(c) (!is_xspecial(c))
#define is_wild(c) ((c) == ANY || (c) == ONE)
#define is_xwild(c) ((c) == ANY || (c) == DIG || (c) == DIGS || (c) == XONE || (c) == XONE2)
#define LOWER(c) ((c) <= 'Z' && (c) >= 'A' ? (c) + ('a'-'A') : c)

/*   Return 1 if "reg" matches "wild" with wildcards '*' and '?', else 0.
     
     '\' is used as a quote character in "wild".
     '*' matches any number of arbitrary characters.
     '?' matches a single arbitrary character.
     Comparison is case sensitive.
     */


int wild_match(char *reg,char *wild)
{
  if (!reg || !wild) {
    if (!reg && !wild) return(TRUE);
    if (!reg && strcmp(wild,"*")==0) return(TRUE);
    else return(FALSE);
  }

  while (*wild || *reg) {
    if (is_normal(*wild)) {
      if (*wild != *reg) return(FALSE);
    }else if (*wild == QUO) {
      wild++;
      if (*wild != *reg) return(FALSE);
    }else if (*wild == ONE) {
      if (*reg == '\0') return(FALSE);
    }else if (*wild == ANY) {
      wild++;
      if (*wild == '\0') return(TRUE);
      do {
	if (wild_match(reg,wild)) return(TRUE);
      } while (*(reg++));
      return(FALSE);
    }
    wild++;reg++;
  }
  return(TRUE);
}
/* ---------------------------------------------------------------------- */

int wild_imatch(char *reg,char *wild)
{

/*   Return 1 if "reg" matches "wild" with wildcards '*' and '?', else 0.
     
     '\' is used as a quote character in "wild".
     '*' matches any number of arbitrary characters.
     '?' matches a single arbitrary character.
     Comparison is case insensitive (except for "quoted" characters).
     */

  if (!reg || !wild) {
    if (!reg && !wild) return(TRUE);
    if (!reg && strcmp(wild,"*")==0) return(TRUE);
    else return(FALSE);
  }

  while (*wild || *reg) {
    if (is_normal(*wild)) {
      if (LOWER(*wild) != LOWER(*reg)) return(FALSE);
    }else if (*wild == QUO) {
      wild++;
      if (*wild != *reg) return(FALSE);
    }else if (*wild == ONE) {
      if (*reg == '\0') return(FALSE);
    }else if (*wild == ANY) {
      wild++;
      if (*wild == '\0') return(TRUE);
      do {
	if (wild_imatch(reg,wild)) return(TRUE);
      } while (*(reg++));
      return(FALSE);
    }
    wild++;reg++;
  }
  return(TRUE);
}

/* ---------------------------------------------------------------------- */

int mwild_xmatch(char *reg,char *wild)
{
  /* the same as wild_xmatch, but ',' in wild means 'or', except if
     quoted by preceding '\' 

     Note: empty string (strlen(wild)==0) matches everything,
     even NULL  !
  */

  int r;
  char *p,*c;

  r=FALSE;
  if (!wild) return(FALSE);
  if (strlen(wild)==0) return(TRUE);
  if (!reg) return(FALSE); 
  p=wild;
  for (c=p;*c!='\0'&&!r;) {
    if (*c=='\\') c++;
    else{
      if (*c==',') {
	*c='\0';
	r=wild_xmatch(reg,p);
	*c=',';
	p=c+1;
      }
    }
    if (*c) c++;
  }
  if (!r&&*p) r=wild_xmatch(reg,p);
  return(r);
}

int wild_xmatch(char *reg,char *wild)
{
/*   Return 1 if "reg" matches "wild" with wildcards '*' and '?', else 0.
     
     This is an X-Plor - like wild match routine.

     '\' is used as a quote character in "wild".
     '*' matches any number of arbitrary characters.
     '%' and '?' match a single arbitrary character.
     '+' matches a single digit.
     '#' matches a sequence of digits (incl. zero length)

     Comparison is case insensitive (except for "quoted" characters).

     */

  if (!reg || !wild) {
    if (!reg && !wild) return(TRUE);
    if (!reg && strcmp(wild,"*")==0) return(TRUE);
    else return(FALSE);
  }

  while (*wild || *reg) {
    if (is_xnormal(*wild)) {
      if (LOWER(*wild) != LOWER(*reg)) return(FALSE);
    }else if (*wild == QUO) {
      wild++;
      if (*wild != *reg) return(FALSE);
    }else if (*wild == DIG) {
      if (*reg < '0' || *reg > '9') return(FALSE);
    }else if (*wild == XONE || *wild == XONE2 ) {
      if (*reg == '\0') return(FALSE);
    }else if (*wild == DIGS) {
      wild++;
      if (*wild == '\0') {
	for (;*reg;reg++) if (*reg < '0' || *reg > '9') return(FALSE);
	return(TRUE);
      }else{
	do {
	  if (wild_xmatch(reg,wild)) return(TRUE);
	  if (*reg < '0' || *reg > '9') return(FALSE);
	} while (*(reg++));
	return(FALSE);
      }
    }else if (*wild == ANY) {
      wild++;
      if (*wild == '\0') return(TRUE);
      do {
	if (wild_xmatch(reg,wild)) return(TRUE);
      } while (*(reg++));
      return(FALSE);
    }
    wild++;reg++;
  }
  return(TRUE);
}
/*-------------------- extfil --------------------*/

char *extfil(char *filext,char *file,char *ext,int replace)
{
  char *path,*start;
  int l;
  
  strcpy(filext,file);
  l=strlen(filext);
  if (l>0) if (*filext=='|') return(filext);           /* it is a pipe */
  if (l>=2) {
    if (streq(filext+l-2,".Z")) {
      if (replace!=YES) return(filext);                /* don't tamper with .Z */
      filext[l-2]='\0';
    }
  }
  path=strrchr(filext,'/');
  if (path==(char *)NULL) path=filext;
  start=strrchr(path,'.');
  
  if (replace==YES && start!=(char *)NULL) {
    if (ext) strcpy(start,ext);
  } else if (start==(char *)NULL) {
    if (ext) strcat(filext,ext);
  }
  return(filext);
}
