#include <stdio.h>
#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "surface.h"
#include "graphics.h"
#include "transpor.h"
#include "bhtree.h"

extern int graphix;

#define CHUNK_SIZE 16192

static void renumberSurface(Surface *S,int *new);

/*-------------------- newSurface --------------------*/

Surface *newSurface(int type)
{
  Surface *S;

  S=(Surface *)malloc(sizeof(Surface));
  if (mvld(S)) {
    S->type=type;
    S->v=(SPoint *)NULL;
    S->t=(STriangle *)NULL;
    S->color=0;                /* no color */
    S->lw=1;
    S->style=WNP_TRG_RENDERED;
    S->pattern=0;
    S->bckclr=0;
    S->nv=0;
    S->nt=0;
  }
  return(S);
}
/*-------------------- surface_info --------------------*/

void surface_info(Surface *S)
{
  float surf,vol;

  if (mvld(S)) {
    switch(S->type) {
    case SURF_TYPE_TRG:
      prtGraphics("Triangulated surface has %d vertices and %d faces.\n",S->nv,S->nt);
      SurfAndVol(S,&surf,&vol);
      prtGraphics("Surface area = %0.3f,"
		  "Volume inside surface = %0.3f\n",surf,vol);
      break;
    case SURF_TYPE_MSH:
      prtGraphics("Surface is of type 'mesh', %d lines.\n",S->nv);
      break;
    default:
      prtGraphics("Surface has unknown type.\n");
      break;
    }
  }else prtGraphics("no surface.\n");
}

/*-------------------- freeSurface --------------------*/

Surface *freeSurface(Surface *S)
{
  if (mvld(S)) {
    if (S->v) {free(S->v);} S->v=(SPoint *)NULL;
    if (S->t) {free(S->t);} S->t=(STriangle *)NULL;
    free(S);
    S=(Surface *)NULL;
  }
  return(S);
}
/*-------------------- readSurface --------------------*/

Surface *readSurface(FILE *f)
{
  Surface *S=(Surface *)NULL;
  char buf[4096];
  int i,k;

  S=newSurface(SURF_TYPE_UNK);
  if (mvld(S)) {
    if (fgets(buf,sizeof(buf)-1,f)&&strneq(buf,"#S:",3)) {
      sscanf(buf+3,"%d %d %d %d %d",
	     &S->color,&S->lw,&S->style,&S->pattern,&S->bckclr);
      if (!fgets(buf,sizeof(buf)-1,f)) *buf='\0';
    }
    sscanf(buf,"%d",&S->nv);
    if (S->nv>0) {
      S->v=(SPoint *)malloc(S->nv*sizeof(SPoint));
      if (mvld(S->v)) {
	for (i=0;i<S->nv;i++) {
	  if (!fgets(buf,sizeof(buf)-1,f)) {
	    err_set(1,"readSurface: premature EOF in input file (vertices).");
	    S=freeSurface(S);
	    break;
	  }
	  sscanf(buf,"%f %f %f %f %f %f",
		 S->v[i].x,S->v[i].x+1,S->v[i].x+2,
		 S->v[i].n,S->v[i].n+1,S->v[i].n+2);
	}
      }else{
	err_set(1,"readSurface: failed to allocate memory (vertices).");
	S=freeSurface(S);
      }
    }else{
      err_set(1,"readSurface: no vertices.");
      S=freeSurface(S);
    }
  }
  if (mvld(S)) { 
    if (fgets(buf,sizeof(buf)-1,f)) sscanf(buf,"%d",&S->nt);
    if (S->nt>0) {
      S->type=SURF_TYPE_TRG;
      S->t=(STriangle *)malloc(S->nt*sizeof(STriangle));
      if (mvld(S->t)) {
	for (i=0;i<S->nt;i++) {
	  if (!fgets(buf,sizeof(buf)-1,f)) {
	    err_set(1,"premature EOF in input file (triangles).");
	    S->nt=i;
	    break;
	  }
	  sscanf(buf,"%d %d %d",S->t[i].v,S->t[i].v+1,S->t[i].v+2);
	  for (k=0;k<3;k++) S->t[i].v[k]=clamp(S->t[i].v[k],0,S->nv-1);
	}
      }else{
	err_set(1,"readSurface: failed to allocate memory (faces).");
	S=freeSurface(S);
      }
    }else S->type=SURF_TYPE_MSH;  
  }
  return(S);
}
/*-------------------- writeSurface --------------------*/

int writeSurface(FILE *out,Surface *S)
{
  int i,status=0;

  if (out&&mvld(S)) {
    switch (S->type) {
    case SURF_TYPE_TRG:
      fprintf(out,"#S: %d %d %d %d %d\n",
	      S->color,S->lw,S->style,S->pattern,S->bckclr);
      fprintf(out,"%d points\n",S->nv);
      if (S->v) {
	for (i=0;i<S->nv;i++) {
	  fprintf(out,"%0.3f %0.3f %0.3f %0.3f %0.3f %0.3f\n",
		  S->v[i].x[0],S->v[i].x[1],S->v[i].x[2],
		  S->v[i].n[0],S->v[i].n[1],S->v[i].n[2]);
	}
      }
      fprintf(out,"%d triangles\n",S->nt);
      if (S->t) {
	for (i=0;i<S->nt;i++) {
	  fprintf(out,"%d %d %d\n",S->t[i].v[0],S->t[i].v[1],S->t[i].v[2]);
	}
      }
      if (ferror(out)) status=3;
      break;
    case SURF_TYPE_MSH:
      fprintf(out,"#S: %d %d %d %d %d\n",
	      S->color,S->lw,S->style,S->pattern,S->bckclr);
      fprintf(out,"%d lines\n",S->nv);
      if (S->v) {
	for (i=0;i<S->nv;i++) {
	  fprintf(out,"%0.3f %0.3f %0.3f %0.3f %0.3f %0.3f\n",
		  S->v[i].x[0],S->v[i].x[1],S->v[i].x[2],
		  S->v[i].n[0],S->v[i].n[1],S->v[i].n[2]);
	}
      }
      if (ferror(out)) status=3;
      break;
    default:
      status=2;
      break;
    }
  }else status=1;
  return(status);
}
/*-------------------- getSurfaceFromServer --------------------*/

Surface *getSurfaceFromServer(Dspobj *obj,int *flags)
{
  Surface *S=NULL;
  Qmsg *Msg;
  int size,type;
  int *msg;

  if (is_dsp(obj)&&SrvVersion()>2.1999) {
    send_bheader(graphix,"GGOB",sizeof(obj));
    send_ptr(graphix,obj);
    Msg=waitqmsg("GGOB SCCN");
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	Msg=delqmsg(Msg);
	graphics_down("Server closed connection...");
      }else{
	size=Msg->size/sizeof(int); msg=(int *)Msg->msg;
	if (size>=9) {
	  type=MY_i(*msg++);
	  *flags=MY_i(*msg++);
	  size-=2;
	  switch (type) {
	  case DSP_TRG:
	    S=getTRGSurfaceFromServer(msg,size);
	    break;
	  default:
	    break;
	  }
	}
      }
      delqmsg(Msg);
    }else{
      graphics_down("no response from server.");
    }
  }
  return(S);
}

Surface *getTRGSurfaceFromServer(int *msg,int size)
{
  Surface *S;
  int i,j;

  S=newSurface(SURF_TYPE_TRG);
  if (S) {
    if (size>=7) {
      S->color=MY_i(*msg++);
      S->lw=MY_i(*msg++);
      S->style=MY_i(*msg++);
      S->pattern=MY_i(*msg++);
      S->bckclr=MY_i(*msg++);
      S->nv=MY_i(*msg++);
      S->nt=MY_i(*msg++);
      size-=7;
      if (S->nt>0&&S->nv>0) {
	if ((6*S->nv+3*S->nt)>=size) {
	  S->v=malloc(6*S->nv*sizeof(float));
	  S->t=malloc(3*S->nt*sizeof(int));
	  if (S->v&&S->t) {
	    for (i=0;i<S->nv;i++) {
	      for (j=0;j<3;j++) S->v[i].x[j]=MY_f(*msg++);
	      for (j=0;j<3;j++) S->v[i].n[j]=MY_f(*msg++);
	    }
	    for (i=0;i<S->nt;i++) {
	      for (j=0;j<3;j++) S->t[i].v[j]=MY_i(*msg++);
	    }
	  }else{
	    S=freeSurface(S);
	    err_set(1,"getTRGSurfaceFromServer: failed to allocate memory for trg surface.");
	  }
	}else{
	  S=freeSurface(S);
	  err_set(1,"getTRGSurfaceFromServer: (nv,nt) in header does not match message size from server.");
	}
      }else S->nv=0;
    }
  }
  return(S);
}

/*-------------------- SurfAndVol --------------------*/

void SurfAndVol(Surface *S,float *surf,float *vol)
{
  /* estimate surface area and volume via Gauss's divergence theorem */
  /* assumes that surface is boundary of admissible region           */

  int i,k;
  float *df,s,d1[3],d2[3],t[3],vi=0,vo=0;

  *surf=0; *vol=0;
  if (mvld(S)&&S->type==SURF_TYPE_TRG&&S->nv>0&&S->nt>0&&mvld(S->v)&&mvld(S->t)) {
    df=(float *)malloc(S->nv*sizeof(float));
    if (!df) {
      err_set(1,"failed to allocate memory for surface area calculation.");
      return;
    }
    for (i=0;i<S->nv;i++) 
      for (df[i]=0.0,k=0;k<3;k++) df[i]+=S->v[i].x[k]*S->v[i].n[k];
    for (*surf=0.0,vi=0.0,vo=0.0,i=0;i<S->nt;i++) {
      for (k=0;k<3;k++) {
	d1[k]=S->v[S->t[i].v[1]].x[k]-S->v[S->t[i].v[0]].x[k];
	d2[k]=S->v[S->t[i].v[2]].x[k]-S->v[S->t[i].v[0]].x[k];
      }
      t[0]=d1[1]*d2[2]-d1[2]*d2[1];
      t[1]=d1[2]*d2[0]-d1[0]*d2[2];
      t[2]=d1[0]*d2[1]-d1[1]*d2[0];
      s=sqrt(t[0]*t[0]+t[1]*t[1]+t[2]*t[2]);
      *surf+=s; vo+=s*(df[S->t[i].v[0]]+df[S->t[i].v[1]]+df[S->t[i].v[2]]);
      vi+=(S->v[S->t[i].v[0]].x[0]+S->v[S->t[i].v[1]].x[0]+S->v[S->t[i].v[2]].x[0])*t[0]
	+(S->v[S->t[i].v[0]].x[1]+S->v[S->t[i].v[1]].x[1]+S->v[S->t[i].v[2]].x[1])*t[1]
	+(S->v[S->t[i].v[0]].x[2]+S->v[S->t[i].v[1]].x[2]+S->v[S->t[i].v[2]].x[2])*t[2];
    }
    free(df);
  }
  *surf/=2.0; *vol=(vi+vo)/36.0;
}
/*-------------------- addSurfaceToDisplay --------------------*/

Dspobj *addSurfaceToDisplay(Surface *S,Dspobj *glob,char *name)
{
  Dspobj *t=(Dspobj *)NULL;
  Gfxobj *cont;
  int i;

  if (!mvld(S)||!is_dsp(glob)) return(t);

  switch (S->type) {
  case SURF_TYPE_TRG:
    if (S->nv>0&&S->nt>0&&mvld(S->v)&&mvld(S->t)) {
      t=new_dsp_obj(glob,name,DSP_TRG);
      if (is_dsp(t)) {
	if (cPtr(t->cross)) {
	  send_bheader(graphix,"DSPC",sizeof(int)+sizeof(t));
	  send_bint(graphix,S->color);
	  send_bptr(graphix,t);
	  send_bheader(graphix,"SOLW",sizeof(int)+sizeof(t));
	  send_bint(graphix,S->lw);
	  send_bptr(graphix,t);
	  send_bheader(graphix,"SBFC",sizeof(int)+sizeof(t));
	  send_bint(graphix,S->bckclr);
	  send_bptr(graphix,t);
	  send_bheader(graphix,"TRGS",sizeof(int)+sizeof(t));
	  send_bint(graphix,S->style);
	  send_bptr(graphix,t);
	  send_bheader(graphix,"TRGT",sizeof(int)+sizeof(t));
	  send_bint(graphix,S->pattern);
	  send_bptr(graphix,t);
	}
	send_bheader(graphix,"ATRV,",sizeof(t)+6*S->nv*sizeof(int));
	send_bptr(graphix,t);
	for (i=0;i<S->nv;i++) {
	  send_bfloata(graphix,3,S->v[i].x);
	  send_bfloata(graphix,3,S->v[i].n);
	}
	send_bheader(graphix,"ATRF,",sizeof(t)+3*S->nt*sizeof(int));
	send_bptr(graphix,t);
	for (i=0;i<S->nt;i++) {
	  send_bint(graphix,S->t[i].v[0]);
	  send_bint(graphix,S->t[i].v[1]);
	  send_bint(graphix,S->t[i].v[2]);
	}
      }
    }
    break;
  case SURF_TYPE_MSH:
    if (S->nv>0&&mvld(S->v)) {
      cont=OpenGfxObject(glob,2000);
      if (mvld(cont)) {
	t=cont->Obj;
	if (is_dsp(t)) {
	  t->dsptype=DSP_CNT;
	  set_node_name((Node *)t,name);
	  for (i=0;i<S->nv;i++) AddMD(cont,S->v[i].x,S->v[i].n,S->color);
	}
	CloseGfxObject(cont);
	send_bheader(graphix,"SOLW",sizeof(int)+sizeof(cont));
	send_bint(graphix,S->lw);
	send_bptr(graphix,cont);
      }
    }
    break;
  default:
    break;
  }
  return(t);
}
/*-------------------- surfaceNormals --------------------*/

void surfaceNormals(Surface *S,Grid *G)
{
  int i,j,k,l;
  float n[3],*x[3],d1[3],d2[3],s;

  if (!mvld(S)||!mvld(S->v)||!mvld(S->t)) return;
  if (S->type==SURF_TYPE_TRG) {
    if (mvld(G)) {
      /* analytical normals (not implemented... *) */
    }else{
      /* Compute normals at vertices by accumulating face normals */
    
      for (i=0;i<S->nv;i++) {
	S->v[i].n[0]=0.0;
	S->v[i].n[1]=0.0;
	S->v[i].n[2]=0.0;
      }
    
      for (k=0;k<S->nt;k++) {
	for (l=0;l<3;l++) x[l]=S->v[S->t[k].v[l]].x;
	for (j=0;j<3;j++) {
	  d1[j]=x[1][j]-x[0][j];
	  d2[j]=x[2][j]-x[0][j];
	}
	n[0]=d1[1]*d2[2]-d1[2]*d2[1];
	n[1]=d1[2]*d2[0]-d1[0]*d2[2];
	n[2]=d1[0]*d2[1]-d1[1]*d2[0];
	s=sqrt(n[0]*n[0]+n[1]*n[1]+n[2]*n[2]);
	if (s>0) {
	  s=1/s;
	  n[0]*=s; n[1]*=s; n[2]*=s;
	}else{
	  n[0]=1; n[1]=0; n[2]=0;
	}
	for (l=0;l<3;l++) {
	  i=S->t[k].v[l];
	  for (j=0;j<3;j++) S->v[i].n[j]+=n[j];
	}
      }
    
      for (i=0;i<S->nv;i++) {
	s=sqrt(S->v[i].n[0]*S->v[i].n[0]+S->v[i].n[1]*S->v[i].n[1]+S->v[i].n[2]*S->v[i].n[2]);
	if (s>0) {
	  s=1/s;
	  for (j=0;j<3;j++) S->v[i].n[j]*=s;
	}
      }
    }
  }
}
/*-------------------- smoothSurface --------------------*/

void smoothSurface(Surface *S,float bias,int iter)
{
  int i,k,i0,i1,i2;
  int *nf=(int *)NULL;
  float *p=(float *)NULL;;

  bias=clamp(bias,0,1);
  iter=clamp(iter,0,100);
  if (!mvld(S)||!mvld(S->v)||!mvld(S->t)) return;

  if (S->type==SURF_TYPE_TRG ) {
    nf=(int *)malloc(S->nv*sizeof(int));
    p=(float *)malloc(3*S->nv*sizeof(float));
    if (mvld(p)&&mvld(nf)) {
      while (iter>0) {
	iter--;
	for (i=0;i<S->nv;i++) {
	  nf[i]=0;
	  p[3*i]=0.0;
	  p[3*i+1]=0.0;
	  p[3*i+2]=0.0;
	}
	for (i=0;i<S->nt;i++) {
	  for (k=0;k<3;k++) {
	    i0=S->t[i].v[k];
	    i1=S->t[i].v[(k+1)%3];
	    i2=S->t[i].v[(k+2)%3];
	    p[3*i0]+=S->v[i1].x[0]+S->v[i2].x[0];
	    p[3*i0+1]+=S->v[i1].x[1]+S->v[i2].x[1];
	    p[3*i0+2]+=S->v[i1].x[2]+S->v[i2].x[2];	
	    nf[i0]++;
	  }
	}
	for (i=0;i<S->nv;i++) {
	  if (nf[i]>0) {
	    S->v[i].x[0]=(1-bias)*S->v[i].x[0]+bias*p[3*i]/(2*nf[i]);
	    S->v[i].x[1]=(1-bias)*S->v[i].x[1]+bias*p[3*i+1]/(2*nf[i]);
	    S->v[i].x[2]=(1-bias)*S->v[i].x[2]+bias*p[3*i+2]/(2*nf[i]);
	  }
	}
      }
    }else err_set(1,"smoothSurface: failed to allocate scratch space.");
  }
  if (mvld(p)) free(p);
  if (mvld(nf)) free(nf);
}
/*-------------------- AddTMDtoS --------------------*/

int AddTMDtoS(Surface *S,mat_4 T,float *x,float *y)
{
  int i,status=0; 
  float *X,*Y;

  if (mvld(S)) {
    if (S->type==SURF_TYPE_MSH) {
      if (S->nv%CHUNK_SIZE==0) {
	S->v=realloc(S->v,(S->nv+CHUNK_SIZE)*sizeof(SPoint));}
      if (S->v) {
	status=1;
	for (X=S->v[S->nv].x,Y=S->v[S->nv].n,i=0;i<3;i++) {
	  X[i]=T.r[i].c[3]+T.r[i].c[0]*x[0]+T.r[i].c[1]*x[1]+T.r[i].c[2]*x[2];
	  Y[i]=T.r[i].c[3]+T.r[i].c[0]*y[0]+T.r[i].c[1]*y[1]+T.r[i].c[2]*y[2];
	}
	S->nv++;
      }
    }
  }
  return(status);
}

void replaceSurface(Dspobj *obj,Surface *S)
{
  char *name=NULL;
  Dspobj *glbl;

  if (is_dsp(obj)&&(obj->dsptype==DSP_TRG||obj->dsptype==DSP_CNT)&&S) {
    name=Strdup(obj->name);
    glbl=glblobj(obj);
    del_node((Node *)obj);
    addSurfaceToDisplay(S,glbl,name);
    if (name) free(name);
  }
}

void clipSurfaceAtoms(Surface *S,Nodechain *c,float cutoff)
{
  BHtree *T;
  int i,n,grain,nacc,*new=NULL;
  float dist;
  Atom *atom;

  if (c&&c->n>0&&S&&S->nv>0) {
    grain = ( c->n < 2048 ? 1 : c->n / 2048 );
    T=generateBHtree(c,grain);
    if (T) {
      new=(int *)malloc(S->nv*sizeof(int));
      if (new) {
	for (nacc=0,i=0;i<S->nv;i++) {
	  n=findBHcloseAtoms(T,S->v[i].x,cutoff,&atom,&dist,1);
	  if (n>0) new[i]=nacc++;
	  else new[i]=-1;
	}
	if (nacc<S->nv) renumberSurface(S,new);
	free(new);
      }else err_set(1,"clipSurfaceAtoms: failed to allocate scratch data.");
    }else err_set(1,"clipSurfaceAtoms: failed to generate BH tree.");
    freeBHtree(T);
  }
}

void clipSurfaceLargest(Surface *S)
{
  int i,j,np,P,*p=NULL,*r=NULL,nu,s;

  /* retain only the largest connected fragment (measured by the number of points) */

  if (S->nt>0) {
    p=malloc(3*S->nv*sizeof(int));
    if (p) {
      r=p+S->nv;
      for (i=0;i<S->nv;i++) p[i]=i;
      for (i=0;i<S->nt;i++) {
	for (P=p[p[S->t[i].v[0]]],j=1;j<3;j++) {
	  if (p[p[S->t[i].v[j]]]<P) P=p[p[S->t[i].v[j]]];
	}
	for (j=0;j<3;j++) {
	  p[p[S->t[i].v[j]]]=P;
	  p[S->t[i].v[j]]=P;
	}
      }
      for (np=0,i=0;i<S->nv;i++) {
	p[i]=p[p[i]];
	if (p[i]==i) p[i]=np++;
      }
      for (i=0;i<np;i++) r[i]=0;
      for (i=0;i<S->nv;i++) r[p[i]]++;     /* fragment size is now in r */        
      for (P=-1,s=0,i=0,nu=0;i<np;i++) {
	if (r[i]>2) nu++;
	if (r[i]>s) {
	  s=r[i];
	  P=i;
	}
      }
      prtGraphics("%d fragments.\n",nu);
      for (nu=0,i=0;i<S->nv;i++) {         /* P is the largest fragment */
	if (p[i]==P) {
	  r[i]=nu++;
	}else r[i]=-1;
      }
      if (nu<S->nv) renumberSurface(S,r);
      free(p);
    }else err_set(1,"clipSurfaceLargest: failed to accocate scratch data.");
  }
}

static void renumberSurface(Surface *S,int *new)
{
  int i,j,nacc;

  for (nacc=0,i=0;i<S->nv;i++) {
    if (new[i]>=0) {
      if (nacc<i) S->v[nacc]=S->v[i];
      nacc++;
    }
  }
  S->nv=nacc;
  for (nacc=0,i=0;i<S->nt;i++) {
    for (j=0;j<3;j++) {
      S->t[i].v[j]=new[S->t[i].v[j]];
      if (S->t[i].v[j]<0) break;
    }
    if (j==3) {
      if (nacc<i) S->t[nacc]=S->t[i];
      nacc++;
    }
  }
  S->nt=nacc;
  
}
