#include <math.h>
#include <ctype.h>
#include "mystdlib.h"
#include "object.h"
#include "errhand.h"
#include "strutil.h"
#include "swdef.h"

#define PARSE 1
#define DONE 2
#define ERROR 3

#define MX_SWILES_NAT    250
#define MX_SWILES_NBND   250
#define MX_SWILES_RINGC   10
#define MX_SWILES_LPREN   10
#define MX_SWILES_ELTS    12
#define MX_SWILES_BTYPES   5

static char *elt[MX_SWILES_ELTS] =
{  "R",  "X", "C", "H", "N" ,"O", "S", "P", "Cl", "Br", "F", "I" };

static int nve[MX_SWILES_ELTS] =
{   63,   63,   4,   1,   5 ,  6,   6,   5,    7,    7,   7,   7 };

static int nchars_in_elt_code(char *s,int brack);
static int find_elt(char *s);

static float bval[MX_SWILES_BTYPES] = { 0, 1, 2, 3, 1.5 };
static int   arom[MX_SWILES_BTYPES] = { 0, 0, 0, 0, 1   };

/*-------------------- swilesPrule --------------------*/

Prule *swilesPrule(char *swiles,int interp)
{
  /* parse swiles string and convert to Prule structure */
  /* interp: default interpretation (hard/soft/substructure) */
  /* first character overrules default interpretation:   */
  /*    '=' : hard, '~' : soft, '?' : substructure. */

  Prule *r=(Prule *)NULL;
  int i,j,k,nat,cur,prv,nbnd,state,rc,ns,e,nar,nadd,naradd,nlp,mlp,noct,bt=0;
  int ringc[MX_SWILES_RINGC];
  int lpren[MX_SWILES_LPREN],prenlev,brack;
  float v;
  char c;
  char *nxtc,*def;

  PrAtom atom[MX_SWILES_NAT];
  PrBond bond[MX_SWILES_NBND];

  if (!swiles) return((Prule *)NULL);
  if (strlen(swiles)==0) return((Prule *)NULL); 

  nat=0;
  cur=nat;
  nbnd=0;
  prv=-1;
  prenlev=0;
  for (k=0;k<MX_SWILES_RINGC;k++) ringc[k]=-1;

  state=PARSE;
  nxtc=swiles;
  if (*nxtc=='~') {
    nxtc++;
    interp=PR_SOFT;
  }else if (*nxtc=='=') {
    nxtc++;
    interp=PR_HARD;
  }else if (*nxtc=='?') {
    nxtc++;
    interp=PR_SUBST;
  }
  def=nxtc;

  while (state==PARSE) {
    if (*nxtc=='[') {
      brack=TRUE;
      nxtc++;
    }else brack=0;
    if (*nxtc=='\0') break;
    if (nat>=MX_SWILES_NAT) {
      err_set(1,"swilesPrule: too many atoms in fragment");
      state=ERROR;
      break;
    }
    if (state==ERROR) break;
    atom[nat].hc='?';
    atom[nat].chrg='?';
    atom[nat].lig='?';
    atom[nat].val='?';
    atom[nat].nbnd=0;
    atom[nat].sym[0]='\0';
    atom[nat].arom=0;
    atom[nat].sc='?';
    atom[nat].q=0;
    ns=nchars_in_elt_code(nxtc,brack);
    if (ns>0) {
      for (k=0;k<ns;k++) atom[nat].sym[k]=*nxtc++;
      cur=nat;
      atom[nat].sym[k]='\0';
      if (islower(atom[nat].sym[0])!=0) {
	atom[nat].arom=1;
	atom[nat].sym[0]+='A'-'a';
      }else{
	atom[nat].arom=0;
      }
    }else{
      err_set(1,"swilesPrule: atom with illegal chem symbol");
      state=ERROR;
    }
    if (state==ERROR) break;
    while (brack) {
      switch (*nxtc++) {
      case 'H':
	if (isdigit(*nxtc)) atom[nat].hc = *nxtc++ - '0';
	else atom[nat].hc=1;
	err_set(1,"implicit H-count not allowed in swiles"); /* 26.11.01 */
	state=ERROR; brack=FALSE;
	break;
      case '+':
	atom[nat].chrg=1;
	if (isdigit(*nxtc)) atom[nat].chrg = *nxtc++ - '0';
	else while (*nxtc=='+') {atom[nat].chrg++;nxtc++;}
	break;
      case '-':
	atom[nat].chrg=-1;
	if (isdigit(*nxtc)) atom[nat].chrg = '0' - *nxtc++;
	else while (*nxtc=='-') {atom[nat].chrg--;nxtc++;}
	break;
      case ']':
	brack=FALSE;
	break;
      case '\0':
	err_set(1,"missing right bracket");
	state=ERROR; brack=FALSE;
	break;
      default:
	err_set(1,"illegal character in atom specification");
	state=ERROR; brack=FALSE;
	break;
      }
    }
    if (state==ERROR) break;
    if (prv>=0) {
      if (atom[prv].nbnd>=PR_MXBND) {
	err_set(1,"swilesPrule: too many bonds to atom\n");
	state=ERROR;
      }else if (nbnd>=MX_SWILES_NBND) {
	err_set(1,"swilesPrule: too many bonds in swiles string");
	state=ERROR;
      }else{
	if (0<=bt && bt < MX_SWILES_BTYPES) {
	  bond[nbnd].bo=bval[bt];
	  bond[nbnd].arom=arom[bt];
	}else{
	  if (atom[prv].arom&&atom[cur].arom) bond[nbnd].arom=1;
	  else {
	    bond[nbnd].bo=1;
	    bond[nbnd].arom=0;
	  }
	}
	bond[nbnd].from=prv;
	bond[nbnd].to=cur;
	bond[nbnd].cover=1;
	atom[prv].bnd[atom[prv].nbnd]=nbnd;
	atom[prv].nei[atom[prv].nbnd]=cur;
	atom[prv].nbnd++;
	atom[cur].bnd[atom[cur].nbnd]=nbnd;
	atom[cur].nei[atom[cur].nbnd]=prv;
	atom[cur].nbnd++;
	nbnd++;
	bt='?';
      }
    }
    if (state==ERROR) break;
    do {
      while (isdigit(*nxtc)) {
	rc=*nxtc++ - '0';
	if (ringc[rc]>=0) {
	  prv=ringc[rc];
	  if (atom[prv].nbnd>=PR_MXBND || atom[cur].nbnd>=PR_MXBND) {
	    err_set(1,"swilesPrule: to many bonds to atom");
	    state=ERROR;
	  }else if (nbnd>=MX_SWILES_NBND) {
	    err_set(1,"swilesPrule: too many bonds in swiles string");
	    state=ERROR;
	  }else{
	    if (bt=='?') {
	      if (atom[cur].arom && atom[prv].arom) {
		bond[nbnd].bo='?';
		bond[nbnd].arom=1;
	      }else{
		bond[nbnd].bo=1;
		bond[nbnd].arom=0;
	      }
	    }else{
	      bond[nbnd].bo=bval[bt];
	      bond[nbnd].arom=arom[bt];
	    }
	    bond[nbnd].from=prv;
	    bond[nbnd].to=cur;
	    bond[nbnd].cover=1;
	    atom[prv].bnd[atom[prv].nbnd]=nbnd;
	    atom[prv].nei[atom[prv].nbnd]=cur;
	    atom[prv].nbnd++;
	    atom[cur].bnd[atom[cur].nbnd]=nbnd;
	    atom[cur].nei[atom[cur].nbnd]=prv;
	    atom[cur].nbnd++;
	    nbnd++;
	    ringc[rc]=-1;
	    bt='?';
	  }
	}else{
	  ringc[rc]=cur;
	}
      }
      bt='?';
      prv=cur;
      while (*nxtc==')' &&  state==PARSE) {
	if (prenlev<=0) {
	  err_set(3,"swilesPrule: unmatched rpren");
	  state=ERROR;
	}else{
	  prv=lpren[--prenlev];
	  cur=prv;
	  nxtc++;
	}
      }
      while (*nxtc=='(' &&  state==PARSE) {
	if (prenlev>=MX_SWILES_LPREN) {
	  err_set(1,"swilesPrule: max pren level exceeded");
	  state=ERROR;
	}else if (prv<0) {
	  err_set(2,"swilesPrule: may not start with lpren");
	  state=ERROR;
	}else{
	  lpren[prenlev++]=prv;
	  nxtc++;
	}
      }
      if (state==ERROR) break;
      switch (*nxtc) {
      case '~':
	if (interp==PR_HARD) {
	  err_set(1,"any bond (~) not allowed in hard swiles.");
	  state=ERROR;
	}else{
	  bt=0;
	}
	nxtc++;
	break;
      case '-':
	bt=1;
	nxtc++;
	break;
      case '=':
	bt=2;
	nxtc++;
	break;
      case '#':
	bt=3;
	nxtc++;
	break;
      case ':':
	bt=4;
	nxtc++;
	break;
      default:
	break;
      }
    }while (isdigit(*nxtc)&&ringc[*nxtc-'0']>=0);
    nat++;
  }
  if (prenlev>0) {
    err_set(1,"swilesPrule: missing right parenthesis.");
    state=ERROR;
  }

  if (bt!='?') {
    err_set(1,"swilesPrule: dangling bond.");
    state=ERROR;
  }

  for (k=0;k<MX_SWILES_RINGC;k++) {
    if (ringc[k]>=0) {
      err_set(1,"swilesPrule: unmatched ring closure bond");
      state=ERROR;
      break;
    }
  }

  if (state==ERROR) {
    r=(Prule *)NULL;
    if (*nxtc=='\0') {
      err_set(1,"%s/\\",swiles);
    }else{
      c=*nxtc; *nxtc='\0';
      err_set(1,"%s/\\%c%s",swiles,c,nxtc+1);
      *nxtc=c;
    }
  }else{
    if (nat==0) {
      err_set(1,"swilesPrule: no atoms in swiles string");
    }else{
      r=(Prule *)make_node(OBJ_PRULE);
      if (mvld(r)) {
	r->def=Strdup(def);
	r->atom=(PrAtom *)malloc(nat*sizeof(PrAtom));
	if (mvld(r->atom)) {
	  r->nat=nat;
	  for (i=0;i<nat;i++) r->atom[i]=atom[i];
	  if (nbnd>0) {
	    r->bond=(PrBond *)malloc(nbnd*sizeof(PrBond));
	    if (mvld(r->bond)) {
	      r->nbnd=nbnd;
	      for (i=0;i<nbnd;i++) r->bond[i]=bond[i];
	    }else{
	      del_node((Node *)r);
	      r=(Prule *)NULL;
	    }
	  }
	}else{
	  del_node((Node *)r);
	  r=(Prule *)NULL;
	}
      }
    }
  }

  if (!mvld(r)) return(r);
  
  if (interp==PR_HARD) {
    for (i=0;i<r->nat;i++) {
      if (r->atom[i].chrg=='?') {
	if (!streq("X",r->atom[i].sym)&&!streq("R",r->atom[i].sym)) {
	  r->atom[i].chrg=0;
	}
      }
    }

    for (i=0;i<r->nat;i++) {
      e=find_elt(r->atom[i].sym);
      if (e<0) continue;
      if (nve[e]=='?') {
	r->atom[i].val='?';
	r->atom[i].sc='?';
	r->atom[i].lig='?';
	continue;
      }
      for (nar=0,v=0,j=0;j<r->atom[i].nbnd;j++) {
	if (r->bond[r->atom[i].bnd[j]].arom) nar++;
	else v+=r->bond[r->atom[i].bnd[j]].bo;
      }
      if (r->atom[i].arom) {
	if (nar < 2) naradd=2-nar;
	else naradd=0;
	v+=1.5*(nar+naradd);
      }else{
	nar=0;
	naradd=0;
      }
      if (r->atom[i].hc=='?') r->atom[i].hc=0;
      v+=r->atom[i].hc;
      nadd=nve[e]-rint(v)-r->atom[i].chrg;
      if (nve[e]>4) nlp=nve[e]-4;
      else nlp=0;
      mlp=nlp;
      while (nlp > 0 && nadd > 1) {
	nlp--; nadd-=2;
      }
      nlp=mlp-nlp;
      noct=2*rint(v)+2*nadd+2*nlp;
      while (noct < 8 && nlp > 0 ) {
	noct+=2; nlp--; nadd+=2;
      }
      while (noct > 8 && nadd >=2) {
	noct-=2; nlp++; nadd-=2;
      }
      if (nadd>0) {
	r->atom[i].lig=r->atom[i].nbnd+nadd+naradd;
      }else {
	r->atom[i].lig=r->atom[i].nbnd+naradd;
      }
      r->atom[i].val=nve[e];
      r->atom[i].sc=r->atom[i].lig-r->atom[i].nbnd+r->atom[i].hc;
    }
  }  
  if (r) r->interp=interp;
  return(r);
}
/*-------------------- matchPrule --------------------*/

#define LIG(x)  ((x)->lig  != '?' ? (x)->lig  : ((x)->lig  = sane_formlig(x)))
#define CHRG(x) ((x)->chrg != '?' ? (x)->chrg : ((x)->chrg = asg_formchrg(x,LIG(x))))

int matchPrule(Atom *atom,Bond *bnd,Prule *pr,int amask,int bmask)
{
  /* return TRUE iff atom a occurs as first atom in fragment defined 
     by prule pr . If bnd!=NULL, bnd must match the first bond in the rule.

     amask: use only atoms with a->flg&amask == amask (bits set in amask must be set in a->flg)
     bmask: bit mask to use for 'covering' bonds.

     NOTE: This function may assign automatic ligance and charge ... sad ! */

  Atom **a=(Atom **)NULL,*t;
  Bond **b=(Bond **)NULL;
  int  i,j,r,it,*nb=(int *)NULL,*alev=(int *)NULL,*blev=(int *)NULL;

  r=0;

  if (!is_prule(pr)) return(r);
  if (!is_atom(atom)) return(r);
  if (amask!=(amask&atom->flg)) return(r);

  if (pr->nat<1) return(r);
  if (bnd&&pr->nbnd<1) return(r);

  if (!mvld(atom->atp)) return(r);
  if (!streq(pr->atom[0].sym,"R")) {
    if (!streq(pr->atom[0].sym,"X")) {
      if (!streq(pr->atom[0].sym,atom->atp->chs)) return(r);
    }else{
      if (atom->atp->z==1) return(r);
    }
  }
  if (pr->interp==PR_HARD) {
    if (pr->atom[0].arom&&!atom->arom) return(r);
    if (pr->atom[0].lig!='?'&&pr->atom[0].lig!=LIG(atom)) return(r);
    if (pr->atom[0].chrg!='?'&&pr->atom[0].chrg!=CHRG(atom)) return(r);
    if (pr->atom[0].val!='?'&&pr->atom[0].val!=atom->val) return(r);
  }else if (pr->interp==PR_SOFT) {
    if (pr->atom[0].nbnd!=atom->nbnd&&
	!streq(pr->atom[0].sym,"X")&&
	!streq(pr->atom[0].sym,"R")) return(r);
  }else if (pr->interp==PR_SUBST) {
    if (pr->atom[0].arom&&!atom->arom) return(r);
    if (pr->atom[0].chrg!='?'&&pr->atom[0].chrg!=CHRG(atom)) return(r);
  }

  a=(Atom **)malloc(pr->nat*sizeof(Atom *));
  alev=(int *)malloc(pr->nat*sizeof(int));
  if (pr->nbnd>0) {
    b=(Bond **)malloc(pr->nbnd*sizeof(Bond *));
    nb=(int *)malloc(pr->nbnd*sizeof(int));
    blev=(int *)malloc(pr->nbnd*sizeof(int));
  }
  if (!a||!alev||(pr->nbnd>0&&(!b||!nb||!blev))) {
    err_set(1,"matchPrule: memory allocation failed.");
    if(a) free(a);
    if (alev) free(alev);
    if (b) free(b);
    if (blev) free(blev);
    if (nb) free(nb);
    return(r);
  }

  a[0]=atom; alev[0]=-1; atom->flg|=(1<<1);

  for (i=1;i<pr->nat;i++) {
    a[i]=(Atom *)NULL;
    alev[i]=-1;
  }
  for (j=0;j<pr->nbnd;j++) {
    b[j]=(Bond *)NULL;
    nb[j]=-1;
    blev[j]=-1;
  }
  bmask|=(1<<1);
  for (i=0;i<pr->nbnd&&i>=0;i++) {
    for (nb[i]++;nb[i]<a[pr->bond[i].from]->nbnd;nb[i]++) {
      if (i==0&&bnd) {                        /* bnd is the only bond tried for i==0 ... */
	b[i]=bnd;
	nb[i]=a[pr->bond[i].from]->nbnd-1;
      }else b[i]=a[pr->bond[i].from]->bnd[nb[i]];
      if (b[i]->flg&(pr->bond[i].cover?bmask:(1<<1))) continue;
      if (pr->interp==PR_HARD||(pr->interp==PR_SUBST&&pr->bond[i].bo!=0)) {
       if (b[i]->arom != pr->bond[i].arom) continue;
       if (b[i]->bo != pr->bond[i].bo && !b[i]->arom) continue;
      }
      t=(b[i]->f == a[pr->bond[i].from] ? b[i]->t : b[i]->f);
      it=pr->bond[i].to;
      if (a[it]) { /* ring closure bond ... */
	if (a[it]!=t) continue;
      }else{
	if (t->flg&(1<<1)) continue;
	if (amask!=(amask&t->flg)) continue;
	if (!streq(pr->atom[it].sym,"R")) {
	  if (!streq(pr->atom[it].sym,"X")) {
	    if (!streq(pr->atom[it].sym,t->atp->chs)) continue;
	  }else{
	    if (t->atp->z==1) continue;
	  }
	}
	if (pr->interp==PR_HARD) {
	 if (pr->atom[it].arom&&!t->arom) continue; /* not: return(r); */
	 if (pr->atom[it].lig!='?'&&pr->atom[it].lig!=LIG(t)) continue;
	 if (pr->atom[it].chrg!='?'&&pr->atom[it].chrg!=CHRG(t)) continue;
	 if (pr->atom[it].val!='?'&&pr->atom[it].val!=t->val) continue;
	}else if (pr->interp==PR_SOFT) {
	  if (pr->atom[it].nbnd!=t->nbnd&&
	      !streq(pr->atom[it].sym,"X")&&
	      !streq(pr->atom[it].sym,"R")) continue;
	}else if (pr->interp==PR_SUBST) {
	 if (pr->atom[it].arom&&!t->arom) continue; /* not: return(r); */
	 if (pr->atom[it].chrg!='?'&&pr->atom[it].chrg!=CHRG(t)) continue;
	}
	t->flg|=(1<<1);
	a[it]=t; alev[it]=i;
      }
      b[i]->flg|=(1<<1); blev[i]=i;
      break;
    }
    if (nb[i]==a[pr->bond[i].from]->nbnd) {  /* backtrack */
      nb[i]=-1;
      if (i-- > 0) {
	if (alev[pr->bond[i].to]==i) {
	  a[pr->bond[i].to]->flg&=~(1<<1);
	  a[pr->bond[i].to]=(Atom *)NULL;
	}
	if (blev[i]==i) b[i]->flg&=~(1<<1);
      }
      i--;
    }
  }
  if (i>=0) {
    for (j=0;j<pr->nbnd;j++) {
      b[j]->flg&=~(1<<1);
      pr->bond[j].match=b[j];
    }
    for (j=0;j<pr->nat;j++) {
      a[j]->flg&=~(1<<1);
      pr->atom[j].match=a[j];
    }
    r=1;
  }else{
    atom->flg&=~(1<<1);
    r=0;
  }

  if (a) free(a);
  if (alev) free(alev);
  if (b) free(b);
  if (blev) free(blev);
  if (nb) free(nb);
  return(r);
} 
/*-------------------- matchPruleAtom --------------------*/

int matchPruleAtom(Atom *atom,Prule *pr,int amask,int bmask)
{
  int r;
  
  r=matchPrule(atom,(Bond *)NULL,pr,amask,bmask);
  return(r);
}
/*-------------------- matchPruleBond --------------------*/

int matchPruleBond(Bond *bnd,Prule *pr,int amask,int bmask)
{
  int r;

  if (bnd) {
    r=matchPrule(bnd->f,bnd,pr,amask,bmask);
    if (!r) r=matchPrule(bnd->t,bnd,pr,amask,bmask); 
  }else r=0;
  return(r);
}
/*-------------------- nchars_in_elt_code --------------------*/

static int nchars_in_elt_code(char *s,int brack)
{
  int r;

  r=0;

  if (brack) {
    if (islower(*s)||isupper(*s)) {
      if (islower(*(s+1))) r=2;
      else r=1;
    }
  }else{
    if (*s=='B' && *(s+1)=='r') r=2;
    else if (*s=='C' && *(s+1)=='l') r=2;
    else if (strchr("RXBCNOPSFHIxcn",*s)) r=1;
  }

  return(r);
}
/*-------------------- find_elt --------------------*/

static int find_elt(char *s)
{
  int i,r;
  
  r=-1;

  for (i=0;i<MX_SWILES_ELTS;i++) {
    if (streq(s,elt[i])) {
      r=i;
      break;
    }
  }
  return(r);
}
