#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/param.h>
#include <math.h>
#include "swdef.h"
#include "transpor.h"
#include "parser.h"
#include "mystdlib.h"
#include "object.h"
#include "errhand.h"
#include "graphics.h"
#include "gcalc.h"
#include "strutil.h"
#include "list.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif

extern int graphix;

#define TAFF_FORMAT_XPLOR 1
#define TAFF_FORMAT_NAMD  2

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))

typedef struct TAFFAtomdef {
  float mass,rvdw,eps;
  int neq,hbacc,hbdon,z,nused;
  char name[8],eqname[8];
}TAFFAtomdef;

static int NAtomDefs;
static TAFFAtomdef *AtomDefs;

typedef struct TAFFBondstretch {
  float k,r;
  int t1,t2,bt;
}TAFFBondstretch;

typedef struct EBond {
  int f,t,t1,t2,bt,idx;
}Ebond;

static int NBondStretch;
static TAFFBondstretch *BondStretch;

typedef struct TAFFAngle {
  float k,a;
  int t1,t0,t2;
}TAFFAngle;

typedef struct Eangle {
  int a1,a0,a2,t1,t0,t2,idx;
}Eangle;

static int NAngle;
static TAFFAngle *Angle;

typedef struct TAFFOop {
  float k;
  int t;
}TAFFOop;

static int NOop;
static TAFFOop *Oop;

typedef struct TAFFTors {
  float k,a;
  int t0,t1,t2,t3,bt,n;
}TAFFTors;

typedef struct Etors {
  int a0,a1,a2,a3,t0,t1,t2,t3,t4,bt,idx;
}Etors;

static int NTors;
static TAFFTors *Tors;

static int ReadAtomDefs(void);
static int ReadBondStretch(void);
static int FindAtomDef(int n,TAFFAtomdef *def,char *name);
static int FindBondStretch(int t1,int t2,int bt);
static int FindAngle(int t1,int t0,int t2);
static int FindTors(int t0,int t1,int t2,int t3,int bt);
static int FindOop(int t1);
static int ReadAngle(void);
static int ReadTors(void);
static int ReadOop(void);
static int typeq(int a,int b);
static int bond_sort(const void *a,const void *b);
static void bond_assign(int n,Ebond *bonds);
static int angle_sort(const void *a,const void *b);
static void angle_assign(int n,Eangle *angles);
static int tors_sort(const void *a,const void *b);
static void tors_assign(int n,Etors *tors);
static int oop_sort(const void *a,const void *b);
static void oop_assign(int n,Etors *impr);

/*========================================================================*/

#define sixth2 1.122462  /* 2^1/6 */
#define hbscale 0.7      /* scale for vdw in HBonds */
#define oopscl  0.015    /* conversion of TAFF oop to X-Plor */

/*------------------------------------------------------------------------*/

static float W_bond=1.0;
static float W_angle=1.0;
static float W_plane=1.0;
static float W_torsion=1.0;
static int   F_format=TAFF_FORMAT_XPLOR;

void setTAFFweight(char *term,float w)
{
  w=(w<0 ? 0 : w);
  if (streq(term,"bond")) W_bond=w;
  else if (streq(term,"angle")) W_angle=w;
  else if (streq(term,"plane")) W_plane=w;
  else if (streq(term,"torsion")) W_torsion=w;
}

void setTAFFformat(char *format)
{
  if (streq(format,"X-Plor")) F_format=TAFF_FORMAT_XPLOR;
  else if (streq(format,"NAMD")) F_format=TAFF_FORMAT_NAMD;
}

void cmd_taff(void)
{
  Token *key;
  Node *m;
  char *filename;
  
  filename=(char *)NULL;
  key=(Token *)NULL;
  key=Get_String("filename: ",";");
  if (key->status==TOK_OK) {
    filename=(char *)key->text;
    key->text=(char *)NULL;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad filename token: %s",key->text);
    key->status=TOK_ABORT;
  }
  if (key->status==TOK_OK) {
    key=free_Token(key);
    key=Get_Mol_D("molecule: ",";");
    if (key->status==TOK_OK) {
      m=(Node *)key->ptr;
      write_taff_xplor(filename,m);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad molecule selection: %s",key->text);
    }
  }
  if (mvld(filename)) free(filename);
  free_Token(key);
}
/*------------------------------------------------------------------------*/

int write_taff_xplor(char *filename,Node *m)
{
  int i,j,l1,l2,k,l,h,n,nerr,nat,nbnd,nang,ntors,nimpr,*types,status;
  int uf_bnd,uf_ang,uf_tors,uf_oop;
  float R,eps,sigma,A,A14,B,B14,hbs6,hbs12,d;
  Nodechain ac;
  Atom **Atoms,*a;
  Ebond *bonds;
  Eangle *angles;
  Etors *tors,*impr;
  char buf[MAXPATHLEN],**sel,**atype;
  FILE *out;

  AtomDefs=(TAFFAtomdef *)NULL;
  BondStretch=(TAFFBondstretch *)NULL;
  Angle=(TAFFAngle *)NULL;
  Tors=(TAFFTors *)NULL;
  Oop=(TAFFOop *)NULL;
  Atoms=(Atom **)NULL; nat=0;
  sel=(char **)NULL;
  atype=(char **)NULL;
  types=(int *)NULL;

  NAtomDefs = ReadAtomDefs();
  if (NAtomDefs==0) {
    err_set(1,"Failed to read TAFF parameters.");
    return(1);
  }

  uf_bnd=0;uf_ang=0;uf_tors=0;uf_oop=0;
  status=0;
  NBondStretch=ReadBondStretch();
  NAngle=ReadAngle();
  NTors=ReadTors();
  NOop=ReadOop();

  ac=tree2chain(m,OBJ_ATOM);
  nat=ac.n;
  if (nat<1) {
    err_set(1,"zero atoms in selected moolecule.");
    status=2;
  }

  if (!status) {
    Atoms=(Atom **)chain_to_ptr(&ac,FALSE);
    clear_node_chain(&ac);
    types=(int *)malloc(nat*sizeof(int));
    sel=(char **)malloc(nat*sizeof(char *));
    if (mvld(sel)) for(i=0;i<nat;i++) sel[i]=(char *)NULL;
    atype=(char **)malloc(nat*sizeof(char *));
    if (mvld(atype)) for(i=0;i<nat;i++) atype[i]=(char *)NULL;
    if (!mvld(types)||!mvld(Atoms)||!mvld(sel)||!mvld(atype)) {
      err_set(1,"memeory allocation failed.");
      status=3;
    }
  }

  if (!status) {
    for (nerr=0,i=0;i<nat;i++) {
      Atoms[i]->seqnr=i;
      Atoms[i]->flg|=(1<<1);
      if (mvld(Atoms[i]->atp)) {
	types[i]=FindAtomDef(NAtomDefs,AtomDefs,Atoms[i]->atp->name);
	if (types[i]<0||types[i]>=NAtomDefs) {
	  err_set(1,"missing atom type definition in TAFF.");
	  nerr++;
	}else{
	  AtomDefs[types[i]].nused++;
	}
      }else{
	err_set(1,"missing atom type definition in TAFF.");
	nerr++;
	status=4;
      }
    }
    if (nerr>0) {
      for (i=0;i<nat;i++) Atoms[i]->flg&=~(1<<1);
    }
  }

  if (!status) {
    for (i=0;i<nat;i++) {
      sel[i]=Strdup(xpl_sel(buf,Atoms[i],m));
      if (AtomDefs[types[i]].z==1) {
	for (k=0;k<Atoms[i]->nbnd;k++) {
	  a=neighbour(k,Atoms[i]);
	  if (a->flg&(1<<1)) {
	    if (AtomDefs[types[a->seqnr]].hbdon) Atoms[i]->flg|=(1<<2);
	  }
	}
      }
      if (Atoms[i]->flg&(1<<2)) atype[i]=Strdup(xplr_hb_type(AtomDefs[types[i]].name));
      else  atype[i]=Strdup(xplr_ff_type(AtomDefs[types[i]].name));
    }  /* bug: should check sel[i] and atype[i], i=0...nat-1, else coredump. */
    if (write_xplor_psf(filename,m)) status=5;   /* should integrate... */
    else prtGraphics("psf-file %s written.\n",filename);
    for(i=0;i<nat;i++) Atoms[i]->seqnr=i;   /* write_xplor_psf messes up seqnr */
  }
  if (!status) {
    out=fopen(extfil(buf,filename,".prm",TRUE),"w");
    if (out) {
      fprintf(out,"{ X-Plor Parameter File, generated by WitNotP from TAFF Tables }\n\n");
      if (F_format==TAFF_FORMAT_XPLOR) fprintf(out,"\nParameter\n\n");
      fprintf(out,"  NBONDS rdiel switch eps 1 nbxmod 5 atoms end\n");
      fprintf(out,"\n  set echo=false message=off end\n\n");
      
      /* non-bonded interaction parameters... */
      
      for (i=0;i<NAtomDefs;i++) {
	if (AtomDefs[i].nused<1) continue;
	if (AtomDefs[i].eps<0.0001) continue;
	Strcpy(buf,xplr_ff_type(AtomDefs[i].name));
	sigma=(2*AtomDefs[i].rvdw)/sixth2;
	eps=AtomDefs[i].eps;
	fprintf(out,"  Nonbonded %-4s   %8.4f %8.4f %8.4f %8.4f\n",
		buf,eps,sigma,eps,sigma);
      }
      for (i=0;i<NAtomDefs;i++) {       /* fixes for HBonds... */
	if (AtomDefs[i].z!=1) continue;
	if (AtomDefs[i].nused<1) continue;
	if (AtomDefs[i].eps<0.0001) continue;
	Strcpy(buf,xplr_hb_type(AtomDefs[i].name));
	sigma=(2*AtomDefs[i].rvdw)/sixth2;
	eps=AtomDefs[i].eps;
	fprintf(out,"  Nonbonded %-4s   %8.4f %8.4f %8.4f %8.4f\n",
		buf,eps,sigma,eps,sigma);
	for (k=0;k<NAtomDefs;k++) {
	  if (AtomDefs[k].nused>0 && AtomDefs[k].hbacc && AtomDefs[k].eps>0.0001) {
	    eps=sqrt(AtomDefs[i].eps*AtomDefs[k].eps);
	    R=(AtomDefs[i].rvdw+AtomDefs[k].rvdw);
	    hbs6=pow(hbscale,6); hbs12=hbs6*hbs6;
	    A14=eps*pow(R,12);  A=A14*hbs12;
	    B14=2*eps*pow(R,6); B=B14*hbs6;
	    fprintf(out,"  Nbfix %-4s  %-4s  %10.3f  %10.3f  %10.3f  %10.3f\n",
		    buf,xplr_ff_type(AtomDefs[k].name),A,B,A14,B14);
	  }
	}
      }
      fprintf(out,"\n");
      
      /* bond stretch parameters (type based and atom based) */
      
      for (nbnd=0,i=0;i<nat;i++) {         /* just counting... */
	for(k=0;k<Atoms[i]->nbnd;k++) {
	  a=neighbour(k,Atoms[i]);
	  if (a->flg&(1<<1)) {
	    j=a->seqnr;
	    if (j>i) nbnd++;
	  }
	}
      }
      if (nbnd>0) {
	bonds=(Ebond *)malloc(nbnd*sizeof(Ebond));
	if (mvld(bonds)){
	  for (nbnd=0,i=0;i<nat;i++) {         /* just collecting... */
	    for(k=0;k<Atoms[i]->nbnd;k++) {
	      a=neighbour(k,Atoms[i]);
	      if (a->flg&(1<<1)) {
		j=a->seqnr;
		if (j>i) {
		  bonds[nbnd].f=i;
		  bonds[nbnd].t=j;
		  if (types[i]<=types[j]) {
		    bonds[nbnd].t1=types[i];
		    bonds[nbnd].t2=types[j];
		  }else{
		    bonds[nbnd].t1=types[j];
		    bonds[nbnd].t2=types[i];
		  }
		  bonds[nbnd].bt=Atoms[i]->bnd[k]->bt;
		  bonds[nbnd].idx=NBondStretch;
		  nbnd++;
		}
	      }
	    }
	  }
	  qsort(bonds,nbnd,sizeof(Ebond),bond_sort);
	  bond_assign(nbnd,bonds);
	  for (k=0;k<nbnd;k++) {            /* write parameters to file */
	    if (bonds[k].idx<0) {
	      l=-(bonds[k].idx+1);
	      if (l<NBondStretch) {
		i=bonds[k].f; j=bonds[k].t;
		fprintf(out,"  Bond %s\n     %s %10.3f %7.3f\n",
			sel[i],sel[j],0.5*W_bond*BondStretch[l].k,BondStretch[l].r);
	      }else uf_bnd++;
	    }else if (bonds[k].idx<NBondStretch) {
	      l=bonds[k].idx;
	      i=bonds[k].f; j=bonds[k].t;
	      fprintf(out,"  Bond %-4s %-4s %10.3f %7.3f\n",
		      atype[i],atype[j],0.5*W_bond*BondStretch[l].k,BondStretch[l].r);
	      
	    }
	  }
	  if (uf_bnd>0) err_set(1,"%d unfound bond parameters.",uf_bnd);
	  free(bonds);
	  fprintf(out,"\n");
	}else{
	  err_set(1,"memory allocation failed: bonds.");
	}
      }
      
      /* do angles, always type based */
      
      for (nang=0,i=0;i<nat;i++) {    /* count angles in system */
	for (k=0;k<Atoms[i]->nbnd-1;k++) {
	  a=neighbour(k,Atoms[i]);
	  if (!(a->flg&(1<<1))) continue;
	  for (l=k+1;l<Atoms[i]->nbnd;l++) {
	    a=neighbour(l,Atoms[i]);
	    if (a->flg&(1<<1)) nang++;
	  }
	}
      }
      
      if (nang>0) {
	angles=(Eangle *)malloc(nang*sizeof(Eangle));
	if (mvld(angles)) {
	  for (nang=0,i=0;i<nat;i++) {           /* collect angles in system */
	    for (k=0;k<Atoms[i]->nbnd-1;k++) {
	      a=neighbour(k,Atoms[i]);
	      if (!(a->flg&(1<<1))) continue;
	      l1=a->seqnr;
	      for (l=k+1;l<Atoms[i]->nbnd;l++) {
		a=neighbour(l,Atoms[i]);
		if (!(a->flg&(1<<1))) continue;
		l2=a->seqnr;
		angles[nang].a0=i;
		angles[nang].a1=l1;
		angles[nang].a2=l2;
		if (types[l1]>types[l2]) {
		  angles[nang].t1=types[l2];
		  angles[nang].t2=types[l1];
		}else{
		  angles[nang].t1=types[l1];
		  angles[nang].t2=types[l2];
		}
		angles[nang].t0=types[i];
		angles[nang].idx=NAngle;
		nang++;
	      }
	    }
	  }
	  qsort(angles,nang,sizeof(Eangle),angle_sort);
	  angle_assign(nang,angles);
	  for (i=0;i<nang;i++) {           /* write parameters to file */
	    if (angles[i].idx==-NAngle) uf_ang++;
	    else if (angles[i].idx<NAngle){
	      l=angles[i].idx;
	      fprintf(out,"  Angle %-4s %-4s %-4s  %10.3f %7.3f\n",
		      atype[angles[i].a1],atype[angles[i].a0],atype[angles[i].a2],
		      0.5*3282.8064*W_angle*Angle[l].k,Angle[l].a);
	    }
	  }
	  if (uf_ang>0) err_set(1,"%d unfound angle parameters.",uf_ang);
	  free(angles);
	  fprintf(out,"\n");
	}else{
	  err_set(1,"memory allocation failed: angles.");
	}
      }
      
      /* do torsions, type based and atom based */
      
      for (ntors=0,i=0;i<nat;i++) {    /* count dihedrals in system */
	for (k=0;k<Atoms[i]->nbnd;k++) {
	  a=neighbour(k,Atoms[i]);
	  if (!(a->flg&(1<<1))) continue;
	  j=a->seqnr; if (j<=i) continue;
	  for (l=0;l<Atoms[i]->nbnd;l++) {
	    a=neighbour(l,Atoms[i]);
	    if (!(a->flg&(1<<1))) continue;
	    l1=a->seqnr; if (l1==j) continue;
	    for (h=0;h<Atoms[j]->nbnd;h++) {
	      a=neighbour(h,Atoms[j]);
	      if (!(a->flg&(1<<1))) continue;
	      l2=a->seqnr; if (l2==i || l2==l1) continue;
	      ntors++;
	    }
	  }
	}
      }
      if (ntors>0) {
	tors=(Etors *)malloc(ntors*sizeof(Etors));
	if (mvld(tors)) {
	  for (ntors=0,i=0;i<nat;i++) {    /* collect dihedrals in system */
	    for (k=0;k<Atoms[i]->nbnd;k++) {
	      a=neighbour(k,Atoms[i]);
	      if (!(a->flg&(1<<1))) continue;
	      j=a->seqnr; if (j<=i) continue;
	      for (l=0;l<Atoms[i]->nbnd;l++) {
		a=neighbour(l,Atoms[i]);
		if (!(a->flg&(1<<1))) continue;
		l1=a->seqnr; if (l1==j) continue;
		for (h=0;h<Atoms[j]->nbnd;h++) {
		  a=neighbour(h,Atoms[j]);
		  if (!(a->flg&(1<<1))) continue;
		  l2=a->seqnr; if (l2==i || l2==l1) continue;
		  tors[ntors].a0=l1; tors[ntors].a1=i;
		  tors[ntors].a2=j; tors[ntors].a3=l2;
		  if (types[i]<types[j]) {
		    tors[ntors].t0=types[l1]; tors[ntors].t3=types[l2];
		    tors[ntors].t1=types[i];  tors[ntors].t2=types[j];
		  }else if (types[i]>types[j]) {
		    tors[ntors].t0=types[l2]; tors[ntors].t3=types[l1];
		    tors[ntors].t1=types[j];  tors[ntors].t2=types[i];
		  }else if (types[l1]<types[l2]) {
		    tors[ntors].t0=types[l1]; tors[ntors].t3=types[l2];
		    tors[ntors].t1=types[i];  tors[ntors].t2=types[j];
		  }else{
		    tors[ntors].t0=types[l2]; tors[ntors].t3=types[l1];
		    tors[ntors].t1=types[j];  tors[ntors].t2=types[i];
		  }
		  tors[ntors].bt=Atoms[i]->bnd[k]->bt;
		  ntors++;
		}
	      }
	    }
	  }
	  qsort(tors,ntors,sizeof(Etors),tors_sort);
	  tors_assign(ntors,tors);
	  for (i=0;i<ntors;i++) {         /* write parameters to file */
	    if (tors[i].idx==-NTors-1) uf_tors++;
	    if (tors[i].idx<0) l=-(tors[i].idx+1);
	    else l=tors[i].idx;
	    if (l<NTors) {
	      if (Tors[l].n<0) {
		n=-Tors[l].n;
		d=180.0;
	      }else{
		n=Tors[l].n;
		d=0.0;
	      }
	      if (tors[i].idx<0) {
		fprintf(out,"  Dihed %-4s %-4s\n      %-4s %-4s\n"
			"         %7.3f %2d %5.1f\n",
			sel[tors[i].a0],sel[tors[i].a1],sel[tors[i].a2],sel[tors[i].a3],
			0.5*W_torsion*Tors[l].k,n,d);
	      }else{
		fprintf(out,"  Dihed %-4s %-4s %-4s %-4s  %7.3f %2d %5.1f\n",
			atype[tors[i].a0],atype[tors[i].a1],atype[tors[i].a2],atype[tors[i].a3],
			0.5*W_torsion*Tors[l].k,n,d);
	      }
	    }
	  }
	  if (uf_tors>0) err_set(1,"%d unfound dihedral parameters.",uf_tors);
	  free(tors);
	  fprintf(out,"\n");
	}else{
	  err_set(1,"memory allocation failed: dihedrals.");
	}
      }
      
      /*  out-of-plane via impropers   */
      
      for (nimpr=0,i=0;i<nat;i++) {    /* counting... */
	if (Atoms[i]->nbnd==3) {
	  l=FindOop(types[i]);
	  if (l<NOop && l>-1) {
	    for (k=0;k<Atoms[i]->nbnd;k++) {
	      a=neighbour(k,Atoms[i]);
	      if (!(a->flg&(1<<1))) break;
	    }
	    if (k==Atoms[i]->nbnd) nimpr+=3; /* symmetrize... */
	  }
	}
      }
      if (nimpr>0) {
	impr=(Etors *)malloc(nimpr*sizeof(Etors));
	if (mvld(impr)) {
	  for (nimpr=0,i=0;i<nat;i++) {    /* collecting... */
	    if (Atoms[i]->nbnd==3) {
	      l=FindOop(types[i]);
	      if (l<NOop && l>-1) {
		for (k=0;k<Atoms[i]->nbnd;k++) {
		  a=neighbour(k,Atoms[i]);
		  if (!(a->flg&(1<<1))) break;
		}
		if (k==Atoms[i]->nbnd) {
		  impr[nimpr].t1=types[i]; impr[nimpr].a1=i;
		  for (k=0;k<Atoms[i]->nbnd;k++) {
		    a=neighbour(k,Atoms[i]);
		    j=a->seqnr;
		    if (k==0) {impr[nimpr].t0=types[j]; impr[nimpr].a0=j;}
		    if (k==1) {impr[nimpr].t2=types[j]; impr[nimpr].a2=j;}
		    if (k==2) {impr[nimpr].t3=types[j]; impr[nimpr].a3=j;}
		  }
		  impr[nimpr].bt=0;
		  nimpr++;
		  impr[nimpr]=impr[nimpr-1];
		  impr[nimpr].t0=impr[nimpr-1].t2; impr[nimpr].a0=impr[nimpr-1].a2;
		  impr[nimpr].t2=impr[nimpr-1].t3; impr[nimpr].a2=impr[nimpr-1].a3;
		  impr[nimpr].t3=impr[nimpr-1].t0; impr[nimpr].a3=impr[nimpr-1].a0;
		  nimpr++;
		  impr[nimpr]=impr[nimpr-1];
		  impr[nimpr].t0=impr[nimpr-1].t2; impr[nimpr].a0=impr[nimpr-1].a2;
		  impr[nimpr].t2=impr[nimpr-1].t3; impr[nimpr].a2=impr[nimpr-1].a3;
		  impr[nimpr].t3=impr[nimpr-1].t0; impr[nimpr].a3=impr[nimpr-1].a0;
		  nimpr++;
		}
	      }
	    }
	  }
	  qsort(impr,nimpr,sizeof(Etors),oop_sort);
	  oop_assign(nimpr,impr);
	  for (i=0;i<nimpr;i++) {
	    l=impr[i].idx;
	    if (l==-NOop) uf_oop++;
	    else if (l<NOop) {
	      fprintf(out,"  Impr %-4s %-4s %-4s %-4s %8.4f 2 180.0\n",
		      atype[impr[i].a0],atype[impr[i].a1],atype[impr[i].a2],atype[impr[i].a3],
		      oopscl*W_plane*Oop[l].k);
	    }
	  }
	  if (uf_oop>0) err_set(1,"%d unfound out-of-plane parameters",uf_oop);
	  free(impr);
	  fprintf(out,"\n");
	}else{
	  err_set(1,"memory allocation failed: improper.");
	}
      }
      
      /*  done with parameter file...  */
      
      if (F_format==TAFF_FORMAT_XPLOR) fprintf(out,"\nend\n");
      fprintf(out,"set echo=true message=normal end\n");
      fclose(out);
      prtGraphics("prm-file %s written.\n",filename);
    }else{
      err_set(1,"could not open file %s for write.");
      status=6;
    }
  }

  if (mvld(AtomDefs)) free(AtomDefs);
  if (mvld(BondStretch)) free(BondStretch);
  if (mvld(Angle)) free(Angle);
  if (mvld(Tors))  free(Tors);
  if (mvld(Oop))   free(Oop);
  if (mvld(Atoms)) {
    for(i=0;i<nat;i++) Atoms[i]->flg&=~(1<<1|1<<2);
    free(Atoms);
  }
  if (mvld(sel)) {
    for (i=0;i<nat;i++) if (mvld(sel[i])) free(sel[i]);
    free(sel);
  }
  if (mvld(atype)) {
    for (i=0;i<nat;i++) if (mvld(atype[i])) free(atype[i]);
    free(atype);
  }
  if (mvld(types))  free(types);
  return(status);
}
/*-------------------- ReadAtomDefs --------------------*/

static int ReadAtomDefs(void)
{
  FILE *in;
  char fname[MAXPATHLEN],rec[512],*path,tok[512];
  int ndefs,i,k;
  Parse p;

  ndefs=0;
  path=getenv("MINIMAX_FF_TABLES");
  if (path) {
    sprintf(fname,"%s/ATOM_DEF",path);
    in=fopen(fname,"r");
    if (in) {
      for (ndefs=0;!feof(in);) if (fgets(rec,511,in)) ndefs++;
      if (ndefs>0) {
	AtomDefs=(TAFFAtomdef *)malloc(ndefs*sizeof(TAFFAtomdef));
	if (AtomDefs) {
	  fseek(in,0,SEEK_SET);
	  p.sep=""; p.tok=tok; p.mts=511; p.buf=rec;
	  for(k=0;;) {
	    if (!fgets(rec,511,in)) break;
	    if (strchr(rec,'#')) *strchr(rec,'#')='\0';
	    p.pos=0;
	    if (!get_token(&p)) continue;                    /*  column  */
	    Strcpy(AtomDefs[k].name,tok);                    /* 1        */
	    for (i=0;i<3;i++) get_token(&p);                 /* 2,3,4    */
	    if (!get_token(&p)) continue;		     
	    Strcpy(AtomDefs[k].eqname,tok);                  /* 5        */  
	    if (!get_token(&p)) continue;		     
	    AtomDefs[k].z=str2i(tok);                        /* 6        */             
	    for (i=0;i<4;i++) get_token(&p);                 /* 7,8,9,10 */
	    if (!get_token(&p)) continue;		     
	    AtomDefs[k].hbdon = *tok=='y'||*tok=='Y';        /* 11       */
	    if (!get_token(&p)) continue;		     
	    AtomDefs[k].hbacc = *tok=='y'||*tok=='Y';        /* 12       */           
	    for (i=0;i<2;i++) get_token(&p);                 /* 13,14    */
	    if (!get_token(&p)) continue;
	    if (!streq(tok,"-"))AtomDefs[k].mass=str2f(tok); /* 15       */
	    else AtomDefs[k].mass=0.0;
	    AtomDefs[k].rvdw=-1.0;                       
	    AtomDefs[k].eps=0.001;
	    AtomDefs[k].neq=k;
	    AtomDefs[k].nused=0;
	    k++;
	  }
	  ndefs=k;
	}else ndefs=0;
      }
      fclose(in);
    }else{
      err_set(1,"sorry, could not open %s for read.",fname);
    }
    for (i=0;i<ndefs;i++) {
      if (!streq(AtomDefs[i].name,AtomDefs[i].eqname)) {
	k=FindAtomDef(ndefs,AtomDefs,AtomDefs[i].eqname);
	if (k<ndefs) AtomDefs[i].neq=k;
      }
    }
    if (ndefs>0) {
      sprintf(fname,"%s/TAFF_VDW",path);
      in=fopen(fname,"r");
      if (in) {
	while (fgets(rec,511,in)) {
	  p.pos=0;
	  if (!get_token(&p)) continue;
	  for (k=0;k<ndefs;k++) if (streq(AtomDefs[k].name,tok)) break;
	  if (k<ndefs) {
	    if (get_token(&p)) AtomDefs[k].rvdw=str2f(tok);
	    if (get_token(&p)) AtomDefs[k].eps=str2f(tok);
	  }
	}
	fclose(in);
      }else{
	err_set(1,"sorry, could not open %s for read.",fname);
	free(AtomDefs);
	AtomDefs=(TAFFAtomdef *)NULL;
	ndefs=0;
      }
    }
  }else{
    err_set(1,"sorry, MINIMAX_FF_TABLES is not defined.");
  }
  return(ndefs);
}
/*-------------------- ReadBondStretch --------------------*/

static int ReadBondStretch(void)
{
  FILE *in;
  char fname[MAXPATHLEN],rec[512],*path,tok[512],nm1[8],nm2[8];
  int ndefs,k,t;
  Parse p;
  
  ndefs=0;
  path=getenv("MINIMAX_FF_TABLES");
  if (path) {
    sprintf(fname,"%s/TAFF_BOND_STRETCH",path);
    in=fopen(fname,"r");
    if (in) {
      for (ndefs=0;!feof(in);) if (fgets(rec,511,in)) ndefs++;
      if (ndefs>0) {
	BondStretch=(TAFFBondstretch *)malloc(ndefs*sizeof(TAFFBondstretch));
	if (BondStretch) {
	  fseek(in,0,SEEK_SET);
	  p.sep=""; p.tok=tok; p.mts=511; p.buf=rec;
	  for(k=0;;) {
	    if (!fgets(rec,511,in)) break;
	    if (strchr(rec,'#')) *strchr(rec,'#')='\0';
	    p.pos=0;
	    if (!get_token(&p)) continue;
	    Strcpy(nm1,tok);
	    if (!get_token(&p)) continue;
	    Strcpy(nm2,tok);
	    if (!get_token(&p)) continue;
	    BondStretch[k].bt=BT_number(tok);
	    if (!get_token(&p)) continue;
	    BondStretch[k].r=str2f(tok);
	    if (!get_token(&p)) continue;
	    BondStretch[k].k=str2f(tok);
	    BondStretch[k].t1=FindAtomDef(NAtomDefs,AtomDefs,nm1);
	    BondStretch[k].t2=FindAtomDef(NAtomDefs,AtomDefs,nm2);
	    if (BondStretch[k].t1>BondStretch[k].t2) {
	      t=BondStretch[k].t2;
	      BondStretch[k].t2=BondStretch[k].t1;
	      BondStretch[k].t1=t;
	    }
	    if (BondStretch[k].t1<NAtomDefs && BondStretch[k].t2<NAtomDefs) k++;
	  }
	  ndefs=k;
	}else ndefs=0;
      }
      fclose(in);
    }else{
      err_set(1,"sorry, could not open %s for read.");
    }
  }else{
    err_set(1,"sorry, MINIMAX_FF_TABLES is not defined.");
  }
  return(ndefs);
}

/*-------------------- ReadAngle --------------------*/

static int ReadAngle(void)
{
  FILE *in;
  char fname[MAXPATHLEN],rec[512],*path,tok[512];
  int ndefs,k,t;
  Parse p;
  
  ndefs=0;
  path=getenv("MINIMAX_FF_TABLES");
  if (path) {
    sprintf(fname,"%s/TAFF_ANGLE_BEND",path);
    in=fopen(fname,"r");
    if (in) {
      for (ndefs=0;!feof(in);) if (fgets(rec,511,in)) ndefs++;
      if (ndefs>0) {
	Angle=(TAFFAngle *)malloc(ndefs*sizeof(TAFFAngle));
	if (Angle) {
	  fseek(in,0,SEEK_SET);
	  p.sep=""; p.tok=tok; p.mts=511; p.buf=rec;
	  for(k=0;;) {
	    if (!fgets(rec,511,in)) break;
	    if (strchr(rec,'#')) *strchr(rec,'#')='\0';
	    p.pos=0;
	    if (!get_token(&p)) continue;
	    Angle[k].t1=FindAtomDef(NAtomDefs,AtomDefs,tok);
	    if (!get_token(&p)) continue;
	    Angle[k].t0=FindAtomDef(NAtomDefs,AtomDefs,tok);
	    if (!get_token(&p)) continue;
	    Angle[k].t2=FindAtomDef(NAtomDefs,AtomDefs,tok);
	    if (!get_token(&p)) continue;
	    Angle[k].a=str2f(tok);
	    if (!get_token(&p)) continue;
	    Angle[k].k=str2f(tok);
	    if (Angle[k].t1>Angle[k].t2) {
	      t=Angle[k].t1;
	      Angle[k].t1=Angle[k].t2;
	      Angle[k].t2=t;
	    }
	    if (Angle[k].t0<NAtomDefs && Angle[k].t1<NAtomDefs && Angle[k].t2<NAtomDefs) k++;
	  }
	  ndefs=k;
	}else ndefs=0;
      }
      fclose(in);
    }else{
      err_set(1,"sorry, could not open %s for read.",fname);
    }
  }else{
    err_set(1,"sorry, MINIMAX_FF_TABLES is not defined.");
  }
  return(ndefs);
}
/*-------------------- ReadOop --------------------*/

static int ReadOop(void)
{
  FILE *in;
  char fname[MAXPATHLEN],rec[512],*path,tok[512];
  int ndefs,k;
  Parse p;
  
  ndefs=0;
  path=getenv("MINIMAX_FF_TABLES");
  if (path) {
    sprintf(fname,"%s/TAFF_OOP_BEND",path);
    in=fopen(fname,"r");
    if (in) {
      for (ndefs=0;!feof(in);) if (fgets(rec,511,in)) ndefs++;
      if (ndefs>0) {
	Oop=(TAFFOop *)malloc(ndefs*sizeof(TAFFOop));
	if (Oop) {
	  fseek(in,0,SEEK_SET);
	  p.sep=""; p.tok=tok; p.mts=511; p.buf=rec;
	  for(k=0;;) {
	    if (!fgets(rec,511,in)) break;
	    if (strchr(rec,'#')) *strchr(rec,'#')='\0';
	    p.pos=0;
	    if (!get_token(&p)) continue;
	    Oop[k].t=FindAtomDef(NAtomDefs,AtomDefs,tok);
	    if (!get_token(&p)) continue;
	    Oop[k].k=str2f(tok);
	    if (Oop[k].t<NAtomDefs) k++;
	  }
	  ndefs=k;
	}else ndefs=0;
      }
      fclose(in);
    }else{
      err_set(1,"sorry, could not open %s for read.",fname);
    }
  }else{
    err_set(1,"sorry, MINIMAX_FF_TABLES is not defined.");
  }
  return(ndefs);
}
/*-------------------- ReadTors --------------------*/

static int ReadTors(void)
{
  FILE *in;
  char fname[MAXPATHLEN],rec[512],*path,tok[512];
  int ndefs,k;
  Parse p;
  
  ndefs=0;
  path=getenv("MINIMAX_FF_TABLES");
  if (path) {
    sprintf(fname,"%s/TAFF_TORS",path);
    in=fopen(fname,"r");
    if (in) {
      for (ndefs=0;!feof(in);) if (fgets(rec,511,in)) ndefs++;
      if (ndefs>0) {
	Tors=(TAFFTors *)malloc(ndefs*sizeof(TAFFTors));
	if (Tors) {
	  fseek(in,0,SEEK_SET);
	  p.sep=""; p.tok=tok; p.mts=511; p.buf=rec;
	  for(k=0;;) {
	    if (!fgets(rec,511,in)) break;
	    if (strchr(rec,'#')) *strchr(rec,'#')='\0';
	    p.pos=0;
	    if (!get_token(&p)) continue;
	    Tors[k].t0=FindAtomDef(NAtomDefs,AtomDefs,tok);
	    if (!get_token(&p)) continue;
	    Tors[k].t1=FindAtomDef(NAtomDefs,AtomDefs,tok);
	    if (!get_token(&p)) continue;
	    Tors[k].t2=FindAtomDef(NAtomDefs,AtomDefs,tok);
	    if (!get_token(&p)) continue;
	    Tors[k].t3=FindAtomDef(NAtomDefs,AtomDefs,tok);
	    if (!get_token(&p)) continue;
	    Tors[k].bt=BT_number(tok);
	    if (!get_token(&p)) continue;
	    Tors[k].k=str2f(tok);
	    if (!get_token(&p)) continue;
	    Tors[k].n=str2f(tok);
	    if (Tors[k].t0<NAtomDefs && Tors[k].t1<NAtomDefs && 
		Tors[k].t2<NAtomDefs && Tors[k].t3<NAtomDefs) k++;
	  }
	  ndefs=k;
	}else ndefs=0;
      }
      fclose(in);
    }else{
      err_set(1,"sorry, could not open %s for read.",fname);
    }
  }else{
    err_set(1,"sorry, MINIMAX_FF_TABLES is not defined.");
  }
  return(ndefs);
}

/*-------------------- FindAtomDef --------------------*/

static int FindAtomDef(int n,TAFFAtomdef *def,char *name)
{
  int i;
  
  if (streq("*",name)) i=-1;
  else{
    for (i=0;i<n;i++) if (streq(def[i].name,name)) break;
  }
  return(i);
}
/*-------------------- FindBondStretch --------------------*/

static int FindBondStretch(int t1,int t2,int bt)
{
  int i,r,lf,lb,l,lev;
  
  /* assume : t1 <= t2 */
  
  r=NBondStretch; lev=0;
  for (i=0;i<NBondStretch;i++) {
    if (bt==BondStretch[i].bt) {
      lf=typeq(t1,BondStretch[i].t1)+typeq(t2,BondStretch[i].t2);
      lb=typeq(t1,BondStretch[i].t2)+typeq(t2,BondStretch[i].t1);
      l=(lf>=lb ? lf : lb);
      if (l>lev) {
	r=i;
	lev=l;
      }
    }
  }
  r = ( lev > 0 ? r : NBondStretch);
  return(r);
}
/*-------------------- FindAngle --------------------*/

static int FindAngle(int t1,int t0,int t2)
{
  int i,r,lf,lb,l,lev;
  
  r=NAngle; lev=0;
  for (i=0;i<NAngle;i++) {
    lf=typeq(t1,Angle[i].t1)+typeq(t0,Angle[i].t0)+typeq(t2,Angle[i].t2);
    lb=typeq(t1,Angle[i].t2)+typeq(t0,Angle[i].t0)+typeq(t2,Angle[i].t1);
    l=(lf>=lb ? lf : lb);
    if (l>lev) {
      r=i;
      lev=l;
    }
  }
  r = ( lev > 0 ? r : -NAngle);
  return(r);
}
/*-------------------- FindTors --------------------*/

static int FindTors(int t0,int t1,int t2,int t3,int bt)
{
  int i,r,lf,lb,l,lev;
  
  r=NTors; lev=0;
  for (i=0;i<NTors;i++) {
    if (Tors[i].bt==bt) {
      lf=typeq(t0,Tors[i].t0)+typeq(t1,Tors[i].t1)+typeq(t2,Tors[i].t2)+typeq(t3,Tors[i].t3);
      lb=typeq(t3,Tors[i].t0)+typeq(t2,Tors[i].t1)+typeq(t1,Tors[i].t2)+typeq(t0,Tors[i].t3);
      l=(lf>=lb ? lf : lb);
      if (l>lev) {
	r=i;
	lev=l;
      }
    }
  }
  r = ( lev > 0 ? r : NTors);
  return(r);
}
/*-------------------- FindOop --------------------*/

static int FindOop(int t1)
{
  int i,r,l,lev;
  
  r=NOop; lev=0;
  for (i=0;i<NOop;i++) {
    l=typeq(t1,Oop[i].t);
    if (l>lev) {
      r=i;
      lev=l;
    }
  }
  r = ( lev > 0 ? r : -NOop);
  return(r);
}
/*-------------------- typeq --------------------*/

static int typeq(int a,int b)
{
  int r;
  
  if (a==b) r=100;
  else if (b==AtomDefs[a].neq) r=90;
  else if (b==-1) r=1;
  else r=-100000;
  return(r);
} 
/*-------------------- bond_sort --------------------*/

static int bond_sort(const void *a,const void *b)
{
  Ebond *A,*B;
  int r;
  
  A=(Ebond *)a;
  B=(Ebond *)b;
  
  if (A->t1>B->t1) r=1;
  else if (A->t1<B->t1) r=-1;
  else if (A->t2>B->t2) r=1;
  else if (A->t2<B->t2) r=-1;
  else if (A->bt>B->bt) r=-1;
  else if (A->bt<B->bt) r=1;
  else r=0;
  
  return(r);
}
/*-------------------- bond_assign --------------------*/

static void bond_assign(int n,Ebond *bonds)
{
  int k,l;
  
  /* assign indices to bonds: < 0 atom based , > 0 skip */
  /* assume: bond list sorted, type1 <= type2 */
  
  if (n<1) return;
  bonds[0].idx=FindBondStretch(bonds[0].t1,bonds[0].t2,bonds[0].bt);
  if (bonds[0].idx==NBondStretch) bonds[0].idx=-NBondStretch-1;
  for (l=0,k=1;k<n;k++) {
    if (bonds[k].t1==bonds[l].t1&&bonds[k].t2==bonds[l].t2) {
      if (bonds[k].bt==bonds[l].bt) {
	bonds[k].idx=(bonds[l].idx < 0 ? bonds[l].idx : NBondStretch) ;
      }else{
	bonds[k].idx=-(FindBondStretch(bonds[k].t1,bonds[k].t2,bonds[k].bt)+1);
	l=k;
      }
    }else{
      bonds[k].idx=FindBondStretch(bonds[k].t1,bonds[k].t2,bonds[k].bt); 
      if (bonds[k].idx==NBondStretch) bonds[k].idx=-NBondStretch-1;
      l=k;
    }
  }
}
/*-------------------- angle_sort --------------------*/

static int angle_sort(const void *a,const void *b)
{
  Eangle *A,*B;
  int r;
  
  A=(Eangle *)a;
  B=(Eangle *)b;
  
  if (A->t0>B->t0) r=1;
  else if (A->t0<B->t0) r=-1;
  else if (A->t1>B->t1) r=1;
  else if (A->t1<B->t1) r=-1;
  else if (A->t2>B->t2) r=1;
  else if (A->t2<B->t2) r=-1;
  else r=0;
  
  return(r);
}
/*-------------------- angle_assign --------------------*/

static void angle_assign(int n,Eangle *angles)
{
  int k,l;

  /* assign indices to angles: , always type based. */
  /* assume: bond list sorted, type1 <= type2 */

  if (n<1) return;
  angles[0].idx=FindAngle(angles[0].t1,angles[0].t0,angles[0].t2);
  for (l=0,k=1;k<n;k++) {
    if (angles[k].t1==angles[l].t1&&angles[k].t2==angles[l].t2&&angles[k].t0==angles[l].t0) {
      angles[k].idx=NAngle;
    }else{
      angles[k].idx=FindAngle(angles[k].t1,angles[k].t0,angles[k].t2);
    }
    l=k;
  }
}
/*-------------------- tors_sort --------------------*/

static int tors_sort(const void *a,const void *b)
{
  Etors *A,*B;
  int r;

  A=(Etors *)a;
  B=(Etors *)b;

  if (A->t1>B->t1) r=1;
  else if (A->t1<B->t1) r=-1;
  else if (A->t2>B->t2) r=1;
  else if (A->t2<B->t2) r=-1;
  else if (A->t0>B->t0) r=1;
  else if (A->t0<B->t0) r=-1;
  else if (A->t3>B->t3) r=1;
  else if (A->t3<B->t3) r=-1;
  else if (A->bt<B->bt) r=1;
  else if (A->bt>B->bt) r=-1;
  else r=0;

  return(r);
}
/*-------------------- tors_assign --------------------*/

static void tors_assign(int n,Etors *tors)
{
  int k,l;

  /* assign indices to dihedrals: atom or type based. */
  /* assume: bond list sorted */

  if (n<1) return;
  tors[0].idx=FindTors(tors[0].t0,tors[0].t1,tors[0].t2,tors[0].t3,tors[0].bt);
  if (tors[0].idx==NTors) tors[0].idx=-NTors-1;
  for (l=0,k=1;k<n;k++) {
    if (tors[k].t0==tors[l].t0&&tors[k].t1==tors[l].t1&&
	tors[k].t2==tors[l].t2&&tors[k].t3==tors[l].t3) {
      if (tors[k].bt==tors[l].bt) {
	tors[k].idx=(tors[l].idx < 0 ? tors[l].idx : NTors) ;
      }else{
	tors[k].idx=-(FindTors(tors[k].t0,tors[k].t1,tors[k].t2,tors[k].t3,tors[k].bt)+1);
	l=k;
      }
    }else{
      tors[k].idx=FindTors(tors[k].t0,tors[k].t1,tors[k].t2,tors[k].t3,tors[k].bt); 
      if (tors[k].idx==NTors) tors[k].idx=-NTors-1;
      l=k;
    }
  }
}
/*-------------------- oop_sort --------------------*/

static int oop_sort(const void *a,const void *b)
{
  Etors *A,*B;
  int r;

  A=(Etors *)a;
  B=(Etors *)b;

  if (A->t1>B->t1) r=1;
  else if (A->t1<B->t1) r=-1;
  else if (A->t0>B->t0) r=1;
  else if (A->t0<B->t0) r=-1;
  else if (A->t2>B->t2) r=1;
  else if (A->t2<B->t2) r=-1;
  else if (A->t3>B->t3) r=1;
  else if (A->t3<B->t3) r=-1;
  else r=0;

  return(r);
}
/*-------------------- oop_assign --------------------*/

static void oop_assign(int n,Etors *oop)
{
  int k,l;

  /* assign indices to impropers: atom or type based. */
  /* assume: list sorted */

  if (n<1) return;
  oop[0].idx=FindOop(oop[0].t1);
  for (l=0,k=1;k<n;k++) {
    if (oop[k].t0==oop[l].t0&&oop[k].t1==oop[l].t1&&
	oop[k].t2==oop[l].t2&&oop[k].t3==oop[l].t3) {
      oop[k].idx=NOop;
    }else{
      oop[k].idx=FindOop(oop[k].t1);
    }
    l=k;
  }
}

