#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "xbhtree.h"

#define mvld(p) (p)

/* Barnes - Hut Trees ? not quite, but almost...           */
/* No: more like median K-D trees, apparently invented by  */
/* Bentley: "Multidimensional Binary Search Trees Used for */
/* Associative Searching", ACM Sept. 1975 Vol 18. No 9.    */
/* This is the version that uses points, instead of atoms  */

#define MXTREE 1024

#define NSTEPS 128

static float *findxBHclosestInNode(xBHnode *node,float *x,float cutoff,float *d);
static int   findIxBHclosestInNode(xBHnode *node,float *x,float cutoff,float *d);
static int findxBHclosePointsInNode(xBHnode *node,float *x,float cutoff,
				    float **atom,float *dist,int maxn);
static float findxBHmindistInNode(xBHnode *node,float *x,float cutoff);
static int nrfindxBHclosePointsInNode(xBHnode *node,int idx,float cutoff,
				      float **atom,float *dist,int maxn);

static int CUT=xBH_CUT_MEDIAN;       /* method to determine cutting plane */

/*-------------------- GeneratexBHtree --------------------*/

xBHtree *GeneratexBHtree(int N,float **atoms,int granularity)
{
  /* 3D space-sort points into Barnes-Hut tree with given granularity */
  /* (max number of points per leaf-node) */

  xBHtree *r;
  float **p;
  int i,k;

  /* allocate tree data structure */

  r=(xBHtree *)malloc(sizeof(xBHtree));
  if (!mvld(r)) return(r);
  r->atom=(float **)NULL;
  r->nr=(int *)NULL;
  r->n=0;
  
  /* allocate root node data structure */

  r->root=(xBHnode *)malloc(sizeof(xBHnode));
  if (!mvld(r->root)) {
    freexBHtree(r);
    return((xBHtree *)NULL);
  }

  /* initialize root node data structure */

  r->root->atom=(float **)NULL;
  r->root->n=0; r->root->marked=0;
  r->root->dim=-1;
  r->root->left=(xBHnode *)NULL;
  r->root->right=(xBHnode *)NULL;

  r->n=N;
  if (r->n==0) {
    freexBHtree(r);
    return((xBHtree *)NULL);
  }
  r->atom=atoms;
  if (!mvld(r->atom)) { 
    freexBHtree(r);
    return((xBHtree *)NULL);
  }

  r->nr=(int *)malloc(N*sizeof(int));
  if (!mvld(r->nr)) {
    freexBHtree(r);
    return((xBHtree *)NULL);
  }
  for (i=0;i<r->n;i++) r->nr[i]=i;

  r->root->atom=r->atom;
  r->root->nr=r->nr;
  r->root->n=N;

  /* determine box dimension */

  p=r->root->atom;

  for (k=0;k<3;k++) {
    r->xmin[k]=p[0][k];
    r->xmax[k]=r->xmin[k];
  }

  for (i=1;i<r->root->n;i++) {
    for (k=0;k<3;k++) {
      if (r->xmin[k] > p[i][k]) r->xmin[k] = p[i][k];
      if (r->xmax[k] < p[i][k]) r->xmax[k] = p[i][k];
    }
  }

  /* start recursive 3D space sort */

  dividexBHnode(r->root,r->xmin,r->xmax,granularity);

  /* done... */

  return(r);
}
/*-------------------- find_xBHnode --------------------*/

xBHnode *findxBHnode(xBHtree *tree,float *x)
{
  /* find leaf in BH tree that contains 3D point x */

  xBHnode *r;
  int k;

  if (!mvld(tree)) return((xBHnode *)NULL);

  /* first check if point is in tree */

  for (k=0;k<3;k++) {
    if (x[k]<tree->xmin[k]) return((xBHnode *)NULL);
    if (x[k]>tree->xmax[k]) return((xBHnode *)NULL);
  }

  /* if point is in tree, locate the leaf */

  r=tree->root;
  while (mvld(r)) {
    if (r->dim<0) break;
    if (x[r->dim]<r->cut) r=r->left;
    else r=r->right;
  }

  return(r);
}
/*-------------------- freexBHtree --------------------*/

void freexBHtree(xBHtree *tree)
{
  if (mvld(tree)) {
    freexBHnode(tree->root);
    if (mvld(tree->nr)) free(tree->nr);
    free(tree);
  }
}
/*-------------------- freexBHnode --------------------*/

void freexBHnode(xBHnode *node)
{
  if (mvld(node)) {
    freexBHnode(node->left);
    freexBHnode(node->right);
    free(node);
  }
}
/*-------------------- dividexBHnode--------------------*/

void dividexBHnode(xBHnode *node,float *xmin,float *xmax,int granularity)
{
  /* does not work well for some pathological cases... */

  float cut,dx,xminl[3],xmaxl[3],xminr[3],xmaxr[3];
  int dim,i,j,k,n[NSTEPS],lm,rm,hnr;
  float *a;

  /* if nothing left to divide just return */

  if (!mvld(node)) return;
  if (granularity<1 || node->n <= granularity) return;
  if (!mvld(node->atom)) return;

  /* determine dimension along which to cut */

  dim=0;
  if (xmax[1]-xmin[1] > xmax[dim]-xmin[dim]) dim=1;
  if (xmax[2]-xmin[2] > xmax[dim]-xmin[dim]) dim=2;

  /* determine position of cutting plane */

  switch (CUT) {
  case xBH_CUT_MIDDLE:
    cut=0.5*(xmax[dim]+xmin[dim]);
    if (xmax[dim]-0.1 < cut || xmin[dim]+0.1 > cut) return;
    break;
  case xBH_CUT_MEDIAN:
    dx=(xmax[dim]-xmin[dim])/NSTEPS;
    if (dx<0.0001) return;
    for (i=0;i<NSTEPS;i++) n[i]=0;
    for (j=0;j<node->n;j++) {
      i=(node->atom[j][dim]-xmin[dim])/dx;
      if (i>=0 && i<NSTEPS) n[i]++;
    }
    for (i=1;i<NSTEPS;i++) {
      n[i]+=n[i-1];
      if (n[i]>node->n/2) break;
    }
    if (i==NSTEPS) i--;
    cut=xmin[dim]+i*dx;
    break;
  default:
    cut=0.5*(xmax[dim]+xmin[dim]);
    if (xmax[dim]-0.1 < cut || xmin[dim]+0.1 > cut) return;
    break;

  }

  /* sort atoms into left/right descendants */

  lm=0;
  rm=node->n-1;

  while (lm<rm) {
    for(;lm<node->n;lm++) if (node->atom[lm][dim]>=cut) break;
    for(;rm>=0;rm--)      if (node->atom[rm][dim]<cut) break;
    if (lm<rm) {
      a=node->atom[rm];
      node->atom[rm]=node->atom[lm];
      node->atom[lm]=a;
      hnr=node->nr[rm];
      node->nr[rm]=node->nr[lm];
      node->nr[lm]=hnr;
      rm--;
      lm++;
    }
  }

  if (lm==rm) {
    if (node->atom[rm][dim]<cut) lm++;
    else rm--;
  }

  /* create left/right descendants */

  node->left=(xBHnode *)malloc(sizeof(xBHnode));
  if (!mvld(node->left)) return;
  node->left->dim=-1;
  node->left->left=(xBHnode *)NULL;
  node->left->right=(xBHnode *)NULL;
        
  node->right=(xBHnode *)malloc(sizeof(xBHnode));
  if (!mvld(node->right)) {
    freexBHnode(node->left);
    return;
  }
  node->right->dim=-1;
  node->right->left=(xBHnode *)NULL;
  node->right->right=(xBHnode *)NULL;

  node->left->n=rm+1; node->left->marked=0;
  node->left->atom=node->atom;
  node->left->nr=node->nr;

  node->right->n=node->n-(rm+1); node->right->marked=0;
  node->right->atom=node->atom+lm;
  node->right->nr=node->nr;

  node->cut=cut;
  node->dim=dim;

  /* if descendants are coarse, cut them up... */

  if (node->left->n > granularity) {
    for (k=0;k<3;k++) {
      xminl[k]=xmin[k];
      xmaxl[k]=xmax[k];
    }
    xmaxl[dim]=cut;
    dividexBHnode(node->left,xminl,xmaxl,granularity);
  }

  if (node->right->n > granularity) {
    for (k=0;k<3;k++) {
      xminr[k]=xmin[k];
      xmaxr[k]=xmax[k];
    }
    xminr[dim]=cut;
    dividexBHnode(node->right,xminr,xmaxr,granularity);
  }

  /* done... */

  return;
}
/*-------------------- findxBHmindist --------------------*/

float findxBHmindist(xBHtree *tree,float *x)
{
  /* find minimum distance from point x to points in tree */

  float d;

  d=1.e30;

  if (!mvld(tree)) return(d);
  if (!mvld(tree->root)) return(d);
  if (tree->n<1) return(d);

  d=sqrt((x[0]-tree->atom[0][0])*(x[0]-tree->atom[0][0])+
	 (x[1]-tree->atom[0][1])*(x[1]-tree->atom[0][1])+
	 (x[2]-tree->atom[0][2])*(x[2]-tree->atom[0][2]));
  d=findxBHmindistInNode(tree->root,x,d);
  return(d);
}

/*-------------------- findxBHclosest --------------------*/

float *findxBHclosest(xBHtree *tree,float *x,float dmax,float *d)
{
  /* find atom closest to point x, accept only atoms within dmxax */

  float *a;

  a=(float *)NULL;
  if (!mvld(tree)) return(a);
  if (!mvld(tree->root)) return(a);
  if (tree->n<1) return(a);

  a=findxBHclosestInNode(tree->root,x,dmax,d);
  return(a);
}
/*-------------------- findxBHclosest --------------------*/

int findIxBHclosest(xBHtree *tree,float *x,float dmax,float *d)
{
  /* find atom closest to point x, accept only atoms within dmxax */

  int a;

  a=-1;
  if (!mvld(tree)) return(a);
  if (!mvld(tree->root)) return(a);
  if (tree->n<1) return(a);

  a=findIxBHclosestInNode(tree->root,x,dmax,d);
  return(a);
}
/*-------------------- findxBHclosestInNode --------------------*/

static float *findxBHclosestInNode(xBHnode *node,float *x,float cutoff,float *r)
{
  int j;
  float d[3],D,C;
  float *a,*b;

  a=(float *)NULL;
  if (!mvld(node)) return(a);
  if (node->n < 1) return(a);
  
  if (node->dim<0) {
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      d[0]=x[0]-node->atom[j][0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j][1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j][2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      cutoff=sqrt(D);
      C=D;
      a=node->atom[j];
      *r=cutoff;
    }
  }else{
    if (x[node->dim]<node->cut) {
      a=findxBHclosestInNode(node->left,x,cutoff,r);
      if (a) cutoff= *r;
      if (x[node->dim]>=node->cut-cutoff) {
	b=findxBHclosestInNode(node->right,x,cutoff,r);
	if (b) a=b;
      }
    }else{
      a=findxBHclosestInNode(node->right,x,cutoff,r);
      if (a) cutoff= *r;
      if (x[node->dim]<node->cut+cutoff) {
	b=findxBHclosestInNode(node->left,x,cutoff,r);
	if (b) a=b;
      }
    }
  }
  return(a);
}

/*-------------------- findxBHclosestInNode --------------------*/

static int findIxBHclosestInNode(xBHnode *node,float *x,float cutoff,float *r)
{
  int j,a,b;
  float d[3],D,C;
  
  a=-1;
  if (!mvld(node)) return(a);
  if (node->n < 1) return(a);
  
  if (node->dim<0) {
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      d[0]=x[0]-node->atom[j][0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j][1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j][2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      cutoff=sqrt(D);
      C=D;
      a=node->nr[j];
      *r=cutoff;
    }
  }else{
    if (x[node->dim]<node->cut) {
      a=findIxBHclosestInNode(node->left,x,cutoff,r);
      if (a>=0) cutoff= *r;
      if (x[node->dim]>=node->cut-cutoff) {
	b=findIxBHclosestInNode(node->right,x,cutoff,r);
	if (b>=0) a=b;
      }
    }else{
      a=findIxBHclosestInNode(node->right,x,cutoff,r);
      if (a>=0) cutoff= *r;
      if (x[node->dim]<node->cut+cutoff) {
	b=findIxBHclosestInNode(node->left,x,cutoff,r);
	if (b>=0) a=b;
      }
    }
  }
  return(a);
}

/*-------------------- findxBHclosePoints --------------------*/

int findxBHclosePoints(xBHtree *tree,float *x,float cutoff,
		       float **atom,float *dist,int maxn)
{
  int i;

  /* find points in BH tree that are close to given point x */

  if (maxn<1 || !mvld(tree) || cutoff <=0.0) return(0);
  if (!mvld(tree->root)) return(0);

  for (i=0;i<3;i++) {
    if (x[i] < tree->xmin[i] - cutoff) break;
    if (x[i] > tree->xmax[i] + cutoff) break;
  }
  if (i<3) return(0);

  return(findxBHclosePointsInNode(tree->root,x,cutoff,atom,dist,maxn));
}
/*-------------------- nrfindxBHclosePoints --------------------*/

int nrfindxBHclosePoints(xBHtree *tree,int idx,float cutoff,
		         float **atom,float *dist,int maxn)
{
  int i;

  /* non-redundant find pints in BH tree close to given point atom[idx],
     only finds points that come later in the list than atom idx */

  if (maxn<1 || !mvld(tree) || cutoff <=0.0) return(0);
  if (!mvld(tree->root)) return(0);
  if (idx<0 || idx>=tree->n) return(0);

  for (i=0;i<3;i++) {
    if (tree->atom[idx][i] < tree->xmin[i] - cutoff) break;
    if (tree->atom[idx][i] > tree->xmax[i] + cutoff) break;
  }
  if (i<3) return(0);

  return(nrfindxBHclosePointsInNode(tree->root,idx,cutoff,atom,dist,maxn));
}
/*-------------------- findBHclosePointsInNode --------------------*/

static int findxBHclosePointsInNode(xBHnode *node,float *x,float cutoff,
				    float **atom,float *dist,int maxn)
{
  int j,n;
  float d[3],D,C;

  if (!mvld(node)) return(0);
  if (maxn<0) return(0);
  if (node->n < 1) return(0);
  
  if (node->dim<0) {
    n=0;
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      d[0]=x[0]-node->atom[j][0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j][1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j][2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      if (n<maxn) {
	atom[n]=node->atom[j];
	dist[n]=sqrt(D);
	n++;
      }else{
	n++;
	break;
      }
    }
  }else{
    n=0;
    if (x[node->dim]<node->cut) {
      n+=findxBHclosePointsInNode(node->left,x,cutoff,atom,dist,maxn);
      if (x[node->dim]>=node->cut-cutoff) {
	n+=findxBHclosePointsInNode(node->right,x,cutoff,atom+n,dist+n,maxn-n);
      }
    }else{
      n+=findxBHclosePointsInNode(node->right,x,cutoff,atom,dist,maxn);
      if (x[node->dim]<node->cut+cutoff) {
	n+=findxBHclosePointsInNode(node->left,x,cutoff,atom+n,dist+n,maxn-n);
      }
    }
  }
  return(n);
}
/*-------------------- nrfindBHcloseAtomsInNode --------------------*/

static int nrfindxBHclosePointsInNode(xBHnode *node,int idx,float cutoff,
				      float **atom,float *dist,int maxn)
{
  int j,n,idl,idr,jstart;
  float *x,d[3],D,C;

  if (!mvld(node)) return(0);
  if (maxn<0) return(0);
  if (node->n < 1) return(0);
  if (idx>=node->n) return(0);
  
  x=node->atom[idx];
  if (node->dim<0) {
    n=0;
    C=cutoff*cutoff;
    jstart=(idx < 0 ? 0 : idx+1);
    for (j=jstart;j<node->n;j++) {
      d[0]=x[0]-node->atom[j][0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j][1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j][2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      if (n<maxn) {
	atom[n]=node->atom[j];
	dist[n]=sqrt(D);
	n++;
      }else{
	n++;
	break;
      }
    }
  }else{
    n=0;
    if (x[node->dim]<node->cut+cutoff) {
      idl=idx;
      n+=nrfindxBHclosePointsInNode(node->left,idl,cutoff,atom,dist,maxn);
    }
    if (x[node->dim]>=node->cut-cutoff) {
      idr=idx-node->left->n;
      n+=nrfindxBHclosePointsInNode(node->right,idr,cutoff,atom+n,dist+n,maxn-n);
    }
  }
  return(n);
}
/*-------------------- findxBHmindistInNode --------------------*/

static float findxBHmindistInNode(xBHnode *node,float *x,float cutoff)
{
  int j;
  float d[3],D,C;

  if (!mvld(node)) return(cutoff);
  if (node->n < 1) return(cutoff);
  
  if (node->dim<0) {
    C=cutoff*cutoff;
    for (j=0;j<node->n;j++) {
      d[0]=x[0]-node->atom[j][0];
      if (d[0]>cutoff || d[0]< -cutoff) continue;
      d[1]=x[1]-node->atom[j][1];
      if (d[1]>cutoff || d[1]< -cutoff) continue;
      d[2]=x[2]-node->atom[j][2];
      if (d[2]>cutoff || d[2]< -cutoff) continue;
      D=d[0]*d[0]+d[1]*d[1]+d[2]*d[2];
      if (D>C) continue;
      cutoff=sqrt(D);
      C=D;
    }
  }else{
    if (x[node->dim]<node->cut) {
      cutoff=findxBHmindistInNode(node->left,x,cutoff);
      if (x[node->dim]>=node->cut-cutoff) {
	cutoff=findxBHmindistInNode(node->right,x,cutoff);
      }
    }else{
      cutoff=findxBHmindistInNode(node->right,x,cutoff);
      if (x[node->dim]<node->cut+cutoff) {
	cutoff=findxBHmindistInNode(node->left,x,cutoff);
      }
    }
  }
  return(cutoff);
}
/*-------------------- set_bh_cut_method --------------------*/

int set_xbh_cut_method(int m)
{
  int r;

  r=CUT;
  
  switch (m) {
  case xBH_CUT_MIDDLE:
  case xBH_CUT_MEDIAN:
    CUT=m;
    break;
  default:
    break;
  }

  return(r);
}
