#include <stdio.h>
#include <string.h>
#include "swdef.h"
#include "mystdlib.h"
#include "strutil.h"
#include "errhand.h"
#include "utlsock.h"
#include "parser.h"

#define NAPTIME 20000   /* in micro-seconds */

#define key_is(a) (streq(a,key->key))
static int xmlrpcport=0;

#ifdef withRPC
#include <pthread.h>
#include "rpcmethods.h"
#include <xmlrpc.h>
#include <xmlrpc_server.h>
#include <xmlrpc_abyss.h>

#ifndef MAX_CONN
#define MAX_CONN 16
#endif

static pthread_t xmlrpcthread=0;
static xmlrpc_env env;
static int socketFd;
static TSocket *socketP;
static TServer abyssServer;
static pthread_mutex_t rpcMutex;

void wnpRPCclose(pthread_t thread);
static void *wnpRPCserver(void *arg);
static void wnpRPCcleanup(void *arg);

static void *wnpRPCserver(void *arg)
{
  /* executed by XML-RPC server thread */

  const char *error;
  int port;

  xmlrpc_registry *registryP;
  
  xmlrpc_env_init(&env);
  registryP=xmlrpc_registry_new(&env);

  register_rpcmethods(&env,registryP);

  port=*(int *)arg;
  socketFd=open_socket_on_port(port);
  if (!socketFd) {
    err_dump();
    xmlrpcthread=0;
    xmlrpcport=0;
    pthread_exit(NULL);
  }

  SocketUnixCreateFd(socketFd, &socketP);

  ServerCreateSocket2(&abyssServer, socketP, &error);
  if (error) {
    free((char *)error);
    close_service(socketFd);
    xmlrpcthread=0;
    xmlrpcport=0;
    pthread_exit(NULL);
  }
  
  xmlrpc_server_abyss_set_handlers(&abyssServer,registryP);
  ServerInit(&abyssServer);
#ifdef HAVE_SERVER_SET_MAX_CONN
  ServerSetMaxConn(&abyssServer,MAX_CONN);
#endif
  xmlrpcport=port;

  pthread_cleanup_push(wnpRPCcleanup,(void *)NULL);

  ServerRun(&abyssServer);

  pthread_cleanup_pop(TRUE);
  pthread_exit(NULL);

}

static void wnpRPCcleanup(void *arg)
{
  /* executed by XML-RPC server thread */

  if (xmlrpcthread) {
    ServerFree(&abyssServer);
    SocketDestroy(socketP);
    close_service(socketFd);
    xmlrpcthread=0;
    xmlrpcport=0;
  }
}

#endif

/*==================== may be executed by any thread ====================*/

#ifdef withRPC

int wnpRPCmutexLock(void)
{
  return(pthread_mutex_lock(&rpcMutex));
}

int wnpRPCmutexUnlock(void)
{
  return(pthread_mutex_unlock(&rpcMutex));
}

#endif

/*==================== main thread code ====================*/

void cmd_xmlrpc(void)
{
  /* executed by main thread */

  Token *key=NULL;

  key=Get_Kwd("xmlrpc: ",";","close open");
  if (key->status==TOK_OK) {

    if (key_is("open")) {
      key=free_Token(key);
      key=Get_Int_M("open service on port number: ",";","19937 19938 19939 19940");
      if (key->status==TOK_OK) {
	if (wnpHaveRPC()) {
	  if (wnpRPCstart(key->n)) prtGraphics("XML/RPC server started on port %d\n",wnpRPCport());
	}else err_set(0,"XML/RPC server is not available in this version of WITNOTP");
      }else if (key->status==TOK_NOK) err_set(1,"bad port number: %s",key->text);
    }else if (key_is("close")) {
      wnpRPCshutdown();
    }else err_set(1,"bad xmlrpc command: %s",key->text);
  }else if (key->status==TOK_NOK) err_set(1,"bad xmlrpc command: %s",key->text);
  
  free_Token(key);
}

int wnpHaveRPC(void)
{
#ifdef withRPC
  return(1);
#else
  return(0);
#endif
}

int wnpRPCstart(int port)
{
  int i;
  char buff[256];
#ifdef withRPC
  int status=0;
#endif

#ifdef withRPC
  if (!xmlrpcthread) {
    xmlrpcport=0;
    pthread_mutex_init(&rpcMutex,NULL);
    status=pthread_create(&xmlrpcthread,NULL,wnpRPCserver,&port);
    if (status) err_set(1,"failed to start xmlrpc server thread.");
    if (xmlrpcthread) {
      for (i=0;xmlrpcthread&&!xmlrpcport&&i<250;i++) uSleep(NAPTIME);
      if (xmlrpcthread) {
	if (!xmlrpcport) err_set(1,"giving up waiting for XML/RPC server.");
      }else err_set(1,"failed to start RPC server on port %d",port);
    }else{
      err_set(1,"failed to start RPC server on port %d",port);
      pthread_mutex_destroy(&rpcMutex);
    }
  }else err_set(1,"already connected to xmlrpc server on port %d.",xmlrpcport);
#endif

  /* put current port number into environment */
  i=wnpRPCport();
  buff[0]='\0';
  if (i) sprintf(buff,"%d",i);
  Setenv("WNP_RPC_PORT",buff);
  return(wnpRPCport());
}

void wnpRPCshutdown()
{
#ifdef withRPC
  int i;

  if (xmlrpcthread) {
    ServerTerminate(&abyssServer);
    wnpRPCclose(xmlrpcthread);
    xmlrpcport=0;
    for (i=0;xmlrpcthread&&i<100;i++) uSleep(NAPTIME);
    if (xmlrpcthread) {
      prtGraphics("killing RPC threads...\n");
      pthread_cancel(xmlrpcthread);
      xmlrpcthread=0;
    }
    pthread_mutex_destroy(&rpcMutex);
  }
#endif
}

int wnpRPCport(void)
{
  return(xmlrpcport);
}

int wnpRPCisOn(void)
{
#ifdef withRPC
  return(xmlrpcthread?1:0);
#else
  return(0);
#endif
}

int wnpRPCisOff(void)
{
#ifdef withRPC
  return(xmlrpcthread?0:1);
#else
  return(0);
#endif
}

int wnpRPCmaxCalls(void)
{
  /* maximum number of open RPC calls that we allow */
  /* should be less than the number of open calls that Abyss can handle */

#ifdef withRPC
  return(MAX_CONN-1);   /* MAX_CONN is the max. number of open connections allowed by Abyss */
#else
  return(0);
#endif
}
