#include <stdio.h>
#include "xtal.h"
#include "gcalc.h"
#include "strutil.h"
#include "errhand.h"
#include "mystdlib.h"
#include "swdef.h"
#include <math.h>
#include <ctype.h>

#ifndef M_PI
#define M_PI  3.14159265358979323846
#endif

#define MxM m4xm4
#define MxV m4xv3
#define SYM_TOL (0.001)

/*-------------------- internals --------------------*/

static vec_3 off[8] = {
   {{0.0,0.0,0.0}}, {{1.0,0.0,0.0}}, {{0.0,1.0,0.0}}, {{1.0,1.0,0.0}},
   {{0.0,0.0,1.0}}, {{1.0,0.0,1.0}}, {{0.0,1.0,1.0}}, {{1.0,1.0,1.0}}
};

static double f2int(const double x)
{
/* round x to nearest integer, return as double */

  long r;
  double d;
 
  r=(long)x; d=x-r;
  r=(d>=0.5 ? r+1 : (d<= -0.5 ? r-1 : r));
  return((double)r);
}

static int gcd(int a,int b)
{
/* return gcd(a,b), 0 if min(abs(a),abs(b)) == 0 */

  int r;

  a=(a<0 ? -a : a); b=(b<0 ? -b : b);
  r=(a>b ? b : a);
  a=(a>b ? a : b);
  b=r;

  while (r!=0) {
    b=r;
    r=a%b;
    a=b;
  }
  b=(b>0 ? b : 1);
  return(b);
}
/*-------------------- sym2str --------------------*/

char *sym2str(char *str,mat_4 m)
{
/* convert 4x4 symmetry matrix to character string (SHELX format) 
   Assumes that the rotation matrix coefficients ar -1,0,+1, and
   that the translation components are "simple" fractions (/2, /3, /4, /6, /8)
   Return pointer to resulting string. m should be in fract coord.
*/

  int i,j,t,n,d,c[3],o;
  float f;
  char *s;

  s=str; *s='\0';
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      f=m.r[i].c[j];
      c[j]=(f > 0.5 ? 1 : (f > -0.5 ? 0 : -1));   /* sign of coefficients */
    }

    t=f2int(12*m.r[i].c[3]);
    n=t/gcd(12,t); d=12/gcd(12,t);   /* t normalized to simple fraction */ 
    o=0;
    for (j=0;j<3;j++) {
      if (c[j]>0) {                  /* positive coefficients first */
	if (o) *s++ = '+';
	*s++ = 'X'+j;
	o++; *s='\0';
      }
    }
    for (j=0;j<3;j++) {              /* negative coefficients next */
      if (c[j]<0) {
	*s++ = '-';
	*s++ = 'X'+j;
	o++; *s='\0';
      }
    }
    if (n!=0) {                      /* translation last */
      if (n>0&&o>0) {
	*s++ = '+';
	*s='\0';
      }
      if (d!=1) sprintf(s,"%d/%d",n,d);
      else sprintf(s,"%d",n);
    }
    if (i<2) strcat(s,",");
    s=s+strlen(s);
  }
  return(str);
}
/*-------------------- str2sym --------------------*/  

Symop str2sym(char *string)
{
/* Convert SHELX formatted symmetry operator (string) to 4x4 matrix
   Input is case-insensitive. symop in fract coords. 
*/

  char tok[256],*sym,*c,t;
  Parse p;
  Symop s;
  int i,d,tx,n,slash;
  float f,sign;

  sym=(string ? Strdup(string) : (char *)NULL);
  if (!sym) {
    err_set(1,"str2sym: memory allocation failure.");
    return(zero_4());
  }

/* convert to upper,
   if there are no commas transform to standard form (comma sep) */

  s=zero_4();s.r[3].c[3]=1.0;

  for (n=0,c=sym;*c;c++) {
    *c=(*c >= 'a' && *c <= 'z' ? *c+'A'-'a' : *c);
    if (isspace(*c)) *c=' ';
    if (*c==',') n++;
  }
  if (n==0) { /* insert two commas */
    p.buf=sym; p.sep=","; p.tok=(char *)NULL; p.pos=0;
    for (n=0;n<2;n++) {
      if (!get_token(&p)||p.ltok==0) {
	err_set(2,"str2sym: bad argument(%s).",string);
	free(sym);return(zero_4());
      }
      p.tks[p.ltok]=',';
    }
  }
  if (n!=2) {
    free(sym);
    err_set(2,"str2sym: bad argument(%s).",string);
    return(zero_4());
  }

/* parse modified symm-string */

  p.buf=sym; p.sep="+-/,;"; p.tok=tok; p.mts=255; p.pos=0;

  sign=1.0;
  slash=FALSE;
  n=0;
  t=',';
  do {
    if (!get_token(&p)) {
      n=10;
      break;
    }
    switch (*tok) {
    case 'X':
    case 'Y':
    case 'Z':
      if (strchr("+-,",t)) s.r[n].c[*tok-'X']+=sign;
      else n=10;
      t='X';
      break;
    default:
      if (p.ltok>0) {
	if (strchr("+-/,",t)) {
	  f=str2f(tok);
	  if (f>0) s.r[n].c[3]=(slash ? s.r[n].c[3]/f : sign*f);
	  else n=10;
	}else{
	  n=10;
	}
	t='f';
      }else{
	if (strchr("+-/",t)) n=10;;
      }
    }
    slash=FALSE;
    sign=1.0;
    switch(p.term) {
    case '+':
      sign=1.0;
      break;
    case '-':
      sign= -1.0;
      break;
    case '/':
      if (t!='f') n=10;
      slash=TRUE;
      break;
    default:
      n++;
      if (t==',') n=10;
      break;
    }
    t=p.term;
  }while(n<3);

  if (n==3&&!p.buf[p.pos]) {
    for (i=0;i<3;i++) {                     /* clean up translation */
      tx=f2int(12*s.r[i].c[3]);
      n=tx/gcd(12,tx);d=12/gcd(12,tx);
      s.r[i].c[3]=((double)n)/((double)d);
    }
  }else {
    err_set(3,"str2sym: bad argument(%s).",string);
    s=zero_4();
  }
  if (sym) free(sym);
  return(s);
}
/*-------------------- csd2sym --------------------*/  

Symop csd2sym(char *string)
{
/* Convert CSD formatted symmetry operator (string, 15 chars) to 4x4 matrix
   Result will be in fract coords. 
*/

  int i,j;
  Symop s;
  char *p;

  for (p=string,i=0;i<3;i++,p+=2) {
    s.r[3].c[i]=0.0;
    for (j=0;j<3;j++) s.r[i].c[j]= *p++-'1';
    s.r[i].c[3]=(10*(*p!=' '?*p-'0':0)+*(p+1)-'0')/12.0;
  }
  s.r[3].c[3]=1.0;
  return(s);
}
/*-------------------- cell_volume --------------------*/

float cell_volume(float *l,float *a)
{
  float C[3],f,V;
  int i;

  f=M_PI/180.0;  

  for (i=0;i<3;i++) C[i]=cos(a[i]*f);
  V=(1.0-C[0]*C[0]-C[1]*C[1]-C[2]*C[2]+2.0*C[0]*C[1]*C[2]);
  V=(V>0 ? sqrt(V) : 0.0)*l[0]*l[1]*l[2];
  return(V);
}

/*-------------------- PDB_fta_matrix --------------------*/

mat_4 PDB_fta_matrix(float *l,float *a)
{
/*
   compute transformation matrix m for transformation from
   cell coordinates (cell lengths l, cell angles a, in deg)
   to cartesian angstrom coordinates.  PDB convention:
   x||a , z || aXb.
*/

  mat_4 r;
  float C[3],S[3],f,V;
  int i;

  f=M_PI/180.0;  

  r=ident_4();

  for (i=0;i<3;i++) {
    C[i]=cos(a[i]*f);
    S[i]=sin(a[i]*f);
  }
  V=(1.0-C[0]*C[0]-C[1]*C[1]-C[2]*C[2]+2.0*C[0]*C[1]*C[2]);
  V=(V<0?0.0:sqrt(V));
  V=l[0]*l[1]*l[2]*V;
  if (V<0.0001) {
    err_set(1,"PDB_fta_matrix, singular matrix");
    r.r[0].c[0]=0.0;
  }else{
    r.r[0].c[0]=l[0];
    r.r[0].c[1]=l[1]*C[2]; r.r[1].c[1]=l[1]*S[2];
    r.r[0].c[2]=l[2]*C[1]; r.r[1].c[2]=l[2]*(C[0]-C[1]*C[2])/S[2];
    r.r[2].c[2]=V/(l[0]*l[1]*S[2]);
  }
  return(r);
}

/*-------------------- fract_to_angstrom --------------------*/

mat_4 fract_to_angstrom(float l[3],float a[3])
{
/*
   compute transformation matrix m for transformation from
   cell coordinates (cell lengths l, cell angles a, in deg)
   to cartesian angstrom coordinates.
   c --> z, b --> (z,y)
*/
  float f;
  mat_4 r;

  f=M_PI/180.0;

  r=ident_4();

  r.r[0].c[2]=0.0; r.r[1].c[2]=0.0; r.r[2].c[2]=l[2];

  r.r[0].c[1]=0.0; r.r[1].c[1]=l[1]*sin(f*a[0]); r.r[2].c[1]=l[1]*cos(f*a[0]);

  r.r[2].c[0]=l[0]*cos(f*a[1]);
  r.r[1].c[0]=l[0]*(cos(f*a[2])-cos(f*a[0])*cos(f*a[1]))/sin(f*a[0]);
  r.r[0].c[0]=l[0]*l[0]-r.r[1].c[0]*r.r[1].c[0]-r.r[2].c[0]*r.r[2].c[0];
  if (r.r[0].c[0]>0) r.r[0].c[0]=sqrt(r.r[0].c[0]);
  else {
    r.r[0].c[0]=0.0;
    err_set(1,"fract_to_angstrom: negative argument in sqrt");
  }
    
  return(r);
}
/*-------------------- symA --------------------*/

Symop symA(Xcn *x,Symop s)
{
/* convert symmetry operator from fractional to angstrom ortho,
   assuming that x points to a proper (i.e. correct fta,atf) Xcn */

  return(MxM(x->fta,MxM(s,x->atf)));
}
/*-------------------- symF --------------------*/

Symop symF(Xcn *x,Symop s)
{
/* convert symmetry operator from angstrom ortho to fractional,
   given proper Xcn */

  return(MxM(x->atf,MxM(s,x->fta)));
}
/*-------------------- symopexp --------------------*/

int symopexp(Xcn *x)
{
/*  expand symmetry operators according to x->lattice, x->centric 
    symmetry operators are assumed to be in fractional coordinates */

  int nsym,nsymx,i,k;

  nsymx=x->nsym;
  x->nsymx=nsymx;
  if (x->centric) nsymx*=2;
  switch (x->lattice) {
  case A_latt:
  case B_latt:
  case C_latt:
  case I_latt:
    nsymx*=2;
    break;
  case F_latt:
    nsymx*=4;
    break;
  case H_latt:
    nsymx*=3;
    break;
  default:
    break;
  }

  if (nsymx>MX_XCN_SYMOPS) return(nsymx);
 
  nsym=x->nsym;
  switch (x->lattice) {
  case A_latt:
    for (i=nsym;i<2*nsym;i++) {
      x->sym[i]=x->sym[i-nsym];
      x->sym[i].r[1].c[3]+=0.5; 
      x->sym[i].r[2].c[3]+=0.5;
    }
    break;
  case B_latt:
    for (i=nsym;i<2*nsym;i++) {
      x->sym[i]=x->sym[i-nsym];
      x->sym[i].r[0].c[3]+=0.5; 
      x->sym[i].r[2].c[3]+=0.5;
    }
    break;
  case C_latt:
    for (i=nsym;i<2*nsym;i++) {
      x->sym[i]=x->sym[i-nsym];
      x->sym[i].r[0].c[3]+=0.5; 
      x->sym[i].r[1].c[3]+=0.5;
    }
    break;
  case I_latt:
    for (i=nsym;i<2*nsym;i++) {
      x->sym[i]=x->sym[i-nsym];
      x->sym[i].r[0].c[3]+=0.5; 
      x->sym[i].r[1].c[3]+=0.5; 
      x->sym[i].r[2].c[3]+=0.5;
    }
    break;
  case F_latt:
    for (i=nsym;i<2*nsym;i++) {
      x->sym[i]=x->sym[i-nsym];
      x->sym[i].r[1].c[3]+=0.5; 
      x->sym[i].r[2].c[3]+=0.5;
    }
    for (i=2*nsym;i<3*nsym;i++) {
      x->sym[i]=x->sym[i-2*nsym];
      x->sym[i].r[0].c[3]+=0.5; 
      x->sym[i].r[2].c[3]+=0.5;
    }
    for (i=3*nsym;i<4*nsym;i++) {
      x->sym[i]=x->sym[i-3*nsym];
      x->sym[i].r[0].c[3]+=0.5; 
      x->sym[i].r[1].c[3]+=0.5;
    }
    break;
  case H_latt:
    for (i=nsym;i<2*nsym;i++) {
      x->sym[i]=x->sym[i-nsym];
      x->sym[i].r[0].c[3]+=1.0/3.0;
      x->sym[i].r[1].c[3]+=2.0/3.0;
      x->sym[i].r[2].c[3]+=2.0/3.0;
    }
    for (i=2*nsym;i<3*nsym;i++) {
      x->sym[i]=x->sym[i-2*nsym];
      x->sym[i].r[0].c[3]+=2.0/3.0;
      x->sym[i].r[1].c[3]+=1.0/3.0;
      x->sym[i].r[2].c[3]+=1.0/3.0;
    }
  }

  if (x->centric) {
    for (i=0;i<nsymx/2;i++) {
      for (k=0;k<3;k++) x->sym[i+nsymx/2].r[k]=sxv4(-1.0,x->sym[i].r[k]);
      x->sym[i+nsymx/2].r[3]=x->sym[i].r[3];
    }
  }

  x->nsymx=nsymx;
  return(nsymx);
}
/*-------------------- symopequiv --------------------*/

int symopequiv(Symop a,Symop b)
{
/* return TRUE iff symmetry operators a and b are equivalent.
   two symops are equivalent iff the corresponding rotations are
   the same modulo (-x,-y,-z) .      */
	
  int i,j,r;
  float d;

  for (r=1,i=0;i<3 && r;i++) {       /* Rot(a) == Rot(b) ? */
    for (j=0;j<3 && r;j++) {
      d=(a.r[i].c[j]-b.r[i].c[j]);
      r=d > -SYM_TOL && d < SYM_TOL;
    }
  }
  if (r) return(r);

  for (r=1,i=0;i<3 && r;i++) {       /* Rot(a) == -Rot(b) ? */
    for (j=0;j<3 && r;j++) {
      d=(a.r[i].c[j]+b.r[i].c[j]);
      r=d > -SYM_TOL && d < SYM_TOL;
    }
  }
  return(r);
}
/*-------------------- Fxyz --------------------*/

vec_3 Fxyz(Xcn *c,vec_3 x)
{
/* convert angstrom to fractional */

  if (is_xcn(c)) return(MxV(c->atf,x));
  else return(x);

}
/*-------------------- Axyz --------------------*/

vec_3 Axyz(Xcn *c,vec_3 x)
{

/* convert fractional to angstrom */

  if (is_xcn(c)) return(MxV(c->fta,x));
  else return(x);
}

/*-------------------- F2cell --------------------*/

vec_3 F2cell(vec_3 x)
{
/*  return x' equivalent to x (fract coord) in cell (0:1,0:1,0:1) */
	
  vec_3 r;
  int i;
  double d;
	
  for (i=0;i<3;i++) {
    d=x.c[i];
    r.c[i]=d-(long)d;
    while(r.c[i]<0.0) r.c[i]+=1.0;
  }
  return(r);
}

/*-------------------- A2cell --------------------*/

vec_3 A2cell(Xcn *c,vec_3 x)
{
/*  same as F2cell, but for angstrom orthogonal x, result in a.o. 
    c points to appropriate Xcn */
	
  return(MxV(c->fta,F2cell(MxV(c->atf,x))));
}
/*-------------------- symdo --------------------*/

vec_3 symdo(Symop s,vec_3 x)
{
/*  return symmetry transformed point x as 3D-vector */

  vec_3 r;
  int i,j;

  for (i=0;i<3;i++) {
    r.c[i]=s.r[i].c[3];
    for (j=0;j<3;j++) r.c[i]+=s.r[i].c[j]*x.c[j];
  }
  return(r);
}
/*-------------------- symtra --------------------*/

Symop symtra(Symop s,vec_3 t)
{
  Symop r;
  
  r=s;
  r.r[0].c[3]+=t.c[0];
  r.r[1].c[3]+=t.c[1];
  r.r[2].c[3]+=t.c[2];
  return(r);
}
/*-------------------- lattice_dist --------------------*/

float lattice_dist(Xcn *x,vec_3 a,vec_3 b)
{
/* distance between points a and b (ortho angstrom) modulo lattice x */

  vec_3 d,O,D;
  float lsq,r;
  int i;

  O=MxV(x->fta,off[0]);
  d=F2cell(MxV(x->atf,v3pv3(O,v3mv3(a,b))));
  r=1.0E+30;
  for (i=0;i<8;i++) {
    D=v3mv3(MxV(x->fta,v3mv3(d,off[i])),O);
    lsq=v3sv3(D,D);
    if (r>lsq) r=lsq;
  }
  r=(r>0 ? sqrt(r) : 0.0);
  return(r);
}
/*-------------------- recell --------------------*/

float recell(float *cl,float *ca,float *rl,float *ra)
{
/* Compute reciprocal cell (rl,ca) from (cl,ca), return volume of
   direct cell 

   NOTE: ca is in degrees, ra in radians 
*/

  int i;
  float s[3],c[3],v;

  for (i=0;i<3;i++) {
    c[i]=cos(M_PI*ca[i]/180.0);
    s[i]=sin(M_PI*ca[i]/180.0);
  }
  
  v=cl[0]*cl[1]*cl[2]*sqrt(1.0-c[0]*c[0]-c[1]*c[1]-c[2]*c[2]+
			   2*c[0]*c[1]*c[2]);

  rl[0]=cl[1]*cl[2]*s[0]/v;
  rl[1]=cl[2]*cl[0]*s[1]/v;
  rl[2]=cl[0]*cl[1]*s[2]/v;

  ra[0]=acos((c[1]*c[2]-c[0]/(s[1]*s[2])));
  ra[1]=acos((c[2]*c[0]-c[1]/(s[2]*s[0])));
  ra[2]=acos((c[0]*c[1]-c[2]/(s[0]*s[1])));
  
  return(v);
}
/*-------------------- RotTfact --------------------*/

void RotTfact(Symop s,float b[6])
{
  /* Transform anisotropic temperature factor b; s is in xtal coord. */

  int i,j;
  mat_3 B,S,ST;

  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      S.r[i].c[j]=ST.r[j].c[i]=s.r[i].c[j];
    }
  }

  B.r[0].c[0]=b[0];   B.r[1].c[1]=b[1];   B.r[2].c[2]=b[2];
  B.r[0].c[1]=B.r[1].c[0]=b[5];
  B.r[0].c[2]=B.r[2].c[0]=b[4];
  B.r[1].c[2]=B.r[2].c[1]=b[3];

  B=m3xm3(S,m3xm3(B,ST));

  b[0]=B.r[0].c[0]; b[1]=B.r[1].c[1]; b[2]=B.r[2].c[2];
  b[3]=B.r[1].c[2]; b[4]=B.r[0].c[2]; b[5]=B.r[0].c[1];
}
/*-------------------- symClosest --------------------*/

Symop symClosest(Xcn *xcn,vec_3 X,vec_3 Z,float *d)
{
  vec_3 Xf,xf,dX,x,y,Ymin,Y,dY;
  int i,k,op,opmin=0;
  float dmin,ds,dx[3][2];

  dmin=1.0e20;

  Xf=F2cell(Fxyz(xcn,X));
  dX=v3mv3(X,Axyz(xcn,Xf));

  for (op=0;op<xcn->nsymx;op++) {
    x=symdo(xcn->sym[op],Z);
    xf=F2cell(Fxyz(xcn,x));
    for (i=0;i<3;i++) {
      dx[i][0]=0.0;
      dx[i][1]=(xf.c[i]<Xf.c[i] ? 1.0 : -1.0);
    }
    for (k=0;k<8;k++) {
      for (i=0;i<3;i++) y.c[i]=dx[i][(k&(1<<i))>>i]+xf.c[i];
      Y=v3pv3(dX,Axyz(xcn,y));
      for (ds=0,i=0;i<3;i++) ds+=(Y.c[i]-X.c[i])*(Y.c[i]-X.c[i]);
      if (ds<dmin) {
	Ymin=Y;
	opmin=op;
	dmin=ds;
      }
    }
  }
  
  dY=v3mv3(Ymin,symdo(xcn->sym[opmin],Z));
  *d=sqrt(dmin);

  return(symtra(xcn->sym[opmin],dY));
}
/*-------------------- xcn_transform --------------------*/

void xcn_transform(mat_4 M,Xcn *xcn)
{
  /* Apply transform M to Xcn node xcn */

  int i;

  if (!is_xcn(xcn)) return;
  for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symF(xcn,xcn->sym[i]);
  xcn->fta=MxM(M,xcn->fta);
  xcn->atf=inv_4(xcn->fta);
  for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);
}
/*-------------------- check_fta_matrix --------------------*/

float check_fta_matrix(mat_4 fta,float *cl,float *ca)
{
  /* checks consistency of fract-to-angstrom transformation with given unit cell,
     returns max. error of (1,0,0),(0,1,0),(0,0,1) in A */

  mat_4 M,C;
  float d,e,c[3],a[3];
  int i,j;
  
  for (i=0;i<3;i++) {
    for (c[i]=0,j=0;j<3;j++) c[i]+=fta.r[j].c[i]*fta.r[j].c[i];
    c[i]=sqrt(c[i]);
  }
  for (i=0;i<3;i++) {
    for (a[i]=0,j=0;j<3;j++) a[i]+=fta.r[j].c[(i+1)%3]*fta.r[j].c[(i+2)%3];
    a[i]=acos(a[i]/(c[(i+1)%3]*c[(i+2)%3]))*180.0/M_PI;
  }
  
  M=PDB_fta_matrix(c,a);
  C=PDB_fta_matrix(cl,ca);
  for (d=0,i=0;i<3;i++) {
    for (e=0,j=0;j<3;j++) e+=(M.r[i].c[j]-C.r[i].c[j])*(M.r[i].c[j]-C.r[i].c[j]);
    e=sqrt(e); d=(d<e ? e : d);
  }
  return(d);
}
/*-------------------- fix_fta_matrix --------------------*/

float fix_fta_matrix(Xcn *xcn)
{
  float r,y[4][3],x[4][3];
  mat_4 rot,std;
  int i,j;

  if (!is_xcn(xcn)) {
    err_set(1,"bad argument passed to fix_fta_matrix.");
    return(1.e4);
  }

  std=fract_to_angstrom(xcn->cl,xcn->ca);
  for (j=0;j<3;j++) {
    x[3][j]=xcn->fta.r[j].c[3];
    y[3][j]=std.r[j].c[3];
  }
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      x[i][j]=x[3][j]+xcn->fta.r[j].c[i];
      y[i][j]=y[3][j]+std.r[j].c[i];
    }
  }
  r=rms_best_mat_yx(4,(float *)y,(float *)x,&rot);
  xcn->fta=m4xm4(rot,std);
  xcn->atf=inv_4(xcn->fta);
  return(r);
}
/*-------------------- symopreduce --------------------*/

void symopreduce(Xcn *xcn)
{
  /* reduce "expanded" set of symops, assumes symops are in fract. system */

  int i,j,A,B,C,I,F,H;
  char buff[1024];

  if (!mvld(xcn)) return;
  if (xcn->nsym<1) return;

  /* move Id to begin of list of symops... */

  if (xcn->nsymx<xcn->nsym) xcn->nsymx=xcn->nsym;
  for (i=0;i<xcn->nsymx;i++) if (streq("X,Y,Z",sym2str(buff,xcn->sym[i]))) break;
  if (i<xcn->nsymx) {
    if (i>0) {
      xcn->sym[i]=xcn->sym[0];
      xcn->sym[0]=str2sym("X,Y,Z");
    }
  }else{
    if (xcn->nsymx<MX_XCN_SYMOPS) {
      for (i=xcn->nsymx;i>0;i--) xcn->sym[i]=xcn->sym[i-1];
      xcn->sym[0]=str2sym("X,Y,Z");
      xcn->nsymx++;
      xcn->nsym++;
    }else{
      err_set(1,"symopreduce failed: too many operators.");
      return;
    }
  }
    
  /* check for center at origin */

  xcn->centric=FALSE;
  for (i=1;i<xcn->nsymx;i++) {
    if (streq("-X,-Y,-Z",sym2str(buff,xcn->sym[i]))) {
      xcn->centric=TRUE;
      for (j=i+1;j<xcn->nsymx;j++) xcn->sym[j-1]=xcn->sym[j];
      xcn->nsymx--; i--;
    }
  }

  /* remove symops obtained by face-centering */

  A=FALSE; B=FALSE; C=FALSE; I=FALSE; H=FALSE; F=FALSE;
  xcn->nsym=1;
  for (i=1;i<xcn->nsymx;i++) {
    for (j=0;j<i;j++) if (symopequiv(xcn->sym[i],xcn->sym[j])) break;
    if (j<i) {
      sym2str(buff,m4mm4(xcn->sym[i],xcn->sym[j]));
      if (streq(buff,"1/3,2/3,2/3"))     H=TRUE;
      else if (streq(buff,"2/3,1/3,1/3")) H=TRUE; 
      else if (streq(buff,",1/2,1/2"))    A=TRUE;
      else if (streq(buff,"1/2,,1/2"))    B=TRUE;
      else if (streq(buff,"1/2,1/2,"))    C=TRUE;
      else if (streq(buff,"1/2,1/2,1/2")) I=TRUE;
    }else{
      xcn->sym[xcn->nsym++]=xcn->sym[i];
    }
  }

  if ((A&&B) || (A&&C) || (C&&B)) F=TRUE;
  if (F) xcn->lattice=F_latt;
  else if (H) xcn->lattice=H_latt;
  else if (I) xcn->lattice=I_latt;
  else if (A) xcn->lattice=A_latt;
  else if (B) xcn->lattice=B_latt;
  else if (C) xcn->lattice=C_latt;

  symopexp(xcn);   /* expand again */
}
