#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* Tool to convert X-Plor DCD file to Wit!P .cnf format */
/* DCD: binary trajectory file (positions or velocities)*/
/* These files are FORTRAN unformatted files with four  */
/* control bytes (length of record) at head and tail of */
/* each record .                                        */

#define TRUE  1
#define FALSE 0

#define BYTESEX_UNKNOWN   0
#define BYTESEX_HOMO   1
#define BYTESEX_HETERO 2
#define TIMFAC 0.04888821 /* conversion factor from AKMA time to picoseconds.*/

static int read_header(FILE *F,int *sex,char *type,int *version,int *qcrys,
		       int *nat,int *nfix,int **mob,int *nfrm,
		       int *first,int *nsteps,float *dt,char ***cmnt);
static int read_dvec(FILE *F,int sex,int n,double *v);
static int read_fvec(FILE *F,int sex,int n,float *v);

static char *swap(int n,char *byte);

void ProvideHelp(void)
{
   fprintf(stderr,"\n"
    "\n"
    "Usage: dcd2cnf [-v] [-f from] [-t to] [-s step] [filename]\n"
    "-----\n"
    "\n"
    "Description:\n"
    "-----------\n"
    "Converts X-Plor (or CHARMM, FMD,...) binary trajectory file to\n"
    "Wit!P .cnf format on stdout. If no input filename is given, the\n"
    "binary file will be read from stdin. If the -v flag is set,\n"
    "information from the file header will be written to stderr. The\n"
    "-f, -t and -s flags allow the selection of a subset of the \n"
    "frames (conformations) in the file. \n"
   "\n");
}

/*------------------------------------------------------------------------*/

int main (int argc,char **argv)
{
  FILE *in,*out;
  int verbose,status;
  int i,k,n,nat,nfix,*mob=(int *)NULL,nfrm,first,nsteps,nfree=0,sex;
  int version,qcrys;
  int from=1,to=-1,step=1;
  float dt;
  char type[5],**cmnt,**c,*hdr,*fname=(char *)NULL;
  float *x=(float *)NULL,*y=(float *)NULL,*z=(float *)NULL;
  double crys[6];

  out=stdout;
  in=stdin;
  verbose=FALSE;

  for (i=1;i<argc;i++) {
    if (!strcmp(argv[i],"-v")) {
      verbose=TRUE;
    }else if (!strcmp(argv[i],"-f")) {
      from=atoi(argv[++i]);
    }else if (!strcmp(argv[i],"-t")) {
      to=atoi(argv[++i]);
    }else if (!strcmp(argv[i],"-s")) {
      step=atoi(argv[++i]);
    }else if (!strcmp(argv[i],"-h")||
	      !strcmp(argv[i],"--help")) {
      ProvideHelp();
      return(0);
    }else if (argv[i][0]=='-') {
      ProvideHelp();
      fprintf(stderr,"dcd2cnf, unknown flag: %s\n",argv[i]);
      return(1);
    }else{
      if (fname) {
	ProvideHelp();
	fprintf(stderr,"dcd2cnf, multiple input files not supported\n");
	return(1);
      }else{
	fname=argv[i];
	in=fopen(fname,"r");
	if (!in) {
	  fprintf(stderr,"dcd2cnf, failed to open file %s for read\n",fname);
	  return(1);
	}
      }
    }
  }

  status=read_header(in,&sex,type,&version,&qcrys,
		     &nat,&nfix,&mob,&nfrm,&first,&nsteps,&dt,&cmnt);
  if (status) {
    fprintf(stderr,"dcd2cnf, failed to read trajectory header record (status=%d).\n",status);
    return(status);
  }
  nfree=nat-nfix;
  hdr=(!strcmp(type,"VELD")?"Velocities":
       !strcmp(type,"CORD")?"Conformation":
       "unk");
  if (!strcmp(hdr,"unk")) {
    fprintf(stderr,"dcd2cnf, unknown trajectory type: %s\n",type);
    return(1);
  }

  if (verbose) {
    fprintf(stderr,"Filetype: %s, version %d, %d atoms (%d fixed), timestep %0.4f ps\n",
	    type,version,nat,nfix,dt);
    fprintf(stderr,"Frames: first = %d, %0.4f ps per frame, number of frames",
	    first,nsteps*dt);
    if (nfrm==0) fprintf(stderr," unknown.\n");
    else fprintf(stderr," = %d\n",nfrm);
    if (cmnt) {
      for (c=cmnt;*c;c++) {
	fprintf(stderr,"%s\n",(*c[0]==' '?*c+1:*c));
      }
    }
  }

  x=(float *)malloc(nat*sizeof(float));
  y=(float *)malloc(nat*sizeof(float));
  z=(float *)malloc(nat*sizeof(float));
  if (!x||!y||!z) {
    fprintf(stderr,"dcd2cnf: memory allocation failure\n");
    if (x) free(x); if (y) free(y); if (z) free(z);
    status=1;
    return(status);
  }

  if (mob&&nfree>0) {
    fprintf(out,"Mobile %d\n",nfree);
    for (i=0;i<nfree;) {
      for (k=0;k<10&&i<nfree;k++,i++) fprintf(out," %d",mob[i]+1);
      if (k>0) fprintf(out,"\n");
    }
  }

  for(n=0;to<0||n<to;n++) {
    if (qcrys&&(status=read_dvec(in,sex,6,crys))!=6) break; /* skip crysinf */
    if ((status=read_fvec(in,sex,nat,x))<=0) break;
    if (status!=read_fvec(in,sex,nat,y)) break;
    if (status!=read_fvec(in,sex,nat,z)) break;
    if (n+1-from>=0&&((n+1-from)%step==0)) {
      fprintf(out,"%s %d Energy = 0.0\n",hdr,first+n*nsteps);
      for (i=0;i<status;i++) fprintf(stdout,"%0.5f %0.5f %0.5f\n",x[i],y[i],z[i]);
    }
    status=0;
  }
  if (status) {
    fprintf(stderr,"dcd2cnf: error reading file after %d frames\n",n);
    return(status);
  }
  if (n>0) fprintf(stderr,"processed %d frames\n",n);

  if (cmnt) {
    for (c=cmnt;*c;c++) free(*c);
    free (cmnt);
  }
  if (x) free(x); if (y) free(y); if (z) free(z);
  if (mob) free(mob);
  status=0;
  return(status);
}

/*-------------------- read_header --------------------*/

static int read_header(FILE *F,int *sex,char *type,int *version,int *qcrys,
		       int *nat,int *nfix,int **mob,int *nfrm,
		       int *first,int *nsteps,float *dt,char ***cmnt)
{
  int r=0,n,i,nfree;
  char buff[256],*c,**p;

  *cmnt=(char **)NULL;
  *sex=BYTESEX_UNKNOWN;
  *qcrys=0;

  if (fread(buff,1,92,F)==92) { /* (rl)(84 bytes)(rl) ... */
    *sex=BYTESEX_HOMO;
    n=*(int *)buff;
    if (n!=84) {
      swap(4,buff);
      *sex=BYTESEX_HETERO;
      n=*(int *)buff;
    }
    if (n!=84) r=1;
  }else r=2;
  if (r==0) {
    strncpy(type,buff+4,4);type[4]='\0';
    if (*sex==BYTESEX_HETERO) swap(4,buff+84);
    *version=*(int *)(buff+84);
    if (*sex==BYTESEX_HETERO) {
      swap(4,buff+8);
      swap(4,buff+40);
      swap(4,buff+12);
      swap(4,buff+16);
      if (version==0) swap(8,buff+44);  /* X-Plor and CNS, CNX, maye NAMD */
      else{
	swap(4,buff+44);                /* CHARMM */
	swap(4,buff+48);
      }
    }
    *nfrm=*(int *)(buff+8);
    *nfix=*(int *)(buff+40);
    *first=*(int *)(buff+12);
    *nsteps=*(int *)(buff+16);
    if (*version==0) {
      *dt=*(double *)(buff+44)*TIMFAC;
      *qcrys=0;
    }else{
      *dt=*(float *)(buff+44)*TIMFAC;
      *qcrys=*(int *)(buff+48);
    }
  }

  if (r==0) {                     /* REMARKS block */
    if (fread(buff,1,8,F)==8) {
      if (*sex==BYTESEX_HETERO) {
	swap(4,buff);
	swap(4,buff+4);
      }
      n=*(int *)(buff+4);
      if (n*80+4==*(int *)(buff)) {
	*cmnt=(char **)malloc((n+1)*sizeof(char *));
	if (*cmnt) for (i=0;i<=n;i++) (*cmnt)[i]=(char *)NULL;
	for (i=0;i<n;i++) {
	  if (fread(buff,1,80,F)==80) {
	    if (*cmnt) {
	      buff[81]='\0';
	      for (c=buff+strlen(buff)-1;c>=buff;c--) if (*c!=' ') break;
	      c[1]='\0';
	      (*cmnt)[i]=strdup(buff);
	    }
	  }else r=3;
	}
      }else r=4;
      if (r==0&&fread(buff,1,4,F)!=4) r=5; 
    }else r=6;
  }

  if (r==0) {                       /* Number of Atoms */
    if (fread(buff,1,12,F)==12) { 
      if (*sex==BYTESEX_HETERO) {
	swap(4,buff);
	swap(4,buff+4);
      }
      *nat=*(int *)(buff+4);
      if (*(int *)(buff)!=4) r=7;
    }else r=8;
  }

  if (r==0&&*nfix>0) {              /* free atom list */
    if (fread(buff,1,4,F)==4) {
      if (*sex==BYTESEX_HETERO) swap(4,buff);
      nfree=*nat-*nfix;
      if ((int)*(int *)buff==nfree*4) {
	*mob=(int *)malloc(nfree*sizeof(int));
	if (*mob) {
	  if (fread(*mob,4,nfree,F)==nfree) {
	    if (*sex==BYTESEX_HETERO) for (i=0;i<nfree;i++) swap(4,(char *)((*mob)+i));
	    for (i=0;i<nfree;i++) (*mob)[i]-=1;
	  }else r=9;
	}else r=10;
      }else r=11;
      if (r==0&&fread(buff,1,4,F)!=4) r=12;
    }else r=0;
  }

  if (r) {                          /* clean up */
    if (*cmnt) {
      for (p=*cmnt;*p;p++) free(*p);
      free(*cmnt);
      *cmnt=(char **)NULL;
    }
    if (*mob) free(*mob); *mob=(int *)NULL;
  }
  return(r);
}
/*---------------------read_dvec --------------------*/

static int read_dvec(FILE *F,int sex,int n,double *v)
{
  /* read double-vector, return n-read (success), 0 (eof), -1 (error) */

  char buff[4];
  int i,m;
  
  if (n<1) return(-1);
  if (fread(buff,1,4,F)==4) {
    if (sex==BYTESEX_HETERO) swap(4,buff);
    m=*(int *)buff;
    if (m%8==0&&m/8<=n) {
      m/=8;
      if (fread(v,8,m,F)==m) {
	if (sex==BYTESEX_HETERO) {
	  for (i=0;i<m;i++) swap(8,(char *)(v+i));
	}
	if (fread(buff,1,4,F)!=4) return(-1);
      }else return(-1);
    }else return(-1);
  }else if (feof(F)) return(0);
  else return(-1);
  return (m);
}
/*---------------------read_fvec --------------------*/

static int read_fvec(FILE *F,int sex,int n,float *v)
{
  /* read float-vector, return n-read (success), 0 (eof), -1 (error) */

  char buff[4];
  int i,m;
  
  if (n<1) return(-1);
  if (fread(buff,1,4,F)==4) {
    if (sex==BYTESEX_HETERO) swap(4,buff);
    m=*(int *)buff;
    if (m%4==0&&m/4<=n) {
      m/=4;
      if (fread(v,4,m,F)==m) {
	if (sex==BYTESEX_HETERO) {
	  for (i=0;i<m;i++) swap(4,(char *)(v+i));
	}
	if (fread(buff,1,4,F)!=4) return(-1);
      }else return(-1);
    }else return(-1);
  }else if (feof(F)) return(0);
  else return(-1);
  return (m);
}
/*-------------------- swap --------------------*/

static char *swap(int n,char *byte)
{
  int i,m;
  char c;

  switch (n) {
  case 2:
    c=byte[0]; 
    byte[0]=byte[1]; 
    byte[1]=c;
    break;
  case 4:
    c=byte[0]; 
    byte[0]=byte[3]; 
    byte[3]=c;
    c=byte[1]; 
    byte[1]=byte[2]; 
    byte[2]=c;
    break;
  case 8:
    c=byte[0]; 
    byte[0]=byte[7]; 
    byte[7]=c;
    c=byte[1]; 
    byte[1]=byte[6]; 
    byte[6]=c;
    c=byte[2]; 
    byte[2]=byte[5]; 
    byte[5]=c;
    c=byte[3]; 
    byte[3]=byte[4]; 
    byte[4]=c;
    break;
  default:
    m=n/2;
    for (i=0;i<m;i++) {
      c=byte[i];
      byte[i]=byte[n-i-1];
      byte[n-i-1]=c;
    }
    break;
  }
  return(byte);
}
