#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#if defined(darwin)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include "wnp.h"
#include "wnpwin.h"

static float cursorColor[WNP_MAXCCOL][3] = {
  {0.0,0.0,0.0},                  /* black  */
  {1.0,0.0,0.0},                  /* red    */
  {0.0,1.0,0.0},                  /* green  */
  {0.0,0.0,1.0},                  /* blue   */ 
  {1.0,1.0,0.0},                  /* yellow */
  {1.0,1.0,1.0}                   /* white  */
};

typedef struct dspPatch {              /* structure to hold display patch     */
  char *values;
  float x,y;
  int   width,height;
  int   size;
} dspPatch;

typedef struct WnpCursor {
  int mono,cryst;                      /* cursor in mono/cryst mode           */
  dspPatch (*draw)(int x,int y);       /* pointer draw-cursor function        */
  int color;                           /* cursor, foreground=background       */   
} WnpCursor;

static dspPatch WnpCursor_none(int x,int y);
static dspPatch WnpCursor_DOTBOX(int x,int y);
static dspPatch WnpCursor_dotbox(int x,int y);
static dspPatch WnpCursor_ARROW(int x,int y);
static dspPatch WnpCursor_arrow(int x,int y);

static WnpCursor curtab[WNP_MAX_CURSOR] = {
  {WNP_CURSOR_NONE,   WNP_CURSOR_NONE,   WnpCursor_none,   WNP_CURSOR_RED},
  {WNP_CURSOR_DOTBOX, WNP_CURSOR_dotbox, WnpCursor_DOTBOX, WNP_CURSOR_RED},
  {WNP_CURSOR_DOTBOX, WNP_CURSOR_dotbox, WnpCursor_dotbox, WNP_CURSOR_RED},
  {WNP_CURSOR_ARROW,  WNP_CURSOR_arrow,  WnpCursor_ARROW,  WNP_CURSOR_GREEN},
  {WNP_CURSOR_ARROW,  WNP_CURSOR_arrow,  WnpCursor_arrow,  WNP_CURSOR_GREEN}
};

static int redraw=FALSE;
static int cursx=0,cursy=0;            /* cursor position in GLUT coordinates */
static int hide_cursor = FALSE;        /* should we hide cursor in glbox ?    */
static int current_cursor =  1;        /* current index into cursor tab       */
static int win_geom[4];

static dspPatch cursorPatch;
static dspPatch emptyPatch={NULL,0,0,0,0,0};

static dspPatch fetchDspPatch(dspPatch patch);
static dspPatch restoreDspPatch(dspPatch patch);
static dspPatch clearDspPatch(dspPatch patch);
static void setCursorProjection(void);

/* ======================================================================== */

void WnpInitCursor(void)
{
  WnpTogglePhotoMode(300);            /* hide cursor after 300 ticks (3 sec) */
  glutSetCursor(GLUT_CURSOR_NONE);    /* we will do our own cursor */
  cursorPatch=emptyPatch;
  cursx=-100; cursy=-100;
  hide_cursor=FALSE;
  current_cursor=1;
  redraw=FALSE;
}

int WnpGlBoxShowCursor(void)
{
  int r,o;

  r=hide_cursor;
  hide_cursor=FALSE;
  if (r) {
    o=current_cursor;
    current_cursor=o+1;    /* to force OP in WnpSetGlBoxCursor */
    WnpSetGlBoxCursor(o);
    redraw=TRUE;
  }
  return(r);
}

int WnpGlBoxHideCursor(void)
{
  int r;
  
  r=hide_cursor;
  hide_cursor=TRUE;
  WnpGlRestoreCursorPatch();
  redraw=FALSE;
  return(r);
}

void WnpSetGlBoxCursorColor(int cursor,int color)
{
  int c;

  if (cursor>0 && cursor<WNP_MAX_CURSOR) {
    if (color >=0 && color <WNP_MAXCCOL) {
      c=curtab[cursor].mono;
      if (curtab[c].color!=color) {
	curtab[c].color=color;
      }
      c=curtab[cursor].cryst;
      if (curtab[c].color!=color) {
	curtab[c].color=color;
      }
    }
  }
}

int WnpSetGlBoxCursor(int new)
{
  int old_cursor,crystaleyes;

  crystaleyes=WnpCrystalEyesMode();
  old_cursor=current_cursor;
  if (new>=0 && new < WNP_MAX_CURSOR) {
    new=(crystaleyes ? curtab[new].cryst : curtab[new].mono);
    if (current_cursor!=new) {
      current_cursor=new;
      redraw=TRUE;
    }
  }
  return(old_cursor);
}

void WnpGlSetCursorRedrawFlag(int state)
{
  redraw=state;
}

void WnpGlMoveCursor(int x,int y)
{
  WnpGlRestoreCursorPatch();
  redraw=TRUE;
  cursx=x;
  cursy=y;
}

void WnpGlDrawCursor(int swapped)
{
  if (redraw||swapped) {
    if (!hide_cursor) {
      WnpDrawMenuBoxBB();
      WnpDrawMessageBoxBB();
      setCursorProjection();
      if (!swapped) cursorPatch=restoreDspPatch(cursorPatch);
      else cursorPatch=clearDspPatch(cursorPatch);
      cursorPatch=curtab[current_cursor].draw(cursx,cursy);
    }
    redraw=FALSE;
  }
}

void WnpGlRestoreCursorPatch()
{
  if (cursorPatch.values) {
    setCursorProjection();
    cursorPatch=restoreDspPatch(cursorPatch);
  }
}

static void setCursorProjection(void)
{
  ViewportSet("Cursor",WNP_VWP_FULL); /* full, even in above/below stereo mode... */
  glGetIntegerv(GL_VIEWPORT,win_geom);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0,win_geom[2],0,win_geom[3]);
  glMatrixMode(GL_MODELVIEW);
}
  
static dspPatch restoreDspPatch(dspPatch patch)
{
  if (patch.values) {
    glRasterPos2f(patch.x,patch.y);
    glDrawPixels(patch.width,patch.height,GL_RGB,GL_UNSIGNED_BYTE,patch.values);
    free(patch.values);
    patch.values=(char *)NULL;
    patch.size=0;
    patch.width=0;
    patch.height=0;
  }
  return(patch);
}

static dspPatch fetchDspPatch(dspPatch patch)
{
  patch.size=patch.width*patch.height;
  if (patch.size>0) {
    if (patch.values) free(patch.values);
    patch.values=malloc(4*patch.size*sizeof(char));
    if (patch.values) {
      patch.x=clamp(patch.x,0,win_geom[2]-1);
      patch.y=clamp(patch.y,0,win_geom[3]-1);
      glReadPixels(patch.x,patch.y,
		   patch.width,patch.height,
		   GL_RGB,GL_UNSIGNED_BYTE,
		   patch.values);
    }
  }
  return(patch);
}

static dspPatch clearDspPatch(dspPatch patch)
{
  dspPatch r={NULL,0,0,0,0,0};
  if (patch.values) free(patch.values);
  return(r);
}

/*======================================================================*/

dspPatch WnpCursor_none(int x,int y)
{
  dspPatch r={NULL,0,0,0,0,0};
  
  return(r);     /* that was easy! */
}

dspPatch WnpCursor_DOTBOX(int x,int y)
{
  dspPatch r={NULL,0,0,0,0,0};

  /* 17x17 square box with dot in center at x,y */

  y=win_geom[3]-y;  /* GLUT to GL ... */
  r.x=x-9; r.y=y-9; r.width=20; r.height=20;
  r=fetchDspPatch(r);
  WnpNoColor(); WnpGlLineWidth(1);
  glColor3fv(cursorColor[curtab[current_cursor].color]);
  glBegin(GL_LINE_LOOP);
  glVertex2i(x-8,y-8);
  glVertex2i(x-8,y+8);
  glVertex2i(x+8,y+8);
  glVertex2i(x+8,y-8);
  glEnd();
  glBegin(GL_POINTS);
  glVertex2i(x,y);
  glEnd();
  glFlush();
  return(r);
}

dspPatch WnpCursor_dotbox(int x,int y)
{
  dspPatch r={NULL,0,0,0,0,0};

  /* half-height (stereo) version of DOTBOX */

  y=win_geom[3]-y;  /* GLUT to GL ... */
  r.x=x-9; r.y=y-5; r.width=20; r.height=10;
  r=fetchDspPatch(r);
  WnpNoColor(); WnpGlLineWidth(1);
  glColor3fv(cursorColor[curtab[current_cursor].color]);
  glBegin(GL_LINE_LOOP);
  glVertex2i(x-8,y-4);
  glVertex2i(x-8,y+4);
  glVertex2i(x+8,y+4);
  glVertex2i(x+8,y-4);
  glEnd();
  glBegin(GL_POINTS);
  glVertex2i(x,y);
  glEnd();
  return(r);
}

dspPatch WnpCursor_ARROW(int x,int y)
{
  dspPatch r={NULL,0,0,0,0,0};

  /* 17x17 up-left arrow , point at x,y */

  y=win_geom[3]-y;  /* GLUT to GL ... */
  r.x=x-1; r.y=y-17;r.width=12; r.height=20;
  r=fetchDspPatch(r);
  WnpNoColor(); WnpGlLineWidth(1);
  glColor3fv(cursorColor[curtab[current_cursor].color]);
  glBegin(GL_POLYGON);
  glVertex2i(x,y);
  glVertex2i(x,y-10);
  glVertex2i(x+7,y-7);
  glEnd();
  glBegin(GL_POLYGON);
  glVertex2i(x,y-3);
  glVertex2i(x+5,y-15);
  glVertex2i(x+7,y-14);
  glVertex2i(x+2,y-2);
  glEnd();
  return(r);
}

dspPatch WnpCursor_arrow(int x,int y)
{
  dspPatch r={NULL,0,0,0,0,0};

  /* half height version of ARROW  */

  y=win_geom[3]-y;  /* GLUT to GL ... */
  r.x=x-1; r.y=y-9; r.width=12; r.height=12;
  r=fetchDspPatch(r);
  WnpNoColor(); WnpGlLineWidth(1);
  glColor3fv(cursorColor[curtab[current_cursor].color]);
  glBegin(GL_POLYGON);
  glVertex2i(x,y);
  glVertex2i(x,y-5);
  glVertex2i(x+7,y-4);
  glEnd();
  glBegin(GL_POLYGON);
  glVertex2i(x,y-2);
  glVertex2i(x+5,y-8);
  glVertex2i(x+7,y-7);
  glVertex2i(x+2,y-1);
  glEnd();
  return(r);
}
