#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "wnpgl.h"
#include "wnp.h"

/* Assumption: in model-view matrix mode, ... */

static void updateMatrix(GlblObj *o);
static int objDisplayed(DspObj *obj);

#define MAX_DOTS 3

static int wnpMarkColor=3;      /* color index for atom marks... */
static float wnpAtomLabelSize = 0.2;
static int MVTFlag=TRUE;        /* current MVT matrices ? */

typedef union {
  void *ptr;
  struct {
    unsigned int lsb;
    unsigned int msb;
  } A;
} Addr;

#define NUINTPERPTR (sizeof(void *)/sizeof(GLuint))

static void LoadName(void *ptr);
static void PopName(void);
static void *NameToPtr(GLuint *pickBuffer);

/*------------------------------------------------------------------------*/

void WnpDrawObject(DspObj *obj)
{
  if (obj) {
    if (obj->flags&OBJ_FLG_DSP) {
      switch (obj->type) {
      case DSP_GLBL:
	WnpDrawGlobal((GlblObj *)obj);
	break;
      case DSP_MDL:
	WnpDrawMdl((MdlObj *)obj);
	break;
      case DSP_CSO:
	WnpDrawCSO((CSObj *)obj);
      case DSP_TRG:
	WnpDrawTRG((TRGObj *)obj);
	break;
      }
    }
    obj->flags&=~OBJ_FLG_MOD;
  }
}

void WnpDrawGlobal(GlblObj *r)
{
  DspObj *obj;

  if (r&&r->type==DSP_GLBL) {
    glPushMatrix(); updateMatrix(r); 
    glGetFloatv(GL_MODELVIEW_MATRIX,(float *)r->mvt);
    WnpPlotMVT(r->mvt);
    for (obj=r->fst;obj;obj=obj->nxt) WnpDrawObject(obj);
    glPopMatrix();
  }
}

void WnpDrawMdl(MdlObj *mdl)
{
  if (mdl&&mdl->type==DSP_MDL) {
    if (mdl->flags&OBJ_FLG_RBLD) WnpRebuildMdl(mdl);
    if (mdl->flags&OBJ_FLG_ACTIVE) {
      WnpDrawMdlAtoms(mdl);
      WnpDrawMdlBonds(mdl);
      WnpDrawMdlDots(mdl);
      WnpDrawNOEList(mdl);
    }
  }
}

void WnpDrawMdlBonds(MdlObj *mdl)
{
  WnpBond *b;
  float x[3];
  int cw;

  b=mdl->fstBond->dsp;
  if (b) {
    WnpSetColor(1); WnpSetColor(2);  /* workaround for a (suspected) bug in some versions of libGL */
    if (b->flags&BND_FLG_CYLINDER) WnpPrepareForSolid();      

    /* cylinders... */

    while (b&&(b->flags&BND_FLG_CYLINDER)) {
      if (b->f->color==b->t->color) {
	WnpSetColor(b->f->color);
	WnpDrawCylinder(b->f->x,b->r,b->t->x,b->r);
      }else{
	x[0]=0.5*(b->f->x[0]+b->t->x[0]);
	x[1]=0.5*(b->f->x[1]+b->t->x[1]);
	x[2]=0.5*(b->f->x[2]+b->t->x[2]);
	WnpSetColor(b->f->color);
	WnpDrawCylinder(b->f->x,b->r,x,b->r);
	WnpSetColor(b->t->color);
	WnpDrawCylinder(x,b->r,b->t->x,b->r);
      }
      b=b->dsp;
    }
    if (mdl->fstBond->dsp->flags&BND_FLG_CYLINDER) WnpRestoreAfterSolid();
    cw=1;
    WnpVLineWidth(1);

    while (b) {

      /* lines... */

      WnpGlBegin(GL_LINES);
      while(b&&b->lw==cw) {
	if (b->f->color==b->t->color) {
	  WnpSetColor(b->f->color);
	  WnpVertex(b->f->x);
	  WnpVertex(b->t->x);
	}else{
	  x[0]=0.5*(b->f->x[0]+b->t->x[0]);
	  x[1]=0.5*(b->f->x[1]+b->t->x[1]);
	  x[2]=0.5*(b->f->x[2]+b->t->x[2]);
	  WnpSetColor(b->f->color);
	  WnpVertex(b->f->x);
	  WnpVertex(x);
	  WnpSetColor(b->t->color);
	  WnpVertex(x);
	  WnpVertex(b->t->x);
	}
	b=b->dsp;
      }
      WnpGlEnd();
      if (b) {
	cw=b->lw;
	WnpVLineWidth(cw);
      }
    }
    WnpVLineWidth(1);
  }

  /* bond labels ... */

  if (mdl->fstBond->lab) {
    WnpTextIniCur(5);
    WnpTextSize(wnpAtomLabelSize,1.15*wnpAtomLabelSize);
    for (b=mdl->fstBond->lab;b;b=b->lab) {
      WnpSetColor(b->lcolor);
      x[0]=0.5*(b->f->x[0]+b->t->x[0]);
      x[1]=0.5*(b->f->x[1]+b->t->x[1]);
      x[2]=0.5*(b->f->x[2]+b->t->x[2]);
      WnpDrawText(x,mdl->up->mvt,b->label);
    }
  }
}

void WnpDrawMdlAtoms(MdlObj *mdl)
{
  WnpAtom *a;

  /* solid spheres... */

  if (mdl->fstAtom->sph) {
    WnpPrepareForSolid();
    for (a=mdl->fstAtom->sph;a;a=a->sph) {
      WnpSetColor(a->color);
      WnpDrawSphere(a->x,a->r);
    }
    WnpRestoreAfterSolid();
  }
    
  /* atom marks and isolated atoms ... */

  if (mdl->fstAtom->mid&&mdl->fstAtom->mid->flags&ATM_FLG_MARK) {
    WnpSetColor(wnpMarkColor);
    WnpPointSize(7);
    WnpGlBegin(GL_POINTS);
    for (a=mdl->fstAtom->mid;a&&a->flags&ATM_FLG_MARK;a=a->mid) WnpVertex(a->x);
    WnpGlEnd();
  }else a=mdl->fstAtom->mid;
    
  if (a&&a->flags&ATM_FLG_ISO) {
    WnpPointSize(3);
    WnpGlBegin(GL_POINTS);
    while (a&&a->flags&ATM_FLG_ISO) {
      WnpSetColor(a->color);
      WnpVertex(a->x);
      a=a->mid;
    }
    WnpGlEnd();
  }
  if (mdl->fstAtom->mid) WnpPointSize(1);

  /* atom labels */
  
  if (mdl->fstAtom->lab) {
    WnpTextIniCur(1);
    WnpTextSize(wnpAtomLabelSize,1.15*wnpAtomLabelSize);
    for (a=mdl->fstAtom->lab;a;a=a->lab) {
      WnpSetColor(a->lcolor==WNP_COLOR_BYATOM ? a->color : a->lcolor);
      WnpDrawText(a->x,mdl->up->mvt,a->label);
    }
  }
  WnpTextPos3f(0,0,0);    /*  workaround for problem on Indy... */
}

void WnpDrawMdlDots(MdlObj *mdl)
{
  float *p;
  int color,i,d,dp;
  WnpDots *dots;

  dp=1;if (mdl->fstDot->dot) WnpGlBegin(GL_POINTS);
  for (dots=mdl->fstDot->dot;dots;dots=dots->dot) {
    color = ( dots->color == DOT_BYATOM ? dots->up->color : dots->color);
    if (color>=0) {
      WnpSetColor(color);
      d=clamp(dots->style,0,MAX_DOTS)+1;
      if (d!=dp){ 
	WnpGlEnd(); dp=d;
	WnpPointSize(d);
	WnpGlBegin(GL_POINTS);
      }
      for (i=0,p=dots->xyz;i<dots->n;i++,p+=4) WnpVertex(p);
    }else{
      d=clamp(dots->style,0,MAX_DOTS)+1;
      if (d!=dp){ 
	WnpGlEnd(); dp=d;
	WnpPointSize(d);
	WnpGlBegin(GL_POINTS);
      }
      for (i=0,p=dots->xyz;i<dots->n;i++,p+=4) {
	WnpDotColor(p[3]);
	WnpVertex(p);
      }
    }
  }
  if (mdl->fstDot->dot) {
    WnpGlEnd();
    if (dp>1) WnpPointSize(1);
  }
}

void WnpDrawInterBonds(WnpBond *lst)
{
  /* this will be slow, if the number of inter bonds is large... */

  WnpBond *m,*b;
  WnpMatrix mvt;
  float x[3],y[3],z[3];
  int i,sol,cw;

  if (lst&&lst->nxt) {
    glGetFloatv(GL_MODELVIEW_MATRIX,(float *)mvt);
    WnpPlotMVT(mvt);
    cw=1; sol=FALSE;
    WnpTextIniCur(5);
    WnpTextSize(wnpAtomLabelSize,1.15*wnpAtomLabelSize);
    for (b=lst;m=b->nxt;) {
      if (!(m->flags&BND_FLG_REMOVE)&&(m->f->flags|m->t->flags)&ATM_FLG_DSP) {
	if (objDisplayed((DspObj *)m->f->up)&&objDisplayed((DspObj *)m->f->up)) {
	  WnpTransform(m->f->up->up->mvt,m->f->x,x); 
	  WnpTransform(m->t->up->up->mvt,m->t->x,y);
	  for (i=0;i<3;i++) z[i]=0.5*(x[i]+y[i]);
	  if (m->flags&BND_FLG_CYLINDER) {
	    if (!sol) WnpPrepareForSolid(); sol=TRUE;
	    if (m->f->color==m->t->color) {
	      WnpSetColor(m->f->color);
	      WnpDrawCylinder(x,m->r,y,m->r);
	    }else{
	      WnpSetColor(m->f->color);
	      WnpDrawCylinder(x,m->r,z,m->r);
	      WnpSetColor(m->t->color);
	      WnpDrawCylinder(z,m->r,y,m->r);
	    }
	  }else{
	    if (sol) WnpRestoreAfterSolid(); sol=FALSE;
	    if (cw!=m->lw) {
	      WnpVLineWidth(m->lw);
	      cw=m->lw;
	    }
	    WnpGlBegin(GL_LINES);
	    if (m->f->color==m->t->color) {
	      WnpSetColor(m->f->color);
	      WnpVertex(x);
	      WnpVertex(y);
	    }else{
	      WnpSetColor(m->f->color);
	      WnpVertex(x);
	      WnpVertex(z);
	      WnpSetColor(m->t->color);
	      WnpVertex(z);
	      WnpVertex(y);
	    }
	    WnpGlEnd();
	  }
	  if (m->label) {
	    if (sol) WnpRestoreAfterSolid(); sol=FALSE;
	    if (cw!=1) {
	      WnpVLineWidth(1);
	      cw=1;
	    }
	    WnpSetColor(m->lcolor);
	    WnpDrawText(z,mvt,m->label);
	  }
	}
      }
      if (m->flags&BND_FLG_REMOVE) {
	b->nxt=m->nxt;
	WnpDeleteBond(m);
      }else b=m;
    }
    if (sol) WnpRestoreAfterSolid();
    if (cw!=1) WnpVLineWidth(1);
  }
}

void WnpDrawMonitorList(WnpMonPair *lst)
{
  WnpMonPair *m;
  WnpMatrix mvt;
  float x[3],y[3],d;
  char label[WNP_MAXREC];
  int i;

  if (lst) {
    glGetFloatv(GL_MODELVIEW_MATRIX,(float *)mvt);
    WnpPlotMVT(mvt);
    WnpLineStipple(WNP_DOTTED_LINE);
    WnpTextIniCur(5); WnpTextSize(wnpAtomLabelSize,1.15*wnpAtomLabelSize);
    for (m=lst;m;m=(WnpMonPair *)m->nxt) {
      if (m->dsp&&m->f->flags&m->t->flags&ATM_FLG_DSP) {
	if (objDisplayed((DspObj *)m->f->up)&&objDisplayed((DspObj *)m->t->up)) {
	  WnpTransform(m->f->up->up->mvt,m->f->x,x); 
	  WnpTransform(m->t->up->up->mvt,m->t->x,y);
	  WnpSetColor(m->color);
	  WnpVLineWidth(m->width);
	  WnpGlBegin(GL_LINES);
	  WnpVertex(x);
	  WnpVertex(y);
	  WnpGlEnd();
	  if (m->label) {
	    for (d=0,i=0;i<3;i++) {
	      d+=(x[i]-y[i])*(x[i]-y[i]);
	      x[i]=0.5*(x[i]+y[i]);
	    }
	    sprintf(label,"%0.3f",sqrt(d));
	    WnpLineStipple(WNP_SOLID_LINE);
	    if (m->width>2) WnpVLineWidth(2);
	    WnpDrawText(x,mvt,label);
	    WnpLineStipple(WNP_DOTTED_LINE);
	  }
	}
      }
    }
    WnpVLineWidth(1);
    WnpLineStipple(WNP_SOLID_LINE);
  }
}

void WnpDrawNOEList(MdlObj *mdl)
{
  WnpNOE *m;
  float x[3],xon[3],xoff[3],s;
  char label[WNP_MAXREC];
  int i,j,broken,dotted;

  if (mdl->fstNOE->dsp) {
    broken=mdl->NOEstyle&WNP_NOE_BROKEN;
    dotted=mdl->NOEstyle&WNP_NOE_DOTTED;
    WnpTextIniCur(5); WnpTextSize(wnpAtomLabelSize,1.15*wnpAtomLabelSize);
    for (m=(WnpNOE *)mdl->fstNOE->dsp;m;m=(WnpNOE *)m->dsp) {
      WnpSetColor(m->color);
      WnpVLineWidth(mdl->NOEwidth);
      if (m->f!=m->t) {
	if (broken) {
	  s=m->D/(m->D+m->viol);
	  for (j=0;j<3;j++) {
	    xon[j]=0.5*((1+s)*m->f->x[j]+(1-s)*m->t->x[j]);
	    xoff[j]=0.5*((1-s)*m->f->x[j]+(1+s)*m->t->x[j]);
	  }
	  WnpGlBegin(GL_LINES);
	  WnpVertex(m->f->x);
	  WnpVertex(xon);
	  WnpVertex(xoff);
	  WnpVertex(m->t->x);
	  WnpGlEnd();
	  if (dotted) {
	    WnpLineStipple(WNP_DOTTED_LINE);
	    WnpGlBegin(GL_LINES);
	    WnpVertex(xon);
	    WnpVertex(xoff);
	    WnpGlEnd();
	    WnpLineStipple(WNP_SOLID_LINE);
	  }
	}else{
	  if (dotted) WnpLineStipple(WNP_DOTTED_LINE);
	  WnpGlBegin(GL_LINES);
	  WnpVertex(m->f->x);
	  WnpVertex(m->t->x);
	  WnpGlEnd();
	  if (dotted) WnpLineStipple(WNP_SOLID_LINE);
	}
      }
      switch (mdl->NOElabt) {
      case WNP_NOE_LAB_VIOL:
	for (i=0;i<3;i++) x[i]=0.5*(m->f->x[i]+m->t->x[i]);
	sprintf(label,"%0.3f",m->viol);
	if (mdl->NOEwidth>2) WnpVLineWidth(2);
	WnpDrawText(x,mdl->up->mvt,label);
	break;
      case WNP_NOE_LAB_LIM:
	for (i=0;i<3;i++) x[i]=0.5*(m->f->x[i]+m->t->x[i]);
	sprintf(label,"%0.3f",m->D);
	if (mdl->NOEwidth>2) WnpVLineWidth(2);
	WnpDrawText(x,mdl->up->mvt,label);
	break;
      }
    }
    WnpVLineWidth(1);
  }
}


int WnpPickAtoms(DspObj *root,WnpAtom **atoms,int maxpicks)
{
  int i,j,k,nn,n,nhits;
  GLint old;
  GLuint pickBuffer[WNP_PICK_BUFFER_SIZE];
  WnpAtom *a;
  
  /* returns number of picks; only
     the first maxpick hits will be stored in atoms[].  */

  if (root) {
    glSelectBuffer(WNP_PICK_BUFFER_SIZE,pickBuffer);
    glGetIntegerv(GL_RENDER_MODE,&old);
    glRenderMode(GL_SELECT); glInitNames(); 
    glPushName(1);         /* needed: glLoadName on empty stack generates error */
    WnpAtomPicker(root);
    nhits=glRenderMode(old);
    if (nhits>=0) {
      for (i=0,j=0,k=0;i<nhits&&i<maxpicks;i++) {
	nn=pickBuffer[k];k+=3;
	a=NameToPtr(pickBuffer+k);
	if (a) atoms[j++]=a;
	k+=nn;
      }
      n=nhits;
    }else n=0;
  }else n=0;
  return(n);
}

void WnpAtomPicker(DspObj *obj) 
{
  DspObj *o;
  WnpAtom *a;

  if (obj->flags&OBJ_FLG_DSP) {
    if (obj->type==DSP_GLBL) {
      for (o=((GlblObj*)obj)->fst;o;o=o->nxt) WnpAtomPicker(o);
    }else if (obj->type==DSP_MDL&&obj->flags&OBJ_FLG_ACTIVE) {
      glPushMatrix(); 
      glLoadMatrixf((float *)obj->up->mvt);
      for (a=((MdlObj *)obj)->fstAtom->mid;a;a=a->mid) {
	LoadName(a);
	glRasterPos3fv(a->x);
	PopName();
      }
      glPopMatrix();
    }
  }
}

int WnpPickObjects(DspObj *root,DspObj **objects,int maxpicks)
{
  int i,j,k,nn,n,nhits;
  GLint old;
  GLuint pickBuffer[WNP_PICK_BUFFER_SIZE];
  DspObj *o;
  
  /* returns number of picks in n; only
     the first maxpick hits will be stored in objects[]   */

  if (root) {
    glSelectBuffer(WNP_PICK_BUFFER_SIZE,pickBuffer);
    glGetIntegerv(GL_RENDER_MODE,&old);
    glRenderMode(GL_SELECT); glInitNames(); 
    glPushName(1);         /* needed: glLoadName on empty stack generates error */
    WnpObjectPicker(root);
    nhits=glRenderMode(old);
    if (nhits>=0) {
      for (i=0,j=0,k=0;i<nhits&&i<maxpicks;i++) {
	nn=pickBuffer[k];k+=3;
	o=NameToPtr(pickBuffer+k);
	if (o) objects[j++]=o;
	k+=nn;
      }
      n=nhits;
    }else n=-1;
  }else n=0;
  return(n);
}

void WnpObjectPicker(DspObj *obj) 
{
  DspObj *o;
  WnpAtom *a;

  if (obj->flags&OBJ_FLG_DSP) {
    if (obj->type==DSP_GLBL) {
      for (o=((GlblObj*)obj)->fst;o;o=o->nxt) WnpObjectPicker(o);
    }else if (obj->type==DSP_MDL&&obj->flags&OBJ_FLG_ACTIVE) {
      glPushMatrix(); 
      glLoadMatrixf((float *)obj->up->mvt);
      LoadName(obj);
      /* only atoms are pickable... */
      for (a=((MdlObj *)obj)->fstAtom->mid;a;a=a->mid) glRasterPos3fv(a->x);
      PopName();
      glPopMatrix();
    }
  }
}

void WnpFreezeMVT(DspObj *o,WnpMatrix mvt)
{
  DspObj *s;
  GlblObj *g;
  float piv[3];
  WnpMatrix mat;
  int i,j;
  
  if (o) {
    switch (o->type) {
    case DSP_GLBL:
      g=(GlblObj *)o;
      glPushMatrix(); updateMatrix(g);
      glGetFloatv(GL_MODELVIEW_MATRIX,(GLfloat *)mat);
      for (s=g->fst;s;s=s->nxt) WnpFreezeMVT(s,mat);
      glPopMatrix();
      for (i=0;i<3;i++) 
	for (piv[i]=mat[3][i],j=0;j<3;j++) piv[i]+=g->pivot[j]*mat[j][i];
      for (i=0;i<3;i++) g->pivot[i]=piv[i];
      WnpResetMVT(o);
      break;
    case DSP_MDL:
      WnpTransformMdl((MdlObj *)o,mvt);
      break;
    case DSP_CSO:
      WnpTransformCSO((CSObj *)o,mvt);
      break;
    case DSP_TRG:
      WnpTransformTRG((TRGObj *)o,mvt);
      break;
    }
  }
}

void WnpTransform(WnpMatrix m,float *x,float *y)
{
  int i,j;
  float z[3];

  for (i=0;i<3;i++) 
    for (z[i]=m[3][i],j=0;j<3;j++) z[i]+=x[j]*m[j][i];
  y[0]=z[0]; y[1]=z[1]; y[2]=z[2];
}

void WnpSetAtomLabelSize(float s)
{
  wnpAtomLabelSize=clamp(s,0.001,1000);
}

void WnpDrawCSO(CSObj *obj)
{
  WnpLst *l;
  WnpMDS *mds;
  WnpTOBJ *tob;
  float *p;
  int i;

  if (obj&&obj->type==DSP_CSO) {

    WnpVLineWidth(obj->lw);
    WnpTextSize(3*wnpAtomLabelSize,3*1.15*wnpAtomLabelSize);
    WnpTextIniCur(1);

    for (l=obj->fst;l;l=l->nxt) {
      switch (l->type) {
      case WNP_LST_MDS:
	mds=(WnpMDS *)l;
	if (mds->n>0&&mds->x) {
	  WnpSetColor(mds->color);
	  switch (obj->mds_type) {
	  case WNP_CSO_POINTS:
	    WnpVPointSize(obj->lw);
	    WnpGlBegin(GL_POINTS);
	    for (i=0,p=mds->x;i<mds->n;i++,p+=3) WnpVertex(p);
	    WnpGlEnd();
	    break;
	  case WNP_CSO_LINES:
	    WnpGlBegin(GL_LINES);
	    for (i=0,p=mds->x;i<mds->n;i++,p+=3) WnpVertex(p);
	    WnpGlEnd();
	    break;
	  }
	}
	break;
      case WNP_LST_TOBJ:
	tob=(WnpTOBJ *)l;
	if (tob->text) {
	  WnpSetColor(tob->color);
	  WnpDrawText(tob->x,obj->up->mvt,tob->text);
	}
      }
    }

    WnpVLineWidth(1);
  }
}

void WnpDrawTRG(TRGObj *obj)
{
  WnpLst *l;
  WnpTRG *trg;
  float *N,*V,n[3];
  int *t,k,front,back;
  float spec[]={1.0,1.0,1.0,1.0};
  float epsilon[]={0.1,0.1,0.1,1};
      
  if (obj&&obj->type==DSP_TRG) {

    front=obj->color;
    back =(obj->bckclr > 0 ? obj->bckclr : obj->color);

    switch (obj->style) {
    case WNP_TRG_RENDERED:
      glPushAttrib(GL_ENABLE_BIT|GL_POLYGON_BIT|GL_LIGHTING_BIT);
      glEnable(GL_LIGHTING); glEnable(GL_LIGHT0); glShadeModel(GL_SMOOTH);
      glDisable(GL_NORMALIZE); glDisable(GL_BLEND);

      glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
      glDisable(GL_CULL_FACE);

      glMaterialfv(GL_FRONT,GL_SPECULAR,spec);
      WnpMaterialColor(GL_FRONT,GL_AMBIENT_AND_DIFFUSE,front,1,1.0);
      glMaterialf(GL_FRONT,GL_SHININESS,50.0);
      
      glMaterialfv(GL_BACK,GL_SPECULAR,epsilon);
      WnpMaterialColor(GL_BACK,GL_AMBIENT_AND_DIFFUSE,back,1,0.5);
      glMaterialf(GL_BACK,GL_SHININESS,0.5);

      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

      if (obj->pat>0&&obj->stipple) {
	glEnable(GL_POLYGON_STIPPLE);
	glPolygonStipple((GLubyte *)obj->stipple);
      }

      for (l=obj->fst;l;l=l->nxt) {
	switch (l->type) {
	case WNP_LST_TRG:
	  trg=(WnpTRG *)l;
	  if (trg->v&&trg->t&&trg->nt>0) {
	    V=trg->v; N=V+3;
	    WnpGlBegin(GL_TRIANGLES);
	    for (t=trg->t,k=0;k<trg->nt;k++,t+=3) {
	      glNormal3fv(N+t[0]); WnpVertex(V+t[0]);
	      glNormal3fv(N+t[1]); WnpVertex(V+t[1]);
	      glNormal3fv(N+t[2]); WnpVertex(V+t[2]);
	    }
	    WnpGlEnd();
	  }
	  break;
	}
      }
      glPopAttrib();
      break;

    case WNP_TRG_MESH:
      WnpVLineWidth(obj->lw);
      WnpSetColor(obj->color);
      for (l=obj->fst;l;l=l->nxt) {
	switch (l->type) {
	case WNP_LST_TRG:
	  trg=(WnpTRG *)l;
	  if (trg->v&&trg->t&&trg->nt>0) {
	    V=trg->v; N=V+3;
	    WnpGlBegin(GL_LINES);
	    for (t=trg->t,k=0;k<trg->nt;k++,t+=3) {
	      WnpVertex(V+t[0]); WnpVertex(V+t[1]);
	      WnpVertex(V+t[1]); WnpVertex(V+t[2]);
	      WnpVertex(V+t[2]); WnpVertex(V+t[0]);
	    }
	    WnpGlEnd();
	  }
	  break;
	}
      }
      WnpVLineWidth(1);
      break;

    case WNP_TRG_DOTS:
      WnpVPointSize(obj->lw);
      WnpSetColor(obj->color);
      for (l=obj->fst;l;l=l->nxt) {
	switch (l->type) {
	case WNP_LST_TRG:
	  trg=(WnpTRG *)l;
	  if (trg->v&&trg->nv>0) {
	    V=trg->v;
	    WnpGlBegin(GL_POINTS);
	    for (k=0;k<trg->nv;k++,V+=6) WnpVertex(V);
	    WnpGlEnd();
	  }
	  break;
	}
      }
      WnpVPointSize(1);
      break;

    case WNP_TRG_NORMALS:
      WnpVLineWidth(obj->lw);
      WnpSetColor(obj->color);
      for (l=obj->fst;l;l=l->nxt) {
	switch (l->type) {
	case WNP_LST_TRG:
	  trg=(WnpTRG *)l;
	  if (trg->v&&trg->nv>0) {
	    V=trg->v; N=V+3;
	    WnpGlBegin(GL_LINES);
	    for (k=0;k<trg->nv;k++,V+=6,N+=6) {
	      WnpVertex(V);
	      n[0]=V[0]+N[0]; n[1]=V[1]+N[1]; n[2]=V[2]+N[2]; 
	      WnpVertex(n);
	    }
	    WnpGlEnd();
	  }
	  break;
	}
      }
      WnpVLineWidth(1);
      break;
    }
  }
}
/*-------------------- WnpUpdMVT --------------------*/

void WnpUpdMVT(void)
{
  if (!MVTFlag) {
    WnpZeroViewShift();        /* since we don't have a proper 3D projection ...*/
    glPushAttrib(GL_TRANSFORM_BIT);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    WnpUpdGlblMVT(WnpDisplayRoot());
    glPopMatrix();
    glPopAttrib();
  }
  WnpSetMVTFlag(TRUE);
}

void WnpUpdGlblMVT(GlblObj *r)
{
  DspObj *obj;

  if (r&&r->type==DSP_GLBL) {
    glPushMatrix(); updateMatrix(r); 
    glGetFloatv(GL_MODELVIEW_MATRIX,(float *)r->mvt);
    for (obj=r->fst;obj;obj=obj->nxt) {
      if (obj->type==DSP_GLBL) WnpUpdGlblMVT((GlblObj *)obj);
    }
    glPopMatrix();
  }
}

int WnpSetMVTFlag(int s)
{
  int r;

  r=MVTFlag;
  MVTFlag=s;
  return(r);
}
/*------------------------------------------------------------------------*/

static void updateMatrix(GlblObj *o)
{
  glTranslatef(o->tra[0]+o->pivot[0],
	       o->tra[1]+o->pivot[1],
	       o->tra[2]+o->pivot[2]);
  glMultMatrixf((float *)o->rot);
  glTranslatef(-o->pivot[0],-o->pivot[1],-o->pivot[2]);
}

static int objDisplayed(DspObj *obj)
{
  int r;

  if (obj) r=obj->flags&OBJ_FLG_DSP&&objDisplayed((DspObj *)obj->up);
  else r=TRUE;
  return(r);
}

/*---------------- the following is for picking --------------------------*/

static void LoadName(void *ptr)
{
  int n;
  Addr A;

  A.ptr=ptr;
  if (NUINTPERPTR==1) {
    glLoadName(A.A.lsb);
  }else if (n=2) {
    glLoadName(A.A.lsb);
    glPushName(A.A.msb);
  }
}

static void PopName(void)
{
  int n;
  Addr A;

  if (NUINTPERPTR==2) {
    glPopName();
  }
}

static void *NameToPtr(GLuint *pickBuffer)
{
  Addr A;

  if (NUINTPERPTR==1) {
    A.A.lsb=pickBuffer[0];
  }else if (NUINTPERPTR==2) {
    A.A.lsb=pickBuffer[0];
    A.A.msb=pickBuffer[1];
  }else A.ptr=NULL;

  return A.ptr;
}
