#include <stdio.h>
#include <math.h>
#include "wnp.h"
#include "wnpgl.h"

/* Solid objects (sphere, cone, cylinder... ) */

static int wnpSphereMethod = WNP_SPHERE_GLU;
static int wnpSphereStyle = WNP_SOLID_RENDERED;
static int wnpHalfSpheres = TRUE;
static int wnpSphereQuality = 9;
static WnpMatrix LightRot;

static int wnpCylinderQuality = 6;

/* data for GLU library quadrics ... */

static GLUquadricObj *WnpGluSphere=(GLUquadricObj *)NULL;
static GLUquadricObj *WnpGluCylinder=(GLUquadricObj *)NULL;
static int wnpQuadricStyle[3] = {GLU_FILL, GLU_LINE, GLU_POINT};

/* -------------------- Data for polyhedron tesselation --------------------*/

/* 0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15 */
/*16  17  18  19  20  21  22  23  24  25  26  27  28  29  30  31 */

static int wnpIcoDepth[WNP_MAX_iSPHERE_QUALITY] = {
   0,  0,  1,  0,  0,  1,  2,  1,  1,  2,  3,  2,  2,  3,  4,  3,
   3,  4,  5,  4,  4
};
static char wnpIcoBase[WNP_MAX_iSPHERE_QUALITY] = {
  'T','O','T','I','I','O','T','I','I','O','T','I','I','O','T','I',
  'I','O','T','I','I' 
};

GLuint icoSphere  =  0;
int    icoQuality = -1;

/* Icosahedron data (ref.: OpenGL Programming Guide) */

#define X 0.525731112119133606
#define Z 0.850650808352039932

static float IV[12][3] = {                         /* vertices */
  {X,0,Z}, {-X,0,Z}, {X,0,-Z}, {-X,0,-Z},
  {-0,Z,X}, {0,Z,-X}, {0,-Z,X}, {0,-Z,-X},
  {-Z,X,0}, {Z,X,0}, {-Z,-X,0}, {Z,-X,0}
};

static int IT[20][3] = {                           /* triangles */
  {0,4,1},  {0,9,4},  {9,5,4}, {4,5,8}, {4,8,1},
  {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},
  {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6},
  {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11}
};
/*

Platonic Solids, by Alan Paeth 

Coordinates for these and for their four-dimensional analogs were published 
by HSM Coxeter, first in 1948 in Regular Polytopes, pg. 52-53 (Methuen, London 
and again in subsequent revisions; any/all are highly recommended reading. The
table for (quasi) regular 3D polyhedra is transcribed below. I've posted this
a few times already; perhaps a "frequently asked" entry is in order. 

                        PLATONIC SOLIDS
                (regular and quasi-regular variety,
                Kepler-Poinset star solids omitted)

The orientations minimize the number of distinct coordinates, thereby revealing
both symmetry groups and embedding (eg, tetrahedron in cube in dodecahedron).
Consequently, the latter is depicted resting on an edge (Z taken as up/down). 

    SOLID            VERTEX COORDINATES
    -----------      ----------------------------------
    Tetrahedron      (  1,  1,  1), (  1, -1, -1), ( -1,  1, -1), ( -1, -1,  1)
    Cube             (+-1,+-1,+-1)
    Octahedron       (+-1,  0,  0), (  0,+-1,  0), (  0,  0,+-1)
    Cubeoctahedron   (  0,+-1,+-1), (+-1,  0,+-1), (+-1,+-1,  0)
    Icosahedron      (  0,+-p,+-1), (+-1,  0,+-p), (+-p,+-1,  0)
    Dodecahedron     (  0,+-i,+-p), (+-p,  0,+-i), (+-i,+-p,  0), (+-1,+-1,+-1)
    Icosidodecahedron(+-2,  0,  0), (  0,+-2,  0), (  0,  0,+-2), ...
                     (+-p,+-i,+-1), (+-1,+-p,+-i), (+-i,+-1,+-p)

    with golden mean: p = (sqrt(5)+1)/2; i = (sqrt(5)-1)/2 = 1/p = p-1
*/
#define tetX 0.577350269190

static float TV[4][3] = {
  {tetX,tetX,tetX}, {tetX,-tetX,-tetX}, {-tetX,tetX,-tetX}, {-tetX,-tetX,tetX}
};

static int   TT[4][3] = {
  {0,3,1}, {0,2,3}, {0,1,2}, {1,3,2}
};


static float OV[6][3] = {
  {-1,0,0},{0,-1,0},{1,0,0},{0,1,0},{0,0,-1},{0,0,1}
};

static int   OT[8][3] = {
  {0,1,5}, {1,2,5}, {2,3,5}, {3,0,5},
  {0,4,1}, {1,4,2}, {2,4,3}, {3,4,0}
};


/*-------------------- prototypes for local functions --------------------*/

static GLuint MakeIcoSphere(void);
static int TriangleDiv(int depth,float *a,float *b,float *c);
#ifdef MYSPHERE
static GLuint MakeMySphere(void);
#endif

/*------------------------------------------------------------------------*/

void WnpPrepareForSolid(void)
{
  int i,j;
  float spec[]={1.0,1.0,1.0,1.0};
  float s,lp[4],x[3];
  WnpMatrix mvt;

  /* set up for "solid" rendering */

  WnpNoColor();
  glPushAttrib(GL_ENABLE_BIT|GL_POLYGON_BIT|GL_LIGHTING_BIT);
  glCullFace(GL_BACK);
  if (wnpHalfSpheres) {
    glCullFace(GL_BACK);
    glEnable(GL_CULL_FACE);
  }else glDisable(GL_CULL_FACE);
  if (wnpSphereMethod==WNP_SPHERE_ICO) glEnable(GL_NORMALIZE); 
  else glDisable(GL_NORMALIZE);
  glDisable(GL_BLEND);
  glPolygonMode(GL_FRONT,GL_FILL); 
  glEnable(GL_LIGHTING); glEnable(GL_LIGHT0); 
  glShadeModel(GL_SMOOTH);
  glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE); 
  glEnable(GL_COLOR_MATERIAL);
  glMaterialf(GL_FRONT,GL_SHININESS,50.0);
  glMaterialfv(GL_FRONT,GL_SPECULAR,spec);

  /* compute rotation for constant highlight spot... */ 

  glGetFloatv(GL_MODELVIEW_MATRIX,(float *)mvt);
  glGetLightfv(GL_LIGHT0,GL_POSITION,lp);
  for (s=0,i=0;i<3;i++) s+=lp[i]*lp[i]; for(i=0;i<3;i++) lp[i]/=sqrt(s); lp[2]+=0.8;
  for (i=0;i<3;i++) for (x[i]=0,j=0;j<3;j++) x[i]+=mvt[i][j]*lp[j];
  for (s=0,i=0;i<3;i++) s+=x[i]*x[i]; for(i=0;i<3;i++) x[i]/=sqrt(s);
  glPushMatrix(); glLoadIdentity();
  glRotatef(180*acos(x[2])/M_PI,-x[1],x[0],0);
  glGetFloatv(GL_MODELVIEW_MATRIX,(float *)LightRot);glPopMatrix();
}

void WnpRestoreAfterSolid(void)
{
  /* restore "solid" attributes to previous values */

  glPopAttrib();
}

void WnpDrawSphere(float *x,float r)
{
  /* draw sphere with radius r at position x */

  if (r>0) {
    switch (wnpSphereMethod) {
    case WNP_SPHERE_GLU:                /* use Quadric from GLU Library     */
      if (!WnpGluSphere) {
	WnpGluSphere=gluNewQuadric();
	gluQuadricDrawStyle(WnpGluSphere,wnpQuadricStyle[wnpSphereStyle]);
	gluQuadricOrientation(WnpGluSphere,GLU_OUTSIDE);
      }
      glPushMatrix(); 
      LightRot[3][0]=x[0];LightRot[3][1]=x[1];LightRot[3][2]=x[2];
      glMultMatrixf((float *)LightRot);
      gluSphere(WnpGluSphere,r,(2*wnpSphereQuality+4),wnpSphereQuality+2);
      glPopMatrix();
      break;
    case WNP_SPHERE_ICO:                /* use subdivided polyhedron method */
      glPushMatrix(); 
      LightRot[3][0]=x[0];LightRot[3][1]=x[1];LightRot[3][2]=x[2];
      glMultMatrixf((float *)LightRot);
      glScalef(r,r,r);
      icoSphere=MakeIcoSphere();
      if (icoSphere) glCallList(icoSphere);
      glPopMatrix();
    default:
      break;
    }
  }
}

void WnpDrawGlCylinder(float *x,float rx,float *y,float ry)
{
  float h,t[3];
  int i;

  if (!WnpGluCylinder) {
    WnpGluCylinder=gluNewQuadric();
    gluQuadricDrawStyle(WnpGluCylinder,wnpQuadricStyle[wnpSphereStyle]);
    gluQuadricOrientation(WnpGluCylinder,GLU_OUTSIDE);
  }
  for (h=0,i=0;i<3;i++) {
    t[i]=y[i]-x[i];
    h+=t[i]*t[i];
  }
  h=sqrt(h);
  if (h>0) for (i=0;i<3;i++) t[i]/=h;
  glPushMatrix(); 
  glTranslatef(x[0],x[1],x[2]);
  glRotatef(180*acos(t[2])/M_PI,-t[1],t[0],0);
  gluCylinder(WnpGluCylinder,rx,ry,h,wnpCylinderQuality,1);
  glPopMatrix();
}

void WnpSetSphereMode(int mode,int par)
{
  switch (mode) {
  case WNP_SPHERE_FULL:
    wnpHalfSpheres=FALSE;
    break;
  case WNP_SPHERE_HALF:
    wnpHalfSpheres=TRUE;
    break;
  case WNP_SPHERE_QUALITY:
    wnpSphereQuality=clamp(par,0,WNP_MAX_SPHERE_QUALITY);
    break;
  default:
    break;
  }
}

void WnpSetSphereMethod(int method)
{
  wnpSphereMethod=method;
}

void WnpSetCylinderMode(int mode,int par)
{
  switch (mode) {
  case WNP_CYLINDER_FULL:
    wnpHalfSpheres=FALSE;    /* note: half-spheres/half-cylinders are correlated */
    break;
  case WNP_CYLINDER_HALF:    /* note: half-spheres/half-cylinders are correlated */
    wnpHalfSpheres=TRUE;
    break;
  case WNP_CYLINDER_QUALITY:
    wnpCylinderQuality=clamp(par,3,WNP_MAX_CYLINDER_QUALITY);
    break;
  default:
    break;
  }
}

/*-------------------- local functions... --------------------*/

static GLuint MakeIcoSphere(void)
{
  GLuint s;
  int i,t,depth;

  if (icoSphere&&icoQuality!=clamp(wnpSphereQuality,0,WNP_MAX_iSPHERE_QUALITY-1)) {
    glDeleteLists(icoSphere,1);
    icoSphere=0;
  }
  if (!icoSphere) {
    s=glGenLists(1);
    icoQuality = clamp(wnpSphereQuality,0,WNP_MAX_iSPHERE_QUALITY-1);
    depth=wnpIcoDepth[icoQuality];
    glNewList(s,GL_COMPILE);
    glBegin(GL_TRIANGLES);
    switch (wnpIcoBase[icoQuality]) {
    case 'T':                           /* Tetrahedron */
      for (t=0,i=0;i<4;i++) 
	t+=TriangleDiv(depth,TV[TT[i][0]],TV[TT[i][1]],TV[TT[i][2]]);
      break;
    case 'O':                           /* Octahedron  */
      for (t=0,i=0;i<8;i++) 
	t+=TriangleDiv(depth,OV[OT[i][0]],OV[OT[i][1]],OV[OT[i][2]]);
      break;
    case 'I':                           /* Icosahedron */
      for (t=0,i=0;i<20;i++) 
	t+=TriangleDiv(depth,IV[IT[i][0]],IV[IT[i][1]],IV[IT[i][2]]);
      break;
    default:
      break;
    }
    glEnd();
    glEndList();
  }else s=icoSphere;
  return(s);
}

static int TriangleDiv(int depth,float *a,float *b,float *c)
{
  /* recursive subdivision of spherical triangle (a,b,c)       */
  /* assumes sphere radius of 1. Returns number of Triangles   */

  float v[3][3],s;
  int i,j,n;
  
  if (depth>0) {
    depth--; n=0;
    for (j=0;j<3;j++) {
      v[0][j]=a[j]+b[j];
      v[1][j]=b[j]+c[j];
      v[2][j]=c[j]+a[j];
    }
    for (i=0;i<3;i++) {
      for (s=0,j=0;j<3;j++) s+=v[i][j]*v[i][j];
      for (j=0;j<3;j++) v[i][j]/=sqrt(s);
    }
    n+=TriangleDiv(depth,v[0],v[1],v[2]);
    n+=TriangleDiv(depth,v[0],v[2],a);
    n+=TriangleDiv(depth,v[1],v[0],b);
    n+=TriangleDiv(depth,v[2],v[1],c);
  }else{
    glNormal3fv(a); glVertex3fv(a);
    glNormal3fv(b); glVertex3fv(b);
    glNormal3fv(c); glVertex3fv(c);
    n=1;
  }
  return(n);
}

#ifdef MYSPHERE
static GLuint MakeMySphere(void)  /* not used... */
{
  GLuint s;
  int i,j,m,n;
  float dphi,dpsi,v[3];

  if (icoSphere&&icoQuality!=clamp(wnpSphereQuality,0,WNP_MAX_iSPHERE_QUALITY-1)) {
    glDeleteLists(icoSphere,1);
    icoSphere=0;
  }
  if (!icoSphere) {
    s=glGenLists(1);
    icoQuality = clamp(wnpSphereQuality,3,1000);
    n=2*icoQuality+4; dphi=2*M_PI/n;
    m=(n+1)/2; dpsi=M_PI/m;
    glNewList(s,GL_COMPILE);
    for (j=0;j<m;j++) {
      glBegin(GL_QUAD_STRIP);
      for (i=0;i<=n;i++) {
	v[0]=sin((j+1)*dpsi)*cos(i*dphi); 
	v[1]=sin((j+1)*dpsi)*sin(i*dphi);
	v[2]=-cos((j+1)*dpsi);
	glNormal3fv(v); glVertex3fv(v);
	v[0]=sin(j*dpsi)*cos(i*dphi); 
	v[1]=sin(j*dpsi)*sin(i*dphi);
	v[2]=-cos(j*dpsi);
	glNormal3fv(v); glVertex3fv(v);
      }
      glEnd();
    }
    glEndList();
  }else s=icoSphere;
  return(s);
}
#endif
