/****************************************************************************/
/*   Conjugate Gradient, an attempt to implement the VA14 method in C       */
/*                                                                          */
/*   Reference:                                                             */
/*   M.J.D. Powell, "Restart procedures of the conjugate gradient method",  */ 
/*       Mathematical Programming 2(1977),241-254.                          */
/****************************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "conjgrad.h"

#define TRUE 1
#define FALSE 0

#define min(a,b) ( (a) <= (b) ? (a) : (b) )
#define max(a,b) ( (a) >= (b) ? (a) : (b) )

static int stop_pending=0;               /* not re-entrant, sorry... */

/****************************************************************************/

int conjgrad(
	     
   REAL (*func)(int *n,REAL *x,REAL *g,void *dat),/* the objective function */
   int   n,                              /* number of variables (in)        */
   REAL *x,                              /* n variables (in/out)            */
   REAL *f,                              /* min f (out)                     */
   REAL *g,                              /* gradient at optimum x (out)     */
   REAL  dpred,                          /* the infamous dpred (in,cf.below)*/
   REAL  rms,                            /* desired accuracy (in)           */
   int   maxfc,                          /* max. func evaluations (in)      */
   void *dat)                            /* pointer to data needed in func  */

/* Remarks:

   func:   objective function to be minimized. Note that the first arg is
           a  pointer, for easier use with functions written in FORTRAN.

   dpred:  the first trial step is a steepest descent step. The stepsize is
           chosen such that the expected decrease of the objective is dpred
           (based on the asumption that func is nearly quadratic).

   rms :   if rms(g[i]) <= rms, the procedure has "converged to desired
           accuracy".

   maxfc : if maxfc function evaluations are used before rms(g)<=rms,
           the method will return with a non-zero return status. If
           maxfc=0, the number of func calls is unlimited.

   dat   : pointer to additional data needed for function evaluations

   Return values:

    >=0  : success, optimum function value in f, optimum values of variables
           in x, gradient in g. Return value is iteration count.

     -1  : max number of f-calls exceeded.

     -2  : convergence failure (possible causes are error in definition of
           gradient or value of rms too small).

     -3  : insufficient memory for internal workspace.

     -4  : halted via call to conjgradstop()

*****************************************************************************/
{
  int   ncalls,accepted,i,nrest,mxrest,rstc,lok,niter;
  REAL *dr,*dgr,*d,*gs0,*xt,*gt;        /* pointers to "workspace" vectors  */
  REAL xnew,xold,xmin,xmax,step,dx;
  REAL fmin,f0,ft,ssqg,ssqg0,ginit,gprev,tolg,gmin,gm,gnew,*work_space,beta=0,gamma,gbstar=0;
  REAL ggprv=0;
  REAL c,clt=0.9;
  
  /* convert rms to tolerance on sum of squares of gradient */

  tolg=n*rms*rms;

  /* Get energy and gradient at starting point */

  *f=func(&n,x,g,dat);
  ncalls=1;
  for (ssqg=0,i=0;i<n;i++) {
    ssqg+=g[i]*g[i];
  }

  /* Return, if converged, or if maximum allowd function calls is reached. */

  if (ssqg<=tolg) return(0);                  /* already converged */
  if (ncalls>=maxfc) return(-1);              /* already at maxfc limit */

  /* Allocate dynamic memory for workspace */
  
  work_space=(REAL *)calloc(6*n,sizeof(REAL));
  if (work_space==NULL) return(-3);           /* didn't get what we asked for */
  dr=work_space ;
  dgr=work_space+n ;
  d=work_space+2*n;
  gs0=work_space+3*n;                         /* gradient at start of linesearch */
  xt=work_space+4*n;                          /* scratch space for (x,g) during line search */
  gt=work_space+5*n;
  
  /* Get current position and gradient */

  for (i=0;i<n;i++) {
    xt[i]=x[i];
    gt[i]=g[i];
  }
  ft=*f;
  
  /* Set up for first iteration */  

  gprev=-ssqg;                                 
  xnew=fabs(2*dpred)/ssqg;                    /* determine first step size based on dpred */
  rstc=n+1;                                   /* force steep.desc. restart */
  nrest=0;                                    /* number of restarts so far */
  mxrest=1;                                   /* max. number of restarts we are willing to do */
  niter=0;                                    /* iteration count, and exit status  */

  /* Enter main iteration, abort if exit status niter is < 0 */

  while (ssqg>tolg && niter>=0) {

    niter++;

    if (stop_pending) {                       /* stop if there is a pending request */
      niter=-4;
      break;
    }
 
    if (rstc>n) {                             /* reset search direction very n iterations  */
      for(i=0;i<n;i++) d[i]=-g[i];
    }

    for(ginit=0,i=0;i<n;i++) {
      gs0[i]=g[i];
      ginit+=d[i]*g[i];
    }

    if (ginit<0) {

      /* current search direction is downhill, do line search. */

      xmin=0; xmax=-1.0;                      /* bracket, no upper bound yet */
      fmin=f0=*f;
      ssqg0=ssqg;
      gmin=gm=ginit;
      xnew*=min(1.0,gprev/ginit);             /* our best guess, using information from prev. iteration */
      gprev=ginit;                            /* keep this for next iteration */
  
      for (lok=FALSE; !lok && niter>=0; ) {

	step=xnew-xmin;
	for (dx=0,i=0;i<n;i++) {              /* step to xmin */
	  xt[i]=x[i]+step*d[i];
	  dx+=fabs(xt[i]-x[i]);
	}

	if (dx>0) {
	  if (max(xmax,xmin-step)<0) clt=0.1;
	  ft=func(&n,xt,gt,dat); ncalls++; 
	  for (gnew=0,i=0;i<n;i++) gnew+=d[i]*gt[i];
	  accepted = (ft<fmin || (ft==fmin && fabs(gnew)<=fabs(gmin)));
	  if  (accepted) {
	    *f=ft;                            /* store it */
	    for (ssqg=0,i=0;i<n;i++) {
	      ssqg+=gt[i]*gt[i];
	      x[i]=xt[i];
	      g[i]=gt[i];
	    }
	    gm=gnew;
	    if (gm>ginit) {
	      for (gbstar=0,i=0;i<n;i++) gbstar+=gs0[i]*gt[i];
	      beta=(ssqg-gbstar)/(gm-ginit);
	    }
	  }
	  if (ncalls==maxfc) {                /* reached max. f-calls */
	    niter=-1;
	    break;
	  }
	}else{
	  if (nrest<mxrest) {                 /* dx too small, no progress, force a reset */
	    rstc=n+1; nrest++;
	  }else{
	    niter=-2;
	  }
	  break;
	}

	if ( gm>ginit && fabs(gm)<=clt*fabs(ginit) && fabs(gm*beta)<=clt*ssqg ) lok=TRUE;
	   
	else {
	  clt+=0.3;
	  if (clt>0.8) {
	    if (nrest<mxrest) {
	      rstc=n+1; nrest++;
	    }else{
	      niter=-2;
	    }
	    break;
	  }
	  xold=xnew;
	  if (!accepted || gmin*gnew<=0) xnew=0.5*(xmin+xold); /* change of direction        */
	  else if (xmax<0) xnew=10*xold;  /* continue, still don't have upper bound          */
	  else xnew=0.5*(xold+xmax);      /* continue to mid point of upper half  of bracket */

	  step=gnew-(3*gnew+gmin-4*(ft-fmin)/(xold-xmin))*((xold-xnew)/(xold-xmin));
	  if (accepted) {
	    if (gmin*gnew<=0) xmax=xmin;
	    xmin=xold;fmin=*f;gmin=gnew;
	  }else xmax=xold;     

	  if (step*gmin<0) xnew=(xmin*step-xnew*gmin)/(step-gmin);
	}
      }
  
      if (lok) {

	/* Linesearch converged */
  
	if (*f>=f0 && ssqg>=ssqg0) {
	  if (nrest<mxrest) {                     /* convergence fails */
	    rstc=n+1; nrest++;                    /* force reset */
	    continue;
	  }else{
	    niter=-2;                             /* signal failure */ 
	  }
	}
	if (rstc>=n || fabs(gbstar)>=0.2*ssqg) {
	  rstc=2;                                 /* restart every n iterations, or if d is "off" */
	  ggprv=gm-ginit;
	  for (i=0;i<n;i++) {
	    dr[i]=d[i];
	    dgr[i]=g[i]-gs0[i];
	    d[i]=-g[i]+beta*d[i];
	  }
	}else{
	  for (gamma=0,c=0,i=0;i<n;i++) {         /* normal update of search direction */
	    gamma+=g[i]*dgr[i];
	    c+=g[i]*dr[i];
	  }
	  gamma=gamma/ggprv;
	  if (fabs(beta*gm+gamma*c)>=0.2*ssqg) {
	    rstc=2;
	    ggprv=gm-ginit;
	    for (i=0;i<n;i++) {
	      dr[i]=d[i];
	      dgr[i]=g[i]-gs0[i];
	      d[i]=-g[i]+beta*d[i];}
	  }else{
	    rstc++;
	    for (i=0;i<n;i++) {
	      d[i]=-g[i]+beta*d[i]+gamma*dr[i];
	    }
	  }
	}
      }else if (ssqg<tolg) break;
 
    }else{
     
     /* current search direction is uphill */

     if (nrest<mxrest) {           /* force a restart, if not too many resets. */
	rstc=n+1; nrest++;
      }else{
	niter=-2;
      }
    }

  }

  if (work_space) free(work_space);

  return(niter);
}

/*==========================================================================*/

                     /* ----------------------------- */
                     /* FORTRAN interface to conjgrad */
                     /* ----------------------------- */

int conjgrad_(
	     
   REAL (*func)(int *n,REAL *x,REAL *g,void *dat),
   int  *n,
   REAL *x,
   REAL *f,
   REAL *g,
   REAL *dpred,
   REAL *rms,
   int  *maxfc)
{
  return(conjgrad(func,*n,x,f,g,*dpred,*rms,*maxfc,NULL));
}

void conjgradstop_(void)
{
  stop_pending++;
}

void conjgradclear_(void)
{
  stop_pending=0;
}

int conjgradtest_(void)
{
  return(stop_pending);
}
