#!/usr/bin/env python

_version = "1.0.0"
_usage="""wnprpc.py [options...] [command]

Description: Send 'command' to Wit!P XML/RPC server and write response text
to stdout. If the -p option is set, the Wit!P prompt string at the end of the
execution will be appended to the response. The exit status of the script will
be set to the number of error that occured in the Wit!P during the execution
of the command, or to 127 if the connection to the XML/RPC server could not be
established. The server (host:port) to connect to may be defined using the -s
option. If the -l option is set, commands will be read from stdin, rather than
from the command line, and executed line by line until the end of file is
reached. If the XML/RPC server uses a connection tag, the tag specified by the
-t option must match, or the connection will fail."""

import xmlrpclib, os, random
from subprocess import Popen, PIPE
from time import sleep
from types import *

#========================================================================

class wnpRPC(object):
  """
  Base class for wnpRPCserver and wnpRPCconnection
  """
  def getenv(self,name):
    """
    Returns value of environment varibale 'name' from server (string)
    """
    return(self.server.wnp.getenv(name))

  def getprompt(self):
    """
    Returns current prompt string.
    """
    return self.server.wnp.getprompt()

  def loadPDB(self,name,data=None):
    """
    Load molecule from a PDB datablock [string].
    Returns (status,prompt,response) [int, string, list of strings].
    status: error count during execution of command.
    prompt: command line prompt after execution of command.
    response: text output generated by command, list of LF terminated lines.
    """
    return(self.loadMolTypes(name=name,type="pdb",data=data))

  def loadMolTypes(self,name=None,type=None,data=None):
    """
    Load a molecule from a data block [string[.
    name: name to assign to loaded molecule.
    type: one of "mol","mol2","cif","pdb","xpdb","shelx","mmod","diana",
          "par","maccs","csd" (for "maccs" and "csd" only the first molecule
          will be loaded).
    Returns (status,prompt,response) [int, string, list of strings].
    status: error count during execution of command.
    prompt: command line prompt after execution of command.
    response: text output generated by command, list of LF terminated lines.
    """
    self.loadData(data)
    return(self.server.wnp.loadMOLtypes("%s:%s"%(type,name)))

  def cmd(self,command):
    """
    Sends Wit!P command(s) to server [string],
    Returns (status,prompt,response) [int, string, list of strings].
    status: error count during execution of command(s).
    prompt: command line prompt after execution of command(s).
    response: text output generated by command(s), list of LF terminated lines.
    """
    return(self.server.wnp.cmd(command))

  def loadData(self,data):
    """
    Load data (string) into server's datablock (used by loadMolTypes()).
    Returns number of characers loaded.
    """
    s=0
    if isinstance(data,str):
      slice=500000
      n=len(data)
      for i in range(0,len(data),slice): 
        s+=self.server.wnp.dataBlock(n,data[i:i+slice])
        n=0
    return(s)

  def freeData(self):
    """
    Free server's datablock.
    Returns number of characers loaded.
    """
    return(self.server.wnp.dataBlock(-1,""))

#------------------------------------------------------------------------

class wnpRPCconnection(wnpRPC):
  """

  The wnpRPCconnection class connects to a running Wit!P session through
  XML/RPC.
  """
  def __init__(self,host="localhost",port=None,tag=""):
    """
    Establish connection to Wit!P XML/RPC sever listening at http://host:port/RPC2.
    (host [string]: hostname, port [int]: TCP/IP port of XML/RPC server).
    If the tag [string] does not match the environment variable WNP_RPC_TAG in
    Wit!P's process environment, the connection will be terminated. 
    """
    self.server=connectRPCserver(host,port,tag)
    
#------------------------------------------------------------------------

class wnpRPCserver(wnpRPC):
  """

  The wnpRPCserver class creates a Wit!P session and connects to it through
  XML/RPC.
  """
  def __init__(self,graphics=False,ports=None,wait=5):
    """
    Start Wit!P on local host, with XML/RPC open on the first available port from
    the list of ports. If 'graphics' is True, the Wit!P session will start and
    connect to a graphics server. 'wait' [int] is the maximum time allowed for the
    XML/RPC server to respond to the initial RPC call.
    """
    self.server,self.pipe=startRPCserver(graphics=graphics,ports=ports,wait=wait)

  def __del__(self):
    """
    Destructor: will shut down the XML/RPC server. If Wit!P is not connected to
    a graphics server, this will cause Wit!P to be terminated.
    """
    if isinstance(self.server,xmlrpclib.ServerProxy):
      self.server.wnp.shutdown()
      self.server=None
    if isinstance(self.pipe,Popen):
      self.pipe.stdin.flush()
      self.pipe.stdin.close()
      self.pipe=None

#========================================================================

def queryRPCserver(pipe,server,tag,wait):
  """
  Check if (pipe,server) connects to a Wit!P XML/RPC server with the
  correct tag (value of WNP_RPC_TAG in Wit!P's process environment).
  """
  r=False
  a=None
  T=0
  for t in [0,0,0,0,0,0.1,0.1,0.1,0.1,0.1,0.5,0.5,0.5,0.5,0.5,1,1,1,1,1,1,1,1,1,1]:
    try:
      a=server.wnp.getenv("WNP_RPC_TAG")
      break
    except:
      if (pipe.poll()):
        break
      sleep(t)
      T=T+t
      if (T>wait):
        break

  if (not pipe.poll()):
    if type(a)== StringType and a==tag:
      r=True

  return(r)

def connectRPCserver(host,port,tag):
  """
  Connect to server port@host. Returns xmlrpclib.ServerProxy object, if
  connection could be established and Wit!P has the right tag ($WNP_RPC_TAG).
  """
  server=xmlrpclib.ServerProxy("http://%s:%d/RPC2"%(host,port))
  try:
    a=server.wnp.getenv("WNP_RPC_TAG")
    if type(a)== StringType and a==tag:
      r=server
    else:
      r=None
  except:
    r=None
  return(r)

def startRPCserver(ports,graphics=False,wait=None):
  """
  Description: Starts Wit!P with XML/RPC server open on the first
  available port. Wit!P will run with graphics only if the first
  argument ('graphics') is true. Returns 
  """
  if (wait is None):
    wait=5

  if isinstance(ports,list):
    plist=ports
  else:
    plist=[ports]

  tag="WNP/RPC %d/%d"%(os.getpid(),random.randint(1,1000000))
  os.environ["WNP_RPC_TAG"]=tag
  for port in plist:
    wnpcmd="witnotp rpc %d %s 1>/dev/null 2>&1"%(port,"" if graphics else "host none")
    pipe=Popen(wnpcmd, shell=True, stdin=PIPE)
    server=xmlrpclib.ServerProxy("http://localhost:%d/RPC2"%(port))
    if (queryRPCserver(pipe,server,tag,wait)):
      return server,pipe

  return None,None
  
#========================================================================

if __name__ == '__main__':

  import sys,os
  from optparse import OptionParser

  port=os.environ.get("WNP_RPC_PORT","19001")
  tag=os.environ.get("WNP_RPC_TAG","")

  optionParser=OptionParser(_usage,version='%prog '+_version)
  optionParser.add_option('-l','--loop',dest='loop',action='store_true',default=False,
                          help='read commands from stdin [False]')
  optionParser.add_option('-p','--prompt',dest='prompt',action='store_true',default=False,
                          help='show prompt [False]')
  optionParser.add_option('-s','--server',dest='server',default="localhost:%s"%port,
                          help='server [localhost:%s]'%port)
  optionParser.add_option('-t','--tag',dest='tag',default="",
                          help='tag [""]')

  options,args = optionParser.parse_args()

  if options.server.count(":")>0:
    k=options.server.find(":")
    host=options.server[0:k]
    port=options.server[k+1:]
  else:
    host="localhost"
    port=options.server
  try:
    port=int(port)
  except:
    print >>sys.stderr, "invalid server (%s), port must be an integer."%options.server
    sys.exit(127)

  W=wnpRPCconnection(host=host,port=port,tag=options.tag)

  if W.server is None:
    print >> sys.stderr,"failed to connect to Witnotp/RPC server."
    sys.exit(127)
  
  s=0
  if options.loop:
    if options.prompt: sys.stdout.write(W.getprompt())
    cmd=sys.stdin.readline()
    while cmd:
      try:
        s,p,r=W.cmd(cmd)
      except:
        print >> sys.stderr,"lost connection to server"
        break
      sys.stdout.write("".join(r))
      if options.prompt: sys.stdout.write(p)
      cmd=sys.stdin.readline()
    sys.stdout.write('\n')
  else:
    cmd=" ".join(args)
    s,p,r=W.cmd(cmd)
    sys.stdout.write("".join(r))
    if options.prompt: sys.stdout.write(p)

  sys.exit(s)
